//----- bcmAppl.c ----- 06/28/13 -----(03/05/06)----

//-----------------------------------------------------------------------------
//  Application-specific code run at a regular interval
//
//  Application code to monitor current flow on two separate circuits
//-----------------------------------------------------------------------------
//  Update log (most recent first):
//
//
//  2013-06-28 - Mark Davis
//               - Moved statusBitsStdA and statusBitsStdB to same ModBus
//                 register positions as all other applications.
//
//  2012-10-02 - Mark Davis (*)
//               - Added chanSettings[] which mirrors/supports just the
//                 writable bits of chanStatus[] and replaced chanStatus with
//                 chanSettings in the "settings" file.  Prior to this, changes
//                 to the read-only status bits in chanStatus would cause the
//                 settings cache (and eventually the file on flash) to be
//                 updated, which would result in extra unneeded writes to
//                 flash (which could use up the typical write cycles much
//                 faster).
//
//  2011-06-07 - Mark Davis
//               - Corrected value in table that caused the screen command to
//                 show channel 0's rangeOK flag value for channel 1.
//               - Fixed so the default front end calibrations are set
//                 properly at startup (so if there is no bench calibration
//                 data the defaults at least will be usable).
//
//  2011-05-23 - Mark Davis
//               - The BAD/OK indicator for the "Range" in the state screen
//                 now shows "BAD" in red.
//
//  2011-05-09 - Mark Davis
//               - Fixes for the BCMs wired to have 10x less sensitive
//                 ranges (i.e. to measure larger currents).
//
//  2011-05-06 - Mark Davis
//               - Moved bench calibrations to separate feCalibs file (it was
//                 in the devData file, which was useless as it meant the
//                 bench calibrations would not stay with the front end
//                 boards that they were calibrated to).
//               - Moved the runtime calibration data to the devData file,
//                 since they ARE for the external device and should be
//                 independent of which BCM is connected to it.
//               - Saved the devData file (which now contains the runtime
//                 calibration data) after successful completion of the
//                 runtime calibrations.
//
//  2011-05-04 - Mark Davis
//               - First release of DEBROS version of BCM application code
//
//  2011-04-05 - Mark Davis
//               - Began work on actually porting the BCM appl to DEBROS (this
//                 code was just a mostly-empty skeleton before then used
//                 originally to test RS-485 serial port use).
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2007-03-14 - Mark Davis -
//               - added call to rwIO() (replaces "invisible" code in the
//                 sharedAppl.c file that did this for all applications).
//
//  2007-03-04 - Mark Davis -
//               - reworked X-Macro usage to allow application-specific
//                 register definitions to reside in the same file as the
//                 rest of the application-specific code.
//
//  2006-10-10 - Mark Davis -
//               - updated this skelton using latest starterAppl.c code
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <sys/types.h>



#ifndef  ApplicSource
  #define  ApplicSource  "bcmAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "Beam Current Monitor" };
const char ApplAbbrev[] = { "BCM" };

//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_BCM_SS

//APPLCODE - set the # of input and outputs used
//  (must include any unused ones within ones that ARE used)
#define  NUM_APPL_AIN     2   // analog inputs
#define  NUM_APPL_AOUT    1   // analog outputs
#define  NUM_APPL_DIN     1   // digital inputs
#define  NUM_APPL_DOUT   16   // digital outputs

#define  ADC_MASK_APPL  0x00000003L
#define  DAC_MASK_APPL  0x00000001L

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( 0 )


#define ADC_FULL_SCALE  10.0  // nominal full scale value of ADC


#define SharedINC
#include <sharedAppl.c>


void resetChan(uint chan);


//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//=============== digitalIn bit usage ===============
#define  DI_00               0x000001  // 00: Not used
#define  DI_01               0x000002  // 01: Not used
#define  DI_02               0x000004  // 02: Not used
#define  DI_03               0x000008  // 03: Not used
#define  DI_04               0x000010  // 04: Not used
#define  DI_05               0x000020  // 05: Not used
#define  DI_06               0x000040  // 06: Not used
#define  DI_07               0x000080  // 07: Not used

// next 8 are read by ReadDeviceID() in the arch/hdw/BSP layer
#define  ADDR_0              0x000100  // 08: Addr Switch Bit 0
#define  ADDR_1              0x000200  // 09: Addr Switch Bit 1
#define  ADDR_2              0x000400  // 10: Addr Switch Bit 2
#define  ADDR_3              0x000800  // 11: Addr Switch Bit 3
#define  ADDR_4              0x001000  // 12: Addr Switch Bit 4
#define  ADDR_5              0x002000  // 13: Addr Switch Bit 5
#define  ADDR_6              0x004000  // 14: Addr Switch Bit 6
#define  ADDR_7              0x008000  // 15: Addr Switch Bit 7

#define  DI_16               0x010000L // 16: Not used
#define  DI_17               0x020000L // 17: Not used
#define  DI_18               0x040000L // 18: Not used
#define  DI_19               0x080000L // 19: Not used
#define  DI_20               0x100000L // 20: Not used
#define  DI_21               0x200000L // 21: Not used
#define  DI_22               0x400000L // 22: Not used
#define  NotRangeX10         0x800000L // 23: 0 ==> multiply FSCurrent's by 10

const uchar  LogicOrderDIn[NUM_APPL_DIN] = { 0 };


//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--    --1 state descr--
  NULL,                NULL
};


//=============== digitalOut bit usage ===============
#define  ActivityLED    0x000001  // 00: bit used for the activity LED

#define  CH1_FE_ON      0x000002  // 01: (K1) Front End (extern input) On
#define  CH1_GAIN_A     0x000004  // 02: (K2) Gain Ctl A
#define  CH1_GAIN_B     0x000008  // 03: (K3) Gain Ctl B
#define  CH1_GAIN_C     0x000010  // 04: (K4) Gain Ctl C
#define  CH1_GAIN_D     0x000020  // 05: (K5) Gain Ctl D
#define  CH1_FLTR_OFF   0x000040  // 06: (K6) Filter Off

#define  CH1_TEST_CUR   0x000080  // 07: (K7) enable test curr

#define  DO_08          0x000100  // 08:  unused

#define  CH2_TEST_CUR   0x000200  // 09: (K7) enable test curr

#define  CH2_FE_ON      0x000400  // 10: (K1) Front End (extern input) On
#define  CH2_GAIN_A     0x000800  // 11: (K2) Gain Ctl A
#define  CH2_GAIN_B     0x001000  // 12: (K3) Gain Ctl B
#define  CH2_GAIN_C     0x002000  // 13: (K4) Gain Ctl C
#define  CH2_GAIN_D     0x004000  // 14: (K5) Gain Ctl D
#define  CH2_FLTR_OFF   0x008000  // 15: (K6) Filter Off


const uchar  LogicOrderDOut[NUM_APPL_DOUT] = { 1,  2,  3,  4,  5,  6,  7,
                                              10, 11, 12, 13, 14, 15,  9,
                                               0, 8 };

#define  TurnOnActivityLED()    SetBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   ClearBits(Next.digitalOut, ActivityLED)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--      --1 state descr--
  NULL,                  NULL,                       // 00
  "CH1 FrontEnd OFF",    "CH1 FrontEnd ON",          // 01
  NULL,                  NULL,                       // 02
  NULL,                  NULL,                       // 03
  NULL,                  NULL,                       // 04
  NULL,                  NULL,                       // 05
  "CH1 Filter ON",       "CH1 Filter OFF",           // 06
  "CH1 Test Cur OFF",    "CH1 Test Cur ON",          // 07
  NULL,                  NULL,                       // 08
  "CH2 Test Cur OFF",    "CH2 Test Cur ON",          // 09
  "CH2 Front End OFF",   "CH2 Front End ON",         // 10
  NULL,                  NULL,                       // 11
  NULL,                  NULL,                       // 12
  NULL,                  NULL,                       // 13
  NULL,                  NULL,                       // 14
  "CH2 Filter ON",       "CH2 Filter OFF"            // 15
};




//=============== Analog Input usage ===============
//--- indexes for anaIn[] and readVal[] ---
#define  CH1_READ         0
#define  CH2_READ         1

const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0, 1 };


//=============== Analog Output usage ===============
#define  DUMMY_00        0x0001

const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0 };


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// values that can be different between prev/cur or cur/next during processing

#define  NumRanges  16   // number of current ranges
#define  NumChans   NumLogicalDevices


//--- Format of data in ApplCfg file created from older pre-DEBROS version ---
//--- which combines the benchtop calibration data with the runtime offset ---
//--- values which change all the time.  This increases the risk of ---
//--- corrupting the benchtop data so the newer version separates the ---
//--- the runtime and benchtop calibration data ---
typedef struct  {
  bool         rangeOK;        // range tested OK during BENCHTOP calibration
  bool         inputOn;        // state of Front-End when RTZeroOffset determined
  float        RTZeroOffset;   // Run-Time voltage offset
  float        BCAmpsPerVolt;  // Bench-Calibrated gain using known current
  float        BCZeroOffset;   // Bench-Calibrated offset from zero volts
} OldFmtCalibData;


typedef struct {
  bool         rangeOK;        // range tested OK during BENCHTOP calibration
  float        BCAmpsPerVolt;  // Bench-Calibrated gain using known current
  float        BCZeroOffset;   // Bench-Calibrated offset from zero volts
} BenchCalibData;

typedef struct {
  bool         inputOn;        // state of Front-End when RTZeroOffset determined
  float        RTZeroOffset;   // Run-Time voltage offset
} RunTimeCalibData;


#define  BC_DATA_SIZE  ( sizeof(BenchCalibData) )
#define  RT_DATA_SIZE  ( sizeof(RunTimeCalibData) )

BenchCalibData      benchCalib[NumChans][NumRanges];
RunTimeCalibData  runtimeCalib[NumChans][NumRanges];

RunTimeCalibData  far newCalibs[NumChans][NumRanges];

//
// Range-related values
//
#define LOW_RANGE_START                 8   // start of low-current ranges
#define FILTER_RANGE_START             10   // start of filter-always-on ranges
#define VOLTAGE_OVERRANGE_LIMIT      9.00   // in volts
#define INV_VOLT_LIMIT               0.11111111  // inverse of voltage limit
#define LIMIT_READY_OVERRANGE        1.05   // % of VOLTAGE_OVERRANGE_LIMIT
#define LIMIT_ERROR_OVERRANGE        1.10   // % of VOLTAGE_OVERRANGE_LIMIT
#define LIMIT_OVERRANGE_AUTODROP     0.75   // % of full-scale current
#define MAX_OFFSET_FFV               0.80   // max allowed % of VOLTAGE_OVERRANGE_LIMIT




typedef struct {
  float          FSCurrent;     // Full Scale Current
  float          FSVoltage;     // Full Scale Voltage (determines gain)
  char           RelayBits;     // what relays to set for this range
} RangeInfo;

#define  MinChgTime     500  // Min # ms to wait for chg to take effect
#define  MaxNoChgTime  3000  // Max # ms to wait for no improvement in rdg

//
// Range information
// Pre-initialized, ZWorld will place it in simulated ROM
//
// NOTE:  The hardware supports only 8 physical ranges, but the older
//        BCM's had 15 ranges, so the software ones where thrown in
//        to mimic the older devices
//
//  The FSVoltage values here are default values only
//
const RangeInfo RangeData[NumRanges] = {
/* FSCurrent, FSVoltage, RelayBits */
  {   1E-03,    7.558,     0x04  },  // RANGE 0    1mA       K3
  { 300E-06,    2.266,     0x04, },  // RANGE 1  300uA       K3

  { 100E-06,    9.000,     0x08, },  // RANGE 2  100uA       K4
  {  30E-06,    2.951,     0x08, },  // RANGE 3   30uA       K4

  {  10E-06,    7.306,     0x10, },  // RANGE 4   10uA       K5
  {   3E-06,    2.193,     0x10, },  // RANGE 5    3uA       K5

  {   1E-06,    7.964,     0x00, },  // RANGE 6    1uA   <none>
  { 300E-09,    2.388,     0x00, },  // RANGE 7  300nA   <none>

  { 100E-09,    7.931,     0x06, },  // RANGE 8  100nA   K2 + K3
  {  30E-09,    2.380,     0x06, },  // RANGE 9   30nA   K2 + K3

  {  10E-09,    8.647,     0x0A, },  // RANGE 10  10nA   K2 + K4
  {   3E-09,    2.595,     0x0A, },  // RANGE 11   3nA   K2 + K4

  {   1E-09,    7.290,     0x12, },  // RANGE 12   1nA   K2 + K5
  { 300E-12,    2.819,     0x12, },  // RANGE 13 300pA   K2 + K5

  { 100E-12,    7.940,     0x02, },  // RANGE 14 100pA   K2
  {  30E-12,    2.384,     0x02, }   // RANGE 15  30pA   K2
};

//--- The real full-scale currents.  Defaults to values in the RangeData ---
//--- table, but a jumper can be used to multiply those values by 10 ---
float fsCurrent[NumRanges];

//
// Averaging table (entries are in milliseconds)
//
#define  NumAvgIdxs  8
const unsigned short AvgSelTable[NumAvgIdxs] =
{
  100,   // 0 (000)
   10,   // 1 (001)
   20,   // 2 (010)
   50,   // 3 (011)
  200,   // 4 (100)  default
  300,   // 5 (101)
  500,   // 6 (110)
 1000    // 7 (111)
};

#define  DefaultAvgIdx  4


//---------------------------------------------------------------------

// Each value in the ApplState structure should be throught of as representing
// a "signal" in a clocked logic circuit, with the "applic" task representing
// the logic implemented by the circuit elements such as gates, latches, etc.

// The main "applic" task is the highest priority task and executes the code
// that implements the application logic on a real-time, periodic basis.

//                 ====== IMPORTANT NOTE on LOCKING ======
// Remember to use a semaphore or some other form of locking mechanism to
// protect all access to shared values by more than one task.  The only
// exception to this is when the applic task needs to access a value protected
// by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
// macros).  Because the applic task is the only task that runs at the highest
// priority, no other task can preempt it, and if the scheduler is locked,
// the applic task cannot preempt another task.  But scheduler locks must be
// kept short to avoid creating large latency in the timely execution of other
// tasks (the applic task in particular).


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


#define NUM_EXTRA 7

typedef struct {
  bool       calibState;           // state of runtime Zero calibration

  bool       saveFrontEndOn;       // frontend on/off state before calib
  bool       saveFilterOn;         // filter on/off state before calib
  uint       saveRange;            // range state before calib
  uint       saveAvgingIdx;        // avgingIdx state before calib

  bool       clearOffsets;         // got request to clear the runtime offsets
  bool       doInit;               // got request to initialize a channel

  char       overRange;            // over-range state (OVERRANGE_xxx)
  bool       hwError;              // hardware error state (true or false)

  ulong      settlingStart;        // ms count for calibration settling
  float      variance;             // % variation in recent readings
  float      lowestVariance;       // for monitoring progress

  float      avgRaw;               // averaged RAW ADC value (with frac portion)
  float      avgCur;               // calc & offseted beam current
  float      avgVolts;             // average converted to volts
  float      adjVolts;             // adj volts: avgVolts less noise offset
  float      fracFSVolts;          // Fraction of Full-scale Voltage
  float      fracFSCur;            // Fraction of Full-scale Current

  bool       rangeOK;              // Usable range calibration values
  bool       testCurOn;            // Connect Test Current to Front End
  bool       frontEndOn;           // Connect Front End to external input
  bool       autoRanged;           // (not currently implemented)
  bool       filterOn;             // Enable Front End filter (*)
  uint       range;                // set range # (0-15)
  uint       avgingIdx;            // index into table of # period to avg over

  long       extra1;               // fixed-point value used by extern clients
  long       extra2;               // fixed-point value used by extern clients
  long       extra3;               // fixed-point value used by extern clients
  long       extra4;               // fixed-point value used by extern clients
  long       extra5;               // fixed-point value used by extern clients
  long       extra6;               // fixed-point value used by extern clients
  long       extra7;               // fixed-point value used by extern clients
} ChanState;

// *NOTE:  The Front End filter is forced on for range >= FILTER_RANGE_START
//         regardless of the value of filterOn


//--- values for ChanState.calibState ---
#define  CALIB_READY         0   // Not doing a run-time calibration
#define  CALIB_START         1   // Run-time calibr requested, not started
#define  CALIB_START_RANGE   2   //
#define  CALIB_WAIT_SETTLE   3   //
#define  CALIB_QUIT         99   //

//--- values for ChanState.overRange ---
#define OVERRANGE_OK         0   // not overrange
#define OVERRANGE_READY      1   // overrange, but still valid
#define OVERRANGE_ERROR      2   // overrange, reading likely bad




typedef struct {
  ChanState  chanState[NumChans];

 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------
//  bool   resetRequest;    // reset Interlocks request

 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
 //--- values updated in "cur" instance at start of each pass ---
  ulong  processTime;     // time (in ms) when processing was started
  ulong  digitalIn;       // state of digital inputs

 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.  These values
 //  must  NOT be altered by other tasks (although they can read them as long
 //  as they use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------
//  bool   ovenModeOn;           // Turn on power, enable output soon after
//  bool   ovenOutErrInterlock;  // Oven PS Output state was wrong
//  ulong  ovenOutErrStartTime;  // start of period of Output state error


 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (i.e. it acts like a
 //  clocked latch but is in software and the clock is one pass through
 //  the main application loop).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task AFTER all the other ApplState values are set.  Because they represent
 //  only the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  digitalOut;      // state for digital outputs

} ApplState;



//bool  localMode, closedLoop;

ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );


//
// Structure for storing ADC calibration values
//
typedef struct {
  float  volts;   // voltage reading at time we were told the amps value
  float  amps;    // amperage that corresponds to recorded volts value
} RangeCalibData;

RangeCalibData  far calibPoints[NumChans][NumRanges][2];  // Range calibration point data




//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//APPLCODE - add tasks that service network clients

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC        PRI  -1  config params
//  ------------  -------------  --------  ----------- ---  --  ---------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,         10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,    5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,   6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,         10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);



//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/

#ifndef NumLogicalDevices
  #define  NumLogicalDevices  2
#endif

//-----------------------------------------------------------------------------
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x,  1, setCoil )
COIL0( DoInit,         _,  2, MB_DoInit )
COIL0( KillTelnet,     x,  3, setCoil )  // not applicable
COIL0( AbortCalib,     _,  4, MB_AbortCalib )

COIL0( Unused_05,      x,  5, setCoil )
COIL0( Unused_06,      x,  6, setCoil )
COIL0( Unused_07,      x,  7, setCoil )
COIL0( Unused_08,      x,  8, setCoil )
COIL0( Unused_09,      x,  9, setCoil )
COIL0( Unused_10,      x, 10, setCoil )
COIL0( Unused_11,      x, 11, setCoil )
COIL0( Unused_12,      x, 12, setCoil )
COIL0( Unused_13,      x, 13, setCoil )
COIL0( Unused_14,      x, 14, setCoil )
COIL0( Unused_15,      x, 15, setCoil )
COIL0( Unused_16,      x, 16, setCoil )
COIL0( Unused_17,      x, 17, setCoil )
COIL0( Unused_18,      x, 18, setCoil )
COIL0( Unused_19,      x, 19, setCoil )
COIL0( Unused_20,      x, 20, setCoil )
COIL0( Unused_21,      x, 21, setCoil )
COIL0( Unused_22,      x, 22, setCoil )
COIL0( Unused_23,      x, 23, setCoil )

COIL0( RangeUp,        _, 24, setCoil )  // no longer used?
COIL0( RangeDown,      _, 25, setCoil )  // no longer used?

COIL0( AvgIdxUp,       _, 26, setCoil )  // no longer used?
COIL0( AvgIdxDown,     _, 27, setCoil )  // no longer used?

COIL0( FrontEndOn,     _, 28, setCoil )  // no longer used?
COIL0( FrontEndOff,    _, 29, setCoil )  // no longer used?

COIL0( FilterOn,       _, 30, setCoil )  // no longer used?
COIL0( FilterOff,      _, 31, setCoil )  // no longer used?

COIL0( ClearOffsets,   _, 32, MB_ClearOffsets )
COIL0( DoCalibrate,    _, 33, MB_DoCalibrate )

COIL0( TestCurOn,      _, 34, setCoil )  // no longer used?
COIL0( TestCurOff,     _, 35, setCoil )  // no longer used?

COIL0( SaveCalibs,     _, 36, MB_SaveCalibs )

#undef COIL0
#endif


//-----------------------------------------------------------------------------
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

// standard for older pre-DEBROS applications
REG3( versDate,       x, 30001, U32 )  // older decimal yyyymmdd format
REG3( devType_,       x, 30003, U16 )  // was devTyp pre-DEBROS
REG3( upSecs,         x, 30004, U32 )  // was secCounter pre-DEBROS
REG3( dlmVersn,       x, 30006, U16 )  // closest equiv is not-implemented dlmVersTime

REG3( ctlRate,        x, 30007, U16 )
REG3( comRate,        x, 30008, U16 )
REG3( ttlErrs,        x, 30009, U16 )  // no longer applicable
REG3( digInBits,      x, 30010, U32 )
REG3( _digOutBits,    x, 30012, U32 )  // read-only copy
REG3( rateErrs,       x, 30014, U16 )  // no longer applicable
REG3( serialErrs,     x, 30015, U16 )  // no longer applicable
REG3( _timeZone,      x, 30016, S16 )  // read-only copy
REG3( deviceID_,      x, 30017, U32 )
//REG3( blk3_dateTime,  x, 30019, U32 )  // Doesn't apply to DEBROS applications
REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )

REG3( vin00,          x, 30021, F32 )
REG3( vin01,          x, 30023, F32 )
REG3( vin02,          x, 30025, F32 )
REG3( vin03,          x, 30027, F32 )
REG3( vin04,          x, 30029, F32 )
REG3( vin05,          x, 30031, F32 )
REG3( vin06,          x, 30033, F32 )
REG3( vin07,          x, 30035, F32 )

REG3( vout00_,        x, 30037, F32 )  // read-only copy
REG3( vout01_,        x, 30039, F32 )  // read-only copy
REG3( vout02_,        x, 30041, F32 )  // read-only copy
REG3( vout03_,        x, 30043, F32 )  // read-only copy

// standard DEBROS values with no exact equiv in pre-DEBROS set
REG3( dlmVersTime,    x, 30045, U32 )
REG3( rtosVersTime,   x, 30047, U32 )
REG3( applVersTime,   x, 30049, U32 )
REG3( percentIdle_,   x, 30051, U16 )

REG3( numChans,       x, 30052, U16 )
REG3( coilChanMult,   x, 30053, U16 )
REG3( reg3ChanMult,   x, 30054, U16 )
REG3( reg4ChanMult,   x, 30055, U16 )

REG3( unused_30056,   x, 30056, U16 )
REG3( unused_30057,   x, 30057, U16 )
REG3( statusBitsAppl, x, 30058, U32 )

REG3( latency,        x, 30060, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30061, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30062, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime, x, 30063, U16 )  // # ms used by last kernel pass

REG3( avgVolts,       _, 30064, F32 )
REG3( fracFSCur,      _, 30066, F32 )

#undef REG3
#endif

#ifdef RegMacros
//===== application-specific defintions related to register values =====

  #define  VOLTSIN(idx)  REG_ARRAY(float, vin00, idx)

#endif


//-----------------------------------------------------------------------------
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

#ifdef REG4

//--- BCM appl-specific register values ---
REG4( amps,           _, 40001, F32, RO, roReg)

REG4( rangeIdx,       _, 40003, U16, RW, MB_rangeIdx )
REG4( fullScale,      _, 40004, F32, RO, roReg )  // upper end of range

REG4( avgingSize,     _, 40006, U16, RO, roReg )  // # elements in averaging buffer
REG4( avgingIdx,      _, 40007, U16, RW, MB_avgingIdx )  // # index into list of avg buffer sizes

REG4( chanStatus,     _, 40008, U16, RW, MB_chanStatus )  // See CS_xxx values

REG4( calibPoint1,    _, 40009, F32, RW, MB_calibPoint1 )
REG4( calibPoint2,    _, 40011, F32, RW, MB_calibPoint2 )

REG4( extra1,         _, 40013, S32, RO, roReg )  // long int "debug" data used by Test appl
REG4( extra2,         _, 40015, S32, RO, roReg )
REG4( extra3,         _, 40017, S32, RO, roReg )
REG4( extra4,         _, 40019, S32, RO, roReg )
REG4( extra5,         _, 40021, S32, RO, roReg )
REG4( extra6,         _, 40023, S32, RO, roReg )
REG4( extra7,         _, 40025, S32, RO, roReg )

REG4( variance,       _, 40027, F32, RO, roReg )
REG4( volts,          _, 40029, F32, RO, roReg )
REG4( avgInterval,    _, 40031, U16, RO, roReg )  // # ms over which rdgs are averaged

//--- standard DEBROS registers ---
REG4( timeZone_,      x, 40032, S16, RW, MB_timeZone )
REG4( curTime,        x, 40033, U32, RW, setU32 )

REG4( ipAddr,         x, 40035, U32, RW, setU32 )
REG4( netmask,        x, 40037, U32, RW, setU32 )
REG4( gateway,        x, 40039, U32, RW, setU32 )

REG4( setACoil,       _, 40041, U16, RW, MB_setACoil )
REG4( resetACoil,     _, 40042, U16, RW, MB_resetACoil )

REG4( digOutBits,     x, 40043, U32, RW, setU32 )

REG4( cmdBitsStd,     x, 40045, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,    _, 40047, U32, RW, setU32 )

REG4( vout00,         x, 40049, F32, RW, MB_setVout )
REG4( vout01,         x, 40051, F32, RW, MB_setVout )
REG4( vout02,         x, 40053, F32, RW, MB_setVout )
REG4( vout03,         x, 40055, F32, RW, MB_setVout )
REG4( vout04,         x, 40057, F32, RW, MB_setVout )
REG4( vout05,         x, 40059, F32, RW, MB_setVout )
REG4( vout06,         x, 40061, F32, RW, MB_setVout )
REG4( vout07,         x, 40063, F32, RW, MB_setVout )
REG4( vout08,         x, 40065, F32, RW, MB_setVout )
REG4( vout09,         x, 40067, F32, RW, MB_setVout )
REG4( vout10,         x, 40069, F32, RW, MB_setVout )
REG4( vout11,         x, 40071, F32, RW, MB_setVout )
REG4( vout12,         x, 40073, F32, RW, MB_setVout )
REG4( vout13,         x, 40075, F32, RW, MB_setVout )
REG4( vout14,         x, 40077, F32, RW, MB_setVout )
REG4( vout15,         x, 40079, F32, RW, MB_setVout )

REG4( chanSettings,   _, 40081, U16, RW, MB_chanStatus )  // only writable chanStatus bits

#undef REG4
#endif

#ifdef RegMacros
//===== application-specific defintions related to register values =====

  #define  VOLTSOUT(idx)   REG_ARRAY(float, vout00, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion


//
// chanStatus[] bit usage
//
#define CS_FILTER      0x0001  // RW - FrontEnd filter On?
#define CS_RDY_NORM    0x0002  // RO - Reading below upper limit
#define CS_RDY_OVER    0x0004  // RO - Reading near upper limit
#define CS_INCAL       0x0008  // RO - Calibr field level (runttime) 0-offsets
#define CS_ERR_OVER    0x0010  // RO - Reading above upper limit
#define CS_HW_ERR      0x0020  // RO - Problem with hardware
#define CS_TESTCURR    0x0040  // RW - Enable/disable test current
#define CS_FRONTEND    0x0080  // RW - Connect to external current source
#define CS_RANGEBAD    0x0100  // RO - Range not usable
#define CS_AUTORANGED  0x0200  // RW - Autoranging (not currently implemented)
#define CS_AVG_B0      0x0400  // RO - Bit 0 of averaging table index
#define CS_AVG_B1      0x0800  // RO - Bit 1 of averaging table index
#define CS_AVG_B2      0x1000  // RO - Bit 2 of averaging table index
#define CS_RESET       0x2000  // RW - chan needs INIT cmd

#define CS_RW_BITS     0x3EC1  // bits that can be chgd externally


//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================
//APPLCODE - settings/values we want saved to/restored from files


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------

RESTORE_FUNC(resetChans);


const ParamData far  devDataParams[] = {
//APPLDATA
//                ID   ---value---  --restoredFunc--
//PERSISTENT_VAR(  1,  myVarName,     NULL ),
//                ID   ---value---  granularity  --restoredFunc--
//PERSISTENT_TBL(  2   myArrayName, elemSize,     NULL ),

//finish:  Need to handle reloading the calibration data

  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//----------------------------------------------------

const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --restoreFunc--
//  PERSISTENT_VAR(  1,  ampSet,      PVRF(ampSet) ),
//  PERSISTENT_VAR(  2,  rfOnSetpoint,    NULL ),

  PERSISTENT_TBL(  1,  runtimeCalib,    RT_DATA_SIZE,     PVRF(resetChans) ),
  PERSISTENT_TBL(  2,  chanSettings,    sizeof(chanSettings),    NULL ),
  PERSISTENT_TBL(  3,  rangeIdx,        sizeof(rangeIdx),        NULL ),
  PERSISTENT_TBL(  4,  avgingIdx,       sizeof(avgingIdx),       NULL ),

  END_PARAM_DATA
};


//----------------------------------------------------

const ParamData far  feCalibsParams[] = {
//APPLDATA
//                ID   ---value---  --restoreFunc--
  PERSISTENT_TBL(  1,  benchCalib,    BC_DATA_SIZE,  PVRF(resetChans) ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --restoreFunc--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };

ParamFile feCalibsFile =
    { "feCalibs",       PDC_SBC,  feCalibsParams,0 };



#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile,
  &feCalibsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Declarations for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

//void onRequest(void);
//void offRequest(void);
//void resetRequest(void);

//void newSetpoint(float newAmps);


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//APPLCODE
//--- add definitions for application-specific global variables here ---



//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//APPLCODE - add functions to process writes to ModBus register values


//-----------------------------------------------------------------------------
//  Got a request to initialize the state of one of the input channels
//-----------------------------------------------------------------------------
bool _nearcall MB_DoInit(SetCoilFuncParams)
{

  ChanState  *csNext = Next.chanState + chanNum;

  KLOCK();

  if (csNext->calibState > CALIB_START)  csNext->calibState = CALIB_QUIT;

  csNext->doInit = true;

  KUNLOCK();  return true;
}


//-----------------------------------------------------------------------------
//  Got a request to abort a runtime zero calibration
//-----------------------------------------------------------------------------
bool _nearcall MB_AbortCalib(SetCoilFuncParams)
{

  ChanState  *csNext = Next.chanState + chanNum;

  KLOCK();
  if (csNext->calibState > CALIB_START)  csNext->calibState = CALIB_QUIT;
  KUNLOCK();

  return true;
}


//-----------------------------------------------------------------------------
//  Got a new value for the ClearOffsets bit register
//-----------------------------------------------------------------------------
bool _nearcall MB_ClearOffsets(SetCoilFuncParams)
{

  ChanState  *csNext = Next.chanState + chanNum;

  if (!newVal)  return true;

  KLOCK();

  if (csNext->calibState)  { KUNLOCK();  return false; }

  csNext->clearOffsets = true;

  KUNLOCK();  return true;
}


//-----------------------------------------------------------------------------
//  Got a new value for the DoCalibrate bit register
//-----------------------------------------------------------------------------
bool _nearcall MB_DoCalibrate(SetCoilFuncParams)
{

  ChanState  *csNext = Next.chanState + chanNum;

  if (!newVal)  return true;

  KLOCK();

  if (csNext->calibState)  { KUNLOCK();  return false; }

  csNext->calibState = CALIB_START;

  KUNLOCK();  return true;
}


//-----------------------------------------------------------------------------
//  Got a new value for the SaveCalibs bit register
//-----------------------------------------------------------------------------
bool _nearcall MB_SaveCalibs(SetCoilFuncParams)
{
  ChanState  *csNext = Next.chanState;
  char  buf[256];


  if (csNext[0].calibState OR csNext[1].calibState)  return false;

  fprintf(DiagLog, "Save FrontEnd Calibrations\r\n");

  saveParams(buf, sizeof(buf), &feCalibsFile, ForceFlush);
  saveParams(buf, sizeof(buf), &devDataFile, ForceFlush);

  return true;
}


//-----------------------------------------------------------------------------
//  An external source told us the 1st of 2 actual amperage readings for the
//  current range and specified channel.
//-----------------------------------------------------------------------------
bool _nearcall MB_calibPoint1(SetRegFuncParams)
{
  ChanState  *csCur;
  RangeCalibData  far *rcd;


  csCur = Cur.chanState + chanNum;
  rcd = calibPoints[chanNum][csCur->range];

  rcd[0].volts = getAvgdAnaIn(chanNum);
  rcd[0].amps = MB_VAL(F32);

  return true;
}

//-----------------------------------------------------------------------------
//  An external source told us the 2nd of 2 actual amperage readings for the
//  current range and specified channel.
//
//  The assumption is that this followed a recent write to the calibPoint1
//  value for another calibration data point for the same range and channel,
//  and that we should use these 2 values to compute a new benchtop gain and
//  offset for the current range.
//-----------------------------------------------------------------------------
bool _nearcall MB_calibPoint2(SetRegFuncParams)
{
  bool  rangeOK;
  uint  range;
  float  deltaA, deltaV, ampsPerVolt, zeroOffset;
  BenchCalibData  *bc;
  RunTimeCalibData  *rc;
  ChanState  *csCur;
  RangeCalibData  far *rcd;
  char  buf[64];


  csCur = Cur.chanState + chanNum;

  if (csCur->calibState)  return false;

  range = csCur->range;  rcd = calibPoints[chanNum][range];

  rcd[1].volts = getAvgdAnaIn(chanNum);
  rcd[1].amps = MB_VAL(F32);

  deltaA = rcd[1].amps - rcd[0].amps;
  deltaV = rcd[1].volts - rcd[0].volts;

  ampsPerVolt = deltaA / deltaV;
  zeroOffset = rcd[0].volts - rcd[0].amps/ampsPerVolt;

  rangeOK = true;
  if (FABS(deltaV) < 8.0)  rangeOK = false;
  if (FABS(zeroOffset) > 5.0)  rangeOK = false;

  fprintf(DiagLog, "Amps/Volt: %s   zeroOffset: %s Volts   %s\r\n",
          Engineering(ampsPerVolt, 4, buf),
          Engineering(zeroOffset, 4, buf+16),
          rangeOK ? "OK" : "BAD"
         );

  bc = benchCalib[chanNum] + range;
  rc = runtimeCalib[chanNum] + range;

  KLOCK();

  rc->RTZeroOffset = 0.0;
  bc->rangeOK = rangeOK;
  bc->BCZeroOffset = zeroOffset;
  bc->BCAmpsPerVolt = ampsPerVolt;


  if (! (range & 1))  {  // save for software range (all odd #'s) as well
    ++bc;  ++rc;  ++range;
    rc->RTZeroOffset = 0.0;
    bc->rangeOK = rangeOK;
    bc->BCZeroOffset = zeroOffset;
    bc->BCAmpsPerVolt = ampsPerVolt;
  }

  resetChan(chanNum);

  KUNLOCK();

  return true;
}

//-----------------------------------------------------------------------------
//  Received a new rangeIdx value for a channel
//-----------------------------------------------------------------------------
bool _nearcall MB_rangeIdx(SetRegFuncParams)
{
  ChanState *csNext = Next.chanState + chanNum;
  U16 newIdx = MB_VAL(U16);


  KLOCK();

  if ((newIdx >= NumRanges) OR (csNext->calibState))  {
    KUNLOCK();  return false; }

  csNext->range = newIdx;

  KUNLOCK();  return true;
}


//-----------------------------------------------------------------------------
//  Received a new avgingIdx value
//-----------------------------------------------------------------------------
bool _nearcall MB_avgingIdx(SetRegFuncParams)
{
  ChanState *csNext = Next.chanState + chanNum;
  U16 newIdx = MB_VAL(U16);


  KLOCK();

  if ((newIdx >= NumAvgIdxs) OR (csNext->calibState))  {
    KUNLOCK();  return false; }

  csNext->avgingIdx = newIdx;

  KUNLOCK();  return true;
}


//-----------------------------------------------------------------------------
//  Received a new chanStatus value
//-----------------------------------------------------------------------------
bool _nearcall MB_chanStatus(SetRegFuncParams)
{
  ChanState  *csNext = Next.chanState + chanNum;
  U16  newBits = MB_VAL(U16) & CS_RW_BITS;


  KLOCK();

  if (!(chanSettings[chanNum] ^ newBits))  { KUNLOCK();  return true; }

  if (csNext->calibState)  { KUNLOCK();  return false; }

  csNext->filterOn = BOOL(newBits & CS_FILTER);
  csNext->testCurOn = BOOL(newBits & CS_TESTCURR);
  csNext->frontEndOn = BOOL(newBits & CS_FRONTEND);
//  csNext->autoRanged = BOOL(newBits & CS_AUTORANGED);

  chanStatus[chanNum] = (chanStatus[chanNum] & ~CS_RW_BITS) | newBits;
  chanSettings[chanNum] = newBits;


  KUNLOCK();  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
/*
  ulong  cmdBits;
  EventData  eventMsg;  // used by EVENT macro


  cmdBits = *((ulong *)newVal);

  if (cmdBits & x)  EVENT(MyEventX, chan, newVal, sizeof(int)); // w/ data
  if (cmdBits & y)  EVENT(MyEVentY, chan, NULL, 0);             // w/o data
*/

  return true;
}

//-----------------------------------------------------------------------------
//  routine to set one of the voltage outputs.
//-----------------------------------------------------------------------------
bool _nearcall MB_setVout(SetRegFuncParams)
{
  int  vidx;


  vidx = (REG->regNum - vout00_RegInfo.regNum) / 2;

  if ((vidx < 0) OR (vidx > NumHdwDAC))  return false;

  anaOut[(uint)vidx] = *((float *)newVal);

  setF32(info, chanNum, newVal, eventsfd);  return true;
}



//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

/*
//-----------------------------------------------------------------------------
//  one of the setpoint values needs restoring from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(myVal)
{
  uint  u;
  float  fval;
  char  temp[32];

  for (u=0; u<NumPS; ++u)  if (curVal == (void far *) &ISET(u))  {
    fval = *(float far *)newVal;
    printf("restoring iSet_%u to: %s\r\n", u, fmtFloat(fval, 8,2, temp));
    newSetpoint(u, fval);
  }

}

//-----------------------------------------------------------------------------
//  restore an ampSet value from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ampSet)
{
  newSetpoint(NEW_VAL(float));
}


*/

//-----------------------------------------------------------------------------
//  Reloaded some values that requires reinitializing data for both channels
//-----------------------------------------------------------------------------
RESTORE_FUNC(resetChans)
{
  KLOCK();  fmemcpy(curVal, newVal, count);  KUNLOCK();

  resetChan(0);  resetChan(1);
}

//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>


//=============================================================================
//  Misc Application-specific functions
//=============================================================================
//APPLCODE


//-----------------------------------------------------------------------------
//  Show the BCM Front End calibration data for the specified input channel
//-----------------------------------------------------------------------------
void showFECalibs(uint chan)
{
  uint  range;
  BenchCalibData  *bc;
  RunTimeCalibData  *rc;
  char  buf[32];


  printf("BCM Front End Calibrations: Channel %d\r\n"
         "     --- Run-time ---     ------ Benchtop ------  \r\n"
         "     Offset   w/FE On?    Offset   Gain(A/V)  OK? \r\n"
         "    --------  -------    --------  ---------  --- \r\n", chan );
//        ##.  -#.###V    ###       -#.###V -###.#E-##  ###

  for (range=0; range<NumRanges; range+=2)  {
    bc = benchCalib[chan] + range;
    rc = runtimeCalib[chan] + range;

    printf("%2u.  %sV    %s       %sV %s  %s\r\n",
           range,
           fmtFloat(rc->RTZeroOffset, 6, 3, buf),
           (rc->inputOn ? "YES" : "NO "),
           fmtFloat(bc->BCZeroOffset, 6, 3, buf+8),
           Engineering(bc->BCAmpsPerVolt, 4, buf+16),
           (bc->rangeOK ? "YES" : "NO ")
          );
  }
  printf("\r\n");

}

//-----------------------------------------------------------------------------
//  Show the BCM calibration data for each front end
//-----------------------------------------------------------------------------
void _task feCalib(int argc, char **argv)
{
  uint  chan;

  printf("\r\n");
  for (chan=0; chan<NumChans; ++chan)  showFECalibs(chan);

  Exit(0);
}


//-----------------------------------------------------------------------------
//  Read the benchtop and runtime calibration values from an ApplCfg file.
//
//  Pre-DEBROS code (wirtten using DynamicC) stored persistent data in a
//  reserved portion of flash that (then) ZWorld called the 'User Block'.
//  The DynamicC-based NSCL code used linked lists of blocks of data within
//  the "User Block" to store persistent data (basically a crude file system).
//
//  The DEBROS btof command locates these lists of blocks and converts each
//  of the standard ones to named files that DEBROS code can open and read
//  using stio functions like open(), read(), etc.
//
//  One of the blocks is the "ApplData" block that contains configuration
//  and/or calibration data in an application-specific format.  This command
//  will open the ApplData file (if there is one) and read the benchtop and
//  runtime calibration data from it and store them in the newer format used
//  by the DEBROS version of the application.
//-----------------------------------------------------------------------------
void _task convApplCfg(int argc, char **argv)
{
   int  applfd, chan, range;
   uint  dataVers, dataDevType;
   OldFmtCalibData  ofcd;
   BenchCalibData  *bc;
   RunTimeCalibData  *rc;


   printf("\r\n");

   if ((applfd = open("ApplCfg", O_RDONLY, 0)) < 0)  {
     printf("Unable to open ApplCfg file\r\n");  Exit(-1);  }

   if (read(applfd, (char *)&dataVers, sizeof(uint)) != sizeof(uint))  {
     perror("reading vers");  Exit(-2);  }

   if (read(applfd, (char *)&dataDevType, sizeof(uint)) != sizeof(uint))  {
     perror("reading devType");  Exit(-3);  }

   printf("ApplCfg vers: %u,  devType: %u\r\n", dataVers, dataDevType);

   bc = benchCalib[0];  rc = runtimeCalib[0];
   for (chan=0; chan<2; ++chan)  {  // always 2 chans in older data
     for (range=0; range<16; ++range)  {  // always 16 ranges in older data
       if (read(applfd, (void *)(&ofcd), sizeof(ofcd)) != sizeof(ofcd))  {
         perror("reading ApplCfg");  Exit(-4);  }
       bc->BCAmpsPerVolt = ofcd.BCAmpsPerVolt;
       bc->BCZeroOffset = ofcd.BCZeroOffset;
       bc->rangeOK = ofcd.rangeOK;
       rc->inputOn = ofcd.inputOn;
       rc->RTZeroOffset = ofcd.RTZeroOffset;
       ++bc;  ++rc;
     }
     showFECalibs(chan);
   }

   close(applfd);  Exit(0);
}


//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
//APPLCODE - text and formatting info for application state summary

/*EXAMPLE:
//-----------------------------------------------------------------------------
_nearcall char *fmtMode(DispElem far *elem, char *buf)
{
  return (char *) ModeDesc[ELEM_VAL(U16)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtHdwCfg(DispElem far *elem, char *buf)
{
  return (char *) HdwCfgDesc[ELEM_VAL(U16)];
}
*/

//-----------------------------------------------------------------------------
_nearcall char *fmtRangeFS(DispElem far *elem, char *buf)
{
  uint range = ELEM_VAL(U16);

  return Engineering(fsCurrent[range], elem->prec, buf);
}

//-----------------------------------------------------------------------------
_nearcall char *fmtPercentFS(DispElem far *elem, char *buf)
{
  uint range = ELEM_VAL(U16), chan = elem->width;
  float  percent;

  percent = amps[chan] * 100.0 / fsCurrent[range];
  percent = FABS(percent);

  strcpy(buf, (percent > 95.0) ? RedText : BlueText);

  fmtFloat(percent, 6,2, buf+strlen(buf));

  return buf;
}

//-----------------------------------------------------------------------------
_nearcall char *fmtRangeOK(DispElem far *elem, char *buf)
{
  uint range = ELEM_VAL(U16), chan = elem->width;

  return (benchCalib[chan][range].rangeOK ? (BlueText"OK ") : (RedText"BAD"));
}

//-----------------------------------------------------------------------------
_nearcall char *fmtState(DispElem far *elem, char *buf)
{
  uint chan = elem->width;

  return Next.chanState[chan].calibState ?
    (RedText "CALIB") : (BlueText "RUN  ");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtAvging(DispElem far *elem, char *buf)
{
  uint  idx = ELEM_VAL(U16), chan = elem->width;

//  sprintf(buf, "[%u] %3u samples (%4u ms)",
//          idx, getAvgSamplesADC(chan), getAvgPeriodADC(chan));
  sprintf(buf, "[%u] (%4u ms)",
          idx, getAvgPeriodADC(chan));

  return buf;
}

//-----------------------------------------------------------------------------
_nearcall char *fmtSource(DispElem far *elem, char *buf)
{
  uint  stat = ELEM_VAL(U16) & (CS_FRONTEND | CS_TESTCURR);

  switch (stat) {
    case 0:           return RedText  "NONE       ";
    case CS_FRONTEND: return BlueText "EXTERN     ";
    case CS_TESTCURR: return RedText  "TEST       ";
    default:          return RedText  "EXTERN+TEST";
  }
}

//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"------------ Chan 0 ------------        ------------ Chan 1 ------------ \r\n"
"Range:    ## (-##.##E-##) xxx           Range:    ## (-##.##E-##) xxx \r\n"
"FrontEnd: ...                           FrontEnd: ... \r\n"
"Filter:   ..                            Filter:   .. \r\n"
"State:    .....                         State:    ..... \r\n"
"Voltage: ##.###     (###.##%%var)        Voltage: ##.###     (###.##%%var) \r\n"
"Current: -#.###E-## (###.##%%FS)         Current: -#.###E-## (###.##%%FS) \r\n"
"Avging:   [#] (#### ms)                 Avging:   [#] (#### ms) \r\n"
"Source:   ........                      Source:   ........ \r\n"
"\r\n"
};


const DispElem  far dispElems[] = {
//{ xx,yy, VALUE,                  ATTR,  FmtFunc    width,prec },
  { 10, 4, VFPV(rangeIdx[0]),      BLUE,  fmtU16,       2,0 },
  { 14, 4, VFPV(rangeIdx[0]),      BLUE,  fmtRangeFS,   0,4 },
  { 26, 4, VFPV(rangeIdx[0]),      BLUE,  fmtRangeOK,   0,0 },
  { 10, 7, VFPV(rangeIdx[0]),      "",    fmtState,     0,0 },
  {  9, 8, VFPV(anaIn[CH1_READ]),  BLUE,  fmtF32,       6,3 },
  { 21, 8, VFPV(variance[0]),      BLUE,  fmtF32,       6,2 },
  {  9, 9, VFPV(amps[0]),          BLUE,  fmtEng,       4,0 },
  { 21, 9, VFPV(rangeIdx[0]),      "",    fmtPercentFS, 0,0 },
  { 10,10, VFPV(avgingIdx[0]),     BLUE,  fmtAvging,    0,0 },
  { 10,11, VFPV(chanStatus[0]),    "",    fmtSource,    1,0 },

  { 50, 4, VFPV(rangeIdx[1]),      BLUE,  fmtU16,       2,0 },
  { 54, 4, VFPV(rangeIdx[1]),      BLUE,  fmtRangeFS,   0,4 },
  { 66, 4, VFPV(rangeIdx[1]),      BLUE,  fmtRangeOK,   1,0 },
  { 50, 7, VFPV(rangeIdx[1]),      "",    fmtState,     1,1 },
  { 49, 8, VFPV(anaIn[CH2_READ]),  BLUE,  fmtF32,       6,3 },
  { 61, 8, VFPV(variance[1]),      BLUE,  fmtF32,       6,2 },
  { 49, 9, VFPV(amps[1]),          BLUE,  fmtEng,       4,0 },
  { 61, 9, VFPV(rangeIdx[1]),      "",    fmtPercentFS, 1,1 },
  { 50,10, VFPV(avgingIdx[1]),     BLUE,  fmtAvging,    1,1 },
  { 50,11, VFPV(chanStatus[1]),    "",    fmtSource,    1,0 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};

//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
/*EXAMPLE:
//{ xx,yy,  BIT_MASK,     "0STR"  ,ATTR_NAME,  "1STR"  ,ATTR_NAME },
  { 43, 3,  CLOSED_LOOP,  "OPEN  ",DARK_RED,   "CLOSED",DARK_BLUE },
  { 43, 4,  RF_IS_OFF,    "ON "   ,DARK_BLUE,  "OFF"   ,DARK_RED },
  { 43, 5,  LOC_REM_SW,   "REMOTE",DARK_BLUE,  "LOCAL ",DARK_BLUE },
*/
  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
//{ xx,yy,  BIT_MASK,     "0STR"  ,ATTR_NAME,  "1STR" ,ATTR_NAME },
  { 10, 5,  CH1_FE_ON,    "OFF"  ,RED,         "ON "  ,BLUE },
  { 10, 6,  CH1_FLTR_OFF, "ON "  ,BLUE,        "OFF"  ,RED  },

  { 50, 5,  CH2_FE_ON,    "OFF"  ,RED,         "ON "  ,BLUE },
  { 50, 6,  CH2_FLTR_OFF, "ON "  ,BLUE,        "OFF"  ,RED  },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
//  int  xoffset;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

//    for (chan=0; chan<NumChans; ++chan)  {
//      xoffset = chan * 40;
//    }

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/




//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "cur" OR the "next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "cur" value for the "input", and the "prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "cur" values
//    represent the input values just read (what they are "now") and the
//    "prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "cur" values).
//
//    During a pass the output values in "cur" represent what the outputs
//    where set to AFTER the "cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessive latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

/*Examples
//-----------------------------------------------------------------------------
void onRequest(void)
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON request\n");
}

//-----------------------------------------------------------------------------
void offRequest(void)
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF request\n");
}

//-----------------------------------------------------------------------------
void resetRequest(void)
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET request\n");
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps)
{
  char  temp[16];


  KLOCK();

  if (FABS(newAmps) > maxAmpSet)  newAmps = (float)SIGN(newAmps) * maxAmpSet;

  Next.ampSet = newAmps;

  KUNLOCK();

  fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}

//-----------------------------------------------------------------------------
*/

//-----------------------------------------------------------------------------
//  Reset all the values related to the processing of readings for the
//  specified channel.
//
//  This function must be called whenever the range or the calibration data
//  for a channel is changed.
//-----------------------------------------------------------------------------
void resetChan(uint chan)
{
  uint  range;
  float  defGain, gainFrac;
  BenchCalibData  *bc;
  RunTimeCalibData  *rc;
  const RangeInfo  *ri;
  ChanState  *csNext = Next.chanState + chan;


  range = csNext->range;

  ri = RangeData + range;
  bc = benchCalib[chan] + range;
  rc = runtimeCalib[chan] + range;

  readOffset[chan] = bc->BCZeroOffset + rc->RTZeroOffset;
  readGain[chan] = bc->BCAmpsPerVolt;
  newADCCalib(chan);

  setAvgPeriodADC(chan, AvgSelTable[csNext->avgingIdx]);
  resetAvgDataADC(chan);

  resetNoiseDataADC(chan);  csNext->lowestVariance = 999.0;

  csNext->extra1 = range;
  csNext->extra2 = (long) (FABS(rc->RTZeroOffset) * INV_VOLT_LIMIT * 1000);
  csNext->extra5 = (long) (bc->BCAmpsPerVolt * 10000000000000.0);
  csNext->extra6 = (long) (bc->BCZeroOffset * 1000000.0);
  csNext->extra7 = (long) (rc->RTZeroOffset * 1000000.0);

//--- If the benchtop calibrations differ greatly from the defaults ---
//--- then mark the range as bad ---
  defGain = fsCurrent[range] / ri->FSVoltage;
  gainFrac = readGain[chan] / defGain;  gainFrac = FABS(gainFrac);

  csNext->rangeOK = bc->rangeOK;
  if ((gainFrac < 0.50) OR (gainFrac > 2.0))  csNext->rangeOK = false;
  if (readOffset[chan] > ri->FSVoltage / 2.0)  csNext->rangeOK = false;
}

//-----------------------------------------------------------------------------
//  Update the runtime calibration state of the specified channel.
//
//  The runtime calibration process determines the RTZeroOffset for each range
//  of a channel by testing each range and waiting for the reading to settle.
//  Once the reading has settled, it records the averaged voltage reading and
//  includes it in the offset applied to the readings for future readings from
//  that channel.
//-----------------------------------------------------------------------------
void runtimeCalibTick(uint chan)
{
  uint  range;
  float  offsetA, offsetV, offsetFracFSV;
  BenchCalibData  *bc;
  RunTimeCalibData  *rc;
  RunTimeCalibData  far *rcNew;
  ChanState  *csCur, *csNext;
  char  temp[256];


  csCur = Cur.chanState + chan;
  csNext = Next.chanState + chan;

  switch (csCur->calibState)  {

    case CALIB_START:
      csNext->saveRange = csNext->range;
      csNext->saveFilterOn = csNext->filterOn;
      csNext->saveAvgingIdx = csNext->avgingIdx;

  //      csNext->range = 0;  csNext->filterOn = false;
      csNext->range = csNext->range & 0xfe;  // ensure even-#'d range
      csNext->avgingIdx = DefaultAvgIdx;
      csNext->testCurOn = false;

      printf("------ Starting calibration of channel %u ------\r\n", chan);

      csNext->calibState = CALIB_START_RANGE;
      break;


    case CALIB_START_RANGE:
      resetChan(chan);  csNext->settlingStart = GetMS();
      csNext->calibState = CALIB_WAIT_SETTLE;
      break;


    case CALIB_WAIT_SETTLE:
      if ((uint)msSince(csNext->settlingStart) < getNoisePeriodADC(chan) * 2)
        break;

    //--- keep waiting as long as there is improvement ---
      if (csNext->variance < csNext->lowestVariance)  {
        csNext->lowestVariance = csNext->variance;
        csNext->settlingStart = MS_TIMER;  break;
      }

    //--- wait at least 2 seconds from last improvement ---
      if (msSince(csNext->settlingStart) < 2000)  break;

    //--- reading did settle - update the run-time (RAM) copy for both ---
    //--- the hardware and software ranges (and save the on/off state of ---
    //--- the FrontEnd with it)---
      rcNew = newCalibs[chan] + csNext->range;
      bc = benchCalib[chan] + csNext->range;

      offsetV = csNext->avgVolts - bc->BCZeroOffset;
      offsetA = offsetV * bc->BCAmpsPerVolt;
      offsetFracFSV = FABS(offsetV) * INV_VOLT_LIMIT;

      rcNew->RTZeroOffset = offsetV;
      rcNew->inputOn = csNext->frontEndOn;

      rcNew[1] = rcNew[0];

      printf("%2u/%u: Var:%s%%  Offset:%sV/%sA, (%s%%FS)\r\n",
              csNext->range, chan,
              fmtFloat(csNext->variance, 4,2, temp),
              fmtFloat(offsetV, 6,3, temp+8),
              Engineering(offsetA, 4, temp+16),
              fmtFloat(offsetFracFSV * 100.0, 6,2, temp+32)
            );

    //--- move on to next range (or end calibration if no more) ---
      if ((csNext->range += 2) > 15)  csNext->range = 0;
      if (csNext->range != (csNext->saveRange & 0xfe))  {
        csNext->calibState = CALIB_START_RANGE;  break; }

    //--- All done - save the results and restore the range to channel's ----
    //--- settings to what they were before we started ---
      rcNew = newCalibs[chan];  rc = runtimeCalib[chan];
      for (range=0; range<NumRanges; ++range,++rcNew,++rc)  *rc = *rcNew;

      saveParams(temp, sizeof(temp), &devDataFile, ForceFlush);

      csNext->calibState = CALIB_QUIT;
      break;


    case CALIB_QUIT:
      csNext->range = csNext->saveRange;
      csNext->filterOn = csNext->saveFilterOn;
      csNext->avgingIdx = csNext->saveAvgingIdx;
      csNext->avgCur = 0.0;
      resetChan(chan);
      csNext->calibState = 0;
      break;
  }

}


/*
=========== NOTES on Locking ============

The applic task runs at the highest priority (even above the kernel), so it
cannot be preempted by any other task.  As long as this is true, the applic
task does not need to use locks when accessing global variables.

However, all other tasks MUST disable the scheduler when accessing values
used by the applic task to avoid being preempted while doing so.  Care must
be taken to keep these times short to avoid latency issues.

For REALLY short times, a critical section can be used as this is much more
efficient than disabling and reenabling the scheduler.  But critical sections
disable all interrupts, so it is not practical in most cases.

*/

//=============================================================================
//  Update the "Cur"rent state of the controller
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).  These values
//  are copied to Next as needed and constitute the Hardware Inputs.
//
//  Other tasks can affect the state of the control logic by changing values
//  in Next as well.  These values constitute Software Inputs.
//
//  This function is responsible for the Asynchronous logic that is applied
//  each pass.  It updates values in Next (or other global variables) based
//  on the Software and Hardware Input values in Next.  This portion of the
//  logic is analogous to unclocked gates in an electronic circuit.
//
//  After the Asychronous logic is applied, we set Prev = Cur and Cur = Next.
//  The result is that Prev contains the controller state as it was during the
//  previous pass, and Cur indicates the current state of the controller.
//
//  Also, see the "NOTES on Locking" section above....
//=============================================================================
void updateCurState(void)
{
  uint  chan;
  ChanState  *csNext, *csCur;


  if (current != applTask)  return;  // must only be called by "applic" task


//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The analog and digital hardware inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "Cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "Next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

  Next.processTime = GetMS();  // curTime = time(NULL);

//  msSincePrev = msSince(Cur.processTime);

  Next.digitalIn = digitalIn;

//APPLCODE

  for (chan=0; chan<NumChans; ++chan)  {
    csNext = Next.chanState + chan;
    csCur = Cur.chanState + chan;

    csNext->avgRaw = getAvgdRawIn(chan);
    csNext->avgVolts = ADCtoAIn(chan, csNext->avgRaw);
    csNext->avgCur = AInToRead(chan, csNext->avgVolts);

  //--- Update the variation (noise) value (computed as a percent of the ---
  //--- nominal full scale value the ADC can read) ---
    csNext->variance =
      getNoiseADC(chan, NULL) * FABS(ADCGain(chan)) * (100.0 / ADC_FULL_SCALE);

  //--- Update the fraction of full-scale voltage limit value ---
    csNext->fracFSVolts = FABS(csNext->avgVolts) * INV_VOLT_LIMIT;

  //--- Update the fraction of full-scale current value ---
    csNext->fracFSCur = csNext->avgCur / fsCurrent[csNext->range];
    csNext->fracFSCur = FABS(csNext->fracFSCur);

  //--- Set overrange state based of fraction of fullscale voltage ---
    if (csNext->fracFSVolts < LIMIT_READY_OVERRANGE)
      csNext->overRange = OVERRANGE_OK;
    else if (csNext->fracFSVolts < LIMIT_ERROR_OVERRANGE)
      csNext->overRange = OVERRANGE_READY;
    else
      csNext->overRange = OVERRANGE_ERROR;

  //--- check for a change in the range ---
    if (csNext->range != csCur->range)  resetChan(chan);

  //--- check for a change in the averaging period ---
    if (csNext->avgingIdx != csCur->avgingIdx)
      setAvgPeriodADC(chan, AvgSelTable[csNext->avgingIdx]);

  //--- some fixed-point values still used by external clients ---
    csNext->extra3 = (long) csNext->avgRaw;
    csNext->extra4 = (long) (FABS(csNext->avgVolts) * 1000.0);
  }




// Next we make a copy of the application state before we apply all the
// logic that updates the overall state.  This allows us to have a copy of
// all the relevant values that remain unchanged during the processing,
// avoiding all kinds of order-of-execution dependencies.

// We also make a copy of the state from last time before overwriting it
// so we can see if/how values have changed since the last pass

// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Use of pointers would also result in larger, slower code for every
// reference to Prev, Cur, and Next.  Overall, copying everything here is
// actually faster overall, simpler, and much safer.

  Prev = Cur;  Cur = Next;


//--- log msgs for bit chgs only if not starving other processes for cycles ---
  if (applLoopTime < SchedMS-3)  logDigIOChgs(false);

//APPLCODE


}


//=============================================================================
//  Update the values in "Next" based on the "Cur" and "Prev" values.
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  This function contains the Synchronous logic used to compute the next
//  state we want the controller to be in.
//
//  A few simple rules that apply to all the values used by this function:
//
//    - All values changed within this function are "outputs"
//
//    - Every value read by this function is an "input"
//
//    - There must be no overlap between inputs and outputs
//
//    - The output values from this function must not be altered outside of
//      this function.
//
//  Any global variable can be an input OR an output, but they must NEVER
//  serve both purposes.  Most inputs come from Prev or Cur, while most
//  outputs are in Next.  If the logic requires detecting a change in a
//  value between passes (edge-triggered logic), then the value should be
//  part of the ApplState structure so we can compare its Prev and Cur values.
//
//  Following these simple rules allows the code to be organized as small,
//  self-contained blocks, with well-defined inputs and outputs.  Furthermore,
//  there are no order-of-execution dependencies between blocks of code.  The
//  blocks can be executed in any order:  The result will always be the same.
//
//
//  Each output value (or VERY small "set" of output values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative, data-oriented
//  style of programming and has many advantages including:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result.  Far simpler than debugging code structured around the
//       traditional process/event-driven approach which can contain numerous
//       independent paths of execution leading to an observed result.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The name(s) of the value(s) it modifies
//   - A descr of the meaning of the value(s) and the logic used
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  uint  chan, r;
  ChanState  *csCur, *csNext;


  if (current != applTask)  return;  // must only be called by "applic" task


  for (chan=0; chan<NumChans; ++chan)  {
    csCur = Cur.chanState + chan;
    csNext = Next.chanState + chan;

  //--------------------------------------------------------------------
  // csNext->calibState (and related csNext values)
  //    Description:
  //       If the calibState is non-0, call runtimeCalibTick() so it
  //       the run-time zero calibration process can progress.
  //    Inputs:
  //       csCur->calibState
  //----------------------------------------------------------------------
    if (csCur->calibState)  runtimeCalibTick(chan);


  //--------------------------------------------------------------------
  // csNext->clearOffsets
  //    Description:
  //       Clears the run-time zero-level offsets if requested to.
  //    Inputs:
  //       csCur->clearOffsets
  //----------------------------------------------------------------------
    if (csCur->clearOffsets)  {
      for (r=0; r<NumRanges; ++r)  runtimeCalib[chan][r].RTZeroOffset = 0.0;
      csNext->clearOffsets = false;
      printf("Run-time 0-level offsets cleared for chan %u\r\n", chan);
    }

  //--------------------------------------------------------------------
  // csNext->doInit
  //    Description:
  //       initializes the state of the channel
  //    Inputs:
  //       csCur->doInit
  //----------------------------------------------------------------------
    if (csCur->doInit)  {
      csNext->rangeOK = true;
      csNext->testCurOn = false;
      csNext->frontEndOn = true;
      csNext->autoRanged = false;
      csNext->filterOn = false;
      csNext->range = 0;
      csNext->avgingIdx = DefaultAvgIdx;
      csNext->doInit = false;
      resetChan(chan);  printf("Channel %u reinitialized\r\n", chan);
    }

  }


}



//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
  ChanState  *csNext;

  if (current != applTask)  return;  // must only be called by "applic" task

//--- update output values that are based on current conditions (values ---
//--- in "cur" or any values used in a conditional in the applic logic) ---

/*EXAMPLE:
  // (insured earlier that only ONE of these can be true when setting them)
  if (initialBtnPressed)  setVal[FP_METER] = rfOffSetpoint;  else
  if (finalBtnPressed)    setVal[FP_METER] = rfOnSetpoint;  else
                          setVal[FP_METER] = cavityVolts;


  SetOffApplLEDs(Next.digitalOut);

  if (limited)  SetOnLimitLED(Next.digitalOut);
  if (regErr)   SetOnErrorLED(Next.digitalOut);
*/

//--- update output values based on values that were set by the application ---
//--- logic (e.g. values in "next" or variables that are the target of an ---
//--- assignment statement in the application logic) ---

//APPLCODE

  Next.digitalOut = CH1_FLTR_OFF | CH2_FLTR_OFF;

  csNext = Next.chanState;

  if (csNext->frontEndOn)  SetBits(Next.digitalOut, CH1_FE_ON);
  if ((csNext->filterOn) OR (csNext->range >= FILTER_RANGE_START))
                           ClearBits(Next.digitalOut, CH1_FLTR_OFF);
  if (csNext->testCurOn)   SetBits(Next.digitalOut, CH1_TEST_CUR);
  SetBits(Next.digitalOut, RangeData[csNext->range].RelayBits << 1);

  ++csNext;

  if (csNext->frontEndOn)  SetBits(Next.digitalOut, CH2_FE_ON);
  if ((csNext->filterOn) OR (csNext->range >= FILTER_RANGE_START))
                           ClearBits(Next.digitalOut, CH2_FLTR_OFF);
  if (csNext->testCurOn)   SetBits(Next.digitalOut, CH2_TEST_CUR);
  SetBits(Next.digitalOut, RangeData[csNext->range].RelayBits << 10);

}




//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  bool  filterOn[NumChans];
  U16  status;
  uint  chan, range;
//  BenchCalibData  *bc;
//  RunTimeCalibData  *rc;
  ChanState  *csNext;


  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  _digOutBits = digOutBits = Cur.digitalOut;

  comRate = ctlRate = ctlLoopTicksPerSec;

  _timeZone = timeZone;

//APPLCODE

  vin00 = anaIn[0];
  vin01 = anaIn[1];
  vin02 = anaIn[2];
  vin03 = anaIn[3];
  vin04 = anaIn[4];
  vin05 = anaIn[5];
  vin06 = anaIn[6];
  vin07 = anaIn[7];

  vout00_ = anaOut[0];
  vout01_ = anaOut[1];
  vout02_ = anaOut[2];
  vout03_ = anaOut[3];


  filterOn[0] = !(Next.digitalOut & CH1_FLTR_OFF);
  filterOn[1] = !(Next.digitalOut & CH2_FLTR_OFF);

  for (chan=0; chan<NumChans; ++chan)  {
    csNext = Next.chanState + chan;

    range = csNext->range;

//    bc = benchCalib[chan] + range;
//    rc = runtimeCalib[chan] + range;

    amps[chan] = csNext->avgCur;

    rangeIdx[chan] = range;

    fullScale[chan] = fsCurrent[range];
    avgingIdx[chan] = csNext->avgingIdx;
    avgingSize[chan] = getAvgSamplesADC(chan);
    avgInterval[chan] = getAvgPeriodADC(chan);


    status = 0;

    if (filterOn[chan])  status |= CS_FILTER;

    switch (csNext->overRange) {
      case OVERRANGE_OK:       status |= CS_RDY_NORM;  break;
      case OVERRANGE_READY:    status |= CS_RDY_OVER;  break;
      case OVERRANGE_ERROR:    status |= CS_ERR_OVER;  break;
    }

    if (csNext->calibState)  status |= CS_INCAL;

    if (csNext->hwError)  status |= CS_HW_ERR;

    if (csNext->testCurOn)  status |= CS_TESTCURR;
    if (csNext->frontEndOn)  status |= CS_FRONTEND;

    if (!csNext->rangeOK)  status |= CS_RANGEBAD;

    if (csNext->avgingIdx & 0x01)  status |= CS_AVG_B0;
    if (csNext->avgingIdx & 0x02)  status |= CS_AVG_B1;
    if (csNext->avgingIdx & 0x04)  status |= CS_AVG_B2;

    chanStatus[chan] = status;
    chanSettings[chan] = status & CS_RW_BITS;

    variance[chan] = csNext->variance;

    volts[chan] = anaIn[chan];
    avgVolts[chan] = csNext->avgVolts;
    fracFSCur[chan] = csNext->fracFSCur;

    extra1[chan] = csNext->extra1;
    extra2[chan] = csNext->extra2;
    extra3[chan] = csNext->extra3;
    extra4[chan] = csNext->extra4;
    extra5[chan] = csNext->extra5;
    extra6[chan] = csNext->extra6;
    extra7[chan] = csNext->extra7;
  }


}


//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0; }

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
  fstrcpy(ainLabel[0], "CH0 Read");
  fstrcpy(ainLabel[1], "CH1 Read");

//  fstrcpy(aoutLabel[0], "ao0-Descr");

//  fstrcpy(dinLabel[0], "di0-Descr");
//  fstrcpy(dinLabel[1], "di1-Descr");

  fstrcpy(doutLabel[0], "Activity LED");

  fstrcpy(doutLabel[1], "(K1) CH0 Front End ON");
  fstrcpy(doutLabel[2], "(K2) CH0 Gain Ctl A");
  fstrcpy(doutLabel[3], "(K3) CH0 Gain Ctl B");
  fstrcpy(doutLabel[4], "(K4) CH0 Gain Ctl C");
  fstrcpy(doutLabel[5], "(K5) CH0 Gain Ctl D");
  fstrcpy(doutLabel[6], "(K6) CH0 Filter OFF");
  fstrcpy(doutLabel[7], "(K7) CH0 Test Cur ON");

  fstrcpy(doutLabel[9],  "(K7) CH1 Test Cur ON");
  fstrcpy(doutLabel[10], "(K1) CH1 Front End ON");
  fstrcpy(doutLabel[11], "(K2) CH1 Gain Ctl A");
  fstrcpy(doutLabel[12], "(K3) CH1 Gain Ctl B");
  fstrcpy(doutLabel[13], "(K4) CH1 Gain Ctl C");
  fstrcpy(doutLabel[14], "(K5) CH1 Gain Ctl D");
  fstrcpy(doutLabel[15], "(K6) CH1 Filter OFF");
}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  uint  range, chan;
  BenchCalibData  *bc;
//  RunTimeCalibData  *rc;
  const RangeInfo  *ri;
  ChanState  *csNext = Next.chanState;


  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

  BZERO(runtimeCalib);

  for (range=0; range<NumRanges; range+=2)  {
    ri = RangeData + range;
    for (chan=0; chan<NumChans; ++chan)  {
      bc = benchCalib[chan] + range;

      bc->rangeOK = true;
      bc->BCAmpsPerVolt = fsCurrent[range] / ri->FSVoltage;
      bc->BCZeroOffset = 0.0;
    }
  }


  if (! prevLogInfo.didSoftReset)
    csNext[0].avgingIdx = csNext[1].avgingIdx = DefaultAvgIdx;

  resetChan(0);  resetChan(1);
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

//APPLCODE


  Exit(0);
}


//=============================================================================
//  Use the values in prevLogInfo to determine if we should assert interlocks
//  for any devices any devices that were shut off due to a non-soft reset.
//=============================================================================
void checkPrevOnStates(void)
{

}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
//APPLCODE
  int  range;
  struct tm  tms;
  float  rangeMult;


  #define SharedApplLocals2
  #include <sharedAppl.c>


  allowSoftReset = true;

  localtime_r(&applVersTime, &tms);
  versDate = (tms.tm_year+1900)*10000L + (tms.tm_mon+1)*100 + tms.tm_mday;

  digitalIn = ReadDigInputs();
  rangeMult = (digitalIn & NotRangeX10) ? 1.0 : 10.0;
  for (range=0; range<NumRanges; ++range)
    fsCurrent[range] = RangeData[range].FSCurrent * rangeMult;


  #define SharedApplInit
  #include <sharedAppl.c>

//--- add application initialization code here ---
//APPLCODE


  #define SharedApplInit2
  #include <sharedAppl.c>


  resetChan(0);  resetChan(1);

//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>

  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,     "..." },
  { "feCalib",      feCalib,      "Show BCM Front End calibration data" },
  { "convApplCfg",  convApplCfg,  "Read BCM calib data from ApplCfg file" },
  { "nsclseriald",  nsclSeriald,  "Start NSCL Serial service" },
  { "state",        showState,    "Show state of controller"  }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
