//----- ModularPS-FETAppl.c ----- 12/04/13 -----(03/05/06)----

// EE Documents bundle # 211

//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//  Application code for the Modular Power Supply
//-----------------------------------------------------------------------------
//  Update Log (most recent first)
//
//
//  2013-12-04 - Mark Davis
//               - Added logic to update new onStatesBitMap in the DiagLogInfo
//                 structure.
//               - Added checkPrevOnStates() for setting offDueToReset, which
//                 is used to assert an interlock if the supply shut off due to
//                 an unexpected reset.
//
//  2013-07-01 - Mark Davis
//               - Added PS_CTLR_RESET bit to statusBitsAppl.  This is set
//                 during startup whenever the previous state could not be
//                 restored, and cleared the first time one of the supplies is
//                 successfully turned on.
//
//  2012-11-16 - Mark Davis (*)
//               - Added debounceMS setting and the debounceDigIn() function
//                 developed for the TCP controller.  Prior to this, there
//                 was at least one observed instance of anomalous readings on
//                 the digital inputs (where nearly all the inputs changed from
//                 a 0 to a 1 state for one pass).
//
//  2011-12-28 - Mark Davis
//               - Set SchedMS to 20.
//               - Logging of digital IO signal changes and computation of the
//                 next ampApply value are now based on the SchedMS setting.
//
//  2011-06-24 - Mark Davis
//               - Changed the order of the data elements in the devData file
//                 to deal with order-of-restoration dependencies.
//
//  2010-02-22 - Mark Davis
//               - Changed the ApplAbbrev[] value to "ModPS" (the same as the
//                 the non-FET version) so the log file names can be found
//                 by the DevMgr application (since both types return the
//                 same value for the Device Type).
//
//  2010-02-19 - Mark Davis
//               - Corrected default labels for the TEMP_OK
//
//  2010-02-08 - Mark Davis
//               - Added msgs for the individual bulk supplies turning on and
//                 off
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2009-11-02 - Mark Davis -
//               - Copied contrast code from non-FET version of application
//                 so contrast can be adjusted and the setting saved.
//               - Added configuration for settings persistent values file.
//
//  2009-07-01 - Mark Davis -
//                - The enabled/disabled setting for each supply is now part
//                  of the persistent configuration data.  This was necessary
//                  because many of the controller boards are not completely
//                  stuffed resulting in input values that are erroneous.
//
//  2009-02-04 - Mark Davis
//               - reorganized and made changes as necessary to make it
//                 compatible with current base code standards
//               - added code to skip logging msgs for most input bit changes
//                 if the applic task is starving other processes for CPU cycles
//
//  2007-09-19 - Mark Davis
//               - first properly functioning prototype!
//
//  2007-05-08 - Mark Davis -
//               - Started new project.
//
//-----------------------------------------------------------------------------

/*-----------------------------------------------------------------------------
  Misc thoughts:

    - What if instead of trying "ride out" the glitch that happens when the
      FET's are turned back on at the end of SINK mode, we could eliminate
      the cause?  What if, just before turning on the FET's, we set ISET
      equal to IFDBK?  Would that eliminate the glitch?

    - Coming out of SINK mode seems to happen at a certain IFDBK voltage.
      Why is that?  Should it be based on the relative difference between
      IFDBK and ISET?  If the setpoint goes from a large value down to one
      that results in an IFDBK value well above 1.25 volts, it still has
      to go into and back out of SINK mode.  Does the glitch still happen
      in such cases?

    - Difference between load in SINK mode and supply in SINK mode (?)
      True or false:  The BJT switches polarity when VOUT drops below the
      VLIM_POS/_NEG limits, so the supply then tries to drive the load in
      the correct direction, even though it has some power to dissapate in
      the wrong polarity (?)  So the IFDBK voltage at which this occurs is
      dependant upon the VLIM_XXX settings?
*/

/*=============================================================================
              Notes on how the Modular Power Supply operates
  =============================================================================

    The ModularPS controller supports up to 10 "bulk" power supplies.  The
    output is normally connected to an inductive load (a magnet) through a
    bridge made up of BJT's and FET's.  The BJT's are used to set the output
    level of the supply, while the FET's turn the bridge on or off.

    The controller has an analog feedback loop that uses the following signals:

      ISET:  Active setpoint for the analog feedback loop
      IFDBK: The current through the load
      VOUT:  The voltage across the load
      VLIM:  A level-triggered signal, set by comparing VOUT to the following
             limits for the maximum allowed voltage across the load:
      VLIM_NEG:  The largest negative voltage allowed across the load
      VLIM_POS:  The largest positive voltage allowed across the load

    The combination of ISET and IFDBK provides regulation of the load current,
    while the VLIM value imposes a limit based on the voltage across the load.
    The combined result (VCTRL) is used to bias the BJT's which control the
    combined power output of the bulk supplies.


    The software's role in the control loop is to monitor the state of the
    load and the BJT's and determine when one side of the bridge should be
    turned on via the positive or negative FET driver.  The VCTRL, IFDBK,
    ISET, and VOUT values provide the needed information.


    The polarity of the VCTRL signal determines which polarity BJT driver is
    enabled.  This value, and the difference between ISET and IFDBK allow
    the software to determine when the power in the load is decreasing (SINK
    mode) and when power should be increased or maintained (RAMP and REG
    modes).  This works as follows:

    - If the difference between ISET and IFDBK exceeds some preset level,
      then the power in the load is changing.

      If the BJT polarity is consistent with the direction of change (+ for a
      value that is getting more positive, - for a value that is getting more
      negative), then the power supply is in RAMP mode.  In this mode the
      load voltage (VOUT) is normally at the VLIM_NEG or VLIM_POS limit.

      If the BJT polarity is NOT consistent with the direction of change, then
      the supply is in SINK mode.

      If there is little difference between ISET and IFDBK, then the supply is
      in REGulation mode.  At this point, the voltage across the load (VOUT)
      will also be significantly smaller than the VLIM_NEG/VLIM_POS values.

    - When the supply is in SINK mode, the dump resistors on the bulk supplies
      are sinking (absorbing) power from the load.


    When in RAMP or REG mode, the bridge should be turned on and the polarity
    of the enabled FET driver noramlly must match the active BJT driver.

    In SINK mode, the FET's are disabled to avoid dumping lots of power
    through the bridge.

    When the mode changes from SINK back to RAMP or REG, the appropriate FET
    must be re-enabled.  However, this is where a bit of a problem occurs.

    SINK mode ends when the polarity of the active BJT changes.  But when the
    load is reconnected at the end of the SINK mode, VCTRL spikes, causing the
    active BJT to switch back.  If the software attempts to keep the active
    FET in sync with the active BJT at this time, the result is oscillation.

    To avoid this problem, the changes in the BJT state are ignored for a
    (configurable) period time after coming out of SINK mode.  This avoids
    having the control loop oscillate between SINK and RAMP or REG mode.
    Only a change in polarity of the ISET-IFDBK difference will shorten this
    period and cause the control loop to resume using the BJT state to trigger
    another SINK mode.

===============================================================================
                    Notes on persistent register values
-------------------------------------------------------------------------------

  vlimPlus:          The maximum positive voltage across the load

  vlimMinus:         The maximum negative voltage across the load

  rampRate:          The rate at which the active setpoint changes until it
                     reaches the requested setpoint (in Amps/second)

  vctrlHyst:         The +/- threshold that VCTRL must cross to change the
                     perceived polarity of the BJT drivers (in volts)

  maxErrTime:        The maximum contiguous time normal operating parameters
                     are allowed to be exceeded before an Interlock condition
                     is asserted (in ms)

  maxVctrl:          The maximum value VCTRL is allowed to reach (more longer
                     than maxErrTime) before an Interlock condition is asserted.
                     This is provided as a safeguard against software and
                     hardware problems that might damage the FET/BJT bridge in
                     the power supplies. (in volts)

  forceBJTtime:      The amount of time that the polarity of the active BJT is
                     ignored after SINK mode ends.  This stops the control loop
                     from causing oscillations due to the brief spike in VCTRL
                     that happens when the FET's are turned back on at the end
                     of a SINK interval. (in ms)

  maxCurDiff:        Currently ignored.  Will be used in the future to check
                     for problems with the individual bulk supplies.

  maxVoltDiff:       Currently ignored.  Will be used in the future to check
                     for problems with the individual bulk supplies.

  maxDiffSec:        Currently ignored.  Will be used in the future to check
                     for problems with the individual bulk supplies.

  knobSelectRate:    Controls how much the knob on the front panel has be be
                     turned to change what value is selected.

  maxAmpSet:         The maximum allowable setpoint (in amps)

  minPSonTime:       The minimum amount of time the bulk supplies must be on
                     before normal operation is allowed (in ms)

  minFEToffTime:     Currently ignored.  Possibly used again in the future to
                     force a delay between turning FET's off and on (in ms)

  minBJTstableTime:  Minimum amount of time the polarity of the active BJT
                     must not change before the last known stable state of
                     the BJT's is updated (in ms)  (currently ignored)

  rampThreshold:     The minimum difference between ISET and IFDBK values that
                     indicates the load current is RAMPING or SINKING
                     (in raw volts, NOT amps!!!  The equivalent value in amps
                     varies based on # of supplies connected, so volts is used)

===============================================================================
-------------------------------------------------------------------------------
- LCD <==> modbus client?
- simple way to alter register values that invokes the same funcs that
  are called to handle a write from an external client?
--------------------------------------------------------------------------
NOTES:
--------------------------------------------------------------------------
Signals for each of the 10 supplies:

12-Analog Inputs:
- HOLEC Current
- Bulk Supply Volt

Digital Inputs:
- On/Off status
- Temperature
- Roll Call Jumper

Digital Outputs:
- On/Off Command

-------------------------------------------------
Signals for all supplies:

16-bit Analog Inputs:
- IFDBK     current through load
- VOUT      voltage across load
- VCTRL     output of BJT feedback-ctl circuit
- VLIM      scaled/limited load voltage used for BJT feedback-ctl circuit

16-bit Analog Outputs:
- ISet      setpoint for BJT feeback-ctl circuit
- VLimPos   positive limit for VOUT feedback ctl
- VLimNeg   negative limit for VOUT feedback ctl

Digital Inputs:
- HOLEC OK
- External Interlock

Digital Outputs:
- Activity LED

- Negative Polarity
- Positive Polarity
- Negative Polarity LED
- Positive Polarity LED

------------------------------------------------------------------------------*/
#include <common.h>

#include <unistd.h>
#include <sys/types.h>



#ifndef  ApplicSource
  #define  ApplicSource  "ModularPS-FETAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "ModularPS-FET" };
const char ApplAbbrev[] = { "ModPS" };

//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_ModularPS

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN     26   // analog inputs
#define  NUM_APPL_AOUT     3   // analog outputs
#define  NUM_APPL_DIN     32   // digital inputs
#define  NUM_APPL_DOUT    16   // digital outputs

#define  ADC_MASK_APPL  0x03FFFFFFL
#define  DAC_MASK_APPL  0x00000007L

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( 0x0000 )


#define SharedINC
#include <sharedAppl.c>


#define  NumPS  10  // # of Power Supplies supported



//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//=============== digitalIn bit usage ===============
#define  PS5_IS_ON       0x00000001    // DI-00
#define  PS5_AVAIL       0x00000002    // DI-01
#define  PS5_TEMP_OK     0x00000004    // DI-02

#define  PS10_IS_ON      0x00000008    // DI-03
#define  PS10_AVAIL      0x00000010    // DI-04
#define  PS10_TEMP_OK    0x00000020    // DI-05

#define  PS4_IS_ON       0x00000040    // DI-06
#define  PS4_AVAIL       0x00000080    // DI-07
#define  PS4_TEMP_OK     0x00000100    // DI-08

#define  PS9_IS_ON       0x00000200    // DI-09
#define  PS9_AVAIL       0x00000400    // DI-10
#define  PS9_TEMP_OK     0x00000800    // DI-11

#define  PS3_IS_ON       0x00001000    // DI-12
#define  PS3_AVAIL       0x00002000    // DI-13
#define  PS3_TEMP_OK     0x00004000    // DI-14

#define  PS8_IS_ON       0x00008000    // DI-15
#define  PS8_AVAIL       0x00010000L   // DI-16
#define  PS8_TEMP_OK     0x00020000L   // DI-17

#define  PS2_IS_ON       0x00040000L   // DI-18
#define  PS2_AVAIL       0x00080000L   // DI-19
#define  PS2_TEMP_OK     0x00100000L   // DI-20

#define  PS7_IS_ON       0x00200000L   // DI-21
#define  PS7_AVAIL       0x00400000L   // DI-22
#define  PS7_TEMP_OK     0x00800000L   // DI-23

#define  PS1_IS_ON       0x01000000L   // DI-24
#define  PS1_AVAIL       0x02000000L   // DI-25
#define  PS1_TEMP_OK     0x04000000L   // DI-26

#define  PS6_IS_ON       0x08000000L   // DI-27
#define  PS6_AVAIL       0x10000000L   // DI-28
#define  PS6_TEMP_OK     0x20000000L   // DI-29

#define  HOLEC_OK        0x40000000L   // DI-30
#define  NO_INTLK        0x80000000L   // DI-31

#define  PS_AVAIL_MASK   0x12492492L  // mask for all AVAIL bits

//--- map from Bulk PS index to first bit in a group of digitalIn bits ---
const uint  psDigInIdx[NumPS] = { 24, 18, 12, 6, 0, 27, 21, 15, 9, 3 };

//--- sensible, human-readable order for displaying digital inputs ---
const uchar  LogicOrderDIn[NUM_APPL_DIN] = {
  24,25,26, 18,19,20, 12,13,14,  6,7,8,   0,1,2,
  27,28,29, 21,22,23, 15,16,17,  9,10,11, 3,4,5,
  30, 31
};

//--- msgs to log automatically when a digital input changes state ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//-0 state descr-    -1 state descr-
  "PS1 is OFF",      "PS1 is ON",
  "PS5 NOT Avail",   "PS5 Avail",
  "PS5 Temp BAD",    "PS5 Temp OK",

  "PS10 is OFF",     "PS10 is ON",
  "PS10 NOT Avail",  "PS10 Avail",
  "PS10 Temp BAD",   "PS10 Temp OK",

  "PS4 is OFF",      "PS4 is ON",
  "PS4 NOT Avail",   "PS4 Avail",
  "PS4 Temp BAD",    "PS4 Temp OK",

  "PS9 is OFF",      "PS9 is ON",
  "PS9 NOT Avail",   "PS9 Avail",
  "PS9 Temp BAD",    "PS9 Temp OK",

  "PS3 is OFF",      "PS3 is ON",
  "PS3 NOT Avail",   "PS3 Avail",
  "PS3 Temp BAD",    "PS3 Temp OK",

  "PS8 is OFF",      "PS8 is ON",
  "PS8 NOT Avail",   "PS8 Avail",
  "PS8 Temp BAD",    "PS8 Temp OK",

  "PS2 is OFF",      "PS2 is ON",
  "PS2 NOT Avail",   "PS2 Avail",
  "PS2 Temp BAD",    "PS2 Temp OK",

  "PS7 is OFF",      "PS7 is ON",
  "PS7 NOT Avail",   "PS7 Avail",
  "PS7 Temp BAD",    "PS7 Temp OK",

  "PS1 is OFF",      "PS1 is ON",
  "PS1 NOT Avail",   "PS1 Avail",
  "PS1 Temp BAD",    "PS1 Temp OK",

  "PS6 is OFF",      "PS6 is ON",
  "PS6 NOT Avail",   "PS6 Avail",
  "PS6 Temp BAD",    "PS6 Temp OK",

  "*** HOLEC BAD ***",  "HOLEC OK",
  "*** External Interlock ***",  "External Interlock cleared"
};


//=============== digitalOut bit usage ===============
#define  PS10_SET_ON      0x00000001  // PS 10 ON ctl      DO-00
#define  PS5_SET_ON       0x00000002  // PS  5 ON ctl      DO-01
#define  PS9_SET_ON       0x00000004  // PS  9 ON ctl      DO-02
#define  PS4_SET_ON       0x00000008  // PS  4 ON ctl      DO-03
#define  PS8_SET_ON       0x00000010  // PS  8 ON ctl      DO-04
#define  PS3_SET_ON       0x00000020  // PS  3 ON ctl      DO-05
#define  PS7_SET_ON       0x00000040  // PS  7 ON ctl      DO-06
#define  UNUSED_OUT       0x00000080  // unused output     DO-07

#define  PS2_SET_ON       0x00000100  // PS  2 ON ctl      DO-08
#define  PS6_SET_ON       0x00000200  // PS  6 ON ctl      DO-09
#define  PS1_SET_ON       0x00000400  // PS  1 ON ctl      DO-10
#define  ENABLE_FET_NEG   0x00000800  // FET- ctl          DO-11
#define  ENABLE_FET_POS   0x00001000  // FET+ ctl          DO-12
#define  LED_POS          0x00002000  // Positive LED ctl  DO-13
#define  LED_NEG          0x00004000  // Negative LED ctl  DO-14
#define  ActivityLED      0x00008000  // Activity LED      DO-15

#define  PS_SET_ON_MASK      0x077F  // mask for all SET_ON bits

//--- map from Bulk PS index to digitalOut bits index ---
const uint psDigOutIdx[NumPS] = { 10, 8, 5, 3, 1, 9, 6, 4, 2, 0 };

//--- sensible, human-readable order for displaying digital outputs ---
const uchar  LogicOrderDOut[NUM_APPL_DOUT] = {
  10, 8, 5, 3, 1, 9, 6, 4, 2, 0, 11, 12, 13, 14, 15, 7 };

#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--    --1 state descr--
  "Set PS10 OFF",       "Set PS10 ON",
  "Set PS05 OFF",       "Set PS05 ON",
  "Set PS09 OFF",       "Set PS09 ON",
  "Set PS04 OFF",       "Set PS04 ON",
  "Set PS08 OFF",       "Set PS08 ON",
  "Set PS03 OFF",       "Set PS03 ON",
  "Set PS07 OFF",       "Set PS07 ON",
  NULL,                 NULL,
  "Set PS02 OFF",       "Set PS02 ON",
  "Set PS06 OFF",       "Set PS06 ON",
  "Set PS01 OFF",       "Set PS01 ON",
  "FET [-] Disabled",   "FET [-] Enabled",
  "FET [+] Disabled",   "FET [+] Enabled",
  NULL,                 NULL,
  NULL,                 NULL,
  NULL,                 NULL
};


//--- extraIn bit usage ---
#define  SELECT_BTN      0x0001  // select LCD page to show on LCD
#define  SET_BTN         0x0002  // change value on LCD display
#define  START_BTN       0x0004  // turn ON
#define  STOP_BTN        0x0008  // turn OFF
#define  RESET_BTN       0x0010  // reset interlock (?)
// for future use:
#define  ID_BIT_08       0x0020  // extended ID bit 8
#define  ID_BIT_09       0x0040  // extended ID bit 9
#define  ID_BIT_10       0x0080  // extended ID bit 10

#define  SetBtnPressed     (! (extraIn & SET_BTN))
#define  SelectBtnPressed  (! (extraIn & SELECT_BTN))


//=============== Analog input indexes ================
//--- 16-bit ADC channels ---
#define  AI_IFDBK    0
#define  AI_VOUT     1
#define  AI_VCTRL    2
#define  AI_VLIM     3
//--- 1st 12-bit ADC channels ---
#define  AI_BSV6     4
#define  AI_HECT6    5
#define  AI_BSV1     6
#define  AI_HECT1    7
#define  AI_BSV7     8
#define  AI_HECT7    9
#define  AI_BSV2    10
#define  AI_HECT2   11
#define  AI_BSV8    12
#define  AI_SPARE0  13
#define  AI_SPARE1  14
//--- 2nd 12-bit ADC channels ---
#define  AI_HECT8   15
#define  AI_BSV3    16
#define  AI_HECT3   17
#define  AI_BSV9    18
#define  AI_HECT9   19
#define  AI_BSV4    20
#define  AI_HECT4   21
#define  AI_BSV10   22
#define  AI_HECT10  23
#define  AI_HECT5   24
#define  AI_BSV5    25

//--- sensible, human-readable order for displaying analog inputs ---
const uchar  LogicOrderAIn[NUM_APPL_AIN] = {
   7, 11, 17, 21, 24, 5, 9, 15, 19, 23,  // cur1 - cur10
   6, 10, 16, 20, 25, 4, 8, 12, 18, 22,  // bias1 - bias10
   0,  1,  2,  3,   // ifdbk, vout, vctrl, vlim
  13, 14  // SPARE0, SPARE1
};


//=============== Analog output indexes ===============
#define  AO_ISET       0
#define  AO_VLIM_POS   1
#define  AO_VLIM_NEG   2

//--- sensible, human-readable order for displaying analog outputs ---
const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0, 1, 2 };


//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ---------------

//--- example of standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);



//=============================================================================
//  Definition of data elements needed to keep track of the current state of
//  the device and the controller logic
//=============================================================================
typedef struct  {
//--- values updated in "cur" instance at start of each pass ---
  bool   isOn;       // bulk supply is ON
  bool   isAvail;    // bulk supply is available
  bool   tempOK;     // no termp problem with this supply

  float  biasVolts;  // BIAS voltage readback
  float  current;    // current output readback

//--- values updated in "next" instance during each pass ---
  bool   enabled;    // supply not manually or automatically disabled
  bool   setOn;      // assert ON for this supply
}  BulkSupplyState;


//-------------------------------------------------------------------------


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct  {
  BulkSupplyState  bs[NumPS];  // state of each bulk supply

//--- values updated in "next" instance by tasks other than "applic" ---
//--- these are copied to "cur" at start of the pass and then "reset" ---
//--- (as needed) so we can tell if they are set again ---
  bool   onRequest;
  bool   offRequest;
  bool   resetRequest;

  float  ampSet;      // most recent setpoint

//--- values updated in "cur" instance at start of each pass ---
  ulong  digitalIn;

  bool   externIntlk;

  uint   btnBits;
  uint   psCount;     // # of avail and enabled bulk supplies

//--- values updated in "next" instance during each pass ---
  uint   state;       // STATE_INIT, STATE_OFF, STATE_ON, STATE_INTLK;

  float  ampApply;    // in-use value, ramping toward ampSet

  uint   fetState;    // state for FET drivers (OFF, -, +)
  uint   outMode;     // output mode (OFF, ON, REG+/-, SINK+/-, RAMP+/-)

  ulong  digitalOut;  // state for digital outputs
}  ApplState;


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );

bool  offDueToReset = false, loggedResetIntlk = false;


//--- values for fetState ---
#define  FET_OFF   0   // both FETs OFF
#define  FET_POS   1   // positive FET driver ON
#define  FET_NEG   2   // negative FET driver ON
#define  FET_BOTH  3   // an ERROR!!!!!

const char *FETstateDesc[4] = { "OFF", "[+]", "[-]", "ERR" };


float  last_vlimPlus, last_vlimMinus;

//--- values and flags set by the 1st stage of the control loop when ---
//--- interpreting the state of the inputs and processing the next ---
//--- event msg ---
bool  holecOK;

uint  enabledMask;  // bit-mapped persistent vers of BulkSupplyState[].enabled



//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

/*  Bias Supply Voltage for each supply */
REG3( bsv_1,          x, 30023, F32 )
REG3( bsv_2,          x, 30025, F32 )
REG3( bsv_3,          x, 30027, F32 )
REG3( bsv_4,          x, 30029, F32 )
REG3( bsv_5,          x, 30031, F32 )
REG3( bsv_6,          x, 30033, F32 )
REG3( bsv_7,          x, 30035, F32 )
REG3( bsv_8,          x, 30037, F32 )
REG3( bsv_9,          x, 30039, F32 )
REG3( bsv_10,         x, 30041, F32 )

/* Current total for each supply */
REG3( amps_1,         x, 30043, F32 )
REG3( amps_2,         x, 30045, F32 )
REG3( amps_3,         x, 30047, F32 )
REG3( amps_4,         x, 30049, F32 )
REG3( amps_5,         x, 30051, F32 )
REG3( amps_6,         x, 30053, F32 )
REG3( amps_7,         x, 30055, F32 )
REG3( amps_8,         x, 30057, F32 )
REG3( amps_9,         x, 30059, F32 )
REG3( amps_10,        x, 30061, F32 )

REG3( ampRead,        x, 30063, F32 )
REG3( voltRead,       x, 30065, F32 )
REG3( vctrl,          x, 30067, F32 )
REG3( vlim,           x, 30069, F32 )

REG3( state_1,        x, 30071, U16 )
REG3( state_2,        x, 30072, U16 )
REG3( state_3,        x, 30073, U16 )
REG3( state_4,        x, 30074, U16 )
REG3( state_5,        x, 30075, U16 )
REG3( state_6,        x, 30076, U16 )
REG3( state_7,        x, 30077, U16 )
REG3( state_8,        x, 30078, U16 )
REG3( state_9,        x, 30079, U16 )
REG3( state_10,       x, 30080, U16 )

REG3( outMode,        x, 30081, U16 )

REG3( ampApply,       x, 30082, F32 )

REG3( psCount,        x, 30084, U16 )

REG3( state,          x, 30085, U16 )

REG3( iset,           x, 30086, F32 )
REG3( ifdbk,          x, 30088, F32 )
REG3( vout,           x, 30090, F32 )

REG3( fetState,       x, 30092, U16 )

REG3( bjtStatus,      x, 30093, U16 )  /* combines polBJT and stablePolBJT */

REG3( latency,        x, 30094, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30095, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30096, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime, x, 30097, U16 )  // # ms used by last kernel pass

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//--- values for statusBitsAppl (implemented 07/01/13) ---
#define  PS_CTLR_RESET  0x000000001   // PS off since last CPU reset


//--- values for state_# registers ---
#define  PS_IS_ON     0x0001
#define  PS_AVAIL     0x0002
#define  PS_TEMP_OK   0x0004
#define  PS_SET_ON    0x0008
#define  PS_ENABLED   0x0010

#define  PS_STATE(idx)  ( *((uint *)((&state_1_RegInfo + (idx))->regVal)) )


//--- values for outMode ---
#define  OM_OFF     0   // supplies are off
#define  OM_WAIT    1   // waiting for supplies to stabilize
#define  OM_REG_P   2   // steady with + current flow
#define  OM_REG_N   3   // steady with - current flow
#define  OM_SINK_P  4   // dissipating power, positive current flow
#define  OM_SINK_N  5   // dissipating power, negative current flow
#define  OM_RAMP_P  6   // incr current flow in positive dir
#define  OM_RAMP_N  7   // incr current flow in negative dir

//--- string descr corresponding to outMode values ---
const char *OutModeDesc[8] = {
  "OFF  ", "WAIT ", "REG+ ", "REG- ", "SINK+", "SINK-", "RAMP+", "RAMP-" };


//  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif


//--------------------------------------------------------------
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

#ifdef REG4

REG4( timeZone_,        x, 40001, S16, RW, MB_timeZone )
REG4( curTime,          x, 40002, U32, RW, setU32 )

REG4( ipAddr,           x, 40004, U32, RW, setU32 )
REG4( netmask,          x, 40006, U32, RW, setU32 )
REG4( gateway,          x, 40008, U32, RW, setU32 )

REG4( setACoil,         _, 40010, U16, RW, setU16 )
REG4( resetACoil,       _, 40011, U16, RW, setU16 )

REG4( digOutBits,       x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,       x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,      x, 40016, U32, RW, MB_cmdBitsAppl )

REG4( ampSet,           x, 40018, F32, RW, MB_ampSet )
REG4( vlimPlus,         x, 40020, F32, RW, MB_vlimPlus )
REG4( vlimMinus,        x, 40022, F32, RW, MB_vlimMinus )
REG4( rampRate,         x, 40024, F32, RW, MB_rampRate )

REG4( vctrlHyst,        x, 40026, F32, RW, MB_vctrlHyst )
REG4( maxVctrl,         x, 40028, F32, RW, setF32 )
REG4( unused2,          x, 40030, F32, RW, setF32 )
REG4( unused3,          x, 40032, F32, RW, setF32 )

REG4( maxCurDiff,       x, 40034, F32, RW, MB_maxCurDiff )
REG4( maxVoltDiff,      x, 40036, F32, RW, MB_maxVoltDiff )
REG4( maxDiffSec,       x, 40038, F32, RW, MB_maxDiffSec )

REG4( knobSelectRate,   x, 40040, U16, RW, setU16 )

REG4( maxAmpSet,        x, 40041, F32, RW, setF32 )

REG4( minPSonTime,      x, 40043, U16, RW, setU16 )
REG4( minFEToffTime,    x, 40044, U16, RW, setU16 )
REG4( minBJTstableTime, x, 40045, U16, RW, setU16 )

REG4( rampThreshold,    x, 40046, F32, RW, setF32 )

REG4( maxErrTime,       x, 40048, U16, RW, setU16 )

REG4( forceBJTtime,     x, 40049, U16, RW, MB_forceBJTtime )

REG4( contrast,         x, 40050, U16, RW, setU16 )

REG4( debounceMS,       x, 40051, U16, RW, setU16 )  /* ms period for digIn debouncing */

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion


//----- values for state -----
#define  STATE_INIT   0
#define  STATE_OFF    1
#define  STATE_ON     2
#define  STATE_INTLK  3

const char *CtlStateDesc[4] = { "INIT ", "OFF  ", "ON   ", "INTLK" };


//----- values for cmdBitsAppl -----
#define  CBA_TURN_ON     0x00000001
#define  CBA_TURN_OFF    0x00000002
#define  CBA_CLR_INTLK   0x00000004

#define  CBA_ENABLE_1    0x00000008
#define  CBA_ENABLE_2    0x00000010
#define  CBA_ENABLE_3    0x00000020
#define  CBA_ENABLE_4    0x00000040
#define  CBA_ENABLE_5    0x00000080
#define  CBA_ENABLE_6    0x00000100
#define  CBA_ENABLE_7    0x00000200
#define  CBA_ENABLE_8    0x00000400
#define  CBA_ENABLE_9    0x00000800
#define  CBA_ENABLE_10   0x00001000

#define  CBA_DISABLE_1   0x00002000
#define  CBA_DISABLE_2   0x00004000
#define  CBA_DISABLE_3   0x00008000
#define  CBA_DISABLE_4   0x00010000
#define  CBA_DISABLE_5   0x00020000
#define  CBA_DISABLE_6   0x00040000
#define  CBA_DISABLE_7   0x00080000
#define  CBA_DISABLE_8   0x00100000
#define  CBA_DISABLE_9   0x00200000
#define  CBA_DISABLE_10  0x00400000


//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================

float dummyFloat;

//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data the represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------
/*--- examples from sharedAppl.c ---*
#define  AIN_LABELS     1000
#define  AOUT_LABELS    1001
#define  DIN_LABELS     1002
#define  DOUT_LABELS    1003

#define  TIME_ZONE      1004
#define  SRFS_IP_ADDR   1005

#define  READ_GAINS     1006
#define  READ_OFFSETS   1007
#define  SET_GAINS      1008
#define  SET_OFFSETS    1009
F*-----------------------------------*/

RESTORE_FUNC(enabledMask);

const ParamData far  devDataParams[] = {
  #define SharedDevParams
  #include <sharedAppl.c>
/* Examples from the SharedDevParams entries in sharedAppl.c:
  PERSISTENT_VAR( TIME_ZONE,     timeZone    ),
  PERSISTENT_VAR( SRFS_IP_ADDR,  srfsIPAddr  ),

  PERSISTENT_TBL( AIN_LABELS,    ainLabel,  MAX_LABEL_LEN ),
  PERSISTENT_TBL( AOUT_LABELS,   aoutLabel, MAX_LABEL_LEN ),
  PERSISTENT_TBL( DIN_LABELS,    dinLabel,  MAX_LABEL_LEN ),
  PERSISTENT_TBL( DOUT_LABELS,   doutLabel, MAX_LABEL_LEN ),

  PERSISTENT_TBL( READ_GAINS,    readGain   ,0 ),
  PERSISTENT_TBL( READ_OFFSETS,  readOffset ,0 ),
  PERSISTENT_TBL( SET_GAINS,     setGain    ,0 ),
  PERSISTENT_TBL( SET_OFFSETS,   setOffset  ,0 ),
*/

// NOTE:  The application-specific ones need to come AFTER the generic ones
//        if (as in this case) there are values that depend on the generic
//        ones (in this case, the vlimPlus and vlimMinus depend on the
//        setGain and setOffset values)
//APPLDATA
//                ID   ---value---  --restoredFunc--
  PERSISTENT_VAR(  1,  vlimPlus,         NULL ),
  PERSISTENT_VAR(  2,  vlimMinus,        NULL ),
  PERSISTENT_VAR(  3,  rampRate,         NULL ),
  PERSISTENT_VAR(  4,  vctrlHyst,        NULL ),
  PERSISTENT_VAR(  5,  maxErrTime,       NULL ),
  PERSISTENT_VAR(  6,  maxVctrl,         NULL ),
  PERSISTENT_VAR(  7,  forceBJTtime,     NULL ),
  PERSISTENT_VAR(  8,  maxCurDiff,       NULL ),
  PERSISTENT_VAR(  9,  maxVoltDiff,      NULL ),
  PERSISTENT_VAR( 10,  maxDiffSec,       NULL ),
  PERSISTENT_VAR( 11,  knobSelectRate,   NULL ),
  PERSISTENT_VAR( 12,  maxAmpSet,        NULL ),
  PERSISTENT_VAR( 13,  minPSonTime,      NULL ),
  PERSISTENT_VAR( 14,  minFEToffTime,    NULL ),
  PERSISTENT_VAR( 15,  minBJTstableTime, NULL ),
  PERSISTENT_VAR( 16,  rampThreshold,    NULL ),
  PERSISTENT_VAR( 17,  enabledMask,      PVRF(enabledMask) ),
  PERSISTENT_VAR( 18,  debounceMS,       NULL ),


  END_PARAM_DATA
};



//----------------------------------------------------
RESTORE_FUNC(ampSet);

const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
  PERSISTENT_VAR(  1,  ampSet,       PVRF(ampSet) ),
  PERSISTENT_VAR(  2,  contrast,     NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --restoreFunc--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams, 0 };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Declarations for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

void onRequest(void);
void offRequest(void);
void resetRequest(void);

void enableBulkSupply(uint idx);
void disableBulkSupply(uint idx);

void newSetpoint(float newAmps);


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//--- add definitions for application-specific global variables here ---

float * const bsvX[] = {
  &bsv_1, &bsv_2, &bsv_3, &bsv_4, &bsv_5,
  &bsv_6, &bsv_7, &bsv_8, &bsv_9, &bsv_10
};
float * const ampsX[] = {
  &amps_1, &amps_2, &amps_3, &amps_4, &amps_5,
  &amps_6, &amps_7, &amps_8, &amps_9, &amps_10
};

//--- mapping from supply index to Analog Input #'s ---
const uint BSVidx[NumPS] = {
  AI_BSV1, AI_BSV2, AI_BSV3, AI_BSV4, AI_BSV5,
  AI_BSV6, AI_BSV7, AI_BSV8, AI_BSV9, AI_BSV10
};
const uint HECTidx[NumPS] = {
  AI_HECT1, AI_HECT2, AI_HECT3, AI_HECT4, AI_HECT5,
  AI_HECT6, AI_HECT7, AI_HECT8, AI_HECT9, AI_HECT10
};

//--- list of bit-masks for testing state of each bulk PS ---
const uint SetOnBit[NumPS] = {
  PS1_SET_ON, PS2_SET_ON, PS3_SET_ON, PS4_SET_ON, PS5_SET_ON,
  PS6_SET_ON, PS7_SET_ON, PS8_SET_ON, PS9_SET_ON, PS10_SET_ON
};

const ulong IsOnBit[NumPS] = {
  PS1_IS_ON, PS2_IS_ON, PS3_IS_ON, PS4_IS_ON, PS5_IS_ON,
  PS6_IS_ON, PS7_IS_ON, PS8_IS_ON, PS9_IS_ON, PS10_IS_ON
};

const ulong AvailBit[NumPS] = {
  PS1_AVAIL, PS2_AVAIL, PS3_AVAIL, PS4_AVAIL, PS5_AVAIL,
  PS6_AVAIL, PS7_AVAIL, PS8_AVAIL, PS9_AVAIL, PS10_AVAIL
};

const ulong TempOKBit[NumPS] = {
  PS1_TEMP_OK, PS2_TEMP_OK, PS3_TEMP_OK, PS4_TEMP_OK, PS5_TEMP_OK,
  PS6_TEMP_OK, PS7_TEMP_OK, PS8_TEMP_OK, PS9_TEMP_OK, PS10_TEMP_OK
};


ulong  psOnTime;  // time the bulk supplies where turned on
bool  psReady;    // bulk supplies on long enough to be stable

ulong  lastFETonTime;  // last time one of the FET drivers was ON
bool  offToOnSafe;  // safe to chg a FET driver from OFF to ON

float  ampErr, ctlErr, lastCtlErr, absErr;  // error in load current
bool  polErr;   // is polarity of the error positive?

bool  polBJT, lastPolBJT, stablePolBJT, hystPolBJT, tempPolBJT;
ulong  lastBJTChgTime, sinkModeEndTime;

bool  ramping, sinkMode;

ulong  lastTimeCondsOK;

//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
  uint  ui;
  ulong  mask, cmdBits = MB_VAL(ulong);


  if (cmdBits & CBA_TURN_ON)    onRequest();
  if (cmdBits & CBA_TURN_OFF)   offRequest();
  if (cmdBits & CBA_CLR_INTLK)  resetRequest();

  mask = CBA_ENABLE_1;
  for (ui=0; ui<10; ++ui,mask<<=1)  if (cmdBits & mask)  enableBulkSupply(ui);

  mask = CBA_DISABLE_1;
  for (ui=0; ui<10; ++ui,mask<<=1)  if (cmdBits & mask)  disableBulkSupply(ui);

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_ampSet(SetRegFuncParams)
{
  newSetpoint(MB_VAL(float));

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_vlimPlus(SetRegFuncParams)
{
  float  limit, fval = MB_VAL(float);

//--- make sure value is achievable ---
  limit = SetMin(AO_VLIM_POS);  limit = FABS(limit);
  if (fval < 0.0)  fval = 0.0;
  if (fval > limit)  fval = limit;

  KLOCK();  vlimPlus = fval;  KUNLOCK();

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_vlimMinus(SetRegFuncParams)
{
  float  limit, fval = MB_VAL(float);

//--- make sure value is achievable ---
  limit = SetMax(AO_VLIM_NEG);  limit = FABS(limit);
  if (fval < 0.0)  fval = 0.0;
  if (fval > limit)  fval = limit;

  KLOCK();  vlimMinus = fval;  KUNLOCK();

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_rampRate(SetRegFuncParams)
{
  rampRate = MB_VAL(float);

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for the BJT/VCTRL Hysterisis
//-----------------------------------------------------------------------------
bool MB_vctrlHyst(SetRegFuncParams)
{
  float fval = MB_VAL(float);

  fval = FABS(fval);

  if (fval > 0.50)  fval = 0.50;

  vctrlHyst = fval;

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_maxCurDiff(SetRegFuncParams)
{
  maxCurDiff = MB_VAL(float);

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_maxVoltDiff(SetRegFuncParams)
{
  maxVoltDiff = MB_VAL(float);

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_maxDiffSec(SetRegFuncParams)
{
  maxDiffSec = MB_VAL(uint);

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_forceBJTtime(SetRegFuncParams)
{
  forceBJTtime = MB_VAL(uint);

  if (forceBJTtime > 10000)  forceBJTtime = 10000;

  return true;
}


//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

//-----------------------------------------------------------------------------
//  restore an ampSet value from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ampSet)
{
  newSetpoint(NEW_VAL(float));
}


//-----------------------------------------------------------------------------
//  restore the enabled/disabled settings for each bulk supply
//-----------------------------------------------------------------------------
RESTORE_FUNC(enabledMask)
{
  uint  u, bit, mask;

  mask = NEW_VAL(uint);

  for (u=0,bit=1; u<NumPS; ++u,bit<<=1)
    if (mask & bit)  enableBulkSupply(u);  else  disableBulkSupply(u);
}


//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>



//=============================================================================
//  Misc Application-specific functions
//=============================================================================

//-----------------------------------------------------------------------------
//  Return a # indicating the state of the FETs based on the passed digitalOut
//  value.
//
//  The return value is 0, 1, 2, or 3 with the following meaning:
//   0 - both FET drivers off
//   1 - negative FET driver on
//   2 - positive FET driver on
//   3 - both FET drivers on (an ERROR!)
//-----------------------------------------------------------------------------
uint bitStateFETS(ulong outBits)
{
  uint  state = 0;

  if (outBits & ENABLE_FET_NEG)  state = 1;
  if (outBits & ENABLE_FET_POS)  state |= 2;

  return state;
}




//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================

typedef struct LCDElem_  LCDElem;

typedef char * (*LCDFmtFunc)(LCDElem far *elem, void far *val, char *buf);
typedef bool   (*LCDChgFunc)(LCDElem far *elem, void far *val, float chgBy);
typedef void   (*LCDSetFunc)(LCDElem far *elem, void far *val);

typedef struct LCDElem_ {
  uchar        type;
  uchar        col, row;
  float        chgRate;
  void far    *val;
  int          valSize;
  long         p1, p2;
  LCDFmtFunc   fmtFunc;
  LCDChgFunc   chgFunc;
  LCDSetFunc   setFunc;
};

//--- values for LCDElem.type ---
#define  RO_DATA   0
#define  RW_DATA   1
#define  PAGE_REF  2


typedef struct {
  const char far   *name;      // name of the page
  const char far  **text;      // static text for page
  LCDElem far      *elems;     // elems that make up the page
  LCDElem far      *cur;       // ptr to current element
} PageDef;



#define LCD_ELEM_VAL(type)  ( *((type far *)elem->val) )

#define LCD_VAL(type)  ( *((type far *)val) )

#define VAR(name)   VFPV(name), sizeof(name)

#define ENABLED(idx)  VAR(Next.bs[idx-1].enabled)


#define LCD_FMT_FUNC(name)  \
  _nearcall char *name(LCDElem far *elem, void far *val, char *buf)

#define LCD_CHG_FUNC(name)  \
  _nearcall bool  name(LCDElem far *elem, void far *val, float chgBy)

#define LCD_SET_FUNC(name)  \
  _nearcall void  name(LCDElem far *elem, void far *val)

#define LCD_SET_REG_FUNC(funcname, reg, type)  \
  _nearcall void  funcname(LCDElem far *elem, void far *val)  { \
    type newVal = LCD_VAL(type);  \
    char  temp[64];  \
    printf("new value for %ls: %s\r\n", reg##_RegInfo.name,  \
           fmtRegVal(&reg##_RegInfo, (S16 *)&newVal, 0, temp) );  \
    (reg##_RegInfo.setRegFunc)(&reg##_RegInfo, 0, &newVal, -1);  \
  }

#define  PAGE_TEXT(name)   const char far * name##Text[]

#define  PAGE_ELEMS(name)  const LCDElem far name##Elems[]

#define  PAGE_DEF(name, title)  const PageDef far name##Page = {  \
                title, name##Text, VFP(name##Elems),  VFP(name##Elems) };


bool  lcdBlink;

//-----------------------------------------------------------------------------
LCD_FMT_FUNC(strF32)  {
  return fmtFloat(LCD_VAL(float), (int)elem->p1, (int)elem->p2, buf); }

LCD_CHG_FUNC(chgF32)  {  LCD_VAL(float) += chgBy;  return true; }


LCD_SET_FUNC(savF32)  {
  KLOCK();  LCD_ELEM_VAL(float) = LCD_VAL(float);  KUNLOCK(); }

LCD_FMT_FUNC(strU16)  {
  sprintf(buf, "%*u", (int)elem->p1, LCD_VAL(uint) );  return buf; }


LCD_SET_REG_FUNC(set_ampSet, ampSet, F32)
LCD_SET_REG_FUNC(set_plim, vlimPlus, F32)
LCD_SET_REG_FUNC(set_mlim, vlimMinus, F32)


LCD_FMT_FUNC(strCtlState) {
  uint ui = LCD_VAL(uint);
  if ((ui != STATE_INTLK) OR lcdBlink)  return (char *)CtlStateDesc[ui];
  return "     ";
}

LCD_FMT_FUNC(strOutMode) {
  return (char *) OutModeDesc[ LCD_VAL(uint) ]; }

LCD_FMT_FUNC(fetStr) {
  return (char *) FETstateDesc[LCD_VAL(uint)]; }

LCD_FMT_FUNC(bjtStr) {
  return (LCD_VAL(bool)) ? "[+]" : "[-]"; }

LCD_FMT_FUNC(strHolec) {
  return (LCD_VAL(ulong) & HOLEC_OK) ? "OK " : (lcdBlink?"BAD":"   "); }

LCD_FMT_FUNC(strIlk)  {
  return (LCD_VAL(ulong)  & NO_INTLK) ? "OK   " : (lcdBlink?"FAULT":"     "); }

LCD_FMT_FUNC(pageName)  {
  fstrcpy(buf, LCD_VAL(PageDef).name);  return buf;  }

LCD_FMT_FUNC(bitYN)  {
  return (LCD_VAL(ulong)  & elem->p1) ? "Y" : "N"; }

LCD_FMT_FUNC(bitOB)  {
  return (LCD_VAL(ulong) & elem->p1) ? "OK " : (lcdBlink?"BAD":"   "); }

LCD_FMT_FUNC(bitYN_)  {
  return (LCD_VAL(ulong)  & elem->p1) ? "YES" : "NO "; }

LCD_FMT_FUNC(enaYN)  {
  return LCD_VAL(bool) ? "YES" : "NO "; }

//========================================================================
PAGE_TEXT(main) = {
//----------------------------------------
 "SET:-####.##A  CUR:-####.##A   V:-##.###",
 " ",
 "ISET:-##.###V   +LIM:-##.###V   FET: ###",
 "VLIM:-##.###V   -LIM:-##.###V   BJT: ###",
 "IFDK:-##.###V  VCTRL:-##.###V      #####",
 " ",
 "HOLEC: ###  EXTILK: #####  #####  #PS:##",
//----------------------------------------
 NULL
};

PAGE_ELEMS(main) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RW_DATA,  5,1,  0.001,  VAR(ampSet),      8,2, strF32, chgF32, set_ampSet },
  { RO_DATA, 20,1,  0.0,    VAR(ampRead),     8,2, strF32,       NULL, NULL },
  { RO_DATA, 34,1,  0.0,    VAR(voltRead),    7,3, strF32,       NULL, NULL },

  { RO_DATA,  6,3,  0.0,    VAR(iset),        7,3, strF32,       NULL, NULL },
  { RO_DATA,  6,4,  0.0,    VAR(vlim),        7,3, strF32,       NULL, NULL },
  { RO_DATA,  6,5,  0.0,    VAR(ifdbk),       7,3, strF32,       NULL, NULL },

  { RW_DATA, 22,3,  0.0001, VAR(vlimPlus),    7,3, strF32, chgF32, set_plim },
  { RW_DATA, 22,4,  0.0001, VAR(vlimMinus),   7,3, strF32, chgF32, set_mlim },
  { RO_DATA, 22,5,  0.0,    VAR(vctrl),       7,3, strF32,       NULL, NULL },

  { RO_DATA, 38,3,  0.0,    VAR(fetState),   -1,1, fetStr,       NULL, NULL },
  { RO_DATA, 38,4,  0.0,    VAR(polBJT),     -1,1, bjtStr,       NULL, NULL },
  { RO_DATA, 36,5,  0.0,    VAR(outMode),    -5,5, strOutMode,   NULL, NULL },

  { RO_DATA,  8,7,  0.0,    VAR(digitalIn),  -3,3, strHolec,     NULL, NULL },
  { RO_DATA, 21,7,  0.0,    VAR(digitalIn),  -5,5, strIlk,       NULL, NULL },
  { RO_DATA, 28,7,  0.0,    VAR(state),      -4,4, strCtlState,  NULL, NULL },
  { RO_DATA, 39,7,  0.0,    VAR(psCount),     2,0, strU16,       NULL, NULL },

  { 0,        0,0,  0.0,    NULL,              0,0, NULL, NULL, NULL }
};
PAGE_DEF(main, "MAIN");

//========================================================================
PAGE_TEXT(ps1) = {
//----------------------------------------
 "    --Bias- --Cur--  On  Temp Avail Ena ",
 "  1.-##.### -###.##  #/#  ###  ###  ### ",
 "  2.-##.### -###.##  #/#  ###  ###  ### ",
 "  3.-##.### -###.##  #/#  ###  ###  ### ",
 "  4.-##.### -###.##  #/#  ###  ###  ### ",
 "  5.-##.### -###.##  #/#  ###  ###  ### ",
 " ",
//----------------------------------------
 NULL
};
PAGE_ELEMS(ps1) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RO_DATA,  5,2,  0.0,  VAR(bsv_1),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,3,  0.0,  VAR(bsv_2),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,4,  0.0,  VAR(bsv_3),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,5,  0.0,  VAR(bsv_4),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,6,  0.0,  VAR(bsv_5),       7,3, strF32, NULL, NULL },

  { RO_DATA, 13,2,  0.0,  VAR(amps_1),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,3,  0.0,  VAR(amps_2),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,4,  0.0,  VAR(amps_3),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,5,  0.0,  VAR(amps_4),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,6,  0.0,  VAR(amps_5),      7,3, strF32, NULL, NULL },

  { RO_DATA, 22,2,  0.0,  VAR(digitalOut), PS1_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 22,3,  0.0,  VAR(digitalOut), PS2_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 22,4,  0.0,  VAR(digitalOut), PS3_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 22,5,  0.0,  VAR(digitalOut), PS4_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 22,6,  0.0,  VAR(digitalOut), PS5_SET_ON,0, bitYN, NULL, NULL },

  { RO_DATA, 24,2,  0.0,  VAR(digitalIn), PS1_IS_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,3,  0.0,  VAR(digitalIn), PS2_IS_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,4,  0.0,  VAR(digitalIn), PS3_IS_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,5,  0.0,  VAR(digitalIn), PS4_IS_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,6,  0.0,  VAR(digitalIn), PS5_IS_ON,0, bitYN, NULL, NULL },

  { RO_DATA, 27,2,  0.0,  VAR(digitalIn), PS1_TEMP_OK,0, bitOB, NULL, NULL },
  { RO_DATA, 27,3,  0.0,  VAR(digitalIn), PS2_TEMP_OK,0, bitOB, NULL, NULL },
  { RO_DATA, 27,4,  0.0,  VAR(digitalIn), PS3_TEMP_OK,0, bitOB, NULL, NULL },
  { RO_DATA, 27,5,  0.0,  VAR(digitalIn), PS4_TEMP_OK,0, bitOB, NULL, NULL },
  { RO_DATA, 27,6,  0.0,  VAR(digitalIn), PS5_TEMP_OK,0, bitOB, NULL, NULL },

  { RO_DATA, 32,2,  0.0,  VAR(digitalIn), PS1_AVAIL,0, bitYN_, NULL, NULL },
  { RO_DATA, 32,3,  0.0,  VAR(digitalIn), PS2_AVAIL,0, bitYN_, NULL, NULL },
  { RO_DATA, 32,4,  0.0,  VAR(digitalIn), PS3_AVAIL,0, bitYN_, NULL, NULL },
  { RO_DATA, 32,5,  0.0,  VAR(digitalIn), PS4_AVAIL,0, bitYN_, NULL, NULL },
  { RO_DATA, 32,6,  0.0,  VAR(digitalIn), PS5_AVAIL,0, bitYN_, NULL, NULL },

  { RO_DATA, 37,2,  0.0,  ENABLED(1),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,3,  0.0,  ENABLED(2),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,4,  0.0,  ENABLED(3),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,5,  0.0,  ENABLED(4),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,6,  0.0,  ENABLED(5),       3,0, enaYN, NULL, NULL },

  { 0,        0,0,  0.0,  NULL,             0,0, NULL, NULL, NULL }
};
PAGE_DEF(ps1, "S1-5");


//========================================================================
PAGE_TEXT(ps2) = {
//----------------------------------------
 "    --Bias- --Cur--  On  Temp Avail Ena ",
 "  6.-##.### -###.##  #/#  ###  ###  ### ",
 "  7.-##.### -###.##  #/#  ###  ###  ### ",
 "  8.-##.### -###.##  #/#  ###  ###  ### ",
 "  9.-##.### -###.##  #/#  ###  ###  ### ",
 " 10.-##.### -###.##  #/#  ###  ###  ### ",
 " ",
//----------------------------------------
 NULL
};
PAGE_ELEMS(ps2) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RO_DATA,  5,2,  0.0,  VAR(bsv_6),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,3,  0.0,  VAR(bsv_7),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,4,  0.0,  VAR(bsv_8),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,5,  0.0,  VAR(bsv_9),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,6,  0.0,  VAR(bsv_10),      7,3, strF32, NULL, NULL },

  { RO_DATA, 13,2,  0.0,  VAR(amps_6),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,3,  0.0,  VAR(amps_7),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,4,  0.0,  VAR(amps_8),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,5,  0.0,  VAR(amps_9),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,6,  0.0,  VAR(amps_10),     7,3, strF32, NULL, NULL },

  { RO_DATA, 22,2,  0.0,  VAR(digitalOut), PS6_SET_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 22,3,  0.0,  VAR(digitalOut), PS7_SET_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 22,4,  0.0,  VAR(digitalOut), PS8_SET_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 22,5,  0.0,  VAR(digitalOut), PS9_SET_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 22,6,  0.0,  VAR(digitalOut), PS10_SET_ON,0, bitYN, NULL, NULL },

  { RO_DATA, 24,2,  0.0,  VAR(digitalIn), PS6_IS_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 24,3,  0.0,  VAR(digitalIn), PS7_IS_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 24,4,  0.0,  VAR(digitalIn), PS8_IS_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 24,5,  0.0,  VAR(digitalIn), PS9_IS_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 24,6,  0.0,  VAR(digitalIn), PS10_IS_ON,0, bitYN, NULL, NULL },

  { RO_DATA, 27,2,  0.0,  VAR(digitalIn), PS6_TEMP_OK,0,  bitOB, NULL, NULL },
  { RO_DATA, 27,3,  0.0,  VAR(digitalIn), PS7_TEMP_OK,0,  bitOB, NULL, NULL },
  { RO_DATA, 27,4,  0.0,  VAR(digitalIn), PS8_TEMP_OK,0,  bitOB, NULL, NULL },
  { RO_DATA, 27,5,  0.0,  VAR(digitalIn), PS9_TEMP_OK,0,  bitOB, NULL, NULL },
  { RO_DATA, 27,6,  0.0,  VAR(digitalIn), PS10_TEMP_OK,0, bitOB, NULL, NULL },

  { RO_DATA, 32,2,  0.0,  VAR(digitalIn), PS6_AVAIL,0,  bitYN_, NULL, NULL },
  { RO_DATA, 32,3,  0.0,  VAR(digitalIn), PS7_AVAIL,0,  bitYN_, NULL, NULL },
  { RO_DATA, 32,4,  0.0,  VAR(digitalIn), PS8_AVAIL,0,  bitYN_, NULL, NULL },
  { RO_DATA, 32,5,  0.0,  VAR(digitalIn), PS9_AVAIL,0,  bitYN_, NULL, NULL },
  { RO_DATA, 32,6,  0.0,  VAR(digitalIn), PS10_AVAIL,0, bitYN_, NULL, NULL },

  { RO_DATA, 37,2,  0.0,  ENABLED(6),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,3,  0.0,  ENABLED(7),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,4,  0.0,  ENABLED(8),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,5,  0.0,  ENABLED(9),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,6,  0.0,  ENABLED(10),      3,0, enaYN, NULL, NULL },

  { 0,        0,0,  0.0,  NULL,             0,0, NULL, NULL, NULL }
};
PAGE_DEF(ps2, "S6-10");


//========================================================================
PAGE_TEXT(config) = {
//----------------------------------------
 "Contrast: ### ",
 " ",
 " ",
 " ",
 " ",
 " ",
 " ",
//----------------------------------------
 NULL
};

//------------------------------------------------------------------------
//--- contrast ---
LCD_CHG_FUNC(chgContr)  {
  int ii = LCD_VAL(U16);
  ii += (int)chgBy;
  if (ii < 5)  ii = 5;
  if (ii > 200)  ii = 200;
  if ((uint)ii != LCD_VAL(U16))  { LCD_VAL(U16) = ii;  return true; }
  return false;
}
LCD_SET_REG_FUNC(setContr, contrast, U16)

/*
//--- loadType ---
LCD_CHG_FUNC(chgLTyp)  {
  int ii = LCD_VAL(U16);
  ii += (int)chgBy;
  if (ii < 0)  ii = 0;
  if (ii > 2)  ii = 2;
  if ((uint)ii != LCD_VAL(U16))  { LCD_VAL(U16) = ii;  return true; }
  return false;
}
LCD_SET_REG_FUNC(setLTyp, loadType, U16)

//--- supplyType ---
LCD_CHG_FUNC(chgSTyp)  {
  int ii = LCD_VAL(U16);
  ii += (int)chgBy;
  if (ii < 0)  ii = 0;
  if (ii > 2)  ii = 2;
  if ((uint)ii != LCD_VAL(U16))  { LCD_VAL(U16) = ii;  return true; }
  return false;
}
LCD_SET_REG_FUNC(setSTyp, supplyType, U16)
*/

//------------------------------------------------------------------------
PAGE_ELEMS(config) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RW_DATA, 11,1,  0.005,  VAR(contrast),    3,0, strU16,  chgContr,setContr},
//  { RW_DATA, 12,2,  0.001,  VAR(loadType),    8,8, loadTypeStr,chgLTyp,setLTyp},
//  { RW_DATA, 14,3,  0.001,  VAR(supplyType),  3,3, supTypeStr, chgSTyp,setSTyp},

  { 0,        0,0,  0.0,    NULL,              0,0, NULL, NULL, NULL }
};
PAGE_DEF(config, "CONFIG");


//========================================================================
/*
PAGE_TEXT(pageName) = {
//----------------------------------------
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
//----------------------------------------
 NULL
};

PAGE_ELEMS(pageName) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RW_DATA, ##,#,  #.###,  VAR(varName),    #,#,  xxxxxx,  xxxxxx, xxxxxx },

  { 0,        0,0,  0.0,    NULL,              0,0, NULL, NULL, NULL }
};
PAGE_DEF(pageName, "PAGE NAME");

*/

//========================================================================
PAGE_TEXT(list) = { NULL };

PAGE_ELEMS(list) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { PAGE_REF, 1,1,  0.0,  VAR(mainPage),    0,0,  pageName,  NULL, NULL },
  { PAGE_REF, 1,2,  0.0,  VAR(ps1Page),     0,0,  pageName,  NULL, NULL },
  { PAGE_REF, 1,3,  0.0,  VAR(ps2Page),     0,0,  pageName,  NULL, NULL },
  { PAGE_REF, 1,4,  0.0,  VAR(configPage),  0,0,  pageName,  NULL, NULL },

  { 0,        0,0,  0.0,  NULL,            0,0, NULL, NULL, NULL }
};
PAGE_DEF(list, "LIST");

//========================================================================

/*
........................................
........................................
........................................
........................................
........................................
........................................
........................................


*/


//-----------------------------------------------------------------------------
  #define SetNormal(fd)  { \
        write(fd, "\376\061\001\376\062\000\000\000\001\100", 10); }
  #define SetReverse(fd)  { \
        write(fd, "\376\061\002\376\062\000\000\000\001\100", 10); }
  #define ResetLCD(fd)  {  \
      write(fd, "\376\223\001\376\071\147\376\120\200\376\223\000", 12);  \
      write(fd, "\376\072\024\024\376\130\376\122", 8);  SetNormal(fd);  \
}

//-----------------------------------------------------------------------------
//  Change the current element ptr for a page to the next/prev page element
//  that can be altered
//
//  If no such elements, then the ptr is left unchanged
//
//  delta determines direction (>=0 means next, < 0 means prev)
//-----------------------------------------------------------------------------
void chgFocus(PageDef far *page, int delta)
{
  LCDElem far *elem = page->cur;

  for (;;)  {
    if (delta < 0)  --elem;  else  ++elem;
    if ((elem < page->elems) OR !elem->col)  return;
    if (elem->type != RO_DATA)  { page->cur = elem;  return; }
  }
}


//-----------------------------------------------------------------------------
//  display the static text for an LCD display page
//-----------------------------------------------------------------------------
void updateLCD(fd_t fd, PageDef far *page, void far *modVal)
{
  uchar  i;
  bool  hasFocus;
  char far **ptr;
  char  *str;
  void far  *val;
  LCDElem far *elem;
  char  temp[50];
  static PageDef far *lastPage = NULL;


  SetNormal(fd);

//--- clear screen and draw static text if page changed ---
  if (page != lastPage)  {
  //--- Home + CLS, disable scrolling, set Normal font ---
    write(fd, "\376\130\376\122", 4);  SetNormal(fd);

    temp[41] = 0;   ptr = (char far **) page->text;
    for (i=1; *ptr; ++i,++ptr)  {
      sprintf(temp, "\376\107\001%c%.41ls", i, *ptr);
      write(fd, temp, strlen(temp));
    }

    lastPage = page;
  }

//--- display the data and link elements ---
  for (elem=page->elems; elem->col; ++elem)  {
    hasFocus = (elem == page->cur);
    if (elem->type == RO_DATA)  hasFocus = false;

    val = elem->val;
    if (hasFocus)  { SetReverse(fd);  if (modVal)  val = modVal;  }

    sprintf(temp, "\376\107%c%c", elem->col, elem->row);  // set cursor pos
    str = elem->fmtFunc(elem, val, temp+4);
    if (str != temp+4)  strncpy(temp+4, str, (sizeof temp)-5);
    temp[(sizeof temp)-1] = 0;  write(fd, temp, strlen(temp));

    if (hasFocus)  SetNormal(fd);
  }

}


//-----------------------------------------------------------------------------
//  Process to manage the front panel LCD display and monitor front panel
//  controls
//-----------------------------------------------------------------------------
void _task lcdUI(int argc, char **argv)
{
#pragma warn -775

  fd_t  lcdfd;
  int  delta, knobCount, newCount;
  uint  flowChars, btnBits, prevBtnBits, curContrast, prevContrast;
  bool  updateDisplay;
  ulong  baudRate, lastKnobTime, lastUpdateTime;
  float  chg;
  char  contrByte;
  void far  *modVal = NULL;
  LCDElem far  *elem;
  PageDef far  *page;
  char  *lcdDev;
  char  valBuf[64];


  redirectStdio();

//--- open port for LCD connection in Xon/Xoff mode ---
  lcdDev = (hdwType == HT_MODULAR_PS) ? "/dev/ttyx2" : "/dev/ttyx3";
  if ((lcdfd = open(lcdDev, O_RDWR, 0)) < 0)  {
    printf("Unable to open port for LCD\r\n");  Exit(-3); }

//--- set baud rate and Xon/Xoff chars ---
  //finish - use std termio interface instead of non-std ioctl values
  baudRate = 19200;  ioctl(lcdfd, IOC_SET_BAUD, &baudRate);
  flowChars = 0xFFFE;  ioctl(lcdfd, IOC_SET_FLOW_CHARS, &flowChars);

//--- Initialize LCD: ---
//---   Enable Flow Ctl, Xoff when 20 unused bytes, Xon when 20 unread bytes ---
//---   Home + CLS, disable scrolling
  write(lcdfd, "\376\072\024\024\376\130\376\122", 8);  SetNormal(lcdfd);

  page = (PageDef far *) &mainPage;  updateDisplay = true;

//--- get initial values ---
  prevContrast = 0;
  KLOCK();
  btnBits = Cur.btnBits;  curContrast = contrast;
  KUNLOCK();
  _EnterCritical_();  knobCount = qdCount = 0;  _ExitCritical_();

  lastUpdateTime = GetMS();  updateDisplay = true;

  for (;;)  {
    if (! updateDisplay)  {
      sleepMS(100);
      if (msSince(lastUpdateTime) >= 500)  updateDisplay = true;
    }

  //--- update display periodically or because of user action ---
    if (updateDisplay)  {
      lcdBlink = !lcdBlink;
      updateLCD(lcdfd, page, modVal);  updateDisplay = false;
      lastUpdateTime = GetMS();
    }

  //--- get last recorded state of buttons ---
    prevBtnBits = btnBits;
    KLOCK();
    btnBits = Cur.btnBits;  curContrast = contrast;
    KUNLOCK();

  //--- get # of knob ticks since last pass and reset count ---
    _EnterCritical_();  newCount = qdCount;  qdCount = 0;  _ExitCritical_();


    if (newCount)  {
      lastKnobTime = GetMS();  knobCount += newCount;  }
    else
      if (msSince(lastKnobTime) > 500)  knobCount = 0;

  //--- if knob turned while no buttons are being pressed, move the ---
  //--- highlight to the next or prev selectable element ---
    if (!btnBits AND (abs(knobCount) > knobSelectRate))  {
      delta = (knobCount < 0) ? -1 : 1;
      knobCount -= knobSelectRate * delta;
      chgFocus(page, delta);  updateDisplay = true;
    }

  //--- SELECT button:
  //---   if focus is on page ref, then switch to selected page
  //---   otherwise, switch to page with list of pages
    if ((btnBits == SELECT_BTN) AND (prevBtnBits != SELECT_BTN))  {
      elem = page->cur;
      if (elem->type != PAGE_REF)
        page = (PageDef far *) &listPage;
      else
        page = (PageDef far *) elem->val;
      updateDisplay = true;
    } while (0);


  //--- SET button (with focus on RW element):
  //---   if just pressed, make a copy of the cur elem's value
  //---   while pressed, apply knob movements to copy of cur's value
    if (btnBits == SET_BTN)  do  {
      if ((elem = page->cur)->type != RW_DATA)  break;
      if (!modVal)  {  // copy existing value if just pressed
        fmemcpy(valBuf, elem->val, elem->valSize);  modVal = valBuf;  }
      if (!knobCount)  break;
      chg = (float) knobCount;
      chg = SIGN(knobCount) * chg * chg * elem->chgRate;
      if (elem->chgFunc)  if (elem->chgFunc(elem, modVal, chg))  {
        updateDisplay = true;  knobCount = 0;  }
    } while (0);
    else  if (modVal)  {
      if (elem->setFunc)  elem->setFunc(elem, modVal);
//      printf("done modifying\r\n");
      modVal = NULL;  knobCount = 0;
    }


  //--- if contrast changed, update the display's contrast ---
    if (curContrast != prevContrast)  {
      write(lcdfd, "\376\120", 2);  contrByte = contrast;
      write(lcdfd, &contrByte, 1);
      prevContrast = curContrast;
    }


  //--- if SELECT and SET buttons both pressed at once, then reinitialize ---
  //--- the display (if manual override jumper was in when it was powered ---
  //--- up, this should force it to remember the 19200 baud rate when it ---
  //--- is next powered up without the jumper ---
    if (btnBits == (SELECT_BTN | SET_BTN))   {
      ResetLCD(lcdfd);  updateDisplay = true;  }

#pragma warn 775
  }

}


//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================


//-----------------------------------------------------------------------------
_nearcall char *fmtState(DispElem far *elem, char *buf)
{
  return (char *) CtlStateDesc[ ELEM_VAL(uint) ];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtMode(DispElem far *elem, char *buf)
{
  return (char *) OutModeDesc[ ELEM_VAL(uint) ];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtFET(DispElem far *elem, char *buf)
{
  return (char *) FETstateDesc[ELEM_VAL(uint)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtBJT(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? "[+]" : "[-]");
}

//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"Set:-####.##A    Read:-####.##A       --Bias-  --Cur-- On   Temp Avail Enabled\r\n"
"                                   1. -##.### -###.##  #/#   ###  ###    ###\r\n"
"Volts:-##.###V   Err: -####.##A    2. -##.### -###.##  #/#   ###  ###    ###\r\n"
"                                   3. -##.### -###.##  #/#   ###  ###    ###\r\n"
"ISET: -##.###V   FET: ###          4. -##.### -###.##  #/#   ###  ###    ###\r\n"
"VLIM: -##.###V   BJT: ###          5. -##.### -###.##  #/#   ###  ###    ###\r\n"
"IFDK: -##.###V   #####/#####       6. -##.### -###.##  #/#   ###  ###    ###\r\n"
"                                   7. -##.### -###.##  #/#   ###  ###    ###\r\n"
"VCTRL:-##.###V  +LIM:-##.##V       8. -##.### -###.##  #/#   ###  ###    ###\r\n"
"#PS: ##         -LIM:-##.##V       9. -##.### -###.##  #/#   ###  ###    ###\r\n"
"                                  10. -##.### -###.##  #/#   ###  ###    ###\r\n"
"HOLEC: ###   ExtILK: #####\r\n"
"\r\n"
};

const DispElem  far dispElems[] = {
  {  4, 3, VFPV(ampSet),      DARK_BLUE, fmtF32,    8,2 },
  { 22, 3, VFPV(ampRead),     DARK_BLUE, fmtF32,    8,2 },

  {  6, 5, VFPV(voltRead),    DARK_BLUE, fmtF32,    7,3 },
  { 22, 5, VFPV(ampErr),      DARK_BLUE, fmtF32,    8,2 },

  {  6, 7, VFPV(iset),        DARK_BLUE, fmtF32,    7,3 },
  {  6, 8, VFPV(vlim),        DARK_BLUE, fmtF32,    7,3 },
  {  6, 9, VFPV(ifdbk),       DARK_BLUE, fmtF32,    7,3 },

  { 22, 7, VFPV(fetState),    DARK_BLUE, fmtFET,    3,3 },
  { 22, 8, VFPV(polBJT),      DARK_BLUE, fmtBJT,    3,3 },
  { 17, 9, VFPV(state),       DARK_BLUE, fmtState, -5,5 },
  { 23, 9, VFPV(outMode),     DARK_BLUE, fmtMode,  -5,5 },

  {  6,11, VFPV(vctrl),       DARK_BLUE, fmtF32,    7,3 },
  {  5,12, VFPV(psCount),     DARK_BLUE, fmtU16,    2,2 },

  { 21,11, VFPV(vlimPlus),    DARK_BLUE, fmtF32,    6,2 },
  { 21,12, VFPV(vlimMinus),   DARK_BLUE, fmtF32,    6,2 },

  { 38, 4, VFPV(bsv_1),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 5, VFPV(bsv_2),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 6, VFPV(bsv_3),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 7, VFPV(bsv_4),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 8, VFPV(bsv_5),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 9, VFPV(bsv_6),       DARK_BLUE, fmtF32,    7,3 },
  { 38,10, VFPV(bsv_7),       DARK_BLUE, fmtF32,    7,3 },
  { 38,11, VFPV(bsv_8),       DARK_BLUE, fmtF32,    7,3 },
  { 38,12, VFPV(bsv_9),       DARK_BLUE, fmtF32,    7,3 },
  { 38,13, VFPV(bsv_10),      DARK_BLUE, fmtF32,    7,3 },

  { 45, 4, VFPV(amps_1),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 5, VFPV(amps_2),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 6, VFPV(amps_3),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 7, VFPV(amps_4),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 8, VFPV(amps_5),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 9, VFPV(amps_6),      DARK_BLUE, fmtF32,    8,2 },
  { 45,10, VFPV(amps_7),      DARK_BLUE, fmtF32,    8,2 },
  { 45,11, VFPV(amps_8),      DARK_BLUE, fmtF32,    8,2 },
  { 45,12, VFPV(amps_9),      DARK_BLUE, fmtF32,    8,2 },
  { 45,13, VFPV(amps_10),     DARK_BLUE, fmtF32,    8,2 },

  { 73, 4, VFPV(Cur.bs[0].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73, 5, VFPV(Cur.bs[1].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73, 6, VFPV(Cur.bs[2].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73, 7, VFPV(Cur.bs[3].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73, 8, VFPV(Cur.bs[4].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73, 9, VFPV(Cur.bs[5].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73,10, VFPV(Cur.bs[6].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73,11, VFPV(Cur.bs[7].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73,12, VFPV(Cur.bs[8].enabled), DARK_BLUE, fmtYN, 3,3 },
  { 73,13, VFPV(Cur.bs[9].enabled), DARK_BLUE, fmtYN, 3,3 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};


const DigDispElem far digInElems[] = {
  {  7,14,  HOLEC_OK,  "BAD",DARK_RED,    "OK ",DARK_BLUE },
  { 21,14,  NO_INTLK,  "FAULT",DARK_RED,  "OK   ",DARK_BLUE },

  { 57, 4,  PS1_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 5,  PS2_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 6,  PS3_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 7,  PS4_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 8,  PS5_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 9,  PS6_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57,10,  PS7_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57,11,  PS8_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57,12,  PS9_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57,13,  PS10_IS_ON, "N",DARK_RED,  "Y",DARK_BLUE },

  { 61, 4,  PS1_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 5,  PS2_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 6,  PS3_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 7,  PS4_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 8,  PS5_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 9,  PS6_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61,10,  PS7_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61,11,  PS8_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61,12,  PS9_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61,13,  PS10_TEMP_OK,  "BAD",DARK_RED,  "OK ",DARK_BLUE },

  { 66, 4,  PS1_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 5,  PS2_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 6,  PS3_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 7,  PS4_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 8,  PS5_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 9,  PS6_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66,10,  PS7_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66,11,  PS8_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66,12,  PS9_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66,13,  PS10_AVAIL, "NO ",DARK_RED,  "YES",DARK_BLUE },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
  { 55, 4,  PS1_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 5,  PS2_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 6,  PS3_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 7,  PS4_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 8,  PS5_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 9,  PS6_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55,10,  PS7_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55,11,  PS8_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55,12,  PS9_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55,13,  PS10_SET_ON,  "N",DARK_RED,  "Y",DARK_BLUE },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};


const DigFmtInfo digInFmt[] = {
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
  { -1,-1, "",0, "",0 }
};

const AnalogFmtInfo  adcFmt[] = {
  { -1,-1,  0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};

//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}


//-----------------------------------------------------------------------------
//  Turn supplies On/Off, enable/disable individual supplies, reset interlock
//  condition
//-----------------------------------------------------------------------------
void _task cmd(int argc, char **argv)
{
  bool  validCmd, enableCmd, disableCmd;
  uint  idx;
  char  *cmdStr;


  validCmd = false;  cmdStr = argv[1];

  do  {
    if (argc < 2)  break;

    if (StriEq(cmdStr, "on"))  { onRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "off"))  { offRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "reset"))  { resetRequest();  validCmd = true;  break; }

    enableCmd = StriEq(cmdStr, "enable");
    disableCmd = StriEq(cmdStr, "disable");

    if (argc < 3)  break;

    if (enableCmd OR disableCmd)  {
      idx = ATOI(argv[2]);  if ((idx < 1) OR (idx > NumPS))  break;
      --idx;
      if (enableCmd)  enableBulkSupply(idx);  else  disableBulkSupply(idx);
      validCmd = true;  break;
    }

    if (StriEq(cmdStr, "set"))  {
      newSetpoint(atof(argv[2]));  validCmd = true;  break;  }

  }  while(0);


  if (! validCmd)  {
    printf("Usage:  cmd on|off|reset\r\n"
           "            enable|disable #\r\n"
           "            set #.#\r\n" );
    Exit(-1);
  }

  Exit(0);
}


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  voltsOut[idx] = ISETtoVolts(idx, setpoint);
}
*/


//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "cur" OR the "next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "cur" value for the "input", and the "prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "cur" values
//    represent the input values just read (what they are "now") and the
//    "prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "cur" values).
//
//    During a pass the output values in "cur" represent what the outputs
//    where set to AFTER the "cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

//-----------------------------------------------------------------------------
void onRequest(void)
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON request\n");
}

//-----------------------------------------------------------------------------
void offRequest(void)
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF request\n");
}

//-----------------------------------------------------------------------------
void resetRequest(void)
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET request\n");
}

//-----------------------------------------------------------------------------
void enableBulkSupply(uint idx)
{
  _EnterCritical_();  Next.bs[idx].enabled = true;  _ExitCritical_();

  enabledMask |= (uint)1 << idx;

  fprintf(DiagLog, "Enable Bulk Supply #%d\n", idx+1);
}

//-----------------------------------------------------------------------------
void disableBulkSupply(uint idx)
{
  _EnterCritical_();  Next.bs[idx].enabled = false;  _ExitCritical_();

  enabledMask &= ~((uint)1 << idx);

  fprintf(DiagLog, "Disable Bulk Supply #%d\n", idx+1);
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps)
{
  char  temp[16];


  KLOCK();

  if (FABS(newAmps) > maxAmpSet)  newAmps = (float)SIGN(newAmps) * maxAmpSet;

  Next.ampSet = newAmps;

  KUNLOCK();

  fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}



int  bitCounts[32];

//-----------------------------------------------------------------------------
//  Minimize the effect of noise on the digitial inputs in trade for some
//  added latency in response time to actual "legitimate" changes in state.
//
//  What this func does is add or subtract the periodic timer interval value
//  to/from a running total for each input bit, with a cap on the total being
//  the configured # of ms for debouncing.
//
//  The current state of each input is determined by whether the running total
//  for each input is positive or negative.  A positive value indicates a
//  predominately ON state, while a negative value indicates a predominiately
//  OFF state.
//
//  Inputs that are mostly stable (anomolies not very frequent) will quickly
//  reach the positive or negative limit and stay close to it until the "real"
//  state changes to the opposite state.  When that happens, the total will
//  shrink, cross 0, and grow toward the opposite limit.
//
//  NOTE the latency effect of this approach on the two extremes:
//    - No noise:  No errors, but Max latency (== the debounce period)
//    - Max noise:  Min latency, but constant errors.
//
//  NOTE that debounce period also effectively sets a limit on how short a
//  period that a change in state must last for it to be recognized:  Any
//  change that lasts for less than twice the debounceMS setting will
//  generally not be detected.
//
//  The setting for the debounce period effectively allows you to ajust the
//  inherent compromise between delays introduced for the correct readings vs
//  the probability of getting a false reading.  A setting of 0 effectively
//  disables the filter, while larger values reduce the effect of any noise.
//
//  This approach behaves aproximately like connecting a capacitor across a
//  noisy input.  As long as the noise level is low enough, the capacitor
//  will tend to charge up to the level of the ON voltage, or dischage to
//  0.  Noise will increase or decrease the voltage, but so long as it is
//  less frequent than the "real" value, the voltage across the capacitor
//  will tend to stay near the voltage level for the "real" state.
//-----------------------------------------------------------------------------
void debounceDigIn(void)
{
  int  i;
  int  *bitCount;
  ulong  mask, digIn;


  bitCount = bitCounts;  digIn = 0;  mask = 1;

  for (i=0; i<32;  ++i)  {
    if (digitalIn & mask)
      *bitCount += (int)SchedMS;
    else
      *bitCount -= (int)SchedMS;

    if (*bitCount >  (int)debounceMS)  *bitCount =  (int)debounceMS;
    if (*bitCount < -(int)debounceMS)  *bitCount = -(int)debounceMS;

    if (*bitCount > 0)  digIn |= mask;

    ++bitCount;  mask <<= 1;
  }

  digitalIn = digIn;
}


//-----------------------------------------------------------------------------



//=============================================================================
//  Copy the "cur" state from last time to "prev", and update "cur"
//=============================================================================
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).
//
//  This func first copies "cur" to "prev" and then "next" to "cur" which does
//  2 things for us:
//
//    - We have a copy of the controller's logical state from the last pass
//      (so we can determine what has changed)
//
//    - We have a copy of values set by other tasks indicating events/requests
//      that happend/were made since the beginning of the last pass
//
//  We then reset the values (as needed) that are set by other tasks to
//  indicate events and requests, and update the current state based on the
//  readings just taken and the requests and events communicated by other
//  tasks.
//
//  All of this is in preperation for the updateNextState() function, which
//  implements the main "decision making" portion of the application logic,
//  determining what the next state should be based on the latest conditions
//  (as determined by the updateCurState() function).
//
//=============================================================================
void updateCurState(void)
{
  uint  ui;
  BulkSupplyState  *bsCur;  //, *bsPrev;


  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

//  Next.processTime = GetMS();  curTime = time(NULL);

//  msSincePrev = msSince(Cur.processTime);

  debounceDigIn();  Next.digitalIn = digitalIn;


//--- clear the offDueToReset flag for ON or RESET requests ---
  if (Next.onRequest OR Next.resetRequest)  offDueToReset = false;


// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer.

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---


//--------------------------------------------------------------------
// Next we update the rest of the "cur" values needed by the control
// logic (the ones that do NOT come from "next" values set by other
// tasks)
//
// NOTES:
//  - This portion of the logic has order-of-execution dependencies,
//    so be sure that all values referenced in an expression are ones
//    set before they are used.
//
//  - References to "prev" values are always safe since they are just
//    a copy of values from the last pass
//--------------------------------------------------------------------


//--- log msgs for bit chgs only if not starving other processes for cycles ---
  if (SchedMS - applLoopTime > 2)  logDigIOChgs(false);


//APPLCODE

  Cur.btnBits = ~(uint)extraIn & 0x001F;


//--- Check for front-panel button activity ---
//--- These bits are not in the normal digitalIn so they are not covered ---
//---  by the call to logInputChgs() ---
  if (Cur.btnBits != Prev.btnBits)  {
    if (Cur.btnBits == START_BTN)  {
      Cur.onRequest = true;  printf("\nSTART btn pressed\n");  }
    if (Cur.btnBits == STOP_BTN)   {
      Cur.offRequest = true;  printf("\nSTOP btn pressed\n");  }
    if (Cur.btnBits == RESET_BTN)  {
      Cur.resetRequest = true;  printf("\nRESET intlk btn pressed\n");  }
  }


  if (offDueToReset AND !loggedResetIntlk)  {
    printf("*** Supplies are OFF due to ctlr RESET ***\n");
    loggedResetIntlk = true;
  }


  Cur.externIntlk = !(digitalIn & NO_INTLK);
  holecOK = BOOL(digitalIn & HOLEC_OK);

//--- unscaled input voltage readings ---
  ifdbk = anaIn[AI_IFDBK];
  vout  = anaIn[AI_VOUT];
  vctrl = anaIn[AI_VCTRL];
  vlim  = anaIn[AI_VLIM];

//--- input voltages scaled to engineering units ---
  ampRead = readVal[AI_IFDBK];
  voltRead = readVal[AI_VOUT];


//=================================================
//===== update the state of the bulk supplies =====
//=================================================

//--- update readings for each bulk supply and determine # of supplies ---
//--- that should be operating and the total current they are handling ---
  Cur.psCount = 0;  // ttlBulkCur = 0.0;
  for (ui=0; ui<NumPS; ++ui)  {
//    bsPrev = Prev.bs + ui;
    bsCur = Cur.bs + ui;

    bsCur->biasVolts = readVal[BSVidx[ui]];
    bsCur->current = readVal[HECTidx[ui]];

    bsCur->isOn = BOOL(digitalIn & IsOnBit[ui]);
    bsCur->tempOK = BOOL(digitalIn & TempOKBit[ui]);
    bsCur->isAvail = BOOL(digitalIn & AvailBit[ui]);

    if (!bsCur->isAvail)  continue;

    ++Cur.psCount;  //ttlBulkCur += bsCur->current;
  }

}


//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of ouput values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//     - Drastically reduces the complexity of the code by greatly reducing
//       the number of possible paths through the code.  This simplifies
//       testing and verification processes.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  uint  ui;
  bool  setPolFET, oscillationPeriod;
  BulkSupplyState  *bsCur, *bsNext;


#define  Pos_Pol  true
#define  Neg_Pol  false

#define BJT_Pos  (polBJT)
#define BJT_Neg  (!(polBJT))

#define Hyst_BJT_Pos  (hystPolBJT)
#define Hyst_BJT_Neg  (!(hystPolBJT))

#define Stable_BJT_Pos  (stablePolBJT)
#define Stable_BJT_Neg  (!(stablePolBJT))



  if (current != applTask)  return;

//APPLCODE

//-------------------------------------------------------------------
//  Phase 0:
//     IN:
//        updated state values from updateCurState()
//        curState reflects current state of outputs
//     OUT:
//        iset
//        psReady, psOnTime
//        lastFETonTime, offToOnSafe
//        ampErr, ctlErr, polErr, absErr
//        polBJT, lastPolBJT, lastBJTChgTime, stablePolBJT
//        ramping, sinkMode, sinkModeEndTime, tempPolBJT
//        lastTimeCondsOK

  iset = anaOut[AO_ISET];

  if (Cur.state != STATE_ON)  {
    psReady = false;  psOnTime = lastBJTChgTime = sinkModeEndTime = 0;
    lastTimeCondsOK = GetMS();
  }

//--- bulk supplies on long enough to have stable outputs? ---
  if (Cur.state == STATE_ON)  {
    if (!psOnTime)  psOnTime = GetMS();
    if (!psReady)  if (msSince(psOnTime) > minPSonTime)  psReady = true;
  }
/*
//--- have both FET drivers been off for a bit? ---
  if (Cur.fetState)  {  // one of the FET drivers ON?
    lastFETonTime = GetMS();  offToOnSafe = false; }
  else
    if (!offToOnSafe AND psReady AND (msSince(lastFETonTime) >= minFEToffTime))
       offToOnSafe = true;
*/
//--- polarity and magnitude of error in the load current ---
  ampErr = Cur.ampApply - ampRead;
  lastCtlErr = ctlErr;  ctlErr = iset - ifdbk;
  polErr = (ctlErr >= 0.0) ? Pos_Pol : Neg_Pol;
  absErr = FABS(ctlErr);

//--- is the positive or negative BJT driver currently active? ---
//--- NOTE:  Negative VCTRL means Positive BJTs are ON! ---
  polBJT = (vctrl <= 0) ? Pos_Pol : Neg_Pol;
  if (vctrl < -vctrlHyst)  hystPolBJT = Pos_Pol;
  if (vctrl >  vctrlHyst)  hystPolBJT = Neg_Pol;

//--- update the last known stable BJT polarity ---
  if (polBJT != lastPolBJT)  {
    lastBJTChgTime = GetMS();  lastPolBJT = polBJT; }
  else
    if (!lastBJTChgTime OR (msSince(lastBJTChgTime) >= minBJTstableTime))
      stablePolBJT = polBJT;

//--- are we in SINK mode? ---
//--- NOTE:  sink mode not allowed to be reasserted until change in ---
//---        polarity of error or oscillation due to previous one is over ---
  ramping = (absErr > rampThreshold);
  if (SIGN(ctlErr) != SIGN(lastCtlErr))  {
    sinkMode = false;  sinkModeEndTime = 0;  }
  else  {
    sinkMode |= ((polBJT != polErr) AND ramping AND !sinkModeEndTime);
    if (sinkMode AND (polBJT == polErr))  {
      sinkMode = false;  tempPolBJT = polBJT;  sinkModeEndTime = GetMS(); }
  }



//--------------------------------------------------------------------
// Next.state
// Next.onRequest, Next.offRequest, Next.resetRequest (resets)
//
//  Phase 1:
//     IN:
//        updated state values from updateCurState()
//        offDueToReset
//     OUT:
//        Next.state

//--- chg state from OFF to ON? ---
  if (Cur.onRequest)  do  {
    Next.onRequest = false;
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_ON)  { printf("Already ON\n\n");  break; }
    Next.state = STATE_ON;  printf("Turning ON\n\n");
  } while(0);

//--- chg state from ON to OFF? ---
  if (Cur.offRequest)  do {
    Next.offRequest = false;
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_OFF)  { printf("Already OFF\n\n");  break; }
    Next.state = STATE_OFF;  printf("Turning OFF\n\n");
  } while(0);

//--- chg state from INTLK to OFF? ---
  if (Cur.resetRequest)  do {
    Next.resetRequest = false;
    if (Cur.state != STATE_INTLK)  {
      printf("No Interlock to clear\n\n");  break;  }
    if (Cur.externIntlk)  {
      printf("External Interlock still exists\n\n");  break; }
    if (!holecOK)  {
      printf("HOLEC still reads BAD\n\n");  break; }
    Next.state = STATE_OFF;  printf("Interlock cleared\n\n");
  } while(0);

//--- chg state to INTLK due to external interlock? ---
  if (Cur.externIntlk OR !holecOK OR offDueToReset)  do {
    if (Cur.state == STATE_INTLK)  break;
    Next.state = STATE_INTLK;
  } while(0);


//--- check for exceeding safe operating conditions ---
  if ((FABS(vctrl) < maxVctrl) OR !psReady) lastTimeCondsOK = GetMS();
  if (msSince(lastTimeCondsOK) > maxErrTime)  {
    printf("***** Normal operating parameters exceeded for too long! *****\n\n");
    Next.state = STATE_INTLK;
  }


//-------------------------------------------------------------------
//  Phase 2:
//    IN:
//       updated values from updateCurState()
//       Next.state from Phase 1
//       sinkMode, sinkModeEndTime, forceBJTtime
//    OUT:
//       Next.fetState set for next pass

  oscillationPeriod = false;
  if (sinkModeEndTime)  {
    if (msSince(sinkModeEndTime) < forceBJTtime)
      oscillationPeriod = true;
    else
      sinkModeEndTime = 0;  // allow sink mode to be reasserted in future
  }

//--- determine next state for FET driver outputs ---
  Next.fetState = FET_OFF;
  if (psReady AND !sinkMode)  {
    if (oscillationPeriod)
      setPolFET = tempPolBJT;
    else
      setPolFET = hystPolBJT;
    if (setPolFET)
      Next.fetState = FET_POS;
    else
      Next.fetState = FET_NEG;
  }
/*
    if ( Hyst_BJT_Pos AND Stable_BJT_Pos
         AND (offToOnSafe OR (Cur.fetState == FET_POS)) )
      Next.fetState = FET_POS;
    if ( Hyst_BJT_Neg AND Stable_BJT_Neg
         AND (offToOnSafe OR (Cur.fetState == FET_NEG)) )
      Next.fetState = FET_NEG;
  }
*/

//-------------------------------------------------------------------
//  Phase 3:
//    IN:
//       updated values from updateCurState()
//       Next.state from Phase 1
//       Next.fetState from Phase 2
//    OUT:
//       Next.outMode
//       Next.bs[..]


//--- combine output state logic values in to a single index ---
  do  {
    if (Next.state != STATE_ON)  { Next.outMode = OM_OFF;  break; }
    if (!psReady)  { Next.outMode = OM_WAIT;  break; }
    if (sinkMode) {
      Next.outMode = (BJT_Pos ? OM_SINK_P : OM_SINK_N);  break; }
    if (ramping)  {
      Next.outMode = (BJT_Pos ? OM_RAMP_P : OM_RAMP_N);  break; }
    Next.outMode = (BJT_Pos ? OM_REG_P : OM_REG_N);
  } while(0);


//-------------------------------------------------------------------
// bsNext.setOn,
//    Description:
//       Determines which bulk supplies should have the setOn output
//       enabled at the start of the next pass
//    Inputs:
//       Cur.state
//       bsCur->enabled
//       bsCur->tempOK
//-------------------------------------------------------------------
//--- set next state for each bulk supply ---
  for (ui=0; ui<NumPS; ++ui)  {
    bsCur = Cur.bs + ui;
    bsNext = Next.bs + ui;
    bsNext->setOn =
      ((Cur.state == STATE_ON) AND bsCur->isAvail
       AND bsCur->enabled AND bsCur->tempOK);
  }

}


//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
  uint  ui;
  float  ampsPerPass, diff;


  if (current != applTask)  return;  // must only be called by "applic" task

  Next.digitalOut = 0;

//APPLCODE

  for (ui=0; ui<NumPS; ++ui)
    if (Next.bs[ui].setOn)  Next.digitalOut |= SetOnBit[ui];

  if (Next.fetState == FET_POS)  Next.digitalOut |= (ENABLE_FET_POS | LED_POS);
  if (Next.fetState == FET_NEG)  Next.digitalOut |= (ENABLE_FET_NEG | LED_NEG);

//--- log changes to output mode ---
/*
  if (Next.fetState != Cur.fetState)  {
    printf("[%s] --> [%s]\r\n"
           "  FET:[%s]  BJT: cur:[%c]  stable:[%c]  hyst:[%c]\r\n"
           "  Ramping: %c  SinkMode: %c r\n",
             OutModeDesc[Cur.outMode], OutModeDesc[Next.outMode],
             FETstateDesc[Next.fetState],
             (BJT_Pos ? '+' : '-'), (Stable_BJT_Pos ? '+' : '-'),
             (Hyst_BJT_Pos ? '+' : '-'),
             (ramping ? 'Y' : 'N'), (sinkMode ? 'Y' : 'N')
           );
  }
  */

//--- compute new interim setting to apply based on actual setpoint, ---
//--- currently applied interim setting, and allowed rate of change ---
  ampsPerPass = rampRate * SchedMS * 0.001;  // amps/sec --> amps/SchedMS

  diff = Next.ampSet - Next.ampApply;

  if (Next.state != STATE_ON)
    Next.ampApply = 0.0;
  else
    if (FABS(diff) < ampsPerPass)
      Next.ampApply = Next.ampSet;
    else
      if (diff < 0.0)
        Next.ampApply -= ampsPerPass;
      else
        Next.ampApply += ampsPerPass;


  setVal[AO_ISET] = Next.ampApply;

  setVal[AO_VLIM_POS] = -vlimPlus / 1.5;
  setVal[AO_VLIM_NEG] = vlimMinus / 1.5;
}


//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  uint  ui, bsState, onStatesBitMap = 0;
  BulkSupplyState  *bsCur;


  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//APPLCODE

  fetState = Cur.fetState;
  ampApply = Cur.ampApply;

  state = Cur.state;

  ampSet = Cur.ampSet;


  bjtStatus = (BJT_Pos ? 1 : 0);
  bjtStatus |= (Stable_BJT_Pos ? 2 : 0);

  psCount = Cur.psCount;

  outMode = Cur.outMode;


  for (ui=0; ui<NumPS;  ++ui)  {
    bsCur = Cur.bs + ui;

    *bsvX[ui] = bsCur->biasVolts;
    *ampsX[ui] = bsCur->current;

    bsState = 0;

    if (bsCur->isOn)      bsState |= PS_IS_ON;
    if (bsCur->isAvail)   bsState |= PS_AVAIL;
    if (bsCur->tempOK)    bsState |= PS_TEMP_OK;
//    if (!bsCur->biasErr)  bsState |= PS_BIAS_OK;

    if (bsCur->setOn)     bsState |= PS_SET_ON;
    if (bsCur->enabled)   bsState |= PS_ENABLED;

    PS_STATE(ui) = bsState;

    if (bsCur->setOn AND bsCur->isOn)
      ClearBits(statusBitsAppl, PS_CTLR_RESET);
  }

  if (Cur.state == STATE_ON)  onStatesBitMap = 1;

  diagLogInfo->onStatesBitMap = onStatesBitMap;
}


//=============================================================================
//  Set the initial state we want the controller to start with.  This includes
//  "next" values that will be used to set the output signal states as soon
//  as the main application loop begins.
//=============================================================================
void setInitialState(void)
{
  uint  ui;
  BulkSupplyState  *bsNext;


  if (current != applTask)  return;  // must only be called by "applic" task

  updateCurState();

//---------------------------------------------------------------
//  At this point the analog and digital input values have been
//  read so they can be used in determining what state to start
//  with for the outputs and any other values.
//--------------------------------------------------------------

//--- update current state, both for use here, and so when updateCurState() ---
//--- is called near the start of the application loop, there will be valid ---
//--- data to copy to the previous state ---

  Next.ampSet = Next.ampApply = 0.0;

  Next.state = STATE_INIT;

  for (ui=0; ui<NumPS; ++ui)  {
    bsNext = Next.bs + ui;
    bsNext->enabled = true;  bsNext->setOn = false;
  }

  psOnTime = lastFETonTime = 0;  psReady = offToOnSafe = false;

  polBJT = lastPolBJT = stablePolBJT = hystPolBJT = true;

  dummyFloat = 0;
}



//=============================================================================
#define  BiasInputs    0x02551550L
#define  CurInputs     0x01aa8aa0L
#define  SpareInputs   0x00006000L

//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

//------------------------------------------------------------------
_nearcall void readDefaults(uint ui, float *gain, float *offset)
{
  *gain = 1.0;  *offset = 0.0;
  if (ui == AI_IFDBK)        *gain =  100.0;
  if (ui == AI_VOUT)         *gain =    1.0;
  if ((1L<<ui) & CurInputs)  *gain = 1000.0;
}
//------------------------------------------------------------------
_nearcall void readCalibPoints(uint ui, float *low, float *high)
{
  *low = -9.5;  *high = 9.5;
  if (ui == AI_IFDBK)        { *low = -500.0;  *high = 500.0;  }
  if ((1L<<ui) & CurInputs)  { *low =  -70.0;  *high =  70.0;  }
}
//--------------------------------------------------------
_nearcall char *readUnits(uint ui, char *buf)  {
   return (((1L<<ui) & CurInputs) OR (ui == AI_IFDBK)) ? "A" : "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwDAC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

//-----------------------------------------------------------------------------
_nearcall void setDefaults(uint ui, float *gain, float *offset)
{  *gain = (ui == AO_ISET) ? 100.0 : 1.0;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)
{
  if (ui == AO_ISET)  { *low = -500.0;  *high = 500.0;  return; }
  *low = -9.5;  *high = 9.5;
}
_nearcall char *setUnits(uint ui, char *buf)  {
  return (ui == AO_ISET) ? "A" : "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  uint  ui, uidx;


  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
  fstrcpy(ainLabel[AI_SPARE0], "SPARE0");
  fstrcpy(ainLabel[AI_SPARE1], "SPARE1");

  for (ui=0; ui<NumPS; ++ui)  {
    sprintf(ainLabel[BSVidx[ui]], "bias%u", ui+1);
    sprintf(ainLabel[HECTidx[ui]], "cur%u", ui+1);

    uidx = psDigInIdx[ui];
    sprintf(dinLabel[uidx],   "IS_ON_%u", ui+1);
    sprintf(dinLabel[uidx+1], "AVAIL_%u", ui+1);
    sprintf(dinLabel[uidx+2], "TEMP_OK_%u", ui+1);

    sprintf(doutLabel[psDigOutIdx[ui]], "SET_ON_%u", ui+1);
  }
  fstrcpy(dinLabel[30], "HOLEC_OK");
  fstrcpy(dinLabel[31], "NO_INTLK");

  fstrcpy(ainLabel[AI_IFDBK], "IFDBK");
  fstrcpy(ainLabel[AI_VOUT], "VOUT");
  fstrcpy(ainLabel[AI_VCTRL], "VCTRL");
  fstrcpy(ainLabel[AI_VLIM], "VLIM");

  fstrcpy(aoutLabel[AO_ISET], "ISET");
  fstrcpy(aoutLabel[AO_VLIM_POS], "VLIM_POS");
  fstrcpy(aoutLabel[AO_VLIM_NEG], "VLIM_NEG");

  fstrcpy(doutLabel[7],  "unused");
  fstrcpy(doutLabel[11], "Enable_FET_Minus");
  fstrcpy(doutLabel[12], "Enable_FET_Plus");
  fstrcpy(doutLabel[13], "FET_Plus_LED");
  fstrcpy(doutLabel[14], "FET_Minus_LED");
  fstrcpy(doutLabel[15], "ActivityLED");
}

//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  uint  u;
  BulkSupplyState  *bsNext;


  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

  vlimPlus = vlimMinus = 7.0;

  rampRate = 10.0;  // in amps/sec

  maxCurDiff = 20.0;  maxVoltDiff = 2.0;  maxDiffSec = 10.0;

  knobSelectRate = 20;  maxAmpSet = 2000.0;

  minPSonTime = 1000;  minFEToffTime = minBJTstableTime = 250;  // in ms

  rampThreshold = 0.25;  // in volts difference between ISET and IFDBK

  vctrlHyst = 0.050;  // in volts

  maxErrTime = 1000;  // in ms
  maxVctrl = 999.0;    // in volts

  forceBJTtime = 8000;  // in ms

  contrast = 32;  // LCD contrast, range 0-255

  enabledMask = 0xFFFF;  // all PS enabled unless overridden by stored config
  for (u=0; u<NumPS; ++u)  {
    bsNext = Next.bs + u;
    bsNext->enabled = true;
  }

  debounceMS = 250;  // interval (in ms) for debouncing digitalIn bits
}



//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);


  Exit(0);
}

//=============================================================================
//  Use the values in prevLogInfo to determine if we should assert interlocks
//  for any devices any devices that were shut off due to a non-soft reset.
//=============================================================================
void checkPrevOnStates(void)
{
  if (prevLogInfo.onStatesBitMap)  offDueToReset = true;
}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---


  #define SharedApplLocals2
  #include <sharedAppl.c>



  #define SharedApplInit
  #include <sharedAppl.c>

  BZERO(bitCounts);  // counts for debouncing the digital inputs

  _EnterCritical_();  SchedMS = 20;  _ExitCritical_();

//--- workaround for older devData files with order-of-reload dependencies ---
  if (vlimPlus == 0.0)  loadAllParams(applBuf, sizeof(applBuf));

  setInitialState();

//--- add application initialization code here ---
//APPLCODE


  #define SharedApplInit2
  #include <sharedAppl.c>

//--- start process to manage the LCD display ---
  argv_[0] = "lcdUI";  argv_[1] = NULL;
  createProc(lcdUI, 4, 1, argv_, 0, 1, 2);


  if (!restoredApplState)  SetBits(statusBitsAppl, PS_CTLR_RESET);


//--- The following loop is run at a periodic interval (every SchedMS ms). ---
//--- The application process consists of reading the current state of the ---
//--- digital and analog inputs, changing the state of the analog and ---
//--- digital outputs to the values computed during the previous pass, then ---
//--- computing new values for the outputs and updating values that outside ---
//--- clients can read. ---

  Next.state = STATE_OFF;

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>


  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
const CmdInfo far applCmds[] = {
  { "lcdui",     lcdUI,       "LCD user interface daemon" },
  { "cmd",       cmd,         "Turn On/Off, Reset, enable/disable, etc" },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
