//----- RFAC_Appl.c ----- 08/10/12 -----(02/15/11)----

// Sections that typically require application-specific code are marked with
// the comment:  //APPLCODE

//-----------------------------------------------------------------------------
//  Application specific code for the FRIB RF Amplifier Coordinator (RFAC)
//  controller.
//
//APPLCODE
//  Application code to ...
//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2012-08-10 - Mark Davis
//               - All 3 inputs (OK, TEMP OK, and AVAIL) must all be true for
//                 an amplifier to be considered ready.  Prior to this any
//                 amplifier with AVAIL false was ignored even if the ignore
//                 mask was not configured to ignore it.
//               - At least one amplifier must be ready and not ignored for
//                 the RF to be enabled.
//               - Added resetOnDelay to allow the OK (ON?) input to be
//                 temporarily ignored after a reset.
//
//  2011-02-15 - Mark Davis
//               - Original(skeletal) version of RFAC application code
//
//-----------------------------------------------------------------------------

/*
       ================== Operational description =================

   The RF Amplifier Coordinator (RFAC) controller is a simple device for
   monitoring the state of several inputs and controlling the state of a
   single output.

   The inputs are divided in to 2 groups:  8 groups of 3 inputs each are
   connected to RF amplifiers and provide the following information for
   each supply:

      - PS OK:  The amplifier is not reporting any errors
      - TEMP OK:  The amplifier is not reporting any temperature problem
      - ROLL CALL:  An amplifier is connected to the input

   The other group consists of 10 Temperature Fault inputs.  The connectors
   for each input supports up to 3 daisy-chained inputs.  A break anywhere
   in the chain (excluding wiring that purposely bypasses one or more of
   the inputs) will cause a fault to be asserted.

   To minimize the need to hardwire any inputs that are not used, the
   controller can be configured to ignore all the inputs for an amplifier
   or a temperature fault.

   The single output (excluding the Activity LED) is used to enable or
   disable the amplifiers.  This is NOT an ON/OFF control!  It mearly
   provides the means to enable or disable a group of amplifiers based on
   the state of ALL the amplifier and temperature fault inputs that are
   not being ignored.
*/

//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <sys/types.h>


//APPLCODE - change to match the name of this file
#ifndef  ApplicSource
  #define  ApplicSource  "RFAC_Appl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "RFAC" };
const char ApplAbbrev[] = { "RFAC" };


//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_RFAC

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN      1   // analog inputs
#define  NUM_APPL_AOUT     1   // analog outputs
#define  NUM_APPL_DIN     24   // digital inputs
#define  NUM_APPL_DOUT     8   // digital outputs
#define  NUM_APPL_XIN     24   // extra digital inputs

#define  ADC_MASK_APPL  0x00000000L
#define  DAC_MASK_APPL  0x0000000FL

#define  NumAmps     8   // max # of amplifiler/PS supported
#define  NumFaults  10   // max # of temperature fault inputs


#define SharedINC
#include <sharedAppl.c>


extern void _task rio(int argc, char **argv);


//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//APPLCODE - define macros for each Digital bit and each Analog index

//=============== digitalIn bit usage ===============
#define  PS4_OK          0x00000001  // 00: 1=Amp OK
#define  PS4_TEMP_OK     0x00000002  // 01: 1=Temp OK
#define  PS4_PRESENT     0x00000004  // 02: 1=Amp is present

#define  PS8_OK          0x00000008  // 03: 1=Amp OK
#define  PS8_TEMP_OK     0x00000010  // 04: 1=Temp OK
#define  PS8_PRESENT     0x00000020  // 05: 1=Amp is present

#define  PS3_OK          0x00000040  // 06: 1=Amp OK
#define  PS3_TEMP_OK     0x00000080  // 07: 1=Temp OK
#define  PS3_PRESENT     0x00000100  // 08: 1=Amp is present

#define  PS7_OK          0x00000200  // 09: 1=Amp OK
#define  PS7_TEMP_OK     0x00000400  // 10: 1=Temp OK
#define  PS7_PRESENT     0x00000800  // 11: 1=Amp is present

#define  PS6_OK          0x00001000  // 12: 1=Amp OK
#define  PS6_TEMP_OK     0x00002000  // 13: 1=Temp OK
#define  PS6_PRESENT     0x00004000  // 14: 1=Amp is present

#define  PS5_OK          0x00008000  // 15: 1=Amp OK
#define  PS5_TEMP_OK     0x00010000L // 16: 1=Temp OK
#define  PS5_PRESENT     0x00020000L // 17: 1=Amp is present

#define  PS2_OK          0x00040000L // 18: 1=Amp OK
#define  PS2_TEMP_OK     0x00080000L // 19: 1=Temp OK
#define  PS2_PRESENT     0x00100000L // 20: 1=Amp is present

#define  PS1_OK          0x00200000L // 21: 1=Amp OK
#define  PS1_TEMP_OK     0x00400000L // 22: 1=Temp OK
#define  PS1_PRESENT     0x00800000L // 23: 1=Amp is present

//--- pattern repeated every 3 bits for 8 PS/amplifiers ---
#define  PS_OK          0x01
#define  PS_TEMP_OK     0x02
#define  PS_PRESENT     0x04

#define  PS_OK_BITS     0x03  // OK and TEMP_OK
#define  PS_PRESENT     0x04
#define  PS_READY_BITS  0x07  // OK and TEMP_OK and PRESENT


const uchar  LogicOrderDIn[NUM_APPL_DIN] =  {
   21,22,23, 18,19,20, 6,7,8, 0,1,2, 15,16,17, 12,13,14, 9,10,11, 3,4,5 };
const uint  AmpBitsOrder[NumAmps] = { 4, 8, 3, 7, 6, 5, 2, 1 };
const uint  AmpBitsIndex[NumAmps] = { 7, 6, 2, 0, 5, 4, 3, 1 };

/*EXAMPLE:
#define  ClosedLoop(din)        ( (bool)((din) & CLOSED_LOOP) )
#define  FinalBtnPushed(din)    ( (bool)((din) & FINAL_BTN) )
#define  InitialBtnPushed(din)  ( (bool)((din) & INITIAL_BTN) )
#define  RFisOff(din)           ( (bool)((din) & RF_IS_OFF) )
#define  LocalMode(din)         ( (bool)((din) & LOC_REM_SW) )
*/


//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--    --1 state descr--
  "PS 4 NOT OK",       "PS 4 OK",
  "TEMP 4 NOT OK",     "TEMP 4 OK",
  "PS 4 NOT present",  "PS 4 IS present",

  "PS 8 NOT OK",       "PS 8 OK",
  "TEMP 8 NOT OK",     "TEMP 8 OK",
  "PS 8 NOT present",  "PS 8 IS present",

  "PS 3 NOT OK",       "PS 3 OK",
  "TEMP 3 NOT OK",     "TEMP 3 OK",
  "PS 3 NOT present",  "PS 3 IS present",

  "PS 7 NOT OK",       "PS 7 OK",
  "TEMP 7 NOT OK",     "TEMP 7 OK",
  "PS 7 NOT present",  "PS 7 IS present",

  "PS 6 NOT OK",       "PS 6 OK",
  "TEMP 6 NOT OK",     "TEMP 6 OK",
  "PS 6 NOT present",  "PS 6 IS present",

  "PS 5 NOT OK",       "PS 5 OK",
  "TEMP 5 NOT OK",     "TEMP 5 OK",
  "PS 5 NOT present",  "PS 5 IS present",

  "PS 2 NOT OK",       "PS 2 OK",
  "TEMP 2 NOT OK",     "TEMP 2 OK",
  "PS 2 NOT present",  "PS 2 IS present",

  "PS 1 NOT OK",       "PS 1 OK",
  "TEMP 1 NOT OK",     "TEMP 1 OK",
  "PS 1 NOT present",  "PS 1 IS present"
};

//=============== extraIn bit usage ===============
#define  TEMP_OK_5       0x00000001  // 00: 1=No Temp Fault
#define  TEMP_OK_4       0x00000002  // 01: 1=No Temp Fault

#define  TEMP_OK_3       0x00000004  // 02: 1=No Temp Fault
#define  TEMP_OK_2       0x00000008  // 03: 1=No Temp Fault
#define  TEMP_OK_1       0x00000010  // 04: 1=No Temp Fault
#define  TEMP_OK_6       0x00000020  // 05: 1=No Temp Fault

#define  TEMP_OK_7       0x00000040  // 06: 1=No Temp Fault
#define  TEMP_OK_8       0x00000080  // 07: 1=No Temp Fault
#define  TEMP_OK_9       0x00000100  // 08: 1=No Temp Fault
#define  TEMP_OK_10      0x00000200  // 09: 1=No Temp Fault

//--- all 10 temperature fault bits ---
#define  TEMP_OK_BITS    0x000003FF

const uint FaultBitsOrder[NumFaults] = { 5, 4, 3, 2, 1, 6, 7, 8, 9, 10 };
const uint FaultBitsIndex[NumFaults] = { 4, 3, 2, 1, 0, 5, 6, 7, 8, 9  };


//=============== digitalOut bit usage ===============
#define  UNUSED_DO_00    0x00000001   // 00:
#define  UNUSED_DO_01    0x00000002   // 01:
#define  UNUSED_DO_02    0x00000004   // 02:
#define  UNUSED_DO_03    0x00000008   // 03:
#define  UNUSED_DO_04    0x00000010   // 04:
#define  ActivityLED     0x00000020   // 05: Activity LED
#define  UNUSED_DO_06    0x00000040   // 06:
#define  RF_DISABLE      0x00000080   // 07: 1=Disable all amplifiers



const uchar  LogicOrderDOut[NUM_APPL_DOUT] =  { 0, 1, 2, 3, 4, 5, 6, 7 };

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( RF_DISABLE )


#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--    --1 state descr--
  "",                  "",
  "",                  "",
  "",                  "",
  "",                  "",
  "",                  "",
  "",                  "",
  "",                  "",
  "RF ENABLED",        "RF DISABLED"
};




//=============== Analog inputs ===============
//--- indexes for anaIn[] and readVal[] ---
/*EXAMPLE:
#define  RF_CAV_VOLT      0  // RF cavity voltage
#define  LIMITER_VOLT     1  // Setpoint Limiter voltage
#define  HDW_CFG          2  // Fixed volt indicating HD1 component set
*/

const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0 };


//=============== Anlog outputs ===============
//--- indexes for anaOut[] and setVal[] ---
/*EXAMPLE:
#define  SETPOINT         0  // Setpoint for Cav Voltage Regulator circuit
#define  FP_METER         1  // FrontPanel Meter if INITIAL or SET pressed
*/

const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0 };


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// Each value in the ApplState structure should be throught of as representing
// a "signal" in a clocked logic circuit, with the "applic" task representing
// the logic implemented by the circuit elements such as gates, latches, etc.

// The main "applic" task is the highest priority task and executes the code
// that implements the application logic on a real-time, periodic basis.

//                 ====== IMPORTANT NOTE on LOCKING ======
// Remember to use a semaphore or some other form of locking mechanism to
// protect all access to shared values by more than one task.  The only
// exception to this is when the applic task needs to access a value protected
// by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
// macros).  Because the applic task is the only task that runs at the highest
// priority, no other task can preempt it, and if the scheduler is locked,
// the applic task cannot preempt another task.  But scheduler locks must be
// kept short to avoid creating large latency in the timely execution of other
// tasks (the applic task in particular).


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------
  uint   resetRequest;    // reset Interlocks request


 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
//--- values updated in "cur" instance at start of each pass ---
  ulong  processTime;     // time (in ms) when processing was started
  ulong  digitalIn;       // state of digital inputs
  ulong  extraIn;         // state of extra digital inputs

  bool  psError;          // At least 1 active PS/ampl asserting an error
  bool  tempFault;        // At least 1 temp fault input being asserted
  bool  activeError;      // psError or tempFault is true

  uint  resetStarted;     // time of most recent allowed resetRequest (or 0)

 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task just after reading the state of the hardware inputs.  These values
 //  must  NOT be altered by other tasks (although they can read them as long
 //  as they use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------
//  bool   ovenModeOn;           // Turn on power, enable output soon after
//  bool   ovenOutErrInterlock;  // Oven PS Output state was wrong
//  ulong  ovenOutErrStartTime;  // start of period of Output state error


 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (i.e. it acts like a
 //  clocked latch but is in software and the clock is one pass through
 //  the main application loop).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance of ApplState by the applic
 //  task AFTER all the other ApplState values are set.  Because they represent
 //  only the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  digitalOut;      // state for digital outputs

  bool   interlocked;     // unacknowledged error/fault occured


//  float  activeSetpoint;  // value applied to regulation circuit

//  float  rfOnSetpoint;    // warm/final setpoint
//  float  rfOffSetpoint;   // cold/initial setpoint

} ApplState;


//bool  localMode, closedLoop;

ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );


//ulong  rfOnTime;   // time RF turned on (in milliseconds)

//float  regulationError;  // diff between setpoint and readback


//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//APPLCODE - add tasks that service network clients

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ----------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);



//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


//APPLCODE - set # of devices for which we want duplicate register sets
#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of coils
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read-only registers
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

REG3( latency,        x, 30023, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30024, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30025, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime, x, 30026, U16 )  // # ms used by last kernel pass

//APPLCODE - add registers for externally available read-only values
REG3( extraInBits,    x, 30027, U32 )  // additional digital inputs

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read/write registers
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

/*
  The "standard" functions for processing writes to Modbus register values:

    setS16, setU16, setS32, setU32, setF32

  These functions simply log a message for the change and update the
  value in a critical block (so the change cannot be interrupted).

  In many cases, just allowing any setting to be written to a register (or
  multiple registers comprising a larger value) is not acceptable.


  If any of the following apply, then a custom func must be used instead:

    - The order of changes is important.  For example, if you get ON,
      OFF, and RESET commands in rapid succession, you have to be sure
      they are processed in the order they were recieved.

    - Not all possible settings are acceptable.  The acceptable settings
      for a value are often a subset of the possible ones, and frequently
      are dependent upon the current state of the controller.


  IMPORTANT NOTES:

    The "applic" task periodically executes the code that makes up the
    bulk of the application logic.  Because it runs at the highest priority
    there is no chance that a lower priority task (like the daemon handling
    Modbus register writes) can change a value in the middle of this
    processing.

    However, the applic task CAN (and often does) preempt other processes,
    so any process that does read or make changes to the register values
    must lock the register values before doing so.  This is currently done
    by disabling the scheduler using the KLOCK() and KUNLOCK() macros.

    Use of such a lock insures that changes occur only between passes
    through the application logic, but there are still several issues
    that must be addressed when writing functions to handle writes to
    register values:

      - Priority inversion and latency:  Allowing a lower priority task
        to lock out higher ones (or, in this case, ALL tasks) allows
        the lower priority task to run when a higher-priority one should
        be running.  Care must be taken to keep such periods as short
        as possible to avoid disrupting the timing of the control logic.

      - Loss of encapsulation:  The last thing we want is to break the
        application logic in to small pieces spread through a bunch of
        functions executed by different tasks at different times.

        While the logic for processing writes to register values is often
        dependent upon the state of the control logic, functions outside
        the applic task must NOT directly alter the state of the application
        logic.  To achieve this, such functions must limit changes to
        values that act as input to the applic task.

        To this end, you need to treat the register values that act as
        inputs just like they were knobs, switches, analog inputs, or
        latches.  Like their hardware equivalent, the applic task logic
        reads their state at the beginning of each pass and uses this
        information when determining what the next state of the control
        logic should be.  And just like a hardware latch, values can be
        reset after reading them so the next write to the value can be
        treated as another instance of a request.

    The trickiest circumstances to handle are ones where the order of separate
    writes is important.  The most common example are commands for ON, OFF,
    and RESET.  If, for example, you had 3 separate flags, and simply set
    each once corresponding to which of the 3 commands you received, you
    end up with multiple flags set when the applic logic is processed, and
    no way to determine what the end result should be.  In these cases,
    the logic that controls the state of each flag should be placed in its
    own function and called during the processing of any event that could
    affect the flag's state.

    Ideally, the logic for setting each variable would be triggered each
    time an event occurs that could affect it:  A publish-subscribe type
    of mechanism.  One possibility to do this efficiently?  Use a bit-mask
    "array" to determine which functions to call for each register that is
    changed.  That way it doesn't take many bytes to determine which functions
    to call, and each function can update its corresponding register value
    based on its current state and the event that just occured.

    Or should I simply combine such values in to one (a struct, if necessary),
    and have a single function that is called for all the related values
    and updates the state of the single logical entity based on its
    current state and what event called it?  (But the # of combinations could
    get out of hand easily...)
*/

#ifdef REG4

REG4( timeZone_,       x, 40001, S16, RW, MB_timeZone )
REG4( curTime,         x, 40002, U32, RW, setU32 )

REG4( ipAddr,          x, 40004, U32, RW, setU32 )
REG4( netmask,         x, 40006, U32, RW, setU32 )
REG4( gateway,         x, 40008, U32, RW, setU32 )

REG4( setACoil,        _, 40010, U16, RW, setU16 )
REG4( resetACoil,      _, 40011, U16, RW, setU16 )

REG4( digOutBits,      x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,      x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,     x, 40016, U32, RW, MB_cmdBitsAppl )

REG4( ampIgnoreMask,   x, 40018, U32, RW, setU32 )
REG4( faultIgnoreMask, x, 40020, U32, RW, setU32 )

REG4( ignoreAmp,       x, 40022, U16, RW, MB_ignoreAmp )
REG4( useAmp,          x, 40023, U16, RW, MB_useAmp )

REG4( ignoreFault,     x, 40024, U16, RW, MB_ignoreFault)
REG4( useFault,        x, 40025, U16, RW, MB_useFault)

REG4( resetOnDelay,    x, 40026, U16, RW, setU32)


#undef REG4
#endif

//--- values for cmdBitsAppl ---
#define  CBA_CLR_INTLK    0x00000001

//--- values for ampIgnoreMask ---
#define  IGNORE_AMP_4     0x00000001
#define  IGNORE_AMP_8     0x00000002
#define  IGNORE_AMP_3     0x00000004
#define  IGNORE_AMP_7     0x00000008
#define  IGNORE_AMP_6     0x00000010
#define  IGNORE_AMP_5     0x00000020
#define  IGNORE_AMP_2     0x00000040
#define  IGNORE_AMP_1     0x00000080

//--- values for faultIgnoreMask ---
#define  IGNORE_FAULT_5   0x00000001
#define  IGNORE_FAULT_4   0x00000002
#define  IGNORE_FAULT_3   0x00000004
#define  IGNORE_FAULT_2   0x00000008
#define  IGNORE_FAULT_1   0x00000010
#define  IGNORE_FAULT_6   0x00000020
#define  IGNORE_FAULT_7   0x00000040
#define  IGNORE_FAULT_8   0x00000080
#define  IGNORE_FAULT_9   0x00000100
#define  IGNORE_FAULT_10  0x00000200


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion




//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================
//APPLCODE - settings/values we want saved to/restored from files


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------

//RESTORE_FUNC(myVar);

const ParamData far  devDataParams[] = {
//APPLDATA
//                ID   ---value---       --RESTORE_FUNC--
  PERSISTENT_VAR(  1,  ampIgnoreMask,    NULL ),
  PERSISTENT_VAR(  2,  faultIgnoreMask,  NULL ),
  PERSISTENT_VAR(  3,  resetOnDelay,     NULL ),

//PERSISTENT_VAR( 98,  myVarName,     PVRF(myVar) ),
//                ID   ---value---  granularity  --RESTORE_FUNC--
//PERSISTENT_TBL( 99,  myArrayName, elemSize,     NULL ),

  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//----------------------------------------------------

//RESTORE_FUNC(ampSet);

const ParamData far  settingsParams[] = {
//APPLDATA
//
//                ID   ---value---  --restoreFunc--
//  PERSISTENT_VAR(  1,  ampSet,     PVRF(ampSet) ),
//  PERSISTENT_VAR(  2,  rfOnSetpoint,    NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --RESTORE_FUNC--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Declarations for functions that handle setting "Next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

//void onRequest(void);
//void offRequest(void);
//void resetRequest(void);

//void newSetpoint(float newAmps);


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//APPLCODE
//--- add definitions for application-specific global variables here ---



//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//APPLCODE - add functions to process writes to ModBus register values

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
  ulong  cmdBits = MB_VAL(U32);


  if (cmdBits & CBA_CLR_INTLK)  Next.resetRequest = true;

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "ignoreAmp"
//
//  This register provides a simple way to set a bit in the ampIgnoreMask
//  value.  As the value is not saved, this is essentially a write-only
//  register.
//-----------------------------------------------------------------------------
bool _nearcall MB_ignoreAmp(SetRegFuncParams)
{
  uint  ampIdx = MB_VAL(U16) - 1;

  if (ampIdx >= NumAmps)  return false;

  SetBits(ampIgnoreMask, 1L << AmpBitsIndex[ampIdx]);

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "useAmp"
//
//  This register provides a simple way to clear a bit in the ampIgnoreMask
//  value.  As the value is not saved, this is essentially a write-only
//  register.
//-----------------------------------------------------------------------------
bool _nearcall MB_useAmp(SetRegFuncParams)
{
  uint  ampIdx = MB_VAL(U16) - 1;

  if (ampIdx >= NumAmps)  return false;

  ClearBits(ampIgnoreMask, 1L << AmpBitsIndex[ampIdx]);

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "ignoreFault"
//
//  This register provides a simple way to set a bit in the faultIgnoreMask
//  value.  As the value is not saved, this is essentially a write-only
//  register.
//-----------------------------------------------------------------------------
bool _nearcall MB_ignoreFault(SetRegFuncParams)
{
  uint  faultIdx = MB_VAL(U16) - 1;

  if (faultIdx >= NumFaults)  return false;

  SetBits(faultIgnoreMask, 1L << FaultBitsIndex[faultIdx]);

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "useFault"
//
//  This register provides a simple way to clear a bit in the faultIgnoreMask
//  value.  As the value is not saved, this is essentially a write-only
//  register.
//-----------------------------------------------------------------------------
bool _nearcall MB_useFault(SetRegFuncParams)
{
  uint  faultIdx = MB_VAL(U16) - 1;

  if (faultIdx >= NumFaults)  return false;

  ClearBits(faultIgnoreMask, 1L << FaultBitsIndex[faultIdx]);

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
/*EXAMPLE
bool MB_ampSet(SetRegFuncParams)
{
  newSetpoint(MB_VAL(float));

  return true;
}
*/


//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

/*
//-----------------------------------------------------------------------------
//  one of the setpoint values needs restoring from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(myVal)
{
  uint  u;
  float  fval;
  char  temp[32];

  for (u=0; u<NumAmps; ++u)  if (curVal == (void far *) &ISET(u))  {
    fval = *(float far *)newVal;
    printf("restoring iSet_%u to: %s\r\n", u, fmtFloat(fval, 8,2, temp));
    newSetpoint(u, fval);
  }

}

//-----------------------------------------------------------------------------
//  restore an ampSet value from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ampSet)
{
  newSetpoint(NEW_VAL(float));
}


*/

//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>


//=============================================================================
//  Misc Application-specific functions
//=============================================================================
//APPLCODE



//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
//APPLCODE - text and formatting info for application state summary

/*EXAMPLE:
//-----------------------------------------------------------------------------
_nearcall char *fmtMode(DispElem far *elem, char *buf)
{
  return (char *) ModeDesc[ELEM_VAL(U16)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtHdwCfg(DispElem far *elem, char *buf)
{
  return (char *) HdwCfgDesc[ELEM_VAL(U16)];
}
*/

//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"    ======== Amplifiers ========    == Temp Faults ==   \r\n"
"      OK   TEMP   Avail  Ignore      Input    Ignore \r\n"
"    ----- ------  -----  -------    --------  ------- \r\n"
" 1.  ###    ###    ###     ###       1. ###     ### \r\n"
" 2.  ###    ###    ###     ###       2. ###     ### \r\n"
" 3.  ###    ###    ###     ###       3. ###     ### \r\n"
" 4.  ###    ###    ###     ###       4. ###     ### \r\n"
" 5.  ###    ###    ###     ###       5. ###     ### \r\n"
" 6.  ###    ###    ###     ###       6. ###     ### \r\n"
" 7.  ###    ###    ###     ###       7. ###     ### \r\n"
" 8.  ###    ###    ###     ###       8. ###     ### \r\n"
"                                     9. ###     ### \r\n"
" RF: ########                       10. ###     ### \r\n"
"\r\n"
};

const DispElem  far dispElems[] = {
/*EXAMPLE:
  { 17, 3, VFPV(cavityVolts),     DARK_BLUE, fmtF32,    6,2 },
  { 17, 4, VFPV(activeSetpoint),  DARK_BLUE, fmtF32,    6,2 },
  { 16, 5, VFPV(regulationError), DARK_BLUE, fmtF32,    7,2 },

  { 17, 7, VFPV(limiterVolts),    DARK_BLUE, fmtF32,    7,3 },

  { 17, 9, VFPV(rfOffSetpoint),   DARK_BLUE, fmtF32,    6,2 },
  { 17,10, VFPV(rfOnSetpoint),    DARK_BLUE, fmtF32,    6,2 },

  { 43,10, VFPV(rampMode),        DARK_BLUE, fmtMode,   4,4 },

  { 64, 3, VFPV(hdwCfg),          DARK_BLUE, fmtHdwCfg, 5,5 },
*/
  { 0,0, NULL, NULL, NULL, 0,0 }
};

//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
//{ xx,yy,  BIT_MASK,     "0STR",ATTR_NAME,    "1STR",ATTR_NAME },
  {  5, 6,  PS1_OK,       "BAD",RED,           "OK ",BLUE        },
  { 12, 6,  PS1_TEMP_OK,  "BAD",RED,           "OK ",BLUE        },
  { 19, 6,  PS1_PRESENT,  "NO ",RED,           "YES",BLUE        },

  {  5, 7,  PS2_OK,       "BAD",RED,           "OK ",BLUE        },
  { 12, 7,  PS2_TEMP_OK,  "BAD",RED,           "OK ",BLUE        },
  { 19, 7,  PS2_PRESENT,  "NO ",RED,           "YES",BLUE        },

  {  5, 8,  PS3_OK,       "BAD",RED,           "OK ",BLUE        },
  { 12, 8,  PS3_TEMP_OK,  "BAD",RED,           "OK ",BLUE        },
  { 19, 8,  PS3_PRESENT,  "NO ",RED,           "YES",BLUE        },

  {  5, 9,  PS4_OK,       "BAD",RED,           "OK ",BLUE        },
  { 12, 9,  PS4_TEMP_OK,  "BAD",RED,           "OK ",BLUE        },
  { 19, 9,  PS4_PRESENT,  "NO ",RED,           "YES",BLUE        },

  {  5,10,  PS5_OK,       "BAD",RED,           "OK ",BLUE        },
  { 12,10,  PS5_TEMP_OK,  "BAD",RED,           "OK ",BLUE        },
  { 19,10,  PS5_PRESENT,  "NO ",RED,           "YES",BLUE        },

  {  5,11,  PS6_OK,       "BAD",RED,           "OK ",BLUE        },
  { 12,11,  PS6_TEMP_OK,  "BAD",RED,           "OK ",BLUE        },
  { 19,11,  PS6_PRESENT,  "NO ",RED,           "YES",BLUE        },

  {  5,12,  PS7_OK,       "BAD",RED,           "OK ",BLUE        },
  { 12,12,  PS7_TEMP_OK,  "BAD",RED,           "OK ",BLUE        },
  { 19,12,  PS7_PRESENT,  "NO ",RED,           "YES",BLUE        },

  {  5,13,  PS8_OK,       "BAD",RED,           "OK ",BLUE        },
  { 12,13,  PS8_TEMP_OK,  "BAD",RED,           "OK ",BLUE        },
  { 19,13,  PS8_PRESENT,  "NO ",RED,           "YES",BLUE        },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far ampIgnoreElems[] = {
//{ xx,yy,  BIT_MASK,     "0STR",ATTR_NAME,    "1STR",ATTR_NAME },
  { 27, 6,  IGNORE_AMP_1, "NO ",BLUE,          "YES",RED        },
  { 27, 7,  IGNORE_AMP_2, "NO ",BLUE,          "YES",RED        },
  { 27, 8,  IGNORE_AMP_3, "NO ",BLUE,          "YES",RED        },
  { 27, 9,  IGNORE_AMP_4, "NO ",BLUE,          "YES",RED        },
  { 27,10,  IGNORE_AMP_5, "NO ",BLUE,          "YES",RED        },
  { 27,11,  IGNORE_AMP_6, "NO ",BLUE,          "YES",RED        },
  { 27,12,  IGNORE_AMP_7, "NO ",BLUE,          "YES",RED        },
  { 27,13,  IGNORE_AMP_8, "NO ",BLUE,          "YES",RED        },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};


const DigDispElem far extraInElems[] = {
//{ xx,yy,  BIT_MASK,     "0STR",ATTR_NAME,    "1STR",ATTR_NAME },
  { 40, 6,  TEMP_OK_1,    "BAD",RED,           "OK ",BLUE        },
  { 40, 7,  TEMP_OK_2,    "BAD",RED,           "OK ",BLUE        },
  { 40, 8,  TEMP_OK_3,    "BAD",RED,           "OK ",BLUE        },
  { 40, 9,  TEMP_OK_4,    "BAD",RED,           "OK ",BLUE        },
  { 40,10,  TEMP_OK_5,    "BAD",RED,           "OK ",BLUE        },
  { 40,11,  TEMP_OK_6,    "BAD",RED,           "OK ",BLUE        },
  { 40,12,  TEMP_OK_7,    "BAD",RED,           "OK ",BLUE        },
  { 40,13,  TEMP_OK_8,    "BAD",RED,           "OK ",BLUE        },
  { 40,14,  TEMP_OK_9,    "BAD",RED,           "OK ",BLUE        },
  { 40,15,  TEMP_OK_10,   "BAD",RED,           "OK ",BLUE        },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far faultIgnoreElems[] = {
//{ xx,yy,  BIT_MASK,        "0STR",ATTR_NAME,  "1STR",ATTR_NAME },
  { 48, 6,  IGNORE_FAULT_1,  "NO ",BLUE,        "YES",RED        },
  { 48, 7,  IGNORE_FAULT_2,  "NO ",BLUE,        "YES",RED        },
  { 48, 8,  IGNORE_FAULT_3,  "NO ",BLUE,        "YES",RED        },
  { 48, 9,  IGNORE_FAULT_4,  "NO ",BLUE,        "YES",RED        },
  { 48,10,  IGNORE_FAULT_5,  "NO ",BLUE,        "YES",RED        },
  { 48,11,  IGNORE_FAULT_6,  "NO ",BLUE,        "YES",RED        },
  { 48,12,  IGNORE_FAULT_7,  "NO ",BLUE,        "YES",RED        },
  { 48,13,  IGNORE_FAULT_8,  "NO ",BLUE,        "YES",RED        },
  { 48,14,  IGNORE_FAULT_9,  "NO ",BLUE,        "YES",RED        },
  { 48,15,  IGNORE_FAULT_10, "NO ",BLUE,        "YES",RED        },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};


const DigDispElem far digOutElems[] = {
//{ xx,yy,  BIT_MASK,     "0STR"  ,ATTR_NAME,  "1STR" ,ATTR_NAME },
  {  5,15,  RF_DISABLE,   "ENABLED ",BLUE,     "DISABLED",RED    },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};

//---------------------------------------------------------------
//---- older style - do not use for newer applications -----
const DigFmtInfo digInFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};


const AnalogFmtInfo  adcFmt[] = {
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(ampIgnoreMask, (DigDispElem far *)ampIgnoreElems, blinkOff, temp);
    displayDigElems(faultIgnoreMask, (DigDispElem far *)faultIgnoreElems, blinkOff, temp);

    displayDigElems(extraIn, (DigDispElem far *)extraInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}

//-----------------------------------------------------------------------------
//  Application-specific commands
//-----------------------------------------------------------------------------
void _task cmd(int argc, char **argv)
{
  bool  validCmd = false;
  int  targetNum, maxNum;
  const uint  *ignoreMaskIdx = NULL;
  ulong  *ignoreMask = NULL;
  ulong  mask;


  for (;;)  {
    if (argc < 2)  break;

    if (StriEq(argv[1], "reset"))  {
      Next.resetRequest = true;  validCmd = true;  break; }

    if (argc < 4)  break;

    if (StriEq(argv[1], "amp"))  {
      ignoreMask = &ampIgnoreMask;  ignoreMaskIdx = AmpBitsIndex;
      maxNum = NumAmps;  }
    else  if (StriEq(argv[1], "fault"))  {
      ignoreMask = &faultIgnoreMask;  ignoreMaskIdx = FaultBitsIndex;
      maxNum = NumFaults;  }
    else  break;

    targetNum = atoi(argv[2]);
    if ((targetNum < 1) OR (targetNum > maxNum))  {
      printf("Invalid %s #\r\n", argv[1]);  break; }

    mask = 1L << ignoreMaskIdx[targetNum - 1];

    if (StriEq(argv[3], "use"))  *ignoreMask &= ~mask;
    else  if (StriEq(argv[3], "ignore"))  *ignoreMask |= mask;
    else  break;

    validCmd = true;  break;
  }


  if (! validCmd)  {
    printf("\r\n"
           "Usage:  cmd amp|fault # use|ignore  (enable/disable inputs)\r\n"
           "            reset                   (reset interlocks)\r\n" );
    Exit(-1);
  }


  Exit(0);
}


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/


//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "Cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "Next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "Cur" OR the "Next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "Cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "Next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "Cur" value for the "input", and the "Prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "Cur" values
//    represent the input values just read (what they are "now") and the
//    "Prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "Cur" values).
//
//    During a pass the output values in "Cur" represent what the outputs
//    where set to AFTER the "Cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "Next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessive latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

/*Examples
//-----------------------------------------------------------------------------
void onRequest(void)
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON request\n");
}

//-----------------------------------------------------------------------------
void offRequest(void)
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF request\n");
}

//-----------------------------------------------------------------------------
void resetRequest(void)
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET request\n");
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps)
{
  char  temp[16];


  KLOCK();

  if (FABS(newAmps) > maxAmpSet)  newAmps = (float)SIGN(newAmps) * maxAmpSet;

  Next.ampSet = newAmps;

  KUNLOCK();

  fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}

//-----------------------------------------------------------------------------
*/

/*
=========== NOTES on Locking ============

The applic task runs at the highest priority (even above the kernel), so it
cannot be preempted by any other task.  As long as this is true, the applic
task does not need to use locks when accessing global variables.

However, all other tasks MUST disable the scheduler when accessing values
used by the applic task to avoid being preempted while doing so.  Care must
be taken to keep these times short to avoid latency issues.

For REALLY short times, a critical section can be used as this is much more
efficient than disabling and reenabling the scheduler.  But critical sections
disable all interrupts, so it is not practical in most cases.

*/

//=============================================================================
//  Update the "Cur"rent state of the controller
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).  These values
//  are copied to Next as needed and constitute the Hardware Inputs.
//
//  Other tasks can affect the state of the control logic by changing values
//  in Next as well.  These values constitute Software Inputs.
//
//  This function is responsible for the Asynchronous logic that is applied
//  each pass.  It updates values in Next (or other global variables) based
//  on the Software and Hardware Input values in Next.  This portion of the
//  logic is analogous to unclocked gates in an electronic circuit.
//
//  After the Asychronous logic is applied, we set Prev = Cur and Cur = Next.
//  The result is that Prev contains the controller state as it was during the
//  previous pass, and Cur indicates the current state of the controller.
//
//  Also, see the "NOTES on Locking" section above....
//=============================================================================
void updateCurState(void)
{
  uint  u, readyCount, testBits;
  ulong  inBits, mask, ignoreMask, extraInChgs;


  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The analog and digital hardware inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "Cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "Next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

  Next.processTime = GetMS();  // curTime = time(NULL);

//  msSincePrev = msSince(Cur.processTime);

  Next.digitalIn = digitalIn;  Next.extraIn = extraIn;


//APPLCODE
  if (Next.resetRequest)  Next.resetStarted = GetMS_();



//--- Every 3 bits provides the state of 1 RF amplifier/PS ---
//--- Any amplifier that is not being ignored must be ready, and there must ---
//--- be at least one amplifier that is not ignored (i.e. ignoring them all ---
//--- is an error, even if one or more is ready) ---
  inBits = digitalIn;  Next.psError = false;  ignoreMask = ampIgnoreMask;
  readyCount = 0;
  testBits = Next.resetStarted ? PS_TEMP_OK | PS_PRESENT : PS_READY_BITS;
  for (u=0; u<NumAmps; ++u,inBits>>=3,ignoreMask>>=1)  {
    if ((uchar)ignoreMask & 1)  continue;
    if (((uchar)inBits & testBits) == testBits)  { ++readyCount;  continue; }
    Next.psError = true;  break;
  }
  if (!readyCount)  Next.psError = true;

//--- Separate from the amplifier/PS OK and Temp inputs are 10 temperature ---
//--- fault inputs, which generally do NOT have a 1-to-1 correspondence to ---
//--- the amplifiers (not suprising, since there are more of them than ---
//--- amplifiers, and each of the 10 temperature fault inputs is itself ---
//--- actually 3 daisy-chained inputs) ---
  inBits = extraIn;  Next.tempFault = false;  ignoreMask = faultIgnoreMask;
  for (u=0; u<NumFaults; ++u,inBits>>=1,ignoreMask>>=1)  {
    if ((uchar)ignoreMask & 1)  continue;
    if ((uchar)inBits & 1)  continue;
    Next.tempFault = true;  break;
  }

  Next.activeError = Next.psError OR Next.tempFault;



// Next we make a copy of the application state before we apply all the
// logic that updates the overall state.  This allows us to have a copy of
// all the relevant values that remain unchanged during the processing,
// avoiding all kinds of order-of-execution dependencies.

// We also make a copy of the state from last time before overwriting it
// so we can see if/how values have changed since the last pass

// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Use of pointers would also result in larger, slower code for every
// reference to Prev, Cur, and Next.  Overall, copying everything here is
// actually faster overall, simpler, and much safer.

  Prev = Cur;  Cur = Next;


//--- log msgs for bit chgs only if not starving other processes for cycles ---
  if (applLoopTime < 8)  {
    logDigIOChgs(false);
    if ((extraInChgs = (Cur.extraIn ^ Prev.extraIn)) != 0)  {
      for (u=0,mask=1; u<NumFaults; ++u,mask<<=1)  {
        if (!(extraInChgs & mask))  continue;
        fprintf(DiagLog, "Temp Fault %u %s\n",
                FaultBitsOrder[u], (extraIn & mask ? "CLEAR" : "ASSERTED") );
      }
    }
  }

}


//=============================================================================
//  Update the values in "Next" based on the "Cur" and "Prev" values.
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
//
//  This function contains the Synchronous logic used to compute the next
//  state we want the controller to be in.
//
//  A few simple rules that apply to all the values used by this function:
//
//    - All values changed within this function are "outputs"
//
//    - Every value read by this function is an "input"
//
//    - There must be no overlap between inputs and outputs
//
//    - The output values from this function must not be altered outside of
//      this function.
//
//  Any global variable can be an input OR an output, but they must NEVER
//  serve both purposes.  Most inputs come from Prev or Cur, while most
//  outputs are in Next.  If the logic requires detecting a change in a
//  value between passes (edge-triggered logic), then the value should be
//  part of the ApplState structure so we can compare its Prev and Cur values.
//
//  Following these simple rules allows the code to be organized as small,
//  self-contained blocks, with well-defined inputs and outputs.  Furthermore,
//  there are no order-of-execution dependencies between blocks of code.  The
//  blocks can be executed in any order:  The result will always be the same.
//
//
//  Each output value (or VERY small "set" of output values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative, data-oriented
//  style of programming and has many advantages including:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result.  Far simpler than debugging code structured around the
//       traditional process/event-driven approach which can contain numerous
//       independent paths of execution leading to an observed result.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The name(s) of the value(s) it modifies
//   - A descr of the meaning of the value(s) and the logic used
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

//APPLCODE


//--------------------------------------------------------------------
// Next.interlocked
// Next.resetRequest (reset-only)
//    Description:
//       An amplifier/PS or temperature fault occured that has not
//       yet been cleared.
//    Inputs:
//       Cur.interlocked
//       Cur.activeError
//       Cur.resetRequest
//--------------------------------------------------------------------
  if (Cur.activeError)
    if (!Cur.interlocked)  {
      printf("INTERLOCK ASSERTED\n");  Next.interlocked = true;  }

  if (Cur.resetRequest)  {
    if (Cur.activeError)
      printf("Error still asserted:  Reset request ignored\n");
    else  {
      if (Next.interlocked)
        Next.interlocked = false;
      else
        printf("No interlock to clear\n");
    }
    Next.resetRequest = false;
  }

//--------------------------------------------------------------------
// Next.resetStarted (resetOnly)
//    Description:
//       Sets or clears the start time for the interval during which
//       the On/Off (OK) bit for an amplifier will be ignored.
//    Inputs:
//       Cur.resetStarted
//       Cur.resetRequest
//--------------------------------------------------------------------
  if (Cur.resetStarted)  if (msSince_(Cur.resetStarted) > resetOnDelay)
    Next.resetStarted = 0;
}



//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
void updateOutputState(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

//APPLCODE

//--- update output values that are based on current conditions (values ---
//--- in "Cur" or any values used in a conditional in the applic logic) ---

// NOTE:  Do NOT clear all the bits of a bit-mapped value unless you are
//        going to check the state of every bit every time (i.e. clearing
//        every bit and then forgetting about some macro or piece of code
//        shared by multiple applications to manage a bit is an easy way
//        to mess things up - so do NOT touch bits that aren't specifically
//        handled by the function that clears them!

  if (Next.interlocked)
    SetBits(Next.digitalOut, RF_DISABLE);
  else
    ClearBits(Next.digitalOut, RF_DISABLE);

}




//=============================================================================
// The last thing we do is update values read by other tasks
//
//  Be sure to read the "NOTES on Locking" section above....
//=============================================================================
void updateExternals(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//APPLCODE

  extraInBits = extraIn;

}



//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0; }

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)  {
  *gain = 1.0;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  uint  ps, u;


  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
//--- Fixed size arrays are used so the values can be changed at run time ---
//--- (i.e. if pointers where used instead, and initialized to const string ---
//--- values, then we couldn't change them at run time) ---

//const uchar  LogicOrderDIn[NUM_APPL_DIN] =  {
//   21,22,23, 18,19,20, 6,7,8, 0,1,2, 15,16,17, 12,13,14, 9,10,11, 3,4,5 };

  for (ps=0; ps<NumAmps; ++ps)  {
    u = LogicOrderDIn[ps * 3];
    sprintf(dinLabel[u],    "PS%d_IS_OK",   ps+1);
    sprintf(dinLabel[u+1],  "PS%d_TEMP_OK", ps+1);
    sprintf(dinLabel[u+2],  "PS%d_PRESENT", ps+1);
  }

  fstrcpy(doutLabel[5], "Activty_LED");
  fstrcpy(doutLabel[7], "RF_DISABLE");


/*
  fstrcpy(ainLabel[0], "ai0-Descr");
  fstrcpy(ainLabel[1], "ai1-Descr");

  fstrcpy(aoutLabel[0], "ao0-Descr");
  fstrcpy(aoutLabel[1], "ao1-Descr");

  fstrcpy(dinLabel[0], "di0-Descr");
  fstrcpy(dinLabel[1], "di1-Descr");

  fstrcpy(doutLabel[0], "do0-Descr");
  fstrcpy(doutLabel[1], "do1-Descr");
*/
}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE
  resetOnDelay = 1000;  // delay after reset before OK input must be true
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

//APPLCODE


  Exit(0);
}


//=============================================================================
//  Use the values in prevLogInfo to determine if we should assert interlocks
//  for any devices any devices that were shut off due to a non-soft reset.
//=============================================================================
void checkPrevOnStates(void)
{

}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
//APPLCODE


  #define SharedApplLocals2
  #include <sharedAppl.c>

//--- add local array definitions here ---


//--- place code here to set initial default values ---
//  allowSoftReset = true;

  #define SharedApplInit
  #include <sharedAppl.c>

//--- At this point generic initialization has been done which includes ---
//--- the following: ---
//      - The Prev, Cur, and Next ApplState objects have been zeroed
//      -


//--- add application initialization code here ---
//APPLCODE


  #define SharedApplInit2
  #include <sharedAppl.c>


//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>


  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
//APPLCODE
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,       "..." },
  { "cmd",       cmd,         "Application-specific commands" },
  { "rio",       rio,         "Set/Read Rabbit RIO registers" },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
