//----- RFServoAppl.c ----- 11/30/12 -----(03/05/06)----

// see EE/Controls Bundle #208 for schematics, drawings, parts, etc

// Sections that typically require application-specific code are marked with
// the comment:  //APPLCODE

//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//APPLCODE
//  Application code to manage the state of a Bang-Bang or DC Servo motor that
//  is part of the RF system.
//-----------------------------------------------------------------------------
// UPDATE LOG (most recent first):
//
//
//  2012-11-20 - Mark Davis
//               - Finally back to porting this application to DEBROS platform.
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2007-03-14 - Mark Davis -
//               - added call to rwIO() (replaces "invisible" code in the
//                 sharedAppl.c file that did this for all applications).
//
//  2007-03-04 - Mark Davis -
//               - reworked X-Macro usage to allow application-specific
//                 register definitions to reside in the same file as the
//                 rest of the application-specific code.
//
//  2006-11-14 - Mark Davis -
//               - original version started based on ps8rack code
//
//-----------------------------------------------------------------------------


/*-----------------------------------------------------------------------------
Operational Notes:

- CARDINAL RULE:  Controllers DO NOT MOVE DEVICE if they get a position request
  from a REMOTE source that is the same as the LAST request from a remote
  source!  This is to INSURE the device DOES NOT MOVE when a crate is rebooted!
  It makes no difference if the device is not at the requested position!!!!
  WE DO NOT WANT A WORKING BEAM TO GET MESSED UP JUST BECAUSE A CRATE WAS
  REBOOTED!!!!!

- The Amplifier Enable/Disable switch must be in the Disable position when the
  module is powered up or down to avoid momentarily enabling the servo amp (and
  movement of the motor).  (A disable switch that disconnected the move bits -
  and set the +/- limit inputs high would be a handy addition and make it more
  like the other Servo Ctlrs).

- A Reference (error voltage) Calibration should be done whenever a new
  controller is set up (or swapped) in order to determine the minimum Reference
  (error) voltage necessary to reliably produce movement.  On the NIM module,
  this calibration can be started (or stopped) via the front panel by pressing
  both the Lock and Set buttons while in (panel) Local mode.

- Changes to CIB (Control Interface Board) on older DC Servo controllers:
  - cut or remove R19 (usually a 0-ohm "resistor" or wire) to eliminate the
    non-linear portion of the error amp's transfer function.  This eliminates
    massive chatering and oscillation.
  - Reducing the error amp gain by shorting R17 or R18 will reduce chatter
    and improve accuracy.  This effectively increases the resolution of the
    controller's output voltage and so is preferable to simply reducing the
    Gain setting on the Servo Amplifier or the PID loop parameters.
    (NOTE: Do NOT reduce CIB's error amp gain to less than 10!)
  - The controller checks the gain of the CIB error amp each time it is booted

- For devices with mechanical slop, the PID loop and the Servo Amp need to be
  tuned to eliminate as much overshoot as possible.  Slop means delays in
  changes to the calculated device position value which would cause a PID loop
  to oscillate endlessly (and the device position to overshoot back-and-forth
  endlessly), so the PID loop operates on the motor position, NOT the
  calculated device position.  This means that overshoot corrections are based
  on the motor position only and these corrections (i.e. reverse movement) are
  partly or wholely "absorbed" by the mechanical slop, resulting in little or
  no correction to the device's position (i.e. short of traversing the slop
  zone in reverse, there is no way to undo overshoot - and allowing this would
  cause it oscillate back and forth continuously, correcting overshoot back and
  forth, adding considerable wear to the drive mechanics and probably resulting
  in more slop - not to mention a device that never "settles").

- The Balance on the Servo Amplifier must be adjusted with a zero value
  reference/error voltage.  For the older DC Servos, this can be accomplished
  by simply switching them to LOCAL mode (which disconnects the controller's
  output to the amplifier within the controller box).  For NIM modules, the
  PID loop remains active and the result driving the amplifier, so you must
  use the LOCAL switch at the motor itself to disconnect the ref voltage.

- For the NIM module, the (+) and (-) Drive LED's both blink on and off if the
  device does not respond as expected (i.e. no movement when there should be,
  or movement in the wrong direction).

- The SET button controls what value the Meter output on the DB-25 connector
  has.  When the SET button is held closed, it will equal the current Set
  Point value (i.e. the position the device should be moving toward or the
  position the controller is trying to maintain).  When the SET button is
  not being held closed, it will be the current device position (NOT the
  same as the motor position if slop != 0)

- The DISABLE position of the ENABLE/DISABLE switch causes the controller
  to ignore new position request values.

- Pressing the Lock button will toggle the LOCK state of the controller.
  In the Locked state, the controller will change the motor position as
  needed to maintain a steady voltage on the Delta Phi input (the value
  maintained being the value of Delta Phi at the time the LOCK state was
  initiated).  If RF OK is lost for more than the specified amount of time,
  the device position will be reset to what it was when the lock was engaged
  (the "cold" position).  NOTE that the locked state is not disabled and
  active corrections will resume when RF OK is reasserted.  Also, manual
  change commands (via the knob or the switch) will alter the value of the
  Delta Phi input the controller is trying to maintain, allowing the RF to
  be tuned while in Phase Feedback Lock mode.


-------------------------------------------------------------------------------
Pin usage for DB-25 connector on back of NIM module:

(NOTE that some inputs are inverted before reaching the CPU, so the logic
 may be inverted from what is listed for the digitalIn bits)

 1 - Shield
 2 - DCS Drive        - AOut: +/-10V ref/error voltage to DC Servo Amp
 3 - DCS Drive Common - AOut: Analog Common to Servo Amp
 4 - +10V Ref         - AOut: To (+) side of motor position feedback pot
 5 - Analog Common    - AOut: To (-) side of motor position feedback pot
 6 - Follow Pot Wiper - AIn:  To middle (wiper) of motor position feedback pot
 7 - Analog Common    -
 8 - (+) Limit        - DIn:  Gnd <= OK, Open or +5V <= (+) Limit reached
 9 - (-) Limit        - DIn:  Gnd <= OK, Open or +5V <= (-) Limit reached
10 - +5.25VDC         - AOut: (use as needed for digital logic inputs)
11 - Loc/Rem          - DIn:  Gnd <= Remote, Open or +5V <= Local
12 - Loc/Rem Common   - DIn:  Dig Common for Local/Remote switch
13 - Amp Enable       - DOut: Gnd => Enable Amp, Open or +5 => Disable Amp

14 - Amp Fault        - DIn:  Gnd <= Amp Operating, +5V <= Problem/Disabled
15 - Amp Reset        - DOut: Gnd => Reset Amp Fault, +5V => for normal oper.
16 - -10V Ref         - AOut: For external +/- drive switch (if needed)
17 - RF OK            - DIn:  Gnd. <= RF NOT OK, +5V <= RF OK
18 - Digital Common   - DIn:  Digital Common (for RF OK if needed)
19 - Meter (+)        - AOut: current read or (when SET pressed) setpoint
20 - Meter (-)        - AOut: Analog Common for Meter
21 - BBS Drive (+)    - DOut: Gnd => no (+) drive, +5V => Move motor in (+) dir
22 - BBS Drive (-)    - DOut: Gnd => no (-) drive, +5V => Move motor in (-) dir
23 - BBS Jumper       - DIn:  Gnd for Bang-Bang Servo mode
24 - Digital Common   - DIn:  Digital Common for BBS/DCS jumper
25 - DCS Jumper       - DIn:  Gnd for DC Servo mode

-------------------------------------------------------------------------------
Key things to remember when making programming changes:

- The posReq and lastPosReq variables are used to keep track of values that
  came from an external command, and should NOT be altered except in response
  to an external command (or a process that wishes to be treated as an external
  client).

- motorPos is the current position of the motor - our only real readable
  feedback as to the state of the device.  The actual position of the device
  must be inferred from this value, the amount of slop we belive the device
  has, and knowledge about past movements (i.e. do we know, based on past
  movements, where the motor is relative to the "slop zone"?)

  Think of the motor vs device position like this:

               device position ---> #
                                    #
                            ==================
             slop zone ---> [   #            ] <--- slop zone
                                #
            motor position ---> #
              ===============================================

  Knowing the motor position is important because a PID loop based on the
  device position where the slop zone is non-zero would oscillate madly
  (since corrections for overshoot would keep increasing with no effect on
  device position until the motor position had worked its way back to the
  other edge of the slop zone, at which point the device position would
  suddenly change, and the PID loop would reverse its effort to compensate
  for the sudden spike, etc, etc).

  In short, PID loops are NOT designed to handle the situation where the
  coupling between the feedback and control values has a significant
  amount of hysteresis.

  So the device position is always inferred, and we have to keep track of
  several things in order to know what CHANGE in the motor position to make
  to move the device to a new position, figure out what the NEXT motor
  position should IDEALLY be to achieve the new device position, and then
  use the motor position in the feedback loop.

  Not actually knowing the device position means the end result is less
  accurate, as it depends on a value this is not hard and fast (the size
  of the slop zone) and correcting for overshoot is generally not practical.
  The best we can do is tune the PID loop to a compromise between overshoot
  and responsiveness.  Make it too responsive, and you end up with too
  much overshoot.  Reduce the overshoot too much, and movement becomes too
  "soft" and the accuracy and consistency of the resulting device position
  suffers due to even minor variations in the force needed to move the
  device a small amount relative to its current position and direction of
  movement.

- slopPos is the position of the center of the "slop zone" (the amount the
  motor can move without actually moving the position of the device).  Until
  the motor has moved from its startup position by an amount >= to the
  measured slop, this value can be off by as much as half the amount of slop.

- The algorithm currently makes the following assumptions:
   - Everything is symetrical and consistent (i.e. the energy or force needed
     to move the device a given amount is independent of its current position
     and the direction of movement).
   - There device has no inertia (i.e. it will move ONLY when the motor
     is exerting a force).

  To some degree, any actual inertia can be compenstated for in the PID loop
  tuning by reducing the responsiveness, but it is one more variable that
  will reduce the consistency of the end result.  But no means of countering
  any asymetries or irregularities is currently provided.

*/

//-----------------------------------------------------------------------------
//  For next version:
//
// - support for a specified movement rate?   The velocity-mode PID loop
//   should make that easy (???)
//
//-----------------------------------------------------------------------------
//  Things to test after changes:
//
//  - Movement and status via serial commands
//
//  - Movement and status via ModBus/TCP
//
//  - Movement possible in opposite dir when limit switch active
//
//  - New position cmd received while movement for prev cmd in progress
//
//  - No movement unless CtlWord == 3 (DCServo)
//
//  - No movement if duplicate position command received
//
//  - Duplicate pos cmd does not stop movement for prev cmd with same pos value
//
//  - Amplifier reference volt < min move ref volt when idle (DC/PID modes)
//
//  - Calibration commands produces reasonable values
//
//  - Calibration commands put motor back where it was when done
//
//  - stored values survive reboots and power loss
//
//  - amp disabled when device does not move as expected (e.g. moves the wrong
//    direction or doesn't move when it should)
//
//  - proper recovery from "hdwCtlErr" state
//
//  - rebooting after updating code should not move things
//
//  - local, move, remote, reboot crate should not result in movement (i.e the
//    crate must save position of device when switching from local back to
//    remote
//
//  - approp movement/lack thereof when doing combinations of the following:
//    - switching between local and remote
//    - rebooting the controller (must be in local mode or have amp disabled
//      to avoid momentary movement)
//    - rebooting the crate (with controller in local or remote mode)
//    - disconnecting and reconnecting cables
//    - motor manually moved while in local mode
//
//-----------------------------------------------------------------------------
//
//  Operational and Programming NOTES:
//
//  - The Host expects the most significant bits at lower addresses while the
//    ZWorld keeps the most significant bits at the higher address
//
//  - This code assumes that, when recieving values via ModBus TCP protocol,
//    any values that span multiple registers (such as floating point values)
//    will be recieved via a Write Multiple Registers command rather than
//    multiple Write Single Register commands (the latter can cause anomolous
//    behavior).
//
// - The code assumes that the position pot that provides a position reading
//   (as a voltage on the 1st Analog input channel) is actually measuring the
//   position of the motor, NOT the device being controlled.  Because the
//   linkage between the motor and the device being moved is never perfect,
//   this code now supports a "slop" factor, which is the amount that the motor
//   can move without actually moving the device.
//
//   The code currently assumes the slop is symetrical and that the device does
//   not exert any appreciable amount of force on the motor once the desired
//   position is reached (see discussions below for appropriate improvements
//   that could be made in the future).  The code keeps track of where it
//   believes the motor position is relative to the "slop" zone, and adjusts
//   the needed motor position to achieve the specified device position, and
//   adjusts the read back position according to where it believes the device
//   really is.
//
//   Note that, until the motor is actually moved in one direction by an amount
//   equal to the slop factor, the code cannot know the actual position of the
//   device to an accuracy greater than the amount of slop in the linkage.
//

/*
------------------------------------------------------------------------------

  Food for thought regarding devices with "slop" in them:

    actualPos() assumes that, if the motor position moves within the last
    known range of the slop zone, then the device itself did not move.
    But what if the new position of the motor is due to the device pushing
    it in that direction?  That is to say, if the motor isn't moving because
    we asked it to, then isn't it likely it is due to the device exerting a
    force on it?

    In short, if the motor position drifts when we are not exerting sufficient
    force to move it, then the drift is most likely due to the device pushing
    it, which means it must have drifted to the point where the motor is
    against one edge of the slop zone (and if that edge is the opposite one
    from what where we thought the motor was, then it drifted across the entire
    slop zone).  In any case, taking this in to account would result in more
    accurate device position information.

*/

//-----------------------------------------------------------------------------

#include <common.h>

#include <unistd.h>
#include <sys/types.h>


//APPLCODE - change to match the name of this file
#ifndef  ApplicSource
  #define  ApplicSource  "RFServoAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - set to the application name
const char far applName[] = { "RFServo" };
const char ApplAbbrev[] = { "RFServo" };

#define DeviceType  DT_RFServo

#define  NUM_APPL_AIN      4   // analog inputs
#define  NUM_APPL_AOUT     2   // analog outputs
#define  NUM_APPL_DIN     12   // digital inputs
#define  NUM_APPL_DOUT    11   // digital outputs

#define  ADC_MASK_APPL  0x0000000FL
#define  DAC_MASK_APPL  0x00000003L

#define InitialDigOutState  ( DISABLE_AMP | MOVE_BITS )



#define SharedINC
#include <sharedAppl.c>



#define  MaxStickTime  5000  // max # ms to push with no measurable result


bool  enablePhaseLock(bool remoteCmd);
bool  disablePhaseLock(bool remoteCmd);

bool  newPosReq(float posReq);

float  actualPos(float motor_pos);
float  neededPos(float desiredPos);
float  recentMovement(ushort msPeriod);

float  avgPos(ushort msStartBack, ushort msPeriod);

bool  updateNoiseData(void);

void  updateCtlValues(void);


//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//APPLCODE - define macros for each Digital bit and each Analog index

//=============== digitalIn bit usage ===============
#define  LOCK_BUTTON    0x0001  // 00: 0 = pushed
#define  LIMIT_SW_POS   0x0002  // 01: 1 = Normal, 0 = Limit Activated
#define  LIMIT_SW_NEG   0x0004  // 02: 1 = Normal, 0 = Limit Activated
#define  SET_BUTTON     0x0008  // 03: 0 = pushed

#define  AMP_DISABLED   0x0010  // 04: 0 = Amp Enabled, 1 = Amp Disabled (*)
#define  FAST_SW        0x0020  // 05: 0 = closed/slow
#define  LOC_REM_BTN    0x0040  // 06: 0 = closed (toggle Local/Remote)
#define  AMP_FAULT      0x0080  // 07: 1 = Amp Fault

#define  RF_OK          0x0100  // 08: 0 = RF OK
#define  BBS_JUMPER     0x0200  // 09: 0 ==> BB Servo device
#define  DCS_JUMPER     0x0400  // 10: 0 ==> DC Servo device
#define  EXT_LOC_REM    0x0800  // 11: Loc/Rem sw at motor (1 = Local for DC)

// (*) DI-04 reads state of FP Enable/Disable switch IFF the following is true:
//      - DO-03 is 0 (i.e. the software is trying to enable the amplifier)
//      - The Local/Remote switch at the motor control box is set to Remote

const uchar  LogicOrderDIn[NUM_APPL_DIN] =  {
     0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };

//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--          --1 state descr--
  "LOCK button pushed",      "LOCK button released",   //
  "[+] Limit Asserted",      "[+] Limit Cleared",      //
  "[-] Limit Asserted",      "[-] Limit Cleared",      //
  "SET button pushed",       "SET button released",    //

  "/ENABLE == 0",            "/ENABLE == 1",           //
  "Fast/Slow ==> SLOW",      "Fast/Slow ==> FAST",     //
  "Loc/Rem btn pushed",      "Loc/Rem btn released",   //
  "Ampl FAULT Cleared",      "Ampl FAULT Asserted",    //

  "RF OK",                   "RF BAD",                 //
  "BBS Jumper applied",      "BBS Jumper removed",     //
  "DCS Jumper applied",      "DCS Jumper removed",     //
  "Motor Loc/Rem == 0",      "Motor Loc/Rem = 1"       //
};


//=============== digitalOut bit usage ===============
#define  LOC_REM_LED      0x000001  // 00: 0=LED ON, 1=LED OFF
#define  DRIVE_MINUS_LED  0x000002  // 01: 0=LED ON, 1=LED OFF
#define  DRIVE_PLUS_LED   0x000004  // 02: 0=LED ON, 1=LED OFF
#define  DISABLE_AMP      0x000008  // 03: 0=Enable Amp (goes thru Enable Sw)

#define  LOCKED_LED       0x000010  // 04: 0=LED ON, 1=LED OFF
#define  FAULT_LED        0x000020  // 05: 0=LED ON, 1=LED OFF
#define  AMP_RESET_BIT    0x000040  // 06: 0=Reset Amp (DC Servo)
#define  MOVE_POS_BIT     0x000080  // 07: 0=assert movement in + dir (BBServo)

#define  MOVE_NEG_BIT     0x000100  // 08: 0=assert movement in - dir (BBServo)
#define  CPU_BUSY_LED     0x000200  // 09: 0=LED ON, 1=LED OFF
#define  ActivityLED      0x000400  // 10: activity LED

const uchar  LogicOrderDOut[NUM_APPL_DOUT] =  {
     0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 };

#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

#define  MOVE_BITS        (MOVE_POS_BIT | MOVE_NEG_BIT)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--        --1 state descr--
  NULL,                    NULL,                   // FP LOCAL LED
  "[-] movement started",  "[-] movement ended",   //
  "[+] movement started",  "[+] movement ended",   //
  "Soft Enable Asserted",  "Soft Enable Cleared",  //

  "Locked LED ON",         "Locked LED OFF",       //
  "Fault LED ON",          "Fault LED OFF",        //
  "Reset Amp Asserted",    "Reset Amp Cleared",    //
  "[+] Move Asserted",     "[+] Move disabled",    //

  "[-] Move Asserted",     "[-] Move disabled",    //
  "Busy LED ON",           "Busy LED OFF",         //
   NULL,                    NULL                   // Activity LED
};


const unsigned long LED_ErrPattern[6] =
     { LOCKED_LED, CPU_BUSY_LED,
       DRIVE_PLUS_LED, DRIVE_MINUS_LED,
       FAULT_LED, LOC_REM_LED };


//=============== Analog inputs ===============
//--- indexes for anaIn[] and readVal[] ---
#define  MOTOR_POS        0  // Readback for Motor Position
#define  RF_PHASE         1  // Phase Angle
#define  DRIVE_SW         2  // +/- Drive switch
#define  ADC_REF          3  // Ref voltage for ADC chip

const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0, 1, 2, 3 };


//=============== Analog outputs ===============
//--- indexes for anaOut[] and setVal[] ---
#define  CTL_VOLTS        0  // Ctl voltage to DC Servo Amp
#define  EXTERN_METER     1  // Voltage to external setpoint/position meter

const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0, 1 };


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

//===== values whose that can have different prev/cur or cur/next values

// Each value in the ApplState structure should be throught of as representing
// a "signal" in a clocked logic circuit, with the "applic" task representing
// the logic implemented by the circuit elements such as gates, latches, etc.

// The main "applic" task is the highest priority task and executes the code
// that implements the application logic on a real-time, periodic basis.

//                 ====== IMPORTANT NOTE on LOCKING ======
// Remember to use a semaphore or some other form of locking mechanism to
// protect all access to shared values by more than one task.  The only
// exception to this is when the applic task needs to access a value protected
// by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
// macros).  Because the applic task is the only task that runs at the highest
// priority, no other task can preempt it, and if the scheduler is locked,
// the applic task cannot preempt another task.  But scheduler locks must be
// kept short to avoid creating large latency in the timely execution of other
// tasks (the applic task in particular).


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------
  bool  onRequest;
  bool  offRequest;
  bool  resetRequest;

  bool  driveLock;

 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
  ulong  processTime;     // time (in ms) when processing was started
  ulong  digitalIn;       // state of digital inputs

  float  phaseAngle;

 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.  They must NOT be
 //  altered by other tasks (although they can read them as long as they
 //  use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------
  bool  localMode;
  bool  preMotorLocalMode;   // localMode value before motorLocal was true
  bool  fpEnableSw;          // what we know about the FP ENABLE/DISABLE sw

 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (such as latches).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance by the applic task AFTER
 //  all the other ApplState values are set.  Because they represent only
 //  the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  digitalOut;    // state for digital outputs

  bool  phaseLocked;
  bool  idleRest;

  int  ctlState;

} ApplState;


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState


const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );



//===== values that are ALWAYS treated as "cur" OR "next", but never both =====
//--- (values treated both ways must be in the ApplState structure) ---

  char  *runStateDesc = "READY   ";

  ulong  lastNoiseChkTime, lastNoiseChgTime, lastMoveCmdTime, lastEnableTime,
         lastTimeRFOK, lastCorrectingMsgTime, newPosTime, lastCalibGainTime,
         lastPhaseChgTime, msSincePrev, prevEnableMsgTime;

  long  knobTicks;

  float  lastPosReq, reqMovement, lastAvgPos, newPos, phaseErr, driveSwVolts,
         savedNoise;

  bool  slopPosKnown, movePlusSw, moveMinusSw, motorLocal, cancelCalib,
        plusLimit, minusLimit, ampFault, hdwCtlErr, fastAdjust,
        DCServo, BBServo, bbMoveMinus, bbMovePlus, ampWasEnabled,
        localChgLastPass, rfOK, testingCIBGain, useNewPos,  setBtnPushed,
        trackOverShoot, lockButton, setButton, softDisabled, hardDisabled,
        gainComputed, oscillating, manualMoveCmd, ignoreSettings,
        reducePhaseErr, correctingPos, newCorrectingMsg, reachedGoal,
        adjustPlusMargin, adjustMinusMargin, phaseLocked, newPhaseLockReq,
        phaseLockReqIgnored, restoreSavedPhaseLock, phaseLockActive,
         phaseLockPaused, phaseLockInactive, newPhaseSetting, atALimit,
        ctlrLocal, chgdToLocal, chgdToRemote,
        processedLockAndSet, lockBtnPushed, locRemBtnPushed;

  uint  lockPressedTime = 0;
  bool  newPhaseLockReq, newPhaseSetting, phaseLockReqIgnored;

  bool  softDisabled, hardDisabled, manualMoveCmd, ignoreSettings;

  int  ctlMode, calibState;

  float  showMovement;  // updated ONLY by showState()!


#define HistorySize  500   // five second's worth of data (at 0.01 sec res)
  uint  motorHistory[HistorySize];  // recent motor position history
  int  lastHistIdx;


const char *ModeDesc[] = { "*UNKNOWN*", "DC Servo ", "BB Servo ", "PID Servo" };
//----- values for ctlMode -----
#define  CM_Unknown          0
#define  CM_AnalogPosition   1   // PID loop mgd ctl voltage
#define  CM_DigDirection     2   // BB Servo digital +/- move cmds
#define  CM_AnalogVelocity   3   // PID loop mgd ctl voltage

#define  BBSCtlMode  (ctlMode == CM_DigDirection)

#define  RmtCmdsAllowed  (!Cur.localMode OR BBSCtlMode)


//----- values for calibState -----
#define  CALIB_IDLE               0

#define  CALIB_REF_INIT         100
#define  CALIB_REF_SETTLE       101
#define  CALIB_REF_START        102
#define  CALIB_REF_MOVE         103
#define  CALIB_REF_ABORT        190
#define  CALIB_REF_STOP         191

#define  CALIB_PID_START        400
#define  CALIB_PID_RUN          401
#define  CALIB_PID_STOP         490


//--- values for ApplState.fpEnableSw ---
#define  FP_ENABLE_UNKNOWN       0
#define  FP_ENABLE               1
#define  FP_DISABLE              2



//----- values for ApplState.ctlState -----
#define  CS_OVERRIDE      -1  // temporarily disables smart positioning code
#define  CS_IDLE           0
#define  CS_STEADY_REF     1
#define  CS_END_LOCAL_CHG  2
#define  CS_START_MOVE     3
#define  CS_ACTIVE_CTL     4



//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//APPLCODE - add tasks that service network clients

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ----------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);





//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


//APPLCODE - set # of devices for which we want duplicate register sets
#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of coils
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read-only registers
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

//APPLCODE - add registers for externally available read-only values

REG3( latency,        x, 30023, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30024, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30025, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime, x, 30026, U16 )  // # ms used by last kernel pass


#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read/write registers
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

#ifdef REG4

REG4( motorPos,         x, 40001, F32, RO, setF32 ) // pos of motor driving device (via ADC0)
REG4( motorPosB,        x, 40003, F32, RO, roReg  ) // pos of motor (via ADC1)

REG4( posSet,           x, 40005, F32, RO, roReg  ) // cur pos value to reach/maintain
REG4( posReq,           x, 40007, F32, RW, setF32 ) // last new pos requested

REG4( devStatus,        x, 40009, U16, RO, roReg  ) // bit-mapped status bits

REG4( ctlWord,          x, 40010, U16, RW, setU16 ) // switch for client chgs to posReq

REG4( versDate_,        x, 40011, U32, RO, roReg  ) // mirror of versDate
REG4( devTyp_,          x, 40013, U16, RO, roReg  ) // mirror of versDate

REG4( refVolt,          x, 40014, F32, RO, roReg  ) // voltage actually controlling Servo Amp

REG4( enableAmp,        x, 40016, U16, RO, setU16 ) // soft enable/disable Servo Amp

REG4( slopValue,        x, 40017, F32, RW, setF32 ) // amount of mechanical slop in device
REG4( oldSampleRate,    x, 40019, U16, RO, roReg  ) // SampleRate moved to read-only regs

REG4( realPos,          x, 40020, F32, RO, roReg  ) // pos of device as opposed to motor
REG4( slopPos,          x, 40022, F32, RO, roReg  ) // same as realPos after move > slopValue

REG4( pWeight,          x, 40024, F32, RW, setF32 ) // weighting of Proportional factor
REG4( iWeight,          x, 40026, F32, RW, setF32 ) // weighting of Integral factor
REG4( dWeight,          x, 40028, F32, RW, setF32 ) // weighting of Differential factor

REG4( pFactor,          x, 40030, F32, RO, roReg  ) // current value of Proportional factor
REG4( iFactor,          x, 40032, F32, RO, roReg  ) // current value of Integral factor
REG4( dFactor,          x, 40034, F32, RO, roReg  ) // current value of Differential factor

REG4( knobRateF,        x, 40036, F32, RW, setF32 ) // scaling for fast knob adjustments
REG4( knobRateS,        x, 40038, F32, RW, setF32 ) // scaling for slow knob adjustments

REG4( testAmpl,         x, 40040, F32, RW, setF32 ) // amplitude of test-mode square wave
REG4( testOffset,       x, 40042, F32, RW, setF32 ) // lowest voltage of test-mode sq. wave
REG4( testPeriod,       x, 40044, U32, RW, setU32 ) // period of test-mode sq. wave in ms

REG4( activeNoise,      x, 40046, F32, RW, setF32 ) // normal pos noise during active ctl
REG4( minMoveRefVolt,   x, 40048, F32, RW, setF32 ) // min ref volt required for movement

REG4( ctlVolts,         x, 40050, F32, RO, roReg  ) // voltage to Interface Board or Amp
REG4( unused_40052,     x, 40052, F32, RW, setF32 ) //
REG4( setRefVolt,       x, 40054, F32, RO, roReg  ) // desired refVolt value
REG4( motorPosSet,      x, 40056, F32, RO, roReg  ) // motor pos needed for posSet
REG4( overShoot,        x, 40058, F32, RO, roReg  ) // amount of overshoot during last move
REG4( setRate,          x, 40060, F32, RW, setF32 ) // scale factor for manual pos chgs
REG4( switchRate,       x, 40062, F32, RW, setF32 ) // scale factor for +/- switch
REG4( unused_40064,     x, 40064, F32, RW, setF32 ) //
REG4( maxActiveSecs,    x, 40066, U16, RW, setU16 ) // period of active ctl for new setpoint

REG4( timeZone_,        x, 40067, S16, RW, MB_timeZone )
REG4( curTime,          x, 40068, U32, RW, setU32 )
REG4( secCounter_,      x, 40070, U32, RO, roReg  ) // mirror of secCounter
REG4( dlmVersn_,        x, 40072, U16, RO, roReg  ) // mirror of dlmVers
REG4( comRate_,         x, 40073, U16, RO, roReg  ) // mirror of comRate
REG4( device_ID_,       x, 40074, U32, RO, roReg  ) // mirror of deviceID
REG4( coils31_00,       x, 40076, U32, RW, setU32 ) // alt access to coils
REG4( coils63_32,       x, 40078, U32, RW, setU32 ) // alt access to coils

REG4( rfLockResetDelay, x, 40080, U16, RW, setU16 ) // max ms to maintain Phase Lock w/o RF
REG4( maxPhaseErr,      x, 40081, F32, RW, setF32 ) // phaseLock deadband

REG4( plusMargin,       x, 40083, F32, RW, setF32 ) // err below which (+) BB move cmds stop
REG4( minusMargin,      x, 40085, F32, RW, setF32 ) // err below which (-) BB move cmds stop

REG4( phaseAngle,       x, 40087, F32, RO, roReg  ) // input on front of NIM module

REG4( ignoreHdwErrs,    x, 40089, U16, RW, setU16 ) // ignore hdw errs if set

REG4( desiredPhase,     x, 40090, F32, RW, setF32 ) // desired Phase value for PhaseLock mode

REG4( digOutBits,       x, 40092, U32, RW, setU32 )

REG4( ipAddr,           x, 40094, U32, RW, setU32 )
REG4( netmask,          x, 40096, U32, RW, setU32 )
REG4( gateway,          x, 40098, U32, RW, setU32 )


//REG4( setACoil,         _, 40010, U16, RW, setU16 )
//REG4( resetACoil,       _, 40011, U16, RW, setU16 )

//REG4( digOutBits,       x, 40012, U32, RW, setU32 )

//REG4( cmdBitsStd,       x, 40014, U32, RW, MB_cmdBitsStd )
//REG4( cmdBitsAppl,      x, 40016, U32, RW, MB_cmdBitsAppl )



//APPLCODE - add registers for externally available read/write values

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)


//--- bit masks for devStatus [40009#U16] ---
#define  DS_DISABLE       0x0001  // Amplifier is Disabled
#define  DS_FAULT         0x0002  // Fault signal from amplifier
#define  DS_LOCAL         0x0004  // Local mode (soft or hdw)
#define  DS_PLIMIT        0x0008  // Positive Limit switch activated

#define  DS_NLIMIT        0x0010  // Negative Limit switch activated
#define  DS_PHASE         0x0020  // PHASE-FEEDBACK-IN SIGNAL RECEIVED
#define  DS_CALIBRATE     0x0040  // calibration or test sequence running
#define  DS_HDW_ERROR     0x0080  // Problem detected with hardware

#define  DS_RF_OK         0x0100  // RF OK
#define  DS_BBS_JUMPER    0x0200  // BB Servo device
#define  DS_DCS_JUMPER    0x0400  // DC Servo device
#define  DS_EXTERN_LOCAL  0x0800  // Local override at the motor

#define  DS_PHASELOCKED   0x1000  // Locked to external reference
#define  DS_DRIVELOCK     0x2000  // Drive Lock enabled - NO movement allowed
#define  DS_IDLE          0x4000  // device is "resting" (no active control)



#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion




//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================
//APPLCODE - settings/values we want saved to/restored from files


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------

const ParamData far  devDataParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
//PERSISTENT_VAR(  1,  myVarName,     NULL ),
//                ID   ---value---  granularity  --RESTORE_FUNC--
//PERSISTENT_TBL(  2   myArrayName, elemSize,     NULL ),

  PERSISTENT_VAR(  1,  slopValue,         NULL ),
  PERSISTENT_VAR(  2,  minMoveRefVolt,    NULL ),
  PERSISTENT_VAR(  3,  pWeight,           NULL ),
  PERSISTENT_VAR(  4,  iWeight,           NULL ),
  PERSISTENT_VAR(  5,  dWeight,           NULL ),
  PERSISTENT_VAR(  6,  knobRateF,         NULL ),
  PERSISTENT_VAR(  7,  knobRateS,         NULL ),
  PERSISTENT_VAR(  8,  maxActiveSecs,     NULL ),
  PERSISTENT_VAR(  9,  rfLockResetDelay,  NULL ),
  PERSISTENT_VAR( 10,  maxPhaseErr,       NULL ),


  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};

//----------------------------------------------------
  RESTORE_FUNC(savedNoise);

const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
//PERSISTENT_VAR(  1,  rfOffSetpoint,   NULL ),
//PERSISTENT_VAR(  2,  rfOnSetpoint,    NULL ),

  PERSISTENT_VAR(  1,  savedNoise,      PVRF(savedNoise) ),
  PERSISTENT_VAR(  2,  plusMargin,      NULL ),
  PERSISTENT_VAR(  3,  minusMargin,     NULL ),
  PERSISTENT_VAR(  4,  desiredPhase,    NULL ),
  PERSISTENT_VAR(  5,  phaseLocked,     NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --RESTORE_FUNC--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Definitions for event msgs.  Event msgs are the means that other processes
//  use to inform the application process of an event that it needs to handle.
//
//  In general, allowing multiple processes to concurrently modify the state
//  of the controller is a sure recipie for chaos.  By restricting all such
//  changes to the context of a single process, and sending messages to that
//  process to request changes, we eliminate all such concurrency issues (not
//  the least of which is keeping track of the order of events).
//
//  The most common event is a request for change of state that corresponds to
//  a message received from an external client via a protocol like ModBus.
//=============================================================================

//EVENT_DEF( myEventX )
//EVENT_DEF( myEventY )



//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>


//APPLCODE
//--- add definitions for application-specific global variables here ---


//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//APPLCODE - add functions to process writes to ModBus register values

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
/*
  ulong  cmdBits;
  EventData  eventMsg;  // used by EVENT macro


  cmdBits = *((ulong *)newVal);

  if (cmdBits & x)  EVENT(MyEventX, chan, newVal, sizeof(int)); // w/ data
  if (cmdBits & y)  EVENT(MyEVentY, chan, NULL, 0);             // w/o data
*/

  return true;
}



//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>


//=============================================================================
//  Misc Application-specific functions
//=============================================================================
//APPLCODE


//-----------------------------------------------------------------------------
//  some portion of the analog calibrations were restored from a file
//-----------------------------------------------------------------------------
RESTORE_FUNC(savedNoise)
{
  KLOCK();  activeNoise = savedNoise = NEW_VAL(float);  KUNLOCK();
}


//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//-----------------------------------------------------------------------------
//  return ptr to a string describing the current contol state
//-----------------------------------------------------------------------------
char *ctlStateDesc(char **prefix)
{
  char *stateStr, *colorStr;


  stateStr = NULL;  colorStr = RedText;

  if (hdwCtlErr)       stateStr = "HDCTLERR";  else
//if (testingCIBGain)  stateStr = "CIB-GAIN";  else
  if (calibState)      stateStr = "CALIBRAT";  else
  if (Cur.driveLock)   stateStr = "DRVLOCKD";  else
  if (phaseLocked)     stateStr = "PHA_LOCK";  else
//if (oscillating)     stateStr = "OSCLTERR";  else
  if ((Cur.fpEnableSw != FP_ENABLE) OR ampFault)
    stateStr = (hardDisabled ? "HARD_DIS" : "SOFT_DIS");  else
  if (Cur.idleRest)    stateStr = "IDLEREST";

  if (stateStr)  {
    if (prefix)  *prefix = colorStr;
    return stateStr;
  }


//--- no higher priority msg to show, so just show the normal state ---
  colorStr = BlueText;
  switch (Cur.ctlState)  {
    case CS_IDLE:        stateStr = "IDLE    ";  break;
    case CS_STEADY_REF:  stateStr = "FIXEDREF";  break;
    case CS_START_MOVE:  stateStr = "NEW POS ";  break;
    case CS_ACTIVE_CTL:  stateStr = "ACTIVE  ";  break;
    default:             stateStr = "UNKNOWN ";  colorStr = RedText;
  }

  if (prefix)  *prefix = colorStr;
  return stateStr;
}


//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
//APPLCODE - text and formatting info for application state summary

/*EXAMPLE*
//-----------------------------------------------------------------------------
_nearcall char *fmtHdwCfg(DispElem far *elem, char *buf)
{
  return (char *) HdwCfgDesc[ELEM_VAL(U16)];
}
*/

//-----------------------------------------------------------------------------
_nearcall char *fmtEnableSw(DispElem far *elem, char *buf)
{
  switch (Cur.fpEnableSw)  {
    case FP_ENABLE:   return BlueText"ENABLE ";
    case FP_DISABLE:  return  RedText"DISABLE";
    default:          return  RedText"UNKNOWN";
  }
}

//-----------------------------------------------------------------------------
_nearcall char *fmtAmpFault(DispElem far *elem, char *buf)
{
// ampFault set based on both limits for BBServo
//  if (BBSCtlMode)  return  BlueText"N/A";

  return ampFault ? RedText"YES" : BlueText"NO ";
}

//-----------------------------------------------------------------------------
_nearcall char *fmtMotorLocal(DispElem far *elem, char *buf)
{
  return motorLocal ? RedText"LOCAL " : BlueText"REMOTE";
}

//-----------------------------------------------------------------------------
_nearcall char *fmtCtlMode(DispElem far *elem, char *buf)
{
  strcpy(buf, ctlMode ? BlueText : RedText);
  strcat(buf, ModeDesc[ctlMode]);

  return buf;
}

//-----------------------------------------------------------------------------
_nearcall char *fmtRunState(DispElem far *elem, char *buf)
{
  strcpy(buf, calibState ? RedText : BlueText);
  strcat(buf, runStateDesc);

  return buf;
}

//-----------------------------------------------------------------------------
_nearcall char *fmtLocRem(DispElem far *elem, char *buf)
{
  if (!Cur.localMode)  return BlueText"REMOTE";

  strcpy(buf, RedText);
  strcat(buf, motorLocal ? "MTRLOC" : "FP_LOC");

  return buf;
}

//-----------------------------------------------------------------------------
_nearcall char *fmtCtlState(DispElem far *elem, char *buf)
{
  char  *prefix, *desc;

  desc = ctlStateDesc(&prefix);

  strcpy(buf, prefix);  strcat(buf, desc);  return buf;
}

//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
"Lock Button:    ######      Local LED:     ###      Motor Pos:  -####.## \r\n"
"Positive Limit: ###         Drive (-) LED: ###      Phase Angle: -###.## \r\n"
"Negative Limit: ###         Drive (+) LED: ###      Drive Sw:     -##.## \r\n"
"Set Button:     ######      Disable Amp:   ### \r\n"
"FP Enable Sw:   #######     Locked LED:    ###      Ctl Output:   -##.## \r\n"
"Fast/Slow Sw:   ####        Fault LED:     ###      Meter Out:  -####.## \r\n"
"Loc/Rem Button: ######      Reset Ampl:    ### \r\n"
"Ampl. Fault:    ###         BBS Move (+):  ###      P: (##.##)  ###.#### \r\n"
"RF OK:          ###         BBS Move (-):  ###      I: (##.##)  ###.#### \r\n"
"BBS Jumper:     ###         CPU Busy LED:  ###      D: (##.##)  ###.#### \r\n"
"DCS Jumper:     ###         Activity LED:  ### \r\n"
"Motor Local Sw: ######                              Dev. Slop:  -####.## \r\n"
"                            MinMoveRef:  ######     Req. Pos:   -####.## \r\n"
"                            Pos. Noise:-####.##     Need Pos:   -####.## \r\n"
"                                                    Dev. Pos:   -####.## \r\n"
"Control Mode: [#########]   Overshoot:   ###### \r\n"
"[########][######][########]                        Movement:   -####.## \r\n"
"\r\n"
};

const DispElem  far dispElems[] = {
  { 16, 7, VFPV(Cur.fpEnableSw),         "", fmtEnableSw,   7,7 },
  { 16,10, VFPV(ampFault),               "", fmtAmpFault,   3,3 },
  { 16,14, VFPV(motorLocal),             "", fmtMotorLocal, 6,6 },

  { 15,18, VFPV(ctlMode),                "", fmtCtlMode,    9,9 },

  {  1,19, VFPV(calibState),             "", fmtRunState,   8,8 },
  { 11,19, VFPV(Cur.localMode),          "", fmtLocRem,     6,6 },
  { 19,19, VFPV(Cur.ctlState),           "", fmtCtlState,   8,8 },

  { 64, 3, VFPV(motorPos),              DARK_BLUE, fmtF32,  8,2 },
  { 65, 4, VFPV(Cur.phaseAngle),        DARK_BLUE, fmtF32,  7,2 },
  { 66, 5, VFPV(driveSwVolts),          DARK_BLUE, fmtF32,  6,2 },

  { 66, 7, VFPV(ctlVolts),              DARK_BLUE, fmtF32,  6,2 },
  { 64, 8, VFPV(setVal[EXTERN_METER]),  DARK_BLUE, fmtF32,  8,2 },

  { 39,16, VFPV(activeNoise),           DARK_BLUE, fmtF32,  8,2 },

  { 56,10, VFPV(pWeight),               DARK_BLUE, fmtF32,  5,2 },
  { 56,11, VFPV(iWeight),               DARK_BLUE, fmtF32,  5,2 },
  { 56,12, VFPV(dWeight),               DARK_BLUE, fmtF32,  5,2 },

  { 64,10, VFPV(pFactor),               DARK_BLUE, fmtF32,  8,4 },
  { 64,11, VFPV(iFactor),               DARK_BLUE, fmtF32,  8,4 },
  { 64,12, VFPV(dFactor),               DARK_BLUE, fmtF32,  8,4 },

  { 64,14, VFPV(slopValue),             DARK_BLUE, fmtF32,  8,2 },
  { 64,15, VFPV(posReq),                DARK_BLUE, fmtF32,  8,2 },
  { 64,16, VFPV(posSet),                DARK_BLUE, fmtF32,  8,2 },
  { 64,17, VFPV(realPos),               DARK_BLUE, fmtF32,  8,2 },

  { 64,19, VFPV(showMovement),          DARK_BLUE, fmtF32,  8,2 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};

//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
//{ xx,yy,  BIT_MASK,     "0STR"  ,ATTR_NAME,  "1STR"  ,ATTR_NAME },
//  --,--,  -------------  ------  ---------   -------  ---------
  { 16, 3,  LOCK_BUTTON,  "PUSHED",DARK_RED,   "OPEN  ",DARK_BLUE },
  { 16, 4,  LIMIT_SW_POS, "YES"   ,DARK_RED,   "NO "   ,DARK_BLUE },
  { 16, 5,  LIMIT_SW_NEG, "YES"   ,DARK_RED,   "NO "   ,DARK_BLUE },
  { 16, 6,  SET_BUTTON,   "PUSHED",DARK_RED,   "OPEN  ",DARK_BLUE },
//{ 16, 7,  AMP_DISABLED, "YES"   ,DARK_BLUE,  "NO "   ,DARK_RED  },
  { 16, 8,  FAST_SW,      "SLOW"  ,DARK_BLUE,  "FAST"  ,DARK_BLUE },
  { 16, 9,  LOC_REM_BTN,  "PUSHED",DARK_RED,   "OPEN  ",DARK_BLUE },
//{ 16,10,  AMP_FAULT,    "NO "   ,DARK_BLUE,  "YES"   ,DARK_RED  },
  { 16,11,  RF_OK,        "YES"   ,DARK_BLUE,  "NO "   ,DARK_RED  },
  { 16,12,  BBS_JUMPER,   "YES"   ,DARK_BLUE,  "NO "   ,DARK_BLUE },
  { 16,13,  DCS_JUMPER,   "YES"   ,DARK_BLUE,  "NO "   ,DARK_BLUE },
  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
//{ xx,yy,  BIT_MASK,        "0STR",ATTR_NAME,  "1STR",ATTR_NAME },
//  --,--,  ---------------  -----  ---------   -----  ---------
  { 43, 3,  LOC_REM_LED,     "ON " ,DARK_RED,   "OFF" ,DARK_BLUE },
  { 43, 4,  DRIVE_MINUS_LED, "ON " ,DARK_RED,   "OFF" ,DARK_BLUE },
  { 43, 5,  DRIVE_PLUS_LED,  "ON " ,DARK_RED,   "OFF" ,DARK_BLUE },
  { 43, 6,  DISABLE_AMP,     "NO " ,DARK_BLUE,  "YES" ,DARK_RED  },
  { 43, 7,  LOCKED_LED,      "ON " ,DARK_RED,   "OFF" ,DARK_BLUE },
  { 43, 8,  FAULT_LED,       "ON " ,DARK_RED,   "OFF" ,DARK_BLUE },
  { 43, 9,  AMP_RESET_BIT,   "YES" ,DARK_RED,   "NO " ,DARK_BLUE },
  { 43,10,  MOVE_POS_BIT,    "YES" ,DARK_RED,   "OFF" ,DARK_BLUE },
  { 43,11,  MOVE_NEG_BIT,    "YES" ,DARK_RED,   "OFF" ,DARK_BLUE },
  { 43,12,  CPU_BUSY_LED,    "ON " ,DARK_RED,   "OFF" ,DARK_BLUE },
  { 43,13,  ActivityLED,     "ON " ,DARK_BLUE,  "OFF" ,DARK_BLUE },
  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};

//---------------------------------------------------------------
//---- older style - do not use for newer applications -----
const DigFmtInfo digInFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};


const AnalogFmtInfo  adcFmt[] = {
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    showMovement = recentMovement(1000);

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/


//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "cur" OR the "next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "cur" value for the "input", and the "prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "cur" values
//    represent the input values just read (what they are "now") and the
//    "prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "cur" values).
//
//    During a pass the output values in "cur" represent what the outputs
//    where set to AFTER the "cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//-----------------------------------------------------------------------------
//    Consider this example:
//    - Goal: Keep a car moving at 60 mph
//
//    Process:
//    - Record speed from odometer
//    - Adjust the pressure on the gas pedal to level determined last pass
//    - Based on current speed of car, and pressure now applied to the pedal,
//      determine how much pressure to put on pedal next time.
//
//    The result:
//    - 1st pass:
//       - current speed: 50 mph
//       - 1st pass, so no decision to change pressure from last time
//       - speed less than 60 mph, so increase pressure next time
//    - 2nd pass:
//       - current speed: 50 mph (since no change in pedal pressure yet)
//       - increase pedal pressure (as decided last time)
//       - speed less than 60 mph, so increase pressure next time
//    - 3rd pass:
//       - current speed 60 mph (we increased the pressure last time)
//       - increase pedal pressure (as we decided last time)
//       - speed correct, so don't change pressure next time
//    - 4th pass:
//       - current speed 70 mph (oops!)
//       - don't change pedal pressure (as we decided last time)
//       - speed too fast, decrease pedal pressure next time
//    - 5th pass:
//       - current speed 70mph (we didn't adjust pressure last time!!!)
//       - decrease pressure (as decided last time)
//       - speed too fast, so decrease pressure next time
//    - 6th pass:
//       - current speed 60 mph (ahh....)
//       - decrease pressure (as we decided last time)
//       - speed correct, so don't change pressure next time
//    - 7th pass:
//       - current speed 50 mph (oh no....)
//       - don't change pressure (as decided last time)
//       - speed less than 60 mph, so increase pressure next time
//     (and round and round we go...)
//
//
//     The "obvious" solution would be this:
//       - record speed
//       - decide if we need more or less pressure
//       - adjust the pressure
//
//     Which is fine and good if the time to read, decide, and adjust is very
//     small compared to how often we do it.  But if it takes a significant
//     amount of time to do that compared to how often we do it, we have the
//     same delay problem (the speed is changing according to the current
//     pedal pressure while we are recording, deciding, and changing the
//     pressure - only to find the speed has already changed based on the
//     previous pressure, so we react (to slowly) again, and things still
//     oscillate.
//
//
//     The answer is to adjust the method as follows:
//      - What was the pressure that was being applied at the time we recorded
//        the current speed (which is the pressure applied during the PREVIOUS
//        pass!!!).
//      - How does our current speed compare to the correct speed?
//      - Scale the change in pressure based on changes in the measured
//        speed and how far it is from the desired speed (PID loops use
//        a combination of current error, rate of change, and cummulative error)
//
//     This gives us all the time not used to determine the current speed and
//     to apply a change in the pedal pressure to determine how we want to
//     change the pressure next time.  And by doing it on a consistent
//     schedule, we can expect some consistency in the result and adjust how
//     we compute the amount of change to apply to improve the overall result.
//
//=============================================================================


//=============================================================================
//  Functions called to set "next" values that are used as "input" to the
//  application control logic.  These functions must alter ONLY the "next"
//  values that represent "inputs".  All "output" values in next are entirely
//  the responsibility of the application logic in the updateOutputState()
//  function.
//=============================================================================

/*
//-----------------------------------------------------------------------------
void onRequest()
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();
}

//-----------------------------------------------------------------------------
void offRequest()
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();
}

//-----------------------------------------------------------------------------
void resetRequest()
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();
}
*/

#pragma warn -775

//=============================================================================
//  Copy the "cur" state from last time to "prev", and update "cur"
//=============================================================================
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).
//
//  This func first copies "cur" to "prev" and then "next" to "cur" which does
//  2 things for us:
//
//    - We have a copy of the controller's logical state from the last pass
//      (so we can determine what has changed)
//
//    - We have a copy of values set by other tasks indicating events/requests
//      that happend/were made since the beginning of the last pass
//
//  We then reset the values (as needed) that are set by other tasks to
//  indicate events and requests, and update the current state based on the
//  readings just taken and the requests and events communicated by other
//  tasks.
//
//  All of this is in preperation for the updateNextState() function, which
//  implements the main "decision making" portion of the application logic,
//  determining what the next state should be based on the latest conditions
//  (as determined by the updateCurState() function).
//
//=============================================================================
void updateCurState(void)
{
  bool  locRemWasPushed, wasMotorLocal, wasSoftEnabled;
  float  scaleBy;
  char  temp[32];


  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "Cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "Next" contains the values that the outputs were just set
//   to and any values that were changed by EVENT handlers (values
//   that will be used by the logic during this pass)
//------------------------------------------------------------------

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

  Next.processTime = GetMS();  // curTime = time(NULL);

  updateNoiseData();  //temp

  ampWasEnabled = enableAmp;
  locRemWasPushed = locRemBtnPushed;
  wasMotorLocal = motorLocal;

  msSincePrev = msSince(Cur.processTime);

  Next.digitalIn = digitalIn;


  lockBtnPushed =   !(digitalIn & LOCK_BUTTON);
  plusLimit =       !(digitalIn & LIMIT_SW_POS);
  minusLimit =      !(digitalIn & LIMIT_SW_NEG);
  setBtnPushed =    !(digitalIn & SET_BUTTON);
  fastAdjust =      BOOL(digitalIn & FAST_SW);
  locRemBtnPushed = !(digitalIn & LOC_REM_BTN);
  rfOK =            !(digitalIn & RF_OK);
  BBServo =         !(digitalIn & BBS_JUMPER);
  DCServo =         !(digitalIn & DCS_JUMPER);

//--- Is the Local/Remote switch at the motor set to Local? ---
//--- NOTE:  For BBServo, !motorLocal usually means no Motor Loc/Rem sw ---
  motorLocal = !(digitalIn & EXT_LOC_REM);
  if (DCServo)  motorLocal = !motorLocal;

//--- Determine what we can regarding the state of the FP ENABLE/DISABLE sw ---
  Next.fpEnableSw = FP_ENABLE_UNKNOWN;
  wasSoftEnabled = !(Cur.digitalOut & DISABLE_AMP);  // DO-03 was == 0
  // If /ENABLE was not 0 when the the software-controlled side of the switch
  // is, then the switch must be open
  if (Next.digitalIn & AMP_DISABLED)  {  //  /ENABLE (DI-04) was == 1
    if (wasSoftEnabled)  Next.fpEnableSw = FP_DISABLE; // while DO-03 was == 0
  }
  // If /ENABLE was 0 and there is no known reason for it to have been 0, then
  // we have a hardware problem
  else  {  //  /ENABLE (DI-04) was == 0
    if (!(DCServo AND motorLocal))  // not due to DSServo ctl box set to local
      if (wasSoftEnabled)  // if soft enable asserted, then sw must == ENABLE
        Next.fpEnableSw = FP_ENABLE;
      else  // no known reason for it to be 0
        if (time(NULL) - prevEnableMsgTime > 600)  {
          printf("*** Amplifier reads enabled when it shouldn't be ***\n");
          prevEnableMsgTime = time(NULL);
        }
  }

  if (BBSCtlMode)
    ampFault = (plusLimit AND minusLimit);  // requires wiring changes!
  else
    ampFault = BOOL(digitalIn & AMP_FAULT);

  if (motorLocal)  {  // Local mode set at motor?
    // If just set, then save state of localMode to restore later
    if (!wasMotorLocal)  Next.preMotorLocalMode = Next.localMode;
    Next.localMode = true;
  }
  else  {
    // If motorLocal mode just cleared, restore localMode to its prev value
    if (wasMotorLocal)  Next.localMode = Next.preMotorLocalMode;
    // Toggle Local/Remote mode when FP Local/Remote button is pressed
    if (locRemBtnPushed AND !locRemWasPushed)  Next.localMode = !Next.localMode;
  }

  motorPos = readVal[MOTOR_POS];
  if (rfOK)  Next.phaseAngle = readVal[RF_PHASE];
  driveSwVolts = anaIn[DRIVE_SW];

  moveMinusSw = (driveSwVolts < -1.5);
  movePlusSw = (driveSwVolts > 1.5);


// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer (and is now required by the new softReset code).

  Prev = Cur;  Cur = Next;

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - "Cur" contains the state the inputs were in at the START of
//   of this pass, and the state the outputs were set to at the
//   start of this pass (and will be in during the entire pass).
// - "Prev" contains the state the inputs were at the START of the
//   previous pass, and the state the outputs were set to at the
//   start of and all during the previous pass (i.e.the output
//   states that resulted in the input states in "Cur").
//------------------------------------------------------------------

//--- log msgs for bit chgs only if not starving other processes for cycles ---
  if ((applLoopTime < SchedMS) AND (SchedMS - applLoopTime >= 2))
    logDigIOChgs(false);


// At this point we update values that are used later in this pass and/or
// by other tasks and for which we don't need the "Next" instance.

  ctlMode = CM_Unknown;
  if (BBServo)  ctlMode = CM_DigDirection;
  if (DCServo)  ctlMode = CM_AnalogVelocity;

  atALimit = (plusLimit OR minusLimit) AND (!plusLimit OR !minusLimit);

  ctlrLocal = (Cur.localMode AND !motorLocal);

  chgdToLocal = (!Prev.localMode AND Cur.localMode);
  chgdToRemote = (Prev.localMode AND !Cur.localMode);

//--- Pressing and releasing the LOCK btn requests toggle of Phase-feedback ---
//--- mode.  Holding it for > 2 secs forces it to change the setpoint to ---
//--- current phase reading (delta-phi) rather than using previous setpoint ---
  if (lockBtnPushed)  {
    if (!lockPressedTime)  lockPressedTime = GetMS_();  }  // just pushed
  else if (lockPressedTime)  {  // just released
    newPhaseLockReq = (!setBtnPushed AND ctlrLocal);
    newPhaseSetting = (msSince_(lockPressedTime) > 2000);
    lockPressedTime = 0;
  }

  if (! newPhaseLockReq)  phaseLockReqIgnored = newPhaseSetting = false;

  softDisabled = !enableAmp;
  hardDisabled = (Cur.fpEnableSw != FP_ENABLE) AND !softDisabled;

  if (softDisabled)  ampFault = false;
  if (msSince(lastEnableTime) < 200)  {
    Cur.fpEnableSw = FP_ENABLE_UNKNOWN;  ampFault = false; }

  if (rfOK)  lastTimeRFOK = GetMS();

  _EnterCritical_();  knobTicks = qdCount;  qdCount = 0;  _ExitCritical_();

  manualMoveCmd = Cur.localMode AND (knobTicks OR movePlusSw OR moveMinusSw);

  ignoreSettings = Cur.driveLock OR Cur.phaseLocked;


//--- check for manual/front panel initiation of minMoveRefVolt calibration ---
//--- for NIM module, pressing SET and LOCK buttons together while in local ---
//--- mode will either start a REF calibration or cancel cur calibration ---
  if (!lockBtnPushed OR !setBtnPushed)
    processedLockAndSet = false;
  else do  {
    if (processedLockAndSet)  break;  // don't process same event > once
    processedLockAndSet = true;

    if (ignoreSettings)  break;  // not while DriveLock or PhaseFeedback enabled
    if (!ctlrLocal)  break;  // must be in ctlr local mode
    if (BBSCtlMode)  break;  // not applicable to BB Servos

    if (calibState)
      cancelCalib = true;
    else
      calibState = CALIB_REF_INIT;
  }  while (0);


//--- Check for lock button alone pushed while RF OK ---
//--- This toggles the Phase Feedback Lock mode, where the controller ---
//--- attempts to maintain a steady Phase Feedback value ---
  if (newPhaseLockReq)  do  {
    newPhaseLockReq = false;  phaseLockReqIgnored = true;
    if (Cur.driveLock)  break;  // NO movement/changes allowed!
    if (Cur.phaseLocked)  {
      disablePhaseLock(false);  phaseLockReqIgnored = false;  break; }
    if (!rfOK)  break;  // must have RF to enable Phase Lock!
    if (! BBSCtlMode)  break;  // only for BB Servos right now
    phaseLockReqIgnored = false;  enablePhaseLock(false);
  }  while (false);


//--- If we lost RF for too long, return to "Cold Position" ---
  if (Cur.phaseLocked AND rfLockResetDelay AND !phaseLockInactive)
    if (msSince(lastTimeRFOK) > rfLockResetDelay)  {
      printf("RF off too long - returning to cold position: %s",
             fmtFloat(posSet, 8,2, temp) );
      newPosReq(posSet);  phaseLockInactive = true;
    }

  phaseLockActive = (Cur.phaseLocked AND rfOK);
  if (phaseLockActive)  phaseLockInactive = false;
  phaseLockPaused =
    (Cur.phaseLocked AND !phaseLockActive AND !phaseLockInactive);


//--- update motor position history ---
  motorHistory[lastHistIdx] = adcRaw[MOTOR_POS];
  if (++lastHistIdx == HistorySize)  lastHistIdx = 0;

  realPos = actualPos(motorPos);

//--- disable idleRest if new movement cmd ---
  if (Cur.idleRest AND !Cur.driveLock)  if (useNewPos OR manualMoveCmd) {
    // assume for one pass that fault was due ONLY to idleRest mode
    Cur.idleRest = softDisabled = false;
    Cur.fpEnableSw = FP_ENABLE_UNKNOWN;
    enableAmp = true;  lastMoveCmdTime = GetMS();
  }


  if (Cur.phaseLocked)  {

  //--- apply manual chg cmds to desiredPhase while in Phase Feedback mode ---
    if (manualMoveCmd AND !Cur.driveLock AND Cur.localMode AND phaseLockActive)  {
      scaleBy = fastAdjust ? knobRateF : knobRateS;
      if (movePlusSw)   knobTicks = (long) (switchRate *  10.0 / knobRateF);
      if (moveMinusSw)  knobTicks = (long) (switchRate * -10.0 / knobRateF);
      desiredPhase += (float) knobTicks * scaleBy / 100.0;
      lastPhaseChgTime = GetMS();
    }

  //--- Has Phase Feedback value drifted outside deadzone? ---
  //--- NOTE: Once initiated, the phaseErr value needs to stay constant ---
  //--- until current correction has ended (to avoid stopping correction ---
  //--- before reaching the middle of the deadzone, or a never ending ---
  //--- series of corrections) ---
    if (!reducePhaseErr)  {
      phaseErr = desiredPhase - phaseAngle;
      if (fabs(phaseErr) > maxPhaseErr)  {
        reducePhaseErr = true;
      //--- ensure idleRest is disabled ---
        Cur.idleRest = softDisabled = false;
        Cur.fpEnableSw = FP_ENABLE_UNKNOWN;
        enableAmp = true;  lastMoveCmdTime = GetMS();
      }
    }

  }
  if (!rfOK)  reducePhaseErr = false;



/*EXAMPLE:
//------------------------------------------------------------------
// rampMode, rfOnTime
//    Description:
//       Specifies how changes to Next.activeSetpoint should be
//       handled.
//
//       If the RF is OFF, then it is set to rfOffSetpoint.
//
//       If the RF just changed from OFF to ON, then an interlval
//       of offToOnDelay ms must pass before any changes are made.
//
//       If the RF is ON and offToOnDelay ms have passed, then
//       it is ramped toward the rfOnSetpoint at the rate specified
//       by coldRampRate.
//
//       If the RF is ON, the activeSetpoint has reached the
//       rfOnSetpoint, then any difference between rfOnSetpoint and
//       activeSetepoint is reduced at the rate specified by
//       warmRampRate.
//
//    Inputs:
//       offToOnDelay
//       rampMode
//       Cur.activeSetpoint
//       Cur.rfOnSetpoint
//       Prev.rfIsOff
//       Cur.rfIsOff
//------------------------------------------------------------------
  if (!Cur.rfIsOn)  rampMode = RAMP_OFF;

  switch (rampMode)  {
    case RAMP_OFF:  // waiting for RF ON
      if (Cur.rfIsOn)  {
        rampMode = RAMP_WAIT;  rfOnTime = GetMS();
        printf("Starting OFF-to-ON delay\r\n");
      }
      break;

    case RAMP_WAIT: // waiting for offToOnDelay to finish
      if (msSince(rfOnTime) >= offToOnDelay)  {
        rampMode = RAMP_COLD;
        printf("Ramping using COLD rate\r\n");
      }
      break;

    case RAMP_COLD:  // waiting for cold-mode ramping to finish
      if (Cur.activeSetpoint == Cur.rfOnSetpoint)  {
        rampMode = RAMP_WARM;
        printf("Using WARM rate for changes\r\n");
      }
      break;
  }
*/

}


//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of output values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  float  scaleBy, posChg, curErr;


  if (current != applTask)  return;  // must only be called by "applic" task

//****************************************************************************
//===== Currently using ported version of original pre-DEBROS code that  =====
//===== does NOT adhere to the DEBROS self-contained blocks of logic     =====
//===== model that avoids order-of-execution dependencies                =====
//****************************************************************************

//tdebug #define  AvgRealPos()  (actualPos(avgPos(1000,1000)))
#define  AvgRealPos()  (avgPos(1000,1000))

  if (Cur.driveLock OR Cur.idleRest OR chgdToRemote)  {
    if (Cur.driveLock)  { newPos = posSet = AvgRealPos();  useNewPos = true; }
    if (Cur.driveLock OR Cur.idleRest)  {
      Next.ctlState = CS_IDLE;  enableAmp = false; }
    bbMovePlus = bbMoveMinus = false;  reachedGoal = true;
    pFactor = iFactor = dFactor = setRefVolt = ctlVolts = 0.0;
    return;
  }

  if (chgdToLocal)  {  // just switched to local mode?
    enableAmp = true;  hdwCtlErr = Next.idleRest = false; }


//--- if motor under manual ctl at device, reset ctl values ---
//--- NOTE:  The RF Bang-Bang Servos don't HAVE a motor local switch so we ---
//---        can't tell the difference between drift and movement via the ---
//---        switches at the motor control box. ---
//---        To allow for this, the NIM module should be set to Disabled ---
//---        before moving the motor via the motor control box.  If this ---
//---        is not done, then posSet will NOT follow the current position ---
//---        and the control system will act based on a setpoint that no ---
//---        longer matches the current position (which will result in the ---
//---        motor moving back to the old set point when the controller is ---
//---        taken out of local mode) ---
  if (motorLocal
        OR (BBSCtlMode AND ((Cur.fpEnableSw == FP_DISABLE) OR ampFault)))  {
    bbMovePlus = bbMoveMinus = false;
    pFactor = iFactor = dFactor = setRefVolt = 0.0;
    posSet = AvgRealPos();  motorPosSet = neededPos(posSet);
  }

//--- if new goal position, make sure amp isn't disabled due to previous ---
//--- detection of a hardware/controller error condition ---
  if (useNewPos)  enableAmp = true;

//===== RF Servo controller connected to DC or BB Servo =====
//  When connected to a DC Servo, we have direct control of the ref/velocity
//  voltage fed to the amplifier that drives the motor and can read the state
//  of the amplifier.
//  When connected to a Bang-Bang servo, we are limited to the 2 movement bits.
//  In both cases, there are 2 local modes:  A software one invoked at the NIM
//  module, where we retain direct control over movement of the motor, and
//  a hardware one (at or near the motor), where we can only monitor what the
//  motor is doing.
//=====================================

  do  {
    if (hardDisabled OR ampFault OR motorLocal)  break;  // ctl vals reset above
    if (softDisabled AND !useNewPos)  break;

  //--- Phase Feedback mode active - move motor in dir needed ---
    if (phaseLockActive OR phaseLockPaused)  {
      bbMovePlus = bbMoveMinus = false;
      pFactor = iFactor = dFactor = setRefVolt = 0.0;

      if (useNewPos)  { posSet = newPos;  useNewPos = false; }

      if (phaseLockActive)  {
      //--- stop correcting once we pass the center of the deadzone ---
        curErr = desiredPhase - phaseAngle;
        if (SIGN(curErr) != SIGN(phaseErr))  reducePhaseErr = false;

        if (! reducePhaseErr)  break;

        scaleBy = fastAdjust ? knobRateF : knobRateS;
        if (BBSCtlMode)  {
          if (phaseErr > 0.0)   bbMovePlus = true;   else
          if (phaseErr < 0.0)  bbMoveMinus = true;
          if (bbMovePlus OR bbMoveMinus)  correctingPos = true;
        }
        else  {  // finish - need smarter algorithm to relate error to velocity
          setRefVolt = phaseErr * knobRateS;
        }
      }

      Next.ctlState = CS_STEADY_REF;  break;
    }


  //--- Remote mode: achieve/maintain current set point ---
    if (!Cur.localMode)  { updateCtlValues();  break; }

  //--- Local mode, but no changes via the NIM front panel this pass ---
  //--- In this case, reset ctl values and prep to maintain new pos ---
    if (! manualMoveCmd)  {
      if (!localChgLastPass)
        updateCtlValues();
      else {
        bbMovePlus = bbMoveMinus = false;
        pFactor = iFactor = dFactor = setRefVolt = 0.0;
        Next.ctlState = CS_END_LOCAL_CHG;  localChgLastPass = false;
      }
      break;
    }


  //===== Local mode and changes via the NIM front panel =====

  //--- if SET button is pressed, apply changes to set point ---
    if (setBtnPushed)  {
      if (fastAdjust)  scaleBy = knobRateF;  else  scaleBy = knobRateS;

      if (movePlusSw)   posChg = scaleBy * switchRate * setRate / 100.0;
      if (moveMinusSw)  posChg = scaleBy * switchRate * setRate / -100.0;

      if (knobTicks)  posChg = (float) knobTicks * scaleBy * setRate / 100.0;

      newPos = posSet + posChg;  useNewPos = true;

      updateCtlValues();  break;
    }


  //--- SET button not pressed, so apply changes to motor velocity ---
    bbMovePlus = bbMoveMinus = false;
    pFactor = iFactor = dFactor = setRefVolt = 0.0;

    if (fastAdjust)  scaleBy = knobRateF;  else  scaleBy = knobRateS;
    if (BBSCtlMode)  {
      if (movePlusSw)   bbMovePlus = true;   else
      if (moveMinusSw)  bbMoveMinus = true;  else
      if (knobTicks > 0)  bbMovePlus = true;   else
      if (knobTicks < 0)  bbMoveMinus = true;
    }
    else  {
      setRefVolt = (float) knobTicks * scaleBy;
      if (movePlusSw)   setRefVolt =  10.0 * scaleBy / knobRateF * switchRate;
      if (moveMinusSw)  setRefVolt = -10.0 * scaleBy / knobRateF * switchRate;
    }

    localChgLastPass = true;  Next.ctlState = CS_STEADY_REF;

  } while (false);


  ctlVolts = setRefVolt;
  if (ctlVolts > AOutMax(CTL_VOLTS))  ctlVolts = AOutMax(CTL_VOLTS);
  if (ctlVolts < AOutMin(CTL_VOLTS))  ctlVolts = AOutMin(CTL_VOLTS);


/*EXAMPLE:
//--------------------------------------------------------------------
// Next.activeSetpoint
//    Description:
//       Determines the setpoint to be applied to the regulation
//       circuit at the beginning of the next pass.
//       See rampMode for a description of how changes are handled
//    Inputs:
//       rampMode
//       msSincePrev
//       Cur.activeSetpoint
//       Cur.rfOffSetpoint
//       Cur.rfOnSetpoint
//       coldRampRate
//       warmRampRate
//----------------------------------------------------------------------
  switch (rampMode)  {
    case RAMP_OFF:  // RF is OFF - use rfOffSetpoint
      Next.activeSetpoint = Cur.rfOffSetpoint;  break;

    case RAMP_WAIT: // delay after recent OFF-to-ON
      Next.activeSetpoint = Cur.activeSetpoint;  break;

    case RAMP_COLD: // use the (presumably) slower cold rate
    case RAMP_WARM: // use the (presumably) faster warm rate
      if ((diff = Cur.rfOnSetpoint - Cur.activeSetpoint) == 0.0)  {
        Next.activeSetpoint = Cur.activeSetpoint;  break;  }
      maxChg =
        (rampMode == RAMP_COLD ? coldRampRate : warmRampRate);
      maxChg = maxChg * (float)msSincePrev / 1000.0;
      if (FABS(diff) > maxChg)  diff = (diff < 0.0 ? -maxChg : maxChg);
      Next.activeSetpoint = Cur.activeSetpoint + diff;
      break;
  }
*/

}


//-----------------------------------------------------------------------------
//  Update the values that control the motor movement.
//
//  For the PID (Proportional-Integral-Derivative) algorithm, the voltage
//  feed to the Servo Amplifier is computed.  The control (or reference)
//  voltage is basically a velocity vector, specifying direction (by +/-
//  polairty) and speed (via amplitude)
//
//  For Bang-Bang Servos, all movement is controlled by two drive bits, each
//  one controlling motor movement in one direction.
//
//  !!!!! WARNING !!!!!
//
//  This function is called by the timer-interrupt driven function that reads
//  and writes the Analog and Digital I/O values.  It should be limited to
//  processing current input values and updating output values that will become
//  the active values on the next interrupt (currently, this interuupt occurs
//  every 10ms, so this routine must take considerably less time than that to
//  do EVERYTHING it needs to).  Note that formatting data for display (e.g.
//  sprintf calls) is very time consuming and should be avoided as much as
//  possible.
//
//  Also, any diagnostic output that is generated must be kept to a minimum or
//  it will quickly flood the output stream.
//-----------------------------------------------------------------------------
void updateCtlValues()
{
  float  posDiff, absDiff, lastIFactor;
  bool  sufficientForce, reversedForce, gotCloser;
  static bool  passedGoal, lastMoveMinus, lastMovePlus;
  static ulong  lastTimeCondsOK;
  static float  posErr0, posErr1, posErr2, ctlVal0, ctlVal1, lastRefVolt,
                closestSoFar;
  char  temp[20];


#define  HalfADCRes  ADCGain(MOTOR_POS)


  if (useNewPos)  Next.ctlState = CS_START_MOVE;


  switch (Next.ctlState)  {

    case CS_IDLE:  // nothing to do for now....
      reqMovement = 0.0;  trackOverShoot = false;
      bbMovePlus = bbMoveMinus = false;
      setRefVolt = 0.0;  pFactor = iFactor = dFactor = 0.0;
      reachedGoal = true;
      break;


    case CS_STEADY_REF:  // maintain steady ref voltage
      trackOverShoot = false;
      break;


    case CS_END_LOCAL_CHG:  // local mode just after chgs stopped
      newPos = realPos;

    case CS_START_MOVE:  // new position requested
      if (newPos < ReadMin(MOTOR_POS))  newPos = ReadMin(MOTOR_POS);
      if (newPos > ReadMax(MOTOR_POS))  newPos = ReadMax(MOTOR_POS);

      posSet = newPos;  useNewPos = false;

      motorPosSet = neededPos(posSet);
      posDiff = reqMovement = motorPosSet - motorPos;

      if (BBSCtlMode)  {
        bbMoveMinus =  BOOL(posDiff < -minusMargin);
        bbMovePlus =   BOOL(posDiff >  plusMargin);
        lastMoveMinus = lastMovePlus = false;
      }
      else  {
        posErr0 = posDiff;
        posErr1 = 0.0;

        pFactor = pWeight * posErr0;
        iFactor = 0.0;
        dFactor = 0.0;
        setRefVolt = lastRefVolt = ctlVal0 = pFactor;
      }

      enableAmp = true;
      hdwCtlErr = Next.idleRest = passedGoal = reachedGoal = false;

      if (BBSCtlMode)
        sufficientForce = (bbMoveMinus OR bbMovePlus);
      else
        sufficientForce =
           ((fabs(setRefVolt) > minMoveRefVolt) AND (minMoveRefVolt > 0.05));

      closestSoFar = fabs(posDiff);

      lastNoiseChkTime = lastTimeCondsOK = lastMoveCmdTime = GetMS();

      overShoot = 0.0;  trackOverShoot = true;

      if (slopPosKnown)
        if (reqMovement < 0)
          adjustMinusMargin = true;
        else
          adjustPlusMargin = true;

      Next.ctlState = CS_ACTIVE_CTL;
      break;


    case CS_ACTIVE_CTL:  // update control voltage based on new state
      if ((Cur.fpEnableSw == FP_DISABLE) OR ampFault)  {
        bbMovePlus = bbMoveMinus = adjustPlusMargin = adjustMinusMargin = false;
        setRefVolt = 0.0;  pFactor = iFactor = dFactor = 0.0;
        Next.ctlState = CS_IDLE;  break;
      }


    //--- hit a limit switch appropriate for the dir of movement? ---
      if ( ((reqMovement > 0.0) AND plusLimit) OR
           ((reqMovement < 0.0) AND minusLimit) )  {
        adjustPlusMargin = adjustMinusMargin = false;
        if (! ignoreHdwErrs)  {
          printf("Limit switch triggered at %s\n", fmtFloat(realPos, 7,3, temp));
          motorPosSet = motorPos;  reqMovement = 0.0;  reachedGoal = true;
        }
      }

      posDiff = motorPosSet - motorPos;  absDiff = fabs(posDiff);

      if (BBSCtlMode)  {
        sufficientForce = (bbMoveMinus OR bbMovePlus);
        reversedForce =
           (bbMoveMinus AND lastMovePlus) OR (bbMovePlus AND lastMoveMinus);
        lastMoveMinus = bbMoveMinus;  lastMovePlus = bbMovePlus;
      }
      else  {
        sufficientForce =
           ((fabs(setRefVolt) > minMoveRefVolt) AND (minMoveRefVolt > 0.05));
        reversedForce = (SIGN(setRefVolt) != SIGN(lastRefVolt));
        lastRefVolt = setRefVolt;
        ctlVal1 = ctlVal0;
        posErr2 = posErr1;  posErr1 = posErr0;  posErr0 = posDiff;
      }
      gotCloser = (absDiff < closestSoFar);


    //--- if trying to move AFTER having reached our goal, then it must be ---
    //--- due to a need to correct for drift or overshoot ---
      if (reachedGoal AND sufficientForce)  correctingPos = true;

    //--- determine if we have passed and/or reached our goal ---
      if (SIGN(posDiff) != SIGN(reqMovement))  passedGoal = true;
      if ((absDiff <= HalfADCRes) OR passedGoal)  if (! reachedGoal)  {
        lastCorrectingMsgTime = GetMS();  reachedGoal = true;  }

    //--- keep track of most recent time nominal conditions existed ---
      if (gotCloser OR reversedForce OR !sufficientForce OR reachedGoal)  {
        closestSoFar = absDiff;  lastTimeCondsOK = GetMS();  }

    //--- keep track of how far we overshoot or wander from goal pos ---
      if (trackOverShoot AND passedGoal)
        if (absDiff > overShoot)  overShoot = (overShoot + absDiff) / 2.0;

    //--- if device not responding as expected, stop trying until new pos ---
    //--- command received ---
    //--- NOTE this is also meant to catch miswired or bad limit switches ---
      if ((msSince(lastTimeCondsOK) > MaxStickTime)
          AND (calibState != CALIB_REF_SETTLE))  {
        adjustPlusMargin = adjustMinusMargin = false;
        if (! ignoreHdwErrs)  {
          if ((plusLimit OR minusLimit) AND !reachedGoal) // (hitting expected limit handled above)
            printf("*** Limit swiches appear to be wired backwards ***\n");
          else
            printf("*** Hdw or Ctlr Error: Device not responding to cmds ***\n");
          enableAmp = false;  hdwCtlErr = true;
          Next.ctlState = CS_IDLE;  break;
        }
      }

  //--- limit time we actively control motor after getting a new setpoint ---
    if (maxActiveSecs > 0)
      Next.idleRest = (secSince(lastMoveCmdTime) > maxActiveSecs);


    //--- update values that control motor movement ---

      if (BBSCtlMode)  {
        if (reachedGoal)
          bbMoveMinus = bbMovePlus = false;
        else  {
          bbMoveMinus = BOOL(posDiff < -minusMargin);
          bbMovePlus =  BOOL(posDiff >  plusMargin);
        }
      }
      else  {
      /*--- velocity-based algorithm (does not suffer from Integral windup...) ---
        pFactor = pWeight * posErr0;
        iFactor = iWeight * (posErr0 - 2.0*posErr1 + posErr2) * 0.01;
        dFactor = dWeight * (posErr0 - posErr1) / 0.01;

        ctlVal0 = ctlVal1 + pFactor + iFactor + dFactor;

        if (ctlVal0 < -10.0)  ctlVal0 = -10.0;
        if (ctlVal0 > 10.0)   ctlVal0 =  10.0;
      ------------*/

      /*--- positional-based algorithm (requires Integral windup check) ---*/
        if (iWeight == 0.0)  iFactor = 0.0;
        lastIFactor = iFactor;

        pFactor = pWeight * posErr0;
        iFactor = iFactor + iWeight * posErr0 * 0.01;
        dFactor = dWeight * (posErr0 - posErr1) / 0.01;

        ctlVal0 = pFactor + iFactor + dFactor;

        if (ctlVal0 < -10.0)  { ctlVal0 = -10.0;  iFactor = lastIFactor; }
        if (ctlVal0 >  10.0)  { ctlVal0 =  10.0;  iFactor = lastIFactor; }

        setRefVolt = ctlVal0;
      }

      break;


    default:
      break;
  }

}


//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
  uint  u;
  float  movement;


  if (current != applTask)  return;  // must only be called by "applic" task

//--- update output values that are based on current conditions (values ---
//--- in "cur" or any values used in a conditional in the applic logic) ---

/*EXAMPLE:
  // (insured earlier that only ONE of these can be true when setting them)
  if (initialBtnPressed)  setVal[FP_METER] = rfOffSetpoint;  else
  if (finalBtnPressed)    setVal[FP_METER] = rfOnSetpoint;  else
                          setVal[FP_METER] = cavityVolts;


  SetOffApplLEDs(Next.digitalOut);

  if (limited)  SetOnLimitLED(Next.digitalOut);
  if (regErr)   SetOnErrorLED(Next.digitalOut);
*/

//--- update output values based on values that were set by the application ---
//--- logic (e.g. values in "next" or variables that are the target of an ---
//--- assignment statement in the application logic) ---
//--- update analog outputs ---

  setVal[CTL_VOLTS] = ctlVolts;  // ctl volt to DC Servo amplifier
  setVal[EXTERN_METER] = setBtnPushed ? posSet : realPos;


  Next.digitalOut = 0xffffffffL;

  movement = (avgPos(100,100) - avgPos(600,100)) * 2.0;  // 500 ms period * 2

  if (motorLocal)  {
    if (GetMS_() & 0x100)  ClearBits(Next.digitalOut, LOC_REM_LED);  }
  else
    if (Cur.localMode)  ClearBits(Next.digitalOut, LOC_REM_LED);

  if (hdwCtlErr)  {
    if (GetMS_() & 0x100)  {
      ClearBits(Next.digitalOut, DRIVE_MINUS_LED | DRIVE_PLUS_LED);  }
  }
  else  {
    if (movement < -activeNoise)  ClearBits(Next.digitalOut, DRIVE_MINUS_LED);
    if (movement >  activeNoise)  ClearBits(Next.digitalOut, DRIVE_PLUS_LED);
  }
//--- if attempting local move while locked, turn on drive LED's ---
  if (manualMoveCmd AND Cur.driveLock)  {
    ClearBits(Next.digitalOut, DRIVE_MINUS_LED | DRIVE_PLUS_LED);
    _EnterCritical_();  qdCount = 0;  _ExitCritical_();
  }

  if (phaseLockReqIgnored)  { // rapid blink if ignored
    if (GetMS_() & 0x40)  ClearBits(Next.digitalOut, LOCKED_LED);  }
  else
    if (phaseLocked)  ClearBits(Next.digitalOut, LOCKED_LED);

  if (calibState)  if (GetMS_() & 0x100)  ClearBits(Next.digitalOut, CPU_BUSY_LED);

//  if (Cur.driveLock)  {  // drives Vincent nuts!
//    if (GetMS_() & 0x80)  ClearBits(Next.digitalOut, FAULT_LED);  }
//  else
    if (ampFault OR hardDisabled)  ClearBits(Next.digitalOut, FAULT_LED);


  switch (ctlMode)  {

    case CM_AnalogPosition:
    case CM_AnalogVelocity:
      SetBits(Next.digitalOut, MOVE_BITS); // just to be safe
      if (enableAmp)  ClearBits(Next.digitalOut, DISABLE_AMP);
  //--- if amp fault when it should be enabled, try pulsing the RESET line ---
      if (ampFault AND enableAmp AND (Cur.fpEnableSw == FP_ENABLE))
        if (GetMS_() & 0x200)  ClearBits(Next.digitalOut, AMP_RESET_BIT);
      break;

    case CM_DigDirection:
      ClearBits(Next.digitalOut, DISABLE_AMP);  //to read FP switch
      if (bbMovePlus)  ClearBits(Next.digitalOut, MOVE_POS_BIT);
      if (bbMoveMinus)  ClearBits(Next.digitalOut, MOVE_NEG_BIT);
      break;

    default:  // blink LEDS in pattern to show error
      for (u=0; u<6; ++u)  SetBits(Next.digitalOut, LED_ErrPattern[u]);
      u = (GetMS_()/100) % 6;
      ClearBits(Next.digitalOut, LED_ErrPattern[u]);
  }

  if (enableAmp AND !ampWasEnabled)  lastEnableTime = GetMS();
}



//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

//--- update settings based on values that were USED by the application ---
//--- logic (i.e. "cur" values that serve as "inputs" to the logic) ---

/*EXAMPLE
  activeSetpoint = Cur.activeSetpoint;

  statusBitsAppl = 0;
  if (Cur.rfIsOn)  statusBitsAppl |= STAT_RF_ON;
  if (limited)     statusBitsAppl |= STAT_LIMITED;
  if (regErr)      statusBitsAppl |= STAT_REG_ERR;
  if (!localMode)  statusBitsAppl |= STAT_REMOTE;
  if (closedLoop)  statusBitsAppl |= STAT_CLOSED;
*/

  devStatus = 0;
  if (Cur.fpEnableSw == FP_DISABLE)  devStatus |= DS_DISABLE;
  if (ampFault)           devStatus |= DS_FAULT;
  if (Cur.localMode)      devStatus |= DS_LOCAL;
  if (! softDisabled)  {
    if (plusLimit)        devStatus |= DS_PLIMIT;
    if (minusLimit)       devStatus |= DS_NLIMIT;
  }
  if (calibState)         devStatus |= DS_CALIBRATE;
//  if (flashBad)         devStatus |= DS_HDW_ERROR;
//  if (oscillating)      devStatus |= DS_HDW_ERROR;
  if (hdwCtlErr)          devStatus |= DS_HDW_ERROR | DS_NLIMIT | DS_PLIMIT;
  if (rfOK)               devStatus |= DS_RF_OK;
  if (ctlMode == CM_DigDirection)    devStatus |= DS_BBS_JUMPER;
  if (ctlMode == CM_AnalogVelocity)  devStatus |= DS_DCS_JUMPER;
  if (motorLocal)         devStatus |= DS_EXTERN_LOCAL;
  if (Cur.driveLock)      devStatus |= DS_DRIVELOCK;
  if (Cur.idleRest)       devStatus |= DS_IDLE;
  if (Cur.phaseLocked)    devStatus |= DS_PHASELOCKED;


//--- update settings based on values that were UPDATED by the application ---
//--- logic (i.e. "next" values that serve as "outputs" from the logic) ---

/*EXAMPLE:
  rfOffSetpoint = Next.rfOffSetpoint;
  rfOnSetpoint = Next.rfOnSetpoint;
*/

}



//-----------------------------------------------------------------------------
//  Request to change the position value used when Phase Feedback Lock mode is
//  not enabled.
//-----------------------------------------------------------------------------
bool  newPosReq(float pos)
{
  if (Cur.driveLock)  {
    printf("New pos req: Ignored while Drive Lock enabled\r\n");
    return false;
  }

  _EnterCritical_();  newPos = pos;  useNewPos = true;  _ExitCritical_();

  return true;
}


//-----------------------------------------------------------------------------
//  Enable Phase Feedback Lock mode, which adjusts the position as needed to
//  maintain a constant phase value
//  returns true on success, false if an error
//-----------------------------------------------------------------------------
bool enablePhaseLock(bool remoteCmd)
{
  char  temp[20];


  if (Cur.phaseLocked)  return true;

  printf("Enable PhaseLock request: ");

  if (Cur.localMode AND remoteCmd)  {
    printf("Ignored while in Local mode\r\n");  return false; }

  if (Cur.driveLock)  {
    printf("Ignored while Drive Lock enabled\r\n");  return false; }

  if (!rfOK)  {
    printf("Ignored due to bad RF\r\n");  return false; }


  KLOCK();
  if (newPhaseSetting)  desiredPhase = Cur.phaseAngle;
  Next.phaseLocked = true;
  KUNLOCK();

  printf("Locked to: %s degrees (pos: %s)",
         fmtFloat(desiredPhase, 7,3, temp), fmtFloat(posSet, 7,3, temp+10) );

  return true;
}

//-----------------------------------------------------------------------------
//  disable the Phase Feedback Lock mode
//  returns true on success, false if an error
//-----------------------------------------------------------------------------
bool disablePhaseLock(bool remoteCmd)
{
  char  temp[20];


  if (! Cur.phaseLocked)  return true;

  printf("Disable PhaseLock request: ");

  if (Cur.localMode AND remoteCmd)  {
    printf("Ignored while in Local mode\r\n");  return false; }

  if (Cur.driveLock)  {
    printf("Ignored while Drive Lock enabled\r\n");  return false; }


  _EnterCritical_();  Next.phaseLocked = false;  _ExitCritical_();

  printf("Returning to pos: %s\r\n", fmtFloat(posSet, 7,3, temp) );

  newPosReq(posSet);

  return true;
}



//----------------------------------------------------------------------------
//  Determine the actual position of the device being controlled based on the
//  current position of the motor and the slop zone the motor moves within
//  and update the value of slopPos.
//
//  NOTE:  The "actual" position of the device is defined as the center of
//         the slop zone.
//----------------------------------------------------------------------------
float actualPos(float motor_pos)
{
  float  halfSlopVal;


  if (slopValue == 0.0)  {
    slopPosKnown = true;  return (slopPos = motor_pos);  }


//  NOTE:  While it is true that, in reality, the REAL slopPos (center of
//         the of slop zone) and motorPos values cannot differ more than
//         halfSlopVal, it is possible for the assumed value of slopPos
//         (= motorPos at startup) and the real motorPos values to vary
//         by twice that (e.g. assume the motor is against one edge of the
//         slop zone at startup.  Then it has to move a full slopValue amount
//         from its original position - the assumed slopPos value, starting
//         off wrong by halfSlopVal - before we know for certain the true
//         value of slopPos).

  halfSlopVal = slopPos / 2.0;

//--- check to see if we actually moved the device (for sure) yet ---
  if (!slopPosKnown)  {
  //--- touch negative edge (for sure) yet? ---
    if ((motor_pos < slopPos) AND (slopPos - motor_pos > slopValue)) {
      slopPos = motor_pos + halfSlopVal;  slopPosKnown = true;  }
  //--- touch positive edge (for sure) yet? ---
    if ((motor_pos > slopPos) AND (motor_pos - slopPos > slopValue)) {
      slopPos = motor_pos - halfSlopVal;  slopPosKnown = true;  }
  }

//--- if still don't know slop position, then just return orig motor pos ---
  if (!slopPosKnown)  return slopPos;

//--- did we actually move the device in the neg dir? ----
  if (motor_pos < slopPos - halfSlopVal)
    return (slopPos = motor_pos + halfSlopVal);

//--- did we actually move the device in the pos dir? ---
  if (motor_pos > slopPos + halfSlopVal)
    return (slopPos = motor_pos - halfSlopVal);

//--- just moved within the slop zone, so return last know position ---
  return slopPos;
}

//----------------------------------------------------------------------------
//  Based on current position of the motor and the slop zone, determine the
//  position the motor needs to be moved to in order to move the device to
//  the specified position.
//
//  NOTE:  The "actual" position of the device is defined as the center of
//         the slop zone (see "diagram" in notes near start of file).
//----------------------------------------------------------------------------
float neededPos(float desiredPos)
{
  float  halfSlopVal;


  if (slopValue == 0.0)  return desiredPos;

  halfSlopVal = slopPos / 2.0;

//--- If we don't know where the motor position is within the slop zone, ---
//--- and the max needed movement is less than the amount of slop, then we ---
//--- don't really know how far to move the motor, so just say to move it ---
//--- to the specified position, which will result in an error no larger ---
//--- than half the slopValue ---
  if (!slopPosKnown AND (fabs(desiredPos - slopPos) < slopValue))
    return desiredPos;

//--- We know where the device is and the motor pos relative to the slop ---
//--- zone so we can determine how much to move the motor and in which ---
//--- direction in order to push the device to the desired position ---

//--- move device in neg dir? ---
  if (desiredPos < slopPos)  return desiredPos - halfSlopVal;

//--- move device in pos dir? ---
  if (desiredPos > slopPos)  return desiredPos + halfSlopVal;

//--- no change so just stay at last read position ---
  return  motorPos;
}


//----------------------------------------------------------------------------
//  Return the amount of recent motor movement.  The motor position history
//  stores a new value every 10ms so that is the max resolution.
//----------------------------------------------------------------------------
float recentMovement(ushort msPeriod)
{
  int  i, x, numSamples;
  uint  minPos, maxPos;
  float  posChg;


  numSamples = msPeriod / SchedMS;
  if (numSamples > HistorySize)  numSamples = HistorySize;

//--- determine starting index ---
  x = lastHistIdx - numSamples;
  if (x < 0)  x += HistorySize;

  minPos = maxPos = (uint)motorHistory[(uint)x];

  for (i=0; i<numSamples; ++i)  {
    if ((uint)motorHistory[x] < minPos)  minPos = (uint)motorHistory[x];
    if ((uint)motorHistory[x] > maxPos)  maxPos = (uint)motorHistory[x];
    if (++x >= HistorySize)  x = 0;
  }

  posChg = (float)(maxPos - minPos) * ADCGain(MOTOR_POS) * readGain[MOTOR_POS];

  return fabs(posChg);
}


//----------------------------------------------------------------------------
//  Return the average pos value for the motor.  The motor position history
//  stores a new value every SchedMS ms so that is the max resolution.
//----------------------------------------------------------------------------
float avgPos(ushort msStartBack, ushort msPeriod)
{
  uint  u, numSamples;
  int  x;
  ulong  posSum;
  float  avgRawPos;


#define  ADCtoRead(sig, adcVal)  AInToRead(sig, ADCtoAIn(sig, adcVal))

//--- determine # of samples ---
  numSamples = msPeriod / SchedMS;
  if (numSamples < 2)  return  ADCtoRead(MOTOR_POS, motorHistory[lastHistIdx]);

  if (numSamples > HistorySize)  numSamples = HistorySize;

//--- determine starting index ---
  x = msStartBack / SchedMS;  // # of samples back from lastHistIdx
  if (x > HistorySize)  x = HistorySize;
  x = lastHistIdx - x;
  if (x < 0) x += HistorySize;

  for (u=0,posSum=0; u<numSamples; ++u)  {
    posSum += (ulong) motorHistory[(uint)x];
    if ((++x) >= HistorySize)  x = 0;
  }

  avgRawPos = (float) posSum / numSamples;

  return ADCtoRead(MOTOR_POS, avgRawPos);
}

//---------------------------------------------------------------------------
//  Keep tabs on the amount of readback noise we get when the controller
//  is actively maintaining a fixed position.
//
//  There are two steps to computing the noise value:
//
//  - Wait for a period of negligable movement.  This is determined by
//    comparing the Average position over time.  Averaged values are
//    needed to reduce the noise to a point where we can be fairly certain
//    as to whether or not there has been significant movement.
//
//  - Find the minimum and maximum position values recorded over the period
//    checked.  The absolute difference between these is the noise.
//
//  Note that a badly tuned PID loop will cause considerable variation in
//  the motor position and is indistinguishable from the "real" noise.
//  If we could accurately model the precise effect of the PID loop on the
//  readback values, then we could use that model to improve the PID loop so
//  this wouldn't even be a factor!
//
//  Returns true if the noise value is updated.
//---------------------------------------------------------------------------
bool  updateNoiseData(void)
{
  float  avg_pos, movement, newNoise, newMargin, err, noiseChg;
  float  *margin;
//  bool  updateFlash;


//--- lastNoiseChkTime also set at start of move to insure changes are ---
//--- not made at the beginning of a movement ---
  if (secSince(lastNoiseChkTime) < 5)  return false;  // 5 sec between tests

//--- skip if not active ctl mode ---
  if (Cur.ctlState != CS_ACTIVE_CTL)  return false;

  lastNoiseChkTime = GetMS();

  avg_pos = avgPos(5000, 5000);  // avg pos for last 5 seconds

  movement = fabs(avg_pos - lastAvgPos);  lastAvgPos = avg_pos;

  if (movement > readGain[MOTOR_POS] * 1.5)  return false;  // too much movement


//--- use new value or do weighted avg with prev value ---
  newNoise = fabs(recentMovement(5000));
  if (activeNoise == 0.0)
    activeNoise = newNoise;
  else
    activeNoise = (activeNoise*3.0 + newNoise) / 4.0;

//--- Update persistent value if difference significant enough --
  noiseChg = activeNoise - savedNoise;
  if (fabs(noiseChg) > readGain[MOTOR_POS] * 2.0)  savedNoise = activeNoise;


//--- also adjust the "stop margins" for BB Servo movement ctl ---
  KLOCK();
  if ((adjustPlusMargin OR adjustMinusMargin)
       AND !(bbMoveMinus OR bbMovePlus))  {
    err = motorPosSet - avg_pos;
    margin = adjustPlusMargin ? &plusMargin : &minusMargin;
    if (reqMovement > 0.0)
      newMargin = *margin - err;
    else
      newMargin = *margin + err;
    if ((newMargin >= 0.0) AND (newMargin < 1.0))
      *margin = (*margin * 3.0 + newMargin) / 4.0;
    adjustPlusMargin = adjustMinusMargin = false;
  }
  KUNLOCK();

//--- if significant chg from value in flash, schedule update of flash ---
//finish
//  if (RelativeDif(plusMargin, applData.plusMargin) > 0.30)  {
//    applData.plusMargin = plusMargin;  ChgWait(ONE_HOUR);  }
//  if (RelativeDif(minusMargin, applData.minusMargin) > 0.30)  {
//    applData.minusMargin = minusMargin;  ChgWait(ONE_HOUR);  }


  lastNoiseChgTime = GetMS();  return true;
}

//-----------------------------------------------------------------------------



//=============================================================================
//  Set the initial state we want the controller to start with.  This includes
//  "next" values that will be used to set the output signal states as soon
//  as the main application loop begins.
//=============================================================================
void setInitialState(void)
{
  if (current != applTask)  return;  // must only be called by "applic" task

//---------------------------------------------------------------
//  At this point the analog and digital input values have been
//  read so they can be used in determining what state to start
//  with for the outputs and any other values.
//--------------------------------------------------------------

//--- update current state, both for use here, and so when updateCurState() ---
//--- is called near the start of the application loop, there will be valid ---
//--- data to copy to the previous state ---
  updateCurState();

  setVal[CTL_VOLTS] = ctlVolts = 0.0;  // Don't want movement until we are ready
}


#pragma warn 775

//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)
{
  switch (ui)  {
    case MOTOR_POS:  *gain = 100.0;  break;
    case RF_PHASE:   *gain =  10.0;  break;  // 9V = 90 degrees
    default:         *gain =   1.0;
  }
  *offset = 0.0;
}

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)
{
  switch (ui)  {
    case CTL_VOLTS:     *gain = 100.0;  break;
    case EXTERN_METER:  *gain = 100.0;  break;
    default:            *gain =   1.0;
  }
  *offset = 0.0;
}

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---

//--- Analog IN ---
  fstrcpy(ainLabel[0], "Motor Pos");
  fstrcpy(ainLabel[1], "Phase Angle");
  fstrcpy(ainLabel[2], "+/- Drive Sw");
  fstrcpy(ainLabel[3], "ADC Ref");

//--- Analog OUT---
  fstrcpy(aoutLabel[0], "Ctl Voltage");
  fstrcpy(aoutLabel[1], "Extern Meter");


//--- Digitial IN ---
  fstrcpy(dinLabel[0], "Lock Btn Open");
  fstrcpy(dinLabel[1], "Not Plus Limit");
  fstrcpy(dinLabel[2], "Not Neg Limit");
  fstrcpy(dinLabel[3], "Set Btn Open");

  fstrcpy(dinLabel[4], "Amp Disabled");
  fstrcpy(dinLabel[5], "Fast Chgs");
  fstrcpy(dinLabel[6], "Loc/Rem Btn Open");
  fstrcpy(dinLabel[7], "Amp Fault");

  fstrcpy(dinLabel[8], "RF not OK");
  fstrcpy(dinLabel[9], "Not BB Servo");
  fstrcpy(dinLabel[10], "Not DC Servo");
  fstrcpy(dinLabel[11], "Extern Local Sw");

//--- Digitial OUT ---
  fstrcpy(doutLabel[0], "Local LED Off");
  fstrcpy(doutLabel[1], "Minus LED Off");
  fstrcpy(doutLabel[2], "Plus LED Off");
  fstrcpy(doutLabel[3], "Amp Disabled");

  fstrcpy(doutLabel[4], "NOT Locked");
  fstrcpy(doutLabel[5], "NO Fault");
  fstrcpy(doutLabel[6], "NOT Resetting Amp");
  fstrcpy(doutLabel[7], "Not (+) Move");

  fstrcpy(doutLabel[8], "Not (-) Move");
  fstrcpy(doutLabel[9], "CPU NOT BUSY");
  fstrcpy(doutLabel[10], "CPU LED");
}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

  activeNoise = readGain[MOTOR_POS] * 2.0;

  pWeight = 10.0;
  iWeight =  1.0;
  dWeight =  0.01;

  knobRateF = 1.0;
  knobRateS = 0.2;

  setRate = 1.0;

  switchRate = 1.0;

  maxActiveSecs = 300;

  rfLockResetDelay = 3000;

  maxPhaseErr = 2.0;

  plusMargin = minusMargin = readGain[MOTOR_POS];
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);


  Exit(0);
}


//=============================================================================
//  Use the values in prevLogInfo to determine if we should assert interlocks
//  for any devices any devices that were shut off due to a non-soft reset.
//=============================================================================
void checkPrevOnStates(void)
{

}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
//APPLCODE


  #define SharedApplLocals2
  #include <sharedAppl.c>


  #define SharedApplInit
  #include <sharedAppl.c>

//--- add application initialization code here ---
//APPLCODE

  prevEnableMsgTime = time(NULL);



  setInitialState();


  #define SharedApplInit2
  #include <sharedAppl.c>


//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>

  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();

    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
//APPLCODE
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,       "..." },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
