//--- RevSwitch.C --- 03/12/09 ---(05/23/05)---

#define DeviceVers     10

#ifdef RevSwitch
//Bundle # 082
  #define  DevName     "RevSwitch"
  #define  DeviceType   0x2e
  #define  NumADC    4  // # of A/D inputs used
  #define  NumDAC    1  // # of D/A outputs used
  #if (_BOARD_TYPE_ != RCM3200)
    #error "Compiling for RevSwitch but target board not set to RCM3200!"
  #endif
#endif

//-----------------------------------------------------------------------------
//
//  Copyright (c) 2003-2009 National Superconducting Cyclotron Laboratory,
//            Michigan State University, and Mark Davis
//
//-----------------------------------------------------------------------------
//
//  Program to control Polarity of a Power Supply
//
//  For notes on program structure, see the file:
//    ..\ZWORLD\NSCL_ZWORLD_LIBS\BASE_DLP_DLM_C.LIB
//
//-----------------------------------------------------------------------------
//  !!!!!! DON'T FORGET TO CHANGE DeviceVers ABOVE !!!!!!
//-----------------------------------------------------------------------------
//
//
// ***** DeviceVers 10 *****
//
//  2009-03-12 - Mark Davis -
//               - Increased interval used for debouncing digital inputs from
//                 150ms to 250ms
//               - Logic state no longer changes to OFF due to an overload
//                 condition unless the supply is actually turned off by the
//                 latches that supposedly caused the interlock.
//
// ***** DeviceVers 9 *****
//
//  2008-09-12 - Mark Davis -
//               - Made change to avoid reporting the last known serial reading
//                 as the current reading when a serial supply is first turned
//                 back on (RevSwitch in serial mode)
//
// ***** DeviceVers 8 *****
//
//  2008-09-05 - Mark Davis
//               - added "statedetails" command to allow turning on/off
//                 detailed logging of state change process.  Default is now
//                 to NOT log all the changes (which slowed things down when
//                 getting lots of changes quickly, as is the case with the
//                 new ramp-rate setting and with tuning via knobs).
//
//  2008-09-02 - Mark Davis -
//               - added rampRate to adjust setpoint gradually
//
// ***** DeviceVers 7 *****
//
//  2008-08-06 - Mark Davis -
//               - polarity of current readback now based on polarity of the
//                 load voltage above a small threshold.  Above a larger
//                 threshold, if the polarity of the setpoint and the polarity
//                 of the load voltage do not match, then it is an indication
//                 that the polarity switch did not happen and an interlock
//                 condition is asserted.
//
// ***** DeviceVers 6 *****
//
//  2007-11-01 - Mark Davis -
//               - fixed problem caused by having ampSet change again while the
//                 code was still busy dealing with a previous change to ampSet
//
// ***** DeviceVers 5 *****
//
//  2007-07-19 - Mark Davis -
//               - added debouncing to digital inputs to avoid problems caused
//                 by noisy/spurious signal changes
//
// ***** DeviceVers 4 *****
//
//  2006-09-19 - Mark Davis -
//               - transistor bias voltage is now maintained any time the power
//                 supply is on (not just in the IDLE state) (TCPS version)
//
//  2006-09-15 - Mark Davis -
//               - reversed logic of interlock control bit (should be normally
//                 high - brought low to re-energize relays)
//               - the sign of adcLoadAmps is now reversed like loadAmps is
//                 for the TCPS version
//               - the TCPS version now will turn back on if it was shut off
//                 while a negative setpoint was active
//
// ***** DeviceVers 3 *****
//
//  2006-08-24 - Mark Davis -
//               - Added sanity checking to responses from power supply.
//                 Serial read errors are now reported if it can't get a
//                 valid response.
//               - disabled use of maxVolts and maxAmps reading to set
//                 interlock for RevSwitch version
//
//  2006-08-23 - Mark Davis -
//               - Disabled the serial vs analog reading CALIB_ERR status bit
//
// ***** DeviceVers 2 *****
//
//  2006-07-12 - Mark Davis -
//               - Added ability to use serial or analog (from shunt) readback
//                 values for both the reported current/voltage readings and
//                 determining when it is safe to switch polarity.
//                 (set supplyType == 3 to use ser readings instead of analog)
//
//  2006-07-10 - Mark Davis -
//               - Added support for supplyType == 2 for TCPSs that do not use
//                 the mechanical motor-driven reversing switch (they use SCRs
//                 connected to the Normal and Reverse polarity relays).
//
//  2006-06-26 - Mark Davis -
//               - PS_Service is no longer shut down if the power supply
//                 keeps yacking....
//
//  2006-06-20 - Mark Davis -
//               - reduced required dif between load voltage for the amperage
//                 calibration setpoints to 0.5 volts for simple bench tests
//               - restored "lost" line of code for RevSwitch mode so that
//                 the switch actually reverses again (!!!!)
//
//  2006-03-10 - Mark Davis -
//               - ampSet now maintained while off and reasserted when the
//                 supply is turned back on
//
//  2006-01-24 - Mark Davis -
//               - started adding support for TCPS (TrimCoil Power Supply),
//                 a reversing-switch power supply regulated by an analog PI
//                 circuit on the controller board
//
//  2005-09-21 - Mark Davis -
//               - added internal interlock for load voltage or current
//                 exceeding max values
//               - changed meaning of gain/offset values so all computations
//                 are as follows (for both reads and sets):
//                   Actual = (ADC|DAC) * gain + offset
//                   [so, for sets:  DAC = (Actual - offset) / gain ]
//
//  2005-05-23 - Mark Davis -
//               - original version started using Starter code as base
//
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//
//  Operational and Programming NOTES:
//
//  - The Host expects the most significant bits at lower addresses while the
//    ZWorld keeps the most significant bits at the higher address
//
//  - This code assumes that, when recieving values via ModBus TCP protocol,
//    any values that span multiple registers (such as floating point values)
//    will be recieved via a Write Multiple Registers command rather than
//    multiple Write Single Register commands (the latter can cause anomolous
//    behavior).
//
//-----------------------------------------------------------------------------
//
//  The purpose of the Reversing Switch controller is to provide a bi-polar
//  output from a Power Supply that is NOT capable of bi-polar output.  This
//  is achieved by using a pair of relays to swap the (+) and (-) connections
//  from the power supply to the load.
//
//  To avoid possible damage to the load or the power supply, a change in the
//  output that requires changing the polarity is done by 1st setting the
//  output of the supply to zero and switching the polarity only when the
//  output is close to zero, and then setting the output level to the desired
//  level using the new polarity.
//
//-----------------------------------------------------------------------------
//
//  For the settings and readings to reflect actual conditions, each of the
//  xxxOffsets and xxxGain values need to be calibrated as follows:
//
//  (all commands are entered via a telnet connection to the controller)
//
//  The gain and offsets for the current reading (adcLoadAmpGain/Offset) and
//  the current setting (dacAmpSetGain/Offset or serAmpSetGain/Offset for
//  serial mode) are most easily set by running the supply at two significantly
//  different settings and using the "calib amps" commands to tell the
//  controller the actual/measured output for each setting and when to
//  computed updated values.
//
//  Example:
//        (#.# represents actual, measured PS output in amps for each setting)
//        (set 1st current, about 20% full scale)
//    calib amps 1 #.#  (remember ADC, DAC or Serial values for 1st point)
//        (set 2nd current, about 80% full scale)
//    calib amps 2 #.#  (remember ADC, DAC or Serial values for 2nd point)
//    calib amps set  (update gain/offsets for reading and setting)
//
//  If the default gain/offset for any other scaled values need to be altered,
//  then set conditions as needed to provide 2 different points and record the
//  corresponding ADC and/or DAC values from the telnet interface, compute the
//  corresponding gain and offset, and manually set the register values for
//  the desired values (e.g. adcPSVltOffset/Gain)
//
//-----------------------------------------------------------------------------



//=============================================================================
//  ==================== for RevSwitch controllers ==========================
//=============================================================================
//
//  There are two modes of operation for the controller:  Analog and Serial
//
//  In Analog mode, the current output of the power supply is controlled by
//  an analog voltage, and all readbacks are from analog inputs.  All analog
//  values are scaled as necessary, and the scalings can/must be set to
//  calibrate the values (more on this below).
//
//  In serial mode, control and monitoring of the power supply is done via
//  a serial connection.  The analog inputs are still in operation and are
//  used to monitor what the supply is actually doing to make sure it
//  agrees with what it says it is doing.  Also, the current to the load as
//  determined by reading the voltage across the shunt is used to determine
//  when it is safe to switch the polarity or turn the PS off.
//
//-----------------------------------------------------------------------------
//
//  For the settings and readings to reflect actual conditions, each of the
//  xxxOffsets and xxxGain values need to be calibrated as follows:
//
//  For two different output settings (one close to zero, the other close to
//  max the supply will be used at), record the corresponding input values
//  (the ADC reading and, for serial-controlled supplies, the device readback
//  read via serial) and output values (for Analog supplies, the DAC value;
//  for serial supplies, the device setpoint sent via serial).  The suggested
//  method of recording the values is to open a telnet connection to the
//  controller and make a copy of the main spplication-specific screen for
//  the two settings.
//
//  Using these values, compute and set the corresponding xxxOffset/Gain
//  register values.
//
//  NOTE:  The "calib amps 1|2 #" and "calib amps set" commands provide an
//         easier method for setting the following:
//
//           adcLoadAmpGain, adcLoadAmpOffset
//         PLUS:
//            (for Analog ctl mode):
//              dacAmpSetGain, dacAmpSetOffset
//            (for Serial ctl mode):
//              serAmpSetGain, serAmpSetOffset
//              serAmpMeasGain, serAmpMeasOffset
//
//  To use these commands, set the supply to a low setpoint (e.g. 1/8th the
//  max output), and type the following:
//
//    calib amps 1 #  (where # is the measured/computed actual load current)
//
//  Then set the supply to a high setpoint (e.g. 7/8th of the max output), and
//  type the following command:
//
//    calib amps 2 #  (where # is the measured/computed actual load current)
//
//  Then, to compute and save the Gain and Offset values listed above, the
//  the command:
//
//    calib amps set
//
//  Since these values are the only ones needed for setting and reading the
//  current, the default values for all other gains and offsets can often be
//  left as is.
//
//-----------------------------------------------------------------------------
//
//  The normal (non-energized) state of the relays that connect the power
//  supply to the load is such that the (-) output from the supply is connected
//  to both the (-) and (+) connections to the load, and the (+) output from
//  the supply is connected to nothing.  If both the relays are energized, then
//  the (+) output from the supply is connected to both, and the (-) output is
//  connected to nothing.
//
//  Note that in both these cases, this insures the supply's output sees no
//  load, and the inputs to the load are shorted.  This way, it does not matter
//  which state the relays are in when the controller boots, so long as they
//  are the same.  One OR the other relay must be energized to connect the load
//  to the power supply, and which one is energized is what determines the
//  polarity the load sees.
//
//-----------------------------------------------------------------------------
// Signal usage:
//
//  DI-10 - Ext Interlock - 1 = external interlock, 0 = no external interlock
//
//  DI-11 - On Feedback - 1 = PS has no power, 0 = power connected to PS
//
//  For power to be connected to the power supply, there must be no external
//  interlock, and the ZWorld must assert on ON condition
//
//------------------------------------------------------
//  D0-00 - Normal Polarity LED
//
//  DO-01 - Reverse Polarity LED
//
//  DO-07 - Relay that controls power going TO the Power Supply
//          0 = ground side of relay connected to ground
//          (also requires NO Ext. Interlock for PS power relay to energize)
//
//  DO-08 - Activity LED
//
//  DO-09 - Normal Polarity Relay - energized for Normal Polarity (supply's
//                       (+) goes to load's (+)
//
//  DO-10 - Reverse Polarity Relay - energized for Reversed Polarity (supply's
//                       (+) goes to load's (-)
//
//------------------------------------------------------
//  AI-00 - Current Reading - load current read as voltage across shunt
//                       resistor in series with the load.
//
//  AI-01 - Load Voltage - voltage as read across the load-side of the polarity
//                       polarity-mode relays and the shunt resistor.  Used to
//                       determine if the relays are functioning properly and
//                       that the shunt resistor is intact.
//
//  AI-02 - Supply Voltage - voltage as read from the output of the power supply
//                       (before going through the polarity relays and the shunt
//                       resistor).
//
//  AI-03 - Current Reading - from external source
//
//
//------------------------------------------------------
//  AO-00 - Voltage/Current Output - analog control of Power Supply output
//
//
//------------------------------------------------------
//  Serial D - I/O for control/monitoring of power supplies with a serial
//             interface.
//
//-----------------------------------------------------------------------------



#class auto
#memmap xmem
#nodebug

#ifdef  COMPILE_PRIMARY_PROG
  #error "Cannot compile as primary program."
  #error "Remove COMPILE_PRIMARY_PROG macro in 'Options | Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
#endif

#ifdef  DevelopMode
  #ifdef  COMPILE_SECONDARY_PROG
  #error "DevelopMode set - must NOT compile as secondary program."
  #error "Remove COMPILE_SECONDAY_PROG macro in 'Options | Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
  #endif
#else
  #ifndef  COMPILE_SECONDARY_PROG
  #error "Must compile as secondary program or DevelopMode."
  #error "Put COMPILE_SECONDARY_PROG macro in 'Options Compiler | #defines' options"
  #error "Must Recompile BIOS also. Hit Ctrl-Y then recompile"
  #endif
#endif


#define  NumChans  1  // # logical devices

//#define  UseEncoder  // uses optical encoder knob?


//#define DEBUG  true

#define  ChanMult     100  // multipler for chan number or'd to reg & coil #'s



//#define NSCLSerial     // support for NSCL RS-485 serial commands
//#define ModBusSerial   // support for ModBus RTU Serial commands

#define SERIAL_A_AVAIL
#if (_BOARD_TYPE_ != RCM3200)
  #define SERIAL_B_AVAIL
#endif
#define SERIAL_C_AVAIL
#define SERIAL_D_AVAIL
//#define SERIAL_E_AVAIL
//#define SERIAL_F_AVAIL


//=============================================================================
//                 Definitions the same for all controllers
//=============================================================================


//=============================================================================
//          Common stuff but with Controller-specific variations
//=============================================================================

#ifdef RevSwitch
  //----- bit usage masks for digitalIn -----
  #define  EXT_INTERLOCK  0x0400   // 10: extern ctl of power for controller
  #define  ON_FEEDBACK    0x0800   // 11: extern confirmation of AC to ctlr PS

  //----- bit usage masks for digitalOut -----
  #define  NORM_POLARITY_LED  0x000001  // 00: 0 ==> turns on Normal Polarity LED
  #define  REV_POLARITY_LED   0x000002  // 01: 0 ==> turns on Reverse Polarity LED
  #define  TURN_PS_OFF        0x000080  // 07: 0 ==> enables power to PS
  #define  ACTIVITY_LED       0x000100  // 08: activity LED
  #define  NORM_POLARITY_OFF  0x000200  // 09: disable Normal Polarity output
  #define  REV_POLARITY_OFF   0x000400  // 10: disable Reverse Polarity output

  //----- Analog Input usage -----
  #define  SHUNT_VOLTS        0   // volt across shunt resistor in series w/load
  #define  LOAD_VOLTAGE       1   // volt to load (after relays, shunt resistor)
  #define  SUPPLY_VOLTAGE     2   // volt from P.Supply (before relays/resistor)
  #define  CURRENT_READ       3   // current readback from external device

  //----- Analog Output usage -----
  #define  OUTPUT_CTL         0   // analog control of Power Supply output
#endif



//--- structure for Event messages ----
typedef struct  {
  int   msgID;        // which msg is it?
  unsigned int  seqNum;  // to identify last msg in seq of interdependent ones
  union {
    float newCur;     // new current setting (for NewSetPoint msg)
  } msgData;
}  EventMsg;

//--- values for EventMsg.msgID ---
#define  MSG_NoMsg           0
#define  MSG_TurnOn          1
#define  MSG_TurnOff         2
#define  MSG_ClearInterlock  3
#define  MSG_NewSetPoint     4

const char *cmdDesc[] = { "", "ON", "OFF", "CLEAR INTLK", "NEW SETPOINT" };


#use "Base_DLP_H.lib"


//----- values we want to be PERSISTENT (i.e. stored in flash) -----
//--- (search for PERSISTENT for places code normally added to support them) ---
//-- !!!!! Adding to this struct is OK, but never touch existing items !!!! ---
typedef struct  {
  uint   vers;             // version of this structure
  uint   devType;          // device type this data came from

  float  adcLoadAmpOffset; // offset for scaled shunt-volts --> amps
  float  adcLoadAmpGain;   // gain for scaled shunt-volts --> amps

  float  adcLoadVltOffset; // offset for reading load voltage via analog
  float  adcLoadVltGain;   // gain for reading load voltage via analog

  float  adcPSVltOffset;   // offset for reading PS output via analog
  float  adcPSVltGain;     // gain for reading PS output via analog

  float  adcExtAmpOffset;  // offset for reading cur from ext analog source
  float  adcExtAmpGain;    // gain for reading cur from ext analog source

  float  dacAmpSetOffset;  // offset for analog ctl of PS ouput
  float  dacAmpSetGain;    // gain for analog ctl of PS output

  float  serAmpMeasOffset; // offset for scaling cur reading from PS
  float  serAmpMeasGain;   // gain for scaling cur reading from PS

  float  serVltMeasOffset; // offset for scaling volt reading from PS
  float  serVltMeasGain;   // gain for scaling volt reading from PS

  float  serAmpSetOffset;  // offset for scaling cur setting sent to PS
  float  serAmpSetGain;    // gain for scaling cur setting sent to PS

  float  serVltSetOffset;  // offset for scaling volt setting sent to PS
  float  serVltSetGain;    // gain for scaling volt setting sent to PS

  float  maxVolts;         // max voltage allowed
  float  maxAmps;          // max current allowed

  float  safeSwAmps;       // max current allowed when switching polarity
  ulong  connectLoadDelay; // min delay in ms between PS On & load connected
  short  supplyType;       // type of power supply (see reg def's for values)

  float  biasVoltsOffset;  // offset for scaling reading of transistor bias V
  float  biasVoltsGain;    // gain for scaling reading of transistor bias V

  float  biasVoltsSet;     // transistor bank bias setpoint voltage
  float  biasErrorStart;   // bias volt error > this starts correction
  float  biasErrorStop;    // bias volt error < this stops correction

  float  rampRate;         // rate to adjust setpoint in amps/sec
}  ApplData;
#define  ApplDataVers   7

// search for _PERSISTENT_ to see where code normally added for saved values


//========================================
//===== Coil or ModBus bit registers =====
//========================================
//--- writing to a coil causes an action - values are not actually saved ---
//  See NewCoilValue() for details

//--- NOTE: 0-19 (00001-000020) are reserved for common values ---
#define  TurnOn           20  // [00021]
#define  TurnOff          21  // [00022]
#define  ClearInterlock   22  // [00023]
//                               =======
           #define  NUMCOILS         23


//======================================
//===== Read-Only ModBus registers =====
//======================================
//--- 30001-30050 (0-49) are reserved for generic/common values ---
//--- 30051-3xxxx are for controller specific values ---

RO_REG_F32( activeAmpSet,    30051 )
//                           =====
       #define  NUM_READ_REGS   52


//=======================================
//===== Read/Write ModBus registers =====
//=======================================
//--- Note that read/write registers can be defined as "Read Only" ---
//--- This allows the client to send write commands for a register to ---
//--- initiate some action without actually changing the value of the ---
//--- register.  Also, devices that have read-only values but multiple ---
//--- logical devices need to use the R/W registers, as the Read-Only ---
//--- set does not support multiple devices, and generally is limited ---
//--- to values that apply to all logical devices handled by one CPU ---

//--- Special handling for chgs to reg values is done by DevNewRegValue() ---

//--- The ChanMult value determines how many registers can be defined ---
//--- per logical device.  A value of 100 means 40001-40100 are for ---
//--- device 0, 40101-40200 are for device 1, etc.  Note that only ---
//--- one set of register definitions is required (i.e. define 40001 ---
//--- but not 40101, 40201, etc).

RW_REG_U32( coils31_00,       40001, RW ) /* alt access to coils */
RW_REG_U32( coils63_32,       40003, RW ) /* alt access to coils */

RW_REG_F32( ampSet,           40005, RW )  /* desired output */

RW_REG_U16( passiveSerial,    40007, WL )  /* disable auto non-polling cmds */
RW_REG_U16( supplyType,       40008, RW )  /* power supply type */
RW_REG_U16( devStatus,        40009, RO )  /* bit-mapped status */

RW_REG_F32( adcLoadAmps,      40010, RO )  /* Amps thru shunt measured via ADC */
RW_REG_F32( adcLoadAmpOffset, 40012, WL )  /* offset for shunt-V --> Amps */
RW_REG_F32( adcLoadAmpGain,   40014, WL )  /* gain for shunt-V --> Amps */

RW_REG_F32( adcLoadVolts,     40016, RO )  /* volt to load measured via ADC */
RW_REG_F32( adcLoadVltOffset, 40018, WL )  /* offset for rdg load V via ADC */
RW_REG_F32( adcLoadVltGain,   40020, WL )  /* gain for rdg load V via ADC */

RW_REG_F32( adcPSVolts,       40022, RO )  /* PS output volt measured via ADC */
RW_REG_F32( adcPSVltOffset,   40024, WL )  /* offset for rdg PS output via ADC */
RW_REG_F32( adcPSVltGain,     40026, WL )  /* gain for rdg PS output via ADC */

RW_REG_F32( adcExtAmps,       40028, RO )  /* Amps measured by ext analog dev */
RW_REG_F32( adcExtAmpOffset,  40030, WL )  /* offset for rdg amps from ext dev */
RW_REG_F32( adcExtAmpGain,    40032, WL )  /* gain for rdg amps from ext dev */

RW_REG_F32( dacAmpSetDev,     40034, RO )  /* output for DAC ctl of PS output */
RW_REG_F32( dacAmpSetOffset,  40036, WL )  /* offset for DAC ctl of PS output */
RW_REG_F32( dacAmpSetGain,    40038, WL )  /* gain for DAC ctl of PS output */

RW_REG_F32( serAmpMeasDev,    40040, RO )  /* amp measured by PS in dev units */
RW_REG_F32( serAmpMeasEng,    40042, RO )  /* amp measured by PS in eng units */
RW_REG_F32( serAmpMeasOffset, 40044, WL )  /* dev to engineering units offset */
RW_REG_F32( serAmpMeasGain,   40046, WL )  /* dev to engineering units gain */

RW_REG_F32( serVltMeasDev,    40048, RO )  /* vlt measured by PS in dev units */
RW_REG_F32( serVltMeasEng,    40050, RO )  /* vlt measured by PS in eng units */
RW_REG_F32( serVltMeasOffset, 40052, WL )  /* dev to engineering units offset */
RW_REG_F32( serVltMeasGain,   40054, WL )  /* dev to engineering units gain */

RW_REG_F32( serAmpSetSendDev, 40056, RO )  /* amp val sent to PS in dev units */
RW_REG_F32( serAmpSetReadDev, 40058, RO )  /* val in use by PS in dev units */
RW_REG_F32( serAmpSetReadEng, 40060, RO )  /* val in use by PS in eng units */
RW_REG_F32( serAmpSetOffset,  40062, WL )  /* dev to engineering units offset */
RW_REG_F32( serAmpSetGain,    40064, WL )  /* dev to engineering units gain */

RW_REG_F32( serVltSetSendDev, 40066, RO )  /* vlt val sent to PS in dev units */
RW_REG_F32( serVltSetReadDev, 40068, RO )  /* val in use by PS in dev units */
RW_REG_F32( serVltSetReadEng, 40070, RO )  /* val in use by PS in eng units */
RW_REG_F32( serVltSetOffset,  40072, WL )  /* dev to engineering units offset */
RW_REG_F32( serVltSetGain,    40074, WL )  /* dev to engineering units gain */

RW_REG_F32( maxVolts,         40076, RW )  /* max voltage allowed */
RW_REG_F32( maxAmps,          40078, RW )  /* max current allowed */

RW_REG_F32( safeSwAmps,       40080, WL )  /* max A allowed for polarity swap */
RW_REG_U32( connectLoadDelay, 40082, WL )  /* PS On --> load connected delay */

RW_REG_F32( biasVolts,        40084, RO )  /* transistor bank collector bias */
RW_REG_F32( biasVoltsOffset,  40086, WL )  /* for bias-V ADC --> actual V */
RW_REG_F32( biasVoltsGain,    40088, WL )  /* for bias-V ADC --> actual V */

RW_REG_F32( biasVoltsSet,     40090, RW )  /* desired transistor bank bias V */
RW_REG_F32( biasErrorStart,   40092, RW )  /* threshold to start correcting */
RW_REG_F32( biasErrorStop,    40094, RW )  /* threshold to stop correcting */

RW_REG_F32( rampRate,         40096, RW )  /* setpoint chg rate in amps/sec */
//                            =====
               #define  NUMREGS  97  // don't forget _x32 types use 2 regs!

// search for _PERSISTENT_ to see where code normally added for saved values


//--- bit usage for supplyType ---
#define  PS_TYPE_ANALOG      0  // analog controlled/monitored power supply
#define  PS_TYPE_SGA         1  // Sorensen SGA series via RS-232
#define  PS_TYPE_ANALOG_SCR  2  // TCPS that must turn PS on/off on its own
#define  PS_TYPE_SGA_FULL    3  // like 1, but use serial amp/volt readings

#define  AnalogCtl     ((supplyType != 1) AND (supplyType != 3))
#define  UseSerialRdg  (supplyType == PS_TYPE_SGA_FULL)

//finish
//--- bit usage for devStatus --- (some bits based on old TrimCoil PS device) ---
#define  DS_INTLK      0x0001   // Interlock condition active
#define  DS_ON         0x0002   // PS is ON and connected to the load
#define  DS_NORM_POL   0x0004   // Normal (+) Polarity state in effect
#define  DS_SER_ERR    0x0100   // Serial PS not responding
#define  DS_CALIB_ERR  0x0200   // serial & analog readings vastly different

RegInfo  rwRegs[NUMREGS+1];
RegInfo  roRegs[NUM_READ_REGS+1];

//=============================================================================
//              Definitions generally unique to this controller
//=============================================================================

//===== structures for serial cmd data and responses =====
// *** WARNING ***
// Remember that the host expects data in MSB 1st format while the Rabbit
// processor uses LSB 1st format for multi-byte formats
// *** WARNING ***

//===== function prototypes for controller-specific functions =====

void InterpretState();
void ComputeNextState();
void ChangeState();

#define TurnOnMsg()  {  AddMsg(MSG_TurnOn);  }

#define TurnOffMsg()  {  AddMsg(MSG_TurnOff);  }

#define ClearInterlockMsg()  {  AddMsg(MSG_ClearInterlock);  }

void NewSetPointMsg(float newCur);

void debounceDigIn();

//=============================================================================
//  GLOBAL variables
//=============================================================================

ApplData  applData;  // for storing application specific config data

int Process_PS_Resp(LinkProc *lp);


//==================== device specific stuff ====================

bool  devInitialized, newSetPoint, reversePolarity, enablePS, connectLoad,
      psNoResponse, restartPS_Service, serReadErr, calibErr, validAmpSetRdg,
      gettingReadings, enableSerial, manualOff, runCondsOK, extInterlock,
      clearInterlock, adjustBias, chgPolarity, showAllStateChgs;
int  ctlState, setPointState, psSrvIdx, numPolErrs;
float  setPoint, lastAmpSet, loadAmps, loadVolts;
ulong  lastSerAmpSetTime, lastSerVltSetTime, pendingResps, psOnTime,
       clearPulseStarted;

int  bitCounts[32];

ByteBuf  psCmdBuf;

#define  NumCmds  4
const char *psCmds[NumCmds] = {
  "SOUR:CURR?", "SOUR:VOLT?", "MEAS:CURR?", "MEAS:VOLT?" };


//--- values for ctlState ---
#define  CS_OFF            0
#define  CS_ON             1
#define  CS_EXT_INTLK      2
#define  CS_INT_INTLK      3

const char *CtlStateDesc[] = { "OFF    ", "ON     ", "EXT-ILK", "INT-ILK" };

//--- values for setPointState ---
#define  SPS_IDLE           0
#define  SPS_CONNECTING     1
#define  SPS_NEW_SETPOINT   2
#define  SPS_WAIT_FOR_ZERO  3
#define  SPS_SET_OUTPUT     4
#define  SPS_WAIT_PS_OFF    5
#define  SPS_WAIT_PS_ON     6
#define  SPS_SET_DELAY      7

const char *OnStateDesc[] = { "IDLE   ", "SER-COM", "NEW-SET", "WAIT-0 ",
                              "SET-OUT", "WAITOFF", "WAIT-ON", "SETWAIT" };

//------------------------------------------------------------------------------
#ifdef RevSwitch

xstring DevStateStr {
"ExtInterlock: xxx        Normal LED:  xxx              Shunt Volts: -##.###",
"ON Feedback:  xxx        Reverse LED: xxx              Load Volts:  -##.###",
"                                                       Supply Volt: -##.###",
"                         Turn PS ON:  xxx              Ext.Current: -##.###",
" ",
"                         Normal  Polarity: xxx         Output Ctl:  -##.###",
"                         Reverse Polarity: xxx ",
" ",
"[############] [#####] [####] ",
"           ----------- via Serial -----------     --- via Analog ---",
"           --Set--  (dev)     --Read-- (dev)       --Load--  -Supply-",
"Current:   ###.###(###.###)   ###.###(###.###)    -###.###   **N/A**   Amps",
"Voltage:   ###.###(###.###)   ###.###(###.###)    -###.###  -###.###   Volts",
" ",
"Setpoint: -###.### Amps (-###.###) [.......]",
""
};


const DigFmtInfo digInFmt[] = {
  {  0, 0, ""     ,0,  ""    ,0 },  //  0
  {  0, 0, ""     ,0,  ""    ,0 },  //  1
  {  0, 0, ""     ,0,  ""    ,0 },  //  2
  {  0, 0, ""     ,0,  ""    ,0 },  //  3
  {  0, 0, ""     ,0,  ""    ,0 },  //  4
  {  0, 0, ""     ,0,  ""    ,0 },  //  5
  {  0, 0, ""     ,0,  ""    ,0 },  //  6
  {  0, 0, ""     ,0,  ""    ,0 },  //  7
  {  0, 0, ""     ,0,  ""    ,0 },  //  8
  {  0, 0, ""     ,0,  ""    ,0 },  //  9
  { 14, 3, "NO "  ,0,  "YES" ,1 },  // 10  Interlock (! "On Command")
  { 14, 4, "ON "  ,0,  "OFF" ,1 },  // 11  ON Feedback
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
  { 38, 3, "ON "  ,0,  "OFF" ,0 },  //  0 - Normal Polarity LED
  { 38, 4, "ON "  ,0,  "OFF" ,0 },  //  1 - Reverse Polarity LED
  {  0, 0, ""     ,0,  ""    ,0 },  //  2
  {  0, 0, ""     ,0,  ""    ,0 },  //  3
  {  0, 0, ""     ,0,  ""    ,0 },  //  4
  {  0, 0, ""     ,0,  ""    ,0 },  //  5
  {  0, 0, ""     ,0,  ""    ,0 },  //  6
  { 38, 6, "ON "  ,0,  "OFF" ,1 },  //  7 - PS On/Off
  {  0, 0, ""     ,0,  ""    ,0 },  //  8 - Activity LED
  { 43, 8, "YES"  ,0,  "NO " ,0 },  //  9 - Normal Polarity Enable/Disable
  { 43, 9, "YES"  ,0,  "NO " ,0 },  // 10 - Reverse Polarity Enable/Disable
  { -1,-1, "",0, "",0 }
};

const AnalogFmtInfo  adcFmt[] = {
  { 68, 3, 7,3 },
  { 68, 4, 7,3 },
  { 68, 5, 7,3 },
  { 68, 6, 7,3 },
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { 68, 8, 7,3 },
  { -1,-1, 0,0 }
};

#endif  //RevSwitch
//------------------------------------------------------------------------------


xstring ApplStrs  {
#define   HelpStr                 ApplStr(0)
  "A=ADC fmt, H=Help, S=Sock/Ser, X=eXt cmd, Z=Zero errs",

#define   ExtHelpStr              ApplStr(1)
  "Bye, Calib, Help, ID, IP, Reboot, Reg, Serial, Track, <reg>",

#define   PSNoRespMsg             ApplStr(2)
  "PS Not Responding on serial interface",

#define   UnrecogCmdMsg           ApplStr(3)
  "Unrecognized command: %s",

#define   IncompleteCmdMsg        ApplStr(4)
  "Incomplete command",

""
};   //NewStr


xstring ApplHelpStrs {
  "StateDetails on|off          - show/hide extra state change details",
  "Calib amps 1|2 <amps>        - save state data for 1 of 2 setpoints",
  "Calib amps set               - update calibration based on 2 setpoints",
  "Clear                        - Clear internal interlock",
  "Off                          - Turn Power Supply Off",
  "On                           - Turn PS ON",
#ifdef RevSwitch
  "PS cmd                       - Send cmd to Power Supply and display response",
#endif
  ""
};

#use "Base_DLP_C.lib"


//-----------------------------------------------------------------------------
#ifdef RevSwitch

void PS_Server(LinkProc *lp);
SerialLinkParams  PS_Params; // {  // baud rate not fixed
//  3,               // Port D
//  19200 or ???     // 19200 for Sorenson SGA
//  8,               // data bits
//  PARAM_NOPARITY   // even parity, 1 stop bit
//  0                // no special flag settings
//};
const ServiceDesc  PS_Service = {
  #define  MaxPS_Clients   1    // Max # clients
  "PS_Serial",       // Name
  &PS_Params,        // parameters needed to construct a ComLink object
  LT_ASYNC,          // type of ComLink object/connection
  LogYes,            // log std event messages
  NetClientNo,       // don't incl in count of appl-type network clients
  0,                 // no timeout for idle connections (secs)
  MaxPS_Clients,     // max # of concurrent clients
  &PS_Server         // process that manages this connection
};


#define  RCV_BUF_SIZE  512
#define  MaxCmdSize     64

typedef struct {
  bool   inUse;            // in-use flag
  ulong  cmdCount;         // # cmds rcvd on this connection
  ulong  xmitStartTime;    // time xmit was started
  ulong  startIdleTime;    // last time started waiting for idle period
  ulong  rcvdBeforeCmd;    // rcvd count just before last xmit
  bool   userCmd;          // last cmd sent was entered by the user
  int    cmdLen;           // length of str in cmdBuf[]
  char   cmdBuf[MaxCmdSize];  // last cmd sent
  int    rcvCount;         // # chars of respBuf[] used so far
  char   respBuf[RCV_BUF_SIZE];
} PS_Client;

PS_Client   PS_Clients[MaxPS_Clients];

char  psCmd[MaxCmdSize];

#endif  //RevSwitch
//-----------------------------------------------------------------------------


//=============================================================================
//          Standard functions with APPLICATION SPECIFIC variations
//=============================================================================


//-----------------------------------------------------------------------------
//  Initialization for processing of commands
//-----------------------------------------------------------------------------
int InitCmds(int chanNum)
{
  devInitialized = true;
  ShowMsg("Device Initialized...");

  return 0;
}

//-----------------------------------------------------------------------------
//  Put device in state that is safe for controller to reboot
//-----------------------------------------------------------------------------
void PrepForReboot()
{
  ulong  lastWriteTime;


//... make changes to digitalOut, voltsOut[], etc ...

//--- wait for latest changes to take effect ---
  lastWriteTime = writeTime;  while (writeTime == lastWriteTime);
}


//-----------------------------------------------------------------------------
//  check validity of application/configuration data.  If any of it is not
//  usuable, change it to a reasonable default value and save any changes
//  to flash memory
//-----------------------------------------------------------------------------
void ValidateApplData(bool readFirst)
{
  int  stat;
  bool  dataChgs;
  char  msg[90];


  dataChgs = false;

  if (readFirst)
    if ((stat = Block_Read(BID_APPL_DATA, 0, &applData, sizeof(applData))) < 2)
      sprintf(msg, ApplDataReadErrMsg, stat);

  if (applData.devType != DeviceType)  {
    LogMsg("Resetting Application data...");
    memset(&applData, 0, sizeof(applData));
    applData.vers = ApplDataVers;  applData.devType = DeviceType;
    Block_SetSize(BID_APPL_DATA, sizeof(applData));  dataChgs = true;
  }

#define CheckRange(attr)  {  \
  if ((applData.attr < (attr##Min)) OR (applData.attr > (attr##Max))) {  \
    applData.attr = (attr##Def);  dataChgs = true;  } }

#define CheckRangeNZ(attr)  { \
  if ( (applData.attr < (attr##Min)) OR (applData.attr > (attr##Max))  \
      OR (applData.attr == 0.0) )  {  \
    applData.attr = (attr##Def);  dataChgs = true;  }  \
}


//  if (applData.attr not valid)  {
//    applData.attr = defaultValue;
//    dataChgs = true;
//  }

//  #define attrMin  #.#
//  #define attrMax  #.#
//  #define attrDef  #.#
//    CheckRange( attr );

  #define adcLoadAmpOffsetMin -100.0
  #define adcLoadAmpOffsetMax  100.0
  #define adcLoadAmpOffsetDef    0.0
    CheckRange( adcLoadAmpOffset );

  #define adcLoadAmpGainMin  -500.0
  #define adcLoadAmpGainMax   500.0
  #define adcLoadAmpGainDef   -40.0
    CheckRangeNZ( adcLoadAmpGain );

  #define adcLoadVltOffsetMin  -5.0
  #define adcLoadVltOffsetMax   5.0
  #define adcLoadVltOffsetDef   0.0
    CheckRange( adcLoadVltOffset );

  #define adcLoadVltGainMin  -10.0
  #define adcLoadVltGainMax   10.0
  #define adcLoadVltGainDef   -1.0
    CheckRangeNZ( adcLoadVltGain );

  #define adcPSVltOffsetMin  -5.0
  #define adcPSVltOffsetMax   5.0
  #define adcPSVltOffsetDef   0.0
    CheckRange( adcPSVltOffset );

  #define adcPSVltGainMin  -10.0
  #define adcPSVltGainMax   10.0
  #define adcPSVltGainDef   -1.0
    CheckRangeNZ( adcPSVltGain );

  #define adcExtAmpOffsetMin  -100.0
  #define adcExtAmpOffsetMax   100.0
  #define adcExtAmpOffsetDef     0.0
    CheckRange( adcExtAmpOffset );
  #define adcExtAmpGainMin  -500.0
  #define adcExtAmpGainMax   500.0
  #define adcExtAmpGainDef    50.0
    CheckRangeNZ( adcExtAmpGain );

  #ifdef RevSwitch
    #define dacAmpSetOffsetMin  -500.0
    #define dacAmpSetOffsetMax   500.0
    #define dacAmpSetOffsetDef     0.0
  #endif
    CheckRange( dacAmpSetOffset );
  #ifdef RevSwitch
    #define dacAmpSetGainMin  -100.0
    #define dacAmpSetGainMax   100.0
    #define dacAmpSetGainDef    40.0
  #endif
    CheckRangeNZ( dacAmpSetGain );

  #define serAmpMeasOffsetMin  -5.0
  #define serAmpMeasOffsetMax   5.0
  #define serAmpMeasOffsetDef   0.0
    CheckRange( serAmpMeasOffset );

  #define serAmpMeasGainMin  -10.0
  #define serAmpMeasGainMax   10.0
  #define serAmpMeasGainDef    1.0
    CheckRangeNZ( serAmpMeasGain );

  #define serVltMeasOffsetMin  -5.0
  #define serVltMeasOffsetMax   5.0
  #define serVltMeasOffsetDef   0.0
    CheckRange( serVltMeasOffset );

  #define serVltMeasGainMin  -10.0
  #define serVltMeasGainMax   10.0
  #define serVltMeasGainDef    1.0
    CheckRangeNZ( serVltMeasGain );

  #define serAmpSetOffsetMin  -5.0
  #define serAmpSetOffsetMax   5.0
  #define serAmpSetOffsetDef   0.0
    CheckRange( serAmpSetOffset );

  #define serAmpSetGainMin  -10.0
  #define serAmpSetGainMax   10.0
  #define serAmpSetGainDef    1.0
    CheckRangeNZ( serAmpSetGain );

  #define serVltSetOffsetMin  -5.0
  #define serVltSetOffsetMax   5.0
  #define serVltSetOffsetDef   0.0
    CheckRange( serVltSetOffset );

  #define serVltSetGainMin  -10.0
  #define serVltSetGainMax   10.0
  #define serVltSetGainDef    1.0
    CheckRangeNZ( serVltSetGain );

  #define maxVoltsMin     0.001
  #define maxVoltsMax  1000.0
  #define maxVoltsDef    50.0
    CheckRangeNZ( maxVolts );

  #define maxAmpsMin     0.001
  #define maxAmpsMax  1000.0
  #define maxAmpsDef   500.0
    CheckRangeNZ( maxAmps );

  #define safeSwAmpsMin  0.001
  #define safeSwAmpsMax 10.0
  #define safeSwAmpsDef  0.5
    CheckRangeNZ( safeSwAmps );

  #define connectLoadDelayMin     0
  #define connectLoadDelayMax 60000L
  #define connectLoadDelayDef  2000
    CheckRange( connectLoadDelay );

  #define supplyTypeMin  0
  #define supplyTypeMax  3
  #define supplyTypeDef  PS_TYPE_ANALOG
    CheckRange( supplyType );

  #define biasVoltsOffsetMin -5.0
  #define biasVoltsOffsetMax  5.0
  #define biasVoltsOffsetDef  0.0
    CheckRange( biasVoltsOffset );

  #define biasVoltsGainMin  -10.0
  #define biasVoltsGainMax   10.0
  #define biasVoltsGainDef   -5.5
    CheckRangeNZ( biasVoltsGain );

  #define biasVoltsSetMin  -50.0
  #define biasVoltsSetMax   50.0
  #define biasVoltsSetDef   10.0
    CheckRangeNZ( biasVoltsSet );

  #define biasErrorStartMin  0.1
  #define biasErrorStartMax  10.0
  #define biasErrorStartDef  2.0
    CheckRangeNZ( biasErrorStart );

  #define biasErrorStopMin  0.1
  #define biasErrorStopMax  (applData.biasErrorStart / 2.0)
  #define biasErrorStopDef  (applData.biasErrorStart / 4.0)
    CheckRangeNZ( biasErrorStop );

  #define debounceMSMin  10
  #define debounceMSMax  1000
  #define debounceMSDef  50
//    CheckRange( debounceMS );

  #define rampRateMin  0.01
  #define rampRateMax  100.0
  #define rampRateDef  5.0
    CheckRangeNZ( rampRate );

// search for _PERSISTENT_ to see where code normally added for saved values


  if (dataChgs)  SaveApplData(true);
}



//-----------------------------------------------------------------------------
//  process a multi-character command from the diagnostic terminal/console
//-----------------------------------------------------------------------------
void ProcessStrCmd(char *cmd, LinkProc *lp)
{
  char  *ptr;
  int  i, idx, chan, state, digState;
  ulong  mask;
  bool  validCmd;
  char  *msg, *temp, *cmdStr;
  float volts;
  float  deltaActual, deltaADC, deltaDAC, deltaSerSetDev, deltaSerMeasDev;
  static float  actualAmps[2], adcShunt[2], dacSet[2],
                serialSetDev[2], serialMeasDev[2];


  StackPtr();

  msg = getBuf(90 + 90 + 90);  temp = msg + 90;  cmdStr = temp + 90;

  strcpy(cmdStr, cmd);  ptr = strtok(cmd, " ");
  state = 0;  validCmd = false;
  strcpy(msg, "cmd parser error");

  while (ptr AND (state != 999))  {

    switch (state)  {

      case 0:  // which command did they type?
        if (strcmpi(ptr, "calib") == 0)  { state = 105;  break; }
        if (strcmpi(ptr, "clear") == 0)  {
          ClearInterlockMsg();  state = 999;  *msg = 0;  break; }
        if (strcmpi(ptr, "off") == 0)  {
          TurnOffMsg();  state = 999;  *msg = 0;  break; }
        if (strcmpi(ptr, "on") == 0)  {
          TurnOnMsg();  state = 999;  *msg = 0;  break; }
        #ifdef RevSwitch
          if (strcmpi(ptr, "ps") == 0)  { state = 104;  break; }
        #endif
        if (strcmpi(ptr, "statedetails") == 0)  { state = 107;  break; }
        sprintf(msg, UnrecogCmdMsg, ptr);  state = 999;  break;

    #ifdef RevSwitch
      case 104:  // set next cmd to send to Power Supply
        strncpy(psCmd, ptr, sizeof(psCmd));
        psCmd[sizeof(psCmd)-2] = 0;
        *msg = 0;  validCmd = true;  state = 999;  break;
    #endif

      case 105:  // "calib" cmd
        if (strcmpi(ptr, "amps") == 0)  { state = 106;  break; }
        break;


      case 106:  // "calib amps" cmd
      //--- update gain and offset calibrations based on the 2 setpoints ---
      //--- ("calib amps set") ---
        if (strcmpi(ptr, "set") == 0)  {
          deltaActual = actualAmps[1] - actualAmps[0];
          deltaADC = adcShunt[1] - adcShunt[0];
          if (fabs(deltaADC) < 0.5)  {
            strcpy(msg, "Unusable ADC data points");  state = 999;  break; }

          ++criticalSection;
          applData.adcLoadAmpGain = adcLoadAmpGain =
                 deltaActual / deltaADC;
          applData.adcLoadAmpOffset = adcLoadAmpOffset =
                 actualAmps[1] - adcLoadAmpGain * adcShunt[1];
          --criticalSection;

          if (AnalogCtl)  {
            deltaDAC = dacSet[1] - dacSet[0];
            if (fabs(deltaDAC) < 2.0)  {
              strcpy(msg, "Unusable DAC data points");
              ChgWait(5);  state = 999;  break;
            }
            ++criticalSection;
            applData.dacAmpSetGain = dacAmpSetGain =
                 deltaActual / deltaDAC;
            applData.dacAmpSetOffset = dacAmpSetOffset =
                 actualAmps[1] - dacAmpSetGain * dacSet[1];
            --criticalSection;
          }
          else  {
            deltaSerSetDev = serialSetDev[1] - serialSetDev[0];
            if (fabs(deltaSerSetDev) < 5.0)  {
              strcpy(msg, "Unusable Serial data points");
              ChgWait(5);  state = 999;  break;
            }
            ++criticalSection;
            applData.serAmpSetGain = serAmpSetGain =
                 deltaActual / deltaSerSetDev;
            applData.serAmpSetOffset = serAmpSetOffset =
                 actualAmps[1] - serAmpSetGain * serialSetDev[1];
            --criticalSection;

            deltaSerMeasDev = serialMeasDev[1] - serialMeasDev[0];
            if (fabs(deltaSerMeasDev) < 5.0)  {
              strcpy(msg, "Unusable Serial data points");  state = 999;  break; }
            ++criticalSection;
            applData.serAmpMeasGain = serAmpMeasGain =
                 deltaActual / deltaSerMeasDev;
            applData.serAmpMeasOffset = serAmpMeasOffset =
                 actualAmps[1] - serAmpMeasGain * serialMeasDev[1];
            --criticalSection;
          }
          ChgWait(5);

          LogMsg("New Amp set/read calibration being saved...");
          state = 999;  *msg = 0;  validCmd = true;  break;
        }

      //--- save current state corresponding to the specified actual output ---
      //--- for one of 2 setpoints ("calib amps 1|2 amps")
        idx = atoi(ptr);
        if ((idx < 1) OR (idx > 2))  {
          strcpy(msg, "Invalid point #");  state = 999;  break; }
        --idx;
        if (! (ptr = strtok(NULL, " ")))  break;

        actualAmps[idx] = atof(ptr);
        adcShunt[idx] = voltsIn[SHUNT_VOLTS];
        dacSet[idx] = voltsOut[OUTPUT_CTL];
        serialSetDev[idx] = serAmpSetReadDev;
        serialMeasDev[idx] = serAmpMeasDev;

        state = 999;  *msg = 0;  validCmd = true;  break;

    case 107:  // "statedetails" cmd
      showAllStateChgs = (strcmpi(ptr, "on") == 0);  validCmd = true;
      break;


    }

    if (ptr)  ptr = strtok(NULL, " ");
  }
  strCmdState = false;



//--- what state were we in when we ran out of cmd string to parse? ---
  switch (state)  {

    case 999:  // msg supplied by cmd parser state machine
      break;

    default:
      strcpy(msg, "Incomplete command");
  }

  if (! validCmd)
    ProcessStrCmdCommon(cmdStr, msg, lp);  // not valid dev cmd - check generic ones
  else
    if (*msg)
      if (StreamDisp)
        ShowMsg(msg);
      else  {
        respShownTime = MS_TIMER - 5000;
        DispStr(0,23, msg, true);
        respShownTime = MS_TIMER;
      }

  freeTo(msg);
}


//-----------------------------------------------------------------------------
//  process single-char commands from the diagnostic console
//-----------------------------------------------------------------------------
void CharCmd(char cmd)
{
  CharCmdCommon(cmd);

  cmd |= 0x20;

//--- device-specfic stuff ---
//  if (cmd == '?') ...

}


//-----------------------------------------------------------------------------
//  If the specified digital input bit has changed since last checked, then
//  log a message regarding the change and clear the change bit in digInChanges
//-----------------------------------------------------------------------------
void msgIfChg(ulong bitMask, char *baseMsg, char *onSuffix, char *offSuffix)
{
  char  msg[90];

  if (digInChanges & bitMask)  {
    strcpy(msg, "*** ");
    strcat(msg, baseMsg);
    strcat(msg, (digitalIn & bitMask) ? onSuffix : offSuffix);
    strcat(msg, " ***");
    LogMsg(msg);
    digInChanges ^= bitMask;
  }
}

//-----------------------------------------------------------------------------
//  get next msg in event msg queue - returns true if a msg read, false if
//  no msg or an error occured
//-----------------------------------------------------------------------------
bool getMsg(EventMsg *eventMsg)
{
  eventMsg->msgID = MSG_NoMsg;

  if (Buf_Used(&eventMsgQueue) < 1) return false;

  if (lastMsgProcessed == lastMsgToProcess)  return false;

  if (Buf_Remove(&eventMsgQueue, (char *)eventMsg, sizeof(EventMsg))
      != sizeof(EventMsg))  { LogErr(EventMsgErr);  return false;  }

  lastMsgProcessed = eventMsg->seqNum;

  return true;
}


//-----------------------------------------------------------------------------
//  Update data based on digital and analog input values and determine new
//  values for the digital and analog outputs
//
//  !!!!! WARNING !!!!!
//
//  This function is called by the timer-interrupt driven function that reads
//  and writes the Analog and Digital I/O values.  It should be limited to
//  processing current input values and updating output values that will become
//  the active values on the next interrupt (currently, this interuupt occurs
//  every 10ms, so this routine must take considerably less time than that to
//  do EVERYTHING it needs to).
//-----------------------------------------------------------------------------
nodebug void UpdateDeviceState()
{
  InterpretState();
  ComputeNextState();
  ChangeState();
}


//==============================================================================
#ifdef RevSwitch
//==============================================================================

//-----------------------------------------------------------------------------
//  Compute and update values/flags based on current state and the just-read
//  state of the digital and analog inputs
//
//  All references and interpretation of digital and analog inputs should be
//  done in this function which should set flags and values based on the
//  just-read state of the inputs.
//
//  ComputeNextState() uses the values computed/updated by this function.
//
//  ChangState() will handle all changes to outputs based on the values
//  computed/updated by ComputeNextState()
//-----------------------------------------------------------------------------
nodebug void InterpretState()  // RevSwitch version
{
//--- update values that are dependent only upon input values that are not ---
//--- altered by this code ---
  adcLoadAmps  = voltsIn[SHUNT_VOLTS]    * adcLoadAmpGain + adcLoadAmpOffset;
  adcLoadVolts = voltsIn[LOAD_VOLTAGE]   * adcLoadVltGain + adcLoadVltOffset;
  adcPSVolts   = voltsIn[SUPPLY_VOLTAGE] * adcPSVltGain   + adcPSVltOffset;
  adcExtAmps   = voltsIn[CURRENT_READ]   * adcExtAmpGain  + adcExtAmpOffset;

  debounceDigIn();

  extInterlock = BOOL(digitalIn & EXT_INTERLOCK);

  if (!UseSerialRdg OR !gettingReadings OR !validAmpSetRdg)  {
    loadAmps = adcLoadAmps;  loadVolts = adcLoadVolts;  }
  else  {
    loadAmps = serAmpMeasEng;  loadVolts = serVltMeasEng;  }

  runCondsOK = true;

  msgIfChg(EXT_INTERLOCK, "External Interlock ", "Bad", "OK");
  msgIfChg(ON_FEEDBACK,   "PS Feedback: ",       "On",  "Off");  //backwards...

  if ((fabs(loadAmps) > maxAmps) OR (fabs(loadVolts) > maxVolts))  {
    if (ctlState != CS_INT_INTLK)  LogMsg("Max cur or volt exceeded!");
//    runCondsOK = false;
  }

//--- set values/flags related to serial PS interface ---
  if (AnalogCtl)  {
    calibErr = serReadErr = false;
    serAmpMeasEng = serVltMeasEng = serAmpSetReadEng = serVltSetReadEng = 0.0;
  }
  else  {
//    calibErr = (fabs(adcLoadAmps) - fabs(serAmpMeasEng) > 10.0);
    serReadErr = (pendingResps > 5);
    gettingReadings = (!serReadErr AND validAmpSetRdg AND enablePS);
  }

}  // RevSwitch version

//-----------------------------------------------------------------------------
//  Determine the values for the next state based on the current state
//
//  This code should NOT refer to digital or analog I/O values directly.
//
//  All interpetation of the I/O values should be done in InterpretState()
//  which computes/updates any values this function needs.
//
//  All changes to outputs should be done in ChangeState() based on values
//  computed/updated in this function.
//-----------------------------------------------------------------------------
void ComputeNextState()  // RevSwitch version
{
  int  beforeState, beforeCtlState;
  bool  gotMsg;
  float  ampSetDiff, ampChg;
  char  msg[90];
  EventMsg  eventMsg;
  static bool  spStateChgd, ctlStateChgd;
  static float  nextAmpSet;


  if (!enablePS)  enableSerial = false;

  clearInterlock = false;

//--- update state based on current readings and event msgs ---
  do  {  // RevSwitch version
    if ((gotMsg = getMsg(&eventMsg)))  {
      strcpy(msg, "===== ");  strcat(msg, cmdDesc[eventMsg.msgID]);
      strcat(msg, " command =====");  LogMsg(msg);
    }

    beforeState = setPointState;  beforeCtlState = ctlState;

  //--- state changes/etc that are independent of current state ---
    if (eventMsg.msgID == MSG_TurnOff)  ctlState = CS_OFF;
    if (manualOff)  ctlState = CS_OFF;
    if (extInterlock)  ctlState = CS_EXT_INTLK;
    if (! runCondsOK)  ctlState = CS_INT_INTLK;
    if (eventMsg.msgID == MSG_ClearInterlock)  clearInterlock = true;

  //--- settings common to multiple states ---
    if (ctlState != CS_ON)  {
      setPoint = activeAmpSet = lastAmpSet = nextAmpSet = 0.0;
      reversePolarity = chgPolarity = false;
      setPointState = SPS_IDLE;  pendingResps = 0;
      enablePS = newSetPoint = false;
    }

  //--- state-specific logic ---
    switch (ctlState)  {

      case CS_OFF:
        if (ctlStateChgd)  LogMsg("CS_OFF");
        validAmpSetRdg = gettingReadings = false;
        if (eventMsg.msgID == MSG_TurnOn)  {
          ctlState = CS_ON;  setPointState = SPS_CONNECTING;
          psOnTime = MS_TIMER;
        }
        break;

      case CS_ON:
        if (ctlStateChgd)  LogMsg("CS_ON");
        enablePS = true;

        if (eventMsg.msgID == MSG_NewSetPoint)
          ampSet = eventMsg.msgData.newCur;

        if (activeAmpSet != ampSet)  {
          ampSetDiff = ampSet - activeAmpSet;
          ampChg = rampRate / 100.0;
          if (FABS(ampSetDiff) > ampChg)  {
            if (ampSetDiff < 0.0)  ampChg = -ampChg;
            activeAmpSet += ampChg;
          }
          else
            activeAmpSet = ampSet;
          newSetPoint = true;
        }

        switch (setPointState)  {
          case SPS_IDLE:
            if (showAllStateChgs)  if (spStateChgd)  LogMsg("  IDLE");
            if (newSetPoint)  setPointState = SPS_NEW_SETPOINT;
            break;

          case SPS_CONNECTING:
            if (spStateChgd)  LogMsg("  CONNECTING");
            if (elapsedMS(psOnTime) < connectLoadDelay)  break;
            enableSerial = true;
            if (! AnalogCtl)  {  // wait for confirmation of 0 setting
              if (! validAmpSetRdg)  break;
              if (fabs(serAmpSetReadEng) > safeSwAmps)  break;
            }
            connectLoad = true;  setPointState = SPS_WAIT_FOR_ZERO;
            break;

          case SPS_NEW_SETPOINT:
            if (showAllStateChgs)  LogMsg("  NEW_SETPOINT");
            nextAmpSet = activeAmpSet;
            if (SIGN(nextAmpSet) == SIGN(lastAmpSet))
              setPointState = SPS_SET_OUTPUT;
            else  {
              chgPolarity = true;  setPoint = 0.0;
              setPointState = SPS_WAIT_FOR_ZERO;
            }
            break;

          case SPS_WAIT_FOR_ZERO:
            if (spStateChgd)  LogMsg("  WAIT_FOR_ZERO");
            if (fabs(loadAmps) > safeSwAmps)  break;
            setPointState = SPS_SET_OUTPUT;
            break;

          case SPS_SET_OUTPUT:
            if (showAllStateChgs)  LogMsg("  SET_OUTPUT");
            reversePolarity = (nextAmpSet < 0.0);
            setPoint = lastAmpSet = nextAmpSet;
            if (activeAmpSet == nextAmpSet)  newSetPoint = false;
            setPointState = SPS_IDLE;  break;
        }
        break;


      case CS_EXT_INTLK:  // non-latched Interlock
        if (ctlStateChgd)  LogMsg("CS_EXT_INTLK");
        enablePS = false;
        if (! extInterlock)  ctlState = CS_OFF;
        break;


      case CS_INT_INTLK:  // latched (for RevSwitch) Interlock
        if (ctlStateChgd)  LogMsg("CS_EXT_INTLK");
        enablePS = false;
        if (clearInterlock)  ctlState = CS_OFF;
        break;


      default:
        LogMsg("default?????");
        ctlState = CS_OFF;
    }

    spStateChgd = (setPointState != beforeState);
    ctlStateChgd = (ctlState != beforeCtlState);

  } while (gotMsg);

}  // RevSwitch version


//-----------------------------------------------------------------------------
//  Update the analog and digital outputs based on the values updated by
//  ComputeNextState()
//-----------------------------------------------------------------------------
nodebug void ChangeState()   // RevSwitch version
{
  char  cmd[MaxCmdSize], temp[16];


  if (AnalogCtl)  {
    dacAmpSetDev = (fabs(setPoint) - dacAmpSetOffset) / dacAmpSetGain;
    voltsOut[OUTPUT_CTL] = dacAmpSetDev;
  }
  else  {
    serAmpSetSendDev = fabs((setPoint - serAmpSetOffset) / serAmpSetGain);
    serVltSetSendDev = (maxVolts - serVltSetOffset) / serVltSetGain;

    if (!passiveSerial AND enablePS AND gettingReadings)  {
      if (fabs(serAmpSetReadDev - serAmpSetSendDev) > 0.001)
        if (elapsedMS(lastSerAmpSetTime) > 200)  {
          sprintf(cmd, "SOUR:CURR %s\n", fmtFloat(serAmpSetSendDev,9,4,temp));
          Buf_Add(&psCmdBuf, cmd, strlen(cmd));
          lastSerAmpSetTime = MS_TIMER;
        }
      if (fabs(serVltSetReadDev - serVltSetSendDev) > 0.001)
        if (elapsedMS(lastSerVltSetTime) > 200)  {
          sprintf(cmd, "SOUR:VOLT %s\n", fmtFloat(serVltSetSendDev,9,4,temp));
          Buf_Add(&psCmdBuf, cmd, strlen(cmd));
          lastSerVltSetTime = MS_TIMER;
        }
    }
  }

  digitalOut = 0xffffffff;

  if (enablePS)  {
    ClearBits(digitalOut, TURN_PS_OFF);
    if (connectLoad)
      if (reversePolarity)
        ClearBits(digitalOut, REV_POLARITY_LED | REV_POLARITY_OFF);
      else
        ClearBits(digitalOut, NORM_POLARITY_LED | NORM_POLARITY_OFF);
  }

  devStatus = 0;
  if (extInterlock)               devStatus |= DS_INTLK;
  if (ctlState == CS_INT_INTLK)   devStatus |= DS_INTLK;
  if (ctlState == CS_ON)          devStatus |= DS_ON;
  if (!reversePolarity)           devStatus |= DS_NORM_POL;
  if (serReadErr)                 devStatus |= DS_SER_ERR;
  if (calibErr)                   devStatus |= DS_CALIB_ERR;
}
#endif  //RevSwitch vers



//-----------------------------------------------------------------------------




//-----------------------------------------------------------------------------
//  Do anything specific to this controller that is not time-ciritical.
//  This is called once during each pass of the ComLoop() function
//-----------------------------------------------------------------------------
void BackGroundProcess()
{
  int  i;


  if (recursiveBackGround)  return;
  recursiveBackGround = true;


  recursiveBackGround = false;
}


//-----------------------------------------------------------------------------
//  Do device specific formatting of current state data that cannot be
//  handled by entries in generic table for I/O values
//
//  Returns:
//      true:  last portion of data has been displayed
//      false:  more data to display during next pass
//-----------------------------------------------------------------------------
bool ShowDeviceState()
{
  char  *str;
  static bool  noSerData;
  static devShowState;
  char  msg[90];


  switch (++devShowState)  {

#ifdef RevSwitch
    case 1:
      switch (supplyType)  {
        case PS_TYPE_ANALOG:    str = "-- ANALOG --";  break;
        case PS_TYPE_SGA:       str = "Sorensen SGA";  break;
        case PS_TYPE_SGA_FULL:  str = "Sor.SGA FULL";  break;
        default:
           str = (blinkOff ? "            " : "**UNKNOWN** ");
      }
      DispStr(1,11, str, false);

      noSerData = (AnalogCtl OR !gettingReadings);

      if (devStatus & DS_INTLK)  str = (blinkOff ? "     " : "INTLK");  else
      if (devStatus & DS_ON)     str = " ON  ";  else
                                 str = (blinkOff ? "     " : "*OFF*");
      DispStr(16,11, str, false);

      str = ((devStatus & DS_NORM_POL) ? "NORM" : "REV ");
      DispStr(24,11, str, false);

      DispStr(11,14,
         (noSerData ? "**N/A**" : fmtFloat(serAmpSetReadEng, 7,3, msg) ), false);
      DispStr(11,15,
         (noSerData ? "**N/A**" : fmtFloat(serVltSetReadEng, 7,3, msg) ), false);
      DispStr(19,14,
         (noSerData ? "**N/A**" : fmtFloat(serAmpSetReadDev, 7,3, msg) ), false);
      DispStr(19,15,
         (noSerData ? "**N/A**" : fmtFloat(serVltSetReadDev, 7,3, msg) ), false);

    case 2:
      DispStr(30,14,
         (noSerData ? "**N/A**" : fmtFloat(serAmpMeasEng, 7,3, msg) ), false);
      DispStr(30,15,
         (noSerData ? "**N/A**" : fmtFloat(serVltMeasEng, 7,3, msg) ), false);
      DispStr(38,14,
         (noSerData ? "**N/A**" : fmtFloat(serAmpMeasDev, 7,3, msg) ), false);
      DispStr(38,15,
         (noSerData ? "**N/A**" : fmtFloat(serVltMeasDev, 7,3, msg) ), false);
      break;

    case 3:
      DispStr(50,14, fmtFloat(adcLoadAmps, 8,3, msg), false);
      DispStr(50,15, fmtFloat(adcLoadVolts, 8,3, msg), false);

      DispStr(60,15, fmtFloat(adcPSVolts, 8,3, msg), false);

      DispStr(10,17, fmtFloat(ampSet, 8,3, msg), false);
      DispStr(25,17, fmtFloat(setPoint, 8,3, msg), false);

      DispStr(36,17, OnStateDesc[setPointState], false);
      break;
#endif // RevSwitch

  //--- to catch all non-active state values ---
    default:
      devShowState = 0;  break;
  }

  return (devShowState == 0);
}


//-----------------------------------------------------------------------------
//  Determine what, if any, action should be taken as a result of one of the
//  coil values being changed
//-----------------------------------------------------------------------------
void NewCoilValue(int chanNum, int coilNum, int coilVal, int cmdRow)
{
  char  *msg;
  RegInfo  *reg;


  StackPtr();

  msg = getBuf(80);

  sprintf(msg, "set coil [%05d] = %d", coilNum+1, coilVal);
//  if (GenericDisp)  DispStr(CmdCol,cmdRow, msg, true);
  LogMsg(msg);

  if ((coilNum < 0) || (coilNum >= NUMCOILS))  { freeTo(msg);  return; }

  if (coilVal)  switch (coilNum)  {

    case DoReboot:
      Reboot(10);  break;

    case DoInit:
      InitCmds(0);  break;

    case AbortCalib:
      if (calibrating)  cancelCalib = true;
      break;

    case TurnOn:
      TurnOnMsg();  break;

    case TurnOff:
      TurnOffMsg();  break;

    case ClearInterlock:
      ClearInterlockMsg();  break;
  }
  coilVal = 0;

  freeTo(msg);

  ++criticalSection;
  ClearBit(coils[chanNum], coilNum);
  --criticalSection;
}

//-----------------------------------------------------------------------------
//  Determine what, if any, action should be taken as a result of one of the
//  registers being written to.
//
//  Called by the generic function NewRegValue() in the base code which is
//  responsible for actually copying the <regName>_New value to <regName>
//
//  Returns true if the new value should be used, false if it should not
//
//  If any value used during execution of the periodic ControlLoop() function
//  (besides the new register value) is altered here, you MUST increase the
//  criticalSection counter 1st (it will be restored to its previous value
//  after the new register value is in effect).  This is to avoid having the
//  ControlLoop() processes act on unsynchronized values.
//
//    Example:  DevNewRegValue() alters some non-register value in response
//    to a new register value.  Since the new register value does not go
//    into effect until DevNewRegValue() returns, any such values used within
//    ControlLoop() would be out of sync with the current register value.
//
//  NOTE:  Assumes that, when a register that is part of a larger data type
//         (e.g. a 32-bit float or long integer) that the lower numbered
//         registers will be written first.
//
//  Any string in *msg will be logged after returning from this func.
//-----------------------------------------------------------------------------
bool  DevNewRegValue(int chanNum, int regNum, int regVal,
                     int clientType, LinkProc *lp, char *msg)
{
  int  i, cmdRow;
  ulong  mask;


  cmdRow = CmdRow(LinkProc_Idx(lp));

//--- Check the New register value (or set of registers comprising a ---
//--- single value) and either reject it (in which case the new value ---
//--- will be ignored), or do whatever processing is necessary using the ---
//--- new value.  NOTE: The new value does not actually go in to effect ---
//--- until sometime after this function returns (one of the reasons that ---
//--- proper handling of the criticalSection counter is required). ---

  do  {

//  If_Reg(regName)  {
//    if (regName_New not usable)  {
//      sprintf(msg, "optional msg if default 'ignored' msg insufficient");
//      return false;
//    }
//    ... whatever else needs to be done with regName_New value ...
//    sprintf(msg, "...optional msg about action taken...");
//    (NOTE: regName = regName_New is handled by caller of this function)
//    break;
//  }

#define  IfReg(regName)  if (regNum == regName ## _Idx)

#define  ReadRes  (ADCRes / shuntImpedance)
#define  MaxAmps  (DACMax / ampSetScale)

#define  BoundNew(reg)  { \
    if (reg##_New < reg##Min)  reg ## _New = reg##Min;  \
    if (reg##_New > reg##Max)  reg ## _New = reg##Max;  \
}

#define  CheckApplVal(reg)  {  \
    if (regNum == reg ## _Idx)  {    \
      if (reg ## _New < (reg##Min))  reg ## _New = (reg##Min);  \
      if (reg ## _New > (reg##Max))  reg ## _New = (reg##Max);  \
      applData.reg = reg ## _New;  ChgWait(5);  break;  \
    }  \
}
#define  CheckApplValNZ(reg)  {  \
    if (regNum == reg ## _Idx)  {    \
      if (reg ## _New < (reg##Min))  reg ## _New = (reg##Min);  \
      if (reg ## _New > (reg##Max))  reg ## _New = (reg##Max);  \
      if (reg ## _New == 0.0)  reg ## _New = (reg##Def);  \
      applData.reg = reg ## _New;  ChgWait(5);  break;  \
    }  \
}

  //--- alternate means of setting modbus "coils" ---
    IfReg( coils31_00 )  {
      for (mask=1,i=0; i<32; ++i,mask<<=1)
        if (coils31_00_New & mask)  NewCoilValue(chanNum, i, 1, cmdRow);
      coils31_00_New = 0;  break;
    }

    IfReg( coils63_32 )  {
      for (mask=1,i=0; i<32; ++i,mask<<=1)
        if (coils63_32_New & mask)  NewCoilValue(chanNum, i+32, 1, cmdRow);
      coils63_32_New = 0;  break;
    }


    IfReg( ampSet )  {
      if (AnalogCtl)  if (fabs(ampSet_New) > maxAmps)
        ampSet_New = ((ampSet_New < 0.0) ? -maxAmps : maxAmps);
      NewSetPointMsg(ampSet_New);   break;
    }


    CheckApplVal( adcLoadAmpOffset );
    CheckApplValNZ( adcLoadAmpGain );

    CheckApplVal( adcLoadVltOffset );
    CheckApplValNZ( adcLoadVltGain );

    CheckApplVal( adcPSVltOffset );
    CheckApplValNZ( adcPSVltGain );

    CheckApplVal( adcExtAmpOffset );
    CheckApplValNZ( adcExtAmpGain );

    CheckApplVal( dacAmpSetOffset );
    CheckApplValNZ( dacAmpSetGain );

    CheckApplVal( serAmpMeasOffset );
    CheckApplValNZ( serAmpMeasGain );

    CheckApplVal( serVltMeasOffset );
    CheckApplValNZ( serVltMeasGain );

    CheckApplVal( serAmpSetOffset );
    CheckApplValNZ( serAmpSetGain );

    CheckApplVal( serVltSetOffset );
    CheckApplValNZ( serVltSetGain );

    CheckApplVal( maxVolts );
    CheckApplVal( maxAmps );

    CheckApplVal( safeSwAmps );

    CheckApplVal( connectLoadDelay );

    CheckApplVal( supplyType );

    CheckApplVal( biasVoltsOffset );
    CheckApplValNZ( biasVoltsGain );

    CheckApplValNZ( biasVoltsSet );
    CheckApplValNZ( biasErrorStart );
    CheckApplValNZ( biasErrorStop );

    CheckApplValNZ( rampRate );

  // search for _PERSISTENT_ to see where code normally added for saved values

  }  while (false);

  return true;
}


//---------------------------------------------------------------------------
//  Multiple calibration processes can be supported simply by having a
//  different starting state for each one.
//---------------------------------------------------------------------------
void calib_tick()
{
  int  i;
  ulong  waited;
  static int  returnState, cancelState;
  static ulong  startTime, waitStart, minWaitMS, maxWaitMS, lastReadTime;
  static char  msg[90];


#define  StopBecause(str)  {  \
  strcpy(msg, (str));  calibState = cancelState;  break;  }


  if (recursiveCalib)  return;
  recursiveCalib = true;

//--- wait until we have new values to work with ---
  if (RdTime == lastReadTime)  { recursiveCalib = false;  return; }
  lastReadTime = rdTime;

//--- start a new calibration? ---
  if (startCalibState AND !calibrating)  {
    *msg = 0;  calibState = startCalibState;  cancelState = CALIB_IDLE; }
  startCalibState = CALIB_IDLE;

//--- external cancelation of current calibration? ---
  if (cancelCalib)  {
    if (calibrating)  { strcpy(msg, "Canceled");  calibState = cancelState; }
    cancelCalib = false;
  }


  switch (calibState)  {

    case CALIB_IDLE:
      break;


//  case CALIB_xxx_INIT:  // prep for specific calibration process
//    ... set state of machine as needed to begin calibration ...
//    cancelState = CALIB_xxx_ABORT;   calibState = CALIB_xxx_1;
//    ShowMsg("Starting ... calibration");   break;


//  case CALIB_xxx_1:  // check progress, "wait" if necessary ...
//    if (! (calibration-friendly-conditions))   StopBecause("xxx");
//    if (1st-state-reached)  {
//      ShowMsg("1st state achieved!!!");
//      calibState = CALIB_xxx_2;  break;
//    }
//    tweak-state-as-needed;
//    if (tweaked-too-far)  StopBecause("xxx");
//    calibState = CALIB_xxx_WAIT;  break;


//  case CALIB_xxx_2:  // check progress, "wait" if necessary ...
//    if (! (calibration-friendly-conditions))   StopBecause("xxx");
//    if (1st-state-reached)  {
//      ShowMsg("2nd state achieved!!!");
//      calibState = CALIB_xxx_RETURN;  break;
//    }
//    tweak-state-as-needed;
//    if (tweaked-too-far)  StopBecause("xxx");
//    calibState = CALIB_xxx_WAIT;  break;


//  case CALIB_xxx_WAIT:  // do asynchronous "wait" for an event
//    waited = elapsedMS(busyStart);
//    if (waited < minWaitMS)  break;
//    if ((event-not-yet-happened) AND (waited < maxWaitMS))  break;
//    calibState = returnState;  break;


//  case CALIB_xxx_ABORT:
//    if (*msg)  {  strcat(msg, ": Aborting Calibration...");  LogErr(msg); }
//    calibState = CALIB_xxx_RETURN;

//  case CALIB_xxx_RETURN:
//    put things back the way they were before calibration began
//    calibState = cancelState = CALIB_IDLE;  break;



    default:
      LogErr("Error in calibration state machine");  calibState = CALIB_IDLE;
  }

  recursiveCalib = false;
}



//=============================================================================
//                   Functions unique to this application
//=============================================================================

#define  DebounceMS  250

//-----------------------------------------------------------------------------
//  keep weighted count for each digitalIn bit based on the state of the bit
//  during a recent interval.  If the bit does not change over a short
//  interval the count will cross the 0 level threshold which allows us to
//  determine if a bit has remained at the same state for some interval.
//-----------------------------------------------------------------------------
void debounceDigIn()
{
  int  i;
  int  *bitCount;
  ulong  mask, digIn;


  bitCount = bitCounts;  digIn = 0;  mask = 1;

  for (i=0; i<32;  ++i)  {
    if (digitalIn & mask)  *bitCount += 10;  else  *bitCount -= 10;

    if (*bitCount > DebounceMS)  *bitCount = DebounceMS;
    if (*bitCount < -DebounceMS)  *bitCount = -DebounceMS;

    if (*bitCount > 0)  digIn |= mask;

    ++bitCount;  mask <<= 1;
  }

  digInChanges &= ~(digitalIn ^ digIn);  digitalIn = digIn;
}


//-----------------------------------------------------------------------------
//          ======== functions to generate Event Msgs ==========
//-----------------------------------------------------------------------------
void NewSetPointMsg(float newCur)
{ eventMsg.msgData.newCur = newCur;  AddMsg(MSG_NewSetPoint); }

//-----------------------------------------------------------------------------
//  Parse the response from the Ameritherm Power Supply
//-----------------------------------------------------------------------------
#ifdef RevSwitch
int Process_PS_Resp(LinkProc *lp)
{
  char  *ptr;
  int  len;
  PS_Client  *ps;


  ps = (PS_Client *) lp->srvData;

  if (strncmpi(ps->respBuf, "SORENSEN, SGA", 13) == 0)
    if (supplyType != PS_TYPE_SGA_FULL)  supplyType = PS_TYPE_SGA;

  switch  (supplyType)  {

    case PS_TYPE_SGA:
    case PS_TYPE_SGA_FULL:
      if (! isNumeric(ps->respBuf))  break;

      if (strncmpi(ps->cmdBuf, "SOUR:CURR?", 10) == 0)  {
        ++criticalSection;
        serAmpSetReadDev = atof(ps->respBuf);
        serAmpSetReadEng = serAmpSetReadDev * serAmpSetGain + serAmpSetOffset;
        validAmpSetRdg = true;
        --criticalSection;  return 0;
      }

      if (strncmpi(ps->cmdBuf, "SOUR:VOLT?", 10) == 0)  {
        ++criticalSection;
        serVltSetReadDev = atof(ps->respBuf);
        serVltSetReadEng = serVltSetReadDev * serVltSetGain + serVltSetOffset;
        --criticalSection;  return 0;
      }

      if (strncmpi(ps->cmdBuf, "MEAS:CURR?", 10) == 0)  {
        ++criticalSection;
        serAmpMeasDev = atof(ps->respBuf);
        serAmpMeasEng = serAmpMeasDev * serAmpMeasGain + serAmpMeasOffset;
        if (reversePolarity)  serAmpMeasEng = 0.0 - serAmpMeasEng;
        --criticalSection;  return 0;
      }

      if (strncmpi(ps->cmdBuf, "MEAS:VOLT?", 10) == 0)  {
        ++criticalSection;
        serVltMeasDev = atof(ps->respBuf);
        serVltMeasEng = serVltMeasDev * serVltMeasGain + serVltMeasOffset;
        if (reversePolarity)  serVltMeasEng = 0.0 - serVltMeasEng;
        --criticalSection;  return 0;
      }

      break;

  }

  return -1;
}

//-----------------------------------------------------------------------------
//  send cmds to and interpret responses from a Power Supply via serial link
//-----------------------------------------------------------------------------
void PS_Server(LinkProc *lp)
{
  short  i, y, numBytes, err, cmdLen;
  char  ch;
  bool  showDetails, queryCmd;
  char  *msg, *temp, *statCmd;
  ComLink  *cl;
  PS_Client  *ps;
  SerialLink  *sl;
  static uint  ticks;


#define  PS_SEND_CMD     1
#define  PS_XMIT_ACTIVE  2
#define  PS_WAIT_START   3
#define  PS_WAIT_END     4
#define  PS_PROCESS      5
#define  PS_WAIT_IDLE    6

#define  MaxRespSize  ( sizeof(PS_Clients[0].respBuf) )

  StackPtr();

  msg = getBuf(90 + 90);  temp = msg + 90;

  y = CmdRow(LinkProc_Idx(lp));

  cl = lp->comLink;  sl = (SerialLink *) cl;
  ps = (PS_Client *) lp->srvData;

  showDetails = GenericDisp AND lp->verboseDisp;

  if (err = cl->getErr(cl))  {
    ShowMsg(cl->errDesc(cl, err, temp));  ++serialErrs;  ++ttlErrs;  }

  if (restartPS_Service)  {
    LinkProc_StartClose(lp);  restartPS_Service = false;  }


  switch (lp->srvState)  {

    case SRV_NEW:  // initialize data for new client
      if ((ps = New_(PS_Client)) == NULL)  {
        LinkProc_StartClose(lp);  break; }
      lp->srvData = (void *) ps;  lp->tempUsed = 0;  ticks = 0;
      lp->srvState = PS_SEND_CMD;  ps->xmitStartTime = MS_TIMER;


    case PS_SEND_CMD:  // send a command
      if (AnalogCtl OR !enableSerial)  break;
      if (cl->wrFree(cl) < MaxCmdSize)  break;
      if (ps->userCmd = *psCmd)
        strncpy(ps->cmdBuf, psCmd, MaxCmdSize);
      else  {
        if (passiveSerial)  break;
      //--- flush buffer - if PS won't shut up, then send RESET cmd ---
        if (cl->unread(cl) > 0)  {  // wait until PS is quiet
          cl->read(cl, ps->respBuf, RCV_BUF_SIZE);
          if (elapsedMS(ps->xmitStartTime) < 10000)  break;
          strcpy(ps->cmdBuf, "*RST");
          //LogMsg(ps->cmdBuf);  //tdebug
        }
      //--- if special cmds in the queue, then send the next in the queue ---
        else if ((cmdLen = Buf_LineLen(&psCmdBuf, 0)) > 0)  {
          if (cmdLen > MaxCmdSize-1)  cmdLen = MaxCmdSize-1;
          Buf_Remove(&psCmdBuf, ps->cmdBuf, cmdLen);
          ps->cmdBuf[cmdLen-1] = 0;  // write a 0 over the trailing LF
          //LogMsg(ps->cmdBuf);  //tdebug
        }
      //--- otherwise, just send the next query command in the sequence ---
        else
          if ((supplyType != PS_TYPE_SGA) AND (supplyType != PS_TYPE_SGA_FULL))
            strcpy(ps->cmdBuf, "*IDN?");
          else  {
            if (++ticks >= NumCmds)  ticks = 0;
            strcpy(ps->cmdBuf, psCmds[ticks]);
          }
      }
      strcat(ps->cmdBuf, "\r");
      ps->cmdLen = strlen(ps->cmdBuf);
      cl->write(cl, ps->cmdBuf, ps->cmdLen);  ++pendingResps;
      *psCmd = 0;  ps->xmitStartTime = MS_TIMER;
      ps->rcvdBeforeCmd = cl->readCount;  lp->srvState = PS_XMIT_ACTIVE;
      break;


    case PS_XMIT_ACTIVE:  // wait for transmission to finish
      if (cl->unsent(cl) > 0)  break;
      queryCmd = (ps->cmdBuf[ps->cmdLen-2] == '?');
      lp->srvState = (queryCmd ? PS_WAIT_START : PS_SEND_CMD);
      break;


    case PS_WAIT_START:  // wait for start of next line of resp from supply
      if (cl->unread(cl) <= 0)  {
        if (elapsedMS(cl->lastActive) < 1500)  break;  // timeout for 1st char
        if (cl->readCount == ps->rcvdBeforeCmd)  {
          psNoResponse = true;  // supplyType = PS_TYPE_ANALOG;
          if (GenericDisp)  DispStr(CmdCol,y, "No Response", true);
//        if (!StreamDisp)  DispStr(0,21, blinkOff ? " " : PSNoRespMsg, true);
        }
        lp->srvState = PS_SEND_CMD;  break;
      }
      ch = cl->getc(cl) & 0x7f;
      if ((ch < ' ') OR (ch > '~'))  ch = '*';
      *ps->respBuf = ch;
      if (*ps->respBuf < ' ')  break;  // LF indicates start of msg
      ps->rcvCount = 1;  lp->srvState = PS_WAIT_END;
      break;


    case PS_WAIT_END:  // wait for the end of the current resp line
      if ((numBytes = cl->unread(cl)) <= 0)  // nothing new
        if (elapsedMS(cl->lastActive) > 9)  {  // max 10ms between chars
          ps->respBuf[ps->rcvCount] = 0;
          sprintf(msg, "Timeout: %s", ps->respBuf);
          ShowMsg(msg);  ++serialErrs;  ++ttlErrs;
          lp->srvState = PS_WAIT_IDLE;  break;
        }
      for (i=0; i<numBytes; ++i)  {
        if (ps->rcvCount >= RCV_BUF_SIZE)  {
          ShowMsg("Msg too long");  ++serialErrs;  ++ttlErrs;
          lp->srvState = PS_WAIT_IDLE;  break;
        }
        ch = cl->getc(cl) & 0x7f;
        if (((ch < ' ') AND (ch != 0x0d)) OR (ch > '~'))  ch = '*';
        ps->respBuf[ps->rcvCount] = ch;
        if (ch >= ' ')  { ++ps->rcvCount;  continue; }  // CR marks EOL
        ps->respBuf[ps->rcvCount] = 0;  psNoResponse = false;
        lp->srvState = PS_PROCESS;  break;
      }
      if (elapsedMS(ps->xmitStartTime) > 10000)  // in case of endless resp
        lp->srvState = PS_SEND_CMD;
      break;


    case PS_PROCESS:
      if (showDetails)  DispStr(CmdCol,y, ps->respBuf, true);
      if (ps->userCmd)  {  // display and wait for multiple lines of resp
        LogMsg(ps->respBuf);  ShowMsg(ps->respBuf);
        lp->srvState = PS_WAIT_START;  // in case of multi-line response
      }
      else  // process single-line response
        lp->srvState = PS_SEND_CMD;
      if (Process_PS_Resp(lp) == 0)
        pendingResps = 0;
      else  {
        sprintf(msg, "Invalid resp: %s", ps->respBuf);
        ShowMsg(msg);  ++serialErrs;  ++ttlErrs;
      }
      break;


    case PS_WAIT_IDLE:  // wait for idle/no data rcvd after an error
      if (!ps->startIdleTime)  ps->startIdleTime = MS_TIMER;
      if (cl->unread(cl) > 0)  {
        while (cl->read(cl, ps->respBuf, RCV_BUF_SIZE) > 0);  //flush read buf
        break;
      }
      if (elapsedMS(cl->lastActive) > 100)  {
        ps->startIdleTime = 0;  lp->srvState = PS_SEND_CMD;  }
      if (elapsedMS(ps->xmitStartTime) > 10000)  {
        ShowMsg("PS won't shut up!");  lp->srvState = PS_SEND_CMD; }
      break;


    case SRV_CLEANUP:
      if (ps)  { lp->srvData = NULL;  Free_(ps, PS_Client);  }
      cl->startClose(cl);  lp->srvState = SRV_DONE;  break;
  }


  freeTo(msg);
}
#endif


//-----------------------------------------------------------------------------
//  Handle any device-specific device special file open requests
//-----------------------------------------------------------------------------
int DLP_DevFile_Open(LinkProc *lp, int argc, char *argv[])
{
  return -999;
}


//============================================================================
main()
{
  int  i;
  char  *msg;
  ulong  mask;


  firstTime = true;

  while (true)  {
    CommonInit();

    msg = getBuf(128);   // do NOT free up - "backup" in case of failed getBuf()


  digOutBootState = 0xffffffff;
  #if (_BOARD_TYPE_ != RCM3200)
    ioBoardType = SmartCat;
  #else
    ioBoardType = NSCL16;
    #ifdef RevSwitch
      digOutBootState = 0xffffffff;
    #endif
  #endif

    CommonHdwInit(digOutBootState);  DLPHdwInit();


  //--- make sure output voltages are set as needed at startup ---
    for (i=0; i<NumADC; ++i)  voltsIn[i] = 0.0;
    for (i=0; i<NumDAC; ++i)  voltsOut[i] = 0.0;

    DLPInit();

  //========== Remaining Device-specific Initialization ==========

    devInitialized = false;

    calibState = CALIB_IDLE;  ctlState = CS_OFF;

    minStackPtr = 0xffff;  maxStackUsed = 0;

    newSetPoint = reversePolarity = false;

  #define RestoreReg(regName)  regName = applData.regName

    ampSet = lastAmpSet = setPoint = 0.0;
    enablePS = connectLoad = false;

    RestoreReg( adcLoadAmpOffset );
    RestoreReg( adcLoadAmpGain );

    RestoreReg( adcLoadVltOffset );
    RestoreReg( adcLoadVltGain );

    RestoreReg( adcPSVltOffset );
    RestoreReg( adcPSVltGain );

    RestoreReg( adcExtAmpOffset );
    RestoreReg( adcExtAmpGain );

    RestoreReg( dacAmpSetOffset );
    RestoreReg( dacAmpSetGain );

    RestoreReg( serAmpMeasOffset );
    RestoreReg( serAmpMeasGain );

    RestoreReg( serVltMeasOffset );
    RestoreReg( serVltMeasGain );

    RestoreReg( serAmpSetOffset );
    RestoreReg( serAmpSetGain );

    RestoreReg( serVltSetOffset );
    RestoreReg( serVltSetGain );

    RestoreReg( maxVolts );
    RestoreReg( maxAmps );

    RestoreReg( safeSwAmps );
    RestoreReg( connectLoadDelay );
    RestoreReg( supplyType );

    RestoreReg( biasVoltsOffset );
    RestoreReg( biasVoltsGain );

    RestoreReg( biasVoltsSet );
    RestoreReg( biasErrorStart );
    RestoreReg( biasErrorStop );

    RestoreReg( rampRate );

  // search for _PERSISTENT_ to see where code normally added for saved values

    lastSerAmpSetTime = lastSerVltSetTime = MS_TIMER;
    pendingResps = 0;  validAmpSetRdg = gettingReadings = false;
    serReadErr = calibErr = manualOff = clearInterlock = false;
    adjustBias = chgPolarity = false;

    serAmpMeasEng = serVltMeasEng = serAmpSetReadEng = serVltSetReadEng = 0.0;
    ampSet = activeAmpSet = setPoint = 0.0;  setPointState = 0;
    clearPulseStarted = 0;

    if (firstTime)  Buf_Init(&psCmdBuf, 512);  else  Buf_Flush(&psCmdBuf);

    #ifdef RevSwitch
      PS_Params.port = 3;      // Port D
      PS_Params.baud = 19200;  // different for different supplies
      PS_Params.bits = 8;
      PS_Params.parity = PARAM_NOPARITY;
      PS_Params.flags = 0;     // no special flag settings

      *psCmd = 0;  restartPS_Service = false;
      memset(PS_Clients, 0, sizeof(PS_Clients));
      psSrvIdx = Service_Add( &PS_Service );
    #endif

    passiveSerial = enableSerial = false;  numPolErrs = 0;

    showAllStateChgs = false;

    controlEnabled = true;  //initiate active control of the device

    for (mask=1,i=0; i<32; ++i,mask<<=1)
      bitCounts[i] = (digitalIn & mask) ? DebounceMS : -DebounceMS;

    while (ComLoop());

    freeTo(msg);  firstTime = false;
  }
}

