//----- RevSwitch_Appl.c ----- 12/03/13 -----(02/17/09)----

// EE Bundle #082

// Sections that typically require application-specific code are marked with
// the comment:  //APPLCODE

// watch ser resp cmds

// scope 5 ampApply ampRead serReadA serReadV serProjErr serProjErrAge extAmps extVolts

// To log cmds/resps to/from supply:  diag msgs psCom 2

//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//APPLCODE
//  Application code to manage relay-controlled polarity switch attached to
//  a power supply via analog and/or a serial connection.
//
//  NOTE:  Different models of Sorensen supplies are connected to these
//         controllers, and not all of them support the same set of commands.
//         The manual used as the basis for this code is:
//
//      SG Series IEEE-488.2/RS-232 Option Programming Manual, Rev C (9/8/2004)
//
//   As of this writing, the name of the file in the Documentation folder
//   containing a copy of this manual is:
//
//      Sorensen SG Series Programming Manual Rev C_2004.pdf
//
//
//  Models known to be in use:
//    - SORENSEN, SGA60/417E-2AAA, 0721A02109, 1.03, 1.02
//    - SORENSEN,DCS12-250,0943A01282,2.02-2.00
//
//-----------------------------------------------------------------------------
// UPDATE LOG (most recent first):
//
//
//  2013-12-03 - Mark Davis
//               - Added logic to update new onStatesBitMap in the DiagLogInfo
//                 structure.
//               - Added checkPrevOnStates() for setting offDueToReset, which
//                 is used to assert an interlock if the supply shut off due to
//                 an unexpected reset.
//
//  2013-07-01 - Mark Davis
//               - Added DS_CTLR_RESET bit to devStatus bits.  This is set
//                 during startup whenever the previous state could not be
//                 restored, and cleared the first time one of the supplies is
//                 successfully turned on.
//
//  2013-03-13 - Mark Davis
//               - For serial controlled supplies, there is now a short delay
//                 (set by the macro SerComDelay) between when the power to the
//                 supply is enabled and when we start trying to communicate
//                 with it.  Prior to this, the serial interface could get in a
//                 state where it would NEVER respond.
//
//  2013-03-13 - Mark Davis
//               - Correction to logic that logs unexpected chars from the
//                 supply (i.e. ones that are in the rcv buffer before we
//                 have sent a new command).
//
//  2013-02-05 - Mark Davis
//               - Corrected bug in modified logic for the periodic logging
//                 of the state so it logs the readings for supplies using
//                 analog mode.
//
//  2013-02-04 - Mark Davis
//               - The serIDStr is now saved in the settings file so it gets
//                 restored on startup (which lets us see the last known
//                 value even when the supply is off).
//               - The value of the serIDStr is now logged whenever the
//                 serial link to the supply changes from being inactive to
//                 being active.
//               - The "*** Serial Link not active ***" message is now logged
//                 ONLY if serLinkNotWorking is true.  Previously, this was
//                 logged along with the periodic logging of the current
//                 settings and readings even if the supply was turned off.
//
//  2013-02-01 - Mark Davis
//               - All traffic on the serial connection is now logged for a
//                 short time (at lengthening intervals) whenever the serial
//                 link to the PS is not working as expected.
//               - Added persistent PV/reg value serInitTime.  Previously the
//                 max time it may take for the serial link to be established
//                 after the power to the supply is enabled was not a
//                 configurable value.
//               - Changed psOnTime to psOutputOnTime and added
//                 psPowerEnabledSecs.
//               - ALL problems with serial responses from the supply are now
//                 logged if the PS_MSG_BITS diagnostic msgs bit is set.
//                 Previously, this was only done if the problem occured
//                 multiple consecutive times.
//
//  2013-01-09 - Mark Davis
//               - The periodic state info (minus readings) is now logged for
//                 serial mode even if the serial link isn't active.
//
//  2012-10-25 - Mark Davis
//               - Uncommented if statement so the "cmd send <str>" command
//                 would work again when the serial link is active.
//
//  2012-10-18 - Mark Davis
//               - Logic for projected readbacks when using serial set commands
//                 and readings now finished.
//               - Increased the min diffrence for sending a new voltage limit
//                 value to a serial controlled supply to 0.05 volts.
//
//  2012-10-11 - Mark Davis
//               - switchTime now used to wait after changing the state of the
//                 polarity switches.  This insures the load is connected and
//                 the supply output is stable before asserting a non-0
//                 setpoint.
//
//  2012-10-08 - Mark Davis
//               - Refinements to serial interface so it minimizes the # of
//                 RAMP commands when possible.
//
//  2012-10-03 - Mark Davis
//               - The min difference that causes the controller to send a new
//                 setpoint to the supply has been increased to avoid what the
//                 supply believes is a redundant value (and the resulting
//                 error responses).
//
//  2012-09-28 - Mark Davis
//               - The polarity of the connection to the load no longer
//                 switches before it should when going from (-) to (+).
//               - Serial mode fixes:
//                  - The devStatus DS_SER_ERR bit is now set when the link is
//                    not working.
//                  - The load is no longer connected until after we have
//                    confirmation that the power supply is on and that the
//                    output is enabled.
//                  - The readback is set to 0 when the supply is not enabled
//
//  2012-09-27 - Mark Davis
//               - Corrected typo in setCtlModes() so the setpoint nibble can
//                 be changes from serial mode to external analog mode.
//
//  2012-09-25 - Mark Davis
//               - First fairly well tested version for testing on live ctlrs.
//
//  2012-08-06 - Mark Davis
//               - Began work on DEBROS version using stripped down copy of the
//                 DEBROS version of the TCPS appl code.
//
//-----------------------------------------------------------------------------

/*
  Description of configurable values:

  NOTES:

     The rate at which the controller can get updated readings via the serial
     link is very low.  When the serial readback is the only method available,
     simply reporting these values as-is results in a significant stair-step
     effect which provides a very misleading representation when compared to
     what the supply is actually doing.

     When the serial readings are the only feedback available, the controller
     provides a much more accurate representation by "projecting" what the
     current state of the supply is from the commands it has sent and the
     (few and outdated) readings it is getting back.

     When using an analog readback for the current output of the supply, the
     values below that reference projected values have no effect on the
     reported values.


   ctlModes:
          Determines how to control and monitor the power supply.  The least
          3 significant nibbles (hex digits) each apply to a one signal as
          follows:

                              0x0###
                                 |||
             Current setpoint ---+||
             Current readback ----+|
             Voltage readback -----+

          Each digit is set to one of the following values:

             4: Send/Read the value to/from the supply via the serial port
             2: Read the value from Internal analog signal (N/A to setpoint)
             1: Apply to/Read from the External analog signal

          The most commonly used combinations:

             0x0444: Serial for setpoint and readbacks
             0x0411: Serial for setpoint, external analog for readbacks
             0x0111: External analog for setpoint and readbacks

          NOTES:
             For at least some of the Sorensen supplies, ANY use of the serial
             connection requires that the setpoint use the serial link.  This
             is due to the fact that nearly ANY use of the serial connection
             causes the supply to change to remote mode and ignore the analog
             control signals.

             The rate at which commands and status info can be sent/read via
             the serial link is quite low.  In addition, the returned value
             can be a second or more out of date (usually only this bad at
             the beginning or end of a RAMP command.  So not only is the rate
             of updates very slow, but the "newest" value can be significantly
             different than actual state of the supply at the time.

             To help compensate for this limitation, the code computes what
             the state of the supply should be based on the RAMP commands it
             sent and adjusts its projections as needed based on the readings
             it gets back.  The result is a much smoother (and MORE accurate)
             readback than if it just returned the values returned from the
             supply.

             Due to an error in the controller board circuits, the internal
             analog readbacks are (at the time of this writing) not functional.
             The internal analog voltage and current signals are connected to
             the front panel for diganostic use, but they are not (currently)
             usable by the controller firmware.


    connectLoadDelay:
          The minimum number of milliseconds between when the supply has been
          turned ON and when the load will be connected.  Defaults to 500 ms.


    debounceMS:
          The number of milliseconds over which a digital input must have the
          same value each time it is sampled.  This is used to avoid reacting
          to spurious noise on the digital inputs.  Defaults to 250 ms.


    fixedOverlapMode:
          Some supplies allow a new setpoint command to "overlap" the execution
          of an ongoing RAMP command while others consider this an error.  For
          those that do NOT allow ramping to be interrupted this way, ramping
          must be broken in to a sequence of short-lived intermediate RAMP
          commands.  This mimimizes the problem of not being able to act on a
          newer setpoint received before the previous one was reached.

          Normally, this value should be 0, which allows the firmware to
          determine which method to use based on the IDN string returned by
          the supply.  If there is a problem, the automatic setting can be
          overridden by setting fixedOverlapMode to a non-zero value:

             1: Assume supply supports interrupting an ongoing RAMP
             2: Assume supply does NOT support interrupting an ongoing RAMP


    logStateSecs:
          The minimum # of seconds since the state of the supply was last
          written to the system log while the supply is ON.  When the supply
          is OFF, the interval is 1800 seconds (30 minutes).  Defaults to 600
          seconds (10 minutes).


    maxAmps:
          The maximum setpoint value the firmware will accept.  Defaults to
          100 amps.


    maxProjErr:
          Maximum difference between the most recent serial readback value
          and the projected readback that is considered normal.

          NOTE: The whole process of the supply updating the readback value,
          the firmware requesting the value, and the supply responding to
          the request results in a considerable delay and uncertainty as to
          when the returned value was even relevant.  Add to this the fact
          that the supply often DOESN'T update the value before replying to
          the query and it means there is almost ALWAYS a significant
          difference between the value returned by the supply and the
          projected value at the time the supply responded.

          NOTE:  Testing using an analog readback that is updated every 10ms
          clearly shows that at least one model responds to RAMP commands in
          a timely manner but often returns readings that are 1 second or more
          out of date (usually at the start or end of a RAMP command).

          Defaults to 2 amps.


    maxProjErrAge:
          The maximum period of time for which the difference between the
          serial and projected readback values can consistently exceed the
          value of maxProjErr.  When this time limit is exceeded, the firmware
          logs the error and resets the projection logic to start from the
          most recent reading.  Defaults to 2000 ms.


    maxVolts:
          The maximum voltage setting for the supply.  Currently applies
          only to use of the serial link for the upper voltage limit.  Values
          larger than the maximum value returned by the supply have no effect.
          Defaults to 50 volts.


    rampRate:
          The rate at which the supply's current output should be changed when
          ramping to a new setpoint.  Defaults to 50 amps/sec.


    safeSwAmps:
          The maximum current level below which it is safe to change the state
          of the connection to the load.  Defaults to 2 amps.


    stepRampFactor:
          For supplies that do not allow an ongoing RAMP command to be
          interrupted, ramping is done by sending a series of shorter-lived
          RAMP commands.  To compensate for the effect of the delays between
          each RAMP command on the overall average rate, the period for
          each of the individual commands is shortened by multiplying it
          by stepRampFactor, which must be a value > 0 and < 1.  Smaller
          values increase the rate.  Defaults to 0.65.


    switchTime:
          The # of milliseconds the firmware waits after each change to the
          state of the polarity switches.  Must be large enough to insure that
          the connections have changed and are stable again.  Defaults to
          700 ms.

    serInitTime:
          The max # of seconds between when the power to the supply is enabled
          and communication via the serial link is established. Applies only to
          ctlMode settings that specify use of the serial link.  Default is
          12000 (12 seconds).

    maxLogAllInterval:
          When the serial link is not functioning as expected, all the traffic
          on the port will periodically be logged (in short bursts) to aid in
          diagnosing the problem.  To avoid continually flooding the system
          log, we limit how long we do it for each time and increase the amount
          of time we wait before doing it again (the later being reset each
          time we do manage to establish communication). maxLogAllInterval sets
          the maximum value (in seconds) the interval may be increased to.

    logAllPeriod:
          How long (in seconds) a period of automatic logging of all serial
          port traffic lasts (see maxLogAllInterval).

*/

//-----------------------------------------------------------------------------
// TO-DO:
//
//   - Keep track of the time each entry in the psCmds[] table was last read?
//     And/Or a flag indicating whether or not it WAS read during the last
//     pass? (in the later case, keeping track of the time for the last pass
//     might be sufficient, although less flexible).
//
//
//   - Try to detect invalid responses to queries and avoid changing the
//     value specified in the table if the format of the response is not
//     as expected.
//
//
//   - Ramp-to-0 for OFF command:
//
//     Currently, the code simply shuts off the power to the supply for an OFF
//     command.  But this puts unecessary wear and tear on the supply and the
//     load.  What it should really be doing is ramping to 0 and then disabling
//     the output of the supply.  For supplies with a serial connection, this
//     also means we can continue to communicate with the supply.
//
//     For analog controlled suplies, IF there is a signal that controls the
//     state of the supply's output, then it would also make sense to add that
//     capability as well.  But of course, that assumes the controller board
//     has a connection we can use for that, and even it if does, it would
//     require wiring changes.
//
//     Notes on required code/logic changes:
//
//     Logic currently simply shuts off the power to the supply for an OFF
//     command.  What is SHOULD do is ramp the output to zero then disconnect
//     the load, leaving the power to the supply enabled.  A new command for
//     actually disabling the power to the load should be added, but that
//     should be fairly simple to add once the rest of the changes are
//     completed.
//
//     xxx.assertOn:
//       Change name and logic to enablePSPower
//
//     xxx.assertPolarity:
//       Needs to be expanded to include disconnecting the load while the power
//       to the supply remains enabled.
//
//     xxx.state:
//       Add STATE_STANDBY?
//
//     xxx.ampSet:
//       The most recent setpoint, this also currently servers as the target
//       for the PS output.  But we need to replace most references to it with
//       ampTarget because the target output will need to be 0 for an OFF
//       command and we don't want to overwrite the current setpoint.
//
//     xxx.ampSetPolarity:
//       Similar to the changes for xxx.ampSet, we need a ampTargetPolarity
//       to use in place of the ampSetPolarity in most cases.
//
//     xxx.ampApply:
//       Add support for ramping to 0 for purposes of going in to standby mode.
//       Currently uses xxx.ampSet for its goal, but that holds the most recent
//       setpoint, which we don't want to overwrite just to go in to standby,
//       so it will need to use xxx.ampTarget instead.
//
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//  Serial interface vs using an analog setpoint:
//
//     In LOCAL mode, the Sorensen supplies use the analog inputs to set the
//     current and voltage limits.  In REMOTE mode, these inputs are ignored.
//
//     Sending pretty much ANY command to a Sorensen supply via one of the
//     serial interfaces causes the supply to change to REMOTE mode.  Even
//     most of the queries that have no effect on the state of the supply
//     will cause this to happen.  The only exceptions are the SYST:LOCAL
//     commands and queries and (for at least one model) the MEAS commands.
//     ANYTHING else (even something fundamental like *IDN?) sets REMOTE mode.
//
//     So, for analog control of the setpoint to be used, the serial interface
//     can only be used to force LOCAL mode (SYST:LOCAL 1) and (possibly) to
//     read the current and voltage outputs using the MEAS: commands.
//
//  NOTE:
//     One possible crude method of allowing the serial interface to be used
//     in the background:  Use the serial interface only when the setpoint
//     is idle and the supply is on.  Always end each pass with the commands
//     to force local mode and query the local mode state (and repeat as
//     needed until local mode is reconfirmed or the passive serial mode has
//     been disabled.
//
//     WARNING:
//       The problem with this approach:  All the other commands (even queries
//       other than the MEAS ones) change the mode back to remote.  The result
//       is that the supply oscillates back and forth between the analog and
//       serial setpoint.  Even if I try to keep the serial settings in sync
//       with the analog, the two will always result in a bit different
//       outputs.  And the lag between a change to the analog setpoint and
//       the serial one means ramping will result in an odd sawtooth or
//       staircase pattern.
//
//     As much as I would like to use the serial interface to get status
//     information to keep tabs on the state of a supply being controlled by
//     an analog input, at least SOME of the Sorensen supplies simply don't
//     seem to have any practical way to support this.  You simply CANNOT
//     use the serial interface for much of anything in a passive read-only
//     mode.
//
//  Bottom line:
//     I can use the serial interface (if available and connected) to force
//     the supply in to local mode when needed, but the only information
//     (besides the current LOCAL/REMOTE mode) I can (possibly) get safely
//     (for some supplies) via the serial interface (while it is in LOCAL
//     mode) is the current and voltage readings.
//
//     So, the analog values can be a backup to the serial readings when
//     serial mode is being used, and the serial readings can (in some cases)
//     be read safely while using analog control, but the only way to get
//     additional status information is to use the serial interface for the
//     setpoint.
//
//-----------------------------------------------------------------------------
//  NOTES on serial-controlled supplies with no analog readback connections:
//
//    At least some of the Sorensen supplies do not provide accurate and
//    timely readbacks (at least at the beginning and end of a RAMP cmd).
//    Add to this the fact that the rate at which we can get a (usually)
//    updated readback value from the supply is quite low.  Together, these
//    issues make it impractical to produce an accurate projection of what
//    the supply is doing between serial readings if we rely entirely on
//    the readback values (as I know all to well, as I tried very had to
//    do just that using ONLY the readings...)
//
//    The alternative (which DOES work quite well) is to use the settings
//    we send to the supply as the basis for projecting the current output
//    of the supply, using the readings only to make minor corrections or
//    detect when the error is great enough to warrant disabling the
//    projection logic.
//-----------------------------------------------------------------------------


//=============================================================================
//
//  The purpose of the Reversing Switch controller is to provide a bi-polar
//  output from a Power Supply that is NOT capable of bi-polar output.  This
//  is achieved by using a pair of relays to swap the (+) and (-) connections
//  from the power supply to the load.
//
//  To avoid possible damage to the load or the power supply, a change in the
//  output that requires changing the polarity is done by 1st setting the
//  output of the supply to zero and switching the polarity only when the
//  output is close to zero, and then ramping the output to the desired level
//  using the new polarity.
//
//=============================================================================

//=============================================================================
//
//  There are two modes of operation for the controller:  Analog and Serial
//
//  In Analog mode, the current output of the power supply is controlled by
//  an analog voltage, and all readbacks are from analog inputs.  All analog
//  values are scaled as necessary, and the scalings can/must be set to
//  calibrate the values.
//
//  In serial mode, control and monitoring of the power supply is done via
//  a serial connection.  The analog inputs are still in operation and can be
//  used to monitor what the supply is actually doing to make sure it
//  agrees with what it says it is doing.
//
//-----------------------------------------------------------------------------
//
//  The normal (non-energized) state of the relays that connect the power
//  supply to the load is such that the (-) output from the supply is connected
//  to both the (-) and (+) connections to the load, and the (+) output from
//  the supply is connected to nothing.  If both the relays are energized, then
//  the (+) output from the supply is connected to both, and the (-) output is
//  connected to nothing.
//
//  Note that in both these cases, this insures the supply's output sees no
//  load, and the inputs to the load are shorted.  This way, it does not matter
//  which state the relays are in when the controller boots, so long as they
//  are the same.  One OR the other relay must be energized to connect the load
//  to the power supply, and which one is energized is what determines the
//  polarity the load sees.
//
//-----------------------------------------------------------------------------
// Signal usage:
//
//  DI-10 - Ext Interlock - 1 = external interlock, 0 = no external interlock
//
//  DI-11 - On Feedback - 1 = PS has no power, 0 = power connected to PS
//
//  For power to be connected to the power supply, there must be no external
//  interlock, and the ZWorld must assert on ON condition
//
//------------------------------------------------------
//  D0-00 - Normal Polarity LED
//
//  DO-01 - Reverse Polarity LED
//
//  DO-07 - Relay that controls power going TO the Power Supply
//          0 = ground side of relay connected to ground
//          (also requires NO Ext. Interlock for PS power relay to energize)
//
//  DO-08 - Activity LED
//
//  DO-09 - Normal Polarity Relay - energized for Normal Polarity (supply's
//                       (+) goes to load's (+)
//
//  DO-10 - Reverse Polarity Relay - energized for Reversed Polarity (supply's
//                       (+) goes to load's (-)
//
//------------------------------------------------------
//  AI-00 - Internal Cur Reading:  Load current read as voltage across shunt
//                       resistor in series with the load.
//
//  AI-01 - Internal Volt Reading: Voltage as read across the load-side of
//                       the polarity-mode relays and the shunt resistor.
//
//  AI-02 - External Volt Reading:  Voltage as read from some external analog
//                       signal (generally across the output of the power
//                       supply), BEFORE going through the polarity relays and
//                       the shunt resistor.
//
//  AI-03 - External Cur Reading:  PS output current, read from some external
//                       analog input.
//
//
//------------------------------------------------------
//  AO-00 - Load Current setpoint: Analog control of Power Supply's output
//
//
//------------------------------------------------------
//  Serial D (/dev/tty3):  For supplies that support it, a serial connection
//                       used to send cmds and read the state of the supply.
//
//=============================================================================


#include <common.h>

#include <unistd.h>
#include <sys/types.h>

#include <math.h>


//APPLCODE - change to match the name of this file
#ifndef  ApplicSource
  #define  ApplicSource  "RevSwitch_Appl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "RevSwitch" };
const char ApplAbbrev[] = { "RevSw" };


//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_RevSwitch

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN      4   // analog inputs
#define  NUM_APPL_AOUT     1   // analog outputs
#define  NUM_APPL_DIN     12   // digital inputs
#define  NUM_APPL_DOUT    11   // digital outputs


#define SharedINC
#include <sharedAppl.c>


//--- the 8 least signif bits of DiagFlags are for appl/cmd-specific events ---
#define  PS_MSG_BITS       0x02L

#define  ShowPSMsgs       (DiagFlags & PS_MSG_BITS)


#define  ID_STR_SIZE   64   // max length of ID str read from PS (DO NOT CHANGE!)

#define  SerComDelay   4    // secs b/w power applied and serial com attempted


//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//APPLCODE - define macros for each Digital bit and each Analog index

//=============== digitalIn bit usage ===============
#define  UNUSED_DI_00    0x00000001  // 00:
#define  UNUSED_DI_01    0x00000002  // 01:
#define  UNUSED_DI_02    0x00000004  // 02:
#define  UNUSED_DI_03    0x00000008  // 03:
#define  UNUSED_DI_04    0x00000010  // 04:
#define  UNUSED_DI_05    0x00000020  // 05:
#define  UNUSED_DI_06    0x00000040  // 06:
#define  UNUSED_DI_07    0x00000080  // 07:
#define  UNUSED_DI_08    0x00000100  // 08:
#define  UNUSED_DI_09    0x00000200  // 09:
#define  EXT_INTERLOCK   0x00000400  // 10: External Interlock: 1 = Interlock
#define  PS_POWER_OFF    0x00000800  // 11: PS Status: 0 = PS ON, no interlocks

#define  DIN_MASK_APPL   0x00000C00

const uchar  LogicOrderDIn[NUM_APPL_DIN] =  {
   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11 };

/*EXAMPLE:
#define  ClosedLoop(din)        ( (bool)((din) & CLOSED_LOOP) )
#define  FinalBtnPushed(din)    ( (bool)((din) & FINAL_BTN) )
#define  InitialBtnPushed(din)  ( (bool)((din) & INITIAL_BTN) )
#define  RFisOff(din)           ( (bool)((din) & RF_IS_OFF) )
#define  LocalMode(din)         ( (bool)((din) & LOC_REM_SW) )
*/

//--- msgs to log automatically when a digital input changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//--0 state descr--     --1 state descr--
   NULL,                 NULL,
   NULL,                 NULL,
   NULL,                 NULL,
   NULL,                 NULL,
   NULL,                 NULL,
   NULL,                 NULL,
   NULL,                 NULL,
   NULL,                 NULL,
   NULL,                 NULL,
   NULL,                 NULL,
  "Ext Intlk Cleared",  "Ext Intlk Asserted",
  "PS power Enabled",   "PS power Disabled"
};

//=============== digitalOut bit usage ===============

#define  NORM_POLARITY_LED  0x00000001  // 00: 0 ==> turns on Normal Polarity LED
#define  REV_POLARITY_LED   0x00000002  // 01: 0 ==> turns on Reverse Polarity LED
#define  UNUSED_DO_02       0x00000004  // 02:
#define  UNUSED_DO_03       0x00000008  // 03:
#define  UNUSED_DO_04       0x00000010  // 04:
#define  UNUSED_DO_05       0x00000020  // 05:
#define  UNUSED_DO_06       0x00000040  // 06:
#define  TURN_PS_OFF        0x00000080  // 07: 0 ==> enables power to PS
#define  ActivityLED        0x00000100  // 08: activity LED
#define  NORM_POLARITY_OFF  0x00000200  // 09: disable Normal Polarity output
#define  REV_POLARITY_OFF   0x00000400  // 10: disable Reverse Polarity output

#define  DOUT_MASK_APPL     0x00000783

#define  POLARITY_LEDS      0x00000003  // both polarity LED outputs


const uchar  LogicOrderDOut[NUM_APPL_DOUT] =  {
   0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 };

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  (  NORM_POLARITY_LED  \
                             | REV_POLARITY_LED   \
                             | TURN_PS_OFF        \
                             | NORM_POLARITY_OFF  \
                             | REV_POLARITY_OFF   \
                            )

#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

 //--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--       --1 state descr--
   NULL,                   NULL,                   // 00
   NULL,                   NULL,                   // 01
   NULL,                   NULL,                   // 02
   NULL,                   NULL,                   // 03
   NULL,                   NULL,                   // 04
   NULL,                   NULL,                   // 05
   NULL,                   NULL,                   // 06
  "Assert PS ON",         "Assert PS OFF",         // 07
   NULL,                   NULL,                   // 08
  "Enable [+] polarity",  "Disable [+] polarity",  // 09
  "Enable [-] polarity",  "Disable [-] polarity"   // 10
};


//=============== Analog inputs ===============
//--- indexes for anaIn[] and readVal[] ---
#define  INT_AMPS_READ    0   // volt across shunt resistor in series w/load
#define  INT_VOLT_READ    1   // volt to load (after relays, shunt resistor)
#define  EXT_VOLT_READ    2   // volt from P.Supply (before relays/resistor)
#define  EXT_AMPS_READ    3   // current readback from external device

#define  ADC_MASK_APPL  0x0000000FL

const uchar  LogicOrderAIn[NUM_APPL_AIN] = { 0, 1, 2, 3 };


//=============== Anlog outputs ===============
//--- indexes for anaOut[] and setVal[] ---
#define  OUTPUT_CTL         0   // setpoint - ctl volt to analog PI circuit

#define  DAC_MASK_APPL  0x00000001L


const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0 };


//=============================================================================
//  Application-specific data structures and definitions
//=============================================================================

// Each value in the ApplState structure should be throught of as representing
// a "signal" in a clocked logic circuit, with the "applic" task representing
// the logic implemented by the circuit elements such as gates, latches, etc.

// The main "applic" task is the highest priority task and executes the code
// that implements the application logic on a real-time, periodic basis.

//                 ====== IMPORTANT NOTE on LOCKING ======
// Remember to use a semaphore or some other form of locking mechanism to
// protect all access to shared values by more than one task.  The only
// exception to this is when the applic task needs to access a value protected
// by other tasks using only a scheduler/kernel lock (the KLOCK and KUNLOCK
// macros).  Because the applic task is the only task that runs at the highest
// priority, no other task can preempt it, and if the scheduler is locked,
// the applic task cannot preempt another task.  But scheduler locks must be
// kept short to avoid creating large latency in the timely execution of other
// tasks (the applic task in particular).


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct {
 //-----------------------------------------------------------------------
 // SOFTWARE SETTING INPUTS
 //
 //  These values represent Software Inputs to the application logic.
 //
 //  These values are set in the "Next" instance by tasks OTHER than the
 //  "applic" task to change some value used by the application logic.
 //  Remember that all tasks besides the main "applic" task must use a
 //  scheduler lock to protect reads and writes to these values.
 //
 //  Note that the application task may alter these values (e.g. to change
 //  an invalid value in to a valid one).
 //-----------------------------------------------------------------------
  bool   onRequest;
  bool   offRequest;
  bool   resetRequest;

  float  ampSet;          // most recent setpoint


 //-----------------------------------------------------------------------
 // READINGS
 //
 //  These values represent "passive" inputs to the application logic.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.
 //
 //  These values generally come from one of two sources:
 //    - The just-read state of analog and/or digital hardware inputs
 //    - Values maintained by other tasks
 //
 //  While some of these values are simply copies of "external" values,
 //  including them in the ApplState structure allows us to see how
 //  they change between cycles.  Also, some of these values are the result
 //  of applying logic to "external" values.  Applying that logic once at
 //  at the beginning of the applic loop and saving the result eliminates
 //  the need to repeat the same logic in multiple blocks of code later on,
 //  avoiding unnecessary redundancy.
 //-----------------------------------------------------------------------
  ulong  processTime;     // time (in ms) when processing was started
  ulong  digitalIn;       // state of digital inputs

  // readings via the serial link to the PS (if used)
  float  serSetA;         // cur limit setting read from PS via serial link
  float  serSetV;         // volt limit setting read from PS via serial link

  uint   opStatus;        //
  uint   alarmStatus;     //
  uint   progStatus;      //

  uint   serProtCond;     // Protection Condition
  uint   serStatusFlags;  // Status Flags

  bool   psOutputOn;      // we know (or assume) the PS output is ON

  bool   serLinkActive;   // getting responses from PS via serial link

 //-----------------------------------------------------------------------
 // OUTPUTS from UN-CLOCKED (asynchronous, gate-level) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a non-clocked logic element (such as a gate).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains ONLY values that are set by a
 //  block of code that represents a NON-clocked logic element.
 //
 //  These values are set in the "Next" instance by the applic task just
 //  after reading the state of the hardware inputs.  They must NOT be
 //  altered by other tasks (although they can read them as long as they
 //  use appropriate locking to do so).
 //
 //  Because these values represent the outputs of un-clocked logic elements,
 //  and such elements often need to be cascaded (the outputs used as inputs
 //  to other non-clocked elements), the result is often dependent on the
 //  order of execution of the blocks of code.
 //-----------------------------------------------------------------------
  bool   unused_01;        // was "interlocks"
  bool   unused_02;        // was "onIsAsserted"

  bool   externIntlk;      // External Interlock is being asserted
  bool   psPowerEnabled;   // power to PS is enabled

  int    ampSetPolarity;   // polarity of current ampSet

 //-----------------------------------------------------------------------
 // OUTPUTS from CLOCKED (latched inputs and outputs) LOGIC ELEMENTS
 //
 //  These values are outputs from one of the blocks of code in the applic
 //  task that represent a clocked logic element (such as latches).
 //
 //  Each value here is set by one (and ONLY one) of the blocks of code in
 //  the applic task.  This group contains only values that are set by a
 //  block of code that represents a CLOCKED logic element.
 //
 //  These values are set in the "Next" instance by the applic task AFTER
 //  all the other ApplState values are set.  Because they represent only
 //  the clocked logic elements, and each block of code reads only values
 //  from the Cur instance, the blocks can be executed in any order and the
 //  result will always be the same.
 //
 //  These values are used as inputs by other blocks and/or to set values
 //  used by other tasks (such as values read by the ModBus server task).
 //
 //  Again, these values must NOT be altered by other tasks (although they
 //  can read them as long as they use appropriate locking to do so).
 //-----------------------------------------------------------------------
  ulong  digitalOut;       // state for digital outputs

  bool   assertReset;      // assert reset cmd for error/"BAD" inputs
  ulong  resetStarted;     // time reset cmd was initiated

  int    state;            // state of application logic

  bool   assertOn;         // assert ON cmd for PS
  int    assertPolarity;   // -1, 0, or 1  (0 means load disconnected)

  float  ampApply;         // value to apply to PS

} ApplState;


float  lastRampRate, maxAmpChg, projectRate, ampTarget,
       projectFrom = 0.0, projectBy = 0.0, maxProjectBy = 0.0, serRampTo = 0.0;

bool  calibErr = false, serLinkRequired = false, serLinkNotWorking = false,
      allowOverlapSets = false, ramping = false, polarityStable = true,
      logAll = false, offDueToReset = false, loggedResetIntlk = false;


U32  rampStartedTime, rampPeriod, lastPolarityChgTime = 0, projectFromTime = 0,
     psOutputOnTime = 0;

U16 psPowerEnabledSecs = 0, lastLogAllSecs = 0, logAllInterval;

int  rdgPolarity = 0;  // the load polarity the load was last connected as


U16  assertOnChgd = 0, lastStateErrMsg = (U16)(-9999),
     lastLogStateTime = 0, lastRamping = 0, oldestProjErrTime = 0;


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState

const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );


//----- state values -----
#define  STATE_OFF     0   // PS off, idle
#define  STATE_ON      1   // PS on or polarity switch in progress
#define  STATE_INTLK   2   // error occured that needs acknowledgement

const char *StateDesc[] = { "OFF", "ON", "INTERLK" };



//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//APPLCODE - add tasks that service network clients

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ----------------

//--- standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);





//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


//APPLCODE - set # of devices for which we want duplicate register sets
#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of coils
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read-only registers
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,     x, 30001, U32 )
REG3( rtosVersTime,    x, 30003, U32 )
REG3( applVersTime,    x, 30005, U32 )
REG3( devType_,        x, 30007, U16 )
REG3( deviceID_,       x, 30008, U32 )
REG3( upSecs,          x, 30010, U32 )
REG3( percentIdle_,    x, 30012, U16 )

REG3( numChans,        x, 30013, U16 )
REG3( coilChanMult,    x, 30014, U16 )
REG3( reg3ChanMult,    x, 30015, U16 )
REG3( reg4ChanMult,    x, 30016, U16 )

REG3( digInBits,       x, 30017, U32 )

REG3( statusBitsStdA,  x, 30019, U16 )
REG3( statusBitsStdB,  x, 30020, U16 )
REG3( statusBitsAppl,  x, 30021, U32 )

REG3( latency,         x, 30023, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,     x, 30024, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,    x, 30025, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime,  x, 30026, U16 )  // # ms used by last kernel pass

//APPLCODE - add registers for externally available read-only values
//
// for serial connection to power supply
REG3( numCmdsSent,     x, 30027, U16 )  // # cmds sent to the PS
REG3( numRespRcvd,     x, 30028, U16 )  // # responses from PS
REG3( numRespLost,     x, 30029, U16 )  // # responses we never got
REG3( consecGoodResp,  x, 30030, U16 )  // # consecutive valid resps from PS
REG3( consecBadResp,   x, 30031, U16 )  // # consecutive invalid resps from PS

REG3( serOnline,       x, 30032, U16 )  //
REG3( serStatusFlags,  x, 30033, U16 )  // Status Flags
REG3( serStatus,       x, 30034, U16 )  //
REG3( serAccumStatus,  x, 30035, U16 )  //
REG3( serFaultsMask,   x, 30036, U16 )  //
REG3( serFaults,       x, 30037, U16 )  //
REG3( serErrors,       x, 30038, U16 )  //
REG3( serModelVolts,   x, 30039, F32 )  //
REG3( serModelAmps,    x, 30041, F32 )  //
REG3( serModelOverV,   x, 30043, F32 )  //
REG3( serProtCond,     x, 30045, U16 )  // Protection Condition

REG3( ampApply,        x, 30046, F32 )  /* desired active setpoint */
REG3( activeSetAmps,   x, 30048, F32 )  /* actual active setpoint */

REG3( ampRead,         x, 30050, F32 )  /* output current */
REG3( voltRead,        x, 30052, F32 )  /* output voltage */

REG3( intAmps,         x, 30054, F32 )  /* Amps thru internal shunt */
REG3( intVolts,        x, 30056, F32 )  /* Volts across ctlr's connection to load */

REG3( extVolts,        x, 30058, F32 )  /* Voltage from extern signal */
REG3( extAmps,         x, 30060, F32 )  /* Amps from external signal */

REG3( serSetA,         x, 30062, F32 )  /* max amp setting to/from PS */
REG3( serSetV,         x, 30064, F32 )  /* max volt setting to/from PS */

REG3( serReadA,        x, 30066, F32 )  /* amp reading from PS */
REG3( serReadV,        x, 30068, F32 )  /* volt reading from PS */

REG3( serProjErr,      x, 30070, F32 )  /* latest projection error */
REG3( serProjErrAge,   x, 30072, U16 )  /* ms since oldest consecutive projection error */

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

//--- values for serStatusFlags ---
#define  REMOTE_MODE   0x0001  // Remote mode was selected
#define  SF_UNUSED_02  0x0002
#define  PON_SRQ       0x0004  // PON Service Req selected by rear panel switch
#define  SRQ_SENT      0x0008  // GPIB Serv Req issued and serial poll not rcvd
#define  EEPROM_LCKD   0x0010  // The EEPROM is locked
#define  TRIP_OCCURRED 0x0020  // Over Volt or Over Temp trip occurred
#define  SF_UNUSED_40  0x0040
#define  SF_UNUSED_80  0x0080
#define  WAIT_TRIG     0x0100  // Volt and Cur output wating for trigger
#define  FOLDBACK_ENA  0x0200  // Foldback protection is enabled
#define  SRQ_ENA       0x0400  // Serv Req generations is enabled
#define  OUTPUT_ON     0x0800  // Output is on
#define  SF_LOG_CHGS   0x073C  //--- which bits to log changes for ---

//--- values for serProtCond ---
#define  CONST_VOLT    0x01  // Constant Voltage operation
#define  CONST_CUR     0x02  // Constant Current operation
#define  PC_UNUSED_04  0x04
#define  OVER_VOLT     0x08  // Overvoltage protection tripped
#define  OVER_TEMP     0x10  // Overtemperatur protection tripped
#define  EXT_SHUTDOWN  0x20  // Supply external shutdown active
#define  FOLDBACK_MODE 0x40  // Foldback mode operation
#define  REM_PROG_ERR  0x80  // Remote programming error

#endif


//--------------------------------------------------------------
//APPLCODE - set max # of registers per set of read/write registers
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

/*
  The "standard" functions for processing writes to Modbus register values:

    setS16, setU16, setS32, setU32, setF32

  These functions simply log a message for the change and update the
  value in a critical block (so the change cannot be interrupted).

  In many cases, just allowing any setting to be written to a register (or
  multiple registers comprising a larger value) is not acceptable.


  If any of the following apply, then a custom func must be used instead:

    - The order of changes is important.  For example, if you get ON,
      OFF, and RESET commands in rapid succession, you have to be sure
      they are processed in the order they were recieved.

    - Not all possible settings are acceptable.  The acceptable settings
      for a value are often a subset of the possible ones, and frequently
      are dependent upon the current state of the controller.


  IMPORTANT NOTES:

    The "applic" task periodically executes the code that makes up the
    bulk of the application logic.  Because it runs at the highest priority
    there is no chance that a lower priority task (like the daemon handling
    Modbus register writes) can change a value in the middle of this
    processing.

    However, the applic task CAN (and often does) preempt other processes,
    so any process that does read or make changes to the register values
    must lock the register values before doing so.  This is currently done
    by disabling the scheduler using the KLOCK() and KUNLOCK() macros.

    Use of such a lock insures that changes occur only between passes
    through the application logic, but there are still several issues
    that must be addressed when writing functions to handle writes to
    register values:

      - Priority inversion and latency:  Allowing a lower priority task
        to lock out higher ones (or, in this case, ALL tasks) allows
        the lower priority task to run when a higher-priority one should
        be running.  Care must be taken to keep such periods as short
        as possible to avoid disrupting the timing of the control logic.

      - Loss of encapsulation:  The last thing we want is to break the
        application logic in to small pieces spread through a bunch of
        functions executed by different tasks at different times.

        While the logic for processing writes to register values is often
        dependent upon the state of the control logic, functions outside
        the applic task must NOT directly alter the state of the application
        logic.  To achieve this, such functions must limit changes to
        values that act as input to the applic task.

        To this end, you need to treat the register values that act as
        inputs just like they were knobs, switches, analog inputs, or
        latches.  Like their hardware equivalent, the applic task logic
        reads their state at the beginning of each pass and uses this
        information when determining what the next state of the control
        logic should be.  And just like a hardware latch, values can be
        reset after reading them so the next write to the value can be
        treated as another instance of a request.

    The trickiest circumstances to handle are ones where the order of separate
    writes is important.  The most common example are commands for ON, OFF,
    and RESET.  If, for example, you had 3 separate flags, and simply set
    each once corresponding to which of the 3 commands you received, you
    end up with multiple flags set when the applic logic is processed, and
    no way to determine what the end result should be.  In these cases,
    the logic that controls the state of each flag should be placed in its
    own function and called during the processing of any event that could
    affect the flag's state.

    Ideally, the logic for setting each variable would be triggered each
    time an event occurs that could affect it:  A publish-subscribe type
    of mechanism.  One possibility to do this efficiently?  Use a bit-mask
    "array" to determine which functions to call for each register that is
    changed.  That way it doesn't take many bytes to determine which functions
    to call, and each function can update its corresponding register value
    based on its current state and the event that just occured.

    Or should I simply combine such values in to one (a struct, if necessary),
    and have a single function that is called for all the related values
    and updates the state of the single logical entity based on its
    current state and what event called it?  (But the # of combinations could
    get out of hand easily...)
*/

#ifdef REG4

REG4( timeZone_,        x, 40001, S16, RW, MB_timeZone )
REG4( curTime,          x, 40002, U32, RW, setU32 )

REG4( ipAddr,           x, 40004, U32, RW, setU32 )
REG4( netmask,          x, 40006, U32, RW, setU32 )
REG4( gateway,          x, 40008, U32, RW, setU32 )

REG4( setACoil,         _, 40010, U16, RW, setU16 )
REG4( resetACoil,       _, 40011, U16, RW, setU16 )

REG4( digOutBits,       x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,       x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,      x, 40016, U32, RW, MB_cmdBitsAppl )

//APPLCODE - add registers for externally available read/write values
REG4( ampSet,           x, 40018, F32, RW, MB_ampSet )  /* desired output */

REG4( devStatus,        x, 40020, U16, RO, setU16 )  /* bit-mapped status */

REG4( maxVolts,         x, 40021, F32, RW, setF32 )  /* max voltage allowed */
REG4( maxAmps,          x, 40023, F32, RW, setF32 )  /* max current allowed */

REG4( safeSwAmps,       x, 40025, F32, RW, setF32 )  /* max A allowed for polarity swap */
REG4( connectLoadDelay, x, 40027, U16, RW, setU16 )  /* PS ON to load connect delay */

REG4( rampRate,         x, 40028, F32, RW, setF32 )  /* setpoint chg rate in amps/sec */

REG4( switchTime,       x, 40030, U32, RW, setU32 )  /* # ms req for polarity switch */

REG4( debounceMS,       x, 40032, U16, RW, setU16 )  /* ms period for digIn debouncing */

REG4( ctlModes,         x, 40033, U16, RW, MB_ctlModes) /* modes for settings/readings */

REG4( logStateSecs,     x, 40034, U16, RW, setU16)

REG4( stepRampFactor,   x, 40035, F32, RW, MB_stepRampFactor) /* to compensate for piecewise ramp mode */

REG4( maxProjErr,       x, 40037, F32, RW, setF32) /* max err in projected readbacks */
REG4( maxProjErrAge,    x, 40039, U16, RW, setU16) /* max interval for consecutive projection errors */

REG4( fixedOverlapMode, x, 40040, U16, RW, setU16) /* override for automatic setting of allowOverlapSets */

REG4( serInitTime,      x, 40041, U16, RW, setU16) /* PS on to ser link working*/

REG4( maxLogAllInterval, x, 40042, U16, RW, setU16) /* max secs b/t ser err logging */
REG4( logAllPeriod,      x, 40043, U16, RW, setU16) /* max secs to log ser errs */

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion


//--- definitions for cmdBitsAppl ---
//#define  CBA_DoReboot   0x00000001  /* reboot ctlr */
//#define  CBA_DoInit     0x00000002  /* tell ctlr to (re)initialize */
#define  CBA_TurnOn     0x00000100  /* connect PS to load, turn it On */
#define  CBA_TurnOff    0x00000200  /* disconnect from load, turn it Off */
#define  CBA_Reset      0x00000400  /* clear Interlocks */


/*------------------- bit usage for ctlModes -----------------------
    3 groups of bits that define how to apply the setpoint and read
    the current and voltage.

    NOTES:
      - Only 1 of the analog bits can be set in each group
      - If both the serial bit and 1 of the analog bits is set, the
        specified analog set or read will be used whenever the serial
        link is not active.
      - "Int" stands for "Internal" (meaning the reading is from the
        internal shunt or across the output terminals of the ctlr case)
      - "Ext" stands for "External" (meaning the reading is from an
        external analog signal)
 *-------------------------------------------------------------------*/
#define  CM_Ser_SetA     0x0400  // apply setpoint via serial link
#define  CM_Ext_SetA     0x0100  // apply setpoint via analog output

#define  CM_Ser_ReadA    0x0040  // use serial link for current readback
#define  CM_Int_ReadA    0x0020  // use shunt voltage for current readback
#define  CM_Ext_ReadA    0x0010  // use external signal for current readback
#define  CM_Ana_ReadA    0x0030     // either analog mode

#define  CM_Ser_ReadV    0x0004  // use serial link for voltage readback
#define  CM_Int_ReadV    0x0002  // use shunt voltage for voltage readback
#define  CM_Ext_ReadV    0x0001  // use external signal for voltage readback
#define  CM_Ana_ReadV    0x0003     // either analog mode

#define  CM_Valid_Bits   0x0577

// bits within each nibble indicating which mode(s) are allowed/enabled
#define  CM_Serial            4  // use serial if serLinkActive
#define  CM_Analog_Int        2  // internal analog signal
#define  CM_Analog_Ext        1  // external analog signal

#define  CM_Serial_Bits  0x0444  // Serial bit mask for all 3 values
#define  CM_Analog_Bits  0x0133  // Analog bit mask for all 3 values

#define  CM_SetA_Bits    0x0500  // Set Amps bit mask for all modes
#define  CM_ReadA_Bits   0x0070  // Read Amps bit mask for all modes
#define  CM_ReadV_Bits   0x0007  // Read Volts bit mask for all modes



//--- bit usage for devStatus --- (some bits based on old TrimCoil PS device) ---
#define  DS_INTLK        0x0001   // Interlock condition active
#define  DS_ON           0x0002   // Logical ON state
#define  DS_NORM_POL     0x0004   // Normal (+) Polarity state in effect
//#define  DS_RTZ          0x0008   // ramping PS to 0
//#define  DS_PS_IS_OFF    0x0010   // PS is not powered up
#define  DS_CTLR_RESET   0x0020   // PS off since last CPU reset

#define  DS_SER_ERR      0x0100   // Serial PS not responding
#define  DS_CALIB_ERR    0x0200   // serial & analog readings vastly different


//--- values for fixedOverlapMode ---
#define  OVERLAP_AUTO      0  // allowOverlapSets based on PS ID string
#define  OVERLAP_ALLOWED   1  // force allowOverlapSets to true
#define  OVERLAP_DISABLED  2  // force allowOverlapSets to false


//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================
//APPLCODE - settings/values we want saved to/restored from files


//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data they represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------
RESTORE_FUNC(ctlModes);

const ParamData far  devDataParams[] = {
//APPLDATA
//                ID   ---value---  --restoreFunc--
//PERSISTENT_VAR(  1,  myVarName,     NULL ),
//                ID   ---value---  granularity  --restoreFunc--
//PERSISTENT_TBL(  2   myArrayName, elemSize,     NULL ),

  PERSISTENT_VAR(  1,  ctlModes,          PVRF(ctlModes) ),
  PERSISTENT_VAR(  2,  maxVolts,          NULL ),
  PERSISTENT_VAR(  3,  maxAmps,           NULL ),
  PERSISTENT_VAR(  4,  safeSwAmps,        NULL ),
  PERSISTENT_VAR(  5,  connectLoadDelay,  NULL ),
  PERSISTENT_VAR(  6,  rampRate,          NULL ),
  PERSISTENT_VAR(  7,  debounceMS,        NULL ),
  PERSISTENT_VAR(  8,  logStateSecs,      NULL ),
  PERSISTENT_VAR(  9,  stepRampFactor,    NULL ),
  PERSISTENT_VAR( 10,  switchTime,        NULL ),
  PERSISTENT_VAR( 11,  maxProjErr,        NULL ),
  PERSISTENT_VAR( 12,  maxProjErrAge,     NULL ),
  PERSISTENT_VAR( 13,  fixedOverlapMode,  NULL ),
  PERSISTENT_VAR( 14,  serInitTime,       NULL ),
  PERSISTENT_VAR( 15,  maxLogAllInterval, NULL ),
  PERSISTENT_VAR( 16,  logAllPeriod,      NULL ),

  #define SharedDevParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};

//----------------------------------------------------
RESTORE_FUNC(ampSet);

extern char  serIDStr[ID_STR_SIZE];  //forward ref

const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --restoreFunc--
//  PERSISTENT_VAR(  1,  rfOffSetpoint,   NULL ),
//  PERSISTENT_VAR(  2,  rfOnSetpoint,    NULL ),

  PERSISTENT_VAR(  1,  ampSet,        PVRF(ampSet) ),
  PERSISTENT_TBL(  2,  serIDStr, sizeof(serIDStr), NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --restoreFunc--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Declarations for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

void onRequest(void);
void offRequest(void);
void resetRequest(void);

void newSetpoint(float newAmps, bool logIt);

bool setCtlModes(U16 newCtlModes);


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//APPLCODE
//--- add definitions for application-specific global variables here ---
char  psCmd[64], psResp[256];

ulong  psLastRespTime = 0, psLastSendRcvTime = 0;

uint  serLastRcvTime;

char  serialNum[16], modelStr[16], serIDStr[ID_STR_SIZE];
bool  logIDStr = false;

int  ser_fflags;


typedef struct {
  int          priority;
  const char  *cmd;
  void        *val;
  uint         size;
  int          valType;
} PS_CmdInfo;


#define  STATUS_BLOCK  (NUM_MB_TYPES)

#define  PS_CMD_DELAY  20   // # ms to wait between commands


const PS_CmdInfo  psCmds[] = {
  { 1, "SOUR:CURR?",       &Next.serSetA,     sizeof(Next.serSetA),     DT_F32 },
  { 1, "MEAS:CURR?",       &serReadA,         sizeof(serReadA),         DT_F32 },
  { 1, "MEAS:VOLT?",       &serReadV,         sizeof(serReadV),         DT_F32 },
  { 2, "SOUR:STAT:BLOC?",  NULL,              0,                  STATUS_BLOCK },
  { 2, "STAT:PROT:COND?",  &Next.serProtCond, sizeof(Next.serProtCond), DT_U16 },
  { 2, "SOUR:VOLT?",       &Next.serSetV,     sizeof(Next.serSetV),     DT_F32 },
  { 9, "*IDN?",            serIDStr,          sizeof(serIDStr),         DT_STR }
};
#define SOUR_CURR_IDX  0   // index of SOUR:CURR? query
#define MEAS_CURR_IDX  1   // index of MEAS:CURR? query
#define IDN_QUERY_IDX  6   // index of *IDN? query

#define  NUM_PS_CMDS  (sizeof(psCmds) / sizeof(PS_CmdInfo))

uint  consecErrsGlobal[NUM_PS_CMDS];
uint  expectedSyntaxErrs;

//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//APPLCODE - add functions to process writes to ModBus register values

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
  ulong  cmdBits = MB_VAL(ulong);

  if (cmdBits & CBA_TurnOn)    onRequest();
  if (cmdBits & CBA_TurnOff)   offRequest();
  if (cmdBits & CBA_Reset)     resetRequest();

  cmdBitsAppl = 0;

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_ampSet(SetRegFuncParams)
{
  newSetpoint(MB_VAL(float), false);

  return true;
}


//-----------------------------------------------------------------------------
//  Set the methods used to control the setpoint and read the current and
//  voltage values.
//-----------------------------------------------------------------------------
bool _nearcall MB_ctlModes(SetRegFuncParams)
{
  return setCtlModes(MB_VAL(U16));
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_stepRampFactor(SetRegFuncParams)
{
  float  fVal = MB_VAL(float);

  if (fVal < 0.1)  fVal = 0.1;
  if (fVal > 1.0)  fVal = 1.0;

  stepRampFactor = fVal;

  return true;
}





//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

//-----------------------------------------------------------------------------
//  restore a ctlModes value from the devdata file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ctlModes)
{
  setCtlModes(NEW_VAL(U16));
}


//-----------------------------------------------------------------------------
//  restore an ampSet value from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ampSet)
{
  newSetpoint(NEW_VAL(float), true);
}


/*
//-----------------------------------------------------------------------------
//  one of the setpoint values needs restoring from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(myVal)
{
  uint  u;
  float  fval;
  char  temp[32];

  for (u=0; u<NumPS; ++u)  if (curVal == (void far *) &ISET(u))  {
    fval = *(float far *)newVal;
    printf("restoring iSet_%u to: %s\r\n", u, fmtFloat(fval, 8,2, temp));
    newSetpoint(fval, true);
  }

}

*/

//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>


//=============================================================================
//  Misc Application-specific functions
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Return a # representing the direction of change of a value:
//     1 if the # is > 0
//     0 if the # is = 0
//    -1 if the # is < 0
//-----------------------------------------------------------------------------
int chgDir(float val)
{
  if (val > 0.0)  return 1;
  if (val < 0.0)  return -1;
  return 0;
}


//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================
//APPLCODE - text and formatting info for application state summary

//--- set each time through showState() loop ---
char serValPrefix[8];
char setReadPrefix[3][8];

//-----------------------------------------------------------------------------
//  Return the string "1" or "0" based on whether a bit is set or not.  The
//  elem->width value is a bit-mask that specifies which bit to test.
//-----------------------------------------------------------------------------
_nearcall char *fmtBinary(DispElem far *elem, char *buf)
{
  uint  val;
#define  BIT_MASK  (elem->width)

  val = ELEM_VAL(uint);
  buf[0] = (val & BIT_MASK) ? '1' : '0';
  buf[1] = 0;

  return buf;
}

//-----------------------------------------------------------------------------
//  Return a string (including a color prefix) indicating the logical state of
//  the controller.
//-----------------------------------------------------------------------------
_nearcall char *fmtState(DispElem far *elem, char *buf)
{
  switch (Cur.state)  {
    case STATE_OFF:    return   RedText"OFF  ";
    case STATE_ON:     return  BlueText"ON   ";
    case STATE_INTLK:  return   RedText"INTLK";
  }

  return RedText"?????";
}

//-----------------------------------------------------------------------------
//  Return a string (including a color prefix) indicating the state of the
//  connection to the load.
//-----------------------------------------------------------------------------
_nearcall char *fmtPolarity(DispElem far *elem, char *buf)
{
  switch (Cur.assertPolarity)  {
    case -1:  return  BlueText"REVERS";
    case  0:  return   RedText"DISCON";
    case  1:  return  BlueText"NORMAL";
  }

  return RedText"????";
}

//-----------------------------------------------------------------------------
//  Return a blank-padded string elem->width chars long
//-----------------------------------------------------------------------------
_nearcall char *fmtStr(DispElem far *elem, char *buf)
{
  uint  len, fill;
  char  *ptr = buf;


  len = (uint) fstrlen(elem->val);
  if (len > elem->width)  len = elem->width;

  fmemcpy(ptr, elem->val, len);

  fill = elem->width - len;  ptr += len;
  while (fill)  { *ptr++ = ' ';  --fill; }

  *ptr = 0;

  return buf;
}

//-----------------------------------------------------------------------------
//  Return string indicating where a the value is actually applied to or
//  read from.
//
//  The width attribute for the DispElem provides the # of bits to shift the
//  ctlModes value to get the desired nibble that tells us this.
//-----------------------------------------------------------------------------
_nearcall char *fmtMode(DispElem far *elem, char *buf)
{
  char  *ptr;
  uint shiftBy = elem->width, mode;


  mode = (ctlModes >> shiftBy) & 7;

  if (mode & CM_Serial)      ptr = "Ser";  else
  if (mode & CM_Analog_Int)  ptr = "Int";  else
  if (mode & CM_Analog_Ext)  ptr = "Ext";  else
                             ptr = "ERR";

  return ptr;
}


//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {

"---Digital In----          --- PS Info via Serial Interface --- \r\n"
"ExtInterlock: ###          PON Serv Req:  #    Const Voltage: # \r\n"
"PS has power: ###          Serv Req sent: #    Const Current: # \r\n"
"                           EEPROM Locked: #    Over Voltage:  # \r\n"
"---Digital Out--           OV/OT tripped: #    Over Temp:     # \r\n"
"Normal LED:   ###          Trigger Wait:  #    Ext. Shutdown: # \r\n"
"Reverse LED:  ###          Foldback Prot: #    Foldback Mode: # \r\n"
"Assert PS ON: ###          Serv Req Enab: #    Cur Set:####.### \r\n"
"Normal  Pol:  ###          Output is ON:  #    Volt Set:###.### \r\n"
"Reverse Pol:  ###          ID:[#############################################] \r\n"
" \r\n"
"---- ADC Inputs ----      Readings   --Ser--   --Int--   --Ext-- \r\n"
"Int Cur Read:-##.###       Current: -###.###  -###.###  -###.### \r\n"
"Int Vlt Read:-##.###       Voltage: -###.###  -###.###  -###.### \r\n"
"Ext Vlt Read:-##.###  \r\n"
"Ext Cur Read:-##.###      Setpoint: -###.### \r\n"
"                           Applied: -###.### A [###]  [#####][######] \r\n"
"---- DAC Output ----      Cur Read: -###.### A [###] \r\n"
"Cur Setpoint:-##.###      Vlt Read: -###.### V [###] \r\n"
" \r\n"
  };

const DispElem  far dispElems[] = {
  { 13,15, VFPV(anaIn[INT_AMPS_READ]),      BLUE, fmtF32,    7,3 },
  { 13,16, VFPV(anaIn[INT_VOLT_READ]),      BLUE, fmtF32,    7,3 },
  { 13,17, VFPV(anaIn[EXT_VOLT_READ]),      BLUE, fmtF32,    7,3 },
  { 13,18, VFPV(anaIn[EXT_AMPS_READ]),      BLUE, fmtF32,    7,3 },

  { 13,21, VFPV(anaOut[OUTPUT_CTL]),        BLUE, fmtF32,    7,3 },

  { 42, 4, VFPV(serStatusFlags),    serValPrefix, fmtBinary, PON_SRQ,1 },
  { 42, 5, VFPV(serStatusFlags),    serValPrefix, fmtBinary, SRQ_SENT,1 },
  { 42, 6, VFPV(serStatusFlags),    serValPrefix, fmtBinary, EEPROM_LCKD,1 },
  { 42, 7, VFPV(serStatusFlags),    serValPrefix, fmtBinary, TRIP_OCCURRED,1 },
  { 42, 8, VFPV(serStatusFlags),    serValPrefix, fmtBinary, WAIT_TRIG,1 },
  { 42, 9, VFPV(serStatusFlags),    serValPrefix, fmtBinary, FOLDBACK_ENA,1 },
  { 42,10, VFPV(serStatusFlags),    serValPrefix, fmtBinary, SRQ_ENA,1 },
  { 42,11, VFPV(serStatusFlags),    serValPrefix, fmtBinary, OUTPUT_ON,1 },

  { 31,12, VFP(serIDStr),           serValPrefix, fmtStr,    45,1 },

  { 62, 4, VFPV(serProtCond),       serValPrefix, fmtBinary, CONST_VOLT,1 },
  { 62, 5, VFPV(serProtCond),       serValPrefix, fmtBinary, CONST_CUR,1 },
  { 62, 6, VFPV(serProtCond),       serValPrefix, fmtBinary, OVER_VOLT,1 },
  { 62, 7, VFPV(serProtCond),       serValPrefix, fmtBinary, OVER_TEMP,1 },
  { 62, 8, VFPV(serProtCond),       serValPrefix, fmtBinary, EXT_SHUTDOWN,1 },
  { 62, 9, VFPV(serProtCond),       serValPrefix, fmtBinary, FOLDBACK_MODE,1 },

  { 55,10, VFPV(serSetA),           serValPrefix, fmtF32,    8,3 },
  { 56,11, VFPV(serSetV),           serValPrefix, fmtF32,    7,3 },

  { 36,15, VFPV(serReadA),          serValPrefix, fmtF32,    8,3 },
  { 36,16, VFPV(serReadV),          serValPrefix, fmtF32,    8,3 },

  { 46,15, VFPV(intAmps),                   BLUE, fmtF32,    8,3 },
  { 46,16, VFPV(intVolts),                  BLUE, fmtF32,    8,3 },

  { 56,15, VFPV(extAmps),                   BLUE, fmtF32,    8,3 },
  { 56,16, VFPV(extVolts),                  BLUE, fmtF32,    8,3 },

  { 36,18, VFPV(ampSet),                    BLUE, fmtF32,    8,3 },
  { 36,19, VFPV(activeSetAmps), setReadPrefix[0], fmtF32,    8,3 },
  { 36,20, VFPV(ampRead),       setReadPrefix[1], fmtF32,    8,3 },
  { 36,21, VFPV(voltRead),      setReadPrefix[2], fmtF32,    8,3 },

  { 48,19, VFPV(ctlModes),      setReadPrefix[0], fmtMode,   8,0 },
  { 48,20, VFPV(ctlModes),      setReadPrefix[1], fmtMode,   4,0 },
  { 48,21, VFPV(ctlModes),      setReadPrefix[2], fmtMode,   0,0 },


  { 55,19, VFPV(Cur.state),                   "", fmtState,    5,5 },
  { 62,19, VFPV(Cur.assertPolarity),          "", fmtPolarity, 6,6 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};

//--- ATTR_NAME macros currently defined in fcntl.h ---
const DigDispElem far digInElems[] = {
//  xx,yy,  BIT_MASK,         "0-STR",ATTR_NAME,   "1-STR",ATTR_NAME

  { 14, 4,  EXT_INTERLOCK,         "NO ",BLUE,     "YES",RED   },
  { 14, 5,  PS_POWER_OFF,          "YES",BLUE,     "NO ",RED   },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
//  xx,yy,  BIT_MASK,          "0-STR",ATTR_NAME, "1-STR",ATTR_NAME

  { 14, 8,  NORM_POLARITY_LED,     "ON ",BLUE,   "OFF",RED },
  { 14, 9,  REV_POLARITY_LED,      "ON ",BLUE,   "OFF",RED },
  { 14,10,  TURN_PS_OFF,           "YES",BLUE,   "NO ",RED },
  { 14,11,  NORM_POLARITY_OFF,     "YES",BLUE,   "NO ",RED },
  { 14,12,  REV_POLARITY_OFF,      "YES",BLUE,   "NO ",RED },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};

//---------------------------------------------------------------
//---- older style - do not use for newer applications -----
const DigFmtInfo digInFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
//{ xx,yy, "0-str"  ,1,  "1-str" ,0 },  //  descr
  { -1,-1, "",0, "",0 }
};


const AnalogFmtInfo  adcFmt[] = {
  { -1,-1, 0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};


//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff, invalid;
  uint  u, shiftBy, cfgMode;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;


  //--- set color prefixes used for several values ---
    strcpy(serValPrefix, (Cur.serLinkActive ? BLUE : RED));

    for (u=0,shiftBy=8; u<3; ++u,shiftBy-=4)  {
      cfgMode = (ctlModes >> shiftBy) & 7;
      invalid = (cfgMode & CM_Serial) AND !Cur.serLinkActive;
    //--- Use Red text if not recent data ---
      strcpy(setReadPrefix[u], (invalid ? RED : BLUE));
    }


    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}


//-----------------------------------------------------------------------------
//  Application-specific commands
//-----------------------------------------------------------------------------
#pragma warn -775

void _task cmd(int argc, char **argv)
{
  bool  validCmd;
  char  *cmdStr;
  uint  u;


#define SetTrue(val)  { \
  _EnterCritical_();  val = true;  _ExitCritical_();  \
  validCmd = true;  break; \
}
#define SetVal(val, str)  { \
  KLOCK();  val = atof(str);  KUNLOCK();  \
  validCmd = true;  break; \
}


  validCmd = false;

  do  {
    if (argc < 2)  break;

    cmdStr = argv[1];

    if (StriEq(cmdStr, "on"))  { onRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "off"))  { offRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "reset"))  { resetRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "set") AND (argc > 2))  {
      newSetpoint(atof(argv[2]), true);  validCmd = true;  break; }

    if (StriEq(cmdStr, "send"))  if (argc > 2)  {
      if (!serLinkRequired) {
        printf("Serial Link disabled during analog ctl\r\n");  Exit(-1); }
      KLOCK();
      *psResp = 0;  strncpy(psCmd, argv[2], sizeof(psCmd)-1);
      KUNLOCK();
      for (u=0; u<20; ++u)  {   // wait up to 2 secs for resp
        if (*psResp)  { printf("[%.256s]\r\n", psResp);  break; }
        sleepMS(100);
      }
      if (argv[2][strlen(argv[2])-1] == '?')  // only query cmds respond
        if (!*psResp)  printf("  <no response>\r\n");
      validCmd = true;
    }

  }  while(0);


  if (! validCmd)  {
    printf("\r\n"
           "Usage:  cmd on|off|reset|set #|send str\r\n"
          );
    Exit(-1);
  }

  Exit(0);
}

#pragma warn 775


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================
//APPLCODE

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  anaOut[idx] = ISETtoVolts(idx, setpoint);
}
*/



//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "cur" OR the "next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "cur" value for the "input", and the "prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "cur" values
//    represent the input values just read (what they are "now") and the
//    "prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "cur" values).
//
//    During a pass the output values in "cur" represent what the outputs
//    where set to AFTER the "cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

//-----------------------------------------------------------------------------
void onRequest(void)
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON request\n");
}

//-----------------------------------------------------------------------------
void offRequest(void)
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF request\n");
}

//-----------------------------------------------------------------------------
void resetRequest(void)
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET request\n");
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps, bool logIt)
{
  char  temp[16];


  KLOCK();

  if (fabs(newAmps) > maxAmps)  newAmps = (float)SIGN(newAmps) * maxAmps;

  Next.ampSet = newAmps;

  KUNLOCK();

  if (logIt)
    fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}

//-----------------------------------------------------------------------------
//  Set the methods used to control the setpoint and read the current and
//  voltage values.
//
//    The ctlModes value consists of 3 groups of bits that define how to apply
//    the setpoint and read the current and voltage.
//
//    NOTES:
//      - Only 1 of the analog bits can be set in each group
//      - If both the serial bit and 1 of the analog bits is set, the
//        specified analog set or read will be used whenever the serial
//        link is not active.
//      - "Int" stands for "Internal" (meaning the reading is from the
//        internal shunt or across the output terminals of the ctlr case)
//      - "Ext" stands for "External" (meaning the reading is from an
//        external analog signal)
//-----------------------------------------------------------------------------
bool setCtlModes(U16 newCtlModes)
{
  newCtlModes &= CM_Valid_Bits;  // eliminate extraneous bits

//--- Any group with no bits set defaults to External Analog mode ---
  if (!(newCtlModes & CM_SetA_Bits))   newCtlModes |= CM_Ext_SetA;
  if (!(newCtlModes & CM_ReadA_Bits))  newCtlModes |= CM_Ext_ReadA;
  if (!(newCtlModes & CM_ReadV_Bits))  newCtlModes |= CM_Ext_ReadV;

//--- If Int and Ext analog bits set, clear the Ext bit ---
  if ((newCtlModes & 0x0030) == 0x0030)  newCtlModes &= ~CM_Ext_ReadA;
  if ((newCtlModes & 0x0003) == 0x0003)  newCtlModes &= ~CM_Ext_ReadV;

//--- The analog setpoint is ignored whenever the serial link is in use, ---
//--- so use of the serial mode for any of the signals means that the ---
//--- setpoint MUST use the serial link ---
  if (newCtlModes & CM_Serial_Bits)  newCtlModes |= CM_Ser_SetA;

//--- If analog and serial bits set for any of the signals, clear the ---
//--- analog mode bits ---
  if ((newCtlModes & CM_Ser_SetA) AND (newCtlModes & CM_Ext_SetA))
    newCtlModes = newCtlModes & ~CM_Ext_SetA;

  if ((newCtlModes & CM_Ser_ReadA) AND (newCtlModes & CM_Ana_ReadA))
    newCtlModes = newCtlModes & ~CM_Ana_ReadA;

  if ((newCtlModes & CM_Ser_ReadV) AND (newCtlModes & CM_Ana_ReadV))
    newCtlModes = newCtlModes & ~CM_Ana_ReadV;


  KLOCK();

  ctlModes = newCtlModes;

  serLinkRequired = BOOL(ctlModes & CM_Serial_Bits);

  KUNLOCK();

  return true;
}

//-----------------------------------------------------------------------------



int  bitCounts[32];

//-----------------------------------------------------------------------------
//  Minimize the effect of noise on the digitial inputs in trade for some
//  added latency in response time to actual "legitimate" changes in state.
//
//  What this func does is add or subtract the periodic timer interval value
//  to/from a running total for each input bit, with a cap on the total being
//  the configured # of ms for debouncing.
//
//  The current state of each input is determined by whether the running total
//  for each input is positive or negative.  A positive value indicates a
//  predominately ON state, while a negative value indicates a predominiately
//  OFF state.
//
//  Inputs that are mostly stable (anomolies not very frequent) will quickly
//  reach the positive or negative limit and stay close to it until the "real"
//  state changes to the opposite state.  When that happens, the total will
//  shrink, cross 0, and grow toward the opposite limit.
//
//  NOTE the latency effect of this approach on the two extremes:
//    - No noise:  No errors, but Max latency (== the debounce period)
//    - Max noise:  Min latency, but constant errors.
//
//  The setting for the debounce period effectively allows you to ajust the
//  inherent compromise between delays introduced for the correct readings vs
//  the probability of getting a false reading.  A setting of 0 effectively
//  disables the filter, while larger values reduce the effect of any noise.
//
//  This approach behaves aproximately like connecting a capacitor across a
//  noisy input.  As long as the noise level is low enough, the capacitor
//  will tend to charge up to the level of the ON voltage, or dischage to
//  0.  Noise will increase or decrease the voltage, but so long as it is
//  less frequent than the "real" value, the voltage across the capacitor
//  will tend to stay near the voltage level for the "real" state.
//-----------------------------------------------------------------------------
void debounceDigIn(void)
{
  int  i;
  int  *bitCount;
  ulong  mask, digIn;


  bitCount = bitCounts;  digIn = 0;  mask = 1;

  for (i=0; i<32;  ++i)  {
    if (! (mask & DIN_MASK_APPL))  {
      digIn |= digitalIn & mask;  mask <<= 1;  continue; }

    if (digitalIn & mask)
      *bitCount += (int)SchedMS;
    else
      *bitCount -= (int)SchedMS;

    if (*bitCount >  (int)debounceMS)    *bitCount =  (int)debounceMS;
    if (*bitCount < (int)(-debounceMS))  *bitCount = -(int)debounceMS;

    if (*bitCount > 0)  digIn |= mask;

    ++bitCount;  mask <<= 1;
  }

  digitalIn = digIn;
}

//-----------------------------------------------------------------------------
//  Return string descr of the state of the polarity switch
//-----------------------------------------------------------------------------
char *polSwState(void)
{
  switch (Cur.assertPolarity)  {
    case -1:  return "REVERS";
    case  0:  return "DISCON";
    case  1:  return "NORMAL";
  }

  return "<err> ";
}


//=============================================================================
//  Copy the "cur" state from last time to "prev", and update "cur"
//=============================================================================
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).
//
//  This func first copies "cur" to "prev" and then "next" to "cur" which does
//  2 things for us:
//
//    - We have a copy of the controller's logical state from the last pass
//      (so we can determine what has changed)
//
//    - We have a copy of values set by other tasks indicating events/requests
//      that happend/were made since the beginning of the last pass
//
//  We then reset the values (as needed) that are set by other tasks to
//  indicate events and requests, and update the current state based on the
//  readings just taken and the requests and events communicated by other
//  tasks.
//
//  All of this is in preperation for the updateNextState() function, which
//  implements the main "decision making" portion of the application logic,
//  determining what the next state should be based on the latest conditions
//  (as determined by the updateCurState() function).
//
//=============================================================================
void updateCurState(void)
{
  bool  logState;
  U16  minLogSecs;
  char  temp[64];


  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "Next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//   since the last pass.
//------------------------------------------------------------------


//  Next.processTime = GetMS();

//  msSincePrev = msSince(Cur.processTime);

// Now we update the values in "Next" that are based solely on the
// state of the hardware inputs we just read and values maintained by
// by other tasks (basically, we deal with READINGS here: Values that
// we need but are not written to "Next" by another task)

  debounceDigIn();  Next.digitalIn = digitalIn;

  if (rampRate <= 0.001)  rampRate = 0.001;

//APPLCODE
  Next.externIntlk  = BOOL((U16)digitalIn & EXT_INTERLOCK );
  Next.psPowerEnabled = !((U16)digitalIn & PS_POWER_OFF);

  intAmps  = readVal[INT_AMPS_READ];
  intVolts = readVal[INT_VOLT_READ];

  extVolts = readVal[EXT_VOLT_READ];
  extAmps  = readVal[EXT_AMPS_READ];

  Next.ampSetPolarity = SIGN(Next.ampSet);

//--- Keep track of when we last changed the state of the polarity switches ---
  if (Next.assertPolarity ^ Cur.assertPolarity)  lastPolarityChgTime = GetMS();
  polarityStable = (msSince(lastPolarityChgTime) >= switchTime);

/*
  Consider Serial link active if:
    - We got a valid looking response within the last 2 seconds, AND
    - The link has been active or we got enough good responses in a row that
      all the values read from the supply are recent
  This stops the active flag from being set after we first turn it on until
  we have updated all the values and also stops it from being cleared when
  we have a momentary communications problem.
*/
  Next.serLinkActive = (msSince(psLastRespTime) < 2000)
               AND ((consecGoodResp >= NUM_PS_CMDS * 2) OR Next.serLinkActive);

//--- Keep track of when power to the PS was enabled ---
  if (!Next.psPowerEnabled)
    psPowerEnabledSecs = 0;
  else
    if (!psPowerEnabledSecs)  psPowerEnabledSecs = GetSeconds_();


//--- If the feedback indicates the internal relays enabled power to the ---
//--- supply, and the serial link is needed, then the serial link should be ---
//--- active shortly after the supply was turned on. ---
  if (!Cur.psPowerEnabled OR !serLinkRequired)
    serLinkNotWorking = false;
  else
    if (SecSince_(psPowerEnabledSecs) > serInitTime)
      serLinkNotWorking = !Next.serLinkActive;


//--- Is the PS output enabled? ---
  if (Next.serLinkActive)
    Next.psOutputOn = BOOL(Cur.serStatusFlags & OUTPUT_ON);
  else
    Next.psOutputOn = (serLinkRequired ? false : Next.psPowerEnabled);

//--- Keep track of when the PS output was enabled ---
  if (!Next.psOutputOn)
    psOutputOnTime = 0;
  else
    if (!psOutputOnTime)  psOutputOnTime = GetMS();


  // This is done in the applic rather than the psCom task because the psCom
  // task does not update it frequently enough, with the result that we would
  // keep projecting the readback based on a RAMP command that had completed.
  if (ramping)  {
    lastRamping = GetMS_();
    if (msSince(rampStartedTime) > rampPeriod)  {
      projectFrom += projectBy;  projectBy = 0.0;
      projectFromTime = 0;  ramping = false;
    }
  }

/*
  ctlModes is a bit-mask that determines what method is used to apply a                                                                  ---
  setpoint and where readings come from.  NOTE:  The serial and external                                                              ---
  readings come from the supply and not from the load or the the connection
  to the load, so if the load is disconnected (assertPolarity == 0), values
  from those sources do not represent the state of the load.
*/
  if (ctlModes & CM_Ser_SetA)
    activeSetAmps = serSetA;
  else
    activeSetAmps = fabs(ampApply);

  // For a serial readback, we compute what the output of the supply is between
  // actual samples while a RAMP command is in effect.
  // NOTES:
  //  - projectFrom, projectBy, and projectFromTime are all updated when we get
  //    a new reading or a send a new ramp or setting command
  //  - serRampTo is updated when we send a new ramp command
  if (ctlModes & CM_Ser_ReadA)  do {
    if (!Cur.assertPolarity)  {
      ampRead = projectFrom = projectBy = 0.0;  projectFromTime = 0;
      ramping = false;  break;
    }
    if (projectFromTime)  {
      projectBy = (float)(msSince(projectFromTime)) * projectRate / 1000.0;
      if (projectBy > maxProjectBy)  projectBy = maxProjectBy;
      if (serRampTo < projectFrom)  projectBy = -projectBy;
    }
    ampRead = projectFrom + projectBy;
  } while(0);
  else  if (ctlModes & CM_Int_ReadA)
    ampRead = intAmps;
  else  if (ctlModes & CM_Ext_ReadA)
    ampRead = Cur.assertPolarity ? extAmps : 0.0;


  if (ctlModes & CM_Ser_ReadV)
    voltRead = Cur.assertPolarity ? serReadV : 0.0;
  else  if (ctlModes & CM_Int_ReadV)
    voltRead = intVolts;
  else  if (ctlModes & CM_Ext_ReadA)
    voltRead = Cur.assertPolarity ? extVolts : 0.0;


//--- Change the polarity of the readings if needed ---
  if (Cur.assertPolarity)  rdgPolarity = Cur.assertPolarity;
  if (ctlModes & (CM_Ser_ReadA | CM_Ext_ReadA))
    if (rdgPolarity < 0)  ampRead = -ampRead;
  if (ctlModes & (CM_Ser_ReadV | CM_Ext_ReadV))
    if (rdgPolarity < 0)  voltRead = -voltRead;



//--- If the rampRate setting changed, update the max amount we should ---
//--- change the active setpoint each pass ---
  if (rampRate != lastRampRate)  {
    maxAmpChg = rampRate * (float)SchedMS / 1000.0;  lastRampRate = rampRate; }

//--- clear the offDueToReset flag for ON or RESET requests ---
  if (Next.onRequest OR Next.resetRequest)  offDueToReset = false;


// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer.

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---


//--- log msgs for digital I/O bit chgs (but only if not starving other ---
//--- processes for cycles) ---
  /* if (applLoopTime < 8) */  logDigIOChgs(false);


//--- log msgs for various changes ---
  if (Cur.serLinkActive != Prev.serLinkActive)  {
    printf("PS is %s responding via serial link\n",
           Cur.serLinkActive ? "NOW" : "NOT");
    if (Cur.serLinkActive)  logIDStr = true;
  }

  if (Cur.assertOn != Prev.assertOn)  assertOnChgd = GetSeconds_();

  if ((Cur.psPowerEnabled != Cur.assertOn) AND (SecSince_(assertOnChgd) >= 5))
    if (SecSince_(lastStateErrMsg) > 300)  {
      printf("*** PS ON input != PS ON output ***\n");
      lastStateErrMsg = GetSeconds_();
    }


  if (Cur.serLinkActive)  {

    if ((Cur.serStatusFlags ^ Prev.serStatusFlags) & SF_LOG_CHGS)
      printf("  PON Serv Req:  %d\r\n"
             "  Serv Req sent: %d\r\n"
             "  EEPROM Locked: %d\r\n"
             "  OV/OT tripped: %d\r\n"
             "  Trigger Wait:  %d\r\n"
             "  Foldback Prot: %d\r\n"
             "  Serv Req Enab: %d\r\n"
             "  Output is ON:  %d\r\n",
               BOOL(Cur.serStatusFlags & PON_SRQ),
               BOOL(Cur.serStatusFlags & SRQ_SENT),
               BOOL(Cur.serStatusFlags & EEPROM_LCKD),
               BOOL(Cur.serStatusFlags & TRIP_OCCURRED),
               BOOL(Cur.serStatusFlags & WAIT_TRIG),
               BOOL(Cur.serStatusFlags & FOLDBACK_ENA),
               BOOL(Cur.serStatusFlags & SRQ_ENA),
               BOOL(Cur.serStatusFlags & OUTPUT_ON)
            );

    if (Cur.serProtCond ^ Prev.serProtCond)
      printf("  Const Voltage: %d\r\n"
             "  Const Current: %d\r\n"
             "  Over Voltage:  %d\r\n"
             "  Over Temp:     %d\r\n"
             "  Ext. Shutdown: %d\r\n"
             "  Foldback Mode: %d\r\n"
             "  Cur:  Set:%s   Read: %s\r\n"
             "  Volt: Set:%s   Read: %s\r\n",
               BOOL(Cur.serProtCond & CONST_VOLT),
               BOOL(Cur.serProtCond & CONST_CUR),
               BOOL(Cur.serProtCond & OVER_VOLT),
               BOOL(Cur.serProtCond & OVER_TEMP),
               BOOL(Cur.serProtCond & EXT_SHUTDOWN),
               BOOL(Cur.serProtCond & FOLDBACK_MODE),
               fmtFloat(serSetA,  8,3, temp),
               fmtFloat(serReadA, 8,3, temp+10),
               fmtFloat(serSetV,  8,3, temp+20),
               fmtFloat(serReadV, 8,3, temp+30)
            );


  }  // Cur.serLinkActive

  if (offDueToReset AND !loggedResetIntlk)  {
    printf("*** Supply is OFF due to ctlr RESET ***\n");
    loggedResetIntlk = true;
  }



//--- Periodically log basic state information ---
  logState = false;

  minLogSecs = ((Cur.state == STATE_ON) ? logStateSecs : 1800);

  if ((U16)upSecs - lastLogStateTime > minLogSecs)  logState = true;
  if (Cur.state ^ Prev.state)  logState = true;
  if (Cur.assertPolarity ^ Prev.assertPolarity)  logState = true;

  if (logState)  {
    printf("Setpoint: %s A    State: %s\n"
           "  Active: %s A   Switch: %s\n",
            fmtFloat(ampSet,         8,3, temp),     StateDesc[Cur.state],
            fmtFloat(activeSetAmps,  8,3, temp+10),  polSwState()
          );
    if (Cur.serLinkActive OR !serLinkRequired)
      printf("    Read: %s A %s V\n",
              fmtFloat(ampRead,        8,3, temp+20),
              fmtFloat(voltRead,       8,3, temp+30)
            );
    else
      if (serLinkNotWorking)  printf("  *** Serial Link not active ***\n");

    lastLogStateTime = (U16)upSecs;
  }

}



//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of output values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  float  actualRead;


  if (current != applTask)  return;  // must only be called by "applic" task

//--------------------------------------------------------------------
// Next.assertReset, Next.resetStarted
// Next.resetRequest (cleared)
//    Description:
//       Should we be asserting a reset pulse for the error inputs?
//    Inputs:
//       Cur.resetRequest
//       Cur.resetStarted
//--------------------------------------------------------------------
  if (Cur.resetRequest)  {
    Next.resetRequest = false;
    Next.resetStarted = GetMS();  Next.assertReset = true;
  }
  if (Cur.resetStarted AND (msSince(Cur.resetStarted) >= 500))  {
    Next.resetStarted = 0;  Next.assertReset = false;  }


//--------------------------------------------------------------------
// Next.state
// Next.onRequest, Next.offRequest (cleared)
//    Description:
//       Determines the next state of for the application logic
//       (STATE_ON, STATE_OFF, STATE_INTLK)
//    Inputs:
//       Cur.onRequest
//       Cur.offRequest
//       Cur.state
//       Prev.assertReset
//       Cur.assertReset
//       Cur.externIntlk
//       offDueToReset
//----------------------------------------------------------------------
//--- chg state from OFF to ON? ---
  if (Cur.onRequest)  do  {
    Next.onRequest = false;
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_ON)  { printf("Already ON\n\n");  break; }
    Next.state = STATE_ON;  printf("Turning ON\n\n");
  } while(0);

//--- chg state from ON to OFF? ---
  if (Cur.offRequest)  do {
    Next.offRequest = false;
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_OFF)  { printf("Already OFF\n\n");  break; }
    Next.state = STATE_OFF;  printf("Turning OFF\n\n");
  } while(0);

//--- chg state from INTLK to OFF? ---
  if (Prev.assertReset AND !Cur.assertReset)  do {  // end of a reset pulse?
    if (Cur.state != STATE_INTLK)  {
      printf("No Interlock to clear\n\n");  break;  }
    if (Cur.externIntlk)  {
      printf("External Interlock still asserted\n\n");  break; }
    Next.state = STATE_OFF;
  } while(0);

//--- chg state to INTLK? ---
  if (Cur.externIntlk OR offDueToReset)  Next.state = STATE_INTLK;


//-------------------------------------------------------------------
// Next.assertOn
//    Description:
//       Should the ON bit (which enables power to the PS) be asserted?
//    Inputs:
//       Cur.state
//-------------------------------------------------------------------
  Next.assertOn = (Cur.state == STATE_ON);


//--------------------------------------------------------------------
// Next.assertPolarity
//    Description:
//       Determines the next state for the polarity switch
//    Inputs:
//       Cur.assertOn
//       serReadA
//       activeSetAmps
//       ampRead
//       safeSwAmps
//       Cur.psOutputOn
//       psOutputOnTime
//       connectLoadDelay
//       Cur.assertPolarity
//       Cur.ampSetPolarity
//----------------------------------------------------------------------
  if (!Cur.assertOn)
    Next.assertPolarity = 0;  // 0 means disconnect the load
  else
//NOTE:  If we are asserting ON, then we leave the current state of the
//       connection to the load alone unless it NEEDS to be changed.  This
//       way, if there is a momentary glitch in our readings, we won't react
//       by disconnecting the load.
    if (Cur.assertPolarity != Cur.ampSetPolarity)  {
      actualRead = (ctlModes & CM_Ser_ReadA) ? serReadA : ampRead;
      if ( (activeSetAmps == 0.0)
          AND (fabs(actualRead) <= safeSwAmps)
          AND Cur.psOutputOn
          AND (msSince(psOutputOnTime) > (ulong)connectLoadDelay) )
        Next.assertPolarity = (Cur.assertPolarity ? 0 : Cur.ampSetPolarity);
    }



//-------------------------------------------------------------------
// Next.ampApply
// freezeAmpApply
//    Description:
//       The setpoint that we want applied to the power supply. NOTE
//       that this value must be 0 while a change in the polarity is
//       in progress.
//    Inputs:
//       Cur.ampSet - desired setpoint when supply is ON
//       Cur.ampApply - cur asserted setpoint for power supply
//       Prev.ampApply - prev asserted setpoint power supply
//       Cur.psPowerEnabled - power to the PS is enabled
//       Cur.assertOn - have we told the PS to turn on
//       Cur.assertPolarity - state of polarity switches
//       maxAmpChg - max change in ampApply each cycle
//       ctlModes - specifies analog or serial mode for set and reads
//-------------------------------------------------------------------
  {
    float  chg;
    static int  lastChgDir = 0;


  //--- If the PS is (or should be) OFF, the load is not connected, or the ---
  //--- asserted polarity switch state may not be real or stable yet, ---
  //--- then apply a 0 setpoint ---
    if  (!Cur.psPowerEnabled OR !Cur.assertOn
         OR !Cur.assertPolarity OR !polarityStable)
      Next.ampApply = 0.0;
  //--- Otherwise, adjust ampApply in the direction of the target setpoint ---
    else  {
      chg = Cur.ampSet - Cur.ampApply;
      if (fabs(chg) > maxAmpChg)  chg = (chg > 0.0) ? maxAmpChg : -maxAmpChg;
      Next.ampApply = Cur.ampApply + chg;

    //--- A change that results in a new setpoint whose polarity is NOT the ---
    //--- same as the current switch polarity is not allowed.  This forces ---
    //--- the applied setpoint to stop at 0 until the polarity of the ---
    //--- switch matches the direction we are trying to go. ---
      if (Cur.assertPolarity != SIGN(Next.ampApply))  Next.ampApply = 0.0;
    }
  }


}  // end updateNextState()



//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
  ulong  polarityBits;


  if (current != applTask)  return;  // must only be called by "applic" task

  Next.digitalOut = 0xffffffffL;


//--- update output values that are based on current conditions (values ---
//--- in "cur" or any values used in a conditional in the applic logic) ---

/*EXAMPLE:
  // (insured earlier that only ONE of these can be true when setting them)
  if (initialBtnPressed)  setVal[FP_METER] = rfOffSetpoint;  else
  if (finalBtnPressed)    setVal[FP_METER] = rfOnSetpoint;  else
                          setVal[FP_METER] = cavityVolts;


  SetOffApplLEDs(Next.digitalOut);

  if (limited)  SetOnLimitLED(Next.digitalOut);
  if (regErr)   SetOnErrorLED(Next.digitalOut);
*/


//--- update output values based on values that were set by the application ---
//--- logic (e.g. values in "Next" or variables that are the target of an ---
//--- assignment statement in the application logic) ---

//APPLCODE

//  if (Next.assertReset)   ClearBits(Next.digitalOut, ALLOW_INTLKS);


  if (Next.assertOn)  {
    ClearBits(Next.digitalOut, TURN_PS_OFF);

    if (Next.assertPolarity < 0)
      polarityBits = REV_POLARITY_OFF | REV_POLARITY_LED;
    else  if (Next.assertPolarity > 0)
      polarityBits = NORM_POLARITY_OFF | NORM_POLARITY_LED;
    else
      polarityBits = 0;

  //--- If serial link is not working when it should be, then blink one or ---
  //--- both the polarity LEDs briefly to indicate the problem ---
    if (serLinkNotWorking)
      if ((GetMS_() & 0x7F) < 20)  // for aprox 20 out of every 128 ms ...
        if (polarityBits)
          polarityBits ^= (polarityBits & POLARITY_LEDS);
        else
          polarityBits = POLARITY_LEDS;

    if (polarityBits)  ClearBits(Next.digitalOut, polarityBits);
  }



  setVal[OUTPUT_CTL] = fabs(Next.ampApply);
}




//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  uint  onStatesBitMap = 0;


  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//APPLCODE
  ampSet = Cur.ampSet;

  devStatus &= DS_CTLR_RESET;  // clear all but CTLR_RESET bit

  if (Cur.externIntlk)             SetBits(devStatus, DS_INTLK);
  if (Cur.state == STATE_INTLK)    SetBits(devStatus, DS_INTLK);
  if (Cur.state == STATE_ON)       SetBits(devStatus, DS_ON);
  if (Cur.assertPolarity >= 0)     SetBits(devStatus, DS_NORM_POL);
  if (Cur.psPowerEnabled AND Cur.assertOn)  ClearBits(devStatus, DS_CTLR_RESET);
  if (serLinkNotWorking)           SetBits(devStatus, DS_SER_ERR);
//  if (calibErr)                    SetBits(devStatus, DS_CALIB_ERR);

  serProtCond = Cur.serProtCond;
  serStatusFlags = Cur.serStatusFlags;

  serSetA = Cur.serSetA;
  serSetV = Cur.serSetV;

  ampApply = Cur.ampApply;

  if (Cur.state == STATE_ON)  onStatesBitMap = 1;

  diagLogInfo->onStatesBitMap = onStatesBitMap;
}




//-----------------------------------------------------------------------------
//  Parse the response from a Sorenson SGA Power Supply
//-----------------------------------------------------------------------------
int Process_SGA_PS_Resp(char *resp, const PS_CmdInfo *cmdInfo)
{
  uint  numFlds, toType, maxSize;
  U16  u16Val;
  float  fVal;
  char  *ptr;
  char  *strs[24];

#define  ToVal(typ)  (*((typ *)cmdInfo->val))


  if (!cmdInfo)  return 0;

  toType = (uint) cmdInfo->valType;

//--- The generic data types require a pointer to the value to be updated ---
  if ((toType < NUM_MB_TYPES) AND !cmdInfo->val)  return 0;


  switch (toType)  {

    case DT_U16:   // a simple 16-bit unsigned int value
      u16Val = atoi(resp);
      _EnterCritical_();  ToVal(U16) = u16Val;  _ExitCritical_();
      return 0;


    case DT_F32:   // a simple floating point value
      fVal = atof(resp);
      _EnterCritical_();  ToVal(F32) = fVal;  _ExitCritical_();
      return 0;


/* --- Sample of response to *IDN? query ---
SORENSEN, SGA60/417E-2AAA, 0721A02109, 1.03, 1.02
SORENSEN,DCS12-250,0943A01282,2.02-2.00
*/
    case DT_STR:   // a string
      maxSize = cmdInfo->size - 1;
      if (((char *)cmdInfo->val == serIDStr) AND logIDStr)  {
        printf("PS ID: [%.128s]\n", resp);  logIDStr = false; }
      KLOCK();
      strncpy(cmdInfo->val, resp, maxSize);
      ((char *)cmdInfo->val)[maxSize] = 0;
      KUNLOCK();
      return 0;


/*--- Sample of response to SOUR:STAT:BLOC? query ---
0 1 2        3  4  5  6  7  8          9     10     11    12         13          14         15          16         17          18         19          20         21          22
1,1,00000881,02,03,00,00,00,0621A00561,60.00,417.00,66.00,9.6096e-04,-3.3346e-01,6.6789e-03,-1.1250e+00,1.0687e-03,-7.3202e-01,9.7788e-04,-1.9448e+00,6.8003e-03,-1.2993e+01,DCS60-417
1,1,00000881,02,02,00,00,00,0943A02767,12.00,250.00,13.20,2.0144e-04,-3.4245e-03,4.1967e-03,-7.1343e-02,2.4042e-04,-2.2984e-01,2.1781e-04,-3.8000e-01,4.5358e-03,-8.4998e+00,DCS12-250
*/
    case STATUS_BLOCK:  // a string with LOTS of values in it
      ptr = resp;  strs[0] = resp;
      for (ptr=resp,numFlds=0; *ptr AND numFlds<24; ++ptr)  {
        if (*ptr != ',')  continue;
        *ptr = 0;  strs[++numFlds] = ptr + 1;
      }
      if (*strs[numFlds])  ++numFlds;

      if (numFlds < 23)  {
        if (Cur.serLinkActive)
          printf("*** incomplete resp to SOUR:STAT:BLOC? ***\n");
        return -1;
      }

      KLOCK();

      serOnline = atoi(strs[1]);
      Next.serStatusFlags = (uint) strtol(strs[2], NULL, 16);
      serStatus = (uint) strtol(strs[3], NULL, 16);
      serAccumStatus = (uint) strtol(strs[4], NULL, 16);
      serFaultsMask = (uint) strtol(strs[5], NULL, 16);
      serFaults = (uint) strtol(strs[6], NULL, 16);
      serErrors = (uint) strtol(strs[7], NULL, 16);
      strncpy(serialNum, strs[8], sizeof(serialNum)-1);
      serModelVolts = atof(strs[9]);
      serModelAmps = atof(strs[10]);
      serModelOverV = atof(strs[11]);
      strncpy(modelStr, strs[22], sizeof(modelStr)-1);

      KUNLOCK();

      return 0;
  }

  return 0;
}


//-----------------------------------------------------------------------------
//  Send a cmd to the PS and wait for a response if cmd is a query
//
//  Returns 0 if no error, < 0 if an error occured.
//-----------------------------------------------------------------------------
int  psSendRcv(int fd, BufInfo *bufInfo, char *cmd, char *resp, uint respSize,
               uint *consecErrs)
{
  bool  queryCmd, showCmd, showErr;
  int  stat;
  uint  u, cmdLen, startTime, secSinceLastLogAll;


  *resp = 0;

//--- The only cmds that will NOT cause the analog setpoint to be ignored ---
  if (!serLinkRequired)  if (strncmp(cmd, "SYST:LOCAL", 10))  return -1;

  if (respSize < 3)  return -1;
  respSize -= 2;

  cmdLen = strlen(cmd);  queryCmd = (cmd[cmdLen-1] == '?');

//--- Show the cmd sent and the resp if diagnostic flag set ---
//--- Show errors if diagnostic flag set or mult consec errs for the cmd ---
  showCmd = (ShowPSMsgs OR (cmd == psCmd));
  showErr = (showCmd OR (Cur.serLinkActive AND *consecErrs));

/*
  If we haven't been able to establish a serial link, then we periodically
  log all traffic on the connection for a short period of time.  To avoid
  continually flooding the system log, we limit how long we do it for each
  time and increase the amount of time we wait before doing it again (the
  later being reset each time we do manage to establish communication).
*/
  do  {
    if (Cur.serLinkActive)  {
      logAll = false;  logAllInterval = serInitTime + 1;  break; }
    if (!serLinkNotWorking)  break;
  //--- time since the start of the last logAll period ---
    secSinceLastLogAll = SecSince_(lastLogAllSecs);
  //--- if interval passed, start logging traffic again and incr interval ---
    if (secSinceLastLogAll >= logAllInterval)  {  // begin logAll
      logAll = true;  lastLogAllSecs = GetSeconds_();
      logAllInterval = (logAllInterval << 2) / 3;  // incr interval by 1/3
      if (logAllInterval > maxLogAllInterval)  // limit length of interval
        logAllInterval = maxLogAllInterval;
      break;
    }
  //--- limit how long we log all traffic for ---
    if (secSinceLastLogAll > logAllPeriod)  logAll = false;  // end logAll
  } while(0);

  if (logAll)  showCmd = showErr = true;


//--- Empty the input buffer (counter and sleep used to avoid excessive CPU ---
//--- usage when bad/no connection causes spurious "incoming" chars) ---
  for (u=0; (stat = readBufCh(bufInfo)) > 0; ++u)  {
    if (logAll)  {
      if (!u)  printf("\nfrom PS[");  // prefix if this is the 1st char
      putchar(stat & 0x7F);
    }
    if ((u & 0x0f) == 0)  sleepMS(10);  // pause every 16 chars
  }
  if (logAll AND u)  printf("]\n");  // suffix if any chars logged


//--- Send the command string followed by a CR ---
  fcntl(fd, F_SETFL, ser_fflags);
  if (write(fd, cmd, cmdLen) != cmdLen)  { perror("write()");  return -1; }
  if (write(fd, "\r", 1) != 1)  { perror("write()");  return -1; }
  psLastSendRcvTime = GetMS();
  fcntl(fd, F_SETFL, ser_fflags | O_NONBLOCK);
  ++numCmdsSent;


//--- If the cmd is not a query, then don't wait for a response ---
  if (!queryCmd)  {
    if (showCmd)  printf("cmd:[%.128s]\n", cmd);
    sleepMS(PS_CMD_DELAY);  return 0;
  }


//--- Wait for the start of the response ---
  startTime = GetMS_();
  for (;;)  {
    if ((stat = readBufCh(bufInfo)) >= 0)  break;
    if (msSince_(startTime) > 1000)  {
      ++numRespLost;  consecGoodResp = consecBadResp = 0;
      if (showErr)
        printf("cmd:[%.128s]  *** No response ***\n", cmd);
      else
        ++expectedSyntaxErrs;
      ++*consecErrs;  return -1;
    }
    sleepMS(10);
  }

  psLastSendRcvTime = GetMS();  // The time we started reading the resp

//--- Read up to the next CR (responses end with CR/LF) ---
  for (u=0; u<respSize; ++u)  {
    serLastRcvTime  = GetMS_();
    if ((resp[u] = (uint)stat & 0x7f) == 0x0D)  break;  // got CR
    while ((stat = readBufCh(bufInfo)) < 0)  {
      if (errno != EAGAIN)  {
        perror("error reading resp from PS");  *resp = 0;  return -1; }
      if (msSince_(serLastRcvTime) > 500)  break;  // max 500ms between chars
      sleepMS(10);
    }
    if (stat < 0)  break;
  }
  ++numRespRcvd;


//--- Error if it didn't end with a CR ---
  if (resp[u] != 0x0D)  {
    resp[u+1] = 0;
    if (showErr)
      printf("cmd:[%.128s]\n"
             "*** incomplete response: [%.256s] ***\n", cmd, resp);
    ++*consecErrs;  ++consecBadResp;  return -1;
  }

//--- Replace the CR with a null terminator ---
  resp[u] = 0;  ++consecGoodResp;  *consecErrs = 0;

  sleepMS(PS_CMD_DELAY);

  if (showCmd)  printf("cmd:[%.128s]  resp:[%.256s]\n", cmd, resp);

  return 0;
}


//-----------------------------------------------------------------------------
//  Read and log any messages in the supply's error msg queue.
//
//  Returns the # of errors read.
//-----------------------------------------------------------------------------
int checkForErrs(int devfd, BufInfo *bufInfo, char *cmd,
                 char *respBuf, uint respBufSize)
{
  uint  u, consecErrs;
  int  numErrs = 0, err, stat;


//--- dump any msgs in the error msg queue ---
  for (u=0; u<10; ++u)  {  // up to 10 consecutive failed queries
    if (!Cur.psPowerEnabled)  break;

    stat = psSendRcv(devfd, bufInfo, "SYST:ERR?", respBuf, respBufSize,
                     &consecErrs);
    if ((stat != 0) OR !*respBuf)  continue;  // no response, try again

    if (*respBuf == '0')  break;  // err code 0 means err msg queue is empty

    err = atoi(respBuf);  // all other valid resps start with a non-0 error code
    if (!err)  continue;

//    if (err == -102)  // ignore expected syntax err msgs
//      if (expectedSyntaxErrs)  { --expectedSyntaxErrs;  continue; }

    if (!ShowPSMsgs AND (cmd != psCmd))  {
      if (!numErrs)  printf("cmd:[%.128s]\n", cmd);
      printf("*** [%.256s] ***\n", respBuf);
    }

    u = 0;  ++numErrs;
  }

  return numErrs;
}


//-----------------------------------------------------------------------------
//  Process to manage comunications with a Sorenson SGA power supply via an
//  RS-232 async serial link
//-----------------------------------------------------------------------------
void _task psCom(int argc, char **argv)
{
  bool  skipErrCheck, isOn, userCmd, useSerReads = true, newProjErr = true;
  int  stat;
  uint  cmdIdx, numSerReadA = 0, lastSerReadTime = 0, consecErrsLocal[8];
  uint  *pErrCount;
  fd_t  devfd;
  ulong  baudRate;
  float  diff, secs, setVolts, targetAmpSet, lastTargetAmpSet, absAmpRead,
         apparentRate, lastSerReadA;
  char  *nextCmd;
  const PS_CmdInfo   *cmdInfo;
  BufInfo  bufInfo;
  char  resp[256], readBuf[256], cmd[64];

/*
Responses to some commands from various Sorensen supplies:

  SOUR:STAT:BLOC?

1,1,00000881,02,03,00,00,00,0715A01363,60.00,417.00,66.00,9.5429e-04,-3.0045e-01,6.6376e-03,-2.1668e+00,1.0548e-03,-6.7719e-01,9.6922e-04,-1.7888e+00,6.7392e-03,-1.2454e+01,DCS60-417
1,1,00000881,02,03,00,00,00,0537A00680,60.00,417.00,66.00,9.5359e-04,-9.2000e-02,6.6408e-03,-2.0913e+00,1.0576e-03,-6.5991e-01,9.6867e-04,-1.7810e+00,6.7325e-03,-1.2366e+01,DCS60-417
1,1,00000881,02,03,00,00,00,0721A00715,60.00,417.00,66.00,9.5425e-04,-3.0658e-01,6.6385e-03,-2.1285e+00,1.0559e-03,-6.1190e-01,9.6968e-04,-1.7998e+00,6.7488e-03,-1.2722e+01,DCS60-417
1,1,00000881,02,03,00,00,00,0621A00558,60.00,417.00,66.00,9.5439e-04,-3.0013e-01,6.6582e-03,-1.4425e+00,1.0578e-03,-5.5337e-01,9.6782e-04,-1.7826e+00,6.7384e-03,-1.2939e+01,DCS60-417
1,1,00000881,02,03,00,00,00,0621A00561,60.00,417.00,66.00,9.6096e-04,-3.3346e-01,6.6789e-03,-1.1250e+00,1.0687e-03,-7.3202e-01,9.7788e-04,-1.9448e+00,6.8003e-03,-1.2993e+01,DCS60-417

1,1,00000881,02,02,00,00,00,0943A02767,12.00,250.00,13.20,2.0144e-04,-3.4245e-03,4.1967e-03,-7.1343e-02,2.4042e-04,-2.2984e-01,2.1781e-04,-3.8000e-01,4.5358e-03,-8.4998e+00,DCS12-250
1,1,00000881,02,03,00,00,00,0943A01914,12.00,250.00,13.20,2.0144e-04,-3.4245e-03,4.1967e-03,-7.1343e-02,2.4099e-04,-2.6581e-01,2.1891e-04,-3.7896e-01,4.5291e-03,-8.0251e+00,DCS12-250
1,1,00000881,02,02,00,00,00,0943A01282,12.00,250.00,13.20,2.0144e-04,-3.4245e-03,4.1967e-03,-7.1343e-02,2.4076e-04,-2.7143e-01,2.1823e-04,-3.9677e-01,4.5236e-03,-8.7164e+00,DCS12-250

   *IDN?

SORENSEN, SGA60/417E-2AAA, 0537A00680, 1.05, 1.03
SORENSEN, SGA60/417E-2AAA, 0621A00561, 1.03, 1.02

SORENSEN,DCS12-250,0943A01282,2.02-2.00
SORENSEN,DCS12-250,0943A02767,2.02-2.00
*/

  redirectStdio();

  numCmdsSent = numRespRcvd = consecGoodResp = consecBadResp = cmdIdx = 0;
  BZERO(consecErrsLocal);  BZERO(consecErrsGlobal);  expectedSyntaxErrs = 0;

  rampStartedTime = GetMS();  rampPeriod = 0;
  lastSerReadA = 0.0;


//--- open and initialize port for communicating with the PS ---
  if ((devfd = open("/dev/tty3", O_RDWR, 0)) < 0)  {
    printf("Unable to open /dev/tty3 for PS\n");  Exit(-3); }

  ser_fflags = fcntl(devfd, F_GETFL, 0);
  fcntl(devfd, F_SETFL, ser_fflags | O_NONBLOCK);

  baudRate = 19200;  ioctl(devfd, IOC_SET_BAUD, &baudRate);

  *resp = 0;

  bufInfo.buf = readBuf;  bufInfo.bufSize = sizeof(readBuf);
  bufInfo.fd = devfd;  bufInfo.nextOut = bufInfo.count = 0;

  lastTargetAmpSet = 0.0;  ramping = false;

//--- Clear all status values including the error msg queue ---
  psSendRcv(devfd, &bufInfo, "*CLS", resp, sizeof(resp), consecErrsLocal);

  for (;;)  {

  //--- If the serial link is not active because it is not powered on, ---
  //--- then set the basic values we get from it to 0 and don't even bother ---
  //--- trying talk to it ---
    if (!Cur.psPowerEnabled AND !Cur.serLinkActive)  {
      KLOCK();
      Next.serSetA = serReadA = serReadV = Next.serSetV = 0.0;
      numCmdsSent = numRespRcvd = numRespLost = consecGoodResp = consecBadResp = 0;
      targetAmpSet = lastTargetAmpSet = serRampTo = 0.0;
      projectFrom = projectBy = 0.0;  projectFromTime = 0;  ramping = false;
      KUNLOCK();
      sleepMS(20);  continue;
    }

  //--- If we just enabled power to the supply, wait a bit before we start ---
  //--- sending stuff to the serial interface. ---
    if (!Cur.psPowerEnabled OR (SecSince_(psPowerEnabledSecs) < SerComDelay)) {
      sleepMS(20);  continue; }


  //--- If we are NOT using the serial link for controlling/monitoring ---
  //--- the supply, then send one of the few cmds that does NOT cause it to ---
  //--- go in to remote mode:  The command to force it to use local mode. ---
    if (! serLinkRequired)  {
      psSendRcv(devfd, &bufInfo, "SYST:LOCAL 1", resp, sizeof(resp),
                consecErrsLocal+1);
      sleepMS(200);  continue;
    }


  //--- Set the volt limit as needed ---
    if (Cur.serLinkActive AND (serModelVolts > 0.0))  {
      _EnterCritical_();  setVolts = maxVolts;  _ExitCritical_();
      if (setVolts > serModelVolts)  setVolts = serModelVolts;
      diff = setVolts - serSetV;
      if ((fabs(diff) > 0.02) AND (consecGoodResp > NUM_PS_CMDS))  {
        strcpy(cmd, "SOUR:VOLT ");  fmtFloat(setVolts, 7, 3, cmd+10);
        printf("Setting max voltage: [%s]\n", cmd);
        psSendRcv(devfd, &bufInfo, cmd, resp, sizeof(resp), consecErrsLocal+3);
      }
    }


  //--- Update state of output if needed ---
    isOn = BOOL(Cur.serStatusFlags & OUTPUT_ON);
    if (Cur.assertOn != isOn)  {
      nextCmd = (Cur.assertOn ? "OUTP:STAT 1" : "OUTP:STAT 0");
      psSendRcv(devfd, &bufInfo, nextCmd, resp, sizeof(resp), consecErrsLocal+4);
    }



  //===== Send RAMP command when needed =====
    // If the supply does not support overriding a previously sent but not yet
    // finished RAMP command, then we limit the timeframe for RAMP cmds by
    // sending using the continuously updated ampApply value as the next target
    // value.  This minimizes the time we are stuck waiting to act on a new
    // setpoint that was received before the previous one was reached.  NOTE
    // that using the ABORT command is not a practical alternative as it causes
    // the supply to immediately jump to the target value in the "aborted"
    // ramp command.

    KLOCK();
    if (!allowOverlapSets)
      targetAmpSet = Cur.ampApply;
    else
      if ((Cur.ampSetPolarity != Cur.assertPolarity) OR (Cur.ampApply == 0.0))
        targetAmpSet = 0.0;
      else
        targetAmpSet = Cur.ampSet;
    KUNLOCK();
    targetAmpSet = fabs(targetAmpSet);

    diff = fabs(targetAmpSet - lastTargetAmpSet);

    do  {
      if (diff < 0.02)  break;

      //NOTE:  The ABORT cmd causes the supply to immediately jump to the
      //       target for the prev RAMP cmd.  So for supplies that allow it,
      //       the only way to STOP a ramp in progress is to change the
      //       setpoint to the value we believe the supply is currently at.
      if (ramping)  {
        if (!allowOverlapSets)  break;
        KLOCK();  absAmpRead = fabs(ampRead);  KUNLOCK();
        strcpy(cmd, "SOUR:CURR ");  fmtFloat(absAmpRead, 7, 2, cmd+10);
        psSendRcv(devfd, &bufInfo, cmd, resp, sizeof(resp), consecErrsLocal+5);
        KLOCK();
        projectFrom = absAmpRead;  projectBy = 0.0;
        projectFromTime = 0;  ramping = false;
        KUNLOCK();
        lastTargetAmpSet = absAmpRead;
        diff = fabs(targetAmpSet - lastTargetAmpSet);
        if (diff < 0.2)  break;
      }

      secs = diff / rampRate;
      if (!allowOverlapSets)  secs *= stepRampFactor;
      if (secs > 99.0)  secs = 99.0;
      if (secs < 0.1)   secs = 0.1;
      strcpy(cmd, "SOUR:CURR:RAMP ####.## ##.#");
      fmtFloat(targetAmpSet, 7, 2, cmd+15);  fmtFloat(secs, 5, 1, cmd+22);

      psSendRcv(devfd, &bufInfo, cmd, resp, sizeof(resp), consecErrsLocal+5);

      KLOCK();
      projectFromTime = rampStartedTime = GetMS();
      rampPeriod = (ulong)(secs * 1000.0) + 50;
      projectFrom = lastTargetAmpSet;  projectBy = 0.0;
      projectRate = diff / secs;
      serRampTo = lastTargetAmpSet = targetAmpSet;
      maxProjectBy = fabs(serRampTo - projectFrom);
      ramping = true;  useSerReads = false;  numSerReadA = 0;
      KUNLOCK();
    }  while(0);


  //===== send a one-time user cmd or the next query cmd from the list =====
    if (*psCmd)  {
      nextCmd = psCmd;  cmdInfo = NULL;  userCmd = true;
      pErrCount = consecErrsLocal+6;
      *pErrCount = 1;  // cause psSendRcv() to show a one-time error
    }
    else  {
      for (;;)  {  // skip low priority queries while ramping
        if (++cmdIdx >= NUM_PS_CMDS)  {
          cmdIdx = 0;  if (ShowPSMsgs)  printf("\n"); }
        if (!ramping OR (psCmds[cmdIdx].priority <= 1))  break;
      }
      nextCmd = (char *)psCmds[cmdIdx].cmd;  cmdInfo = psCmds + cmdIdx;
      pErrCount = consecErrsGlobal + cmdIdx;  userCmd = false;
    }

    stat = psSendRcv(devfd, &bufInfo, nextCmd, resp, sizeof(resp), pErrCount);

  //-- Get the time and projected readback as close as possible to when we --
  //-- last communicated with the PS --
    KLOCK();  absAmpRead = fabs(ampRead);  KUNLOCK();

  //===== If the cmd was one "injected" by the user, then just copy the =====
  //===== response back to the buffer it will be displayed from =====
    skipErrCheck = false;
    if (userCmd)  {
      KLOCK();  *psCmd = 0;
      if (stat == 0)  strncpy(psResp, resp, sizeof(psResp)-1);
      KUNLOCK();
    }
  //=== Otherwise parse the resp and update what we know of the PS's state ===
    else  if (stat == 0)  if (Process_SGA_PS_Resp(resp, cmdInfo) == 0)  {
      _EnterCritical_();  psLastRespTime = psLastSendRcvTime;  _ExitCritical_();

//      if (cmdIdx == SOUR_CURR_IDX)  newRdgSetA = true;

    /*
      Keep track of the difference between the projected readback and the
      lastest readback value from the serial interface.  If we are ramping,
      update the ramp rate used to compute the projected readback so it
      doesn't diverge too much from the serial readings.  NOTE: The serial
      reading is often "wrong" (it is frequently out of date).  The supply
      seems to consider updating this value a low priority (particularly
      at the beginning or end of a RAMP command).
    */
      if (cmdIdx == MEAS_CURR_IDX)  {
        serProjErr = absAmpRead - serReadA;
        ++numSerReadA;  // num of serial readings since last RAMP cmd

        if (ramping)  {
          if (numSerReadA > 1)  {
//            msUsed = msInterval(rampStartedTime, psLastRespTime);
//            msLeft = rampPeriod - msUsed;

            // amps/sec based on last 2 serial readbacks
            apparentRate = (serReadA - lastSerReadA) * 1000.0 /
              (float) ((uint)psLastRespTime - lastSerReadTime);

            KLOCK();
            projectFromTime = GetMS();
            projectFrom = fabs(ampRead);  projectBy = 0;
            projectRate = fabs(serReadA + apparentRate - projectFrom);
            maxProjectBy = fabs(serRampTo - projectFrom);
            KUNLOCK();
          }

          if (fabs(serProjErr) <= maxProjErr)  {
            oldestProjErrTime = serProjErrAge = 0;  newProjErr = true; }
          else  {
            if (!oldestProjErrTime)
              oldestProjErrTime = GetMS_();
            else  {
              serProjErrAge = msSince_(oldestProjErrTime);
              if (serProjErrAge >= maxProjErrAge)  {
                if (newProjErr)  {
                  printf("*** Projected ampRead error:  Projected: %s  Actual: %s ***\n",
                         fmtFloat(absAmpRead, 8,3, resp+32), resp);
                  newProjErr = false;
                }
                KLOCK();
                projectFrom = serReadA;  projectBy = 0;
                projectFromTime = psLastRespTime;
                maxProjectBy = fabs(serRampTo - projectFrom);
                KUNLOCK();
                useSerReads = true;
              }
            }
          }
        }
        else  {
          if (msSince_(lastRamping) >= 1000)  {
            useSerReads = true;  newProjErr = true;
            oldestProjErrTime = serProjErrAge = 0;
          }
          if (useSerReads)  {
            KLOCK();
            projectFrom = serReadA;  projectBy = 0.0;  projectFromTime = 0;
            KUNLOCK();
          }
        }

        lastSerReadTime = (uint) psLastRespTime;
        lastSerReadA = serReadA;
      }


    //--- By default we use the IDN string to determine if we are talking to ---
    //--- a supply that allows interrupting an unfinished RAMP command with ---
    //--- something other than an ABORT command ---
      if (cmdIdx == IDN_QUERY_IDX)
        if (fixedOverlapMode)
          allowOverlapSets = (fixedOverlapMode == OVERLAP_ALLOWED);
        else
          allowOverlapSets = !(strstr(serIDStr, "SGA60"));


      if (ramping)  skipErrCheck = true;
    }


    if (!skipErrCheck)
      if (checkForErrs(devfd, &bufInfo, nextCmd, resp, sizeof(resp)))
        if (userCmd)  {
          KLOCK();  strncpy(psResp, resp, sizeof(psResp)-1);  KUNLOCK(); }
  }
}


//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

_nearcall void readDefaults(uint ui, float *gain, float *offset)
{
  switch (ui)  {
    case INT_AMPS_READ:  *gain = -40.0;  break;
    case INT_VOLT_READ:  *gain =  -1.0;  break;
    case EXT_VOLT_READ:  *gain =  -2.5;  break;
    case EXT_AMPS_READ:  *gain = -50.0;  break;
    default:             *gain =  -1.0;
  }
  *offset = 0.0;
}

_nearcall void readCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5; }

_nearcall char *readUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

_nearcall void setDefaults(uint ui, float *gain, float *offset)
{
  *gain = (ui == OUTPUT_CTL) ? -40.0 : 1.0;
  *offset = 0.0;
}

_nearcall void setCalibPoints(uint ui, float *low, float *high)  {
  *low = -9.5;  *high = 9.5;  }

_nearcall char *setUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
  fstrcpy(ainLabel[0], "Int Cur Read");
  fstrcpy(ainLabel[1], "Int Volt Read");
  fstrcpy(ainLabel[2], "Ext Volt Read");
  fstrcpy(ainLabel[3], "Ext Cur Read");

  fstrcpy(aoutLabel[0], "Ctl Setpoint");

  fstrcpy(dinLabel[0], "DI-00");
  fstrcpy(dinLabel[1], "DI-01");
  fstrcpy(dinLabel[2], "DI-02");
  fstrcpy(dinLabel[3], "DI-03");
  fstrcpy(dinLabel[4], "DI-04");
  fstrcpy(dinLabel[5], "DI-05");
  fstrcpy(dinLabel[6], "DI-06");
  fstrcpy(dinLabel[7], "DI-07");
  fstrcpy(dinLabel[8], "DI-08");
  fstrcpy(dinLabel[9], "DI-09");
  fstrcpy(dinLabel[10], "Ext Interlock");
  fstrcpy(dinLabel[11], "PS Power is OFF");

  fstrcpy(doutLabel[0], "Norm Pol LED OFF");
  fstrcpy(doutLabel[1], "Rev Pol LED OFF");
  fstrcpy(doutLabel[2], "DO-02");
  fstrcpy(doutLabel[3], "DO-03");
  fstrcpy(doutLabel[4], "DO-04");
  fstrcpy(doutLabel[5], "DO-05");
  fstrcpy(doutLabel[6], "DO-06");
  fstrcpy(doutLabel[7], "Assert PS OFF");
  fstrcpy(doutLabel[8], "Activity LED");
  fstrcpy(doutLabel[9], "Disable Norm Pol");
  fstrcpy(doutLabel[10], "Disable Rev Pol");
}


//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE
//  supplyType = PS_TYPE_ANALOG;

  maxVolts = 50.0;
  maxAmps = 100.0;

  safeSwAmps = 2.0;

  connectLoadDelay = 500;  // # of ms to wait after turing on supply

  rampRate = 50.0;

  switchTime = 700;  // ms for polarity switches to change state and stabilize

  debounceMS = 250;  // interval (in ms) for debouncing digitalIn bits

//  maxRTZms = 10000;  // max ms to wait for PS to reach zero before turning Off */

  ctlModes = 0x222;  // use internal analog set and reads

  logStateSecs = 600;  // # secs between logging the state when the PS is ON

  stepRampFactor = 0.65;

  maxProjErr = 2.0;  // threshold for projected readback errors
  maxProjErrAge = 2000;  // max period for consecutive projected readback errors

  fixedOverlapMode = OVERLAP_AUTO;  // automatically set allowOverlapSets

  serInitTime = 12;  // max # secs from psPowerEnabed to serLinkActive

  logAllInterval = serInitTime + 1;
  maxLogAllInterval = 300;  // max # secs between auto log of ser err traffic
  logAllPeriod = 3;  //  # secs for each ser err traffic log "spurt"
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);

//APPLCODE

  Exit(0);
}


//=============================================================================
//  Use the values in prevLogInfo to determine if we should assert interlocks
//  for any devices any devices that were shut off due to a non-soft reset.
//=============================================================================
void checkPrevOnStates(void)
{
  if (prevLogInfo.onStatesBitMap)  offDueToReset = true;
}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---
//APPLCODE


  #define SharedApplLocals2
  #include <sharedAppl.c>



  strcpy(serIDStr, "<_unknown_>");

  #define SharedApplInit
  #include <sharedAppl.c>

  BZERO(bitCounts);  // counts for debouncing the digital inputs

//--- add application initialization code here ---
//APPLCODE
  allowSoftReset = true;


  #define SharedApplInit2
  #include <sharedAppl.c>

  lastRampRate = rampRate;
  maxAmpChg = rampRate * (float)SchedMS / 1000.0;


  serLastRcvTime = 0;
  BZERO(serialNum);  BZERO(modelStr);


//--- set averaging period for monitoring bias voltage for errors ---
//  setAvgPeriodADC(BIAS_VOLTS, 500);  // 500 ms


//--- task for serial com to Sorenson SGA PS (uses Port D or /dev/tty3) ---
  argv_[0] = "psCom";  argv_[1] = NULL;
  createProc(psCom, 2, 1, argv_, 0, 1, 2);


  if (!restoredApplState)  SetBits(devStatus, DS_CTLR_RESET);


//--- The following loop is run at a periodic interval (normally 100 times ---
//--- per second, or every 10ms).  The application process normally ---
//--- consists of reading the current state of the digital and analog ---
//--- inputs, changing the state of the analog and digital outputs to the ---
//--- values computed during the previous pass, then computing new values ---
//--- for the outputs and updating values that outside clients can view ---
//--- (currently accessable as a set of ModBus registers) ---

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>


  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
//APPLCODE
const CmdInfo far applCmds[] = {
//  { "misc",        miscTest,       "..." },
  { "cmd",       cmd,         "Application-specific commands" },
  { "psCom",     psCom,       "serial link to Sorenson SGA PS" },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
