//----- sharedAppl.c ----- 11/25/13 -----(10/26/06)----

//-----------------------------------------------------------------------------
//  Code shared by all application processes
//-----------------------------------------------------------------------------
//  Update Log:  (most recent first)
//
//
//  2013-11-26 - Mark Davis
//               - Added call to new application-specific function
//                 checkPrevOnStates(). This function is called if we didn't do
//                 a softReset but we do have the DiagLogInfo data from before
//                 the restart.  The DiagLogINfo structure includes the new
//                 onStatesBitMap value that is updated by each application
//                 near the end of each applic task cycle and is used by
//                 checkPrevOnStates() to determine  what (if any) devices were
//                 on just before the restart.
//
//  2013-11-22 - Mark Davis (*)
//               - rwIO() now sets digOutState in the DiagLogInfo structure.
//                 Prior to this, that value was set only before doing a
//                 softReset.  This change was done to allow new application-
//                 specific logic to determine if an Interlock should be
//                 asserted during startup (because something was shut off due
//                 to a reset of the CPU).
//
//  2012-12-05 - Mark Davis
//               - Moved MB_cmdBitsStd() and MB_timeZone() from modbus.c
//                 to sharedAppl.c
//
//  2012-07-31 - Mark Davis
//               - Changed all MS_TIMER references to GetMS() or GetMS_() and
//                 changed appropriate variables to uint where needed.  This
//                 eliminates several unsigned long operations in the main
//                 applic loop.
//               - New ca_tcpd task is now launched for all applications.
//
//  2011-06-24 - Mark Davis (*)
//               - setValMin[], setValMax[], readValMin[], readValMax[] are
//                 now properly initialized using the default values during
//                 startup.  Prior to this they could end up with 0 values
//                 for a bit during a portion of startup which could cause
//                 problems when restoring persistent values that are
//                 dependent on them.
//
//  2011-05-04 - Mark Davis
//               - Added startupComplete that is set true just before the
//                 main application loop starts.  This was added to allow
//                 certain log messages to be supressed to minimize the
//                 liklihood of overflowing the system log buffer during
//                 startup.
//
//  2011-05-02 - Mark Davis (*)
//               - Added call to resetWDT() in the loop that waits for the
//                 network to come up during startup.  Prior to this, the
//                 watchdog timer reboot would be tripped if the network did
//                 not come up during startup (e.g. if the cable was not
//                 connected).  This is a new bug caused by the changes made
//                 on 03/29/11 (that stopped the kernel from blocking when
//                 changing the network configuration).
//
//  2011-04-08 - Mark Davis
//               - Moved functions for tracking and managing averaged and noise
//                 values for analog inputs to the hdwDefs module.
//
//  2011-04-06 - Mark Davis
//               - Changed how initModBusRegs() handles an error in the lists
//                 of registers.  If the primary boot image is running, then
//                 it reboots and runs the backup image.  If the backup image
//                 is running, then it shuts down and stays in the boot loader.
//                 Prior to this it would shutdown only if running the primary
//                 image, and it would stay in the boot loader.
//
//  2011-02-25 - Mark Davis
//               - Removed handling of forced bits for extraIn (now handled by
//                 the ReadDigInputs() function for each hdw platform that
//                 uses extraIn).
//               - Added initialization of extraInLabel[][]
//               - Added setGenericLabels()
//
//  2011-01-07 - Mark Davis
//               - getAvgdAnaIn() now returns the forced ADC value if one is
//                 in effect.
//
//  2011-01-06 - Mark Davis
//               - Added allowSoftReset so individual applications can control
//                 when a soft reset is allowed.
//
//  2011-01-04 - Mark Davis (*)
//               - First version of saveApplState() and restoreApplState() that
//                 work and use BBRAM to save and restore the application state
//                 data.
//
//  2010-12-15 - Mark Davis (*)
//               - Fixed some bugs in the code for initializing the data
//                 structures used to compute averaged ADC values.  These bugs
//                 only showed up when working on the "soft reset" capability.
//
//  2010-12-14 - Mark Davis (*)
//               - Changed _TimerSeconds + secTimerOffset to TIME()
//               - Added saveApplState(), restoreApplState().
//
//  2010-07-01 - Mark Davis (*)
//               - The functionality that used to call DLM_Reboot() now clears
//                 the runBackup flag instead.  This provides the previously
//                 intended functionality of causing the backup boot image
//                 to be loaded if a problem occurs soon after a reboot.
//
//  2010-06-17 - Mark Davis
//               - Removed call to DLM_Reboot() for now because it ALWAYS
//                 overwrites the 128 bytes from 0x7FF80 to 0x7FFFF, which
//                 could have been allocated for some other use by the time
//                 it is called by the applic task.
//
//  2010-03-02 - Mark Davis
//               - Added startup of epics_cad process
//
//  2010-02-27 - Mark Davis
//               - removed references to scheduleTickXXX variables
//
//  2010-02-08 - Mark Davis (*)
//               - On the off chance that the 2009-12-07 change (occasionally
//                 reassert the current DAC settings to be "safe") is the
//                 cause of the ECROven supply unexplained drouputs, this
//                 feature is now disabled for the BL2105 platform (the one
//                 used for the ECROven controller).
//
//  2010-01-11 - Mark Davis (*)
//               - Added applicReady flag that is set to true at the end of
//                 each pass through the main applic task loop.  This was added
//                 for use by tasks that are dependent on the values maintained
//                 or initialized by the applic task.  The need for this was
//                 first discovered when readWave tasks for the PhaseCtlr
//                 application were allowed to start and process before the
//                 applic task had initialized the table of locks and other
//                 values.
//
//  2010-01-07 - Mark Davis (*)
//               - restoredCalibs() now copies the new value so it is not
//                 ignored.  Prior to this fix, the setOffset[] array (for
//                 which this function is called when reloading values from
//                 the devData file) was not being restored to the saved
//                 values.
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2009-12-07 - Mark Davis
//               - The raw DAC value for each analog output is occasionally
//                 rewritten (on a round-robin basis) just to make sure
//                 DAC's can't have the wrong value (for whatever reason)
//                 for any significant length of time.
//
//  2009-09-29 - Mark Davis
//               - setAvgPeriodADC() now uses ADCSampleInterval rather than
//                 SchedMS to compute the number of samples to avg.
//
//  2009-06-10 - Mark Davis -
//               - Initial state of digital inputs and outputs of interest are
//                 now logged
//
//  2009-05-26 - Mark Davis -
//               - Added code to keep track of min/max raw ADC values (must
//                 be reset by external process)
//
//  2009-05-11 - Mark Davis -
//               - The readVal[] value (analog input scaled to application
//                 values) is now set when an analog input value is forced.
//               - Changed name of readInputs() to initIOvalues() to avoid
//                 confusion in the future of the purpose of this func.
//
//  2009-04-30 - Mark Davis -
//               - The OS is rebooted if we are running the primary boot image
//                 and there is a problem with the Modbus registers.
//
//  2009-04-13 - Mark Davis -
//               - applic task now waits 8 seconds after rcving SIGTERM signal
//                 before calling applicShutdown()
//
//  2008-01-16 - Mark Davis -
//               - Added readInputs() to provide initial raw and scaled input
//                 values during startup, and to set values that will force
//                 the DAC's to be written to during the first pass of the
//                 application loop.
//
//  2007-03-14 - Mark Davis -
//               - moved reading and writing of digital and analog I/O signals
//                 to a separate func <rwIO()> and placed the call in the
//                 non-shared application-specific source files.  This makes
//                 it more plain when the I/O is being done and allows
//                 individual applications to contol when or if it is done.
//
//  2007-03-13 - Mark Davis -
//               - added DISABLE_AUTO_IO to allow application to disable the
//                 normal reading and writing of the digital and analog I/O
//                 signals by the shared code.
//
//  2006-10-10 - Mark Davis -
//               - first version
//
//-----------------------------------------------------------------------------

#ifdef SharedINC

  #include <common.h>

  #include <sys/types.h>
  #include <unistd.h>
  #include <sys/socket.h>
  #include <fcntl.h>
  #include <sys/stat.h>
  #include <signal.h>
  #include <errno.h>

  #include <string.h>

  #include <sys/backup.h>
  #include <sys/hdwDefs.h>

  #include <sys/srfs.h>

  #include <Rabbit.h>

  #include <sys/spi.h>
  #include <sys/serial.h>
  #include <sys/socket.h>
  #include <sys/epics_cad.h>


  #include <DLMiii.h>
  #include <SHD_DLM.h>

  #include <stcpip.h>

  #include <inetd.h>

  #include <sys/farprintf.h>

  #include <busybox.h>

  #include <applic.h>

  #include <tsh.h>

  #include <sys/backup.h>


  #include <ctype.h>


  void _task timeClient(int argc, char **argv);
  void _task configd(int argc, char **argv);

  void _task tsh(int argc, char **argv);
  void _task fastdatad(int argc, char **argv);


  void setDefaultConfig(void);

  RESTORE_FUNC(ReadSetCalibs);
  RESTORE_FUNC(hostName);



  AsyncParams consCfg = { 115200L, SER_RS232 };


  const uint NumApplAIN  = NUM_APPL_AIN;
  const uint NumApplAOUT = NUM_APPL_AOUT;
  const uint NumApplDIN  = NUM_APPL_DIN;
  const uint NumApplDOUT = NUM_APPL_DOUT;
  const uint NumExtraIn  = EXTRA_IN_BITS;


//--- conversion between voltage and appl/engineering units ---
  float  readGain[NUM_APPL_AIN], readOffset[NUM_APPL_AIN];
  float  setGain[NUM_APPL_AOUT], setOffset[NUM_APPL_AOUT];

  float  readVal[NUM_APPL_AIN], setVal[NUM_APPL_AOUT],
                                lastSetVal[NUM_APPL_AOUT];

  char far  ainLabel[NUM_APPL_AIN][MAX_LABEL_LEN];
  char far aoutLabel[NUM_APPL_AOUT][MAX_LABEL_LEN];
  char far  dinLabel[NUM_APPL_DIN][MAX_LABEL_LEN];
  char far doutLabel[NUM_APPL_DOUT][MAX_LABEL_LEN];
  char far extraInLabel[EXTRA_IN_BITS][MAX_LABEL_LEN];


  void far *stateDataBuf;  // BB RAM buffer for saving/restoring appl state


//=== The following list of shadow registers comes from Rabbit.h ===
//=== If you need to remove a register, REPLACE it with "&dummy" ===
//=== so the position of the rest of the registers in the list ===
//=== remain unchanged.  You can add registers at any time so long ===
//=== as they are added to the END of the list ===
/*
  typedef uchar * const ConstCharPtr;
  uchar  dummy;

  ConstCharPtr far shadowRegPtrs[] = {
    &srRTCCR, &srWDTCR, &srWDTTR, &srGOCR,  &srGCDR,  &srMB0CR, &srMB1CR,
    &srMB2CR, &srMB3CR, &srSPCR,  &srPADR,  &srPBDR,  &srPCDR,  &srPCFR,
    &srPDDR,  &srPDCR,  &srPDFR,  &srPDDCR, &srPDDDR, &srPDB0R, &srPDB1R,
    &srPDB2R, &srPDB3R, &srPDB4R, &srPDB5R, &srPDB6R, &srPDB7R, &srPEDR,
    &srPECR,  &srPEFR,  &srPEDDR, &srPEB0R, &srPEB1R, &srPEB2R, &srPEB3R,
    &srPEB4R, &srPEB5R, &srPEB6R, &srPEB7R, &srIB0CR, &srIB1CR,
    &srIB2CR, &srIB3CR, &srIB4CR, &srIB5CR, &srIB6CR, &srIB7CR, &srI0CR,
    &srI1CR,  &srTACSR, &srTAT1R, &srTACR,  &srTAT4R, &srTAT5R, &srTAT6R,
    &srTAT7R, &srTBCSR, &srTBCR,  &srTBM1R, &srTBL1R, &srTBM2R, &srTBL2R,
    &srTBCMR, &srTBCLR, &srSAAR,  &srSACR,  &srSBAR,  &srSBCR,  &srSCAR,
    &srSCCR,  &srSDAR,  &srSDCR,

    &srGCM0R, &srGCM1R, &srBDCR,  &srGPSCR, &srMTCR,  &srPFDR,  &srPFCR,
    &srPFFR,  &srPFDCR, &srPFDDR, &srPBDDR, &srPGDR,  &srPGCR,  &srPGFR,
    &srPGDCR, &srPGDDR, &srICCR,  &srICT1R, &srICS1R, &srICT2R, &srICS2R,
    &srPWL0R, &srPWM0R, &srPWL1R, &srPWM1R, &srPWL2R, &srPWM2R, &srPWL3R,
    &srPWM3R, &srQDCR,  &srQDCSR, &srTAPR,  &srTAT2R, &srTAT8R, &srTAT3R,
    &srTAT9R, &srTAT10R, &srSAER, &srSECR,  &srSEER,  &srSBER,  &srSFCR,
    &srSFER,  &srSCER,  &srSDER
  };
  #define  NUM_SHADOW_REGS  ( sizeof(shadowRegPtrs) / sizeof(char *) )

  uchar far shadowRegValues[NUM_SHADOW_REGS];
*/


  #undef SharedINC
#endif



//-----------------------------------------------------------------------------
#ifdef SharedServices
//-----------------------------------------------------------------------------


//#include <services.h>

//  NAME          PORT           PROTOCOL  FUNC        PRI  -1  Config Params
//  ------------  -------------  --------  ----------- ---  --  ---------------
  { "console",    TTY0,          ASYNC,    tsh,         10, -1, &consCfg },
  { "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,    5, -1, NULL },
  { "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,   6, -1, NULL },
  { "tsh",        TELNET_PORT,   ProtoTCP, tsh,         10, -1, NULL }

  #undef SharedServices
#endif

//-----------------------------------------------------------------------------
//  values that should be restored on bootup by whatever controller is
//  servicing a specific device
//
//  WARNING:  Adding ID's is OK, but never change what an existing one
//            stands for, even if it is no longer used!
//-----------------------------------------------------------------------------
#ifdef SharedDevParams

#define  AIN_LABELS     1000
#define  AOUT_LABELS    1001
#define  DIN_LABELS     1002
#define  DOUT_LABELS    1003

#define  TIME_ZONE      1004
#define  SRFS_IP_ADDR   1005

#define  READ_GAINS     1006
#define  READ_OFFSETS   1007
#define  SET_GAINS      1008
#define  SET_OFFSETS    1009

#define  HOST_NAME      1010

#define  SRFS_ALT_ADDR  1011

#define  EXTRAIN_LABELS 1012

#define  IP_NETMASK     1013
#define  IP_GATEWAY     1014

//                -----ID-----   ---value,granularity---      --restoreFunc--
  PERSISTENT_VAR( TIME_ZONE,     timeZone,                    NULL ),
  PERSISTENT_VAR( SRFS_IP_ADDR,  srfsIPAddr,                  NULL ),

  PERSISTENT_TBL( AIN_LABELS,    ainLabel, MAX_LABEL_LEN,     NULL ),
  PERSISTENT_TBL( AOUT_LABELS,   aoutLabel,MAX_LABEL_LEN,     NULL ),
  PERSISTENT_TBL( DIN_LABELS,    dinLabel, MAX_LABEL_LEN,     NULL ),
  PERSISTENT_TBL( DOUT_LABELS,   doutLabel,MAX_LABEL_LEN,     NULL ),

  PERSISTENT_TBLN( READ_GAINS,   float,readGain,  0,          NULL ),
  PERSISTENT_TBLN( READ_OFFSETS, float,readOffset,0,          NULL ),
  PERSISTENT_TBLN( SET_GAINS,    float,setGain,   0,          NULL ),
  PERSISTENT_TBLN( SET_OFFSETS,  float,setOffset, 0,          PVRF(ReadSetCalibs) ),

  PERSISTENT_TBL( HOST_NAME,     hostName,0,                  PVRF(hostName) ),

  PERSISTENT_VAR( SRFS_ALT_ADDR, srfsAltAddr,                 NULL ),

  PERSISTENT_TBL( EXTRAIN_LABELS, extraInLabel,MAX_LABEL_LEN, NULL ),

  PERSISTENT_VAR( IP_NETMASK,    netmask,                     NULL ),
  PERSISTENT_VAR( IP_GATEWAY,    gateway,                     NULL ),

  #undef SharedDevParams

#endif


//-----------------------------------------------------------------------------
#ifdef SharedParams
//-----------------------------------------------------------------------------

//        BASE NAME      CATEGORY      ParamData
//    -----------------  --------    -------------
ParamFile devDataFile =
    { "devData",         PDC_DEV,    devDataParams,0 };

ParamFile hdwCfgFile =
    { "hdwcfg",          PDC_SBC,    hdwCfgParams,0 };

  #undef SharedParams
#endif

//---------------------------------------
#ifdef SharedParamsList

  &hdwCfgFile,
  &devDataFile,

  #undef SharedParamsList
#endif


//-----------------------------------------------------------------------------
#ifdef SharedGlobals
//-----------------------------------------------------------------------------

  uint  ctlLoopTicks, ctlLoopTicksPerSec;
  uint  lastCtlLoopTime, lastReadTime;
  uint  readTicks, writeTicks, rwTicks;  //timing

  ulong  ADC_Mask, DAC_Mask;

  static ulong  lastIdleMS;
  static uint  lastIdleTime;

  char far applDTM[32];

  const ulong DigOutInitState = InitialDigOutState;

  char  applBuf[256];

  bool  applicReady = false, restoredApplState = false,
        allowSoftReset = false, startupComplete = false;



  #undef SharedGlobals
#endif


//-----------------------------------------------------------------------------
#ifdef SharedFuncs
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//  Scan the lists of ModBus registers for errors and shutdown if there is
//  an error.  If the primary boot image is the run running, then it will
//  reboot using the backup boot image.
//-----------------------------------------------------------------------------
void initModBusRegs(void)
{
  C0ChanMult = C0ChanMult_;  NumC0Chans = NumC0Chans_;
  R3ChanMult = R3ChanMult_;  NumR3Chans = NumR3Chans_;
  R4ChanMult = R4ChanMult_;  NumR4Chans = NumR4Chans_;

  if (pvInitLists())  {
    doShutdown = true;
    if (runningPrimary)  doReboot = runBackup = true;
  }
}


//-----------------------------------------------------------------------------
//  New value received for the timezone
//-----------------------------------------------------------------------------
bool _nearcall MB_timeZone(SetRegFuncParams)
{
  int  intVal;

  intVal = *((S16 *)newVal);

  if ((intVal < -12) OR (intVal > 12))  return false;

  timeZone = intVal;
  if (startupComplete)  fprintf(DiagLog, "new timezone value: %d\n", timeZone);

  _EnterCritical_();
  ((int *)REG->regVal)[chanNum] = *((int *)newVal);
  _ExitCritical_();

  return true;
}

//-----------------------------------------------------------------------------
//  New value received for cmdBitsStd
//-----------------------------------------------------------------------------
extern ulong  cmdBitsStd;
extern void _task _update(int argc, char **argv);

bool _nearcall MB_cmdBitsStd(SetRegFuncParams)
{
  ulong  cmdBits;
  char  buf[128];


//NOTE:  Don't save any of the bits that are simply one-time commands.
//      (i.e. only save ones that indicate a state going forward)

  cmdBits = MB_VAL(U32);

  if (cmdBits & CMD_REBOOT)  doShutdown = doReboot = true;

  if (cmdBits & CMD_SHUTDOWN)  doShutdown = true;

  if (cmdBits & CMD_SAVE_DEVDATA)
    saveParams(buf, sizeof(buf), &devDataFile, ForceFlush);
  if (cmdBits & CMD_RELOAD_DEVDATA)  loadParamsMsg(&devDataFile);

  if (cmdBits & CMD_SAVE_HDWCFG)
    saveParams(buf, sizeof(buf), &hdwCfgFile, ForceFlush);
  if (cmdBits & CMD_RELOAD_HDWCFG)  loadParamsMsg(&hdwCfgFile);
/*
  if (cmdBits & CMD_UPDATE_BOOTIMG)  {
    argv_[0] = "update";  argv_[1] = NULL;
    createProc(_update, 0, 1, argv_, ?, ?, ?);
  }
*/

  return true;
}


//-----------------------------------------------------------------------------
//  Read the state of the input signals, update the output signals to the
//  most recent values set by the application task
//-----------------------------------------------------------------------------
void rwIO(void)
{
  static int  ioTicks;
  uint  ui, chgdMask, chgBit;
  float far  *aout;
  bool  *update;
  ulong  startTicks, endTicks, rwStartTicks, rwEndTicks;  //timing



  ++ioTicks;

  rwStartTicks = clockTicks();  //timing

//--- check range on output voltage values and update raw DAC values ---
  chgdMask = 0;  chgBit = 1;
  if (DAC_Mask)  for (ui=0; ui<NumHdwDAC; ++ui,chgBit<<=1)  {
    update = updateAOut + ui;

    if (ui < NUM_APPL_AOUT)  if (setVal[ui] != lastSetVal[ui])  {
      lastSetVal[ui] = setVal[ui];  *update = true;  }

    aout = anaOut + ui;
    if (*aout != lastAnaOut[ui])  *update = true;

    if (!*update)  continue;
    *update = false;

    if (dacForced[ui])
      *aout = dacForcedVal[ui];
    else
      if (ui < NumApplAOUT)  *aout = SetToAOut(ui, setVal[ui]);
    if (*aout < aoutMin[ui])  *aout = aoutMin[ui];
    if (*aout > aoutMax[ui])  *aout = aoutMax[ui];
    lastAnaOut[ui] = *aout;

    dacRaw[ui] = AOutToDAC(ui, *aout);  chgdMask |= chgBit;
  }
  // force occasional update of DAC values in round-robin fashion
  // (just to be safe)
  if (hdwType != HT_BL2105)  chgdMask |= (1 << (ioTicks & 0x1f));

//--- read raw analog and digital inputs ---
  lastReadTime = GetMS_();
  startTicks = clockTicks();  //timing
  if (ADC_Mask)  readAnalogInputs(ADC_Mask);
  endTicks = clockTicks();  //timing
  digitalIn = ReadDigInputs();
  readTicks = (uint) UL_Interval(startTicks, endTicks);  //timing

//--- write latest values to analog and digital outputs ---
  startTicks = clockTicks();  //timing
  if (DAC_Mask)  writeAnalogOutputs(DAC_Mask & chgdMask);
  endTicks = clockTicks();  //timing
  WriteDigOutputs(digitalOut);
  writeTicks = (uint) UL_Interval(startTicks, endTicks);  //timing
  diagLogInfo->digOutState = digitalOut;  // added 2013-11-22
//--- update analog readings ---
  if (ADC_Mask)  for (ui=0; ui<NumHdwADC; ++ui)  {
    update = updateAIn + ui;
    if (adcRaw[ui] != lastadcRaw[ui])  {
      lastadcRaw[ui] = adcRaw[ui];  *update = true; }
    if (!*update)  continue;
    *update = false;
    anaIn[ui] = (adcForced[ui] ? adcForcedVal[ui] : ADCtoAIn(ui, adcRaw[ui]));
    if (ui < NumApplAIN)  readVal[ui] = AInToRead(ui, anaIn[ui]);
    newRdg[ui] = true;
  }

//--- update the running averages for raw ADC values ---
  updateAvgsADC();


  rwEndTicks = clockTicks();

  rwTicks = (uint) UL_Interval(rwStartTicks, rwEndTicks);
}


//-----------------------------------------------------------------------------
//  log messages for significant changes in the digital inputs
//-----------------------------------------------------------------------------
void logDigIOChgs(bool firstTime)
{
  uint  ui, offset;
  ulong  digInChgs = curState.digitalIn ^ prevState.digitalIn;
  ulong  digOutChgs = curState.digitalOut ^ prevState.digitalOut;
  ulong  mask;
  char far *msg;


  if (firstTime)  printf("\r\n===== Initial Digital Input states =====\r\n");


  if (firstTime OR digInChgs)
    for (mask=1,ui=0; ui<NumApplDIN; ++ui)  {
      if ((digInChgs & mask) OR firstTime)  {  // dig bit change?
        offset = ui << 1;  if (curState.digitalIn & mask) ++offset;
        msg = (char far *) digInDescr[offset];
        if (msg)  printf("%ls\r\n", msg);
      }
      mask <<= 1;
    }

  if (firstTime)  printf("\r\n===== Initial Digital Output states =====\r\n");

  if (firstTime OR digOutChgs)
    for (mask=1,ui=0; ui<NumApplDOUT; ++ui)  {
      if ((digOutChgs & mask) OR firstTime)  {  // dig bit change?
        offset = ui << 1;  if (curState.digitalOut & mask) ++offset;
        msg = (char far *) digOutDescr[offset];
        if (msg)  printf("%ls\r\n", msg);
      }
      mask <<= 1;
    }

  if (firstTime)  printf("\r\n");
}

//-----------------------------------------------------------------------------
//  Event-Handler to handle updating in-RAM copies of persistent values from
//  a ParamFile that was just restored from a more recent version.
//-----------------------------------------------------------------------------
void _nearcall EVENT_RestoredParamFile(EventMsg *event)
{
  uint  idx;
  ParamFile  *paramFile;


  idx = *(uint *)(event->data);
  paramFile = paramFiles[idx];

  fprintf(DiagLog, "restore from param file: %.24s\n", paramFile->baseName);

  loadParams(applBuf, sizeof(applBuf), paramFile, false);
}

//-----------------------------------------------------------------------------
//  the hostname value is being restored from a file
//-----------------------------------------------------------------------------
RESTORE_FUNC(hostName)
{
  SetHostName(newVal, strlen(newVal));
}

//-----------------------------------------------------------------------------
//  some portion of the analog calibrations were restored from a file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ReadSetCalibs)
{
  uint  ui;

//--- first we have to save the new value ---
  KLOCK();  fmemcpy(curVal, newVal, count);  KUNLOCK();

//--- then we need to update values related to the restored calib values ---
  for (ui=0; ui<NumHdwADC; ++ui)  newADCCalib(ui);
  for (ui=0; ui<NumHdwDAC; ++ui)  newDACCalib(ui);
}

//-----------------------------------------------------------------------------
//  set the date/time this module was compiled
//-----------------------------------------------------------------------------
void  setApplDTM(void)
{
  sprintf(applDTM, "%s %s", __DATE__, __TIME__);

  applVersTime = compileTime(__DATE__, __TIME__);
}

//-----------------------------------------------------------------------------
//  Save the state of the application so it can be restored after a "soft"
//  reset (i.e. a restart w/o a CPU reset).  For applications that support it,
//  this provides a way to reload the firmware and resume operation in response
//  to some unrecoverable error without resetting the state of the outputs.
//
//  Returns 0 on success, -1 if an error
//-----------------------------------------------------------------------------
int saveApplState(void)
{
  uint  dataSize;
  char far *ptr;
  ApplStateDataHdr far *hdr;


  if (!stateDataBuf)  return -1;

  KLOCK();

  fprintf(DiagLog, "*** Saving ApplState ***\r\n");

  dataSize = ApplStateDataSize;

  hdr = stateDataBuf;  ptr = (char far *)hdr + sizeof(ApplStateDataHdr);

  fmemcpy(ptr, &prevState, sizeof(ApplState));  ptr += sizeof(ApplState);
  fmemcpy(ptr, &curState,  sizeof(ApplState));  ptr += sizeof(ApplState);
  fmemcpy(ptr, &nextState, sizeof(ApplState));  ptr += sizeof(ApplState);

  fmemcpy(ptr, setVal, sizeof(setVal));  ptr += sizeof(setVal);

  ptr = (char far *)hdr + sizeof(ApplStateDataHdr);

  hdr->dataSize = dataSize;
  hdr->crc = memCRC16((ulong)ptr, dataSize);
  hdr->resetTime = time(NULL);


//--- Copy the cur values of all the shadow registers to a single struct ---
//  for (ui=0; ui<NUM_SHADOW_REGS; ++ui)
//    shadowRegValues[ui] = *shadowRegPtrs[ui];

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Allocate a block of Battery Backed RAM for saving/restoring the state of
//  the application logic across a "soft" reset (i.e. one that does not
//  involve a CPU reset).
//
//  If the allocated block contains a valid state from before the reset, and
//  the last reset was a soft reset, then restore the state data.
//
//  WARNING:  This function must ONLY be called at startup before there are any
//            tasks running.  If successful
//
//  Returns: < 0: No buffer to restore from
//             0: application state data found and restored
//           > 0: Buffer exists, but state data not usable
//-----------------------------------------------------------------------------
int restoreApplState(void)
{
  uint  dataSize;
  char far *ptr;
  ApplStateDataHdr far *hdr;


  if (!stateDataBuf)  return -1;

  dataSize = ApplStateDataSize;  hdr = stateDataBuf;

//--- Check to see if the buffer contains a valid header and data from a ---
//--- run that ended less than 60 seconds ago ---
  if (hdr->dataSize != dataSize)  return 1;  // applState struct size chgd?
  if (time(NULL) - hdr->resetTime > 60)  return 2;
  ptr = (char far *)hdr + sizeof(ApplStateDataHdr);
  if (hdr->crc != memCRC16((ulong)ptr, dataSize))  return 3;

  fmemcpy(&prevState, ptr, sizeof(ApplState));  ptr += sizeof(ApplState);
  fmemcpy(&curState,  ptr, sizeof(ApplState));  ptr += sizeof(ApplState);
  fmemcpy(&nextState, ptr, sizeof(ApplState));  ptr += sizeof(ApplState);

  fmemcpy(setVal, ptr, sizeof(setVal));  ptr += sizeof(setVal);

//finish:  This is too simplistic:  Many of the registers are probably
//         modified LONG before we get here - and yet we need the shadow
//         register values to agree with the actual in-use values!
//--- Restore the values of all the shadow registers ---
//  for (ui=0; ui<NUM_SHADOW_REGS; ++ui)
//    *shadowRegPtrs[ui] = shadowRegValues[ui];

  curState.digitalIn = nextState.digitalIn = digitalIn;
  digitalOut = curState.digitalOut;

  printf("\r\n*** Restored Application State ***\r\n\n");

  return 0;
}

//-----------------------------------------------------------------------------
//  Set labels for digital I/O signals to generic values
//-----------------------------------------------------------------------------
void setGenericLabels(void)
{
  uint  ui;

  for (ui=0; ui<NumApplAIN; ++ui)  {
    fmemset(ainLabel[ui], 0, MAX_LABEL_LEN);
    sprintf(ainLabel[ui], "AIN-%u", ui);
  }
  for (ui=0; ui<NumApplAOUT; ++ui)  {
    fmemset(aoutLabel[ui], 0, MAX_LABEL_LEN);
    sprintf(aoutLabel[ui], "AOUT-%u", ui);
  }

  for (ui=0; ui<NumApplDIN; ++ui)  {
    fmemset(dinLabel[ui], 0, MAX_LABEL_LEN);
    sprintf(dinLabel[ui], "DIN-%u", ui);
  }
  for (ui=0; ui<NumApplDOUT; ++ui)  {
    fmemset(doutLabel[ui], 0, MAX_LABEL_LEN);
    sprintf(doutLabel[ui], "DOUT-%u", ui);
  }
  for (ui=0; ui<NumExtraIn; ++ui)  {
    fmemset(extraInLabel[ui], 0, MAX_LABEL_LEN);
    sprintf(extraInLabel[ui], "ExtraIN-%u", ui);
  }
}

//-----------------------------------------------------------------------------
  #undef SharedFuncs
#endif


//-----------------------------------------------------------------------------
//  pieces included in the application process code
//-----------------------------------------------------------------------------
#ifdef SharedApplLocals

  ulong  idleMS, secs, nanoSecs, terminateTime, startTime;
  uint  u, ui, intTime, idleInterval, numBytes, ms, loopCountdown;
  int  stat;
  fd_t  eventsfd;
  char  *argv_[3];
  sigset_t  sigSet;
  RegInfo  *reg;

  #undef SharedApplLocals
#endif


#ifdef SharedApplLocals2
  char  buf[64];

  #undef SharedApplLocals2
#endif


//-----------------------------------------------------
#ifdef SharedApplInit
//-----------------------------------------------------

  BZERO(&prevState);  BZERO(&curState);  BZERO(&nextState);

  numScopeRegs = 0;

  resetWDT();  // reset 2-second countdown Watchdog timer

//--- which ADCs/DACs to read/write ---
  ADC_Mask = ADC_MASK_APPL;  DAC_Mask = DAC_MASK_APPL;


  capturePtr = NULL;


//--- set defaults for gains, offsets, and labels ---
  for (ui=0; ui<NumHdwADC; ++ui)  {
    adcDefaults(ui, adcGain+ui, adcOffset+ui);
    lastadcRaw[ui] = 0x0BAD;
    newADCCalib(ui);
  }
  for (ui=0; ui<NumHdwDAC; ++ui)  {
    dacDefaults(ui, dacGain+ui, dacOffset+ui);
    newDACCalib(ui);
  }
  for (ui=0; ui<NumApplAIN; ++ui)  {
    readDefaults(ui, readGain+ui, readOffset+ui);
    readVal[ui] = 0.0;
    if (ui <= NumHdwADC)  newADCCalib(ui);  // sets readValMin/Max[]
  }
  for (ui=0; ui<NumApplAOUT; ++ui)  {
    setDefaults(ui, setGain+ui, setOffset+ui);
    setVal[ui] = 0.0;
    if (ui <= NumHdwDAC)  newDACCalib(ui);  // sets setValMin/Max[]
  }

  setGenericLabels();

//--- block the TERMinate signal so we can check for it and do an orderly ---
//--- cleanup before the shutdown begins ---
  sigSet = sigmask(SIGTERM);  sigprocmask(SIG_BLOCK, &sigSet, NULL);

  printf("applic process started\r\n");

  initModBusRegs();

  devType = DeviceType;


//--- Default network parameters.  These were zeroed out by initModBusRegs() ---
//--- and will be overwritten by initParamModule() if prev values were saved ---
  inet_ptoh(AF_INET, DEF_NETMASK, &netmask);
  inet_ptoh(AF_INET, DEF_GATEWAY, &gateway);


//--- create named pipe for receiving appl event msgs ---
  mkfifo(EventsFIFO, 0);
  eventsfd = open(EventsFIFO, O_RDONLY | O_NONBLOCK, 0);


  lastIdleTime = GetMS_();  lastIdleMS = 0;

//  _EnterCritical_();
//  scheduleTicks = scheduleTicksSeen = scheduleTicksLate = 0;
//  _ExitCritical_();

  ctlLoopTicks = 0;  lastCtlLoopTime = 0;

  terminateTime = 0;

//--- set default values for device calibiration/configurable values ---
  setDefaultLabels();  setDefaultConfig();

//--- Get initial/actual state of inputs ---
  initIOvalues();

/* If this startup is due to a Soft Reset:
      - Restore the appl state (which includes the analog and digital
        I/O values in the cur/prev/next ApplState structures)
      - Update the Cur and Next input values with the values just read
      - Change the assumed digitalOut state to the before-reset value

  WARNING: The restoration of the application state must be done AFTER all
           the default initial values are applied so any values restored
           from before the Soft Reset are not overwritten.  The only values
           that are safe to overwrite (and that we WANT overwritten) are
           the initial readings for the input signals (in case they changed
           while we were restarting).
*/
  if (prevLogInfo.didSoftReset)  {
    stat = restoreApplState();
    if (stat)  {
      printf("*** restoreApplState() failed (%d):  Doing CPU reset ***\r\n",
             stat);
      if (runningPrimary)  runBackup = false;
      REBOOT(true);
    }
    restoredApplState = true;
  }


//--- reload all persistent data values from files ---
  initParamModule(applBuf, sizeof(applBuf));

//--- Note that restoreApplState() sets digitalOut and cur/next digitalIn ---
  if (!restoredApplState)  {
    prevState.digitalIn = curState.digitalIn  = nextState.digitalIn = digitalIn;
    prevState.digitalOut = curState.digitalOut = nextState.digitalOut = digitalOut;
  }

//--- If we didn't restore the previous state, but we did restore the ---
//--- DiagLogInfo structure, then the application logic can use it to ---
//--- determine if devices were on just before a reset and decide if it ---
//--- should assert interlocks (to indicate there is a known reason for ---
//--- them shutting off) ---
  if (restoredLog AND !restoredApplState)  checkPrevOnStates();


  startNetUp(IF_ETH0);  startTime = GetMS();
  while ((msSince(startTime) < 10000) AND (ifpending(IF_ETH0) != IF_UP))
    resetWDT();
  if (ifpending(IF_ETH0) != IF_UP)
    printf("\r\n***** Network did not come up! *****\r\n");
  else
    showNetState(IF_ETH0);


  for (ui=0; ui<NumHdwADC; ++ui)  newADCCalib(ui);
  for (ui=0; ui<NumHdwDAC; ++ui)  newDACCalib(ui);


  #undef SharedApplInit
#endif

//------------------------------------------------------
#ifdef SharedApplInit2
//------------------------------------------------------
  resetWDT();  // reset 2-second countdown Watchdog timer


  logDigIOChgs(true);


  argv_[0] = "inetd";  argv_[1] = NULL;
  createProc(inetd, 3, 1, argv_, 0, 1, 2);

  argv_[0] = "configd";  argv_[1] = NULL;
  createProc(configd, 7, 1, argv_, 0, 1, 2);

  argv_[0] = "backupd";  argv_[1] = NULL;
  createProc(backupd, 4, 1, argv_, 0, 1, 2);

  argv_[0] = "ca_tcpd";  argv_[1] = NULL;
  createProc(ca_tcpd, CA_TCPD_PRIORITY, 1, argv_, 0, 1, 2);


  loopCountdown = 1500;

  resetWDT();  // reset 2-second countdown Watchdog timer

  applicTicks = 0;

  intTime = GetMS_();  applLoopTime = 0;

  printf("free memory: %ld\r\n", xavail());


/* //tdebug
  for (u=0,task=tasks; u<NUM_TASKS; ++u,++task)  {
    stackAddr = ((ulong)(task->sseg) << 12) + (ulong)(task->stack_start);
    printf(" task %2d stack: %02X:%04X-%04X  (%06lX-%06lX)\r\n",
           u, (uint)task->sseg, (uint)(task->stack_start),
           (uint)(task->stack_start) + task->stack_size - 1,
           stackAddr, stackAddr + task->stack_size - 1);
//           sseg = (uint) (((ulong) absAddr - 0xD000l) >> 12);
//           logAddr = 0xD000 | (((uint) absAddr) & 0x0FFF);
  }
*/

  startupComplete = true;

  #undef SharedApplInit2
#endif



//-----------------------------------------------------
#ifdef SharedApplLoopStart
//-----------------------------------------------------

  ++applicTicks;

  //--- If we made it through the main appl loop 1500 times (~ 15/30 secs) ---
  //--- then tell the boot manager that the version of code it loaded ---
  //--- appears to be stable ---
    #ifdef SHD_DLM
      if (loopCountdown)
        if (--loopCountdown == 0)  runBackup = false;
    #endif

    resetWDT();

    applLoopTime = msSince_(intTime);

  //--- sleep until woken up by timer for next pass ---
    EndInterval(TrackApplLoop);
    current->state = TASK_UNINTERRUPTIBLE;  schedule();
    StartInterval(TrackApplLoop);

//    ++scheduleTicksSeen;

    intTime = GetMS_();

  //--- check for timing errors ---
    _EnterCritical_();  latency = intTicks;  _ExitCritical_();
//    if (latency > 1)  { ++scheduleTicksLate;  continue; }
//    lostTicks = (uchar) (scheduleTicks - scheduleTicksSeen);


  //--- keep track of how often we get run ---
    ++ctlLoopTicks;
    if (UI_Interval(lastCtlLoopTime, intTime) >= 2000)  {
      ctlLoopTicksPerSec = ctlLoopTicks >> 1;
      ctlLoopTicks = 0;  lastCtlLoopTime = intTime;
    }

  #undef SharedApplLoopStart
#endif



//-----------------------------------------------------
#ifdef SharedApplLoopEnd
//-----------------------------------------------------
  //--- update device ID ---
    deviceID = ReadDeviceID();

  //--- update registers ---
    if ((idleInterval = msSince_(lastIdleTime)) >= 3000)  {
      idleMS = idleTask->msUsed;
      percentIdle = (uint)(((idleMS - lastIdleMS) * 100) / idleInterval);
      lastIdleMS = idleMS;  lastIdleTime = GetMS_();
    }

    devType_ = devType;
    deviceID_ = deviceID;
    upSecs = SEC_TIMER;
    percentIdle_ = percentIdle;

    timeZone_ = timeZone;
    curTime = time(NULL);

    digInBits = digitalIn;
    digOutBits = digitalOut;

    if (sigpending(&sigSet) == 0)  if (sigismember(&sigSet, SIGTERM))
      if (! terminateTime)  terminateTime  = GetMS();

    if (terminateTime AND (msSince(terminateTime) > 8000))  applicShutdown();

  //--- blink Activity LED ---
    if (GetMS_() & 0x40)
      TurnOnActivityLED();
    else
      TurnOffActivityLED();


  //--- copy current "SoftScope" channel values to circular buffer ---
    if (capturePtr)  {
      //finish:  switch to gettimeofday() when implemented?  But that adds
      //         the overhead of a far call and another multiply...
      _EnterCritical_();
      secs = TIME();  ms = milliTicks;
      _ExitCritical_();

      //--- timestamp with seconds and nanoseconds ---
      nanoSecs = (ulong)ms * 1000000L;
      *(ulong far *)capturePtr = htonl(secs);  capturePtr += 4;
      *(ulong far *)capturePtr = htonl(nanoSecs);  capturePtr += 4;

      for (u=0,reg=scopeList[0]; u<numScopeRegs; ++u,reg=scopeList[u])  {
        numBytes = reg->numRegs << 1;
        if (numBytes == 2)
          *((ushort far *)capturePtr) = htons(*reg->regVal);
        else
          *((ulong far *)capturePtr) = htonl(*((ulong *)reg->regVal));
        capturePtr += (ulong)numBytes;
      }
      ++captureCount;

      if (++captureIdx == scopeBufSampleSize)  {
        captureIdx = 0;  capturePtr = scopeBuf; }
    }

    applicReady = true;

  #undef SharedApplLoopEnd
#endif

//-----------------------------------------------------------------------------
