//----- backup.c ----- 11/22/13 -----(07/19/06)-----

//-----------------------------------------------------------------------------
//  Process to automatically handle the following:
//   - backup and restore "persistent data" values
//   - copy all new entries the in-RAM log to a remote file
//   - keep the controller's clock in sync with an external time server
//-----------------------------------------------------------------------------
// Persistent values:
//
//  Each application defines a "paramFiles" table for data that it wants to
//  be "persistent".  Persistent, in this case, means values that can be
//  changed while the application is running and need to be restored to the
//  value they last had whenever the controller is rebooted.
//
//  This is accomplished by use of 2 structures:
//
//  ParamData:  Each ParamData structure references a single logical value
//              or data structure.  An array of ParamData structures is used
//              to define a group of related values that will be backed up
//              in to a single file.
//
//  ParamFile:  A ParamFile structure defines the relationshsip between an
//              array of ParamData structures and a file for storing a copy
//              of the current values referenced by the ParamData elements.
//              On startup, the controller attempts to open each file in
//              a list of BackupFile elements and restore any values found
//              that match the corresponding ParamData elements.
//
//  Use of the ParamData structure to reference individual elements makes
//  it much simpler to alter the persistent data elements used by an
//  applicaton.  By organizing data elements in to paramterized lists rather
//  than fixed monolithic structures, each version of an application can
//  use the elements it understands and ignore those it doesn't.  Some elements
//  (such as board-specific values like ADC/DAC calibration values) can even
//  be accessed independent of the application.
//
//  Newer versions of the application can read a file created by an older
//  version, using the elements that still apply, and ignoring values that are
//  no longer needed.  Older versions can read a file written by a newer
//  version, ignoring newer elements that they don't know about.
//
//  Not perfect, of course, since values might be missing either way.  But it
//  is far better than having it crash because some fixed structure changed and
//  the data elements don't line up (or wasting space with placeholders for
//  elements no longer in use).
//
//  NOTE:  Because some values may NOT be in the file (or the file might not
//         even exist yet), all persistent values should be initialized to
//         reasonable default values on startup before attempting to restore
//         them from the corresponding file.
//-----------------------------------------------------------------------------
//  Update Log (most recent first)
//
// ---------------------------------------------------
//   *** TO-DO ***
//               - Add locks to avoid possible concurrent calls (by the backupd
//                 task and commands like save and reload) to the functions
//                 that access the param files.
// ---------------------------------------------------
//
//  2013-11-22 - Mark Davis
//               - Fixed saveParams() so it uses the size of the buffer passed
//                 to it rather than sizeof() the pointer to the buf.  Previous
//                 to this it would only write 2 bytes at a time, which was
//                 rather inefficient (and entailed a higher risk of ccorrupted
//                 data if an error occured while writing to flash).
//  (fixed 2013-10-04 on Linux vesion)
//               - The log file is now closed and reopened if the value of
//                 deviceID changes.
//
//  2012-10-02 - Mark Davis (*)
//               - Modified the parameters for how often changes to cached
//                 copies of persistent values may be flushed to flash.  Prior
//                 to this, flash updates were happing frequently enough that
//                 the limited # of writes to a block of flash was likely to
//                 be exceeded on some devices before too long.
//
//  2012-03-07 - Mark Davis
//               - diagLogInfo->hdwResetTime is now updated when the global
//                 hdwResetTime is
//
//  2011-12-08 - Mark Davis (*)
//               - Until I can add proper locking, and get all the opens and
//                 closes balanced/in sync, I changed the functions that
//                 need to open a file to close it first if it is already open.
//                 Without this, the last change resulted in errors because
//                 the opens/closes were not balanced, causing EBUSY errors
//                 when an attempt to open a file found it already/still open.
//
//  2011-12-07 - Mark Davis (*)
//               - Moved the call to closeParamFile() in checkCache() so the
//                 param file will be closed when the next attempt to open it
//                 occurs.  This is necessary because the previous set of
//                 changes uses the short-term (incomplete) method of guarding
//                 against concurrent use of a file by having the functions
//                 that open a param file fail if it is already open.
//
//  2011-11-21 - Mark Davis (*)
//               - Changes to eliminate "orphaned" fd's for param files.  Prior
//                 to this, a param file of the wrong size would result in
//                 repeated loops through the logic to rewrite the file that
//                 would accumulate "orphaned" open fd's for the file, quickly
//                 making it impossible to access ANY flash file.
//
//  2010-12-14 - Mark Davis
//               - Changed secTimerOffset to hdwResetTime
//
//  2010-06-29 - Mark Davis
//               - Added log message for memory allocation calls
//
//  2010-05-02 - Mark Davis (*)
//               - Fixed bug in new code that stops a cached copy of settings
//                 data from being flushed to the flash file if the file size
//                 was not the same size as the cache.
//
//  2010-04-01 - Mark Davis
//               - Parameter lists with no entries are now ignored.  This
//                 avoids the creation of 0-length files and reduces the
//                 overhead a bit of such "lists" (some lists are standard and
//                 expected to exist even if they are empty).
//
//  2010-03-30 - Mark Davis
//               - A BBRAM (Battery-Backed RAM) cache is now created and used
//                 for all parameter lists of type PDC_SET (e.g. settings)
//                 to avoid excessive wear on the flash.  backupd now copies
//                 updated runtime values to the cache frequently, and flushes
//                 any newer cache values to flash much less frequently.  The
//                 previous version had the barest beginning of this and, in
//                 its barely started state, was writing all the changes it
//                 saw to flash, causing lots of wear and slowing down
//                 response times to fast series of updates.
//
//  2010-03-23 - Mark Davis
//               - The version date/time for the OS and the application are
//                 now written to the log file each time it is (re)opened.
//               - Added size info to data logged for remote/local parameter
//                 files
//
//  2010-03-12 - Mark Davis
//               - getTime() now checks the state of the network interface
//                 before bothering to try to get a connection to the time
//                 server.  This eliminates pointless and frequent messages
//                 in the log.
//
//  2010-02-24 - Mark Davis
//               - When renaming a remote backup copy of the devData file,
//                 the last modification date and time of the file are used
//                 as a suffix (rather than simply using .bak all the time).
//
//  2009-12-18 - Mark Davis
//               - Increased max # of params readParams() can handle to 128
//
//  2009-08-24 - Mark Davis
//               - Parameter files for settings (PDC_SET) files are created
//                 if they don't already exist.
//               - readParams() now returns a positve number not just if one
//                 of the saved values != an in-RAM value, but if any of the
//                 in-RAM values are not saved in the file.  This insures that
//                 changes to settings are actually saved.
//
//  2009-08-06 - Mark Davis
//               - updateParamFiles() no longer tries to continue if the call
//                 to loadParams() returns an error (which happens if there
//                 is no devData file).
//
//  2009-06-21 - Mark Davis
//               - The time the log file is opened by a module is now included
//                 in the line written directly to the file at the time.
//
//  2009-06-19 - Mark Davis
//               - The Core Module ID (last 3 bytes of MAC addr) is now written
//                 to the remote copy of the log file each time the module
//                 reopens the connection to the remote file.
//
//  2009-06-11 - Mark Davis
//               - Added month to the name used for the remote copy of the
//                 system log so we will always have a least a year's worth
//                 the log files.
//
//  2009-05-07 - Mark Davis
//               - Corrected value used compute ZWorld RTC value from the UTC
//                 value returned by the time server.
//
//  2009-04-13 - Mark Davis
//               - Application now waits 5 seconds from time a SIGTERM signal
//                 is rcvd before shutting down (to allow entries that were
//                 likely added to the system log to be saved to the server).
//
//  2009-03-23 - Mark Davis
//               - 10 attempts (one per sec) are now made (if necessary) to
//                 get the date/time for a new remote backup copy of a file
//                 to be used for setting the local copy's date/time.
//               - If we fail to get the date/time for the new remote file,
//                 the date/time of the local file is not changed (before it
//                 was changed to whatever value happened to be in the
//                 struct used to return the remote's date/time value).  This
//                 could account for the invalid date/times seen in messages
//                 in the system log.
//
//  2009-03-13 - Mark Davis
//               - saveParamsToFile() now checks the return value of unlink()
//                 and retries for up to 5 seconds if needed to successfully
//                 delete the file.  If it cannot delete the file it does not
//                 try to overwrite it.  This eliminates the most likely case
//                 of problems caused by concurrent access to parameter files
//                 like devdata.  Prior to this it was not uncommon to end up
//                 with a corrupted devData file from doing multiple saves,
//                 causing it to be open by both backupd and the save cmd.
//
//  2008-04-07 - Mark Davis
//               - a size of 0 for a parameter is now considered invalid and
//                 a sign of a corrupt param file
//
//  2008-02-28 - Mark Davis
//               - age of log files now correct (the # of hours was off by a
//                 factor of 10)
//
//  2008-02-20 - Mark Davis
//               - Added 5 sec delay after file I/O errors
//
//  2007-12-12 - Mark Davis
//               - Added logic to automatically save changes to PERSISTENT
//                 (ParamData/ParamFile) values (up to 5 times in a 15-minute
//                 window, and no more often then every 30 seconds)
//
//  2007-12-11 - Mark Davis
//               - Fixed interval for checking on time when it fails so it
//                 doesn't do it continuously
//
//  2007-10-02 - Mark Davis
//               - ParamData struct now includes a pointer to a func to call
//                 whenever the value is restored from a file
//
//  2007-08-01 - Mark Davis
//               - update of remote backup copy of a file now requires that
//                 the local copy was successfully opened
//
//  2007-07-13 - Mark Davis
//               - added code to automatically maintain remote files with a
//                 copy of system log entries
//
//  2006-07-19 - Mark Davis
//               - original version started
//
//-----------------------------------------------------------------------------


#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>

#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>

#include <string.h>

#include <sys/backup.h>

#include <sys/modbus.h>

#include <time.h>

#include <sys/srfs.h>

#include <services.h>

#include <sys/farprintf.h>

#include <sys/stat.h>
#include <utime.h>
#include <sys/time.h>

#include <applic.h>



EVENT_DEF(RestoredParamFile);

static ulong timeChkInterval, lastTimeChk, lastDateChk, lastChgChk;

static fd_t  logfd;
static uint  logDay, logMonth;
static ulong  logFileDevID;

static char  logFileName[PATH_MAX];


#define  HdrSize  (sizeof(BBM_Cache_Hdr))


//-----------------------------------------------------------------------------
//  query the SRFS server for the time of day
//-----------------------------------------------------------------------------
void getTime(void)
{
  int  sockfd;
  ulong  diff;
  time_t  curSecs, newSecs;
  uchar  buf[4];
  struct tm  tms;
  struct sockaddr_in  srvAddr;
  IFConfig  curConfig;


  if (msSince(lastTimeChk) < timeChkInterval)  return;

  lastTimeChk = GetMS();
  timeChkInterval = 5000;  // used if we exit this func due to an error

  getIFConfig(IF_ETH0, &curConfig);
  if (curConfig.state != IF_UP)  return;

  errno = 0;

  srvAddr.sin_family = AF_INET;
  srvAddr.sin_port   = htons(TIME_PORT);
  srvAddr.sin_addr   = htonl(srfsIPAddr);

  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)  {
    perror("getTime: socket()");  return;  }

  if (connect(sockfd, (struct sockaddr *) &srvAddr, sizeof(srvAddr)) < 0)  {
    perror("getTime: connect()");  close(sockfd);  return;  }

  if (read(sockfd, buf, 4) != 4)  {
    perror("getTime: read()");  close(sockfd);  return; }

  close(sockfd);

  time(&curSecs);
  newSecs = ntohl(*((ulong *)buf)) - 2208988800L;

  diff = ULDIF(curSecs, newSecs);

  if ((diff > 5) OR !gotTime)  {
    gmtime_r(&newSecs, &tms);
    printf("UTC time: %04d-%02d-%02d %02d:%02d:%02d (%lu)\r\n",
            tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday,
            tms.tm_hour, tms.tm_min, tms.tm_sec, newSecs );
    gotTime = true;
  }

  stime(&newSecs);

//--- use a longer wait time between successful queries ---
  timeChkInterval = 300000L + ((ntohl(srvAddr.sin_addr) & 0x000000ff) << 4);
}


//-----------------------------------------------------------------------------
//  return the file name to use for the specified paramFile entity
//-----------------------------------------------------------------------------
char *paramFileName(ParamFile *paramFile, char *fileName, uint maxLen)
{
  *fileName = 0;


//--- add prefix for device-specific files ---
  if ( (paramFile->category == PDC_DEV) OR (paramFile->category == PDC_SET) )
    sprintf(fileName, "%.12s_x%02lX_", ApplAbbrev, deviceID);

  fstrncat(fileName, paramFile->baseName, maxLen-strlen(fileName)-1);

  return fileName;
}


//-----------------------------------------------------------------------------
//  Return the # of bytes needed to store a copy of the values in the
//  specified list of parameters.
//-----------------------------------------------------------------------------
uint paramStreamSize(ParamFile *paramFile)
{
  uint  count = 0;
  const ParamData far *param;

  for (param=paramFile->params; param->id; ++param)
    count += param->size + 4;  // add 4 bytes for 2-byte ID & 2-byte size

  return count;
}


//-----------------------------------------------------------------------------
//  close the fd for the parameter file (if one is open)
//-----------------------------------------------------------------------------
void closeParamFile(ParamFile *paramFile)
{
  if (paramFile->fd < 0)  return;

  close(paramFile->fd);  paramFile->fd = -1;
}

//-----------------------------------------------------------------------------
//  Allocate BBRAM for the parameter files that may be written to often so we
//  can use it as a cache to minimize the # of writes to flash (yet still
//  have the most recent values if we reboot before the cache is flushed to
//  the flash file).
//
//  Once we have allocated a block, check to see if it contains data from
//  before the reboot (and if it is newer than the flash file)
//-----------------------------------------------------------------------------
int initParamModule(char *buf, uint bufSize)
{
  bool  useCache;
  uint  u, ttlSize, dataSize;
  char far *cache;
  BBM_Cache_Hdr far *hdr;
  ParamFile *paramFile;
  struct stat  fileInfo;
  char fileName[NAME_MAX];


//=== For each file of category PDC_SET (i.e. runtime settings), allocate ===
//=== the space for a cache in BBRAM and see if it contains valid data ===
//=== from the previous run ===

  for (u=0; u<NumParamFiles; ++u)  {
    paramFile = paramFiles[u];

    BZERO(paramFile->writeCount);
    paramFile->fd = -1;
    paramFile->bbmCache = NULL;

    if (paramFile->category != PDC_SET) continue;

    if (paramStreamSize(paramFile) < 1)  continue;

  //--- allocate a cache in BBRAM (if file contains run-time settings) ---
    dataSize = paramStreamSize(paramFile);  ttlSize = dataSize + HdrSize;
    fprintf(DiagLog, "For '%.32s' file cache:\r\n", paramFile->baseName);
    hdr = (BBM_Cache_Hdr far *) bb_xalloc(ttlSize);
    if (!hdr)  {
      printf("Unable to allocate BBRAM cache for %.32s file\r\n",
             paramFile->baseName);
      continue;
    }
    paramFile->cacheSize = dataSize;  paramFile->bbmCache = hdr;
    cache = (char far *)hdr + HdrSize;

  //--- check to see if there is data from last time and if it is newer ---
  //--- than the data in the flash file ---
    useCache = true;
    if (hdr->devType != devType)  useCache = false;
    else if (hdr->size != dataSize)  useCache = false;
    else if (hdr->crc != memCRC16((ulong)cache, dataSize))  useCache = false;
    else {
      paramFileName(paramFile, fileName, sizeof(fileName));
      stat(fileName, &fileInfo);
      if (fileInfo.st_mtime > hdr->time)  useCache = false;
    }

    if (!useCache)  {
      fmemset(hdr, 0, ttlSize);
      hdr->devType = devType;  hdr->size = dataSize;
    }

    printf("%s %u byte BBRAM cache for '%s' at: 0x%05lX\r\n",
           (useCache ? "Found" : "Created"), dataSize,
           paramFile->baseName, (ulong)cache );
  }


//--- load all the values from the parameter files (or their cache) in to ---
//--- the runtime values ---
  loadAllParams(buf, bufSize);

  return 0;
}

//-----------------------------------------------------------------------------
//  Call to prepare a paramaterized list's flash file for writing.
//
//  If the file already exists and is the correct size, it is left alone.
//
//  If the file already exists but is not the correct size, it is deleted
//  and a new empty one is created.
//
//  If the file doesn't exist, a new empty one is created.
//
//  Returns the fd the file was opened on, or < 0 if an error
//-----------------------------------------------------------------------------
int initParamFile(ParamFile *paramFile)
{
  int  fd, status;
  ulong  startTime;
  struct stat  fileInfo;
  char  fileName[NAME_MAX];


  paramFileName(paramFile, fileName, sizeof(fileName));

  fd = paramFile->fd;

//--- open the existing file (if any) ---
  if (fd < 0)  fd = open(fileName, O_WRONLY, 0);
  if (fd < 0)  if (errno != ENOENT)  { perror(fileName);  return -1; }

//--- if the file exists and it is the wrong size, then delete it ---
  if (fd >= 0)  {
    if (fstat(fd, &fileInfo))  {
      perror(fileName);  close(fd);  paramFile->fd = -1;  return -1; }
    if (fileInfo.st_size != paramStreamSize(paramFile))  {
      close(fd);  paramFile->fd = fd = -1;
      // wait for up to 5 seconds to delete an existing file
      for (startTime=GetMS(); msSince(startTime)<5000; )  {
        status = unlink(fileName);  if (!status)  break;
        if (errno == ENOENT)  { status = 0;  break; }
        if (errno != EBUSY)  break;
        sleepMS(50);
      }
      if (status)  { perror(fileName);  return -1; }
    }
  }

//--- create a new empty file if necessary ---
  if (fd < 0)  fd = open(fileName, O_RDWR | O_CREAT, 0666);

  if (fd < 0)  { perror(fileName);  return -1; }

  paramFile->fd = fd;  return fd;
}


//-----------------------------------------------------------------------------
//  Call in preperation for saving a copy of the run-time values in a list
//  to either its BBRAM cache or its flash file.
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int prepForWrites(ParamFile *paramFile)
{
//  if (paramFile->fd >= 0)  { errno = EBUSY;  return -1; }
  if (paramFile->fd >= 0)  closeParamFile(paramFile);

  paramFile->offset = 0;

//--- just use the cache if one exists for this file ---
  if (paramFile->bbmCache)  if (paramFile->bbmCache->time)  return 0;

//--- otherwise get the file ready for writing ---
  if (initParamFile(paramFile) >= 0)  return 0;

  return -1;
}

//-----------------------------------------------------------------------------
//  Call in preperation for reading a copy of the run-time values in a list
//  from either its BBRAM cache or its flash file.
//
//  Returns 0 on sucess, < 0 if an error.
//-----------------------------------------------------------------------------
int prepForReads(ParamFile *paramFile)
{
  char  fileName[NAME_MAX];


//  if (paramFile->fd >= 0)  { errno = EBUSY;  return -1; }
  if (paramFile->fd >= 0)  closeParamFile(paramFile);

  paramFile->offset = 0;

  if (paramFile->bbmCache)  if (paramFile->bbmCache->time)  return 0;

  paramFileName(paramFile, fileName, sizeof(fileName));

  if ((paramFile->fd = open(fileName, O_RDONLY, 0)) < 0)   return -1;

  return 0;
}


//-----------------------------------------------------------------------------
//  Write a block of bytes to the cache or the file copy.
//
//  Returns the number of bytes written or < 0 if an error
//-----------------------------------------------------------------------------
int writeParamBytes(ParamFile *paramFile, char *from, uint count)
{
  char far  *cachePtr, *endCache;
  int  status;
  BBM_Cache_Hdr far  *hdr = paramFile->bbmCache;


  if (!from)  return -1;

  if (hdr)  {
    cachePtr = (uchar far *)hdr + HdrSize;
    endCache = cachePtr + paramFile->cacheSize;
    cachePtr += paramFile->offset;
    if (cachePtr + count > endCache)  { ERRNO = ENOSPC;  return -1; }
    fmemcpy(cachePtr, from, count);
    paramFile->offset += count;
    return count;
  }

  status = write(paramFile->fd, from, count);

  if (status > 0)  paramFile->offset += status;

  return status;
}


//-----------------------------------------------------------------------------
//  Read a block of bytes from the cache or the file copy.
//
//  Returns the number of bytes read or < 0 if an error
//-----------------------------------------------------------------------------
int readParamBytes(ParamFile *paramFile, char *to, uint count)
{
  char far  *cachePtr, *endCache;
  int stat;
  BBM_Cache_Hdr far  *hdr = paramFile->bbmCache;


  if (hdr)  if (hdr->time)  {
    cachePtr = (uchar far *)hdr + HdrSize;
    endCache = cachePtr + paramFile->cacheSize;
    cachePtr += paramFile->offset;
    if (cachePtr + count > endCache)  { ERRNO = ENODATA;  return -1; }
    if (to)  fmemcpy(to, cachePtr, count);
    paramFile->offset += count;
    return count;
  }

  if (to)
    stat = read(paramFile->fd, to, count);
  else
    stat = skipBytes(paramFile->fd, count, 500);

  if (stat > 0)  paramFile->offset += stat;

  return stat;
}


//-----------------------------------------------------------------------------
//  If the specified parameter list has a BBRAM cache, and the cached copy of
//  the data is newer, then copy its contents to the appropriate file.
//-----------------------------------------------------------------------------
void flushParamCache(char *buf, uint bufSize, ParamFile *paramFile)
{
  int  fd;
  uint  count, countIdx;
  char far  *cache;
  BBM_Cache_Hdr far  *hdr;
  struct stat  fileInfo;


  if (!gotTime)  return;

  hdr = paramFile->bbmCache;  if (!hdr)  return;
  if (!hdr->time)  return;

  if ((fd = initParamFile(paramFile)) < 0)   return;

  if (!fstat(fd, &fileInfo) AND
      ((hdr->time > fileInfo.st_mtime)
        OR (fileInfo.st_size != paramStreamSize(paramFile))) )  {
    printf("Flushing '%.24s' to flash...\r\n", paramFile->baseName);

    countIdx = (uint)(GetSeconds() / 60) & 0x0f;  // 15 minute sliding window
    ++paramFile->writeCount[countIdx];

  //--- Copy the contents of the cache to the file and set the time for ---
  //--- the cached copy equal to the modification time of the file ---
    count = hdr->size;  cache = (char far *)hdr + HdrSize;
//  printf("  write %d bytes from 0x%05lX using fd %d\r\n", count, cache, fd);  //tdebug
    if (fwrite_(fd, cache, count, buf, bufSize, 5000) != count)
      perror("flushParamCache()");
    else
      if (!fstat(fd, &fileInfo))  hdr->time = fileInfo.st_mtime;
  }

  closeParamFile(paramFile);
}


//-----------------------------------------------------------------------------
//  Update the specified file (or cache) of parameterized values from the list
//  of runtime in-RAM values.
//-----------------------------------------------------------------------------
int saveParams(char *buf, uint bufSize, ParamFile *paramFile, bool forceFlush)
{
  uchar far  *data;
  uchar  *ptr;
  short  i, id, size, remain, count;
  BBM_Cache_Hdr far  *hdr;
  const ParamData far *params = paramFile->params;


  if (paramStreamSize(paramFile) < 1)  return 0;

  if (!gotTime)  {
    printf("Can't save persistent data until the clock is synchronized.\r\n"
           "Try again later.\r\n");
    return -1;
  }

  if (bufSize < 64)  {
    printf("saveParams(): Insufficient buffer size: %d\r\n", bufSize);
    return -1;
  }

  if (prepForWrites(paramFile))  { perror("saveParams()");  return -1; }

  hdr = paramFile->bbmCache;

  printf("Updating parameter %s: %s\r\n",
         ((hdr AND !forceFlush) ? "cache" : "file"), paramFile->baseName);


  for (i=0; params[i].id; ++i)  {
    id = params[i].id;  size = params[i].size;  data = params[i].data;

  //--- write the id and the size, MSB 1st for both values ---
    ptr = buf;  addShort(&ptr, id);  addShort(&ptr, size);
    if (writeParamBytes(paramFile, buf, 4) != 4)  {
      sprintf(buf, "writeParams(): ID:%d  size:%d (%.24s)",
              id, size, paramFile->baseName);
      perror(buf);  closeParamFile(paramFile);  return -1;
    }

  //--- write the data ---
    remain = size;
    while (remain > 0)  {
      if ((count = remain) > bufSize)  count = bufSize;
      fmemcpy(buf, data, count);
      if (writeParamBytes(paramFile, buf, count) != count)  {
        sprintf(buf, "writeParams(): ID:%d  size:%d (%.24s)",
                id, size, paramFile->baseName);
        perror(buf);  closeParamFile(paramFile);  return -1;
      }
      remain -= count;  data += count;
    }

    if (ShowDebugMsgs)
      fprintf(DiagLog, "wrote param:%d  size:%d  addr: %05lX\n",
              id, size, (ulong)data );
  }


  if (hdr)  {
    hdr->size = paramStreamSize(paramFile);
    hdr->crc = memCRC16((ulong)hdr + HdrSize, hdr->size);
    hdr->time = time(NULL);
  }


  if (forceFlush)  flushParamCache(buf, bufSize, paramFile);

  closeParamFile(paramFile);

  return 0;
}

//-----------------------------------------------------------------------------
//  Read parameterized data for the specified list of values.  For each data
//  element in the list, the value is read from the file or its cache.
//  Depending upon the value of the 'justCompare' flag, each value read is
//  either compared to the current runtime value or overwrites the current
//  runtime value.
//
//  The list of parameters is terminated by a parameter with an id of 0.
//
//  Returns:
//     0 if it reaches EOF and no errors occured
//     < 0 if an error occured
//     > 0 if in-RAM != saved copy (including any values missing from the file)
//-----------------------------------------------------------------------------
int loadParams(char *buf, uint bufSize, ParamFile *paramFile, bool justCompare)
{
#define  MaxParams 128
  int  status;
  bool  useCache = false;
  short  i, idx, id, size, extra, minRead, ttl, count, remain, numParams;
  uchar  *ptr;
  char far  *curVal;
  uint  offset;
  BBM_Cache_Hdr far *hdr;
  const ParamData far *param, *params = paramFile->params;
  bool verified[(MaxParams+7)/8];


#define  SetVerified(x)  ( verified[(x)>>3] |= (1 << ((x) & 7)) )
#define  IsVerified(x)   ( verified[(x)>>3] & (1 << ((x) & 7)) )

  if (paramStreamSize(paramFile) < 1)  return 0;

  if (bufSize < 64)  {
    printf("loadParams(): Insufficient buffer size: %d\r\n", bufSize);
    return -1;
  }

  if (prepForReads(paramFile))  {
    if (errno != ENOENT)  perror("loadParams()");
    return -1;
  }

  //fix:  Move this to BEFORE prepForReads()
//--- determine how many parameters there are ---
  for (numParams=0; params[numParams].id; ++numParams);
  if (numParams > MaxParams)  {
    printf("loadParams(): Max # parameters exceeded\r\n");  return -1; }

  BZERO(verified);

  hdr = paramFile->bbmCache;
  if (hdr)  if (hdr->time)  useCache = true;


  if (!justCompare)
    printf("\r\nreloading '%.24s' from %s\r\n",
           paramFile->baseName, (useCache ? "BBRAM cache" : "Flash") );


//--- For each value in the file, read it and (if doing a compare only) ---
//--- determine if it matches the current in-RAM value ---
  for (status=0;;)  {

  //--- get id and size of next parameter ---
    offset = paramFile->offset;
    if (readParamBytes(paramFile, buf, 4) != 4)  break;

    ptr = buf;  id = getShort(&ptr);  size = getShort(&ptr);

    if ((id <= 0) OR (size <= 0))  {
      printf("loadParams(): offset: %u  ID:%d  size:%d (%.24s)\r\n",
             offset, id, size, paramFile->baseName);
      printf("*** invalid ID or size\r\n");
      status = -1;  break;
    }

  //--- see if the parameter just read is one of the ones in the list ---
    param = NULL;
    for (i=0; params[i].id; ++i)
      if (params[i].id == id)  { param = params + i;  idx = i;  break; }

  //--- if not in the list, just skip it ---
    if (! param)  {
      if (!justCompare)  {
        printf("loadParams(): offset: %u  ID:%d  size:%d (%.24s)\r\n",
               offset, id, size, paramFile->baseName);
        printf("skipping unknown param\r\n");
      }
      if (readParamBytes(paramFile, NULL, size) < size)  {
        printf("loadParams(): offset: %u  ID:%d  size:%d (%.24s)\r\n",
               offset, id, size, paramFile->baseName);
        perror(NULL);  status = -1;  break;
      }
      continue;  // successfully skipped this param
    }

  //--- Set size to smaller of provided size/expected size and set the ---
  //--- minimum # of bytes we must read/copy each time to avoid major ---
  //--- corruption of the values ---
    extra = 0;
    if (param->size < size)  {
      extra = size - param->size;  size = param->size; }
    if ((minRead = param->gran) < 1)  minRead = size;
    if (size <= bufSize)  minRead = size;

  //--- Is the smallest logical-piece (defined by the granularity) too big ---
  //--- for our buffer? (could read in pieces, but then we would be ---
  //--- taking a chance that we might not get the entire value, which ---
  //--- would corrupt the state of the parameter we were reading)
    if (minRead > bufSize)  {
      printf("loadParams(): offset: %u  ID:%d  size:%d (%.24s)\r\n",
             offset, id, size, paramFile->baseName);
      printf("*** granularity too large: %d\r\n", minRead);
      status = -1;  break;
    }

  //--- read the value (in safe-size chunks if necessary) ---
    count = minRead;  remain = size;  ttl = 0;
    if (!justCompare)  printf("  restoring: %ls\r\n", param->name);

    while (remain)  {
      if (ShowDebugMsgs)
        printf("loadParams(): offset: %u  ID:%d  size:%d (%.24s)\r\n",
               offset, id, size, paramFile->baseName);

      if (count > remain)  count = remain;

      if (readParamBytes(paramFile, buf, count) != count)  {
        if (!justCompare)  {
          printf("loadParams(): offset: %u  ID:%d  size:%d (%.24s)\r\n",
                 offset, id, size, paramFile->baseName);
          perror(NULL);
        }
        status = -1;  break;
      }
      if (status)  break;

    //--- if just comparing in-memory to saved copy, then simply return ---
    //--- a positive value if they don't match ---
      curVal = (char far *)param->data + (uint)ttl;
      if (justCompare)  {
        if (fmemcmp(curVal, buf, count) != 0)  {
          printf("Runtime value of %.24s.%ls has changed\r\n",
                 paramFile->baseName, param->name);
          //fix:  do:  status = idx + 1;  break;  instead of a return!
          return idx + 1;
        }
      }
      else  {
    //--- Read a self-contained chunk without an error, so it is safe to ---
    //--- overwrite the existing value with the one just read.  If the value ---
    //--- has a restore func, then call it.  If not, check to see if it is ---
    //--- a externally accessible (register/PV) value.  If neither of those ---
    //--- is the case, then just copy the new value. ---
        if (param->restoreVal)  // value-specific handling of new values?
          (*param->restoreVal)(curVal, buf, count);
        else
          if (!restoreIfRegValue(curVal, buf, count))  { // reg value?
            KLOCK();  fmemcpy(curVal, buf, count);  KUNLOCK();  }
      }

      ttl += count;  remain -= count;
    }

    SetVerified(idx);

    if (extra)  if (readParamBytes(paramFile, NULL, extra) < extra)  {
      printf("loadParams(): offset: %u  ID:%d  size:%d (%.24s)\r\n",
             offset, id, size, paramFile->baseName);
      perror(NULL);  status = -1;  break;
    }

    if (ShowDebugMsgs)  printf("read param:%d size:%d\r\n", id, size);
  }

  closeParamFile(paramFile);

  if (!justCompare)  printf("\r\n");

  //fix:  do  if (status)  return status;  instead
  if (status < 0)  return status;

//--- Got through all the values in the file without finding any that ---
//--- didn't match the ones in memory, but we still need to make sure ---
//--- that all the ones in memory were in the file ---
  for (i=0; i<numParams; ++i)  if (!IsVerified(i))  return i+1;

  status = 0;

  return status;
}


//-----------------------------------------------------------------------------
//  If the specified parameter list has a BBRAM cache, then update it as
//  necessary and then determine if the file for the same list is also in
//  need of updating.  If the file has older data than the cache, and we
//  haven't updated much recently, then do so now.
//-----------------------------------------------------------------------------
void checkCache(char *buf, uint bufSize, ParamFile *paramFile)
{
  bool  updateFile;
  uint  ui, countIdx, numWrites;
  int  fd;
  ulong  now;
  BBM_Cache_Hdr far  *hdr;
  struct stat  fileInfo;


  if (!gotTime)  return;

  hdr = paramFile->bbmCache;  if (!hdr)  return;

//--- limit checks to one of the affected files every 5 seconds in a ---
//--- round-robin fashion to minimize the overhead incurred ---
  if (msSince(lastChgChk) < 5000)  return;
  lastChgChk = GetMS();

//printf("--- checking on [%.20s] ---\r\n", paramFile->baseName);  //tdebug

//--- if the run-time values have changed then update the cache copy ---
  if (loadParams(buf, bufSize, paramFile, JustCompare) > 0)  {
    saveParams(buf, bufSize, paramFile, DontFlush);  return; }


  if ((fd = initParamFile(paramFile)) < 0)  return;

  if (fstat(fd, &fileInfo))  {
    perror("checkCache()");  closeParamFile(paramFile);  return; }

//--- 16 counters, one always set to 0, used in a round-robin fashion ---
//--- to keep track of # of writes over a 15-minute sliding window ---
  countIdx = (uint)(GetSeconds() / 60) & 0x0f;  // 15 minute sliding window
  paramFile->writeCount[(countIdx + 1) & 0x0f] = 0;  // 16th pos always 0

  now = time(NULL);

  for (;;)  {
    updateFile = true;

  //--- if the file is not the right size, then update it.  This includes ---
  //--- new files which may have a newer time than the cached data. ---
    if (fileInfo.st_size != hdr->size)  break;

  //--- The cached data must be stable for at least 60 secs before flushing ---
    if (now - hdr->time < 60)  { updateFile = false;  break; }

  //--- if the file copy is newer, then don't update it ---
    if (fileInfo.st_mtime >= hdr->time)  { updateFile = false;  break; }
//    printf("  file time: %lu  cache time: %lu\r\n", fileInfo.st_mtime, hdr->time);  //tdebug

  //--- Don't flush more often than once every 15 minutes ---
    for (ui=0, numWrites=0; ui<16; ++ui) numWrites += paramFile->writeCount[ui];
//    printf("  file '%s':  fd: %d  numWrites: %d,  age: %ld\r\n",  //tdebug
//       paramFile->baseName, fd, numWrites, (time(NULL) - fileInfo.st_mtime));  //tdebug
    if (numWrites)  updateFile = false;

    break;
  }

  closeParamFile(paramFile);

//--- copy the contents of the cache to the file ---
  if (updateFile)  flushParamCache(buf, bufSize, paramFile);
}



//-----------------------------------------------------------------------------
//  Load values from all the parameter files
//-----------------------------------------------------------------------------
void loadAllParams(char *buf, uint bufSize)
{
  for (uint i=0; i<NumParamFiles; ++i)
    loadParams(buf, bufSize, paramFiles[i], Reload);
}


//-----------------------------------------------------------------------------
//  Send a msg to the applic process to reload all the parameter files
//-----------------------------------------------------------------------------
int loadAllParamsMsg(void)
{
  uint  i;
  fd_t  eventsfd;
  EventMsg  eventMsg;


  if ((eventsfd = open(EventsFIFO, O_WRONLY, 0)) < 0)  return -1;

  for (i=0; i<NumParamFiles; ++i)  {
     *(uint *)(eventMsg.data) = i;
     EVENT(RestoredParamFile, 0, NULL, sizeof(uint));
  }

  close(eventsfd);  return 0;
}

//-----------------------------------------------------------------------------
//  Send a msg to the applic process to reload a specific ParamFile
//-----------------------------------------------------------------------------
int loadParamsMsg(ParamFile *paramFile)
{
  uint  i;
  fd_t  eventsfd;
  EventMsg  eventMsg;


  if ((eventsfd = open(EventsFIFO, O_WRONLY, 0)) < 0)  return -1;

  for (i=0; i<NumParamFiles; ++i)  {
    if (paramFiles[i] != paramFile)  continue;
     *(uint *)(eventMsg.data) = i;
     EVENT(RestoredParamFile, 0, NULL, sizeof(uint));
  }

  close(eventsfd);  return 0;
}


//-----------------------------------------------------------------------------
//  Save values for all the parameter files
//-----------------------------------------------------------------------------
void saveAllParams(char *buf, uint bufSize)
{
  for (uint i=0; i<NumParamFiles; ++i)
    saveParams(buf, bufSize, paramFiles[i], ForceFlush);
}


//-----------------------------------------------------------------------------
//  Flush any caches with data that is newer than the corresponding file
//-----------------------------------------------------------------------------
void flushAllParams(char *buf, uint bufSize)
{
  for (uint i=0; i<NumParamFiles; ++i)
    flushParamCache(buf, bufSize, paramFiles[i]);
}


//-----------------------------------------------------------------------------
//  set the initial values for logDay and logMonth
//-----------------------------------------------------------------------------
void setLogDayMonth(void)
{
  time_t  dateTime;
  struct tm  tms;

  time(&dateTime);  localtime_r(&dateTime, &tms);

  logDay = (uint)tms.tm_mday;  logMonth = (uint) tms.tm_mon + 1;
}

//-----------------------------------------------------------------------------
//  return the current day of the month (1-31) using local time
//-----------------------------------------------------------------------------
uint  dayOfMonth(void)
{
  time_t  dateTime;
  struct tm  tms;

  time(&dateTime);  localtime_r(&dateTime, &tms);

  return (uint) tms.tm_mday;
}

//-----------------------------------------------------------------------------
//  (re)open the event log file.
//
//  On return, the global values logfd and logFileDevID are set (logfd is < 0
//  if an error occured).
//-----------------------------------------------------------------------------
void reopenLogFile(char *buf)
{
  uint  len;
  ulong  hoursOld;
  bool  err;
  struct timeval  tv;
  struct stat  finfo;
  char  temp[32];


  err = false;  logFileDevID = deviceID;


//===== if no open file descriptor then open one now =====
  if (logfd < 0)  do  {
  //--- set name for log file ---
    setLogDayMonth();
    sprintf(logFileName, "/srfs/%.10s_x%02lX_%02u-%02u.log",
            ApplAbbrev, logFileDevID, logMonth, logDay);

  //--- if file already exists, and is > 72 hours old, then delete it ---
    if (stat(logFileName, &finfo) == 0)  {
      hoursOld = (time(NULL) - finfo.st_mtime) / 3600;
      if (hoursOld > 72)  {
        printf("deleting log file: %s (%ld hours old)\r\n",
               logFileName, hoursOld);
        unlink(logFileName);
      }
    }

  //--- open file for appending, creating it if it doesn't exist yet ---
    if ((logfd = open(logFileName, O_WRONLY | O_CREAT | O_APPEND, 0666)) < 0) {
      err = true;  break; }

    printf("opened log file: %s\r\n", logFileName);

    gettimeofday(&tv, NULL);

    sprintf(buf,
            "\r\n\n  ======== Module %s opened log file at %s ========\r\n\n",
            moduleID, dateTimeMS(temp, tv.tv_sec, (uint)(tv.tv_usec/1000)));
    len = strlen(buf);

    if (write(logfd, buf, len) != len)  err = true;

    sprintf(buf, "  RTOS: %ls  %ls\r\n", osDTM, osName);  len = strlen(buf);
    if (write(logfd, buf, len) != len)  err = true;

    sprintf(buf, "  APPL: %ls  %ls\r\n", applDTM, applName);  len = strlen(buf);
    if (write(logfd, buf, len) != len)  err = true;

  }  while(0);


  if (err)  {
    perror(logFileName);
    if (logfd >= 0)  { close(logfd);  logfd = -1; }
    sleepMS(5000);  return;
  }

}


//-----------------------------------------------------------------------------
//  task to keep backup copies of device data up to date, and restore missing
//  or outdated local copies from backup
//-----------------------------------------------------------------------------
void _task backupd(int argc, char **argv)
{
  uint  ui, idx, dirLen;
  uint  toIdx, nextIdx;
  int  copied;
  time_t  fileTime;
  ulong  lastBackupCheck, terminateTime;
  fd_t  localfd, remotefd, eventsfd;
  sigset_t  sigSet;
  ParamFile  *paramFile;
  char  *errStr, *localName;
  struct stat  localInfo, remoteInfo;
  struct utimbuf  fileTimes;
  EventMsg  eventMsg;
  struct tm  tms;
  char  fileName[PATH_MAX], temp[PATH_MAX], dataBuf[128];
  static uint  fileIdx = 0;


  redirectStdio();  // so printf, perror, etc don't end up in data stream


//--- block Terminate signal so we can shut down cleanly if needed ---
  sigSet = sigmask(SIGTERM);  sigprocmask(SIG_BLOCK, &sigSet, NULL);


//finish:  insure only one copy of this process can run!

  lastTimeChk = lastChgChk = 0;
  timeChkInterval = 2000;  // made much longer after successful query


  if ((eventsfd = open(EventsFIFO, O_WRONLY, 0)) < 0)  {
    perror(EventsFIFO);  Exit(-1);  }

  strcpy(fileName, "/srfs/");
  dirLen = strlen(fileName);  localName = fileName + dirLen;

  localfd = remotefd = logfd = -1;  setLogDayMonth();

  lastBackupCheck = lastDateChk = GetMS();  terminateTime = 0;


//===== several chores to tend to in a loop =====
  for (;;)  {

  //--- check for pending Terminate signal ---
    if (sigpending(&sigSet) == 0)  if (sigismember(&sigSet, SIGTERM))
      if (! terminateTime)  {
        printf("cleaning up...\r\n");
        terminateTime = GetMS();
        flushAllParams(dataBuf, sizeof(dataBuf));
      }

  //--- if terminate signal rcvd, wait 5 secs to allow log data to be saved ---
    if (terminateTime AND (msSince(terminateTime) > 5000))  {
      if (eventsfd >= 0)  close(eventsfd);
      if (logfd >= 0)  close(logfd);
      if (localfd >= 0)  close(localfd);
      if (remotefd >= 0)  close(remotefd);
      Exit(-1);
    }

    sleepMS(20);

  //--- periodically get time from SRFS server ---
    getTime();

  //--- check for changes to persistent values ---
    fileIdx = (fileIdx + 1) % NumParamFiles;
    checkCache(dataBuf, sizeof(dataBuf), paramFiles[fileIdx]);


  //================================================
  //===== keep a remote copy of syslog entries =====
  //================================================

    if (gotTime)  {

    //--- if date changed, then close the current log file ---
      if (msSince(lastDateChk) >= 500)  {
        if (logDay != dayOfMonth())  {
          printf("***** new day *****\n");
          if (logfd >= 0)  { close(logfd);  logfd = -1; }
        }
        lastDateChk = GetMS();
      }

    //--- if deviceID changed, then close the current log file ---
      if ((logFileDevID != deviceID) AND (logfd >= 0))  {
          close(logfd);  logfd = -1; }

    //--- reopen log file if needed ---
      if (logfd < 0)  reopenLogFile(dataBuf);

    //--- copy any new data in the syslog buffer to a remote file ---
      if (logfd >= 0)  {
        _EnterCritical_();
        toIdx = diagLogInfo->toIdx;  nextIdx = diagLogInfo->fromIdx;
        _ExitCritical_();
        for (ui=0; (nextIdx!=toIdx) AND (ui<sizeof(dataBuf)); ++ui)  {
          dataBuf[ui] = diagLogBuf[nextIdx];
          if (++nextIdx == DiagLogSize)  nextIdx = 0;
        }
        if (ui)  {
          copied = write(logfd, dataBuf, ui);
          if (copied < 0)  {
            perror(logFileName);  close(logfd);  logfd = -1;  }
          if (copied > 0)  {
            KLOCK();
            if ((diagLogInfo->fromIdx += copied) >= DiagLogSize)
              diagLogInfo->fromIdx -= DiagLogSize;
            KUNLOCK();
          }
        }
      }

    }  // end: if (gotTime)


  //=========================================================================
  //===== automatically backup or restore the device configuration file =====
  //===== as needed =====
  //=========================================================================
    if (msSince(lastBackupCheck) < 30000)  continue;

    lastBackupCheck = GetMS();

    for (idx=0; idx<NumParamFiles; ++idx)  {

    //--- close any open file descriptors ---
      if (localfd >= 0)  { close(localfd);  localfd = -1; }
      if (remotefd >= 0)  { close(remotefd);  remotefd = -1; }

      paramFile = paramFiles[idx];


    //--- For device-specific data, compare the timestamp on the local file ---
    //--- with the remote backup copy.  If one is newer than the other, ---
    //--- then update the one that is older ---
      if (paramFile->category != PDC_DEV)  continue;

    //--- generate unique name using base name, device type, and device id ---
      paramFileName(paramFile, localName, sizeof(fileName)-dirLen);

    //--- get date/time and size info for local and remote copy of the file ---
      memset(&localInfo, 0, sizeof(localInfo));
      memset(&remoteInfo, 0, sizeof(remoteInfo));

      if ((localfd = open(localName, O_RDONLY, 0)) >= 0)
        if (fstat(localfd, &localInfo))  {
          perror(localName);  sleepMS(5000);  continue; }

      if ((remotefd = open(fileName, O_RDONLY, 0)) >= 0)  {
        if (fstat(remotefd, &remoteInfo))  {
          perror(fileName);  sleepMS(5000);  continue; } }
      else
        if (errno != ENOENT)  { perror(fileName);  sleepMS(5000);  continue; }

      if (ShowDebugMsgs)  {
        printf("Checking file: %s\r\n", localName);
        printf("  local:  fd: %d, mtime: %lu  size: %lu\r\n",
               localfd, localInfo.st_mtime, localInfo.st_size);
        printf("  remote: fd: %d, mtime: %lu  size: %lu\r\n",
               remotefd, remoteInfo.st_mtime, remoteInfo.st_size);
      }


    //=== if backup copy of device-specific data is newer, then restore ===
    //=== the backup copy ===
      if ((remotefd >= 0) AND (localInfo.st_mtime < remoteInfo.st_mtime))  {
        printf("restore from backup: %s\r\n", fileName);
        if (localfd >= 0)
          printf("  local:  %.24s  size: %lu\r\n",
                 ctime_r(&localInfo.st_mtime, dataBuf), localInfo.st_size);
        printf(" remote:  %.24s  size: %lu\r\n",
               ctime_r(&remoteInfo.st_mtime, dataBuf), remoteInfo.st_size);

      //--- delete any old local archive ---
        strcpy(temp, localName);  strcat(temp, ".bak");  unlink(temp);

      //--- rename any existing local as a backup ---
        if (localfd >= 0)
          if (rename(localName, temp))  {
            perror(localName); sleepMS(5000); continue; }

      //--- copy remote-backup to new local file ---
      if (copyFileByName(fileName, localName, false,
                         dataBuf, sizeof(dataBuf), &errStr))  {
        perror(errStr);  sleepMS(5000);  continue; }

      //--- set date/time on restored file to match backup copy ---
        fileTimes.actime = fileTimes.modtime = remoteInfo.st_mtime;
        utime(localName, &fileTimes);

      //--- send restored-param-file msg to applic process ---
        *(uint *)(eventMsg.data) = idx;
        EVENT(RestoredParamFile, 0, NULL, sizeof(uint));

        continue;
      }


    //=== if local copy of parameter data is newer, then update the ===
    //=== remote backup copy ===
      if (localfd >= 0)  if (localInfo.st_mtime > remoteInfo.st_mtime)  {
        printf("update backup copy: %s\r\n", fileName);
        printf("  local:  %.24s  size: %lu\r\n",
               ctime_r(&localInfo.st_mtime, dataBuf), localInfo.st_size);
        if (remotefd >= 0)
          printf("  remote: %.24s  size: %lu\r\n",
                 ctime_r(&remoteInfo.st_mtime, dataBuf), remoteInfo.st_size);

      //--- rename remote-backup copy if one exists ---
        if (remotefd >= 0)  {
          fileTime = remoteInfo.st_mtime;  fileTime += (long)(timeZone * 3600);
          gmtime_r(&fileTime, &tms);
          strcpy(temp, fileName);
          sprintf(temp+strlen(temp), "_%04d-%02d-%02d_%02d%02d%02d",
                  tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday,
                  tms.tm_hour, tms.tm_min, tms.tm_sec);
          if (rename(fileName, temp) < 0)  {
            perror(fileName);  sleepMS(5000);  continue; }
        }

      //--- copy local to new remote-backup ---
        if (copyFileByName(localName, fileName, false,
                           dataBuf, sizeof(dataBuf), &errStr))  {
          printf("copyFileByName(): ");
          perror(errStr);  sleepMS(5000);  continue;
        }

      //--- get date/time for new backup copy and set local to match ---
        remoteInfo.st_mtime = fileTimes.modtime;
        for (ui=0; ui<10; ++ui)  {
          if (stat(fileName, &remoteInfo) == 0)  break;
          printf("stat(): ");  perror(fileName);  sleepMS(1000);
        }
        fileTimes.actime = fileTimes.modtime = remoteInfo.st_mtime;
        utime(localName, &fileTimes);

      //--- "thin-out" remote archive ---
        //finish: don't let # remote backup files expand forever...
        continue;
      }

    }

    if (localfd >= 0)  { close(localfd);  localfd = -1; }
    if (remotefd >= 0)  { close(remotefd);  remotefd = -1; }
  }

}

//-----------------------------------------------------------------------------
