//----- configd.c ----- 10/18/13 -----(08/24/06)-----

/*-----------------------------------------------------------------------------
  Copyright (c) 2004-2010, NSCL and Mark Davis.  All rights reserved.
 ------------------------------------------------------------------------------
  Process to manage the auto configuration process.

  The basic operation of this process is as follows:

  - Periodically send broadcast packets to announce the presence and the
    configuration of the controller

  - Listen for and process Announcement packets sent by other controllers
    looking for conflicts

  - Listen for and process Configuration packets sent by the control system
    and maintenance/diagnostic applications

 ------------------------------------------------------------------------------
  Version log:

  2013-10-18 - Mark Davis (*)
               - Changed MaxPktSize back to 400.  Prior to this, the configd
                 task would occasionally overflow its stack (at least on
                 the digital PhaseCtlr modules).

  2013-08-28 - Mark Davis
               - Changed name of PID_UPSECS to PID_UPSECS_HARD and added
                 PID_UPSECS_SOFT and PID_FIRMWARE

  2012-03-23 - Mark Davis
               - Changed NameSize from 16 to HostNameLen.  Prior to this, the
                 max length hostname that could be handled was 15 characters.
               - Added MaxPktSize macro and set it to 512 (max prior to this
                 was 400).

  2010-03-12 - Mark Davis
               - Now checks the state of the network interface before bothering
                 to try to use any socket connections.  This eliminates
                 pointless and frequent messages in the log.

  2010-03-03 - Mark Davis
               - Changed so UDP socket is normally set to non-blocking mode,
                 changing it to blocking mode only when sending a packet.

  2010-03-02 - Mark Davis
               - Changes to use a single UDP broadcast socket for sending and
                 receiving all packets.

  2009-11-05 - Mark Davis
               - added resetAnnounceInterval()
               - The kernel task is now responsible for calling this function
                 when the DeviceID changes (rather than this module being
                 responsible for detecting thd change).

  2009-08-13 - Mark Davis
               - extractParams() now returns -1 if it encounters an invalid
                 mode or parameter ID so the caller doesn't try to process
                 a corrupt or incomplete packet.

  2009-05-04 - Mark Davis
               - The announcement packet interval is now set to 500ms rather
                 than 0 in response to a terminate signal.

  2009-04-28 - Mark Davis
               - The announcement packet interval is now reset when we get a
                 terminate signal, causing the controller to send one or
                 more packets as it is shutting down.

  2009-04-15 - Mark Davis
               - Added PID_UPSECS, providing the # of seconds the device has
                 been running since it was last booted

  2009-04-14 - Mark Davis
               - Added PID_CMD_STR, providing ability to execute a command
                 via the config service

  2009-03-23 - Mark Davis
               - Changed max interval between announcement packets from 120
                 seconds to 60 seconds.
               - A message is now generated when the announcement interval
                 is reset to the minimum value if the appropriate "diag msgs"
                 bit is set
               - Added checking of announce packets for conflicts with this
                 controller's configuration.

  2009-03-11 - Mark Davis
               - Disabled changes made to support using different ports for
                 sending and receiving broadcasts when I realized that ARP
                 has nothing to do with this, and the change eliminated the
                 possibility of having controllers check for conflicts
                 themselves.

  2009-02-20 - Mark Davis
               - Changed to use a different port to send announce packets
                 so controllers don't process all the announce packets from
                 other controllers (an approach that caused lots of thrashing
                 of the small ARP tables in the controllers and generated lots
                 of ARP traffic on the network).
               - Changed to NOT ask for the from addr for broadcast packets
                 unless diagnostic info is requested (to avoid thrashing of
                 the ARP table)

  2008-01-11 - Mark Davis
               - devCfg table now initialized on startup

  2007-10-02 - Mark Davis
               - increased max packet size from 128 bytes to 400 bytes

  2007-08-03 - Mark Davis
               - added PID_APPL_DTS parameter, PID_RTOS_DTS

  2007-02-02 - Mark Davis
               - added some checks for invalid values

  2006-09-05 - Mark Davis
               - adapted from earlier autoConfig.c version

-----------------------------------------------------------------------------*/


#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <sys/socket.h>

#include <fcntl.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdio.h>
#include <services.h>

#include <string.h>

#include <time.h>

#include <tsh.h>

#include <busybox.h>



#ifdef RABBIT
  #include <sys/farprintf.h>
  #include <SHD_DLM.h>
  extern ulong  applVersTime, rtosVersTime, srfsIPAddr;
#endif


#define  ShowProcess  (DiagFlags & 0x00000001)
#define  ShowPackets  (DiagFlags & 0x00000002)



/*----------------- Platform-specific definitions ---------------*/

#define  MaxDevices  1  /* max # devices the config service can handle */


typedef  struct in_addr      INET_ADDR;
typedef  struct sockaddr    *SAP;
typedef  struct sockaddr_in  SAI;

typedef  struct ether_addr  EtherAddr;


#define  SleepSec(sec)                sleep(sec)
#define  SleepMS(ms)                  sleepMS(ms)
#define  Socket(domain, type, prot)   socket(domain, type, prot)
//#define  Close(sock)                  close(sock)


#ifdef RABBIT
  #define  Sethostname(nam)  SetHostName((nam), strlen(nam))
  #define  FAR  far
  #define  SIN_ADDR  sin_addr
#else  //finish
  #define  SIN_ADDR  sin_addr.s_addr
  #define  FAR
  static short  timeZone = -4;
  static short  devType = 1;
  static long  deviceID = 0x9999;
  static ulong  DiagFlags = ErrorBits;

  typedef struct {
    ulong   ipAddr;   // in host byte order
    ulong   bcAddr;   // in host byte order
    ulong   netmask;  // in host byte order
    uchar   macAddr[6];
    uint    mtu;
    uchar   state;   //  IF_DOWN, IF_COMING_UP, IF_UP, IF_COMING_DOWN
  } IFConfig;


  static int getIFConfig(int iface, IFConfig *ifConfig)
  {
    memset(ifConfig, 0, sizeof(IFConfig));

    //finish
    ifConfig->state = 2;
    inet_ptoh(AF_INET, "192.168.4.34", &ifConfig->ipAddr);
    inet_ptoh(AF_INET, "192.168.255.255", &ifConfig->bcAddr);
    inet_ptoh(AF_INET, "255.255.0.0",     &ifConfig->netmask);
    ifConfig->mtu = 1500;

    return 0;
  }

  int setIFConfig(int iface, IFConfig *ifConfig)  { return 0; }

  void bringNetUp(void)  {  }
  void startNetDown(void)  {  }

  #define  IF_ETH0  0

  #define  Sethostname(nam)  {  }

  #define  CFG_CMD_PORT    2001
//  #define  CFG_REPLY_PORT  2005

#endif

/*--------------------------------------------------------------*/



#define  PktFmtVers     1

/*--- max time between announcement packets ---*/
#define  MaxInterval    60000L  /* Max age for event data (60 seconds) */
#define  MinCmdInterval  5000   /* Time between cmds for non-responding ctlr */

#define DefaultIP  0xC0A8FAFA   /* 192.168.250.250 */


/*------------
  Each NSCL configuration packet starts with a 4 byte header that specifies
  the packet-format version and type of packet.  The header is followed by a
  series of parameterized values:

    pktVers:  2 bytes
    pktType:  2 bytes  (ANOUNCE, SET, or QUERY)
    ------------------
    paramID:  2 bytes
    valSize:  2 bytes
    value:    valSize bytes
    ------------------
    ... repeat parameterized values as needed ...

    A parameter with an id and length of 0 is used to end the list.

    For ANNOUNCE packets, all parameter values are the current values in use
    by the controller that sent the packet.

    For QUERY packets, all parameter values specify a condition.  Each
    controller whose values match the given ones is expected to respond with
    an ANNOUNCE packet giving its current in-use values.  NOTE: In the
    future I may refine this to specify what values should be included in the
    response (probably by including parameters without values to specify
    which values are to be in the response packet).

    For SET packets, parameter values can be part of the settings or part of
    the conditions that specify which controllers are the target for the new
    settings.  In fact, the same parameter could be included in both the
    settings and the conditions with different values.

    To control which parameter values are conditions and which are settings,
    a special MODE parameter is provided.  A MODE parameter specifies the
    context of all parameters received after it (up to the next MODE parameter).

------------*/

/*--- values for "pktType" ---*/
/*--- tells client what to do for each parameter ---*/
#define  PT_ANNOUNCE  1   /* simply announcing an active configuration */
#define  PT_SET       2   /* apply values where relevant */
#define  PT_QUERY     3   /* request annnouncement pkts from 1 or more ctlrs */

static char *pktTypeName[3] = { "ANNOUNCE", "SET", "QUERY" };

/*--- values for "paramID" ---*/
#define  PID_END_PARAMS    0  /* indicates end of params for a pkt */
#define  PID_MODE          1  /* sets context for params until next PID_MODE */
#define  PID_DEVTYPE       2
#define  PID_DEVID         3
#define  PID_MACADDR       4
#define  PID_IPADDR        5
#define  PID_NETMASK       6
#define  PID_GATEWAY       7
#define  PID_NAME          8
#define  PID_TIMEZONE      9
#define  PID_APPL_DTS     10  /* Application Date/Time Stamp */
#define  PID_RTOS_DTS     11  /* RTOS Date/Time Stamp */
#define  PID_SRFS_ADDR    12  /* IPv4 Addr of SRFS server */
#define  PID_CMD_STR      13  /* cmd to execute (PT_SET packets only) */
#define  PID_UPSECS_HARD  14  /* # of secs since last hard (re)boot */
/* Added 08/28/13 */
#define  PID_UPSECS_SOFT  15  /* # of secs since last soft restart */
#define  PID_FIRMWARE     16  /* primary/secondary boot image files info */
/* NOTE:  paramID values must not exceed 31 !!! */

/*--- indexes in to DeviceConfig.event[] ---*/
#define  LAST_SEEN           0   /* time we last heard from controller */
#define  CTLR_ID_CONFLICT    1   /* multiple ctlrs using same type+ID */
#define  CTLR_ADDR_CONFLICT  2   /* multiple ctlrs using same IP Addr */
#define  CTLR_NAME_CONFLICT  3   /* multiple ctlrs using same name */
#define  CSYS_ID_CONFLICT    4   /* typ+ID conflict in ctl sys database */
#define  CSYS_ADDR_CONFLICT  5   /* IP Addr conflict in ctl sys database */
#define  CSYS_NAME_CONFLICT  6   /* ctlr name conflict in ctl sys database */
/*-------------------------------*/
#define  NumEvents           7

static char *EventName[NumEvents] = {
  "LAST SEEN",
  "CTLR ID", "CTLR ADDR", "CTLR NAME",
  "CSYS ID", "CSYS ADDR", "CSYS NAME"
};


/*--- values for PID_MODE parameter ---*/
#define  MODE_ANNOUNCE  0
#define  MODE_MATCH     1
#define  MODE_SET       2


/*--- values for PID_FIRMWARE parameter ---*/
#define  FW_HAS_BACKUP        0x0001
#define  FW_RUNNING_PRIMARY   0x0002
#define  FW_PRIMARY_NEWER     0x0004
#define  FW_BACKUP_NEWER      0x0008


#define  MaxPktSize  400   // This is on the stack so can't be larger!!!

//#define  NameSize  16
#define  NameSize  HostNameLen

typedef struct {
  ulong  time;           /* time packet was rcvd */
  ulong  sentCmdTime;    /* time cmd pkt was sent to deal with conflict */
  short  devType;        /* dev type from sender */
  ulong  devID;          /* dev id from sender */
  uchar  macAddr[6];     /* MAC addr of sender */
  ulong  ipAddr;         /* ip addr of sender */
  ulong  netmask;        /* netmask in use by sender */
  ulong  gateway;        /* gateway in use by sender */
  char   name[NameSize]; /* name used by sender */
  short  timeZone;       /* offset (in hours) from UMT time */
  ulong  paramMask;      /* bitmask: which params were in the pkt */
  ulong  applTime;       /* date/time stamp for application code */
  ulong  rtosTime;       /* date/time stamp for RTOS code */
  ulong  srfsAddr;       /* IPv4 addr of SRFS server */
  char  *cmdStr;         /* a string command */
}  EventData;

typedef struct {
  bool   inUse;           /* this slot is in use */
  char   name[NameSize];  /* name associated with target device */
  short  devType;         /* device type of target device */
  ulong  devID;           /* device ID of target device */
  ulong  ipAddr;          /* IP addr controller should be using */
  ulong  netmask;         /* netmask controller should be using */
  int    plcNum;          /* handle assigned by modrdrTaskCreatePLC() */
  ulong  timeAdded;       /* time device was added to the list */
  EventData  event[NumEvents];  /* timestamp/etc for various events */
}  DeviceConfig;



static int setAddr(int *sockfd,
                   ulong ipAddr, EventData FAR *event, uchar *buf, short bufSize);

static int sendQuery(DeviceConfig FAR *dc, uchar *buf, short bufSize);

static bool conditionsMatch(EventData FAR *matchVals);


/*---------------------------------------------------------------------------*
 *  Global variables
 *---------------------------------------------------------------------------*/

EventData  FAR pktData[3];

DeviceConfig  FAR devCfg[MaxDevices];

ulong  announceInterval, intervalOffset;
static bool  configSent, resetInterval;



/*===========================================================================*/
/*=========== Common funcs with platform-specific implementations ===========*/
/*===========================================================================*/

/*---------------------------------------------------------------------------*
 *  Update the # of seconds between announcement packets
 *---------------------------------------------------------------------------*/
void resetAnnounceInterval(void)
{
  _EnterCritical_();  resetInterval = true;  _ExitCritical_();
}

/*---------------------------------------------------------------------------*
 *  Update the # of seconds between announcement packets
 *---------------------------------------------------------------------------*/
void updateInterval(void)
{

//--- reset the Interval between announce packets if flag set ---
  if (resetInterval)  {
    announceInterval = 1000 + intervalOffset;
    configSent = resetInterval = false;
    if (ShowProcess)  printf("Reset Announcement interval\r\n");
    return;
  }

//--- if last change to the announcement interval expired, then lengthen ---
//--- the interval (up to a maximum) ---
  if (configSent)  {
    announceInterval = announceInterval + announceInterval/2 + intervalOffset;
    if (announceInterval > MaxInterval)  announceInterval = MaxInterval;
    configSent = false;
  }

}


/*---------------------------------------------------------------------------*
 *  Update the status of entries in our list of devices based on the data
 *  from a just-received packet
 *---------------------------------------------------------------------------*/
static int processAnnouncePkt(EventData FAR paramVals[3])
{
  ulong  paramMask;
  bool   gotType, gotID, gotIP, gotName;
  EventData annVals;
  IFConfig  ifConfig;
  char  temp[32];


  annVals = *paramVals;

  paramMask = annVals.paramMask;

  gotType = BOOL(paramMask & (1 << PID_DEVTYPE));
  gotID   = BOOL(paramMask & (1 << PID_DEVID));
//  gotMAC  = BOOL(paramMask & (1 << PID_MACADDR));
  gotIP   = BOOL(paramMask & (1 << PID_IPADDR));
  gotName = BOOL(paramMask & (1 << PID_NAME));


//--- Check to see if the configuration conflicts with ours ---
  if (gotIP)  {
    getIFConfig(IF_ETH0, &ifConfig);
    if (annVals.ipAddr == ifConfig.ipAddr)
      printf("IP Address conflict from: %s\n",
             mactoa(temp, annVals.macAddr) );
  }

  if (gotType AND gotID)  {
    if ((annVals.devType == devType) AND (annVals.devID == deviceID))
      printf("Device Type + ID conflict from: %s\r\n",
             mactoa(temp, annVals.macAddr) );
  }

  if (gotName)  if (! StriEq(annVals.name, "unknown"))  {
    gethostname(temp, sizeof(temp));  temp[sizeof(temp)-1] = 0;
    if (strncmp(temp, annVals.name, NameSize) == 0)
      printf("Device Name conflict from: %s\r\n",
             mactoa(temp, annVals.macAddr) );
  }

  return 0;
}


/*---------------------------------------------------------------------------*
 *  If the conditions in the query match this controller, then respond with
 *  the requested values (for now it just resets the announce interval, which
 *  will send a broadcast pkt with all our values)
 *---------------------------------------------------------------------------*/
static int processQueryPkt(EventData FAR paramVals[3])
{
  if (! conditionsMatch(paramVals + MODE_MATCH))  return 0;

  resetAnnounceInterval();

  return 0;
}


/*---------------------------------------------------------------------------*
 *  If the conditions for these settings match this controller, then apply
 *  the settings
 *---------------------------------------------------------------------------*/
static int processSetPkt(EventData FAR paramVals[3])
{
  short  i, apply;
  ulong  mask;
  bool  settingsOK;
  EventData  FAR *setData;
  IFConfig  ifConfig;
  char  hname[NameSize+1];


  if (! conditionsMatch(paramVals + MODE_MATCH))  return 0;

  resetAnnounceInterval();  // shorten time between announce pkts to minimum

  getIFConfig(IF_ETH0, &ifConfig);
  gethostname(hname, sizeof(hname));  hname[sizeof(hname)-1] = 0;

  setData = paramVals + MODE_SET;  settingsOK = true;

/*--- do 2 passes - 1st one to verify all settings are valid, 2nd to ---*/
/*--- actually apply the settings ---*/

  for (apply=0; apply<2; ++apply)  {
    for (mask=4,i=2; i<31; ++i,mask<<=1)  {
      if (setData->paramMask & mask)  switch (i)  {

        case PID_IPADDR:
          if (setData->ipAddr != ifConfig.ipAddr)
            if ( (!(paramVals[MODE_MATCH].paramMask & (1 << PID_MACADDR)))
                OR FixedAddr() )  {
              settingsOK = false;  break;  }
          if (apply)  chgIPAddr(IF_ETH0, setData->ipAddr);
          break;

        case PID_NAME:
          if (! *setData->name)  { settingsOK = false;  break; }
          if (apply)  {
            printf("Set hostname to: %.*s\r\n", NameSize, setData->name);
            strcpyfn(hname, setData->name);  Sethostname(hname);
          }
          break;

        case PID_TIMEZONE:
          if ((setData->timeZone < -12) OR (setData->timeZone > 12))  {
            settingsOK = false;  break; }
          if (apply)  {
            printf("Set timeZone to: %d\r\n", setData->timeZone);
            timeZone = setData->timeZone;  //finish - safe/std method to chg?
          }
          break;

        case PID_CMD_STR:
          if (! *setData->cmdStr)  { settingsOK = false;  break; }
          if (apply)  {
            printf("Execute cmd: [%.256s]\r\n", setData->cmdStr);
            cmdLine(setData->cmdStr);
          }
          break;
      }
    }

    if (! settingsOK)  break;
  }


  if (ShowProcess)
    if (settingsOK)
      printf("New settings applied\r\n");
    else
      printf("Invalid settings\r\n");

  return 0;
}


/*---------------------------------------------------------------------------*
 *  Send commands as needed to deal with conflicts or to configure a new
 *  controller
 *---------------------------------------------------------------------------*/
static int sendCmdPkts(uchar *buf, short bufSize)
{
  return 0;
}




/*===========================================================================*/
/*========== functions that share identical code on all platforms ===========*/
/*===========================================================================*/

/*---------------------------------------------------------------------------*
 *  Compare the given parameter values to our own.  Returns a non-0 value
 *  if they match, 0 if they do not.
 *---------------------------------------------------------------------------*/
static bool conditionsMatch(EventData FAR *matchVals)
{
  short  i;
  ulong  mask;
  IFConfig  ifConfig;
  char  hname[NameSize+1];


  getIFConfig(IF_ETH0, &ifConfig);
  gethostname(hname, sizeof(hname));  hname[sizeof(hname)-1] = 0;

  for (mask=4,i=2; i<31; ++i,mask<<=1)  if (matchVals->paramMask & mask)

    switch (i)  {
      case PID_DEVTYPE:
        if (matchVals->devType != devType)  return false;
        break;

      case PID_DEVID:
        if (matchVals->devID != deviceID)  return false;
        break;

      case PID_MACADDR:
        if (fmemcmp(matchVals->macAddr, ifConfig.macAddr, 6))  return false;
        break;

      case PID_IPADDR:
        if (matchVals->ipAddr != ifConfig.ipAddr)  return false;
        break;

      case PID_NAME:
        if (fstrcmp(matchVals->name, hname))  return false;
        break;

      case PID_APPL_DTS:
        if (matchVals->applTime != applVersTime)  return false;
        break;

      case PID_RTOS_DTS:
        if (matchVals->rtosTime != rtosVersTime)  return false;
        break;

      case PID_SRFS_ADDR:
        if (matchVals->srfsAddr != srfsIPAddr)  return false;
        break;
    }

   if (ShowProcess)  printf("Conditions match\r\n");

   return true;
}

/*---------------------------------------------------------------------------*
 *  If we have not recently sent a command packet to the ctlr indicated in
 *  the given event, then send one now to set its IP address
 *---------------------------------------------------------------------------*/
/*
static int setAddr(int *sockfd, ulong ipAddr, EventData FAR *event,
                   uchar *buf, short bufSize)
{
  short  count;
  int  stat;
  uchar  *ptr;
  IFConfig  ifConfig;
  char  macAddr[6];


  if (event->sentCmdTime)
    if (msSince(event->sentCmdTime) < MinCmdInterval)  return 0;


  fmemcpy(macAddr, event->macAddr, 6);

  if (ShowProcess)  {
    printf("set IP address for %ls  %s to %s\n",
           event->name, mactoa(buf, event->macAddr),
           inet_htop(AF_INET, &ipAddr, buf+20, bufSize-20) );
  }

  getIFConfig(IF_ETH0, &ifConfig);

  ptr = buf;

  addShort(&ptr, PktFmtVers);  addShort(&ptr, PT_SET);

  addShortParam(&ptr, PID_MODE,     MODE_MATCH);
  addStrParam(&ptr,   PID_MACADDR,  macAddr, 6);
  addShortParam(&ptr, PID_MODE,     MODE_SET);
  addLongParam(&ptr,  PID_IPADDR,   ntohl(ipAddr));

  addShort(&ptr, 0);  addShort(&ptr, 0);  // id,len == 0 ends the list

  count = ptr - buf;

  stat = sendBCastPkt(sockfd, CFG_CMD_PORT, ifConfig.bcAddr, buf, count);

  if (stat == 0)  event->sentCmdTime = GetMS();

  return stat;
}
*/

/*---------------------------------------------------------------------------*
 *  Send a query to try to get a specific controller to respond
 *---------------------------------------------------------------------------*/
/*
static int sendQuery(int *sockfd,
                     DeviceConfig FAR *dc, uchar *buf, short bufSize)
{
  short  count;
  int  stat;
  uchar  *ptr;
  IFConfig  ifConfig;
  EventData  FAR *event;


  event = dc->event + LAST_SEEN;

  if (event->sentCmdTime)
    if (msSince(event->sentCmdTime) < MinCmdInterval)  return 0;

  if (ShowProcess)
    printf("send query for device: %ls  %02X:%04lX\r\n",
           dc->name, dc->devType, dc->devID);

  getIFConfig(IF_ETH0, &ifConfig);

  ptr = buf;

  addShort(&ptr, PktFmtVers);  addShort(&ptr, PT_QUERY);

  addShortParam(&ptr, PID_MODE,     MODE_MATCH);
  addShortParam(&ptr, PID_DEVTYPE,  ntohs(dc->devType));
  addLongParam(&ptr,  PID_DEVID,    ntohl(dc->devID));

  addShort(&ptr, 0);  addShort(&ptr, 0);  // id,len == 0 ends the list

  count = ptr - buf;

  stat = sendBCastPkt(sockfd, CFG_CMD_PORT, ifConfig.bcAddr, buf, count);

  if (stat == 0)  event->sentCmdTime = GetMS();

  return stat;
}
*/

/*---------------------------------------------------------------------------*
 *  Send a packet with the controller or server's configuration
 *---------------------------------------------------------------------------*/
static int announceConfig(int *sockfd, uchar *buf, short bufSize)
{
  short  count, bootImgStat;
  int  status;
  uchar  *ptr;
  IFConfig  ifConfig;
  char  primaryName[NAME_MAX], backupName[NAME_MAX];
  ulong  backupTime;
  struct stat  fsInfo;
  char  hname[NameSize+1];


  gethostname(hname, sizeof(hname));  hname[sizeof(hname)-1] = 0;

  getIFConfig(IF_ETH0, &ifConfig);


  bootImgStat = 0;  backupTime = 0;

  bootFileNames(primaryName, backupName, false);

  if (stat(backupName, &fsInfo) == 0)  {
    bootImgStat |= FW_HAS_BACKUP;
    backupTime = fsInfo.st_mtime;
  }

  if (runningPrimary)  bootImgStat |= FW_RUNNING_PRIMARY;

  if (stat(primaryName, &fsInfo) == 0)  {
    if (fsInfo.st_mtime > backupTime)  bootImgStat |= FW_PRIMARY_NEWER;
    if (fsInfo.st_mtime < backupTime)  bootImgStat |= FW_BACKUP_NEWER;
  }


  ptr = buf;

  addShort(&ptr, PktFmtVers);  addShort(&ptr, PT_ANNOUNCE);

  addShortParam(&ptr, PID_DEVTYPE,      devType);
  addLongParam(&ptr,  PID_DEVID,        deviceID);
  addStrParam(&ptr,   PID_MACADDR,      ifConfig.macAddr, 6);
  addLongParam(&ptr,  PID_IPADDR,       ifConfig.ipAddr);
  addLongParam(&ptr,  PID_NETMASK,      ifConfig.netmask);
  addStrParam(&ptr,   PID_NAME,         hname, strlen(hname));
  addShortParam(&ptr, PID_TIMEZONE,     timeZone);
  addLongParam(&ptr,  PID_APPL_DTS,     applVersTime);
  addLongParam(&ptr,  PID_RTOS_DTS,     rtosVersTime);
  addLongParam(&ptr,  PID_SRFS_ADDR,    srfsIPAddr);
  addLongParam(&ptr,  PID_UPSECS_HARD,  getSeconds());
  addLongParam(&ptr,  PID_UPSECS_SOFT,  time(NULL) - rebootTime);
  addShortParam(&ptr, PID_FIRMWARE,     bootImgStat);


  addShort(&ptr, 0);  addShort(&ptr, 0);  /* id,len == 0 ends the list */

  count = ptr - buf;

  status = sendBCastPkt(sockfd, CFG_CMD_PORT, ifConfig.bcAddr, buf, count);

  if (status == 0)  configSent = true;

/*--- adjust the offset used to spread out the timing of packets from ---*/
/*--- different controllers ---*/
  intervalOffset = GetMS() & 0x000003FFl;

  return status;
}


/*---------------------------------------------------------------------------*
 *  Extract parameter values from a packet, sorted by mode
 *
 *  Parameters:
 *    paramVals[3]: An array for storing parameter values extracted from
 *                  a packet.  Each of the 3 elements corresponds to 1 of
 *                  the 3 possible modes for parameter values:
 *                    MODE_ANNOUNCE, MODE_MATCH, MODE_SET
 *
 *    pkt:  a buffer containing the packet just received
 *
 *    pktSize:  The length of the packet pointed to by pkt
 *
 *
 *  On Return:
 *    If an error detected:
 *    - return value < 0
 *
 *    If no errors detected:
 *    - return value = type of packet (PT_ANNONCE, PT_SET, PT_QUERY)
 *
 *    - paramVals[3] contains any values corresponding to parameter values
 *      found in the packet, sorted by mode
 *
 *---------------------------------------------------------------------------*/
static int extractParams(EventData FAR paramVals[3], char *pkt, short pktSize)
{
  ulong  curTime;
  short  i, pktVers, pktType, id, size, mode, offset;
  uchar  *ptr, *endPkt, *valPtr;
  static struct tm  tms;
  uchar  temp[32];
  EventData  modeVals;


  fmemset(paramVals, 0, sizeof(EventData)*3);

  curTime = GetMS();
  for (i=0; i<3; ++i)  paramVals[i].time = curTime;

  ptr = pkt;  endPkt = ptr + pktSize;  mode = MODE_ANNOUNCE;

  pktVers = getShort(&ptr);  pktType = getShort(&ptr);

  if (pktVers > 1)  {
    printf("Unknown packet-format version: %d\r\n", pktVers);  return -1;  }
  if ((pktType < PT_ANNOUNCE) OR (pktType > PT_QUERY))  {
    printf("Unknown packet type: %d\r\n", pktType);    return -1;  }

  if (ShowPackets)
    printf("  pktVers: %d, pktType: %.9s\r\n", pktVers, pktTypeName[pktType-1]);


  while (ptr < endPkt)  {
    if ((mode < 0) OR (mode > 2))  {
      printf("Invalid parameter mode: %d\r\n", mode);  return -1;  }

    fmemcpy(&modeVals, paramVals + mode, sizeof(EventData));

    if ((id = getShort(&ptr)) == 0)  break;  /* id == 0: end of parameters */

    if ((id < 0) OR (id > 31))  {
      printf("Invalid parameter ID: %d\r\n", id);  return -1; }

    size = getShort(&ptr);
    if (ptr+size >= endPkt)  {
      if (ShowErrors)
        printf("invalid/unusable param size: ID:%d, size:%d\r\n", id, size);
      return -1;
    }
    valPtr = ptr;

    modeVals.paramMask |= (1 << id);  /* keep track of which params we got */
/*
    allZeroValue = false;
    for (i=0; i<size; ++i)  if (valPtr[i])  allZeroValue = false;
*/
    switch (id)  {

      case PID_MODE:
        mode =  getShort(&valPtr);
        if (ShowPackets)  printf("  mode: %d\r\n", mode);
        break;

      case PID_DEVTYPE:
        modeVals.devType = getShort(&valPtr);
        if (ShowPackets)  printf("  devType: 0x%02X\r\n", modeVals.devType);
        break;

      case PID_DEVID:
        modeVals.devID = getLong(&valPtr);
        if (ShowPackets)  printf("  devID: 0x%04lX\r\n", modeVals.devID);
        break;

      case PID_MACADDR:
        fmemcpy(modeVals.macAddr, valPtr, 6);
        if (ShowPackets)
          printf("  MAC Addr: %s\r\n", mactoa(temp, modeVals.macAddr));
        for (i=0; i<6 AND !valPtr[i]; ++i);  /* all 0 MAC value? */
        if (i == 6)  modeVals.paramMask &= ~(1 << PID_MACADDR);
        break;

      case PID_IPADDR:
        modeVals.ipAddr = getLong(&valPtr);
        if (ShowPackets)  {
          inet_htop(AF_INET, &modeVals.ipAddr, temp, sizeof(temp));
          printf("  IP Addr: %.20s\r\n", temp);
        }
        break;

      case PID_NETMASK:
        modeVals.netmask = getLong(&valPtr);
        if (ShowPackets)  {
          inet_htop(AF_INET, &modeVals.netmask, temp, sizeof(temp));
          printf("  Netmask: %.20s\r\n", temp);
        }
        break;

      case PID_GATEWAY:
        modeVals.gateway = getLong(&valPtr);
        if (ShowPackets)  {
          inet_htop(AF_INET, &modeVals.gateway, temp, sizeof(temp));
          printf("  Gateway: %.20s\r\n", temp);
        }
        break;

      case PID_NAME:
        if (size > NameSize-1)  size = NameSize-1;
        memcpy(modeVals.name, valPtr, size);  modeVals.name[size] = 0;
        if (ShowPackets)  printf("  Name: %.*s\r\n", NameSize, modeVals.name);
        break;

      case PID_TIMEZONE:
        modeVals.timeZone = getShort(&valPtr);
        if (ShowPackets)
          printf("  timeZone: %d\r\n", modeVals.timeZone);
        break;

      case PID_APPL_DTS:
        modeVals.applTime = getLong(&valPtr);
        if (ShowPackets)  {
          localtime_r(&modeVals.applTime, &tms);
          printf("  applTime: %04d-%02d-%02d %02d:%02d:%02d (%lu)\r\n",
                tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday,
                tms.tm_hour, tms.tm_min, tms.tm_sec, modeVals.applTime);
        }
        break;

      case PID_RTOS_DTS:
        modeVals.rtosTime = getLong(&valPtr);
        if (ShowPackets)  {
          localtime_r(&modeVals.rtosTime, &tms);
          printf("  rtosTime: %04d-%02d-%02d %02d:%02d:%02d (%lu)\r\n",
                tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday,
                tms.tm_hour, tms.tm_min, tms.tm_sec, modeVals.rtosTime);
        }
        break;

      case PID_SRFS_ADDR:
        modeVals.srfsAddr = getLong(&valPtr);
        if (ShowPackets)  {
          inet_htop(AF_INET, &modeVals.srfsAddr, temp, sizeof(temp));
          printf("  SRFS Addr: %.20s\r\n", temp);
        }
        break;

      case PID_CMD_STR:  /* must be a 0-terminated string */
        offset = valPtr - pkt;
        if (offset + size > pktSize)  size = pktSize - offset;
        modeVals.cmdStr = valPtr;  valPtr[size-1] = 0;
        if (ShowPackets)
          printf("  Cmd Str: [%.*s]\r\n", size, modeVals.cmdStr);
        break;
    }

    ptr += size;
    fmemcpy(paramVals+mode, &modeVals, sizeof(EventData));
  }


  return pktType;
}

/*---------------------------------------------------------------------------*/
/*  check for and process any packets received on the specified socket       */
/*---------------------------------------------------------------------------*/
void checkForPacket(int *sockfd, uchar *packetBuf, int bufSize)
{
  int  stat, pktType;
  struct sockaddr_in  fromAddr;
  socklen_t  addrLen;
  uchar  temp[32];


/*--- check for a packet from a controller or a config server ---*/
  memset(packetBuf, 0, bufSize);

  if (ShowPackets)  {
    addrLen = sizeof(fromAddr);
    stat = recvfrom(*sockfd, (void *)packetBuf, bufSize, 0,
                      (struct sockaddr *)&fromAddr, &addrLen);
  }
  else  {
    addrLen = 0;
    stat = recvfrom(*sockfd, (void *)packetBuf, bufSize, 0, NULL, &addrLen);
  }


  if (stat == 0)  {
    printf("UDP config socket closed by other end\r\n");
    close(*sockfd);  *sockfd = -1;  return;
  }

  if (stat < 0)  {
    if (errno == EAGAIN)  { SleepMS(50);  return; }
    perror("recvfrom()");  close(*sockfd);  *sockfd = -1;  return;
  }

  if (ShowPackets)
    printf("got %d byte packet from: %s\r\n", stat,
           inet_ntop(AF_INET, &fromAddr.sin_addr, temp, sizeof(temp)) );

  {  /*tdebug
    int  i;
    printf("  ");
    for (i=0; i<stat; ++i)  {
      printf("%02x ", (uchar)packetBuf[i]);
      if ((i & 0x000f) == 0x000f)  printf("\r\n  ");
    }
    printf("\r\n\n");
    */
  }

  if (stat < 6)  { printf("incomplete packet\r\n");  return; }

//--- extract values from the packet just received ---
  if ((pktType = extractParams(pktData, packetBuf, stat)) < 0)  return;

//--- process the data from the packet ---
  switch (pktType)  {
    case PT_ANNOUNCE:  processAnnouncePkt(pktData);  break;
    case PT_SET:       processSetPkt(pktData);  break;
    case PT_QUERY:     processQueryPkt(pktData);  break;
  }

}


/*---------------------------------------------------------------------------*/
/*  Task to manage discovery and config of NSCL ethernet-based devices       */
/*---------------------------------------------------------------------------*/
#ifdef EPICS
  int configd(void)
#else
  void _task configd(int argc, char **argv)
#endif
{
  int  sockfd, fflags;
  short  i, j;
//  uint  sp;
  ulong  lastAnnounceTime, lastCmdTime, curTime, terminateTime;
  sigset_t  sigSet;
  DeviceConfig  FAR *dc;
  EventData  FAR *event;
  struct sockaddr_in  locAddr;
  IFConfig  curConfig;
  uchar  packetBuf[MaxPktSize];
//  uint  dummyStackUseTest = 0xDEAD;

  static int  so_broadcast = TRUE;


/*finish - don't allow more than 1 instance to run!!!!*/

//--- block Terminate signal so we can send a packet just before exiting ---
  sigSet = sigmask(SIGTERM);  sigprocmask(SIG_BLOCK, &sigSet, NULL);
  terminateTime = 0;

  redirectStdio();  // so printf, perror, etc don't end up in data stream

  fmemset(devCfg, 0, sizeof(devCfg));

  #ifndef RABBIT
  {
    IFConfig  ifConfig;

    printf("Starting configd task\n");

    getIFConfig(IF_ETH0, &ifConfig);

    printf("Addr of Interface to talk to ctlrs: %s\n\n", hltoa(ifConfig.ipAddr));
    printf("Broadcast addr is %s\n\n", hltoa(ifConfig.bcAddr));

    timeZone = -4;  /*finish*/

    configdInitialized = TRUE;
  }
  #endif


  sockfd = -1;  resetInterval = false;

  lastAnnounceTime = GetMS();  announceInterval = 1000;  configSent = false;
  lastCmdTime = lastAnnounceTime;


/*--- processing for configuration announcement packets ---*/
  for (;;)  {
    sleepMS(20);

  //--- Check for pending Terminate signal.  If we get a new one, ---
  //--- then reset announceInterval and wait 4 seconds before terminating ---
    if (sigpending(&sigSet) == 0)  if (sigismember(&sigSet, SIGTERM))  {
      if (! terminateTime)  terminateTime = GetMS();
      announceInterval = 500;
    }
    if (terminateTime AND (msSince(terminateTime) > 4000))  Exit(-1);

    errno = 0;

    getIFConfig(IF_ETH0, &curConfig);
    if (curConfig.state != IF_UP)  { sleepMS(100);  continue; }

  //--- (re)create and bind socket to specific port, using broadcast mode ---
    if (sockfd < 0)  {
    //--- settings for local "addr" for socket to receive broadcasts ---
      locAddr.sin_family        = AF_INET;
      locAddr.sin_port          = htons(CFG_CMD_PORT);
      locAddr.SIN_ADDR          = htonl(INADDR_ANY);  /*htonl(bcIPv4);*/

      if ((sockfd = Socket(AF_INET, SOCK_DGRAM, 0)) < 0)  {
        perror("sockfd socket() error");  SleepMS(10000);  continue; }

      if ((setsockopt(sockfd, SOL_SOCKET, SO_BROADCAST,
                        (char *)&so_broadcast, sizeof(so_broadcast)) ) < 0)  {
        perror("setsockopt()");
        close(sockfd);  sockfd = -1;  SleepMS(10000);  continue;
      }

      if (bind(sockfd, (SAP)&locAddr, sizeof(locAddr)) < 0)  {
        perror("sockfd bind() error");
        close(sockfd);  sockfd = -1;  SleepMS(10000);  continue;
      }

    //--- make socket non-blocking ---
      fflags = fcntl(sockfd, F_GETFL, 0);
      fcntl(sockfd, F_SETFL, fflags | O_NONBLOCK);

      printf("Created listener on UDP port %d\r\n", CFG_CMD_PORT);
    }

  /*--- announce our own configuration on occasion ---*/
    updateInterval();
    if (msSince(lastAnnounceTime)  > announceInterval)  {
      announceConfig(&sockfd, packetBuf, sizeof(packetBuf));
      lastAnnounceTime = GetMS();
    }

  //--- periodically send cmd pkts to try to deal with conflicts ---
    if (msSince(lastCmdTime) > 1000)  {
      sendCmdPkts(packetBuf, sizeof(packetBuf));  lastCmdTime = GetMS(); }

  //--- check the age of the event data for each device in the list ---
    curTime = GetMS();
    for (i=0; i<MaxDevices; ++i)  {
      dc = devCfg + i;  if (! dc->inUse)  continue;
      for (j=0; j<NumEvents; ++j)  {
        event = dc->event + j;  if (! event->time)  continue;
        if (msInterval(event->time, curTime) <= MaxInterval*2)  continue;
        fmemset(event, 0, sizeof(EventData));
      }
    }


  /*--- check for cmds/queries ---*/
    checkForPacket(&sockfd, packetBuf, sizeof(packetBuf));

  //tdebug
/*
    {
      sp = _asm( [_HL], "ld hl,0", "add hl, sp", _HL );
      printf("---%04X\r\n", sp);
      computeMemCRCs(false, true);
      sp = _asm( [_HL], "ld hl,0", "add hl, sp", _HL );
      printf("   %04X\r\n", sp);
    }
*/
  }


  #ifdef EPICS
    return (0);
  #endif
}

/*---------------------------------------------------------------------------*/

