//----- epics_cad.c ----- 11/07/13 -----(03/02/10)-----

//  Typical diag msgs output settings:   diag msgs epics_cad 0x13a

/*=============================================================================
  Copyright (c) 2010-2012, NSCL and Mark Davis.  All rights reserved.
 ------------------------------------------------------------------------------
  Process to provide access to data elements via EPICS Channel Access protocol.

  The basic operation of this process is as follows:

    Task that handles all UDP traffic:

      - Generate beacon packets

      - Listen for and respond to CA SEARCH packets

    Task that handles all TCP traffic:

      - Listen for and establish new Virtual Circuit connections on request

      - Process commands received on each Virtual Circuit

      - Maintain lists of Virtual Circuits, Channels, and Subscriptions

      - Periodically check for events that apply to each Subscription and send
        out messages as needed

===============================================================================
  Version log (most recent first):


  2013-11-07 - Mark Davis
               - Corrected last argument to inet_ntop() call in the function
                 showAllowedList().

  2012-08-01 - Mark Davis
               - Tweaked settings and modified code to reduce overhead of
                 ca_tcpd task, especially when there are no clients with
                 subscriptions.

  2012-07-30 - Mark Davis
               - ca_udpd now sleeps during the loop if the processing of UDP
                 packets is disabled.  Prior to this, the CPU usage would
                 actually increase when processing of packets was disabled.
               - ca_tcpd now sleeps at least one tick each pass.
               - Improved handling of various cas command options to properly
                 deal with having, none, one, or both the TCP and UDP service
                 tasks running.

  2012-07-27 - Mark Davis
               - Added support for "pending" subscription msgs when the lock on
                 the PV value cannot be obtained without blocking.

  2012-06-14 - Mark Davis
               - Modified to use a separate process to listen for and respond
                 to UDP packets.

  2012-04-24 - Mark Davis
               - Actual # of data elements is now returned for CREATE_CHAN msgs

  2012-04-19 - Mark Davis
               - The "cas stats" command now uses a new line for each second's
                 counts so you can see the history.

  2012-03-27 - Mark Davis
               - Began work on changes to support PV values that are NOT
                 defined via the REG3 and REG4 macros.

  2012-02-29 - Mark Davis
               - Fixed the logic for the temporary boosting of the limit on the
                 time spent each cycle processing UDP packets (so it will
                 actually DO it now).  The boost is now controlled by two
                 separate values that specify the period (over multiple cycles)
                 that the limit is boosted, and the amount the limit is first
                 increased by (which decreases linearly over time).
               - Added anychg option to generate event monitor msgs when there
                 is any change at all for a value

  2012-02-28 - Mark Davis
               - Added the summary option to the cas command.
               - Modified code so it deals with the fact that sendto() doesn't
                 currently support blocking mode.

  2012-02-21 - Mark Davis
               - Like the UDP case, consecutive outgoing messages for a given
                 TCP connection are now accumulated in the outgoing packet
                 buffer first to help minimize the number of packets that are
                 actually used to send them over the physical link layer (and
                 the overhead and delays caused by multiple write calls when
                 we don't buffer them ourselves).

  2012-02-17 - Mark Davis
               - Added ability to disable reception and processing of UDP
                 packets, to generate event monitor msgs for all subscriptions
                 or just ones that have changed, and to set the max rate at
                 which event monitor msgs are generated.

  2012-02-09 - Mark Davis
               - All basic functions working.  UDP packets are now read and
                 processed (for a specified small set of allowed clients) in
                 sufficient numbers that clients can reconnect fairly quickly.
               - Some simple optimzations added, more to come.

  2012-02-06 - Mark Davis
               - Writes working

  2012-02-02 - Mark Davis
               - Basic search, connection, channel creation, reads, and
                 subscription mechanisms working.
               - STS (status) and TIME extended data formats functioning.

  2012-01-20 - Mark Davis
               - Started filling out the basic functionality.

  2010-03-02 - Mark Davis
               - Initial skeleton to provide a starting point, using pieces
                 of code copied from configd.c

===============================================================================
                     NOTES on efficiency and limitations
===============================================================================

  The Rabbit modules DEBROS currently runs on are relatively slow and very
  limited in resources.  We therefore have to limit the number of simultaneous
  connections we allow.  We also limit our processing of packets to those that
  came from an address in a list, which provides two important features:

    - We effectively reserve a connection for each client that MUST be able
      to connect so they never find us out of available connections.  In
      general, the addresses are those of machines running the EPICS gateway
      software, which can service hundreds of clients simultaneously rather
      than just the few that the controller is limited to.

    - We eliminate the overhead of processing every UDP packet we receive,
      allowing us to concentrate on ones from clients we wish to allow to
      connect.

  In a large system such as the NSCL, there are frequently many unconnected
  channels, which results in lots of CA broadcast packets from the IOCs and
  gateways that are periodically searching for them.

  The controllers DEBROS is running on at the time of this writing simply
  cannot read, much less process, every one of these packets.  And while the
  the CA protocol is designed to deal with some loss of packets, if a server
  misses too many of the ones meant for it, then clients won't be able to
  discover it is the source of the channels it is looking for and so won't
  open a connection to it.

  While there is no (practical) way for this service to limit the reception
  (and reading) of UDP packets from all clients, we CAN avoid PROCESSING ones
  from addresses that are not in the list.  This greatly reduces the time
  spent on the packets we wish to ignore, allowing us to read and process
  more of the ones we need to.

  The end result:  Even with the limitations of the current hardware, we are
  able to catch and process enough of the packets from the clients in the
  list that they are able to discover what channels we service and open
  a connection to access those channels in a reasonably short period.

  NOTES:

     Two ways that connections can be established with even less delay:

     - The CA protocol supports the concept of name servers, where machines
       can respond to SEARCH requests for channels that are served by other
       machines.  Even a low-end PC running Linux will be able to capture
       and respond to far more UDP packets with SEARCH requests than a
       Rabbit-based controller, which would provide a faster and more
       reliable means for the clients to discover the IP address of the
       controller that serves a group of channels.

     - Limiting the processing of broadcast packets to a small number of
       gateways means clients can use the name of a controller and the PVs
       it serves and read the data "through" one of the gateways.  While
       this can be handy for diagnostic purposes, the primary use of the
       CA service on the controllers will likely be for an EPICS SIOC to get
       the data it needs using event monitors (subscriptions) on the channels
       the controller provides.  These channels could be created via the
       gateways, but the SIOC could also simply skip the discovery stage and
       open a virtual circuit directly to the controller, getting the
       address for the controller from a PV in an EPICS record.

 ______________________________________________________________________________

  Small vs Large PV values:

     PV's come in 2 varieties:

        - Small, simple numeric values are defined via ModBus register macros.
          These macros not only create a RegInfo structure that describes the
          value, but also declare a global variable that contains the value.

        - Larger values are defined elsewhere (often as arrays) and kept track
          of by a PV_Desc structure created using macros like PV_STRING and
          PV_ARRAY.  At present, the declaration of the global variable is
          separate from the PV_xxx macros used to track them.

     Both the RegInfo and PV_Desc structures contain an eventsOccured mask
     which is used by the CA_ProcessSubscriptions() func to determine which
     PV values have changed since it was last called.  The processing of
     subscriptions starts with a call to pvCheckAllEvents() which updates
     the eventsOccured masks for all the PVs at once.


     For the smaller PVs, the eventsOccured mask is updated by keeping a
     copy of the last value sent to CA clients and comparing it to the
     current value.  Because these values are so small, they never have
     to be locked for more than a moment, so locking can be kept simple
     and efficient by using a critical section.


     For the larger PV values, this approach is generally not practical
     for one or more of the following reasons:

        - Memory usage:  Keeping a separate copy means doubling the amount
          of memory required for the value.

        - Detecting a change:  Even if the memory is available, comparing
          hundres or thousands of bytes each time pvCheckAllEvents() is
          called would require too many CPU cycles.

        - Lock interval:  Larger values require more time to examine, copy,
          and/or update.  This means the method used to lock the value during
          such operations has to be far more selective than the use of a
          critical section or even disabling the scheduler.


     To make matters more complex, how each of these issues is addressed
     affects the others:


        Memory usage vs locking and delays:

           Keeping a separate copy of a large value means increased memory
           usage, but can significantly reduce how long and how often the
           value needs to be locked.

           The "live" value has to be locked against changes whenever it is
           being updated or being read.  A separate copy (or lack of one)
           has no affect on how LONG we need to hold the lock during the
           process of updating the value.  However, it DOES have an affect
           on how OFTEN and how LONG we need to lock the value to read it.

           If we have a separate copy, then the "live" PV value only has to
           be locked long enough to test and (possibly) update the copy.
           Once the copy is updated, it can be read as long and as often as
           needed without interfering with the updating of the live value.

           If we don't have a separate copy, then the value has to be locked
           for however long it takes to send each new value to each client
           that has a subscription to the value.

           If the value is fairly small (e.g. a string), then this may
           not be a significant issue.  But for larger values (especially
           those larger than a single ethernet packet), sending the value to
           just ONE client requires far more CPU time than a simple buffer-
           to-buffer copy.  More time means a higher likelihood of being
           pre-empted, causing the lock to be held even longer.  And if it
           needs to be sent to multiple clients, then it just multiplies
           the problem.

           The bottom line:  Separate copies can use up precious memory, but
           not using them can result in MUCH longer and more frequent delays
           for any process that updates the value (which is usually the
           main applic code, the one task that often must NOT be delayed).


        Detecting a change vs memory usage vs code complexity:

           Comparing the current value with the last one sent makes the
           detection of a change very simple and does not require any
           changes to any other code.  But for larger values, this not only
           requires much more memory, but requires too much CPU time.

           The only other option is to make the code that changes the value
           responsible for updating some smaller values that can be used
           to quickly detect that the value has changed.  One of these
           values is the newEvents bitmask in the PV_Desc structure.  In
           addition to this, another value that serves to uniquely identify
           the value (usually a counter or a date/time stamp) is helpful
           or required to support this process.

           Example:
             The PhaseCtlr controller has 2 copies of each of its waveform
             values.  Each copy has its own lock and time stamp.

             The applic task updates the 1st copy whenever the DSP has new
             data for the waveform and the buffer is not locked by another
             task.  The 2nd copy is updated by whichever task is the first
             to get a lock on both copies and finds that the time stamps
             for the copies are different.  The function called to do this
             also sets the appropriate bit in the newEvents mask for the
             value whenever it updates the 2nd copy.

           This approach requires that any code that changes the value also
           updates a smaller value (generally a counter or date/time stamp)
           that uniquely identifies the value.  But this small increase in
           complexity means we now have two small values that can be quickly
           compared to detect a change without the need for obtaining one or
           more locks.


        Locking delays:

           One thing that can't be eliminated for larger values is the need
           to use a finer-grained locking method.  Critical sections and
           even the "BKL" (Big Kernel Lock) approach are OK (and even have
           some advantages) when the timeframe is VERY small.  But disabling
           the operating system long enough to copy hundreds or thousands of
           bytes is definitely NOT a practical approach.

           A more appropriate method for these longer timeframe locks is to
           use semaphores.  One of the useful features of the standard Unix/
           Linux semaphore API is the ability to NOT block when attempting
           to get a lock.  What this means for the CA server is that it can
           avoid delaying the processing of other values when it tries to
           process one that is currently locked by another task.

           This is, of course, not an ideal solution.  It means that it
           can take longer for clients to be informed of a change in a PV's
           value.  But the alternative is to ALLOW the CA server task to
           block whenever it needs to process a new value, which can cause
           delays of indeterminate length for the processing of ALL the
           not-yet processed new values.

           Where practical, the tasks that update the values can also use
           this same try-again-later approach to avoid delays.  But of
           course this is only practical if the task that needs to update
           the locked value can get the information needed to do so at a
           later time without causing other delays or problems.

           For example, if the data for the locked PV is in the middle of a
           stream of data for many PVs, then the task would either have to
           stop reading the stream until it got the lock (which is just
           another way to block waiting to get the lock), or read the
           data and store it somewhere until it could get the lock.

           But no matter how many values you can save, or how long you can
           save them for, eventually you could run out of space and have
           to deal with the same problem:  Either stop reading and wait
           for the lock, or loose one of the not-yet-processed values.

           The PhaseCtlr uses the try-again-later approach by simply not
           reading the newest data from the DSP until it can lock the
           buffer it saves the new data in.  But that just means the problem
           has been pushed down one more level, as the DSP will not generate
           a newer value until the one in it ITS buffer has been read.

           The bottom line:  Whether or not the try-again-later approach
           is/can be used, the fact is that values will generally change
           much faster than we can process them.  The advantage of the
           try-again-later approach is that it processes as many changes
           as it can as fast as it can, so it actually minimizes this
           problem.


       When considering all these issues, you should also keep in mind that
       EPICS itself and the CA protocol are a "best-effort" design.  There
       is no way that either can guarantee that every change to every value
       will be seen, transmitted, received, and processed by every entity
       that uses the values.

       The bottom line is that each entity has to work with the information
       it has at any given moment.  For any 2 entities, even ones using the
       same set of variables, this information will frequently be slightly
       different for any given instant in time.  But EPICS is designed to
       minimize these differences as much as possible and provide the latest
       information to each entity as quickly as possible.

===============================================================================

  TO-DO List

  ====================
   CONCURRENCY ISSUES
  ====================


  ===================
   EFFICIENCY ISSUES
  ===================

   - Use a global (shared by all connections) array of CA_Sub objects and
     replace the connection-specific array of CA_Sub objects with an index
     (or pointer) to a CA_Sub object in the global array.  This should allow
     me to have fewer (total) CA_Sub objects and avoid lots of unused ones
     in EACH CA_Con object (particulary the ones that aren't in use, in which
     case ALL the CA_Sub objects for that connection are unused).


   - Is it possible to block on read from a UDP socket?  If so, this could
     improve response times for SEARCH messages.

        NOTE:  I don't currently use the callback mechanism for UDP sockets,
        so that would probably need to be added properly support blocking for
        UDP sockets.  In fact, at the moment, I suspect if you did a blocking
        read on a UDP socket that it could block the caller forever if the
        socket stayed open and no data was received and no error occured.


   - Currently the throughput bottleneck caused by larger PV values is reduced
     by limiting the # of bytes of extended PV data we allow each pass.  But
     even with this limit in place supporting larger values significantly
     slows down throughput.

     Originally I figured I would have 2 separate processes and have
     one of them relegated to handling all the larger values.  But that
     adds considerable complexity and there is a much simpler alternative
     that would be just about as effective (although it does require the
     cooperation of the client).

     To avoid having the larger PV values slow down the rate at which the
     service can send updates to the smaller values, the larger PV values
     must be handled by a separate connection for each client.

     The service can enforce this by not accepting the creation of
     channels for the larger PVs on the same connection as the smaller
     PVs.  One way this might still be usable by standard CA clients that
     don't know about this arrangement would be to use different IP
     addresses for the 2 connections so the client would not try to create
     channels for both the larger and smaller PVs on the same connection
     (I am assuming that clients generally use a single connection for
     all PVs hosted by the same IP address).

     As long as all the larger PVs support a lock whose scope is just
     that value, then the CA service can spread the sending of the larger
     values across multiple passes, processing the smaller ones as fast
     as it can on the separate connections.  (Spreading the sending of
     a single value across multiple passes avoids blocking the single
     CA server task, which slows down throughput for all connections).

     An extenstion of this approach COULD even be applied to the smaller
     values:  The service could check the status of a connection and not
     write the next value to a socket if it would block in doing so (just
     mark the subscription pending if it can't be sent without blocking).
     The current value for the same PV could still be written to other
     connections for other clients as soon as they have room.

     Of course, this assumes I can check how much space is left in the
     output buffer for an arbitrary fd (something DEBROS supports, but
     how portable is this really?).  BUT: If I have the memory for one
     respPkt[] for each connection, then I could test how empty IT is.
     That is completely portable, and provides the advantage of buffered
     output for each connection.

     In this way, each client gets updates as fast as the service can
     write them and the throughput for each client is primarily a function
     of how many PVs it has subscribed to.  Of course, there are a limited
     # of CPU cycles to go around, and the network library has an upper
     limit on the total # of bytes that can be pumped out even if they are
     split up amoung all the output buffers, but this approach would at
     least minimize the affect one client can have on any single client.

     The one downside, of course, is that the more PVs any given client
     has subscribed to the fewer updates per second it will get for any
     given channel.  But it is very unlikely that ANY client will actually
     see ALL the changes to a value unless the change threshold is set
     fairly high.  That is a built-in limitation to CA itself, and one
     that is simply not practical to avoid.  Any attempt to insure all
     clients see the same sequence of values just means you ensure all
     clients get updates at the rate of the slowest or busiest connection.
     And that just means that, instead of some clients missing fewer of
     the changes, they ALL miss the same (large) # of changes as the
     slowest/busiest connection.

   NOTES about this approach:

     The "slowest-common-denominator" effect just discussed still DOES
     apply to some degree to the connections for the  larger PV values.

     Locks on the larger PV values will often exist across 2 or more
     passes.  This means that, as long as a read lock exists for any single
     connection, a PV value cannot be updated.

     What this means is that if one client has only a few subscriptions
     on larger PV values, and other clients have several that overlap
     them, there is a much higher chance that there will be at least one
     read-only lock on a value.  Which means the write-lock necessary
     to update the value will take longer to get, and so the value will
     be updated less frequenly.

     So even though a client has only a few subscriptions, updates to
     those values can be delayed due to the number of other clients that
     subscribe to the same values.

     Of course, the subscription logic doesn't sit around waiting for a
     value to change.  It always sends any updated values for subscriptions
     that it can get a read-only lock on.  If a value is locked for writing,
     (meaning it could change in the middle of a read) it simply moves to the
     next value in a client's list.  It NEVER blocks if it doesn't HAVE to.
     It will always look for the next productive thing it can do (which
     generally means writing any new values for PVs it CAN get a read-only
     lock on).

     The likely result is that each client will get a different subset of
     all the values the PVs have gone through.  So long as we mark a
     subscription as pending when we can't send a new value, there there
     is little chance that any given client will NOT get an updated value
     for a PV shortly after it has changed.  It may have changed multiple
     times since the last value that was sent, but each client with a
     subscription will at least get the current value not long after the
     first change in a series of changes.


  - If I have one respPkt[] per connection, I could modify the way
    subscriptions are handled so it writes the same msg to the respPkt[]
    for each connection that has an identical subscription (although that
    would have to include the # elements, the dbrType, etc).  But in
    the scenario this service will be used, this probably rarely happens,
    so it is probably is of no use (and may actually ADD overhead).


  Question:  Is there a standard Unix/Linux method to avoiding
             buffer-to-buffer copying when calling functions like
             send()?  Unless such functions are smart enough to
             avoid this when called with close to a packet's worth
             of bytes, they most likely do what my sendRespPkt()
             func does (add more to the existing buffer until it
             is full and then send it and start filling it again -
             but with the naggle logic thrown in as well).

             If there was a way to let the user-space code tell
             some equivalent of send() to immediately process the
             contents of the buffer it passed a pointer to (i.e.
             skip the buffering part), that could be very handy.

             Of course, while it reduces the overhead of doing
             buffer-to-buffer copies, it also more tightly binds
             the timing of the application code to the timing
             of the network layer.  Something that buffering
             helps to smooth out, allowing the user-space code
             to do something productive while the network stack
             handles sending the data queued up in a buffer
             in the background.



  - Test response times to make sure WRITES are processed quickly enough for
    tuning using knob controls.


  - When sending updates for subscriptions, is there any way I can take
    advantage of the fact that I pretty much KNOW I am generating a bunch
    of event messages?  Filling in as many fields as possible in the resp
    message is one thing that comes to mind, although that is such a small
    portion of the CPU cycles that I doubt it would be significant.  Also
    keep in mind that the dbrType for each subscription can be different.


  - Regarding reconnect times after rebooting a controller:

       How about this:  Have a separate service (based on this code) act as
       a directory service.  If I had a CA server task running on a Linux
       box and using the option in the SEARCH request to return the IP
       address of a server, then IT could respond to the requests that the
       controllers are missing and tell the clients the IP address of the
       controllers.

       Note that such a service would not need any configuration files.  It
       could keep track of the existence and the health of the DEBROS
       controllers using the broadcasts they are already doing for the
       smartConfig configuration management service.

       NOTE:  Unless they had a list of PVs by controller type, they would
       have to be based strictly on the host/record name portion of the PV
       name.  And if I wanted a controller to respond to other names, then
       this external service WOULD need to know about those names.  What
       would probably be best would be to able to query the controllers
       for a list of the PV names they host.

       Question:  Are there not passive servers out there like this already?
       But CA servers respond directly to the client that sent a query, so
       how would such a server see the responses needed to build and maintain
       a database of PV names?  Even if you assumed it read the raw socket
       data, the response packets would not normally be forwarded to the
       server by the switch it was connected to.  So how DO they work?

       Possibilities:
         - I was wrong about there being "passive" servers that discover PV
           names just be being plugged in to the network.

         - They work by being "proxies" or "gateways" to the CA servers.  If
           all search requests are sent to the name servers, and only the
           name servers send the queries to the CA servers, then the name
           servers would know what the PV names are and which ones are (or
           at least recently were) "alive".

           This could be done either by having the name servers literally be
           gateways between the CA clients and servers, or by configuring
           network switches so the CA servers don't get any such broadcast
           messages from anything but the name servers.

         - They are semi-passive:  They get the list of PVs from the EPICS
           database.  This wouldn't tell them what IP address to associate
           with each PV, but they would know what PVs are defined (and
           possibly more depending on how much they understand of the
           configuration of each record).  And they could always create
           channels for each of the PVs to keep tabs on which ones are "live".
                   ----------------------------------------


==========================
 UNFINISHED FUNCTIONALITY
==========================

  - ca_udpd task should sleep most the time when the processing of UDP
    packets is disabled.


  - Add numElems to the detailed list of subscriptions


  - Add the allowedClients[] list to the persistent variables stored in
    the devData file


  - Need to save and restore the minChg values for all PVs (this will be
    done in another module)


  - Add support for different output levels to the "cas state" option:
      - Default lists each connect and # of channels and subscriptions
      - Next level shows list of channels and # subscriptions on each
      - Next level shows everything (including IDs for each channel
        and subscription)


  - Make the cas add and del options sub-commands to the list option


  - When deleting an address from the allowedClients[] list, add option
    to shutdown any connections from that address


  - Option to not allow TCP connections from addresses not in the
    allowedClients[] list


  - Provide a means for each application to specify what value is used for
    the VAL field


  - Support extended types as much as possible.

    NOTE: The DBR_GR_ENUM and DBR_CTRL_ENUM formats seem rather impractical
    for use with subscriptions/monitors for any PV that changes value very
    often (because they include 416 bytes of string data with each msg).


  - What should I do about DOUBLEs (64-bit floats)?  Return error responses
    saying that format is not supported?  Always use 32-bit float formats
    in our responses and assume the client will deal with that fact that we
    responded using a format other than the one it requested?  Even though we
    can avoid use of the double formats in our own code, there may be
    applications that we want to occasionaly use and connect directly to the
    controller with, and THEY may not handle a response in a 32-bit float
    format when they requested a 64-bit one.


==============================================================================*/


#include <common.h>

#include <sys/types.h>
#include <unistd.h>

#include <sys/socket.h>

#include <sys/stat.h>
#include <fcntl.h>

#include <sys/modbus.h>
#include <sys/epics_cad.h>


#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdio.h>
#include <services.h>

#include <string.h>

#include <time.h>

#include <tsh.h>

#include <poll.h>

#include <busybox.h>

#include <ctype.h>



#define  CA_MAX_CONS       4
#define  CA_MAX_SUBSCR    64
#define  CA_MAX_CHANS     64

#define  USER_NAME_LEN    24
#define  HOST_NAME_LEN    24



#define  IPPORT_USERRESERVED  5000


#define CA_MAJOR_VERSION       4
#define CA_MINOR_VERSION      11
#define CAS_PROTO_VERSION    411


#define CA_PORT_BASE      IPPORT_USERRESERVED + 56U
#define CA_SERVER_PORT    (CA_PORT_BASE + CA_MAJOR_VERSION * 2u)
#define CA_REPEATER_PORT  (CA_PORT_BASE + CA_MAJOR_VERSION * 2u + 1u)

#define EPICS_CAS_BEACON_PORT   5065  // dest port for UDP beacon packets

#define CA_INIT_BEACON_PERIOD    25   // initial time between beacons (in ms)
#define CA_MAX_BEACON_PERIOD  15000   // max time between beacons (in ms)
                                         //  must be < 65500


#define DONT_REPLY    5u
#define DO_REPLY     10u





#ifdef RABBIT
  #include <sys/farprintf.h>
  #include <SHD_DLM.h>
  extern ulong  applVersTime, rtosVersTime, srfsIPAddr;
#endif


// This is the lower limit that can be set via the "cas events ###" cmd.
// It is the shortest allowed interval between testing for changes to PVs with
// active subscriptions.  This determines the maximum rate at which READ_NOTIFY
// messages are sent to clients.  Smaller values result in lower latency and
// more frequent messages, but can drastically increase CPU usage.
#define  MIN_EVENT_PERIOD  20u

// The default value for the minimum # of ms that must pass between testing for
// changes in and generation of event messages for channels that have event
// monitors.  Smaller values mean faster updates but higher CPU usage.
// This value can be changed using the "cas events ###" command.
#define  CA_EVENT_PERIOD  125


// The MINIMUM # of ms between iterations of ca_tcpd's main loop.  This sets
// an UPPER limit on how frequently the main loop executes.  It has no effect
// when the processing of messages during a pass takes longer than this.
// Larger values reduce CPU usage but can increase response time.
#define  MIN_LOOP_TIME    50u

// All the read and write operations on TCP socket connections in this module
// use the following idle timeouts.  These value specify the max # of ms a
// blocking read or write can be idle (i.e. no additional bytes received or
// written) without returning a timeout error.  Note that, as long as no SINGLE
// idle period exceeds this value, the operation will continue to block until
// it has successfully read or written the requested # of bytes.  So successful
// reads and writes CAN block for much longer than these values, so long as
// progress doesn't stop for a period exceeding these values.
#define  RD_TIMEOUT   2000u
//#define  WR_TIMEOUT   2000u

// The # of ms to wait after all connections are closed and before exiting the
// process to help insure the connections are cleanly closed
#define  CAS_STOP_WAIT_TIME   2500

// The # of ms to wait after the last failed attempt to create a new listening
// socket.  This reduces wasted CPU time and the number of error messages in
// the syslog when there is a problem (e.g. being out of socket buffers).
#define  LISTEN_RETRY    30000


// This sets the size of the input and output buffers used read UDP packets
// and accumulate response packets for both UDP and TCP connections.  It must
// be large enough to hold the largest UDP packet that we might receive and
// to hold all the responses to it.  To avoid fragementation, it should also
// not be LARGER than the largest block of data that will sent in a single
// link-layer (Ethernet) packet  In general, the value we want is the MSS
// value used by the TCP layer.  While this can vary depending upon several
// factors, the most commonly used value is probably 1460, which is the
// computed as follows:
//
//   MSS:  Max Segment Size (refers only to the data portion of the packet)
//   MTU:  Max Transmission Unit (usually 1500 for Ethernet)
//   IPh:  Header size for IP packet (20)
//   TCPh: Header size for TCP packet (commonly 20, but can be larger)
//
//   MSS = MTU - IPh - TCPh = 1500 - 20 - 20 = 1460
//
#define  IO_BUF_SIZE   1460   // set to most commonly used MSS




typedef  struct in_addr      IN_ADDR;    // IPv4 addr in network byte order
typedef  struct sockaddr    *SAP;        // pointer to type, port, addr struct
typedef  struct sockaddr_in  SAI;        // type, port, addr struct

typedef  struct ether_addr  EtherAddr;


//--- to make code more readable ---
#define S16_(ptr)   (*(S16 *)(ptr))
#define U16_(ptr)   (*(U16 *)(ptr))
#define S32_(ptr)   (*(S32 *)(ptr))
#define U32_(ptr)   (*(U32 *)(ptr))
#define F32_(ptr)   (*(F32 *)(ptr))



/*
 * !! Dont use this - it may vanish in the future !!
 *
 * Provided only for backwards compatibility with
 * db_access.h
 *
 */
#define MAX_STRING_SIZE 40
typedef char            epicsOldString[MAX_STRING_SIZE];

// These are part of the protocol spec and must NOT be changed !!!
#define MAX_UNITS_SIZE         8
#define MAX_ENUM_STRING_SIZE  26
#define MAX_ENUM_STATES       16

typedef char            epicsInt8;
typedef unsigned char   epicsUInt8;
typedef short           epicsInt16;
typedef unsigned short  epicsUInt16;
typedef epicsUInt16     epicsEnum16;
typedef long int        epicsInt32;
typedef unsigned long   epicsUInt32;
typedef float           epicsFloat32;
typedef double          epicsFloat64;
typedef unsigned long   epicsIndex;
typedef epicsInt32      epicsStatus;


typedef epicsOldString  dbr_string_t;
typedef epicsUInt8      dbr_char_t;
typedef epicsInt16      dbr_short_t;
typedef epicsUInt16     dbr_ushort_t;
typedef epicsInt16      dbr_int_t;
typedef epicsUInt16     dbr_enum_t;
typedef epicsInt32      dbr_long_t;
typedef epicsUInt32     dbr_ulong_t;
typedef epicsFloat32    dbr_float_t;
typedef epicsFloat64    dbr_double_t;

typedef epicsUInt16     dbr_put_ackt_t;
typedef epicsUInt16     dbr_put_acks_t;
typedef epicsOldString  dbr_stsack_string_t;
typedef epicsOldString  dbr_class_name_t;


typedef struct epicsTimeStamp {
    epicsUInt32    secPastEpoch;   /* seconds since 0000 Jan 1, 1990 */
    epicsUInt32    nsec;           /* nanoseconds within second */
} epicsTimeStamp;


// Compound types for CA messages that include a PV value

//! CA basic data types  (just the Value)
#define DBR_STRING          0
#define DBR_SHORT           1
#define DBR_INT            DBR_SHORT
#define DBR_FLOAT           2
#define DBR_ENUM            3
#define DBR_CHAR            4
#define DBR_LONG            5
#define DBR_DOUBLE          6

  // Status:  Status + Alarm Severity + Value
#define DBR_STS_STRING      7
#define DBR_STS_SHORT       8
#define DBR_STS_INT        DBR_STS_SHORT
#define DBR_STS_FLOAT       9
#define DBR_STS_ENUM       10
#define DBR_STS_CHAR       11
#define DBR_STS_LONG       12
#define DBR_STS_DOUBLE     13

  // Time:  Status + Alarm Severity + Time Stamp + Value
#define DBR_TIME_STRING    14
#define DBR_TIME_SHORT     15
#define DBR_TIME_INT       DBR_TIME_SHORT
#define DBR_TIME_FLOAT     16
#define DBR_TIME_ENUM      17
#define DBR_TIME_CHAR      18
#define DBR_TIME_LONG      19
#define DBR_TIME_DOUBLE    20

  // Graphical:  Status + Alarm Severity + Alarm/Warn limits + Value
#define DBR_GR_STRING      21
#define DBR_GR_SHORT       22
#define DBR_GR_INT         DBR_GR_SHORT
#define DBR_GR_FLOAT       23
#define DBR_GR_ENUM        24
#define DBR_GR_CHAR        25
#define DBR_GR_LONG        26
#define DBR_GR_DOUBLE      27

  // Control:  Status + Alarm Severity + Alarm/Warn limits + Ctl limits + Value
#define DBR_CTRL_STRING    28
#define DBR_CTRL_SHORT     29
#define DBR_CTRL_INT       DBR_CTRL_SHORT
#define DBR_CTRL_FLOAT     30
#define DBR_CTRL_ENUM      31
#define DBR_CTRL_CHAR      32
#define DBR_CTRL_LONG      33
#define DBR_CTRL_DOUBLE    34

  // Extra:
#define DBR_PUT_ACKT       35    // 16-bit value
#define DBR_PUT_ACKS       36    // 16-bit value
#define DBR_STSACK_STRING  37    // 40-char string value
#define DBR_CLASS_NAME     38    // 40-char string value

#define LAST_BUFFER_TYPE  DBR_CLASS_NAME

#define VALID_DB_REQ(x)    (((x) >= 0) AND ((x) <= LAST_BUFFER_TYPE))
#define INVALID_DB_REQ(x)  (((x) <  0) OR  ((x) >  LAST_BUFFER_TYPE))


#define  EXT_TYPE_NONE    0
#define  EXT_TYPE_STS     1
#define  EXT_TYPE_TIME    2
#define  EXT_TYPE_GR      3
#define  EXT_TYPE_CTRL    4
#define  EXT_TYPE_EXTRA   5


const far char far *baseTypeName[] = {
  "STRING", "SHORT", "FLOAT", "ENUM", "CHAR", "LONG", "DOUBLE" };

const far char far *extTypeName[] = {
  "", "STS", "TIME", "GR", "CTRL", "EXTRA" };


//--- conversion from base DBR value types to PV interface types ---

typedef struct {
  uint  pvType;  // PV data type equiv to base DBR type
  uint  size;    // The size of the DBR data type
  bool  reorder; // Do the bytes need to be reordered (from host to network)?
}  TypeData;

const far TypeData dbfToPV[] = {
  { DT_STR,  40,  false },  // DBR_STRING
  { DT_U16,   2,  true  },  // DBR_SHORT
  { DT_F32,   4,  true  },  // DBR_FLOAT
  { DT_U16,   2,  true  },  // DBR_ENUM
  { DT_U16,   2,  true  },  // DBR_CHAR
  { DT_U32,   4,  true  },  // DBR_LONG
  { DT_F32,   8,  true  }   // DBR_DOUBLE
};


const far uint  pvTypeToDBF[] = {
  DBR_SHORT,  //  DT_S16  16-bit signed integer
  DBR_SHORT,  //  DT_U16  16-bit unsigned integer
  DBR_LONG,   //  DT_S32  32-bit signed integer
  DBR_LONG,   //  DT_U32  32-bit unsigned integer
  DBR_FLOAT,  //  DT_F32  32-bit IEEE Float
  DBR_STRING, //  DT_STR  null-terminated string (for conversions only)
  DBR_CHAR,   //  DT_U8   8-bit unsigned integer
  DBR_CHAR    //  DT_S8   8-bit signed integer
};


//--- The offset of the data value in each of the extended type structures ---
const far uint  dbrValOffset[] = {
// STR SHRT  FLT ENUM CHAR LONG  DBL
    0,   0,   0,   0,   0,   0,   0,  // EXT_TYPE_NONE
    4,   4,   4,   4,   5,   4,   8,  // EXT_TYPE_STS
   12,  14,  12,  14,  15,  12,  16,  // EXT_TYPE_TIME
    0,  24,  40, 422,  19,  36,  64,  // EXT_TYPE_GR
    0,  28,  48, 422,  21,  44,  80,  // EXT_TYPE_CTRL
    0,   0,   8,   0                  // EXT_TYPE_EXTRA
};



//===== VALUES WITH STATUS STRUCTURES =====

/* structure for a  string status field */
struct dbr_sts_string {      // DBR_STS_STRING
  dbr_short_t   status;            /* status of value */
  dbr_short_t   severity;          /* severity of alarm */
  dbr_string_t  value;             /* current value */
};

/* structure for a short status field */
struct dbr_sts_short {       // DBR_STS_SHORT, DBR_STS_INT
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_short_t  value;              /* current value */
};

/* structure for a  float status field */
struct dbr_sts_float {       // DBR_STS_FLOAT
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_float_t  value;              /* current value */
};

/* structure for a  enum status field */
struct dbr_sts_enum {        // DBR_STS_ENUM
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_enum_t   value;              /* current value */
};

/* structure for a char status field */
struct dbr_sts_char {        // DBR_STS_CHAR
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_char_t   RISC_pad;           /* RISC alignment */
  dbr_char_t   value;              /* current value */
};

/* structure for a long status field */
struct dbr_sts_long {        // DBR_STS_LONG
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_long_t   value;              /* current value */
};

/* structure for a double status field */
struct dbr_sts_double {      // DBR_STS_DOUBLE
  dbr_short_t   status;            /* status of value */
  dbr_short_t   severity;          /* severity of alarm */
  dbr_long_t    RISC_pad;          /* RISC alignment */
  dbr_double_t  value;             /* current value */
};


//===== VALUES WITH STATUS AND TIME STRUCTURES =====

/* structure for a  string time field */
struct dbr_time_string {     // DBR_TIME_STRING
  dbr_short_t     status;          /* status of value */
  dbr_short_t     severity;        /* severity of alarm */
  epicsTimeStamp  stamp;           /* time stamp */
  dbr_string_t    value;           /* current value */
};

/* structure for an short time field */
struct dbr_time_short {      // DBR_TIME_SHORT, DBR_TIME_INT
  dbr_short_t     status;          /* status of value */
  dbr_short_t     severity;        /* severity of alarm */
  epicsTimeStamp  stamp;           /* time stamp */
  dbr_short_t     RISC_pad;        /* RISC alignment */
  dbr_short_t     value;           /* current value */
};

/* structure for a  float time field */
struct dbr_time_float {      // DBR_TIME_FLOAT
  dbr_short_t     status;          /* status of value */
  dbr_short_t     severity;        /* severity of alarm */
  epicsTimeStamp  stamp;           /* time stamp */
  dbr_float_t     value;           /* current value */
};

/* structure for a  enum time field */
struct dbr_time_enum {       // DBR_TIME_ENUM
  dbr_short_t     status;          /* status of value */
  dbr_short_t     severity;        /* severity of alarm */
  epicsTimeStamp  stamp;           /* time stamp */
  dbr_short_t     RISC_pad;        /* RISC alignment */
  dbr_enum_t      value;           /* current value */
};

/* structure for a char time field */
struct dbr_time_char {       // DBR_TIME_CHAR
  dbr_short_t     status;          /* status of value */
  dbr_short_t     severity;        /* severity of alarm */
  epicsTimeStamp  stamp;           /* time stamp */
  dbr_short_t     RISC_pad0;       /* RISC alignment */
  dbr_char_t      RISC_pad1;       /* RISC alignment */
  dbr_char_t      value;           /* current value */
};

/* structure for a long time field */
struct dbr_time_long {       // DBR_TIME_LONG
  dbr_short_t     status;          /* status of value */
  dbr_short_t     severity;        /* severity of alarm */
  epicsTimeStamp  stamp;           /* time stamp */
  dbr_long_t      value;           /* current value */
};

/* structure for a double time field */
struct dbr_time_double {     // DBR_TIME_DOUBLE
  dbr_short_t     status;          /* status of value */
  dbr_short_t     severity;        /* severity of alarm */
  epicsTimeStamp  stamp;           /* time stamp */
  dbr_long_t      RISC_pad;        /* RISC alignment */
  dbr_double_t    value;           /* current value */
};


//===== VALUES WITH STATUS AND GRAPHIC STRUCTURES =====

/* structure for a graphic string */
  /* not implemented; use struct dbr_sts_string */

/* structure for a graphic short field */
struct dbr_gr_short {        // DBR_GR_SHORT, DBR_GR_INT
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_short_t  upper_disp_limit;   /* upper limit of graph */
  dbr_short_t  lower_disp_limit;   /* lower limit of graph */
  dbr_short_t  upper_alarm_limit;
  dbr_short_t  upper_warning_limit;
  dbr_short_t  lower_warning_limit;
  dbr_short_t  lower_alarm_limit;
  dbr_short_t  value;              /* current value */
};

/* structure for a graphic floating point field */
struct dbr_gr_float {        // DBR_GR_FLOAT
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_short_t  precision;          /* number of decimal places */
  dbr_short_t  RISC_pad0;          /* RISC alignment */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_float_t  upper_disp_limit;   /* upper limit of graph */
  dbr_float_t  lower_disp_limit;   /* lower limit of graph */
  dbr_float_t  upper_alarm_limit;
  dbr_float_t  upper_warning_limit;
  dbr_float_t  lower_warning_limit;
  dbr_float_t  lower_alarm_limit;
  dbr_float_t  value;              /* current value */
};

/* structure for a graphic enumeration field */
struct dbr_gr_enum {         // DBR_GR_ENUM
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_short_t  no_str;             /* number of strings */
  char    strs[MAX_ENUM_STATES][MAX_ENUM_STRING_SIZE];  /* state strings */
  dbr_enum_t   value;              /* current value */
};

/* structure for a graphic char field */
struct dbr_gr_char {         // DBR_GR_CHAR
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_char_t   upper_disp_limit;   /* upper limit of graph */
  dbr_char_t   lower_disp_limit;   /* lower limit of graph */
  dbr_char_t   upper_alarm_limit;
  dbr_char_t   upper_warning_limit;
  dbr_char_t   lower_warning_limit;
  dbr_char_t   lower_alarm_limit;
  dbr_char_t   RISC_pad;           /* RISC alignment */
  dbr_char_t   value;              /* current value */
};

/* structure for a graphic long field */
struct dbr_gr_long {         // DBR_GR_LONG
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_long_t   upper_disp_limit;   /* upper limit of graph */
  dbr_long_t   lower_disp_limit;   /* lower limit of graph */
  dbr_long_t   upper_alarm_limit;
  dbr_long_t   upper_warning_limit;
  dbr_long_t   lower_warning_limit;
  dbr_long_t   lower_alarm_limit;
  dbr_long_t   value;              /* current value */
};

/* structure for a graphic double field */
struct dbr_gr_double {       // DBR_GR_DOUBLE
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_short_t  precision;          /* number of decimal places */
  dbr_short_t  RISC_pad0;          /* RISC alignment */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_double_t  upper_disp_limit;  /* upper limit of graph */
  dbr_double_t  lower_disp_limit;  /* lower limit of graph */
  dbr_double_t  upper_alarm_limit;
  dbr_double_t  upper_warning_limit;
  dbr_double_t  lower_warning_limit;
  dbr_double_t  lower_alarm_limit;
  dbr_double_t  value;             /* current value */
};



//===== VALUES WITH STATUS, GRAPHIC and CONTROL STRUCTURES =====

/* structure for a control string */
  /* not implemented; use struct dbr_sts_string */

/* structure for a control integer */
struct dbr_ctrl_short {      // DBR_CTRL_SHORT, DBR_CTRL_INT
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_short_t  upper_disp_limit;   /* upper limit of graph */
  dbr_short_t  lower_disp_limit;   /* lower limit of graph */
  dbr_short_t  upper_alarm_limit;
  dbr_short_t  upper_warning_limit;
  dbr_short_t  lower_warning_limit;
  dbr_short_t  lower_alarm_limit;
  dbr_short_t  upper_ctrl_limit;   /* upper control limit */
  dbr_short_t  lower_ctrl_limit;   /* lower control limit */
  dbr_short_t  value;              /* current value */
};

/* structure for a control floating point field */
struct dbr_ctrl_float {      // DBR_CTRL_FLOAT
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_short_t  precision;          /* number of decimal places */
  dbr_short_t  RISC_pad;           /* RISC alignment */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_float_t  upper_disp_limit;   /* upper limit of graph */
  dbr_float_t  lower_disp_limit;   /* lower limit of graph */
  dbr_float_t  upper_alarm_limit;
  dbr_float_t  upper_warning_limit;
  dbr_float_t  lower_warning_limit;
  dbr_float_t  lower_alarm_limit;
  dbr_float_t  upper_ctrl_limit;   /* upper control limit */
  dbr_float_t  lower_ctrl_limit;   /* lower control limit */
  dbr_float_t  value;              /* current value */
};

/* structure for a control enumeration field */
struct dbr_ctrl_enum {       // DBR_CTRL_ENUM
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  dbr_short_t  no_str;             /* number of strings */
  char  strs[MAX_ENUM_STATES][MAX_ENUM_STRING_SIZE];  /* state strings */
  dbr_enum_t   value;              /* current value */
};

/* structure for a control char field */
struct dbr_ctrl_char {       // DBR_CTRL_CHAR
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_char_t   upper_disp_limit;   /* upper limit of graph */
  dbr_char_t   lower_disp_limit;   /* lower limit of graph */
  dbr_char_t   upper_alarm_limit;
  dbr_char_t   upper_warning_limit;
  dbr_char_t   lower_warning_limit;
  dbr_char_t   lower_alarm_limit;
  dbr_char_t   upper_ctrl_limit;   /* upper control limit */
  dbr_char_t   lower_ctrl_limit;   /* lower control limit */
  dbr_char_t   RISC_pad;           /* RISC alignment */
  dbr_char_t   value;              /* current value */
};

/* structure for a control long field */
struct dbr_ctrl_long {       // DBR_CTRL_LONG
  dbr_short_t  status;             /* status of value */
  dbr_short_t  severity;           /* severity of alarm */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_long_t   upper_disp_limit;   /* upper limit of graph */
  dbr_long_t   lower_disp_limit;   /* lower limit of graph */
  dbr_long_t   upper_alarm_limit;
  dbr_long_t   upper_warning_limit;
  dbr_long_t   lower_warning_limit;
  dbr_long_t   lower_alarm_limit;
  dbr_long_t   upper_ctrl_limit;   /* upper control limit */
  dbr_long_t   lower_ctrl_limit;   /* lower control limit */
  dbr_long_t   value;              /* current value */
};

/* structure for a control double field */
struct dbr_ctrl_double {     // DBR_CTRL_DOUBLE
  dbr_short_t   status;            /* status of value */
  dbr_short_t   severity;          /* severity of alarm */
  dbr_short_t   precision;         /* number of decimal places */
  dbr_short_t   RISC_pad0;         /* RISC alignment */
  char    units[MAX_UNITS_SIZE];   /* units of value */
  dbr_double_t  upper_disp_limit;  /* upper limit of graph */
  dbr_double_t  lower_disp_limit;  /* lower limit of graph */
  dbr_double_t  upper_alarm_limit;
  dbr_double_t  upper_warning_limit;
  dbr_double_t  lower_warning_limit;
  dbr_double_t  lower_alarm_limit;
  dbr_double_t  upper_ctrl_limit;  /* upper control limit */
  dbr_double_t  lower_ctrl_limit;  /* lower control limit */
  dbr_double_t  value;             /* current value */
};


//===== MISC =====

/* structure for a  string status and ack field */
struct dbr_stsack_string {   // DBR_STSACK_STRING
  dbr_ushort_t  status;            /* status of value */
  dbr_ushort_t  severity;          /* severity of alarm */
  dbr_ushort_t  ackt;              /* ack transient? */
  dbr_ushort_t  acks;              /* ack severity */
  dbr_string_t  value;             /* current value */
};


// CA Access Rights
#define CA_NO_ACCESS          0
#define CA_READ_ACCESS        1
#define CA_WRITE_ACCESS       2
#define CA_READ_WRITE_ACCESS  3



//--- Bitmasks settable via the "diag msgs" command.  These generate ---
//--- diagnostic messages that are written to syslog ---
#define  ShowUDPPkts  (DiagFlags_ & 0x01)
#define  ShowTCPPkts  (DiagFlags_ & 0x02)
#define  ShowUDPCmds  (DiagFlags_ & 0x04)
#define  ShowTCPCmds  (DiagFlags_ & 0x08)

#define  ShowSearch   (DiagFlags_ & 0x10)
#define  ShowCons     (DiagFlags_ & 0x20)
#define  ShowEvents   (DiagFlags_ & 0x40)




/*
//--- Info about a CA Channel ---
typedef struct  {
  uint  pvid;  // PV ID value returned by pvGetID()  (0xFFFF ==> unused)
  ulong  cid;  // client value provided by the client
} CA_Chan;
*/

typedef  uint  PVID;
typedef  uint  SID;

#define  CHAN_UNUSED  ((PVID)(-1))
#define  SID_UNUSED   ((SID)(-1))


//--- Info about a CA Subscription ---
typedef struct  {
  SID sid;         // chan SID (idx in to CA_Con.chans[])  (0xFFFF ==> unused)
  ulong subID;     // SubscriptionID provided by the client
  uchar monMask;   // monitor mask provided by the client
  uchar dbrType;   // format for asynchronous event messages
  uint  numElems;  // # of elements
} CA_Sub;

//--- Info about a CA connection (Virtual Circuit) ---
typedef struct  {
  fd_t   fd;                       // fd for connection; <0 means unused entry
  bool   subEnabled;               // Subscriptions enabled/disabled flag
  PVID   chans[CA_MAX_CHANS];      // pvIDs for open Chans: (0xFFFF ==> unused)
  uint   maxChanNum;               // largest index (+1) used in chans[]
  CA_Sub subs[CA_MAX_SUBSCR];      // list of Subscriptions
  uint   maxSubNum;                // largest index (+1) used in subs[]
  char   hostName[HOST_NAME_LEN];  // host name provided by client
  char   userName[USER_NAME_LEN];  // user name provided by client
  SAI    in_Addr;                  // client addr
} CA_Con;

CA_Con far caCons[CA_MAX_CONS];



typedef struct  {
  fd_t   fd;          // socket fd to read from/write to
  uint   type;        // specifies what (if any) I/O buffers to use
  CA_Con far *con;    // NULL for UDP connection
  bool   showCmds;
  bool   versRespPending;
} ConInfo;

//--- values for ConInfo.type ---
#define  UDP_CON  0    // use rcvdPkt[] and respPktUDP[] to buffer input/output
#define  TCP_BUF  1    // use respPktTCP[] to buffer output
#define  TCP_DIR  2    // read/write directly from/to ConInfo.fd


//--- rcv and send buffers for UDP messages ---
uchar far rcvdPkt[IO_BUF_SIZE];  // for msgs rcvd in a UDP broadcast packet
uint  rcvdPktSize;        // size of packet currently in rcvdPkt
uint  nextRcvdPktByte;    // offset of next unused byte in rcvdPkt[]

uchar far respPktUDP[IO_BUF_SIZE];  // buffer for msgs sent to a client via UDP
uint  respPktUsedUDP;        // # bytes used/ready to send
SAI  respAddrUDP;            // ptr to net address to send respPkt to


//--- send buffer for smaller PV values ---
uchar far respPktTCP[IO_BUF_SIZE];  // buffer for msgs sent to a client via TCP
uint  respPktUsedTCP;        // # bytes used/ready to send



//--- List of IP addresses for which we will process UDP packets ---
ulong  allowedClients[CA_MAX_CONS];  // stored in network-byte order




typedef struct caMsgHdr {
  U16  cmd;    // see values below
  U16  size;   // # bytes sent after header that are part of this msg
  U16  fld3;   // msg dependent
  U16  fld4;   // msg dependent
  U32  fld5;   // msg dependent
  U32  fld6;   // msg dependent
  //---------- end of header -------------
  uchar  data[128];   // for faster responses when payload is small enough
  int    dataSize;    // # of bytes in data[]
}  CA_MsgHdr;


// max # payload size that will fit in the CA_MsgHdr structure.  For payloads
// larger than this (e.g. arrays, waveform data, long strings), the reading
// and writing of the payload has to be handled separately from the msg header
#define  MAX_MSG_DATA    ( sizeof(((CA_MsgHdr *)0)->data) )


//--- CA_MsgHdr.cmd values (msg types) ---
#define CA_PROTO_VERSION         0u   // set minor version and priority (used to be NOOP cmd)
#define CA_PROTO_EVENT_ADD       1u   // add an event
#define CA_PROTO_EVENT_CANCEL    2u   // cancel an event
#define CA_PROTO_READ            3u   // --- depricated ---
#define CA_PROTO_WRITE           4u   // write a channel value
#define CA_PROTO_SNAPSHOT        5u   // --- obsolete ---
#define CA_PROTO_SEARCH          6u   // IOC channel search
#define CA_PROTO_BUILD           7u   // --- obsolete ---
#define CA_PROTO_EVENTS_OFF      8u   // flow control
#define CA_PROTO_EVENTS_ON       9u   // flow control
#define CA_PROTO_READ_SYNC       10u  // --- depricated ---
#define CA_PROTO_ERROR           11u  // an operation failed
#define CA_PROTO_CLEAR_CHANNEL   12u  // free chan resources
#define CA_PROTO_RSRV_IS_UP      13u  // CA server has joined the net
#define CA_PROTO_NOT_FOUND       14u  // channel not found
#define CA_PROTO_READ_NOTIFY     15u  // add a one shot event
#define CA_PROTO_READ_BUILD      16u  // --- obsolete ---
#define REPEATER_CONFIRM         17u  // registration confirmation
#define CA_PROTO_CREATE_CHAN     18u  // client creates channel in server
#define CA_PROTO_WRITE_NOTIFY    19u  // notify after write chan value
#define CA_PROTO_CLIENT_NAME     20u  // CA V4.1 identify client
#define CA_PROTO_HOST_NAME       21u  // CA V4.1 identify client
#define CA_PROTO_ACCESS_RIGHTS   22u  // CA V4.2 asynch access rights chg
#define CA_PROTO_ECHO            23u  // CA V4.3 connection verify
#define REPEATER_REGISTER        24u  // register for repeater fan out
#define CA_PROTO_SIGNAL          25u  // --- obsolete ---
#define CA_PROTO_CREATE_CH_FAIL  26u  // unable to create chan resource in server
#define CA_PROTO_SERVER_DISCONN  27u  // server deletes PV (or channel)

#define CA_PROTO_LAST_CMMD  CA_PROTO_SERVER_DISCONN



// prototypes for funcs to process CA commands
#define  CMD_FUNC_ARGS   ConInfo *ci, CA_MsgHdr *msg

#define  PROTO_CMD_FUNC(name)    int name(CMD_FUNC_ARGS)

PROTO_CMD_FUNC( CA_Version       );
PROTO_CMD_FUNC( CA_EventAdd      );
PROTO_CMD_FUNC( CA_EventCancel   );
PROTO_CMD_FUNC( CA_Write         );
PROTO_CMD_FUNC( CA_Search        );
PROTO_CMD_FUNC( CA_EventsOff     );
PROTO_CMD_FUNC( CA_EventsOn      );
PROTO_CMD_FUNC( CA_ClearChannel  );
PROTO_CMD_FUNC( CA_ReadNotify    );
PROTO_CMD_FUNC( CA_CreateChan    );
PROTO_CMD_FUNC( CA_ClientName    );
PROTO_CMD_FUNC( CA_HostName      );
PROTO_CMD_FUNC( CA_Echo          );

PROTO_CMD_FUNC( ProtocolError    );


typedef int (* CA_CmdFunc)(CMD_FUNC_ARGS);

const far CA_CmdFunc  cmdFuncs[] = {
  CA_Version,
  CA_EventAdd,
  CA_EventCancel,
  CA_ReadNotify,   // CA_Read (depricated)
  CA_Write,
  ProtocolError,   // CA_Snapshot (obsolete)
  CA_Search,
  ProtocolError,   // CA_Build (obsolete)
  CA_EventsOff,
  CA_EventsOn,
  ProtocolError,   // CA_ReadSync (depricated)
  ProtocolError,   // CA_Error  (not sent by clients)
  CA_ClearChannel,
  ProtocolError,   // CA_RsrvIsUp  (not sent by clients)
  ProtocolError,   // CA_NotFound  (not sent by clients)
  CA_ReadNotify,
  ProtocolError,   // CA_ReadBuild (obsolete)
  ProtocolError,   // RepeaterConfirm  (shouldn't be sent to DEBROS)
  CA_CreateChan,
  CA_Write,        // CA_WriteNotify uses same func as CA_Write
  CA_ClientName,
  CA_HostName,
  ProtocolError,   // CA_AccessRights  (not sent by clients)
  CA_Echo,
  ProtocolError,   // RepeaterRegister  (shouldn't be sent to DEBROS)
  ProtocolError,   // CA_Signal (obsolete)
  ProtocolError,   // CA_CreateChFail  (not sent by clients)
  ProtocolError,   // CA_ServerDisconn  (not sent by clients)
};


//------------------------- from EPICS caerr.h --------------------------

/*  CA Status Code Definitions   */

#define CA_K_WARNING    0   /* unsuccessful */
#define CA_K_SUCCESS    1   /* successful */
#define CA_K_ERROR      2   /* failed- continue */
#define CA_K_INFO       3   /* successful */
#define CA_K_SEVERE     4   /* failed- quit */
#define CA_K_FATAL  CA_K_ERROR | CA_K_SEVERE

#define CA_M_MSG_NO     0x0000FFF8
#define CA_M_SEVERITY   0x00000007
#define CA_M_LEVEL      0x00000003
#define CA_M_SUCCESS    0x00000001
#define CA_M_ERROR      0x00000002
#define CA_M_SEVERE     0x00000004

#define CA_S_MSG_NO     0x0D
#define CA_S_SEVERITY   0x03

#define CA_V_MSG_NO     0x03
#define CA_V_SEVERITY   0x00
#define CA_V_SUCCESS    0x00

/* Define MACROS to extract/insert individual fields from a status value */

#define CA_EXTRACT_MSG_NO(code)    ( ((code) & CA_M_MSG_NO )  >> CA_V_MSG_NO   )
#define CA_EXTRACT_SEVERITY(code)  ( ((code) & CA_M_SEVERITY) >> CA_V_SEVERITY )
#define CA_EXTRACT_SUCCESS(code)   ( ((code) & CA_M_SUCCESS ) >> CA_V_SUCCESS  )

#define CA_INSERT_MSG_NO(code)     ( ((code)<< CA_V_MSG_NO)   &  CA_M_MSG_NO   )
#define CA_INSERT_SEVERITY(code)   ( ((code)<< CA_V_SEVERITY) &  CA_M_SEVERITY )
#define CA_INSERT_SUCCESS(code)    ( ((code)<< CA_V_SUCCESS)  &  CA_M_SUCCESS  )

#define DEFMSG(SEVERITY, NUMBER)  \
                    ( CA_INSERT_SEVERITY(SEVERITY) | CA_INSERT_MSG_NO(NUMBER) )

/*
 * In the lines below "defunct" indicates that current release
 * servers and client library will not return this error code, but
 * servers on earlier releases that communicate with current clients
 * might still generate exceptions with these error constants
 */
#define ECA_NORMAL          DEFMSG(CA_K_SUCCESS,    0) /* success */
#define ECA_MAXIOC          DEFMSG(CA_K_ERROR,      1) /* defunct */
#define ECA_UKNHOST         DEFMSG(CA_K_ERROR,      2) /* defunct */
#define ECA_UKNSERV         DEFMSG(CA_K_ERROR,      3) /* defunct */
#define ECA_SOCK            DEFMSG(CA_K_ERROR,      4) /* defunct */
#define ECA_CONN            DEFMSG(CA_K_WARNING,    5) /* defunct */
#define ECA_ALLOCMEM        DEFMSG(CA_K_WARNING,    6)
#define ECA_UKNCHAN         DEFMSG(CA_K_WARNING,    7) /* defunct */
#define ECA_UKNFIELD        DEFMSG(CA_K_WARNING,    8) /* defunct */
#define ECA_TOLARGE         DEFMSG(CA_K_WARNING,    9)
#define ECA_TIMEOUT         DEFMSG(CA_K_WARNING,   10)
#define ECA_NOSUPPORT       DEFMSG(CA_K_WARNING,   11) /* defunct */
#define ECA_STRTOBIG        DEFMSG(CA_K_WARNING,   12) /* defunct */
#define ECA_DISCONNCHID     DEFMSG(CA_K_ERROR,     13) /* defunct */
#define ECA_BADTYPE         DEFMSG(CA_K_ERROR,     14)
#define ECA_CHIDNOTFND      DEFMSG(CA_K_INFO,      15) /* defunct */
#define ECA_CHIDRETRY       DEFMSG(CA_K_INFO,      16) /* defunct */
#define ECA_INTERNAL        DEFMSG(CA_K_FATAL,     17)
#define ECA_DBLCLFAIL       DEFMSG(CA_K_WARNING,   18) /* defunct */
#define ECA_GETFAIL         DEFMSG(CA_K_WARNING,   19)
#define ECA_PUTFAIL         DEFMSG(CA_K_WARNING,   20)
#define ECA_ADDFAIL         DEFMSG(CA_K_WARNING,   21) /* defunct */
#define ECA_BADCOUNT        DEFMSG(CA_K_WARNING,   22)
#define ECA_BADSTR          DEFMSG(CA_K_ERROR,     23)
#define ECA_DISCONN         DEFMSG(CA_K_WARNING,   24)
#define ECA_DBLCHNL         DEFMSG(CA_K_WARNING,   25)
#define ECA_EVDISALLOW      DEFMSG(CA_K_ERROR,     26)
#define ECA_BUILDGET        DEFMSG(CA_K_WARNING,   27) /* defunct */
#define ECA_NEEDSFP         DEFMSG(CA_K_WARNING,   28) /* defunct */
#define ECA_OVEVFAIL        DEFMSG(CA_K_WARNING,   29) /* defunct */
#define ECA_BADMONID        DEFMSG(CA_K_ERROR,     30)
#define ECA_NEWADDR         DEFMSG(CA_K_WARNING,   31) /* defunct */
#define ECA_NEWCONN         DEFMSG(CA_K_INFO,      32) /* defunct */
#define ECA_NOCACTX         DEFMSG(CA_K_WARNING,   33) /* defunct */
#define ECA_DEFUNCT         DEFMSG(CA_K_FATAL,     34) /* defunct */
#define ECA_EMPTYSTR        DEFMSG(CA_K_WARNING,   35) /* defunct */
#define ECA_NOREPEATER      DEFMSG(CA_K_WARNING,   36) /* defunct */
#define ECA_NOCHANMSG       DEFMSG(CA_K_WARNING,   37) /* defunct */
#define ECA_DLCKREST        DEFMSG(CA_K_WARNING,   38) /* defunct */
#define ECA_SERVBEHIND      DEFMSG(CA_K_WARNING,   39) /* defunct */
#define ECA_NOCAST          DEFMSG(CA_K_WARNING,   40) /* defunct */
#define ECA_BADMASK         DEFMSG(CA_K_ERROR,     41)
#define ECA_IODONE          DEFMSG(CA_K_INFO,      42)
#define ECA_IOINPROGRESS    DEFMSG(CA_K_INFO,      43)
#define ECA_BADSYNCGRP      DEFMSG(CA_K_ERROR,     44)
#define ECA_PUTCBINPROG     DEFMSG(CA_K_ERROR,     45)
#define ECA_NORDACCESS      DEFMSG(CA_K_WARNING,   46)
#define ECA_NOWTACCESS      DEFMSG(CA_K_WARNING,   47)
#define ECA_ANACHRONISM     DEFMSG(CA_K_ERROR,     48)
#define ECA_NOSEARCHADDR    DEFMSG(CA_K_WARNING,   49)
#define ECA_NOCONVERT       DEFMSG(CA_K_WARNING,   50)
#define ECA_BADCHID         DEFMSG(CA_K_ERROR,     51)
#define ECA_BADFUNCPTR      DEFMSG(CA_K_ERROR,     52)
#define ECA_ISATTACHED      DEFMSG(CA_K_WARNING,   53)
#define ECA_UNAVAILINSERV   DEFMSG(CA_K_WARNING,   54)
#define ECA_CHANDESTROY     DEFMSG(CA_K_WARNING,   55)
#define ECA_BADPRIORITY     DEFMSG(CA_K_ERROR,     56)
#define ECA_NOTTHREADED     DEFMSG(CA_K_ERROR,     57)
#define ECA_16KARRAYCLIENT  DEFMSG(CA_K_WARNING,   58)
#define ECA_CONNSEQTMO      DEFMSG(CA_K_WARNING,   59)
#define ECA_UNRESPTMO       DEFMSG(CA_K_WARNING,   60)



//----------------- Platform-specific definitions ---------------


#define  SleepSec(sec)                sleep(sec)
#define  SleepMS(ms)                  sleepMS(ms)
#define  Socket(domain, type, prot)   socket(domain, type, prot)
//#define  Close(sock)                  close(sock)

#ifdef RABBIT
  #define  Sethostname(nam)  SetHostName((nam), strlen(nam))
  #define  SIN_ADDR  sin_addr
#else  //finish
  #define  SIN_ADDR  sin_addr.s_addr
  #define  FAR
  static short  timeZone = -4;
  static short  devType = 1;
  static long  deviceID = 0x9999;
  static ulong  DiagFlags = ErrorBits;

  typedef struct {
    ulong   ipAddr;   // in host byte order
    ulong   bcAddr;   // in host byte order
    ulong   netmask;  // in host byte order
    uchar   macAddr[6];
    uint    mtu;
    uchar   state;   //  IF_DOWN, IF_COMING_UP, IF_UP, IF_COMING_DOWN
  } IFConfig;


  static int getIFConfig(int iface, IFConfig *ifConfig)
  {
    memset(ifConfig, 0, sizeof(IFConfig));

    //finish
    ifConfig->state = 2;
    inet_ptoh(AF_INET, "192.168.4.34", &ifConfig->ipAddr);
    inet_ptoh(AF_INET, "192.168.255.255", &ifConfig->bcAddr);
    inet_ptoh(AF_INET, "255.255.0.0",     &ifConfig->netmask);
    ifConfig->mtu = 1500;

    return 0;
  }

  #define  IF_ETH0  0

#endif


//-----------------------------------------------------------------------------
//  Global variables
//-----------------------------------------------------------------------------


char  servName[HostNameLen];
int  servNameLen;
bool  servNameMatch;  // last call to getPVID() matched at least the servName


bool  processUDP = true;

//--- Values for valChgMode, which determines what criteria is used to ---
//--- generate event monitor messages for channels ---
#define  VCM_PERIODIC    0   // at a fixed rate even if no change in value
#define  VCM_THRESHOLD   1   // only if change exceeds the set threshold
#define  VCM_ANY         2   // for any change, no matter how small

uint  valChgMode = VCM_THRESHOLD;

static const char *chgModeDesc[] = { "periodic", "threshold", "anychg" };


uint  eventPeriod = CA_EVENT_PERIOD;  // default # ms between sending Event Monitor msgs

ulong  beaconID;

uint  numUDPPkts, numProcPkts, numSearchMsgs, numEvents, numWrites, numPSub,
      udpPktRate, procPktRate, searchMsgRate, eventsRate, writesRate, psubRate;

uint  consecutivePkts;  //tdebug

uint  lastBeaconTime;

PID  ca_tcp_pid = 0, ca_udp_pid = 0;

ulong  epochTimeSecs, epochTimeNS;


//-----------------------------------------------------------------------------
//  Show/log msg about the connection currently being processed.
//-----------------------------------------------------------------------------
void conMsg(int fd, const char far *msg)
{
  FDEV  far *file;
  void  *ioDev;
  char  buf[128];


  file = _FileFromFD_(fd);  ioDev = File_Dev(file);
  file->ioFuncs->descr(ioDev, buf, sizeof(buf));

  printf("\r\n%ls on fd: %d (GFD: %d)\r\n%.128s\n\n", msg, fd, GFD(fd), buf );
}


//-----------------------------------------------------------------------------
void closeCon(ConInfo *ci)
{
  conMsg(ci->fd,
         (errno == ENOLINK) ?
            "Connection closed" : "*** Closing due to error ***");

  close(ci->fd);

  if (ci->con)  ci->con->fd = -1;
}


//-----------------------------------------------------------------------------
//  Return the next 'numBytes' bytes of msg data for the specified connection.
//
//  Uses global variables:
//     nextRcvdPktByte, rcvdPktSize,
//
//  Returns the # of bytes actually read (which may be < numBytes) or < 0 if
//  an error.
//-----------------------------------------------------------------------------
int readBytes(ConInfo *ci, char *buf, int numBytes, long timeout)
{
  uint  bytesLeft;


  if (numBytes < 1)  { errno = EINVAL;  return -1; }

//--- if not a UDP connection, read it directly from the socket ---
  if (ci->type != UDP_CON)  return read_(ci->fd, buf, numBytes, timeout);


//--- Otherwise, copy from the next unprocessed bytes in rcvdPkt[] ---
  if (nextRcvdPktByte > rcvdPktSize)  { errno = ENODATA;  return -1; }

  bytesLeft = rcvdPktSize - nextRcvdPktByte;

  if (bytesLeft < numBytes)  numBytes = bytesLeft;

  if (!numBytes)  { errno = EAGAIN;  return -1; }

  fmemcpy(buf, rcvdPkt+nextRcvdPktByte, numBytes);
  nextRcvdPktByte += numBytes;

  return numBytes;
}


//-----------------------------------------------------------------------------
//  Send a response to a msg rcvd via UDP or TCP
//-----------------------------------------------------------------------------
int sendResp(ConInfo *ci, uchar far *buf, uint count)
{
  int  stat, addrLen;
  SAP  addr;
  char  temp[64], addrStr[16];


  if (ci->type == UDP_CON)  {
    addr = (SAP)&respAddrUDP;  addrLen = sizeof(SAI);  }
  else {
    addr = NULL;  addrLen = 0;  }


  while (count > 0)  {
    stat = sendto(ci->fd, buf, count, 0, addr, addrLen);

    if (stat == count)  break;

    if (stat < 0)  {
      if (errno == EAGAIN)  { sleepMS(5);  continue; }
      inet_ntop(AF_INET, &addr->SIN_ADDR, addrStr, sizeof(addrStr));
      sprintf(temp, "\n*** %s:%u: sendto(%d, x, %d, ...)",
              addrStr, ntohs(addr->sin_port), ci->fd, count);
      perror(temp);  printf("\n");  return -1;
    }

    if (!stat)  { sleepMS(5);  continue; }

    // send some but not all, but UDP packets must NOT be fragmented!!!
    if (ci->type == UDP_CON)  { errno = EMSGSIZE;  return -1; }

    buf += (uint)stat;  count -= stat;
  }

  return 0;
}


//-----------------------------------------------------------------------------
//  Send the contents of the outgoing packet buffer, which contains one or
//  more messages (either responses to one or more request messages or
//  periodically generated event monitor messages with new values).
//
//  Note that if the responses are for requests received via a UDP broadcast
//  packet, then we must send the response to the address and port that the
//  packet was sent from.  For TCP, the socket is already connected and it
//  is considered an error to specify the destination (since it cannot be
//  anything other than the address it is already connected to).
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int sendRespPkt(ConInfo *ci)
{
  int  fd;
  static int  so_setBcast = true, so_clearBcast = false;


  fd = ci->fd;

  if (fd < 0)  { errno = EBADF;  return -1; }

  errno = 0;


//--- if not UDP, send the contents of the shorter-msg TCP resp buffer ---
  if (ci->type != UDP_CON)  {
    if (sendResp(ci, respPktTCP, respPktUsedTCP))  return -1;
    respPktUsedTCP = 0;  return 0;
  }


//--- For UDP packets, we need to temporarily disable broadcast mode ---
  if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST,
                (char *)&so_clearBcast, sizeof(so_clearBcast)) < 0)  {
      perror("UDP setsockopt()");  return -1;  }

//--- send the contents of the UDP resp buffer ---
  if (sendResp(ci, respPktUDP, respPktUsedUDP))  return -1;
  respPktUsedUDP = 0;

//--- Restore broadcast mode ---
  if (setsockopt(fd, SOL_SOCKET, SO_BROADCAST,
                 (char *)&so_setBcast, sizeof(so_setBcast)) < 0)  {
      perror("UDP setsockopt()");  return -1;  }

  return 0;
}


//-----------------------------------------------------------------------------
//  Process bytes to be sent to the client either by sending them now or
//  adding them to a buffer that will be sent when it is full or on demand by
//  a call to flushResps().
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int _nearcall addToRespPkt(ConInfo *ci, void far *buf, uint numBytes)
{
  uint  count, unused;


  switch (ci->type)  {

  /*---------------------------------------------------------------------*
    For UDP, we MUST accumulate msgs in a buffer because we have to be
    sent in 1 packet.  If there isn't enough space left, it is an error.
  *----------------------------------------------------------------------*/
    case UDP_CON:
      if (numBytes > sizeof(respPktUDP) - respPktUsedUDP)  return -1;
      fmemcpy(respPktUDP + respPktUsedUDP, buf, numBytes);
      respPktUsedUDP += numBytes;  return 0;

  /*---------------------------------------------------------------------*
    For TCP connections that are using only smaller PVs, the outgoing
    msgs consist mostly of new values for subscriptions, which means we
    generate a whole bunch of small msgs fairly quickly.  In this case,
    we add the msgs to a buffer and flush it as needed.  For these shorter
    messages, this is worth the extra overhead for 2 reasons:

      - It minimizes the # of calls to the socket layer
      - It minimizes the # of packets the responses are sent in

    The naggle algorithm in the network stack helps with the second issue,
    but we generally don't process or generate msgs fast enough for it to
    be very effective, so a series of messages still tends to get broken
    up in to multiple packets (and more packets means more CPU time spent
    sending them).

    The end result:  Even though we add another buffer to the sequence of
    buffer-to-buffer copies that occurs before a msg is actually sent, the
    result is still greater throughput and less CPU time than if we don't.
  *----------------------------------------------------------------------*/
    case TCP_BUF:
      unused = sizeof(respPktTCP) - respPktUsedTCP;
      for (;;)  {
        count = (numBytes > unused) ? unused : numBytes;
        fmemcpy(respPktTCP + respPktUsedTCP, buf, count);
        respPktUsedTCP += count;
        if (! (numBytes -= count))  break;  // nothing left to copy
        if (sendRespPkt(ci))  return -1;  // buffer must be full, so send it
        unused = sizeof(respPktTCP);  // buffer now empty
        buf = (uchar far *)buf + count;  // but still more to copy to it
      }
      return 0;

  /*---------------------------------------------------------------------*
    The last case is connections that use larger PV values.  For these,
    copying the data to another buffer rather than just writing directly
    to the socket just uses up more memory with no gain.  Depending on the
    size of the messages, it can result in a net loss to add another
    buffer-to-buffer copy for each message.  So for this case, we just
    write the data directly to the socket.
  *----------------------------------------------------------------------*/
    case TCP_DIR:
      return sendResp(ci, buf, numBytes);
  }


  return -1;
}


//-----------------------------------------------------------------------------
//  "skip over" the specified number of bytes in the current msg data stream.
//  The main purpose of this is to skip over the payload for a msg so we are
//  ready to process the next msg.
//
//  Returns the # of bytes actually skipped (which may be < numBytes) or < 0
//  if an error.
//-----------------------------------------------------------------------------
int skipMsgBytes(ConInfo *ci, int numBytes)
{
  int  avail, count;


  if (numBytes < 1)  return 0;

  if (ShowErrors OR ci->showCmds)
    printf("*** skipping %d payload bytes ***\n", numBytes);

  if (ci->type != UDP_CON)  return skipBytes(ci->fd, numBytes, RD_TIMEOUT);

//--- NOTE: rcvdPkt is used ONLY for UDP packets ---
  avail = rcvdPktSize - nextRcvdPktByte;

  count = (avail > numBytes) ? avail : numBytes;

  nextRcvdPktByte += count;  return count;
}


//-----------------------------------------------------------------------------
//  Check the PV name for a name of the form host.field, where the host part
//  must match the hostname and the field one of the values in the lists of
//  values exposed to external clients.
//
//  If a match is found, it returns a value >= 0 that is used to reference the
//  value in many of the other PV interface functions.  If no match is found,
//  or there is an error, it returns < 0.
//-----------------------------------------------------------------------------
int getPVID(char *pvName, int pvNameLen)
{
  char  *fldName;


  fldName = NULL;  servNameMatch = false;

//--- first part of PV name must match the hostname ---
  if (strnicmp(servName, pvName, servNameLen))  return -1;

//--- if ALL we have is the host name then assume a fld name of "VAL" ---
  if (pvNameLen == servNameLen)
    fldName = "VAL";
//--- otherwise the rest must be a '.' followed by a field name ---
  else
    if (pvName[servNameLen] == '.')  fldName = pvName + servNameLen + 1;

  if (!fldName)  return -1;

  servNameMatch = true;

  return pvGetID(fldName);
}




//-----------------------------------------------------------------------------
//  Send a CA_PROTO_ERROR msg over a TCP connection
//
//  REturns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int sendErrResp(ConInfo *ci, CA_MsgHdr *msg, ulong cid,
                int errCode, const char far *descr)
{
  int  count;
  uint  descrLen;
  CA_MsgHdr  resp, *msgCopy;


  if (ShowErrors)
    printf("*** Error: fd:%d  cmd:[%2d]  CID:0x%08lX  code:%04X\n"
           "           %ls\n", ci->fd, msg->cmd, cid, errCode, descr);

  descrLen = (uint) fstrlen(descr);
  if (descrLen > MAX_MSG_DATA-17)  descrLen = MAX_MSG_DATA - 17;

  resp.cmd = htons(CA_PROTO_ERROR);
  resp.size = htons(16 + descrLen + 1);
  resp.fld3 = 0;
  resp.fld4 = 0;
  resp.fld5 = htonl(cid);
  resp.fld6 = htonl(errCode);

  msgCopy = (CA_MsgHdr *)resp.data;

  msgCopy->cmd  = htons(msg->cmd);
  msgCopy->size = htons(msg->size);
  msgCopy->fld3 = htons(msg->fld3);
  msgCopy->fld4 = htons(msg->fld4);
  msgCopy->fld5 = htonl(msg->fld5);
  msgCopy->fld6 = htonl(msg->fld6);

  fmemcpy(msgCopy->data, descr, descrLen);
  msgCopy->data[descrLen] = 0;

  count = resp.size + 16;

  return addToRespPkt(ci, &resp, count);
}

//-----------------------------------------------------------------------------
//  send a beacon out using the UDP socket configured for broadcast
//-----------------------------------------------------------------------------
int sendBeacon(int fd)
{
  IFConfig  ifConfig;
  CA_MsgHdr  resp;


  getIFConfig(IF_ETH0, &ifConfig);

  resp.cmd  = htons(CA_PROTO_RSRV_IS_UP);
  resp.size = 0;
  resp.fld3 = htons(CA_MINOR_VERSION);  // doco says THIS should be the port!
  resp.fld4 = htons(CA_SERVER_PORT);    // doco says it MUST be 0!
  resp.fld5 = htonl(beaconID);
  resp.fld6 = htonl(ifConfig.ipAddr);

  if (sendBCastPkt(&fd, EPICS_CAS_BEACON_PORT, ifConfig.bcAddr,
                   (char far *)&resp, 16))  {
    perror("sendBCastPkt()");  return -1; }

  lastBeaconTime = GetMS_();  ++beaconID;

  return 0;
}


//-----------------------------------------------------------------------------
//  Validate the values that are required for a sendValue() call.
//
//  Returns 0 if they are all valid, < 0 if they are not.
//
//  If the ConInfo is valid but the dbrType or sid are not, it sends a
//  response with the BADTYPE or BADCHID status code.
//-----------------------------------------------------------------------------
int validateParams(ConInfo *ci, uint cmd, ulong eventID,
                   uint dbrType, uint numElems, SID sid)
{
  PVID  pvid;
  uint  statusCode = ECA_NORMAL;
  PVID far  *chan;
  PV_Info  pvInfo;
  CA_MsgHdr  resp;
  CA_Con far *con;


//--- Verify ConInfo refers to a TCP (Virtual Circuit) connection ---
  con = ci->con;  if (!con)  return -1;


//--- Verify dbrType is valid and doesn't exceed our limits ---
  if (dbrType > LAST_BUFFER_TYPE)
    statusCode = ECA_BADTYPE;
  else
    if (dbrValOffset[dbrType] > sizeof(CA_MsgHdr) - 32)  {
      if (ShowErrors)
        printf("*** Extended data type exceeds our capabilities ***\n");
      statusCode = ECA_UNAVAILINSERV;
    }


//--- Make sure SID is valid and get internal PV ID ---
  if (sid >= con->maxChanNum)  {
    pvid = CHAN_UNUSED;  chan = NULL; }
  else  {
    chan = con->chans + sid;
    if ((pvid = *chan) == CHAN_UNUSED)  chan = NULL;
  }

  if (!chan)
    statusCode = ECA_BADCHID;
  else
    if (pvGetInfo(pvid, &pvInfo))  statusCode = ECA_BADCHID;


//--- return 0 if ci, dbrType, and sid are all valid and usable ---
  if (statusCode != ECA_NORMAL)  return 0;


//--- Send msg with the error code ---
  resp.cmd  = htons(cmd);
  resp.size = 0;
  resp.fld3 = htons(dbrType);
  resp.fld4 = htons(numElems);
  resp.fld5 = htonl(statusCode);   // Doco says SID, but that is wrong!!! ***
  resp.fld6 = htonl(eventID);

  addToRespPkt(ci, &resp, 16);

  return -1;
}


//-----------------------------------------------------------------------------
//  Send the value of an existing channel.  Used for READ_NOTIFY msgs and for
//  immediate response to and subsequently generated EVENT_ADD msgs.
//
//  The eventID is either the subscription ID from an EVENT_ADD msg or the
//  IOID from a READ_NOTIFY msg.
//
//  WARNING:  The ci, dbrType, and sid values must all have been previously
//            validated (by validateParams() or equivalent checks).
//            Validation was removed from this function to reduce overhead for
//            the most frequent use:  To send updates for subscriptions.
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int sendValue(ConInfo *ci, uint cmd, ulong eventID,
              uint dbrType, uint numElems, SID sid)
{
  bool  sendRaw;
  PVID  pvid;
  uint  payloadSize, count, baseType, extType, statusCode, dataOffset,
        hdrPlusDataSize, payloadFill;
  uchar  *ptr;
  PVID far  *chan;
  div_t  qr;
  TypeData  typeData;
  PV_Info  pvInfo;
  CA_MsgHdr  resp;
  CA_Con far *con;

  if (ShowEvents OR ci->showCmds)  printf("    dbrType:%2u\n", dbrType);  //tdebug

//--- extract the base and extended types from the dbrType ---
  qr = div(dbrType, 7);  baseType = qr.rem;  extType = qr.quot;
#ifdef RABBIT
  if (baseType == DBR_DOUBLE)  baseType = DBR_FLOAT;
#endif
  if (extType > EXT_TYPE_TIME)  extType = EXT_TYPE_STS;
  dbrType = extType * 7 + baseType;

  if (ShowEvents OR ci->showCmds)  printf("    dbrType:%2u\n", dbrType);  //tdebug

//--- all this must have been validated before by call to validateParams() ---
  con = ci->con;  chan = con->chans + sid;  pvid = *chan;
  pvGetInfo(pvid, &pvInfo);


  statusCode = ECA_NORMAL;  sendRaw = false;

  payloadSize = 0;  ptr = resp.data;

//--- All extended DBR_xxx data structs start with status and severity ---
  if (extType)  {
    U16_(ptr) = htons(0);  ptr += 2;  // status
    U16_(ptr) = htons(0);  ptr += 2;  // severity
    if (extType == EXT_TYPE_TIME)  {
      U32_(ptr) = htonl(epochTimeSecs);  ptr += 4;
      U32_(ptr) = htonl(epochTimeNS);  ptr += 4;
    }
  }

//--- Add any filler bytes needed ---
  dataOffset = dbrValOffset[dbrType];
  count = dataOffset - (ptr - resp.data);
  while (count)  { --count;  *ptr++ = 0; }

//--- Can't return more elements than we have ---
  if (numElems > pvInfo.numElems)  numElems = pvInfo.numElems;

  /*-----------------------------------------------------------------------
   NOTE:
       For PVs with larger data values, we cannot simply copy the entire
       value to the message buffer and send it all at once.  To handle
       such values we need to send the header first then call a function
       to get the PV's current value (in pieces if necessary).  But the
       header tells the client how big the payload for the message is,
       which means we have to know NOW how many bytes will be sent.

       If the current PV value is not simply stored in a buffer, and/or we
       need to convert it to a format the client requested (or at least to
       something it can deal with), determining the correct value for the
       size may not be practical.  For example, if the value can only be
       determined by actually doing the conversion, then we might have to
       do it once now (to compute the correct size value) and again later
       after we have sent the message header.

       As even simple conversions can take a lot of CPU cycles for larger
       data sets, this is generally not a practical solution.  And if the
       result could be different each time, then there may not BE a way
       to determine the size the value will be the next time we read it.

       So, for PVs whose values that are too large to handle by simply
       copying the value to the message buffer, there must be a way to
       determine its current size in the format we will be sending it in.

       And because even simple conversions take far more CPU time than
       a direct buffer-to-buffer copy, conversion is not currently
       supported for these larger PV values.  Such conversion COULD be
       supported by providing a function to call that returns the size
       given the format we WILL send it in.  But for now, we are only
       handling sending such values in their "native" format and leaving
       it up to the client to deal with any conversion.
   ----------------------------------------------------------------------*/

  do  {

  //--- Is the value something we shouldn't attempt to or is too costly ---
  //--- to convert? ---

//    msgSpaceLeft = resp.data + sizeof(resp.data) - ptr - 8;
//    if ((numElems > 1) OR (pvInfo.size > msgSpaceLeft))  {

    if ((numElems > 1) OR (pvInfo.flags & PV_SEND_RAW))  {
      dbrType = pvTypeToDBF[pvInfo.dataType] + extType * 7;
      if (ShowEvents OR ci->showCmds)  printf("    dbrType:%2u\n", dbrType);  //tdebug
      hdrPlusDataSize = ptr - resp.data + numElems * pvInfo.elemSize;  // see NOTE above
      payloadSize = (hdrPlusDataSize + 7) & 0xFFF8; // must be mult of 8
      payloadFill = payloadSize - hdrPlusDataSize;  sendRaw = true;
      break;
    }

  // (NOTE: The rest of this block is skipped for large or complex values)

  //--- Add the current value for the PV converted to the requested format ---
    typeData = dbfToPV[baseType];
    if (pvGetVal(pvid, typeData.pvType, typeData.size, ptr, CUR_VALUE) < 0)  {
      if (ShowErrors)  printf("*** Unable to get or convert data ***\n");
      payloadSize = 0;  numElems = 0;  ptr = resp.data;
      statusCode = ECA_INTERNAL;  break;
    }

  //--- reorder the bytes if necessary ---
    if (typeData.reorder)  {
      if (typeData.size == 2)
        U16_(ptr) = htons(U16_(ptr));
      else
        U32_(ptr) = htonl(U32_(ptr));
    }

    ptr += typeData.size;


  //--- Compute total payload size (before filling out to mult of 8) ---
    payloadSize = ptr - resp.data;

  //--- Add filler until we have a multiple of 8 ---
    while (payloadSize & 7)  { ++payloadSize;  *ptr++ = 0; }

  } while (0);


  if (ShowEvents OR ci->showCmds OR ((statusCode != ECA_NORMAL) AND ShowErrors))
    printf("    Sending: dbrType:%2u  #elem:%u  payload: %u  SID:%3u  eventID:0x%08lX  status:%u  pvID:%3d [%ls]\n\n",
           dbrType, numElems, payloadSize, (uint)sid, eventID, statusCode, pvid, pvInfo.name);


//--- Send msg with an error code or as a prefix for the PV's current value. ---
//--- If the value is already part of the payload, then it gets sent too. ---
  resp.cmd  = htons(cmd);
  resp.size = htons(payloadSize);
  resp.fld3 = htons(dbrType);
  resp.fld4 = htons(numElems);
  resp.fld5 = htonl(statusCode);   // Doco says SID, but that is wrong!!! ***
  resp.fld6 = htonl(eventID);

  count = ptr - (uchar *)&resp;

  if (addToRespPkt(ci, &resp, count))  return -1;

  if (!sendRaw OR (statusCode != ECA_NORMAL))  return 0;


/*
  If we  got to this point, the PV was not small and simple enough to be
  handled by mearly copying its value to the fixed size buffer and doing
  whatever data conversion is required.  We have sent the header and any
  extended data structure prefix, but we still need to send the PV's value
  as the remainder of the payload.

  This is done by passing the PV layer a pointer to a func it must call to
  send the data (in this case, addToRespPkt).  The function does not return
  until it has processed all the bytes provided or an error occurs, so the PV
  layer can call it once with a pointer to the whole PV value, or multiple
  times, each time with a pointer to the next piece of the PV value.  This
  avoids the need for another buffer large enough to hold the entire value.
  And in cases where it is not too costly, the PV layer could even generate
  or transform the value a piece at a time (again, without the need for a
  single large buffer to hold the entire result).

  NOTE that, for the current version of this module, the CA data service will
  be blocked for however long it takes to send the value.  If the size and/or
  number of PVs grows to the point where this "blocking" approach becomes a
  performance issue, the method will need to be changed to work more like
  the NSCL DirectCA module which runs on the IOCs and never blocks during a
  read or write, minimizing the effect any one connection can have on all the
  others.  This is achieved by keeping track of the state of each connection
  and allowing the reading or writing of a single message to span multiple
  passes.

  Of course, the downside to the non-blocking approach is that it increases
  the time that larger or more complex values will be locked, which can
  affect the performance of other processes (including the main application
  task, which usually does have some important timing constraints).
*/

  if (pvCopyVal(pvid, numElems, (void *)ci, addToRespPkt))  return -1;

//--- Add filler if needed (computed earlier to insure the payload size is ---
//--- a multiple of 8 bytes, as required by the protocol) ---
  if (payloadFill)  {
    memset(resp.data, 0, payloadFill);
    if (addToRespPkt(ci, resp.data, payloadFill))  return -1;
  }

  return 0;
}


//-----------------------------------------------------------------------------
//  Send the CA protocol version # we are using
//-----------------------------------------------------------------------------
int sendVersion(ConInfo *ci)
{
  CA_MsgHdr  resp;


  memset(&resp, 0, sizeof(resp));

  resp.cmd  = htons(CA_PROTO_VERSION);
  resp.fld4 = htons(CA_MINOR_VERSION);

  if (addToRespPkt(ci, &resp, 16))  return -1;

  ci->versRespPending = false;

  return 0;
}


//=============================================================================
//************  funcs to handle specific CA protocol messages  ****************
//=============================================================================

//-----------------------------------------------------------------------------
//  CA_PROTO_VERSION:  Send the version of the protocol this server is using
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_Version )
{

  if (ci->showCmds)
    printf("VERSION: Client:4.%d  Server:4.%d\n",
           msg->fld4, CA_MINOR_VERSION);

  ci->versRespPending = true;

//--- All UDP broadcast packets with SEARCH msgs start with a VERSION msg, ---
//--- so we must NOT respond to a VERSION request rcvd via UDP broadcast. ---
//--- If all the CA servers did this, then every client would be bombarded ---
//--- with VERSION msg responses from all the CA servers. ---
  if (ci->type == UDP_CON)  return 0;

  return sendVersion(ci);
}

//-----------------------------------------------------------------------------
//  CA_PROTO_EVENT_ADD:  Create or Update a subscription on an existing channel
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_EventAdd )
{
  time_t  curSecs;
  int  stat;
  uint  pvid, curMS, u, dbrType, numElems, monMask, newSubNum, maxSubNum,
        errCode = 0;
  SID  sid;
  ulong  subID;
  CA_Sub far  *sub, *newSub;
  PVID far  *chan = NULL;
  CA_Con far *con;
  CA_MsgHdr  resp;
  const char far *errMsg = "ERROR";


  con = ci->con;

  dbrType = msg->fld3;
  numElems = msg->fld4;
  sid = (SID) msg->fld5;
  subID = msg->fld6;

  monMask = ntohs(U16_(msg->data + 12));

  if (!monMask OR (monMask != monMask & DBE_MASK_BITS))  {
    if (ShowErrors OR ci->showCmds)  errMsg = "Invalid Mask";
    errCode = ECA_BADMASK;
  }

  if (dbrType > LAST_BUFFER_TYPE)  {
    if (ShowErrors OR ci->showCmds)  errMsg = "Invalid DBR type";
    errCode = ECA_BADTYPE;
  }

  if (sid < con->maxChanNum)  chan = con->chans + sid;

  if (!chan OR (*chan == CHAN_UNUSED))  {
    if (ShowErrors OR ci->showCmds)  errMsg = "Invalid SID";
    errCode = ECA_BADCHID;
  }

//--- find an unused subscription slot or one with a matching ID ---
  newSub = NULL;  maxSubNum = con->maxSubNum;
  for (u=0,sub=con->subs; u<maxSubNum; ++u,++sub)  {
    if (!newSub)  if (sub->sid == CHAN_UNUSED)  newSub = sub;  // 1st unused slot
    if (sub->subID == subID)  {  // ID matches existing subscription ID
      if (sub->sid != sid)  {
        if (ShowErrors OR ci->showCmds)  errMsg = "SIDs don't match";
        errCode = ECA_BADCHID;
      }
      break;
    }
  }
  if (u == maxSubNum)  {  // no matching existing subscription
    if (!newSub AND (maxSubNum < CA_MAX_SUBSCR))  newSub = sub;
    sub = NULL;
  }


//--- No free slot and no match? ---
  if (!newSub AND !sub)  {
    if (ShowErrors OR ci->showCmds)  errMsg = "Max Subscriptions exceeded";
    errCode = ECA_ALLOCMEM;
  }


  if (ci->showCmds OR (errCode AND ShowErrors))  {
    printf("EVENT_ADD: mask:0x%02X  dbrType:%2u", monMask, dbrType);
    if (errCode)  printf("*** %ls ***\n", errMsg);
    if (ci->showCmds)  printf("\n");
  }


  if (errCode)  {
    resp.cmd = htons(CA_PROTO_EVENT_ADD);
    resp.size = 0;
    resp.fld3 = htons(dbrType);
    resp.fld4 = 0;
    resp.fld5 = htonl(errCode);
    resp.fld6 = htonl(subID);

    if (addToRespPkt(ci, &resp, 16))  return -1;

    return 0;
  }


//--- update existing or add new subscription ---
  if (sub)  {  // existing
    sub->monMask = monMask;
    sub->dbrType = dbrType;
    sub->numElems = numElems;
  }
  else  {  // new
    newSub->sid = sid;
    newSub->subID = subID;
    newSub->monMask = monMask;
    newSub->dbrType = dbrType;
    newSub->numElems = numElems;
    // track largest subscription # used
    newSubNum = (uint)(newSub - con->subs) + 1;
    if (newSubNum > con->maxSubNum)  con->maxSubNum = newSubNum;
  }


//--- update the mask used to generate subscription event msgs ---
  pvCheckEvents(*chan, monMask);

//--- send a msg with the PV's current value ---
  _EnterCritical_();  curSecs = TIME();  curMS = milliTicks;  _ExitCritical_();
  epochTimeSecs = curSecs - 631152000lu; // EPICS uses 1990-01-01 00:00:00
  epochTimeNS = curMS * 1000000lu;

  pvid = *chan;

  if (pvLockVal(pvid, LM_WaitLock, LT_READ) < 0)  return -1;
  stat = sendValue(ci, CA_PROTO_EVENT_ADD, subID, dbrType, numElems, sid);
  pvLockVal(pvid, LM_Unlock, LT_READ);

  return stat;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_EVENT_CANCEL:  Remove an event subscription created by an earlier
//  EVENT_ADD msg.
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_EventCancel )
{
  uint  u, maxSubNum;  //, dbrType, numElems;
  SID  sid;
  ulong  subID;
  CA_Sub far  *sub;
  CA_Con far *con;
  CA_MsgHdr  resp;


  con = ci->con;

//  dbrType = msg->fld3;
//  numElems = msg->fld4;
  sid = (SID) msg->fld5;
  subID = msg->fld6;

  if (ci->showCmds)
    printf("EVENT_CANCEL:  SID:0x%04X  subID:0x%08lX\r\n", sid, subID);

//--- Locate the subscription in the list for this connection ---
  maxSubNum = con->maxSubNum;
  for (u=0,sub=con->subs; u<maxSubNum; ++u,++sub)
    if (sub->subID == subID)  break;

//--- No subscription with the specified ID? ---
  if (u == maxSubNum)
    return sendErrResp(ci, msg, (ulong)(-1), ECA_BADMONID, "Invalid subID");

//--- send resp, which uses the EVENT_ADD cmd and values from the subscrip ---
  resp.cmd = htons(CA_PROTO_EVENT_ADD);
  resp.size = 0;
  resp.fld3 = htons(sub->dbrType);
  resp.fld4 = 0;
  resp.fld5 = htonl(sid);
  resp.fld6 = htonl(subID);

//--- clear the subscription ---
  fmemset(sub, 0xff, sizeof(CA_Sub));

  if (addToRespPkt(ci, &resp, 16))  return -1;

  return 0;
}



//-----------------------------------------------------------------------------
//  CA_PROTO_WRITE and CA_PROTO_WRITE_NOTIFY:
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_Write )
{
  bool  notify;
  uint  pvid, dbrType, numElems, baseType, errCode = 0; //, extType;
  SID  sid;
  uchar  *ptr;
  PVID far  *chan = NULL;
  CA_Con far *con;
  const char far *errMsg = "ERROR";
//  div_t  qr;
  TypeData  typeData;
  PV_Info  pvInfo;
  CA_MsgHdr  resp;


  con = ci->con;

  dbrType = msg->fld3;
  numElems = msg->fld4;
  sid = (SID) msg->fld5;

  notify = (msg->cmd == CA_PROTO_WRITE_NOTIFY);  // is a response expected?

  ++numWrites;

  if (dbrType > LAST_BUFFER_TYPE)  {
    errMsg = "Invalid DBR type";  errCode = ECA_BADTYPE; }
  else  {
//    qr = div(dbrType, 7);  baseType = qr.rem;  extType = qr.quot;
    baseType = dbrType % 7;
    ptr = msg->data + dbrValOffset[dbrType];
  }

  chan = (sid < con->maxChanNum) ? con->chans + sid : NULL;
  pvid = chan ? *chan : CHAN_UNUSED;

  if (pvid == CHAN_UNUSED)  { errMsg = "Invalid SID";  errCode = ECA_BADCHID; }

  if (ci->showCmds OR (errCode AND ShowErrors))  {
    if (pvGetInfo(pvid, &pvInfo))  pvInfo.name = "<invalid>";
    printf("WRITE%s:  dbrType:%2u  numElems:%u  SID:3u  pvID:%3d [%ls]\n",
           notify ? "_NOTIFY" : "", dbrType, numElems, sid, pvid, pvInfo.name);
  }

  if (msg->dataSize != msg->size)  {
    if (skipMsgBytes(ci, msg->size) != msg->size)  return -1;
    errCode = ECA_PUTFAIL;  errMsg = "Invalid dataSize";
  }

  if (!errCode)  {
    typeData = dbfToPV[baseType];
  //--- reorder the bytes if necessary ---
    if (typeData.reorder)  {
      if (typeData.size == 2)
        U16_(ptr) = htons(U16_(ptr));
      else
        U32_(ptr) = htonl(U32_(ptr));
    }
  //finish:  must handle values > size of CA_MsgHdr.data[]
    if (pvSetVal(pvid, typeData.pvType, typeData.size, ptr))  {
      errCode = ECA_PUTFAIL;  errMsg = "Unable to set or convert data"; }
  }

  if (errCode AND (ci->showCmds OR ShowErrors))
    printf("*** %ls ***\n", errMsg);

  if (notify)  {
    resp.cmd = htons(msg->cmd);
    resp.size = 0;
    resp.fld3 = htons(dbrType);
    resp.fld4 = htons(numElems);
    resp.fld5 = htonl(errCode ? errCode : ECA_NORMAL);
    resp.fld6 = htonl(msg->fld6);

    if (addToRespPkt(ci, &resp, 16))  return -1;
  }

  return 0;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_SEARCH:  Check to see if this server has the specified PV name
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_Search )
{
  bool udpPacket;
  int  pvid, respLen, cliVers;
  ulong  cid;
  CA_MsgHdr  resp;


  cliVers = msg->fld4;
  cid = msg->fld5;

  udpPacket = (ci->type == UDP_CON);

  ++numSearchMsgs;


//--- check PV name to see if it matches any of the ones we provide ---
  pvid = getPVID(msg->data, msg->size);  // >= 0 if a match is found


  if (ci->showCmds OR (ShowSearch AND servNameMatch))  {
    if (!ci->showCmds)  printf("  fd:%d  cmd:[%2u]  ", ci->fd, msg->cmd);
    printf("SEARCH: %s_REPLY  CAVers:4.%d  CID:0x%08lX  pvID:%3d  [%s]\n",
           (msg->fld3 == DO_REPLY) ? "DO" : "DONT",
           cliVers, cid, pvid, msg->data);
  }


  respLen = 0;

  if (pvid >= 0)  {
    if (ci->versRespPending)  if (sendVersion(ci))  return -1;

    resp.cmd  = htons(CA_PROTO_SEARCH);
    resp.size = htons(8);   // payload size
    resp.fld3 = htons(CA_SERVER_PORT);
    resp.fld4 = 0;
    resp.fld5 = (ulong)(-1);
    resp.fld6 = htonl(cid);

    memset(resp.data, 0, 8);
    U16_(resp.data) = htons(CA_MINOR_VERSION);
    respLen = 24;
  }
  else  {
    if (!udpPacket AND (msg->fld3 == DO_REPLY))  {
      resp.cmd  = htons(CA_PROTO_NOT_FOUND);
      resp.size = 0;
      resp.fld3 = htons(DO_REPLY);
      resp.fld4 = htons(cliVers);
      resp.fld5 = htonl(cid);
      resp.fld6 = htonl(cid);

      respLen = 16;
    }
  }

  if (respLen)  if (addToRespPkt(ci, &resp, respLen))  return -1;

  return 0;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_EVENTS_OFF:
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_EventsOff )
{
  if (ci->showCmds)  printf("EVENTS_OFF:\n");

  ci->con->subEnabled = false;

  return 0;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_EVENTS_ON:
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_EventsOn )
{
  if (ci->showCmds)  printf("EVENTS_ON:\n");

  ci->con->subEnabled = true;

  return 0;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_CLEAR_CHANNEL:
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_ClearChannel )
{
  uint  u, pvid, maxSubNum, errCode = 0;
  SID  sid;
  ulong  cid;
  PVID far  *chan;
  CA_Sub far  *sub;
  CA_Con far *con;
  PV_Info  pvInfo;
  const char far *errMsg = "ERROR";
  CA_MsgHdr  resp;


  con = ci->con;

  sid = (SID) msg->fld5;
  cid = msg->fld6;


  chan = (sid < con->maxChanNum) ? con->chans + sid : NULL;
  pvid = chan ? *chan : CHAN_UNUSED;

  if (pvid == CHAN_UNUSED)  {
    if (ShowErrors OR ci->showCmds)  errMsg = "Invalid SID";
    errCode = ECA_BADCHID;
  }

  if (ci->showCmds OR (errCode AND ShowErrors))  {
    if (pvGetInfo(pvid, &pvInfo))  pvInfo.name = "<invalid>";
    printf("CLEAR_CHANNEL:  SID:0x%04X  CID:0x%08lX  pvID:%3d [%ls]\n",
           sid, cid, pvid, pvInfo.name);
    if (errCode)  printf("*** %ls ***\n", errMsg);
  }

  if (errCode)  return sendErrResp(ci, msg, cid, errCode, errMsg);

//--- free all subscriptions for the channel ---
  maxSubNum = con->maxSubNum;
  for (u=0,sub=con->subs; u<maxSubNum; ++u,++sub)
    if (sub->sid == sid)  sub->sid = SID_UNUSED;

//--- free up the channel ---
  *chan = CHAN_UNUSED;

  resp.cmd = htons(msg->cmd);
  resp.size = 0;
  resp.fld3 = 0;
  resp.fld4 = 0;
  resp.fld5 = htonl(sid);
  resp.fld6 = htonl(cid);

  if (addToRespPkt(ci, &resp, 16))  return -1;

  return 0;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_READ_NOTIFY:  Respond with the current value of a channel
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_ReadNotify )
{
  int  stat;
  uint  pvid, curMS, dbrType;
  time_t  curSecs;
  SID  sid;
  PVID far  *chan;
  CA_Con far *con;


  if (ci->showCmds)  printf("READ_NOTIFY:\n");

  dbrType = msg->fld3;
  sid = (SID) msg->fld5;

  if (validateParams(ci, CA_PROTO_READ_NOTIFY, msg->fld6, dbrType,
                     msg->fld4, sid))  return 0;

//--- get pvid ---
  con = ci->con;  chan = con->chans + sid;  pvid = *chan;


  _EnterCritical_();  curSecs = TIME();  curMS = milliTicks;  _ExitCritical_();
  epochTimeSecs = curSecs - 631152000lu; // EPICS uses 1990-01-01 00:00:00
  epochTimeNS = curMS * 1000000lu;

  if (pvLockVal(pvid, LM_WaitLock, LT_READ) < 0)  return -1;
  stat = sendValue(ci, CA_PROTO_READ_NOTIFY, msg->fld6, dbrType, msg->fld4, sid);
  pvLockVal(pvid, LM_Unlock, LT_READ);

  return stat;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_CREATE_CHAN:  Client is requesting a new channel for a PV
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_CreateChan )
{

  int  pvid;
  SID  sid;
  uint  dbrType, newChanNum;
  ulong  cid;
  PVID far  *chan;
  CA_Con far *con;
  PV_Info  pvInfo;
  CA_MsgHdr  resp;


  con = ci->con;

  sid = SID_UNUSED;
  cid = msg->fld5;

//--- If we have the requested PV, then find an unused channel slot for it ---
  if ((pvid = getPVID(msg->data, msg->size)) >= 0)  {
    for (sid=0,chan=con->chans; sid<CA_MAX_CHANS; ++sid,++chan)
      if (*chan == CHAN_UNUSED)  break;  // indicates an unused chan
    if (sid == CA_MAX_CHANS)  { sid = SID_UNUSED;  chan = NULL; }
  }

  if (ci->showCmds)
    printf("CREATE_CHAN:  Client Vers:4.%ld  CID:0x%08lX  SID:0x%04X  pvID:%3d [%.32s]\n",
           msg->fld6, cid, sid, pvid, msg->data);


//--- No matching PV name or out of available channels? ---
  if ((pvid < 0) OR (sid == SID_UNUSED) OR pvGetInfo(pvid, &pvInfo)) {
    resp.cmd  = htons(CA_PROTO_CREATE_CH_FAIL);
    resp.size = 0;
    resp.fld3 = 0;
    resp.fld4 = 0;
    resp.fld5 = htonl(cid);
    resp.fld6 = 0;

    if (addToRespPkt(ci, &resp, 16))  return -1;

    return 0;
  }



//--- If all is well, we send an ACCESS_RIGHTS msg first ---
  resp.cmd  = htons(CA_PROTO_ACCESS_RIGHTS);
  resp.size = 0;
  resp.fld3 = 0;
  resp.fld4 = 0;
  resp.fld5 = htonl(cid);
  resp.fld6 = htonl(pvInfo.flags & 3);

//  if (ci->showCmds)
//    printf("  fd:%d  cmd:[22]  sending: ACCESS_RIGHTS:\n", fd);

  if (addToRespPkt(ci, &resp, 16))  return -1;


  dbrType = pvTypeToDBF[pvInfo.dataType];

//--- And then we respond with the SID for the channel ---
  resp.cmd  = htons(CA_PROTO_CREATE_CHAN);
  resp.size = 0;
  resp.fld3 = htons(dbrType);
  resp.fld4 = htons(pvInfo.numElems);
  resp.fld5 = htonl(cid);
  resp.fld6 = htonl(sid);

  *chan = pvid;
//  chan->cid = msg->fld5;

  newChanNum = sid + 1;
  if (newChanNum > con->maxChanNum)  con->maxChanNum = newChanNum;

  if (addToRespPkt(ci, &resp, 16))  return -1;

  return 0;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_CLIENT_NAME:  Client sent a user name for the connection
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_ClientName )
{
  uint  count;
  CA_Con far *con;


  con = ci->con;

  if (ci->showCmds)  printf("CLIENT_NAME: ");

  if (!msg->dataSize)  {
    if (skipMsgBytes(ci, msg->size) != msg->size)  return -1;
    printf("<too long>\n");  return 0;
  }

  if (ci->showCmds)  printf("[%s]\n", msg->data);

  count = (msg->dataSize >= USER_NAME_LEN) ? USER_NAME_LEN : msg->dataSize;

  if (con)  {
    fmemcpy(con->userName, msg->data, count);
    con->userName[count-1] = 0;  // insure trailing termination char
  }

  return 0;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_HOST_NAME:  Client sent a host name for the connection
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_HostName )
{
  uint  count;
  CA_Con far *con;


  con = ci->con;

  if (ci->showCmds)  printf("HOST_NAME: ");

  if (!msg->dataSize)  {
    if (skipMsgBytes(ci, msg->size) != msg->size)  return -1;
    printf("<too long>\n");  return 0;
  }

  if (ci->showCmds)  printf("[%s]\n", msg->data);

  count = (msg->dataSize >= HOST_NAME_LEN) ? HOST_NAME_LEN : msg->dataSize;

  if (con) {
    fmemcpy(con->hostName, msg->data, count);
    con->hostName[count-1] = 0;  // insure trailing termination char
  }

  return 0;
}


//-----------------------------------------------------------------------------
//  CA_PROTO_ECHO:  Just send back an ECHO msg
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( CA_Echo )
{
  CA_MsgHdr  resp;


  if (ci->showCmds)  printf("ECHO\n");

  if (ci->type == UDP_CON)  return 0;

  memset(&resp, 0, sizeof(resp));

  resp.cmd  = htons(CA_PROTO_ECHO);

  if (addToRespPkt(ci, &resp, 16))  return -1;

  return 0;
}


//-----------------------------------------------------------------------------
//  We received a command that we shouldn't have
//-----------------------------------------------------------------------------
PROTO_CMD_FUNC( ProtocolError )
{
  if (ci->showCmds OR ShowErrors)
    printf("*** Protocol Error:  Unexpected command ***\n");

  if (msg->dataSize != msg->size)
    if (skipMsgBytes(ci, msg->size) != msg->size)  return -1;

  return 0;
}





//=============================================================================
// ************* Funcs to manage reading and dispatching of msgs *************
//=============================================================================

//-----------------------------------------------------------------------------
//  Process the CA msgs just read (from a UDP socket) or available to read
//  (from a TCP socket).
//
//  NOTES:
//    - ci->con will be NULL for msgs received via UDP!!!
//
//  Returns the # of messages read or < 0 if an error.
//-----------------------------------------------------------------------------
int CA_ProcessMsgs(ConInfo  *ci)
{
  int  stat, fd;
  uint  numMsgs = 0;
  CA_MsgHdr  msg;



  fd = ci->fd;

  for (;;)  {

  //--- read the header for the next msg ---
    stat = readBytes(ci, (char *)&msg, 16, 0);

  //--- did we get enough bytes for the header? ---
    if (stat != 16)  {
      if ((stat < 0) AND (errno == EAGAIN))  return numMsgs;
      if ((stat > 0) AND (stat < 16))
        conMsg(fd, "*** Incomplete msg ***");
      else
        perror("readBytes()");
      return -1;
    }

  //--- convert values from network to host format ---
    msg.cmd  = ntohs(msg.cmd);
    msg.size = ntohs(msg.size);
    msg.fld3 = ntohs(msg.fld3);
    msg.fld4 = ntohs(msg.fld4);
    msg.fld5 = ntohl(msg.fld5);
    msg.fld6 = ntohl(msg.fld6);

    *msg.data = 0;  msg.dataSize = 0;

  //--- if the payload will fit in the msg buffer, then read it now ---
    if (msg.size)  {
      if (msg.size < sizeof(msg.data))  {
        if (readBytes(ci, msg.data, msg.size, RD_TIMEOUT) != msg.size) {
          perror("reading payload");  return -1; }
        msg.data[msg.size] = 0;  msg.dataSize = msg.size;
      }
      else
        if ((msg.cmd != CA_PROTO_WRITE)
            AND (msg.cmd != CA_PROTO_WRITE_NOTIFY))  {
          if (skipMsgBytes(ci, msg.size) != msg.size)  {
            perror("skipMsgBytes()");  return -1; }
        }
    }


    if (ci->showCmds)
      printf("  fd:%d  cmd:[%2u]  data:%2u: ", fd, msg.cmd, msg.size);

    if (msg.cmd > CA_PROTO_LAST_CMMD)  {
      if (ShowErrors)
        printf("*** Ignoring Invalid CA cmd: [%2u] ***\n", msg.cmd);
      skipMsgBytes(ci, msg.size);  ++numMsgs;  continue;
    }

    if ((*cmdFuncs[msg.cmd])(ci, &msg) < 0)  {
      if (ShowErrors)  perror("*** ERROR ***");
      return -1;
    }

    ++numMsgs;
  }

}


//-----------------------------------------------------------------------------
//  Create a UDP listener socket
//
//  Returns the new socket fd on success, or < 0 on error
//-----------------------------------------------------------------------------
int  CA_CreateUDPListener(void)
{
  int  fd;
  int  fflags;
  SAI  locAddr;
  IFConfig  curConfig;

  static int  so_broadcast = TRUE;
  static ulong  lastErrTime = (-LISTEN_RETRY);


//--- avoid generating lots of pointless error msgs if the network is down ---
  getIFConfig(IF_ETH0, &curConfig);
  if (curConfig.state != IF_UP)  return -1;

  if (msSince(lastErrTime) < LISTEN_RETRY)  return -1;

  if ((fd = Socket(AF_INET, SOCK_DGRAM, 0)) < 0)  {
    perror("UDP socket() error");  lastErrTime = GetMS();  return -1; }

  if ((setsockopt(fd, SOL_SOCKET, SO_BROADCAST,
                    (char *)&so_broadcast, sizeof(so_broadcast)) ) < 0)  {
    perror("UDP setsockopt()");  close(fd);
    lastErrTime = GetMS();  return -1;
  }

//--- settings for local "addr" for socket to receive broadcasts ---
  locAddr.sin_family = AF_INET;
  locAddr.sin_port   = htons(CA_SERVER_PORT);
  locAddr.SIN_ADDR   = htonl(INADDR_ANY);  //htonl(bcIPv4);

  if (bind(fd, (SAP)&locAddr, sizeof(locAddr)) < 0)  {
    perror("UDP bind() error");
    lastErrTime = GetMS();  close(fd);  return -1;
  }

//--- make socket non-blocking ---
  fflags = fcntl(fd, F_GETFL, 0);
  fcntl(fd, F_SETFL, fflags | O_NONBLOCK);

  lastErrTime = 0;

  printf("Created UDP listener on fd:%d  port %d\r\n", fd, CA_SERVER_PORT);

  return fd;
}


//-----------------------------------------------------------------------------
//  Create a TCP listener socket
//
//  Returns the new socket fd on success, or < 0 on error
//-----------------------------------------------------------------------------
int CA_CreateTCPListener(void)
{
  int  fd;
  int  fflags;
  SAI  locAddr;
  IFConfig  curConfig;

  static ulong  lastErrTime = (-LISTEN_RETRY);


//--- avoid generating lots of pointless error msgs if the network is down ---
  getIFConfig(IF_ETH0, &curConfig);
  if (curConfig.state != IF_UP)  return -1;

  if (msSince(lastErrTime) < LISTEN_RETRY)  return -1;

  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0)  {
    perror("TCP listener socket() error");  lastErrTime = GetMS();  return -1; }

  locAddr.sin_family = AF_INET;
  locAddr.sin_port   = htons(CA_SERVER_PORT);
  locAddr.SIN_ADDR   = htonl(INADDR_ANY);

  if (bind(fd, (SAP) &locAddr, sizeof(locAddr)) < 0)  {
    perror("TCP listener bind() error");
    lastErrTime = GetMS();  close(fd);  return -1;
  }

  if (listen(fd, 2) < 0)  {
    perror("listen()");  close(fd);  lastErrTime = GetMS();  return -1; }

//--- make socket non-blocking ---
  fflags = fcntl(fd, F_GETFL, 0);
  fcntl(fd, F_SETFL, fflags | O_NONBLOCK);

  lastErrTime = 0;

  printf("Created TCP listener on fd:%d  port %d\r\n", fd, CA_SERVER_PORT);

  return fd;
}


//-----------------------------------------------------------------------------
//  Check for new connections on the specified listening TCP socket.
//
//  Returns:
//    1 if a new connection established
//    0 if no new connection
//   -1 if there was a problem with the listening socket
//-----------------------------------------------------------------------------
int CA_CheckNewConnections(fd_t fd)
{
  uint  u;
  int  newfd;
  int  addrLen;
  CA_Con far  *con;
  SAI  remAddr;


  addrLen = sizeof(remAddr);

  newfd = accept(fd, (SAP)&remAddr, &addrLen);

  if (newfd < 0)  {
    if (errno == EAGAIN)  return 0;
    perror("accept()");  return -1;
  }

  /* if (ShowCons) */  conMsg(newfd, "New Virtual Circuit");


//--- find an unused slot for the new connection ---
  for (u=0; u<CA_MAX_CONS; ++u)  {
    con = caCons + u;
    if (con->fd >= 0)  continue;
    FBZERO(con);
    fstrcpy(con->hostName, "<unknown>");
    fstrcpy(con->userName, "<unknown>");
    fmemset(con->chans, 0xff, sizeof(con->chans));
    fmemset(con->subs, 0xff, sizeof(con->subs));
    con->fd = newfd;
    con->in_Addr = remAddr;
    break;
  }

  if (u >= CA_MAX_CONS)  {
    printf("*** Max CA connections exceeded ***\n");
    close(newfd);  return 0;
  }

  return newfd;
}


//-----------------------------------------------------------------------------
//  Check for and respond to any msgs received via UDP.
//
//  Returns the number of msgs processed or < 0 if an error.
//-----------------------------------------------------------------------------
int CA_ProcessUDPPackets(int fd)
{


  return 0;
}


//-----------------------------------------------------------------------------
//  Process any commands received on existing open connections
//
//  Returns the # of currently open connections or < 0 if an error
//-----------------------------------------------------------------------------
int CA_ProcessTCPCmds(void)
{
  int  unreadBytes;
  uint  u, numCons = 0;
  CA_Con far  *con;
  SAI  respAddrTCP;
  ConInfo  ci;
  char  temp[24];


//--- check each connection for unread data and process any found ---
  for (u=0,con=caCons; u<CA_MAX_CONS; ++u,++con)  {
    if ((ci.fd = con->fd) < 0)  continue;

    ci.type = TCP_BUF;  //finish
    ci.con = con;
    ci.showCmds = ShowTCPCmds;
    ci.versRespPending = false;

    unreadBytes = 0;
    if (ioctl(ci.fd, FIONREAD, &unreadBytes))  { closeCon(&ci);  continue; }

    ++numCons;

    if (!unreadBytes)  continue;

    respAddrTCP = con->in_Addr;  respPktUsedTCP = 0;

  //--- got some data, so process whatever msgs the client sent ---
    if (ShowTCPPkts)  {
      printf("\n=== got data on fd %d (con %u) from: %s:%u via TCP ===\n",
             ci.fd, (uint)(con-caCons),
             inet_ntop(AF_INET, &respAddrTCP.SIN_ADDR, temp, sizeof(temp)),
             ntohs(respAddrTCP.sin_port) );
    }

    if (CA_ProcessMsgs(&ci) < 0)  {
      closeCon(&ci);  --numCons;  respPktUsedTCP = 0; }

    if (respPktUsedTCP)  if (sendRespPkt(&ci))  {
      closeCon(&ci);  --numCons;  respPktUsedTCP = 0; }

    if (ShowTCPPkts OR ShowTCPCmds)  printf("\n");
  }

  return numCons;
}


//-----------------------------------------------------------------------------
//  Test for changes to any of the PVs with open subscriptions and generate
//  messages for each client subscribed to the changed values.
//-----------------------------------------------------------------------------
int CA_ProcessSubscriptions(void)
{
  bool  checkForChg;
  int  events, stat;
  uint  u, s, curMS, maxSubNum;
  SID  sid;
  PVID  pvid;
  time_t  curSecs;
  CA_Con far  *con;
  PVID far  *chan, *chans;
  CA_Sub far  *sub;
  PV_Info  pvInfo;
  ConInfo  ci;
  static ulong  lastBugMsgTime = 0;


  if (ShowDebugMsgs)  printf("\r\n--- ProcessSubscriptions ---\r\n");  //tdebug

//--- Update the event masks for all PVs with at least one subscription and ---
//--- clear the events to check for the next time. ---
  checkForChg = valChgMode != VCM_PERIODIC;
  if (checkForChg)  pvCheckAllEvents(&curSecs, &curMS, valChgMode == VCM_ANY);

// ----- From recent changes/patch to EPICS code -----
// The EPICS epoch begins midnight first thing on 1st January 1990 and is in UTC.
// We convert all EPICS timestamps to the Python epoch.  This is not defined in
// the language documentation but in practice is the ANSI epoch, midnight 1st
// January 1970.  Strictly we should compute
//   EPICS_epoch = calendar.timegm((1990, 1, 1, 0, 0, 0, 0, 0, 0))
// but that pulls in an extra module dependency and the number is constant:
// EPICS_epoch = 631152000             # Seconds from 1970 to 1990

  //--- set timestamp to be used for any READ_NOTIFY msgs we generate ---
  epochTimeSecs = curSecs - 631152000lu; // EPICS uses 1990-01-01 00:00:00
  epochTimeNS = curMS * 1000000lu;

//--- Check the updated event mask for each subscription's PV.  Note that ---
//--- pvCheckEvents() also adds to the events that need to be checked the ---
//--- next time. ---
  for (u=0,con=caCons; u<CA_MAX_CONS; ++u,++con)  {
    if ((ci.fd = con->fd) < 0)  continue;

    respPktUsedTCP = 0;  //respAddrTCP = con->in_Addr;

    ci.type = TCP_BUF;  //finish
    ci.con = con;
    ci.showCmds = ci.versRespPending = false;  //finish

    chans = con->chans;  maxSubNum = con->maxSubNum;

    for (s=0,sub=con->subs; s<maxSubNum; ++s,++sub)  {
      if ((sid = sub->sid) == SID_UNUSED)  continue;

      chan = chans + sid;  pvid = *chan;

      //finish:  If subscr is "pending", then don't bother checking for a chg
      if (checkForChg AND !(sub->monMask & DBE_PENDING))  {
        if (!(events = pvCheckEvents(pvid, sub->monMask & ~DBE_PENDING)))  continue;
        if (events < 0)  if (msSince_(lastBugMsgTime) > 600000u)  {
          printf("*** Invalid subscr SID or chan pvID: SID:%u  pvID:%u ***\n",
                 sid, *chan);
          lastBugMsgTime = GetMS_();  continue;
        }
      }

      if (ShowEvents)  {
        if (pvGetInfo(pvid, &pvInfo))  pvInfo.name = "<invalid>";
        printf("Events: %u  Con:%u  SubID:0x%08lX  [%ls]\n",
               events, u, sub->subID, pvInfo.name);
      }

      if (pvLockVal(pvid, LM_TryLock, LT_READ) < 0)  {
        sub->monMask |= DBE_PENDING;  //mark subscr msg as "pending"
        if (ShowEvents)
          printf("*** subscr msg pending [%ls] ***\n", pvInfo.name);
        continue;
      }
//      if (pvLockVal(pvid, LM_WaitLock, LT_READ) < 0)  {
//        printf("*** Unable to get lock on PV [%ls] ***\n", pvInfo.name);
//        continue;
//      }
      stat = sendValue(&ci, CA_PROTO_EVENT_ADD, sub->subID,
                       sub->dbrType, sub->numElems, sid);

      if (sub->monMask & DBE_PENDING)  {
        sub->monMask &= ~DBE_PENDING;
        if (ShowEvents AND !stat)
          printf("*** pending subscr msg sent [%ls] ***\n", pvInfo.name);
      }

      pvLockVal(pvid, LM_Unlock, LT_READ);

      if (stat)  {
        closeCon(&ci);
        if (ci.type == TCP_BUF)  respPktUsedTCP = 0;
        break;
      }

      ++numEvents;
    }

    if ((ci.type == TCP_BUF) AND respPktUsedTCP)
      if (sendRespPkt(&ci))  closeCon(&ci);
  }

  ++numPSub;

  return 0;
}

//-----------------------------------------------------------------------------
void zeroCounts(void)
{
  numUDPPkts = numProcPkts = numSearchMsgs = numEvents = numWrites = numPSub =
  udpPktRate = procPktRate = searchMsgRate = eventsRate = writesRate =
  psubRate = 0;
}


//-----------------------------------------------------------------------------
//  DEBROS implementation of the EPICS Channel Access (CA) protocol service
//-----------------------------------------------------------------------------
void _task ca_udpd(int argc, char **argv)
{
  int  stat;
  uint  u, loopCount, beaconPeriod;
  sigset_t  sigSet;
  socklen_t  addrLen;
  ConInfo  ci;
  uchar  temp[24];


  if (ca_udp_pid AND (_activeTaskIdx_(ca_udp_pid) >= 0))  {
    printf("\r\n  CA %s service already running: pid: %d\r\n",
           "UDP", ca_udp_pid);
    Exit(-1);
  }

  printf("  CA UDP pid: %u\r\n", ca_udp_pid = getpid());


//--- If the list of allowed clients is blank, add the EPICS gateways ---
//finish:  Get values from a file stored on the server?
  if (!allowedClients[0])  {
    allowedClients[0] = htonl(0xC0A80405lu);  // 192.168.4.5
    allowedClients[1] = htonl(0xC0A80406lu);  // 192.168.4.6
  }


//--- block Terminate signal so we can send a packet just before exiting ---
  sigSet = sigmask(SIGTERM);  sigprocmask(SIG_BLOCK, &sigSet, NULL);

  redirectStdio();  // so printf, perror, etc don't end up in data stream
  close(0);  close(1);  close(2);  // in case this was started from a shell

  beaconPeriod = CA_INIT_BEACON_PERIOD;  lastBeaconTime = GetMS_();

  beaconID = 0;  loopCount = 0xff;

  ci.fd = -1;
  ci.type = UDP_CON;
  ci.con = NULL;
  ci.showCmds = false;


//===== main processing loop =====
  for (;;)  {

    ci.versRespPending = false;

  //--- update our copy of the hostname occasionally ---
    if (! (++loopCount & 0xff))  {
      gethostname(servName, sizeof(servName));
      servNameLen = strlen(servName);
    }

  //--- Check for a pending Terminate signal.  If we get a new one, ---
  //--- then close all our connections and exit ---
    if (sigpending(&sigSet) == 0)  if (sigismember(&sigSet, SIGTERM))  {
      printf("=== CA UDP server shutting down ===\n");
    //--- close UDP connection ---
      if (ci.fd >= 0)  { close(ci.fd);  ci.fd = -1; }
    //--- wait for just a bit to help insure connections close gracefully ---
      sleepMS(CAS_STOP_WAIT_TIME);  ca_tcp_pid = 0;  Exit(0);
    }


  //--- (re)create or close UDP broadcast socket as needed ---
    if (processUDP)  {
      if (ci.fd < 0)  ci.fd = CA_CreateUDPListener(); }
    else  {
      if (ci.fd >= 0)  { close(ci.fd);  ci.fd = -1; }
      sleepMS(200);  continue;
    }


  /*
    Clients use UDP broadcast packets to discover the server that hosts a PV,
    and to monitor the state of servers.  For clients to establish and maintain
    connections to a server, the server must respond to packets with search
    requests for its PVs in a timely manner.  Given the limitations of the
    hardware this code currently runs on, this is a bit of a challenge.

    UDP packets with SEARCH requests are generally numerous and quite frequent.
    The current hardware simply doesn't have the ability to capture and process
    ALL the UDP packets.

    The first line of defense against this onslaught of packets is to limit which
    we process based on a list of allowed client addresses.  We can't stop the
    reception of broadcast packets based on their content, but by ignoring ones
    from addresses not in a short list we free up CPU time to deal with the rest.

    This selective processing is a big help.  There are still a lot of packets
    that we never even see, but if we spend enough time each cycle on processing
    the ones we aren't deliberately ignoring, clients are generally able to
    connect new channels quickly, and reconnect disconnected ones without too
    long a delay.  But this still requires more CPU time than we can afford to
    spend (or even have available in some cases).

    The second method used to try to limit CPU usage while still allowing clients
    to (re)connect in a timely manner is to TEMPORARILY increase the amount of
    CPU time we allow for this processing.  The code that decides when to do this
    is based on the following theories about when packets of interest are more
    likely to occur:

      - Beacons affect when and how often a client with disconnected channels
        sends search requests

      - Search requests for PVs with the same record name are grouped together.
        In the same packet where possible, but across multiple packets when
        necessary.  In the later case, this means if we received a packet with
        requests for PVs we host, then it is possible another one followed it
        but we missed it.  In that case we want to increase the probability
        that we won't miss it the next time.

    The max time we will process UDP packets each call is a descreasing value
    based on how long it has been since the most recent occurance of one of
    these events.

    NOTE that because we are using only 16-bit ms values, this will  will
    "wrap-around" aprox once per minute, which is actually useful as insurance
    for clients with disconnected channels that didn't get reconnected following
    the boost triggered by the most recent event.
  */

    errno = 0;

    rcvdPktSize = nextRcvdPktByte = respPktUsedUDP = 0;

  //--- check for a packet ---
    addrLen = sizeof(respAddrUDP);
    stat = recvfrom(ci.fd, rcvdPkt, sizeof(rcvdPkt), 0,
                    (SAP)&respAddrUDP, &addrLen );

    if (stat == 0)  {
      printf("UDP SEARCH socket closed by other end\r\n");
      close(ci.fd);  ci.fd = -1;  sleepMS(100);  continue;
    }

    if (stat < 0)  {
      if (errno == EAGAIN)  { sleepMS(10);  continue; }
      perror("recvfrom()");
      close(ci.fd);  ci.fd = -1;  sleepMS(100);  continue;
    }

  //--- got a packet ---
    ++numUDPPkts;

  //--- Ignore the packet unless it is from one of our allowed clients. ---
  //--- This "filtering" is required to avoid missing so many packets that ---
  //--- clients often cannot reconnect. ---
    for (u=0; u<CA_MAX_CONS; ++u)
      if (allowedClients[u] == respAddrUDP.SIN_ADDR)  break;
    if (u == CA_MAX_CONS)  continue;

    ++numProcPkts;

    if (ShowUDPPkts)
      printf("=== %d byte packet from: %s via UDP ===\n",
             stat,
             inet_ntop(AF_INET, &respAddrUDP.SIN_ADDR, temp, sizeof(temp)) );

    rcvdPktSize = stat;

    if (processUDP AND (ci.fd >= 0))  {
      ci.showCmds = ShowUDPCmds;
      if (CA_ProcessMsgs(&ci) < 0)  {
        close(ci.fd);  ci.fd = -1;  sleepMS(100);  continue; }
    }

    if (ShowUDPPkts OR ShowUDPCmds)  printf("\n");

    if (respPktUsedUDP)  if (sendRespPkt(&ci))  {
      close(ci.fd);  ci.fd = -1;  sleepMS(100);  continue; }


  //--- Announce our own existence occasionally ---
    if (ci.fd < 0)
      beaconPeriod = CA_INIT_BEACON_PERIOD;
    else
      if (msSince_(lastBeaconTime) >= beaconPeriod)
        if (sendBeacon(ci.fd) < 0)  {
          ci.fd = -1;  beaconPeriod = CA_INIT_BEACON_PERIOD; }
        else
          if ((beaconPeriod *= 2) > CA_MAX_BEACON_PERIOD)
            beaconPeriod = CA_MAX_BEACON_PERIOD;
  }

}



//-----------------------------------------------------------------------------
//  DEBROS implementation of the EPICS Channel Access (CA) protocol service
//-----------------------------------------------------------------------------
void _task ca_tcpd(int argc, char **argv)
{
  int  tcpfd;
  int  stat;
  uint  u, numCons, loopTime, startLoopTime, lastTestedEvents, lastSampleTime,
        loopCount, sleepTime;
  sigset_t  sigSet;
  CA_Con far  *con;


  if (ca_tcp_pid AND (_activeTaskIdx_(ca_tcp_pid) >= 0))  {
    printf("\r\n  CA %s service already running: pid: %d\r\n",
           "TCP", ca_tcp_pid);
    Exit(-1);
  }

  printf("  CA TCP pid: %u\r\n", ca_tcp_pid = getpid());


//--- block Terminate signal so we can send a packet just before exiting ---
  sigSet = sigmask(SIGTERM);  sigprocmask(SIG_BLOCK, &sigSet, NULL);

  redirectStdio();  // so printf, perror, etc don't end up in data stream
  close(0);  close(1);  close(2);  // in case this was started from a shell

  lastTestedEvents = 0;

  tcpfd = -1;  numCons = 0;

//--- initialize connection data ---
  for (u=0; u<CA_MAX_CONS; ++u)  {
    con = caCons + u;
    fmemset(con, 0, sizeof(CA_Con));
    fmemset(con->chans, 0xff, sizeof(con->chans));
    fmemset(con->subs, 0xff, sizeof(con->subs));
    con->fd = -1;
  }

  zeroCounts();

  lastSampleTime = GetMS_();  loopCount = 0xff;


//===== main processing loop =====
  for (;;)  {
    startLoopTime = GetMS_();

    if (msSince_(lastSampleTime) >= 1000u)  {
      udpPktRate = numUDPPkts;  procPktRate = numProcPkts;
      searchMsgRate = numSearchMsgs;  eventsRate = numEvents;
      writesRate = numWrites;  psubRate = numPSub;
      numUDPPkts = numProcPkts = numSearchMsgs = numEvents = numWrites =
         numPSub = 0;
      lastSampleTime = GetMS_();
    }


  //--- Check for a pending Terminate signal.  If we get a new one, ---
  //--- then close all our connections and exit ---
    if (sigpending(&sigSet) == 0)  if (sigismember(&sigSet, SIGTERM))  {
      printf("=== CA TCP server shutting down ===\n");
    //--- close all connections ----
      if (tcpfd >= 0)  { close(tcpfd);  tcpfd = -1; }
      for (u=0; u<CA_MAX_CONS; ++u)  {
        con = caCons + u;
        if (con->fd < 0)  continue;
        close(con->fd);  con->fd = -1;
      }
      zeroCounts();
    //--- wait for just a bit to help insure connections close gracefully ---
      sleepMS(CAS_STOP_WAIT_TIME);  ca_tcp_pid = 0;  Exit(0);
    }


  //--- update our copy of the hostname occasionally ---
    if (! (++loopCount & 0xff))  {
      gethostname(servName, sizeof(servName));
      servNameLen = strlen(servName);
    }


  //--- Check for commands rcvd on the Virtual Circuits ---
    if ((stat = CA_ProcessTCPCmds()) >= 0)  numCons = stat;

//    if (consecutivePkts > 1)  printf("=== %u ===\n", consecutivePkts);  //tdebug

  //--- Check for new Virtual Circuit requests ---
    if (tcpfd >= 0)
      if (CA_CheckNewConnections(tcpfd) < 0)  { close(tcpfd);  tcpfd = -1; }


  //--- Check for events that apply to one or more Subscriptions and send ---
  //--- msgs as needed on the corresponding Virtual Circuits ---
    if (numCons)  if (msSince_(lastTestedEvents) >= eventPeriod)  {
      lastTestedEvents = GetMS_();  CA_ProcessSubscriptions(); }


  //--- Close or (re)open the TCP listening socket as needed ---
    if (numCons >= CA_MAX_CONS)  {
      if (tcpfd >= 0)  { close(tcpfd);  tcpfd = -1; }  }
    else
      if (tcpfd < 0)  tcpfd = CA_CreateTCPListener();


  //--- In case we don't have any Virtual Circuits to do a blocked read on, ---
  //--- or there is a problem that would otherwise let this loop use all ---
  //--- non-idle CPU cycles, we put a limit on how often this loop can run ---
    if (numCons)  {
      loopTime = msSince_(startLoopTime);
      sleepTime = (loopTime < MIN_LOOP_TIME ? MIN_LOOP_TIME - loopTime : 5);
    }
    else
      sleepTime = 50;
    sleepMS(sleepTime);
  }

}



//-----------------------------------------------------------------------------
//  Show some basic statistics for the CA server
//-----------------------------------------------------------------------------
void CA_showStats(void)
{
  printf("\r\n"
         "Counts for the previous second:\r\n"
         "\r\n"
         "  UDP pkts    Search Msgs    Events   ProcSub    Writes \r\n"
        //  ###/###         ###         ###       ###       ###
         );


  for (;;)  {
    printf("  %3u/%3u         %3u         %3u       %3u       %3u \r\n",
           procPktRate, udpPktRate, searchMsgRate, eventsRate, psubRate,
           writesRate);
    fflush(stdout);

    CheckETX();  sleepMS(990);
  }

}


//-----------------------------------------------------------------------------
//  Show the state of the CA server
//
//               Circuits   Channels  Subscriptions
//               --------   --------  -------------
//  levels:  0:    count      count       count        <not implemented yet>
//           1:    list       count       count
//           2:    list       list        count        <not implemented yet>
//           3:    list       list        list
//-----------------------------------------------------------------------------
void CA_showStatus(int level)
{
  uint  u, w, baseType, extType, port, chanCount, subCount;
  SID  sid;
  div_t  qr;
  ulong  in_Addr;
  CA_Con far  *con;
  PVID far  *chan;
  CA_Sub far  *sub;
  char  temp[24];
  PV_Info  pvInfo;



  for (u=0,con=caCons; u<CA_MAX_CONS; ++u,++con)  {
    if (con->fd < 0)  continue;

    in_Addr = con->in_Addr.SIN_ADDR;  // IPv4 addr in network byte order
    port = ntohs(con->in_Addr.sin_port);

    printf("\r\n"
           "  ===== Virtual Circuit: %d =====\r\n"
           "    fd: %d   IP: %s:%u\r\n"
           "    Host: [%ls]   User: [%ls]\r\n",
           u, con->fd,
           inet_ntop(AF_INET, &in_Addr, temp, sizeof(temp)), port,
           con->hostName, con->userName );

    if (level > 1)  printf(
           "    ----- Channels -----\r\n");

//    printf("    maxChanNum:%u   maxSubNum:%u\r\n\n",  //tdebug
//           con->maxChanNum, con->maxSubNum );

    chanCount = subCount = 0;

    for (sid=0,chan=con->chans; sid<con->maxChanNum; ++sid,++chan)  {
      if (*chan == CHAN_UNUSED)  continue;

      if (level >= 2)  {
        if (pvGetInfo(*chan, &pvInfo))  pvInfo.name = "<invalid>";
        printf("    Chan:  SID:0x%04X  pvID:%3d  [%ls]\r\n",
               sid, *chan, pvInfo.name );
      }
      ++chanCount;

      for (w=0,sub=con->subs; w<con->maxSubNum; ++w,++sub)  {
        if (sub->sid != sid)  continue;

        if (level >= 3)  {
          qr = div(sub->dbrType, 7);  baseType = qr.rem;  extType = qr.quot;
          printf("         subID:0x%08lX  mask:%2u  dbrType:DBR_%ls_%ls  #elems:%u\r\n",
                 sub->subID, sub->monMask,
                 baseTypeName[baseType], extTypeName[extType],
                 sub->numElems );
        }
        ++subCount;
      }

      if (level >= 2)  printf("\n");
    }

    if (level == 1)
      printf("    %u Channels, %u Subscriptions\r\n", chanCount, subCount);
  }

  Exit(0);
}


//-----------------------------------------------------------------------------
//  Show a list of the client address for which we will process UDP packets
//-----------------------------------------------------------------------------
void showAllowedList(void)
{
  uint  u;
  ulong  *pAddr;
  char  temp[20];


  printf("\r\n  Allowed SEARCH clients:\r\n");

  for (u=0,pAddr=allowedClients; *pAddr AND u<CA_MAX_CONS; ++u,++pAddr)  {
    if (!inet_ntop(AF_INET, pAddr, temp, sizeof(temp)))
      printf("    ?.?.?.?\r\n");
    else
      printf("    %s\r\n", temp);
  }

  printf("\r\n");
}

//-----------------------------------------------------------------------------
//  Command to manage the DEBROS EPICS Channel Access (CA) protocol service
//-----------------------------------------------------------------------------
void _task cas(int argc, char **argv)
{
  uint  u, v;
  ulong  *pAddr;
  ulong  ipAddr;
  char  *argv_[3];


  if (showHelp(2, argc, argv))  {
    printf("\nusage: cas start|stop|chans|stats|summary\r\n"
             "           clist\r\n"
             "           add <client_addr>\r\n"
             "           del <client_addr>|all\r\n"
             "           queries respond|ignore\r\n"
             "           events anychg|threshold|periodic|<period in ms>\r\n"
          );
    Exit(-1);
  }


//--- see if any previous instances died ---
  if (ca_tcp_pid AND (_activeTaskIdx_(ca_tcp_pid) < 0))  ca_tcp_pid = 0;
  if (ca_udp_pid AND (_activeTaskIdx_(ca_udp_pid) < 0))  ca_udp_pid = 0;



//=== show a summary of the state of the CA service ===
  if (fstrcmp(argv[1], "summary") == 0)  {

    printf("\r\n");

    if (!ca_tcp_pid)
      printf("  CA %s service not running\r\n", "TCP");
    else  {
      printf("  CA %s service pid: %d\r\n", "TCP", ca_tcp_pid);

      if (!ca_udp_pid)
        printf("  CA %s service not running\r\n\n", "UDP");
      else  {
        printf("  CA %s service pid: %d\r\n", "UDP", ca_udp_pid);

        showAllowedList();

        printf("  %s SEARCH requests\r\n",
               processUDP ? "responding to" : "ignoring" );
      }

      printf("  Event monitors: %s every %u ms\r\n",
             chgModeDesc[valChgMode], eventPeriod);

      CA_showStatus(1);
    }

    Exit(0);
  }


//===== start the CA server =====
  if (fstrcmp(argv[1], "start") == 0)  {
    if (ca_tcp_pid)
      printf("\r\n  CA %s service already running: pid: %d\r\n",
             "TCP", ca_tcp_pid);
    else  {
      argv_[0] = "ca_tcpd";  argv_[1] = NULL;
      createProc(ca_tcpd, CA_TCPD_PRIORITY, 1, argv_, 0, 1, 2);
    }
    if (ca_udp_pid)
      printf("\r\n  CA %s service already running: pid: %d\r\n",
             "UDP", ca_udp_pid);
    else  {
      argv_[0] = "ca_udpd";  argv_[1] = NULL;
      createProc(ca_udpd, CA_UDPD_PRIORITY, 1, argv_, 0, 1, 2);
    }
    Exit(0);
  }


//===== stop the CA server =====
  if (fstrcmp(argv[1], "stop") == 0)  {
    if (!ca_tcp_pid AND !ca_udp_pid)  {
      printf("\r\n  services are not running.\r\n");  Exit(-1); }
    printf("Terminating CA server(s)\r\n");
    if (ca_udp_pid)  kill(ca_udp_pid, SIGTERM);
    if (ca_tcp_pid)  kill(ca_tcp_pid, SIGTERM);
    Exit(0);
  }


//===== show the state of the CA server =====
  if (fstrcmp(argv[1], "chans") == 0)  {
    if (!ca_tcp_pid)  {
      printf("\r\n  service is not running.\r\n");  Exit(-1); }
    CA_showStatus(3);  Exit(0);
  }


//===== show stats for CA server =====
  if (fstrcmp(argv[1], "stats") == 0)  {
    if (!ca_tcp_pid)  {
      printf("\r\n  service is not running.\r\n");  Exit(-1); }
    CA_showStats();  Exit(0);
  }


//===== show the current list of allowed client addresses =====
  if (fstrcmp(argv[1], "clist") == 0)  {  showAllowedList();  Exit(0);  }


  if (argc < 3)  { printf("Invalid or incomplete cmd\r\n");  Exit(-1); }



//===== add an addr to the list of allowed client addresses =====
  if (fstrcmp(argv[1], "add") == 0)  {
    inet_pton(AF_INET, argv[2], &ipAddr);
    if (!ipAddr)  { printf("Invalid address\r\n");  Exit(-1); }

    KLOCK();
    for (u=0,pAddr=allowedClients; u<CA_MAX_CONS; ++u,++pAddr)
      if (!*pAddr)  { *pAddr = ipAddr;  KUNLOCK();  Exit(0); }

    KUNLOCK();
    printf("Max # of allowed clients already set\r\n");  Exit(-1);
  }


//===== remove an addr from the list of allowed client addresses =====
  if (fstrcmp(argv[1], "del") == 0)  {
    if (fstrcmp(argv[2], "all") == 0)  { BZERO(allowedClients);  Exit(0); }

    inet_pton(AF_INET, argv[2], &ipAddr);
    if (!ipAddr)  { printf("Invalid address\r\n");  Exit(-1); }

    // if found, slide the rest of the list down one and zero the last entry ---
    KLOCK();
    for (u=0,pAddr=allowedClients; *pAddr AND u<CA_MAX_CONS; ++u,++pAddr)  {
      if (*pAddr != ipAddr)  continue;
      for (v=u; v<CA_MAX_CONS-1; ++v,++pAddr)  *pAddr = pAddr[1];
      *pAddr = 0;  KUNLOCK();  Exit(0);
    }

    KUNLOCK();  printf("Addr not in list\r\n");  Exit(-1);
  }


//===== enable/disable UDP bcast pkt processing (used for SEARCH queries) =====
  if (fstrcmp(argv[1], "queries") == 0)  {
    if (fstrcmp(argv[2], "respond") == 0)  { processUDP = true;  Exit(0); }
    if (fstrcmp(argv[2], "ignore") == 0)  { processUDP = false;  Exit(0); }
  }


//===== set event processing parameters =====
  if (fstrcmp(argv[1], "events") == 0)  {
    for (u=0; u<3; ++u)
      if (StriEq(argv[2], chgModeDesc[u]))  { valChgMode = u;  Exit(0); }
    if (isdigit(*argv[2]))  {
      if ((eventPeriod = ATOI(argv[2])) < MIN_EVENT_PERIOD)
        eventPeriod = MIN_EVENT_PERIOD;
      printf("Event monitor period now %u ms\r\n", eventPeriod);  Exit(0);
    }
  }


  printf("\r\nInvalid option\r\n");  Exit(-1);
}

//-----------------------------------------------------------------------------

