//----- inetd.c ----- 06/06/12 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//  deamon process modeled after the older Unix/Linux inetd "super-deamon"
//-----------------------------------------------------------------------------
// Update Log (most recent first)
//
//  Entries marked with (*) indicate ones that improve the stability and/or
//  reliability of the code.
//
//
//  2012-06-15 - Mark Davis
//               - Modified formatting of log messages to improve readability.
//
//  2010-03-22 - Mark Davis (*)
//               - Fixed bug in main inetd loop where the index used for the
//                 outer loop was used in a smaller inner loop.
//
//  2010-03-12 - Mark Davis
//               - Changes to avoid logging the same error message every time
//                 through the loop when there is a consistent problem.
//
//  200x-xx-xx - Mark Davis
//               - other changes I didn't log for some reason....
//
//  2006-03-14 - Mark Davis
//               - Started original version (?)
//
//-----------------------------------------------------------------------------
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>
#include <linux/sched.h>
#include <sys/serial.h>
#include <errno.h>


#include <string.h>

#include <stcpip.h>
#include <inetd.h>

#include <sys/farprintf.h>

#include <tsh.h>


//-----------------------------------------------------------------------------
//  If we got a new error while creating a listener for a service, or if has
//  been a while since we showed the error we have been getting, then show
//  the message for the current error.
//-----------------------------------------------------------------------------
void serviceErr(ServiceInfo *srv, char *context)
{
  if ((errno == srv->lastErr) AND (msSince(srv->lastErrTime) < 10000))  return;

  srv->lastErr = errno;  srv->lastErrTime = GetMS();

  printf("while creating listener for %.32s:\r\n", srv->srvName);

  perror(context);
}


//-----------------------------------------------------------------------------
//  Attempt to create a listener for the specified service if there isn't one
//
//  Returns the fd the listener is open on or < 0 if there is an error
//-----------------------------------------------------------------------------
fd_t getListener(ServiceInfo *srv)
{
  fd_t  devfd;
  int  fflags;
  struct sockaddr_in  srvAddr;
  AsyncParams  *ap;
  ulong  baudRate;
  IFConfig  curConfig;
  char  temp[16];


  if ((devfd = srv->devfd) >= 0)  return devfd;


  switch (srv->protocol)  {

    case ASYNC:
      ap = (AsyncParams *)(srv->cfgParams);

      if (!ap)  {
         baudRate = 115200L;  /* mode = SER_RS232; */ }
      else  {
         baudRate = ap->baud;  /* mode = ap->mode; */ }

      sprintf(temp, "/dev/tty%d", srv->port);

      KLOCK();  // to keep devfd = open(..) and noGettyOnA in sync

      if ((devfd = open(temp, O_RDWR, 0)) < 0)  {
        KUNLOCK();
        if (errno == EBUSY)  break;
        perror(temp);  break;
      }

      ioctl(devfd, IOC_SET_BAUD, &baudRate);

      if (srv->port == 0)  noGettyOnA = true;

      KUNLOCK();

      srv->devfd = devfd;

      printf("fd: %d (GFD: %d) port %u: listening for %.32s\r\n",
              devfd, GFD(devfd), srv->port, srv->srvName);
      break;


    case ProtoTCP:
      getIFConfig(IF_ETH0, &curConfig);
      if (curConfig.state != IF_UP)  break;

      if ((devfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)  {
        serviceErr(srv, "socket() error");  break; }

      srvAddr.sin_family = AF_INET;
      srvAddr.sin_port   = htons(srv->port);
      srvAddr.sin_addr   = htonl(INADDR_ANY);

      if (bind(devfd, (struct sockaddr *) &srvAddr, sizeof(srvAddr)) < 0)  {
        serviceErr(srv, "bind error");  close(devfd);  break;  }

      if (listen(devfd, 2) < 0)  {
        serviceErr(srv, "listen error");  close(devfd);  break;  }

      fflags = fcntl(devfd, F_GETFL, 0);
      fcntl(devfd, F_SETFL, fflags | O_NONBLOCK);

      srv->devfd = devfd;  srv->lastErr = 0;

      printf("fd: %d (GFD: %d) port %u: listening for %.32s\r\n",
              devfd, GFD(devfd), srv->port, srv->srvName);
      break;

    case ProtoUDP:
      break;
  }


  return srv->devfd;
}


//-----------------------------------------------------------------------------
//  process to listen on specified ports and create processes to service
//  incoming connections on those ports
//-----------------------------------------------------------------------------
void _task inetd(int argc, char **argv)
{
  ServiceInfo  *srv;
  uint  u, srvIdx;
  bool  istty;
  fd_t  devfd, newDevfd;
  int  addrLen, fflags, /*mode,*/ readCount;
  struct sockaddr_in  cliAddr;
  FDEV  far *file;
  void  *ioDev;
  char  *argv_[2];
  char  temp[128];


  redirectStdio();  // so printf, perror, etc don't end up in data stream

//--- initialize state of each service ---
  for (u=0; u<NumServices; ++u)  inetdConf[u].devfd = -1;


//--- repeatedly scan the list of services.  If there is no listening link, ---
//--- then create one.  If a connection has been established for a link ---
//--- then create a process to service it ---

  for (;;)  for (srvIdx=0; srvIdx<NumServices; ++srvIdx)  {
    if (!srvIdx)  sleepMS(250);

    srv = inetdConf + srvIdx;

  //===== make sure there is a listening descriptor for each service =====
    if ((devfd = srv->devfd) < 0)  { getListener(srv);  continue; }

    istty = StrEq(srv->srvName, "tsh");  //finish


  //===== check for a connection established on a listening link =====
    switch (srv->protocol)  {

      case ASYNC:
      //--- read any characters that were received ---
        fflags = fcntl(devfd, F_GETFL, 0);
        fcntl(devfd, F_SETFL, fflags | O_NONBLOCK);  // don't block on read
        readCount = read(devfd, temp, sizeof(temp));
        fcntl(devfd, F_SETFL, fflags);  // restore flags

      //--- if an error occured, then close the descriptor ---
        if (readCount < 0)  {
          if (errno == EAGAIN)  break;
          perror(srv->srvName);  close(devfd);  srv->devfd = -1;
          if (srv->port == 0)  noGettyOnA = true;
        }

      //--- if any of the received chars is a CR, then launch a new ---
      //--- process using the connection for stdio ---
        for (u=0; u<readCount; ++u)  if (temp[u] == 0x0D)  {
          if (srv->port == 0)  noGettyOnA = false;

          argv_[0] = srv->srvName;  argv_[1] = NULL;
          if (istty)  setIstty(devfd, true);

          createProc(srv->procMain, srv->priority, 1, argv_,
                     devfd, devfd, devfd);

          close(devfd);  srv->devfd = -1;
        }
        break;


      case ProtoTCP:
        addrLen = sizeof(cliAddr);
        newDevfd = accept(devfd, (struct sockaddr *)&cliAddr, &addrLen);
        if (newDevfd >= 0)  {
          file = FD[userPix][newDevfd];  ioDev = file->ioDev;
          file->ioFuncs->descr(ioDev, temp, sizeof(temp)-1);

          printf("opened %.32s connection on fd: %d (GFD: %d)\r\n"
                 "%.128s\r\n",
                 srv->srvName, newDevfd, GFD(newDevfd), temp );

          argv_[0] = srv->srvName;  argv_[1] = NULL;
          if (istty)  setIstty(newDevfd, true);

          createProc(srv->procMain, srv->priority, 1, argv_,
                     newDevfd, newDevfd, newDevfd);

          close(newDevfd);
        }
        else  if (errno != EAGAIN)  {
          printf("accept() for %.32s: fd: %d (GFD %d): ",
                 srv->srvName, devfd, GFD(devfd));
          perror(NULL);
          close(devfd);  srv->devfd = -1;
        }
        break;
    }

  }

}

//-----------------------------------------------------------------------------
