//----- tsh.c ----- 09/26/13 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//  Tiny shell
//-----------------------------------------------------------------------------
//  Change log (most recent first)
//
//
//  2013-09-26 - Mark Davis
//               - Added check for no arguments in cmdLine() following call to
//                 parseArgs().  Prior to this, it would crash the Linux
//                 version of the application if cmdLine() was called with a
//                 command line that consisted of nothing but spaces.
//
//  2007-07-17 - Mark Davis
//               - Added support for pipes
//
//  2006-03-14 - Mark Davis
//               - Original version started
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <fcntl.h>

#include <string.h>

#include <busybox.h>

#include <sys/backup.h>
#include <applic.h>


#define MAX_ARGS  10

#define MAX_LINE_LEN  160


#include <sys/farprintf.h>

#include <tsh.h>


//-----------------------------------------------------------------------------
//  Get a cmd line string from the user.  If histSize != 0, then history must
//  point to a buffer of histSize that contains a series of null-terminated
//  commands, and idx must be the index of the null char at the end of the
//  most recent command.
//-----------------------------------------------------------------------------
char *Gets(char *str, char *history, uint idx, uint histSize)
{
  char  *ptr;
  int  count, fflags;
  uint  ui, cmdLen;
  uchar  escCount, ch;

#define DECR(ui)  { if (ui) --(ui);  else (ui) = histSize-1; }
#define INCR(ui)  { if ((++ui) == histSize) (ui) = 0; }


  ptr = str;   fflags = fcntl(0, F_GETFL, 0);  escCount = 0;

  fflush(stdout);

  for (;;)  {
    fcntl(0, F_SETFL, fflags | O_NONBLOCK);
    count = read(0, ptr, 1);
    fcntl(0, F_SETFL, fflags & (~O_NONBLOCK));

    if (count == 0)  { fprintf(DiagLog, "connection closed\n");  Exit(-1); }

    if (count < 0)  {
      if (errno == EAGAIN)  { sleepMS(50);  continue; }
      fprintf(DiagLog, "read error: %ls\n", strerror(errno));  Exit(-2);
    }

  //--- just ignore/filter out telnet cmds for now ---
    if ((uchar)*ptr == 0xFF)  {  // telnet IAC (Interpret As Command)
      read(0, ptr+1, 2);
      ptr[1] = 252;   write(1, ptr, 3);  // just say WONT for now
      continue;
    }


    ch = *ptr &= 0x7f;

  //--- if ESC char, possible start of arrow key ---
    if (ch == 0x1B)  {
      escCount = 1;  *ptr = 0;  continue;  }

  //--- check for 2nd char of arrow key ---
    if ((escCount == 1) AND (ch == '['))  {
      escCount = 2;  *ptr = 0;  continue;  }

  //--- check for 3rd char of arrow key ---
    if (escCount == 2)  {

      if (histSize)  {

      //--- up-arrow - extract last intact cmd (if any) from history buffer ---
        if (ch == 'A')  {  // up arrow - get prev cmd (if any)
        //--- erase any existing cmd line chars ---
          while (ptr > str)  { write(1, "\010 \010", 3);  --ptr; }

        //--- find null char preceeding prev cmd (if any) ---
          ui = idx;  DECR(ui);
          for (cmdLen=0; history[ui]; ++cmdLen)  DECR(ui);

        //--- copy and display the cmd if one found ---
          if (cmdLen)  {
            idx = ui;  INCR(ui);
            while (history[ui])  {
              *ptr = history[ui];  write(1, ptr, 1);  INCR(ui);  ++ptr; }
          }
        }  else

      //--- down arrow ---
        if (ch == 'B')  {
        //--- erase any existing cmd line chars ---
          while (ptr > str)  { write(1, "\010 \010", 3);  --ptr; }

        //--- copy and display the next cmd in the circular list ---
          INCR(ui);
          for (cmdLen=0; history[ui]; ++cmdLen)  {
            *ptr = history[ui];  write(1, ptr, 1);  INCR(ui);  ++ptr; }
          if (cmdLen)  idx = ui;
        }  else

      //--- right arrow ---
        if (ch == 'C')  {
//          write(1, "{right}", 7);
        }  else

      //--- left arrow ---
        if (ch == 'D')  {
  //        write(1, "{left}", 6);
        }

      }  //end: if (histSize)

      *ptr = 0;  escCount = 0;  continue;
    }

  //--- if printable char, add it to the current line ---
    if (*ptr >= ' ')  if ((ptr-str) < (MAX_LINE_LEN-4))  {
      write(1, ptr, 1);  ++ptr;  continue; }

  //--- if CR, then return the current line ---
    if (*ptr == 0x0D)  { write(1, "\r\n", 2);  *ptr = 0;  return str; }  // CR

  //--- if BS, then remove a char from the current line ---
    if (*ptr == 0x08)  if (ptr > str)  {  // BS
      write(1, "\010 \010", 3);  --ptr;  continue; }

  //--- if Ctrl-C, then exit the current task ---
    if (*ptr == 0x03)  { write(1, "\r\n", 2);  Exit(-1); }  // Ctrl-C
  }
}

//-----------------------------------------------------------------------------
//  Scan a list of commands looking for a match.  If one is found, then
//  create a new process to execute it.
//
//  Returns:
//       the pid of the process created if a match is found and a process
//       successfully created to execute it
//     < 0 if no match or an error occured
//-----------------------------------------------------------------------------
int checkCmds(const CmdInfo far *cmdList, int numCmds, int argc, char **argv,
              fd_t infd, fd_t outfd, fd_t errfd)
{
  uint  i;
  int  myPriority;
  PID  pid;
  const CmdInfo far *cmdInfo;


  for (i=0; i<numCmds; ++i)  {
    cmdInfo = cmdList + i;
    if (fstrcmp(argv[0], cmdInfo->cmdName))  continue;
    if (! cmdInfo->procMain)  continue;

    myPriority = getpriority(PRIO_PROCESS, getpid());

    fprintf(DiagLog, "Cmd: ");
    for (i=0; i<argc; ++i)  fprintf(DiagLog, "%s ", argv[i]);
    fprintf(DiagLog, "\n");

    pid = createProc(cmdInfo->procMain, myPriority, argc, argv,
                     infd, outfd, errfd);
    if (pid < 0)  { printf("Unable to execute command\r\n");  return 1; }

    return pid;
  }

  return -1;
}

//-----------------------------------------------------------------------------
//  execute the cmd with parameters specified in argv[].
//
//  Returns:
//    pid of the child process created on success, or < 0 if an err
//-----------------------------------------------------------------------------
int execute(int argc, char **argv, fd_t infd, fd_t outfd, fd_t errfd)
{
  int  childpid;


  if (argc < 1)  return -1;

//--- check for "internal" commands ---
  if (StrEq(argv[0], "cls") OR StrEq(argv[0], "clear"))  {
    printf(clsCmd);  return -1; }

  if (StrEq(argv[0], "exit"))  Exit(0);

//--- check for commands that require a new process ---
  if ((childpid = checkCmds(commonCmds, NumCommonCmds, argc, argv,
                            infd, outfd, errfd)) > 0)  return childpid;
  if ((childpid = checkCmds(applCmds, NumApplCmds, argc, argv,
                            infd, outfd, errfd)) > 0)  return childpid;

  printf("%s: Unknown command\r\n", argv[0]);  return -1;
}

//-----------------------------------------------------------------------------
//  process a command line string
//-----------------------------------------------------------------------------
void cmdLine(char *buf)
{
#define  MaxPIDs  4
  uint  ui, numTasks, first;
  int  argc;
  int  infd, outfd, pipefds[2];
  int  pids[MaxPIDs];
  char  ch;
  char  *argv[MAX_ARGS+1];



//--- if 1st non-space char is '#' then ignore the command ---
  ch = *buf;
  for (ui=0; ch; )  {
    if ((ch > ' ') AND (ch != '#'))  break;
    if (ch == '#')  return;
    ch = buf[++ui];
  }


  parseArgs(buf, MAX_ARGS, argv, &argc);  argv[argc] = NULL;

//--- start one task for each cmd separated by a pipe symbol and create ---
//--- a pipe to use for the stdout|stdin for each paired cmd ---
  infd = 0;  outfd = 1;  numTasks = 0;
  for (first=ui=0; ui<argc; ++ui)  if (StrEq(argv[ui], "|"))  {
  //--- create pipe to use for stdout for this cmd and stdin for next cmd ---
    if (pipe(pipefds))  { perror("Could not create pipe");  Exit(-1); }
    outfd = pipefds[1];  // use pipe writer for stdout for this cmd

    pids[numTasks] = execute(ui-first, argv+first, infd, outfd, 2);

    if (infd > 2)  close(infd);  // if stdin was pipe close our copy of it
    infd = pipefds[0];  // use pipe reader for stdin on next cmd
    close(outfd);  outfd = 1;  // close our copy of last cmd's stdout

    ++numTasks;  first = ui + 1;
  }
  if (first < ui)  {
    pids[numTasks] = execute(ui-first, argv+first, infd, outfd, 2);
    ++numTasks;
  }
  if (infd > 2)  close(infd);  // if stdin was pipe close our copy of it


  if (strcmp(argv[argc-1], "&"))  if (pids[numTasks-1] > 0)
    waitpid(pids[numTasks-1], NULL, 0);
}

//-----------------------------------------------------------------------------
//  tiny shell task
//-----------------------------------------------------------------------------
void _task tsh(int argc_, char **argv_)
{
  uint  histIdx;
  char  *ptr;
  char  buf[MAX_LINE_LEN], history[256];


  printf("\r\n%ls\r\n RTOS: %ls  %ls\r\n APPL: %ls  %ls\r\n%ls\r\n",
         asterisks, osDTM, osName, applDTM, applName, asterisks);

  histIdx = 0;  memset(history, 0, sizeof(history));

  for (;;)  {
    gethostname(buf, sizeof(buf));  buf[sizeof(buf)-1] = 0;

    printf("\r\n%s> ", buf);
    Gets(buf, history, histIdx, sizeof(history));
    if (! *buf)  continue;

  //--- copy cmd to history buffer ---
    for (ptr=buf; ; ++ptr)  {
      if (++histIdx == sizeof(history))  histIdx = 0;
      history[histIdx] = *ptr;
      if (!*ptr)  break;
    }

    cmdLine(buf);
  }
}

//-----------------------------------------------------------------------------
