//----- LLRF2.c ----- 01/09/13 -----(09/02/11)-----

//-----------------------------------------------------------------------------
//  Board Support Package (BSP) for LLRF2 NIM module which includes the
//  following hardware:
//
//  - Single Board Computer (RCM3200: 44 MHz 8-bit CPU, 768K RAM, 512K flash)
//  - 2 Serial EEPROMs MicroChip SST 25VF016B-50 (2048K bytes each)
//  - Xilinx Spartan-3A-DSP 3400 FPGA
//  - Texas Instruments TMS320C6748 DSP
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2013-01-09 - Mark Davis
//               - Made same efficiency improvements to readPlot() in this
//                 module that I had done earlier in the PhaseCtlr application
//                 code (the predecessor to the LLRF2 prototype module).
//
//  2012-11-16 - Mark Davis
//               - Added write_FPGA_Regs()
//
//  2012-11-15 - Mark Davis
//               - Reduced timeouts for locks.
//               - Added read_FPGA_Regs()
//
//  2012-03-06 - Mark Davis
//               - All checks for locking errors now include the value 0.  This
//                 is mostly just a safe-coding precaution as the locking func
//                 should never return a value of 0).
//
//  2011-12-09 - Mark Davis
//               - Writes to the DSP are now ended by pulsing the clock signal
//                 high once after the _SPI1_DSP_EN line is set high.
//
//  2011-09-02 - Mark Davis
//               - Original version started.
//
//-----------------------------------------------------------------------------


#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <sys/backup.h>

#include <sys/spi.h>
#include <sys/hdwDefs.h>

#include <busybox.h>

#include <Rabbit.h>
#include <dcdefs.h>

#include <stcpip.h>

#include <sys/socket.h>
#include <sys/srfs.h>
#include <sys/flashFile.h>
#include <sys/asflash.h>


#include <sys/modbus.h>

#include <LLRF2_Appl.h>
#include <LLRF2.h>


#include <sys/farprintf.h>



//--- see hdwInit() for Digital I/O pin usage ---

const char ArchName[] = { "LLRF2" };

#define  NUM_ADC  1
#define  NUM_DAC  1

#define  MAX_DIN   1
#define  MAX_DOUT  2


//--- these values can be set LOWER by the application ---
const uint  NumHdwADC = NUM_ADC;
const uint  NumHdwDAC = NUM_DAC;

const uint  MaxDIN =  MAX_DIN;
const uint  MaxDOUT = MAX_DOUT;


unsigned char  __brdInitFlag;

#include <hdwDefs.c>



//--- Parallel port B bitmasks ---
#define  SPI1_CLK         0x01      // PB0
#define  DSP_RESET_       0x40      // PB6
#define  PROG_B_          0x80      // PB7
#define  DSP_FPGA_RESETS  0xC0      // PB6 and PB7

//--- Parallel port D bitmasks ---
#define  FPGA_EEPROM_CS_  0x10      // PD4
#define  DSP_EEPROM_CS_   0x20      // PD5

//--- Parallel port E bitmasks ---
#define  CPU_RDY          0x02      // PE1
#define  ACTY             0x08      // PE3
#define  FPGA_DONE        0x10      // PE4
#define  FPGA_OE          0x20      // PE5
#define  FPGA_WE          0x40      // PE6
#define  FPGA_CLK         0x80      // PE7

//--- Parallel port F bitmasks ---
#define  _SPI1_DSP_EN     0x01      // PF0
#define  SPI2_CLK         0x02      // PF1



//--- instruction codes usable by both ATMEL & SST SPI Serial EEPROMs ---
#define  EEPROM_WRSR          0x01   // WRite Status Register
#define  EEPROM_WRITE_BYTE    0x02   // Write a single byte
#define  EEPROM_READ_LF       0x03   // Read bytes (low frequency)
#define  EEPROM_WRDI          0x04   // WRite DIsable latch
#define  EEPROM_RDSR          0x05   // ReaD Status Register
#define  EEPROM_WREN          0x06   // WRite ENable
#define  EEPROM_READ          0x0B   // Read bytes
#define  EEPROM_ERASE_4K      0x20   // Erase a 4K Byte block
#define  EEPROM_ERASE_32K     0x52   // Erase a 32K Byte block
#define  EEPROM_ERASE_ALL     0x60   // Erase entire chip
#define  EEPROM_RDID          0x9F   // Read device ID
#define  EEPROM_ERASE_64K     0xD8   // Erase a 64K Byte block

//--- ATMEL-specific SPI Serial EEPROM commands ---
#define  EEPROM_PROT          0x36   // PROTect sector
#define  EEPROM_UNPR          0x39   // UNPRotect sector
#define  EEPROM_RD_PROT_REGS  0x3C   // Read Sector Protection Registers
#define  EEPROM_POWER_RESUME  0xAB   // Resume from Deep-Power-Down
#define  EEPROM_SEQ_PROG      0xAF   // Write a sequence of bytes
#define  EEPROM_POWER_DOWN    0xB9   // Deep-Power-Down

//--- SST-specific SPI Serial EEPROM commands ---
#define  EEPROM_EWSR          0x50   // Enable Write Status Register
#define  EEPROM_AAI_PROG      0xAD   // Write a sequence of bytes


uchar  csEEPROM;

#define  EEPROM_Enable()   outi(PDDR, srPDDR &= ~csEEPROM)  // Chip Select low
#define  EEPROM_Disable()  outi(PDDR, srPDDR |=  csEEPROM)  // Chip Select high

//--- Manufacturer IDs returned by RD_ID command ---
#define  ID_ATMEL   0x1F
#define  ID_SST     0xBF


/*
  Each transmission to the DSP on SPI1 must be 32-bits.  Setting the enable
  signal low connects the DSP to the SPI1 bus as a slave and prepares it to
  receive the 32-bit packet.  Setting the signal back high signals the end of
  the transmission and causes the DSP to process the packet.

  The normal sequence of packets is to send a 32-bit ID of the value we want
  to read or write, and then the 32-bit value to be written.
*/

#define  RespOffset  2  // # packets DSP responses are delayed by


#define  WriteSPI1(buf, len)    _spiWrite_(spi1Port, buf, len, false)
//#define  WriteSPI1(buf, len)   write(spi1fd, buf, len)
#define  ReadSPI1(buf, len)     _spiRead_(spi1Port, buf, len, false)
//#define  ReadSPI12(buf, len)    read(spi1fd, buf, len)

#define  WriteSPI2(buf, len)    _spiWrite_(spi2Port, buf, len, false)
//#define  WriteSPI2(buf, len)   write(spi2fd, buf, len)
#define  ReadSPI2(buf, len)     _spiRead_(spi2Port, buf, len, false)
//#define  ReadSPI2(buf, len)    read(spi2fd, buf, len)



fd_t  spi1fd, spi2fd;
SerialInfo  *spi1Port, *spi2Port;  // can't share fd's between processes yet...



//=============================================================================
_nearcall float adcGetRaw(uint ui)  { return (float) adcRaw[ui]; }

_nearcall float adcMinRaw(uint ui)  { return 0.0; }
_nearcall float adcMaxRaw(uint ui)  { return (float)0xffff; }

_nearcall void adcDefaults(uint ui, float *gain, float *offset)
{ *gain = 0.0003046;  *offset = 32786.0; }

_nearcall void adcCalibPoints(uint ui, float *low, float *high)
{ *low = -9.5;  *high = 9.5; }

_nearcall char *adcUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================
_nearcall float dacGetRaw(uint ui)  { return (float) dacRaw[ui]; }

_nearcall float dacMinRaw(uint ui)  { return 0.0; }
_nearcall float dacMaxRaw(uint ui)  { return (float)0xffffL; }

_nearcall void dacDefaults(uint ui, float *gain, float *offset)
{  *gain = 0.000311;  *offset = 32768.0;  }

_nearcall void dacCalibPoints(uint ui, float *low, float *high)
  {  *low = -9.5;  *high = 9.5; }

_nearcall char *dacUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================


//-----------------------------------------------------------------------------
//--- set _DSP_EN low ---
#define  DSP_Packet_Start()  \
    outi(PFDR, srPFDR &= (uchar)(~_SPI1_DSP_EN))

//-----------------------------------------------------------------------------
void DSP_Packet_End(void)
{
  outi(PFDR, srPFDR |= (uchar)(_SPI1_DSP_EN));  // set _DSP_EN high

  outi(SBCR, srSBCR &= ~0x0C);  // make PB0 normal I/O signal

  outi(PBDR, srPBDR |=  (uchar)(SPI1_CLK));  // set SPI1_CLK high
  outi(PBDR, srPBDR &= ~(uchar)(SPI1_CLK));  // set SPI1_CLK low

  outi(SBCR, srSBCR |= 0x0C);   // make PB0 serial B clock
}


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
Function: int _readExtioPort ( int port );

Note:
This function is setup to provide additional hold time when
accessing external I/O addresses 0000 and 0002. This routine
must stay         and not be changed, for the additional
address hold-time to be valid.

Improved by Softools to not require a copy to RAM because
real tools have align and other stuff for real programmers.
************************************************************/
/*
unsigned readExtIO (unsigned port)      // Port is 0 or 2
{
#asm
    push    ip
    ipset   3
    ld      a,l
    and     a,0x03
    jr      z,..read0
    jr      ..skip
    .align  4           ; Align to clear A0 & A1
..read0:
    nop                 ; A0 = 0, A1 = 0
    nop                 ; A0 = 1, A1 = 0
    ioe     ld l,(hl)   ; A0 = 0, A1 = 1
    pop     ip          ; A0 = 0, A1 = 0
    jr      ..exit
    .align  4           ; Align to clear A0 & A1
..skip:
    ioe     ld l,(hl)   ; A0 = 0, A1 = 0 (2 bytes)
    pop     ip          ; A0 = 0, A1 = 1
..exit:
#endasm
}
*/

//-----------------------------------------------------------------------------
//  Determine the device ID based on external hardware settings
//-----------------------------------------------------------------------------
//extern bool useLastID;  // to allow temp disabling of "reading" the device ID

ulong ReadDeviceID()
{
  uchar  inA, inB;
  uint  devID;


//  if (useLastID)  return (ulong) devID;

  inA = ~ini(PADR);
  inB = ~ini(PBDR);

//--- device ID comes from 12-bit switch inputs ----
  devID = (((uint)inB << 6) & 0x0F00) | (uint)inA;

  return (ulong) devID;
}

//-----------------------------------------------------------------------------
//  return the current state of the digital inputs
//-----------------------------------------------------------------------------
ulong ReadDigInputs(void)
{
  uchar  inE;

  inE = ini(PEDR);

  return (ulong)((inE >> 4) & 1);
}


//-----------------------------------------------------------------------------
//  set the state of the digital outputs
//-----------------------------------------------------------------------------
void WriteDigOutputs(ulong digOut_)
{
  uchar  byt = 0;
  ulong  digOut;


  digOut = (digOut_ & ~forcedDigOutMask) | forcedDigOutBits;

  if ((uchar)digOut & 0x01)  byt  = 0x02;  // CPU_RDY
  if ((uchar)digOut & 0x02)  byt |= 0x08;  // ACTY (Activity LED)

  outi(PEDR, srPEDR = (srPEDR & ~0x0A) | byt);
}


//-----------------------------------------------------------------------------
void readAnalogInputs(ulong mask)
{
}


//-----------------------------------------------------------------------------
//  Update the state of the AD5542 DAC chips
//-----------------------------------------------------------------------------
void writeAnalogOutputs(ulong mask)
{
}


//-----------------------------------------------------------------------------
/*
float ADCtoVolts(uchar chan, uint rawADC)
{
  return (ADCOffset(chan) - (float)(rawADC)) * ADCGain(chan);
}
*/

//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to transmit mode
//-----------------------------------------------------------------------------
void setExt485Tx(void)
{
}

//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to receive mode
//-----------------------------------------------------------------------------
void setExt485Rx(void)
{
}

//-----------------------------------------------------------------------------
//  functions to enable/disable the transmitter for an RS-485 connection
//-----------------------------------------------------------------------------
void _nearcall SerSetDirA(int enableXmit)  {  }
void _nearcall SerSetDirB(int enableXmit)  {  }
void _nearcall SerSetDirC(int enableXmit)  {  }
void _nearcall SerSetDirD(int enableXmit)  {  }


//-----------------------------------------------------------------------------
void hdwInit(ulong digOutState)
{
  hdwType = HT_LLRF2;

  __brdInitFlag = TRUE;  // useLastID = false;

// (NI) = Normal Input
// (NO) = Normal Output
// (AI) = Alternate Input
// (AO) = Alternate Output
// (xx) = Used by ZWorld core - do not change

//--- Port A --- (*must be inputs due to bug in earlier Rabbit 3000's)
//  PA7 - (NI)* ADDR7:  device ID bit 07
//  PA6 - (NI)* ADDR6:  device ID bit 06
//  PA5 - (NI)* ADDR5:  device ID bit 05
//  PA4 - (NI)* ADDR4:  device ID bit 04
//  PA3 - (NI)* ADDR3:  device ID bit 03
//  PA2 - (NI)* ADDR2:  device ID bit 02
//  PA1 - (NI)* ADDR1:  device ID bit 01
//  PA0 - (NI)* ADDR0:  device ID bit 00
  outi(SPCR, srSPCR = 0x80);  // PortA: byte input

//--- Port B ---
//  PB7 - (NO)  PROG_B:  low holds FPGA in reset
//  PB6 - (NO)  /DSP_RESET:  low holds DSP in reset
//  PB5 - (NI)  ADDR11: device ID bit 11
//  PB4 - (NI)  ADDR10  device ID bit 10
//  PB3 - (NI)  ADDR9:  device ID bit 09
//  PB2 - (NI)  ADDR8:  device ID bit 08
//  PB1 - (AO)  Reserved:  Serial-A (Diag/Prog) clock
//  PB0 - (AO)  SPI1_CLK:  Serial-B (SPI1) clock
  outi(PBDR,  srPBDR = 0x00);   // All outputs low
  outi(PBDDR, srPBDDR = 0xC3);  // dir: OOII-IIOO (1100-0011)

//--- Port C ---
//  PC7 - (xI)  Reserved:  Serial-A (Diag/Prog) receive
//  PC6 - (AO)  Reserved:  Serial-A (Diag/Prog) transmit
//  PC5 - (xI)  SPI1_MISO:  Serial-B (SPI1) receive
//  PC4 - (AO)  SPI1_MOSI:  Serial-B (SPI1) transmit
//  PC3 - (xI)  SPI2_MISO:  Serial-C (SPI2) receive
//  PC2 - (AO)  SPI2_MOSI:  Serial-C (SPI2) transmit
//  PC1 - (xI)  not used:  Serial-D (not used)
//  PC0 - (AO)  not used:  Serial-D (not used)
  // (NOTE: All Port C signals have fixed Input/Output modes)
  outi(PCFR, srPCFR |= 0x55);  // func: xAxA-xAxA (x1x1-0x1x1)

//--- Port D ---
//  PD7 - (xx)  Reserved:  Ethernet Controller (TPIN+)
//  PD6 - (xx)  Reserved:  Ethernet Controller (TPIN-)
//  PD5 - (NO)  SPI2_DSPen:  /CS for EEPROM used by DSP
//  PD4 - (NO)  SPI2_FPGAen: /CS for EEPROM used by FPGA
//  PD3 - (xx)  Reserved:  Ethernet Controller (TPOUT+)
//  PD2 - (xx)  Reserved:  Ethernet Controller (TPOUT-)
//  PD1 - (xx)  Not Avail: (pulled high - not even connected???)
//  PD0 - (xx)  Reserved:  Ethernet Controller Reset
  outi(PDDCR, srPDDCR |= 0x30);  // drive: xxZZ-xxxx  (xx11-xxxx)
  outi(PDDDR, srPDDDR |= 0x30);  // dir:   xxOO-xxxx  (xx11-xxxx)
  outi(PDFR,  srPDFR  &= 0xCF);  // func:  xxNN-xxxx  (xx00-xxxx)
  outi(PDDR,  srPDDR &= ~0x30);  // Both /CS's in Hi-Z state (due to open drain)

//--- Port E ---
//  PE7 - (NO)  FPGA_CLK:  (not currently used)
//  PE6 - (NO)  FPGA_WE:   Write Enable:  Pulse high to write to FPGA
//  PE5 - (NO)  FPGA_OE:   Output Enable:  Set high to read from FPGA
//  PE4 - (NI)  FPGA_DONE:  FPGA is done loading config from EEPROM
//  PE3 - (NO)  Activity LED
//  PE2 - (xx)  Reserved:  Ethernet Controller Chip Select (output)
//  PE1 - (NO)  CPU_RDY:   Rabbit booted and running application code
//  PE0 - (xx)  Reserved:  Interrupt Input 0A (Ethernet ctlr Interrupt)
  outi(PEDDR, srPEDDR = (srPEDDR & 0x05) | 0xEA); // dir:  OOOI-OxOx (1110-1x1x)
  outi(PEFR, srPEFR & 0x05);                      // func: NNNN-NxNx (0000-0x0x)
  outi(PEDR, srPEDR &= 0x05);                     //State: (000x-0x0x)

//--- Port F --- reads Device ID switches ---
//  PF7 - (NO)  FPGA_A5:  Address bit 5 for FPGA read/write
//  PF6 - (NO)  FPGA_A4:  Address bit 4 for FPGA read/write
//  PF5 - (NO)  FPGA_A3:  Address bit 3 for FPGA read/write
//  PF4 - (NO)  FPGA_A2:  Address bit 2 for FPGA read/write
//  PF3 - (NO)  FPGA_A1:  Address bit 1 for FPGA read/write
//  PF2 - (NO)  FPGA_A0:  Address bit 0 for FPGA read/write
//  PF1 - (NO)  SPI2_CLK:  SPI2 bus clock (use Hi-Z state when not in use)
//  PF0 - (NO)  SPI1_DSP_EN:  Connect DSP to thhe SPI1 bus as slave device
  outi(PFDDR, srPFDDR = 0xFF);  // dir:  OOOO-OOOO (1111-1111)
  outi(PFFR,  srPFFR  = 0x00);  // func: NNNN-NNNN (0000-0000)
  outi(PFDCR, srPFDCR = 0x02);  //drive: NNNN-NNZN (0000-0010) (open-drain CLK)

//--- Port G --- used as generic digital outputs ---
//  PG7 - (NI)  FPGA_D7:  Data bit 7 for parallel read/write from/to FPGA
//  PG6 - (NI)  FPGA_D6:  Data bit 6 for parallel read/write from/to FPGA
//  PG5 - (NI)  FPGA_D5:  Data bit 5 for parallel read/write from/to FPGA
//  PG4 - (NI)  FPGA_D4:  Data bit 4 for parallel read/write from/to FPGA
//  PG3 - (NI)  FPGA_D3:  Data bit 3 for parallel read/write from/to FPGA
//  PG2 - (NI)  FPGA_D2:  Data bit 2 for parallel read/write from/to FPGA
//  PG1 - (NI)  FPGA_D1:  Data bit 1 for parallel read/write from/to FPGA
//  PG0 - (NI)  FPGA_D0:  Data bit 0 for parallel read/write from/to FPGA
  outi(PGDDR, srPGDDR = 0x00);  // dir:  IIII-IIII (0000-0000)
  outi(PGFR,  srPGFR  = 0x00);  // func: NNNN-NNNN (0000-0000)
  outi(PGDCR, srPGDCR = 0x00);  // no open-drain outputs


//--- set state of digital outputs ---
  digitalOut = digOutState;  WriteDigOutputs(digOutState);



  flashSize = 524288L;

  printf("\r\nInitialized %s hdw\r\n", ArchName);
}


//-----------------------------------------------------------------------------
//  Initialize and register hardware-dependent file systems
//-----------------------------------------------------------------------------
void registerFileSystems(void)
{
  socketInit_();
  srfsInit_();
  serialInit_();
#ifdef ASFLASH
  asflashInit_();
#endif

  flashFileInit_();
}

//-----------------------------------------------------------------------------
//  Initialize the network hardware
//-----------------------------------------------------------------------------
void netHdwInit()
{
//  Select ethernet hardware:
//  usingAsix, usingDevKit, usingRealtek, usingSMSC
//  usingAll() can be used to support all cmd core modules, but req's more code

  usingAsix();
}


//-----------------------------------------------------------------------------
//  allocate any resources that should be owned by the kernel process
//
//  To be called ONLY by the kernel process!!!!
//-----------------------------------------------------------------------------
void hdwInit2(void)
{
  int  clockRate;


//  char  *argv_[3];
//  uint  argc_;

//  argv_[0] = "getty";  argv_[1] = "19200";  argv_[2] = "tty3";  argc_ = 3;
//  createProc(getty, 5, argc_, argv_);

//--- open serial ports for SPI connections ---
  clockRate = CLK_1MHz;

  spi1fd = open("/dev/spi1", O_WRONLY, 0660);  spi1Port = NULL;
  printf("spi1fd: %d\r\n", spi1fd);
  if (spi1fd < 0)
    printf("*** Unable to open '/dev/spi1' ***\r\n");
  else  {
    ioctl(spi1fd, IOC_SET_CLK_RATE, &clockRate);
    spi1Port = _FileFromFD_(spi1fd)->ioDev;
  }

  spi2fd = open("/dev/spi2", O_WRONLY, 0660);  spi2Port = NULL;
  printf("spi2fd: %d\r\n", spi2fd);
  if (spi2fd < 0)
    printf("*** Unable to open '/dev/spi2' ***\r\n");
  else  {
    ioctl(spi2fd, IOC_SET_CLK_RATE, &clockRate);
    spi2Port = _FileFromFD_(spi2fd)->ioDev;
  }


}


//=============================================================================
//  Hardware platform-specifc functions.  The simLLRF2.c module contains
//  hardware-independent versions of these that may provide some minimal
//  simulation for the purpose of doing basic testing of the application
//  logic on other platforms.
//=============================================================================

//-----------------------------------------------------------------------------
//  Get the APPL_LOCK and initialize the SPI port used to talk to the DSP
//
//  Returns lock depth > 0 on success, <= 0 if an error
//-----------------------------------------------------------------------------
int getLockDSP(ulong timeOut)
{
  uchar  clockMode = NORM_LOW_LATCH_RISE;
  int  lockDepth;


  if (hdwType == HT_LLRF2)  if (!spi1Port)  { errno = EBADF; return -1; }

  lockDepth = getLock(APPL_LOCK, timeOut);
  if (lockDepth <= 0)  return lockDepth;

  if (hdwType == HT_LLRF2)  if (lockDepth == 1)
  //  ioctl(spi1fd, IOC_SET_CLK_MODE, &clockMode);
    _spiConfig_(spi1Port, CLK_1MHz, clockMode);

  return lockDepth;
}

//-----------------------------------------------------------------------------
//  release the APPL_LOCK and allow other tasks to use the DSP interface
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int releaseLockDSP(void)
{
  return releaseLock(APPL_LOCK);
}


//-----------------------------------------------------------------------------
//  Reset the DSP so it (re)loads its config from the EEPROM.
//
//  Returns 0 on success, <0 if an error occurs.
//-----------------------------------------------------------------------------
int resetDSP(void)
{
  uint  startTime;


  if (getLock(APPL_LOCK, 5000) <= 0)  return -1;

  prepForReset();

  outi(PBDR, srPBDR &= ~DSP_RESET_);  // set /DSP_RESET signal low

  startTime = GetMS_();  while (msSince_(startTime) < 500);  // wait 500 ms

  outi(PBDR, srPBDR |= DSP_RESET_);  // set /DSP_RESET signal high

  releaseLock(APPL_LOCK);  return 0;
}



//-----------------------------------------------------------------------------
//  Place the DSP and FPGA in a reset state for the purpose of allowing the
//  Rabbit to access their respective EEPROMs via the SPI2 bus.  It is the
//  responsibility of the caller to obtain the proper lock BEFORE calling this
//  function.  Failure to do so can result in multiple master devices trying to
//  use the SPI2 bus at the same time with unpredictable results!
//-----------------------------------------------------------------------------
int stop_DSP_FPGA(void)
{
  if (NotLockOwner(APPL_LOCK))  return -1;

  printf("Placing DSP and FPGA in RESET mode ...\r\n");

//--- set /DSP_RESET and PROG_B low to place the DSP and FPGA in reset mode ---
  outi(PBDR, srPBDR &= ~DSP_FPGA_RESETS);  // set PB6 and PB7 low

//--- set high/low drive mode for SPI2 CLK and /CS for both EEPROMS ---
  outi(PFDCR, srPFDCR = 0x00);    // drive: NNNN-NNNN (0000-0000)
  outi(PDDCR, srPDDCR &= ~0x30);  // drive: xxNN-xxxx  (xx00-xxxx)

  fpgaInitialized = false;

  return 0;
}

//-----------------------------------------------------------------------------
//  Reset the DSP and FPGA, causing them to load and use the configurations
//  currently stored in the EEPROMs.  To reset the DSP and FPGA, call
//  stop_DSP_FPGA() and then restart_DSP_FPGA()
//
//  Returns 0 on success, or <0 if an error
//-----------------------------------------------------------------------------
int restart_DSP_FPGA(ulong waitMS)
{
  ulong  startTime;


  if (NotLockOwner(APPL_LOCK))  return -1;
  if (LockDepth(APPL_LOCK) != 1)  return -1;

  printf("Restarting DSP and FPGA ...\r\n");

//--- back to using open-drain mode for SPI2 CLK and /CS for both EEPROMS ---
  outi(PFDCR, srPFDCR = 0x02);   // drive: NNNN-NNZN (0000-0010)
  outi(PDDCR, srPDDCR |= 0x30);  // drive: xxZZ-xxxx  (xx11-xxxx)

//--- release FPGA from reset and wait for it to (re)load its config ---
  outi(PBDR, srPBDR |= PROG_B_);  // set PB7 (PROG_B) high

  startTime = GetMS();
  for (;;)  {
    resetWDT();  // so WatchDog timer doesn't reboot us...
//if (diagLogInfo->wdtTicks > 10)  //tdebug
//  printf("=== %lu %u\r\n", msSince(startTime), diagLogInfo->wdtTicks);  //tdebug
    if (ini(PEDR) & FPGA_DONE)  break;
    if (msSince(startTime) > waitMS)  break;
    sleepMS(250);
  }
  if (! (ini(PEDR) & FPGA_DONE))  {
    printf("FPGA never released SPI bus\r\n");  return -1; }

  printf("FPGA restarted\r\n");

  fpgaInitialized = true;

//--- release DSP from reset so it will (re)load its config ---
  outi(PBDR, srPBDR |= DSP_RESET_);

  return 0;
}


//-----------------------------------------------------------------------------
//  write a single 32-bit register value to the DSP (without any additional
//  cycles to read a response, which can interfere with any larger sequence
//  of packets this may be imbedded in).
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int writeDSPRegQuick(ulong regNum, ulong value)
{
  ulong  xmitWord, rcvWord;


  if (getLockDSP(1000) <= 0)  return -1;

  regNum |= 1L;

//--- send the register # in the 1st packet ---
  xmitWord = htonl(regNum);
  DSP_Packet_Start();
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  DSP_Packet_End();

//--- followed by the value to be written in the 2nd packet ---
  xmitWord = htonl(value);
  DSP_Packet_Start();
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  DSP_Packet_End();

  releaseLockDSP();

  return 0;
}


//-----------------------------------------------------------------------------
//  write a single 32-bit register value to the DSP and read back the echoed
//  addr and value (which at least lets us know we are talking to the DSP).
//
//  Returns 0 on success, < 0 if an error
//-----------------------------------------------------------------------------
int writeDSPReg(ulong regNum, ulong value, bool showErrs)
{
  ulong  xmitWord, rcvWord;


  if (getLockDSP(1000) <= 0)  return -1;

  regNum |= 1L;

//--- send the register # in the 1st packet ---
  xmitWord = htonl(regNum);
  DSP_Packet_Start();
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  DSP_Packet_End();
  if (showFiller)  printf("  [%08lX", ntohl(rcvWord));

//--- followed by the value to be written in the 2nd packet ---
  xmitWord = htonl(value);
  DSP_Packet_Start();
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  DSP_Packet_End();
  if (showFiller)  printf("  %08lX]\r\n", ntohl(rcvWord));

//--- read back the echoed register # ---
  xmitWord = 0;
  DSP_Packet_Start();
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  DSP_Packet_End();
  rcvWord = ntohl(rcvWord);

  if (rcvWord != regNum)  {
    if (showErrs)
      printf("  echoed write address error: 0x%08lX  echoed for 0x%08lX\r\n",
             rcvWord, regNum);
    releaseLockDSP();  return -2;
  }

//--- read back the echoed value ---
  xmitWord = 0;
  DSP_Packet_Start();
  _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
  DSP_Packet_End();
  rcvWord = ntohl(rcvWord);

  if (rcvWord != value)  {
    if (showErrs)
      printf("  echoed value error: 0x%08lX  echoed for 0x%08lX\r\n",
             rcvWord, value);
    releaseLockDSP();  return -3;
  }

  releaseLockDSP();

  return 0;
}


//-----------------------------------------------------------------------------
//  Read a single 32-bit register value from the DSP
//
//  Returns 0 and sets *regVal on success, < 0 if an error
//-----------------------------------------------------------------------------
int readDSPReg(ulong regNum, ulong *regVal)
{
  uint  i, wordCount;
  ulong  xmitWord, rcvWord;  //, respRegNum;
//  char  temp[32];


  if (getLockDSP(1000) <= 0)  return -1;

  regNum &= 0xFFFFFFFEL;

  wordCount = 1 + RespOffset;

  xmitWord = htonl(regNum);

  for (i=0; i<wordCount; ++i,regNum+=2)  {
    if (hdwType != HT_LLRF2)  {
      *((float *)&rcvWord) = 0.1234 + (float)(uint)(GetMS() & 0x3F);
      rcvWord = htonl(rcvWord);
    }
    else  {
      DSP_Packet_Start();
      _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
      DSP_Packet_End();
    }
    xmitWord = 0;
  }

  releaseLockDSP();

  rcvWord = ntohl(rcvWord);

//  respRegNum = regNum - (ulong)(RespOffset * 2);
//  printf("  0x%08LX:  0x%08LX  %12ld  %.32s\r\n",
//         respRegNum, rcvWord, rcvWord, fmtDSPval(rcvWord, respRegNum, temp));

  *regVal = rcvWord;

  return 0;
}


//-----------------------------------------------------------------------------
//  Write and Read a group of long integers to/from the DSP
//
//  The counts are the # of long (32-bit) words to send/rcv
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  Exchange of data with the DSP is done using a bi-directional (simultaneous
//  send/recv) clocked SPI interface.  All data is exchanged in 32-bit (4-byte)
//  packets.  While we are sending a 32-bit packet to the DSP, it is using the
//  same clock pulses to simultaneously send a 32-bit packet in response.
//
//  Each exchange of a 32-bit packet consists of setting the DSP /Enable signal
//  low, writing (while simultaneously reading) 32-bits of data, and then
//  setting the DSP /Enable signal high.
//
//  Due to the way the DSP buffers are set up, all responses from the DSP are
//  always 2 packets "behind" the ones that contained the commands that they
//  are a response to.  We can write as many commands as we want in a
//  sequence of packets, but the response to each one will always be received
//  2 packets later.  Given that we generally want to read the response to
//  a set of commands as soon as possible, we normally have to add some
//  dummy or filler packets to the sequence to get the desired response
//  packets.  For extended responses (like reading of waveform data), we must
//  send lots of filler packets to get all the data.
//-----------------------------------------------------------------------------
int readPlot(ulong regNum, ulong far *rcvBuf, uint rcvCount)
{
  uint  u;
  ulong  xmitWord, rcvWord;


  if (hdwType != HT_LLRF2)  return 0;

  if (!spi1Port)  return -1;

  if (getLockDSP(1000) <= 0)  return -1;

  if (_spiLock_(spi1Port))  { releaseLockDSP();  return -1; }

  xmitWord = htonl(regNum);

//--- "prime the pump":  Ignore the first group of readings ---
  for (u=0; u<RespOffset; ++u)  {
    DSP_Packet_Start();
    _spiWrRdFast_((char *)&xmitWord, (char *)&rcvWord, 4);
    DSP_Packet_End();
  }
  for (u=0; u<rcvCount; ++u)  {
    DSP_Packet_Start();
    _spiWrRdFast_((char *)&xmitWord, (char *)&rcvWord, 4);
    DSP_Packet_End();
    *rcvBuf = rcvWord;  ++rcvBuf;
  }

  _spiUnlock_(spi1Port);

  releaseLockDSP();

  return 0;
}

//-----------------------------------------------------------------------------
//  Read the values for the RO DSP registers and update the value of the
//  corresponding ModBus registers.
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  See WrRdDSP() above for details on how the DSP interface works
//-----------------------------------------------------------------------------
int read_DSP_Regs(const RD_DSP_REG far *rdDSPregs)
{
  uint  i, wordCount;
  ulong  xmitWord, dspVal;
  RegInfo *mbReg;
  const RD_DSP_REG far *reqReg;
  const RD_DSP_REG far *dspReg;


  if (getLockDSP(1000) <= 0)  return -1;

  wordCount = Num_RD_DSP_Regs + RespOffset;

  for (i=0,reqReg=rdDSPregs; i<wordCount; ++i,++reqReg)  {

    xmitWord = (i < Num_RD_DSP_Regs) ? htonl(reqReg->id) : 0;

    if (hdwType != HT_LLRF2)  {
      *((float *)&dspVal) = 0.1234 + (float)(uint)(GetMS() & 0x3F);
      dspVal = htonl(dspVal);
    }
    else  {
      DSP_Packet_Start();
      _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&dspVal, 4);
      DSP_Packet_End();
    }

    if (i < RespOffset)  continue;  // responses are offset by RespOffset pkts

    dspReg = reqReg - RespOffset;

  //--- convert value as needed and update the ModBus register value ---
    mbReg = dspReg->mbReg;  dspVal = ntohl(dspVal);
    dsp2mb(mbReg, mbReg->regVal, dspReg, &dspVal);
  }

  releaseLockDSP();

  return 0;
}

//-----------------------------------------------------------------------------
//  Send the current values to the DSP for each DSP R/W register that
//  represents a run-time setting (i.e. ones with writeMode == WM_EVERY).
//
//  Returns 0 on success, < 0 if an error
// ---------------------------------------------------------------------------
//  See WrRdDSP() above for details on how the DSP interface works
//-----------------------------------------------------------------------------
int write_DSP_Regs(const WR_DSP_REG far *wrDSPregs)
{
  uint  i;
  ulong  xmitWord, rcvWord, dspVal, regNum;
  const WR_DSP_REG far *dspReg;
  RegInfo  *mbReg;


  if (hdwType == HT_LLRF2)  if (!spi1Port)  return -1;

  if (getLockDSP(1000) <= 0)  return -2;

  for (i=0,dspReg=wrDSPregs; i<Num_WR_DSP_Regs; ++i,++dspReg)  {
    if (dspReg->writeMode != WM_EVERY)  continue;

    regNum = dspReg->id | 1L;

  //--- convert ModBus value to DSP format ---
    mbReg = dspReg->mbReg;
    mb2dsp(mbReg, mbReg->regVal, dspReg, &dspVal);

  //--- send the register # in the 1st packet ---
    xmitWord = htonl(regNum);
    DSP_Packet_Start();
    _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
    DSP_Packet_End();

  //--- send the value in the 2nd packet ---
    xmitWord = htonl(dspVal);
    DSP_Packet_Start();
    _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
    DSP_Packet_End();
  }

  releaseLockDSP();

  return 0;
}




//-----------------------------------------------------------------------------
//  Read an 8-bit value from the FPGA
//
//  Returns the value (>= 0) on success, < 0 if an error
//-----------------------------------------------------------------------------
int readFPGA(uint regNum)
{
  uchar  regVal;


  if (getLock(APPL_LOCK, 100) <= 0)  return -1;

  outi(PFDR, srPFDR = ((uchar)regNum << 2));  // assert 6-bit address

  outi(PEDR, srPEDR | FPGA_OE);  // set OE high
  regVal = ini(PGDR);  // read 8-bit data value
  outi(PEDR, srPEDR & ~FPGA_OE);  // set OE low

  releaseLock(APPL_LOCK);

  return (int)regVal;
}


//-----------------------------------------------------------------------------
//  Read a sequence of FPGA registers.
//
//  Returns 0 on success, <0 if an error.
//-----------------------------------------------------------------------------
int read_FPGA_Regs(U16 *regVal[], uint firstReg, uint numRegs)
{
  uint  regNum, endReg;


  if (getLock(APPL_LOCK, 100) <= 0)  return -1;

  endReg = firstReg + numRegs;

  for (regNum=firstReg; regNum<endReg; ++regNum,++regVal)  {
    outi(PFDR, srPFDR = ((uchar)regNum << 2));  // assert 6-bit address

    outi(PEDR, srPEDR | FPGA_OE);  // set OE high
    **regVal = ini(PGDR);  // read 8-bit data value
    outi(PEDR, srPEDR & ~FPGA_OE);  // set OE low
  }

  releaseLock(APPL_LOCK);

  return 0;
}



//-----------------------------------------------------------------------------
//  Write an 8-bit value to the FPGA
//
//  Returns 0 on success, -1 if an error
//-----------------------------------------------------------------------------
int writeFPGA(uint regNum, uint regVal)
{
  if (getLock(APPL_LOCK, 1000) <= 0)  return -1;

  outi(PFDR, srPFDR = ((uchar)regNum << 2));  // assert 6-bit address
  outi(PGDDR, srPGDDR = 0xFF);  // set dir of Port G bits to all outputs
  outi(PGDR, srPGDR = (uchar)regVal);  // assert 8-bit data value
  outi(PEDR, srPEDR | FPGA_WE);  outi(PEDR, srPEDR);  // pulse WE high
  outi(PGDDR, srPGDDR = 0x00);  // set dir of Port G bits to all inputs

  releaseLock(APPL_LOCK);

  return 0;
}

//-----------------------------------------------------------------------------
//  Check the list of FPGA register values.  For each one that is not 0,
//  write the value to the FPGA, then set it to 0 and wirte it again
//  (effectively "pulsing" any non-0 values).
//
//  Returns 0 on success, <0 if an error.
//-----------------------------------------------------------------------------
int write_FPGA_Regs(U16 *regVal[], uint firstReg, uint numRegs)
{
  uint  regNum, endReg;


  if (getLock(APPL_LOCK, 100) <= 0)  return -1;

  endReg = firstReg + numRegs;

  for (regNum=firstReg; regNum<endReg; ++regNum,++regVal)  {
    if (! **regVal)  continue;

    outi(PFDR, srPFDR = ((uchar)regNum << 2));  // assert 6-bit address

    outi(PGDDR, srPGDDR = 0xFF);  // set dir of Port G bits to all outputs

    outi(PGDR, srPGDR = (uchar)regVal);  // assert 8-bit data value
    outi(PEDR, srPEDR | FPGA_WE);  outi(PEDR, srPEDR);  // pulse WE high

    **regVal = 0;  // set the value back to 0

    outi(PGDR, srPGDR = (uchar)regVal);  // assert 8-bit data value
    outi(PEDR, srPEDR | FPGA_WE);  outi(PEDR, srPEDR);  // pulse WE high

    outi(PGDDR, srPGDDR = 0x00);  // set dir of Port G bits to all inputs
  }

  releaseLock(APPL_LOCK);

  return 0;
}



//-----------------------------------------------------------------------------
//  write and then read specified # of bytes to/from the EEPROM
//-----------------------------------------------------------------------------
char *writeReadEE(char *send, int sendCount, char *rcv, int rcvCount)
{
  EEPROM_Enable();

  WriteSPI2(send, sendCount);
  if (rcvCount)  ReadSPI2(rcv, rcvCount);

  EEPROM_Disable();

  return rcv;
}

//-----------------------------------------------------------------------------
//  write a single byte command to the EEPROM, and read and return the
//  value of the status register
//-----------------------------------------------------------------------------
char writeCmdEE(uchar cmd)
{
  char  readStatCmd, statReg;

  if (cmd)  writeReadEE(&cmd, 1, NULL, 0);

  readStatCmd = EEPROM_RDSR;  writeReadEE(&readStatCmd, 1, &statReg, 1);

  return statReg;
}

//-----------------------------------------------------------------------------
const char SectorAddr[11][3] = {
  { 0x00,0x00,0x00 },  //  0
  { 0x01,0x00,0x00 },  //  1
  { 0x02,0x00,0x00 },  //  2
  { 0x03,0x00,0x00 },  //  3
  { 0x04,0x00,0x00 },  //  4
  { 0x05,0x00,0x00 },  //  5
  { 0x06,0x00,0x00 },  //  6
  { 0x07,0x00,0x00 },  //  7
  { 0x07,0x80,0x00 },  //  8
  { 0x07,0xA0,0x00 },  //  9
  { 0x07,0xC0,0x00 }   // 10
};

//-----------------------------------------------------------------------------
//  Wait up to the specified # of ms for the BUSY bit in the status register
//  for the currently selected EEPROM to clear.  Returns true if the bit
//  cleared before the specified timeout elapsed.
//-----------------------------------------------------------------------------
bool busyCleared(ulong maxWaitMS)
{
  ulong startTime = GetMS();

  while (msSince(startTime) < maxWaitMS)  {
    if (! (writeCmdEE(0) & 0x01))  return true;
    printf(".");  fflush(stdout);  sleepMS(500);
  }

  return false;
}


//-----------------------------------------------------------------------------
//  Update the firmware image in one of the EEPROMs.
//
//  The stopAt parameter specifies the address after which nothing in the
//  memory should be touched.  A value of 0 means the entire chip will be
//  erased, while values > 0 means nothing after the specified address will
//  be erased or written over.  The value will be rounded down to the nearest
//  multiple of 32K.
//
//  Returns 0 on success, < 0 if an error occured (with errno set)
//-----------------------------------------------------------------------------
int updateEEPROM(uint romNum, fd_t fromfd, ulong stopAt,
                 bool showProgress, bool verify)
{
#define  BufSize  256  // can't be very large without overflowing the stack!
                       // AND it MUST be an even # of bytes!

  uchar  clockMode = NORM_LOW_LATCH_RISE;
  bool  downloadOK;
  uint  manufacID, errCount;
  int  i, status, count;
  ulong  ttlBytes, fileSize;
  bool  busy, eraseOK, protectOK;
  char  statReg;
  struct stat  fsInfo;
  char  cmdBuf[6];
  char  buf[BufSize], buf2[BufSize];


  if (hdwType != HT_LLRF2)  return 0;

  if (!spi2Port OR (romNum > 1))  return -1;

  if (getLock(APPL_LOCK, 10000) <= 0)  return -1;

  downloadOK = false;


  stopAt &= 0xFF8000L;  // round down to nearest 32K boundary

//  ioctl(spi2fd, IOC_SET_CLK_MODE, &clockMode);
  _spiConfig_(spi2Port, CLK_1MHz, clockMode);

//--- stop the FPGA and DSP so we can access the EEPROM via the SPI2 bus ---
  stop_DSP_FPGA();

//--- make sure EEPROM is ready to accept commands ---
  csEEPROM = (romNum == DSP_EEPROM) ? DSP_EEPROM_CS_ : FPGA_EEPROM_CS_;
//  printf("*** EEPROM: %u:  CS: 0x%02X\r\n", romNum, csEEPROM);  //tdebug

  EEPROM_Disable();  writeCmdEE(EEPROM_WRDI);

  do {
  //--- read device ID ---
    *cmdBuf = EEPROM_RDID;  writeReadEE(cmdBuf, 1, buf2, 4);
    printf("  Device ID: %02X %02X %02X %02X\r\n",
           buf2[0], buf2[1], buf2[2], buf2[3] );
    manufacID = buf2[0];

  //===== unprotect all sectors =====
    printf("  Unprotecting all sectors...\r\n");  eraseOK = false;
    switch (manufacID)  {

      case ID_ATMEL:
        *cmdBuf = EEPROM_UNPR;
        for (i=0; i<11; ++i)  {
          if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;
          memcpy(cmdBuf+1, SectorAddr[i], 3);  writeReadEE(cmdBuf, 4, NULL, 0);
        }
        if (i == 11)  eraseOK = true;
        break;

      case ID_SST:
        if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;
        cmdBuf[0] = EEPROM_WRSR;  cmdBuf[1] = 0x00;
        writeReadEE(cmdBuf, 2, NULL, 0);
        if ((writeCmdEE(0) & 0x3C) == 0x00)  eraseOK = true;
        break;
    }

    if (!eraseOK)  { printf("  Unable to erase EEPROM\r\n");  break; }

  //===== erase part or all of the selected EEPROM =====
    if (stopAt)  {
      printf("  Erasing up to 0x%06lX (%lu)", stopAt, stopAt);
      memset(cmdBuf, 0, sizeof(cmdBuf));  cmdBuf[0] = EEPROM_ERASE_32K;
      for (ttlBytes=0; ttlBytes<stopAt; ttlBytes+=32768L)  {
        if (! (writeCmdEE(EEPROM_WREN) & 0x02))  { eraseOK = false;  break; }
        writeReadEE(cmdBuf, 4, NULL, 0);  cmdBuf[2] += (uchar)0x10;
        if (!busyCleared(5000))  { eraseOK = false;  break; }
      }
    }
    else  {
      printf("  Erasing entire EEPROM");
      if (! (writeCmdEE(EEPROM_WREN) & 0x02))  { eraseOK = false;  break; }
      if (! (statReg = writeCmdEE(EEPROM_ERASE_ALL) & 0x01))  {
        printf("\r\nUnable to initiate Chip Erase: Status: 0x%02X\r\n", statReg);
        eraseOK = false;  break;
      }
      eraseOK = busyCleared(12000);
    }

    printf("\r\n");
    if (!eraseOK)  { printf("Erase timed out\r\n");  break; }

    if (fstat(fromfd, &fsInfo) < 0)  { perror(NULL);  Exit(-2); }
    fileSize = (ulong)fsInfo.st_size;
    printf("  File size: %lu bytes\r\n", fileSize);


  //===== copy the firmware file from the source fd to the EEPROM =====
    printf("  Copying file to EEPROM...\r\n");
    ttlBytes = 0;  busy = false;

    if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;

    memset(cmdBuf, 0, sizeof(cmdBuf));

    cmdBuf[0] = (manufacID == ID_ATMEL) ? EEPROM_SEQ_PROG : EEPROM_AAI_PROG;

    while (((count = read(fromfd, buf, BufSize))) > 0)  {

      if (manufacID == ID_ATMEL)  {  // 1 byte at a time
        for (i=0; i<count; ++i)  {
          if (! ttlBytes)  {  // 1st byte requires cmd code and 3 addr bytes
            cmdBuf[4] = buf[0];  writeReadEE(cmdBuf, 5, NULL, 0);  }
          else  {  // subsequent bytes require only cmd code
            cmdBuf[1] = buf[i];  writeReadEE(cmdBuf, 2, NULL, 0);  }
//        busy = writeCmdEE(0) & 0x01;  if (busy)  break;
          ++ttlBytes;
          if (stopAt AND (ttlBytes >= stopAt))  break;
        }
      }

      if (manufacID == ID_SST)  { // 2 bytes at a time
        if (count & 1)  buf[count] = 0xFF;  // filler if odd # bytes
        for (i=0; i<count; i+=2)  {
          if (! ttlBytes)  {  // 1st word requires cmd code and 3 addr bytes
            cmdBuf[4] = buf[0];  cmdBuf[5] = buf[1];
            writeReadEE(cmdBuf, 6, NULL, 0);
          }
          else  {  // subsequent words require only cmd code
            cmdBuf[1] = buf[i];  cmdBuf[2] = buf[i+1];
            writeReadEE(cmdBuf, 3, NULL, 0);
          }
//        busy = writeCmdEE(0) & 0x01;  if (busy)  break;
          ttlBytes += 2;
          if (stopAt AND (ttlBytes >= stopAt))  break;
        }
      }

      if (busy)  break;
      if (showProgress)  { printf("%10lu\r", ttlBytes);  fflush(stdout); }
    }
    writeCmdEE(EEPROM_WRDI);

    if (busy OR (ttlBytes < fileSize) OR (ttlBytes > fileSize+1))
      printf("Error writing to EEPROM\r\n");
    else
      if (showProgress)  printf("  Wrote %lu bytes to EEPROM\r\n", ttlBytes);


  //===== protect all sectors =====
    printf("  Protecting all sectors...\r\n");  protectOK = false;
    switch (manufacID)  {

      case ID_ATMEL:
        *cmdBuf = EEPROM_PROT;
        for (i=0; i<11; ++i)  {
          if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;
          memcpy(cmdBuf+1, SectorAddr[i], 3);  writeReadEE(cmdBuf, 4, NULL, 0);
        }
        if (i == 11)  protectOK = true;
        break;

      case ID_SST:
        if (! (writeCmdEE(EEPROM_WREN) & 0x02))  break;
        cmdBuf[0] = EEPROM_WRSR;  cmdBuf[1] = 0x1C;
        writeReadEE(cmdBuf, 2, NULL, 0);
        if ((writeCmdEE(0) & 0x1C) == 0x1C)  protectOK = true;
        break;
    }

    if (!protectOK)  { printf("  Unable to protect EEPROM\r\n");  break; }



  //===== verify contents of EEPROM =====
    if (verify)  {
      printf("  Verifying EEPROM copy...\r\n");

      if (lseek(fromfd, 0, SEEK_SET) != 0)  {
        printf("Unable to reset source file pos to verify EEPROM\r\n");  break; }

      memset(cmdBuf, 0, sizeof(cmdBuf));  cmdBuf[0] = EEPROM_READ_LF;
      EEPROM_Enable();  WriteSPI2(cmdBuf, 4);

      ttlBytes = 0;  errCount = 0;

      while (((count = read(fromfd, buf, BufSize))) > 0)  {
        ReadSPI2(buf2, count);
        for (i=0; i<count; ++i)  {
//          if (!((uchar)ttlBytes & 0x0f))  printf("\r\n%06lX:", ttlBytes);
//          printf(" %02X", buf2[i]);
          if (buf[i] != buf2[i])  {
            printf("Mismatch at offset 0x%06lX:  EEPROM: 0x%02X,  File: 0x%02X\r\n",
                   ttlBytes+i, buf2[i], buf[i]);
            if (++errCount >= 100)  break;
          }
        }
        if (errCount >= 100)  {
          printf("*** Over 100 errors! ***\r\n");  break; }
        ttlBytes += count;
        if (showProgress)  { printf("%10lu\r", ttlBytes);  fflush(stdout); }
        if (ttlBytes >= fileSize)  break;  //tdebug
      }
      printf("\r\n");
      EEPROM_Disable();

      if ((ttlBytes == fileSize) AND !errCount)  {
        downloadOK = true;  printf("Verified %lu bytes\r\n", ttlBytes);  }
      else
        printf("Copies do not match!\r\n");
    }

  } while(0);

  if (downloadOK)  printf("FPGA firmware download complete!\r\n");


//--- Restart the FPGA and DSP so they (re)load their firmware from their ---
//--- respective EEPROMs. ---
  status = restart_DSP_FPGA(30000);

  releaseLock(APPL_LOCK);

  return status;
}




//-----------------------------------------------------------------------------
//  "dspread": command to read one or more 32-bit values from the DSP
//-----------------------------------------------------------------------------
void _task dspRead(int argc, char **argv)
{
  ulong  regNum;
  uint  i, count, wordCount;
  ulong  xmitWord, rcvWord, respRegNum;
  char  temp[64];


  if (showHelp(3, argc, argv))  {
    printf("Usage:  dspread <reg> <count>\r\n");  Exit(-1); }

  regNum = ATOUL(argv[1]) & 0xFFFFFFFEL;
  count = (uint) ATOI(argv[2]);

  if (getLockDSP(5000) <= 0)  { perror("getLockDSP");  Exit(-1); }

  wordCount = count + RespOffset;

  for (i=0; i<wordCount; ++i,regNum+=2)  {
    xmitWord = (i >= count) ? 0 : htonl(regNum);

    if (hdwType != HT_LLRF2)  {
      *((float *)&rcvWord) = 0.1234 + (float)(uint)(GetMS() & 0x3F);
      rcvWord = htonl(rcvWord);
    }
    else  {
      DSP_Packet_Start();
      _spiWrRd_(spi1Port, (char *)&xmitWord, (char *)&rcvWord, 4);
      DSP_Packet_End();
    }

    if (i < RespOffset)  {
      if (showFiller)  printf("  filler: 0x%08lX\r\n", ntohl(rcvWord));
      continue;
    }

    rcvWord = ntohl(rcvWord);

    respRegNum = regNum - (ulong)(RespOffset * 2);

    printf("  0x%08LX:  0x%08LX  %.32s\r\n",
           respRegNum, rcvWord, fmtDSPval(rcvWord, respRegNum, temp));
  }
  printf("\r\n");

  releaseLockDSP();

  Exit(0);
}

//-----------------------------------------------------------------------------
//  "dspwrite": command to write a value to the DSP
//-----------------------------------------------------------------------------
void _task dspWrite(int argc, char **argv)
{
  ulong  regNum, setVal;


  if (showHelp(3, argc, argv))  {
    printf("Usage:  dspwrite <reg> <value> [-f]\r\n");  Exit(-1); }

  if (!spi1Port)  { fprintf(DiagLog, "SPI port not open!\r\n");  Exit(-1); }

  regNum = ATOUL(argv[1]) | 1L;

  if (strstr(argv[2], "."))
    *(float *)&setVal = atof(argv[2]);
  else
    setVal = ATOUL(argv[2]);

  printf("  write 0x%08lX to 0x%08lX\r\n", setVal, regNum);

  writeDSPReg(regNum, setVal, true);

  //NOTE:  Can't just read back the result to check if it worked because
  //       reading from some registers triggers events and processing
  //       within the DSP.

  Exit(0);
}


//-----------------------------------------------------------------------------

