//----- SparkDetector.c ----- 02/09/10 -----(03/26/06)-----

//-----------------------------------------------------------------------------
//  Code to handle hardware-specific implementation of generic hdw functions
//-----------------------------------------------------------------------------
//
//  2010-02-09 - Mark Davis -
//               - Fixes to ReadDigInputs(), WriteDigOutputs() & hdwInit() to
//                 avoid race conditions
//
//  2008-10-22 - Mark Davis -
//               - removed inversion of digital output bits
//
//  2008-05-21 - Mark Davis -
//               - original version created using ps8rack as starting point
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <sys/backup.h>

#include <sys/spi.h>
#include <sys/hdwDefs.h>

#include <busybox.h>

#include <Rabbit.h>

#include <dcdefs.h>

#include <stcpip.h>

#include <sys/socket.h>
#include <sys/srfs.h>
#include <sys/flashFile.h>

#include <sys/farprintf.h>

//--- see hdwInit() for Digital I/O pin usage ---

const char ArchName[] = { "SparkDetector" };

#define  NUM_ADC   1
#define  NUM_DAC   1

#define  MAX_DIN    8
#define  MAX_DOUT   8


//--- these values can be set LOWER by the application ---
const uint  NumHdwADC = NUM_ADC;
const uint  NumHdwDAC = NUM_DAC;

const uint  MaxDIN =  MAX_DIN;
const uint  MaxDOUT = MAX_DOUT;


//--- NSCL 16-bit analog RCM3200-based ctlrs use port B for SPI to ADCs, DACs
#define  ADC_DAC_SPIPort  1  // for most RCM3200-based NSCL ctlr's w/Analog


unsigned char  __brdInitFlag;

#include <hdwDefs.c>


SerialInfo  *spiPort;


//=============================================================================
_nearcall float adcGetRaw(uint ui)  { return (float) adcRaw[ui]; }

_nearcall float adcMinRaw(uint ui)  { return 0.0; }
_nearcall float adcMaxRaw(uint ui)  { return (float)0xffff; }

_nearcall void adcDefaults(uint ui, float *gain, float *offset)
{ *gain = 0.0003046;  *offset = 32786.0; }

_nearcall void adcCalibPoints(uint ui, float *low, float *high)
{ *low = -9.5;  *high = 9.5; }

_nearcall char *adcUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================
_nearcall float dacGetRaw(uint ui)  { return (float) dacRaw[ui]; }

_nearcall float dacMinRaw(uint ui)  { return 0.0; }
_nearcall float dacMaxRaw(uint ui)  { return (float)0xffffL; }

_nearcall void dacDefaults(uint ui, float *gain, float *offset)
{  *gain = 0.000311;  *offset = 32768.0;  }

_nearcall void dacCalibPoints(uint ui, float *low, float *high)
  {  *low = -9.5;  *high = 9.5; }

_nearcall char *dacUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
Function: int _readExtioPort ( int port );

Note:
This function is setup to provide additional hold time when
accessing external I/O addresses 0000 and 0002. This routine
must stay         and not be changed, for the additional
address hold-time to be valid.

Improved by Softools to not require a copy to RAM because
real tools have align and other stuff for real programmers.
************************************************************/
/*
unsigned readExtIO (unsigned port)      // Port is 0 or 2
{
#asm
    push    ip
    ipset   3
    ld      a,l
    and     a,0x03
    jr      z,..read0
    jr      ..skip
    .align  4           ; Align to clear A0 & A1
..read0:
    nop                 ; A0 = 0, A1 = 0
    nop                 ; A0 = 1, A1 = 0
    ioe     ld l,(hl)   ; A0 = 0, A1 = 1
    pop     ip          ; A0 = 0, A1 = 0
    jr      ..exit
    .align  4           ; Align to clear A0 & A1
..skip:
    ioe     ld l,(hl)   ; A0 = 0, A1 = 0 (2 bytes)
    pop     ip          ; A0 = 0, A1 = 1
..exit:
#endasm
}
*/

//-----------------------------------------------------------------------------
//  determine the device ID based on external hardware settings
//-----------------------------------------------------------------------------
ulong ReadDeviceID()
{
  uchar  inA, devID;


  inA = RdPortI(PADR);

//--- device ID comes from 8-bit switch input ----
  devID = ~inA;

  return (ulong) devID;
}

//-----------------------------------------------------------------------------
//  Return the current state of the digital inputs
//
//  The digital inputs are connected to a series of octal line drivers whose
//  outputs are connected to the parallel G signal lines.  To read all the
//  inputs, we enable the drivers one at a time and read the state of the G
//  lines, adding the result to previously read bits.
//
//  The digital inputs and outputs are address in groups of 8 bits.  Which
//  group of 8 bits we are setting or reading is controlled by a 3-to-8 line
//  decoder, where toggling one of its 8 outputs low latches the desired
//  data.
//
//  To avoid activating the wrong one of the 8 outputs even for a moment, we
//  have to make sure the 3 inputs that specify which output to toggle low
//  are all stable BEFORE we enable the outputs and while we are disabling
//  the outputs.
//-----------------------------------------------------------------------------
ulong ReadDigInputs()
{
  ulong  inBits;
  uchar  otherBits = PBDRShadow & 0x87;


//--- read input bits 00-07 --- (Y1)
  WrPortI(PBDR, &PBDRShadow, otherBits | 0x60 );  // selct /Y1
  WrPortI(PBDR, &PBDRShadow, otherBits | 0x40 );  // /Y1 low to enable driver
  inBits = (ulong)RdPortI(PGDR);                  // read the data
  inBits = (inBits << 8) | (ulong)RdPortI(PGDR);  // read the data
  WrPortI(PBDR, &PBDRShadow, otherBits | 0x60 );  // /Y1 high again


  WrPortI(PBDR, &PBDRShadow, otherBits | 0x20);   // deselect all drivers

  inBits = (inBits & ~forcedDigInMask) | forcedDigInBits;

  return inBits;
}


//-----------------------------------------------------------------------------
//  set the state of the digital outputs
//
//  The digital inputs and outputs are address in groups of 8 bits.  Which
//  group of 8 bits we are setting or reading is controlled by a 3-to-8 line
//  decoder, where toggling one of its 8 outputs low latches the desired
//  data.
//
//  To avoid activating the wrong one of the 8 outputs even for a moment, we
//  have to make sure the 3 inputs that specify which output to toggle low
//  are all stable BEFORE we enable the outputs and while we are disabling
//  the outputs.
//-----------------------------------------------------------------------------
void WriteDigOutputs(ulong digOut)
{
  uchar  otherBits = PBDRShadow & 0x87;


  digOut = (digOut & ~forcedDigOutMask) | forcedDigOutBits;

//--- set digital I/O bus (port G) to output mode ---
  WrPortI(PGDDR, &PGDDRShadow, 0xFF);

//--- update output bits 00-07 --- (Y0)
  WrPortI(PGDR, &PGDRShadow, (uchar)digOut);         // put value on bus
  WrPortI(PBDR, &PBDRShadow, otherBits | 0x20);      // select /Y0
  WrPortI(PBDR, &PBDRShadow, otherBits );            // toggle it low
  WrPortI(PBDR, &PBDRShadow, otherBits | 0x20);      // then high again

//--- set digital I/O bus (port G) to input mode ---
  WrPortI(PGDDR, &PGDDRShadow, 0x00);
}


//-----------------------------------------------------------------------------
void readAnalogInputs(ulong mask)
{
}


//-----------------------------------------------------------------------------
//  Update the state of the AD5542 DAC chips
//-----------------------------------------------------------------------------
void writeAnalogOutputs(ulong mask)
{
}


//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to transmit mode
//-----------------------------------------------------------------------------
void setExt485Tx(void)
{
}
//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to receive mode
//-----------------------------------------------------------------------------
void setExt485Rx(void)
{
}

//-----------------------------------------------------------------------------
//  functions to enable/disable the transmitter for an RS-485 connection
//-----------------------------------------------------------------------------
void _nearcall SerSetDirA(int enableXmit)  {  }
void _nearcall SerSetDirB(int enableXmit)  {  }
void _nearcall SerSetDirC(int enableXmit)  {  }
void _nearcall SerSetDirD(int enableXmit)
{  // PF5 controls RS-485 line driver for port D(3) RS-485 connections
  char  val;
  val = enableXmit ? (PFDRShadow | 0x20) : (PFDRShadow & ~0x20);
  WrPortI(PFDR, &PFDRShadow, val);
}


//-----------------------------------------------------------------------------
void hdwInit(ulong digOutState)
{
  hdwType = HT_SparkDetector;

  __brdInitFlag = TRUE;

// (NI) = Normal Input
// (NO) = Normal Output
// (AI) = Alternate Input
// (AO) = Alternate Output
// (xx) = Used by ZWorld core - do not change

  //--- Port A --- (*must be inputs due to bug in earlier Rabbit 3000's)
  //  PA0 - (NI) *
  //  PA1 - (NI) *
  //  PA2 - (NI) *
  //  PA3 - (NI) *
  //  PA4 - (NI) *
  //  PA5 - (NI) *
  //  PA6 - (NI) *
  //  PA7 - (NI) *
    WrPortI(SPCR, &SPCRShadow, SPCRShadow & ~0x0c);  // PortA: byte input

  //--- Port B ---
  //  PB0 - (AO)  Serial-B (SPI) clock
  //  PB1 - (AO)  Serial-A (Diag/Prog) clock
  //  PB2 - (NO)  /OE input on all Digital output line drivers
  //  PB3 - (NO)  A1 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB4 - (NO)  A2 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB5 - (NO)  OE2A input on 3-to-8 line decoder for Dig I/O Bus
  //  PB6 - (NO)  A0 input on 3-to-8 line decoder for Dig I/O Bus
  //  PB7 - (NO)  G2A on 3-to-8 decoder for Analog output selector
    WrPortI(PBDDR, &PBDDRShadow, PBDDRShadow|0xFC); // dir: OOOO-OOxx

  //--- Port C ---
  //  PC0 - (AO)  Serial-D (RS-232/485) transmit
  //  PC1 - (AI)  Serial-D (RS-232/485) receive
  //  PC2 - (AO)  Serial-C (RS-232) transmit
  //  PC3 - (AI)  Serial-C (RS-232) receive
  //  PC4 - (AO)  Serial-B (SPI) transmit (to all DAC's)
  //  PC5 - (AI)  Serial-B (SPI) receive (from 1st ADC in the chain)
  //  PC6 - (AO)  Serial-A (Diag/Prog) transmit
  //  PC7 - (AI)  Serial-A (Diag/Prog) receive
    WrPortI(PCFR, &PCFRShadow, 0xFF);  // func: AAAA-AAAA

  //--- Port D ---
  //  PD0 - (xx)  Ethernet Controller Reset
  //  PD1 - (xx)  Not Avail (pulled high - not even connected???)
  //  PD2 - (xx)  Ethernet Controller (TPOUT-)
  //  PD3 - (xx)  Ethernet Controller (TPOUT+)
  //  PD4 - (NO)  B input on 3-to-8 decoder for Analog output selector
  //  PD5 - (NO)  C input on 3-to-8 decoder for Analog output selector
  //  PD6 - (xx)  Ethernet Controller (TPIN-)
  //  PD7 - (xx)  Ethernet Controller (TPIN+)
    WrPortI(PDDDR, &PDDDRShadow, PDDDRShadow | 0x30);   // dir:  xxOO-xxxx
    WrPortI(PDFR, &PDFRShadow, PDFRShadow & ~0x30);     // func: xxNN-xxxx

  //--- Port E ---
  //  PE0 - (xx)  Interrupt Input 0A (Ethernet ctlr Interrupt)
  //  PE1 - (AI)  Interrupt Input 1A (buffered from external source)
  //  PE2 - (xx)  Ethernet Controller Chip Select (output)
  //  PE3 - (NO)  A on 3-to-8 decoder for Analog output selector
  //  PE4 - (AI)  Interrupt Input 0B (buffered from external source)
  //  PE5 - (AI)  Interrupt Input 1B (buffered from external source)
  //  PE6 - (NO)  ADR0 on all ADC's (output)
  //  PE7 - (NO)  ADR1 on all ADC's (output)
    WrPortI(PEDDR, &PEDDRShadow, (PEDDRShadow|0xC8)&~0x32); // dir:  OOII-OxIx
    WrPortI(PEFR, &PEFRShadow, (PEFRShadow|0x32)&~0xC8);    // func: NNAA-NxAx

  //--- Port F ---
  //  PF0 - (AO)  Serial-D clock (to RS-485/RS-232 port)
  //  PF1 - (AO)  Serial-C clock (to RS-485/RS-232 port)
  //  PF2 - (NO)  to CS on all ADC's (output)
  //  PF3 - (NO)  to R/C on all ADC's (output)
  //  PF4 - (NO)  xmit enable for Serial C (output)
  //  PF5 - (NO)  xmit enable for Serial D (output)
  //  PF6 - (AO)  PWM1 (buffered ouput to external device)
  //  PF7 - (AO)  PWM0 (buffered ouput to external device)
    WrPortI(PFDDR, &PFDDRShadow, 0xFF);  // dir:  OOOO-OOOO
    WrPortI(PFFR, &PFFRShadow, 0xC3);    // func: AANN-NNAA

  //--- Port G ---
  //  PG0 - (NIO)  Digital I/O bus
  //  PG1 - (NIO)  Digital I/O bus
  //  PG2 - (NIO)  Digital I/O bus
  //  PG3 - (NIO)  Digital I/O bus
  //  PG4 - (NIO)  Digital I/O bus
  //  PG5 - (NIO)  Digital I/O bus
  //  PG6 - (NIO)  Digital I/O bus
  //  PG7 - (NIO)  Digital I/O bus
    WrPortI(PGDDR, &PGDDRShadow, 0x00);  // dir:  IIII-IIII (for now)
    WrPortI(PGFR, &PGFRShadow, 0x00);    // func: NNNN-NNNN


  //--- all parallel port outputs driven hi/lo (no open-drain outputs) ---
    WrPortI(PDDCR, &PDDCRShadow, 0x00);
    WrPortI(PFDCR, &PFDCRShadow, 0x00);
    WrPortI(PGDCR, &PGDCRShadow, 0x00);

  //--- deselect all I/O devices ---
    WrPortI(PBDR, &PBDRShadow, PBDRShadow | 0xA0);  // PB7 and PB5 high
    WrPortI(PFDR, &PFDRShadow, PFDRShadow | 0x04);  // PF2 high

  //--- set state of digital outputs ---
    digitalOut = digOutState;  WriteDigOutputs(digitalOut);

  //--- enable output latches so digital output states take effect ---
    WrPortI(PBDR, &PBDRShadow, PBDRShadow & ~0x04);  // bring PB2 low

//    ser485Rx();  // disable the RS485 transmitter
//    setExt485Rx();  // set external 2/4-wire RS-485 adapter to Rx mode

    spiPort = NULL;  // set later by a call from the kernel process



  flashSize = 524288L;

//  resetADC();

  printf("\r\nInitialized %s hardware\r\n", ArchName);
}


//-----------------------------------------------------------------------------
//  Initialize and register hardware-dependent file systems
//-----------------------------------------------------------------------------
void registerFileSystems(void)
{
  socketInit_();
  srfsInit_();
  serialInit_();

  flashFileInit_();
}

//-----------------------------------------------------------------------------
//  Initialize the network hardware
//-----------------------------------------------------------------------------
void netHdwInit()
{
//  Select ethernet hardware:
//  usingAsix, usingDevKit, usingRealtek, usingSMSC
//  usingAll() can be used to support all cmd core modules, but req's more code

  usingAsix();
}


//-----------------------------------------------------------------------------
//  allocate any resources that should be owned by the kernel process
//
//  To be called ONLY by the kernel process!!!!
//-----------------------------------------------------------------------------
void hdwInit2(void)
{
  fd_t  spifd;


  spifd = _serial_(ADC_DAC_SPIPort, 0, SER_SPI, SER_NO_FLOW);

  spiPort = _FileFromFD_(spifd)->ioDev;

  printf("spiPort: %04lX (%d)\r\n", (ulong)spiPort, spiPort->port);  //tdebug
}

//-----------------------------------------------------------------------------

