//----- bl2105.c ----- 10/25/12 -----(03/26/06)-----

//-----------------------------------------------------------------------------
//  Hardware-specific code to implement standard operations for a ZWorld
//  BL21xx series SBC (Single Board Computer )
//-----------------------------------------------------------------------------
//  Update log (most recent first):
//
//
//  2012-10-25 - Mark Davis (*)
//               - Cleaned up hdwInit() so it works reliably for softReset
//                 mode.  Prior to this it would sometimes do a CPU reset in
//                 the middle of the call.
//
//  2012-09-11 - Mark Davis
//               - Commented out code that initializes the analog outputs
//                 to 0 so soft resets (restarting w/o affecting the state of
//                 the digital and analog output signals) will work.
//
//  2011-02-15 - Mark Davis
//               - Changed inefficient WrPortI() and WrPortE() calls to use
//                 outi() and oute() (with assignment statements for the 2nd
//                 parameter to update the shadow registers).  This creates
//                 much smaller, faster code for these operations.
//
//  2010-02-16 - Mark Davis
//               - Removed call to digOutConfig() (which seems to set the
//                 outputs BACKWARDS from what is requested?!?!) and added
//                 the code to hdwInit() to enable the external high-current
//                 outputs.
//
//  2009-05-27 - Mark Davis
//               - Added code to check for failed return value from anaInDriver
//
//  2009-04-29 - Mark Davis
//               - Added code to support RS-485 interface
//
//  2009-04-22 - Mark Davis
//               - Fixes to get the ADC/DAC read/write values correct again
//
//  2007-03-17 - Mark Davis
//               - reworking the analog input functions to drastically reduce
//                 the number of CPU cycles used
//
//-----------------------------------------------------------------------------
//  Notes on reading analog inputs:
//
//    One I/O cycle for the TLC2543 ADC consists of telling the chip which
//    input to do a conversion on and then reading the data for the channel
//    specified during the previous I/O cycle.
//
//    The DC-style functions treat the process of reading each input as an
//    independent operation.  When reading more than one input, this is very
//    wasteful, as it nearly doubles the number of cycles required.
//
//    When multiple inputs are to be read, a much faster approach is to read
//    them in a sequence of n+1 cycles, where n is the number of inputs to
//    be read.  This works as follows:
//
//    - cycle 1:  specify the 1st input to read and throw away the data
//                read from the chip (it is outdated or garbage)
//
//    - cycle 2:  specify the 2nd input to read and save the data read as the
//                value for the 1st input
//
//    - cycle n:  specify the nth input to read and save the data read as the
//                value for input n-1
//
//    - cycle n+1: specify any input to read and save the data read as the
//                value for input n
//
//    Also, the current code appears to do the serial I/O "manually" - i.e.
//    rather than use the Rabbit's serial capabilities, it toggles digital
//    outputs manually and reads the inputs manually - all of which is
//    probably MUCH slower than using the clocked serial capability of the
//    processor.  Unfortunately, I think the signals they choose to use are
//    NOT ones that can be used by any of the serial ports (at least not in
//    a clocked mode), so we are stuck with this slow approach for this
//    hardware.
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <unistd.h>

#include <sys/backup.h>

#include <sys/spi.h>
#include <sys/hdwDefs.h>

#include <busybox.h>

#include <Rabbit.h>
#include <dcdefs.h>

#include <stcpip.h>

#include <sys/socket.h>
#include <sys/srfs.h>
#include <sys/flashFile.h>


#include <sys/farprintf.h>

//--- see hdwInit() for Digital I/O pin usage ---

const char ArchName[] = { "BL2105" };

#define  NUM_ADC   11
#define  NUM_DAC    4

#define  MAX_DIN   25
#define  MAX_DOUT  16


//--- these values can be set LOWER by the application ---
const uint  NumHdwADC = NUM_ADC;
const uint  NumHdwDAC = NUM_DAC;

const uint  MaxDIN =  MAX_DIN;
const uint  MaxDOUT = MAX_DOUT;


unsigned char  __brdInitFlag;

#include <hdwDefs.c>


//=============================================================================
_nearcall float adcGetRaw(uint ui)  { return (float) adcRaw[ui]; }

_nearcall float adcMinRaw(uint ui)  { return (float)0xfff; }
_nearcall float adcMaxRaw(uint ui)  { return 0.0; }

_nearcall void adcDefaults(uint ui, float *gain, float *offset)
{ *gain = -0.00495;  *offset = 2055.0; }

_nearcall void adcCalibPoints(uint ui, float *low, float *high)
{ *low = -9.5;  *high = 9.5; }

_nearcall char *adcUnits(uint ui, char *buf)  { return "V"; }


//=============================================================================
_nearcall float dacGetRaw(uint ui)  { return (float) dacRaw[ui]; }

_nearcall float dacMinRaw(uint ui)  { return (float)0xfff; }
_nearcall float dacMaxRaw(uint ui)  { return 0.0; }

_nearcall void dacDefaults(uint ui, float *gain, float *offset)
{  *gain = -0.002485;  *offset = 4110.0;  }

_nearcall void dacCalibPoints(uint ui, float *low, float *high)
  {  *low = 0.5;  *high = 9.5; }

_nearcall char *dacUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================


unsigned char PE0Shadow, DOUTLShadow, DOUTHShadow, LEDShadow;

// LCD LED
#define LCDLEDADDR      0xE00B
#define LCDLEDREG           0xE008

// Use Port E bit 7 for LCD strobe
#define LCDSTROBE           0x80
#define LCDCSREGISTER   IB7CR
#define LCDCSSHADOW     IB7CRShadow
#define LCDCSCONFIG     0x78



// Speed control bit, 1=fast mode, 0=slow mode
#define DACSCB          6

// Power control bit, 1=power down, 0=normal operation
#define DACPCB          5

// Write data to dac B and buffer mask
#define WRDACB          0xA0

// Write data to dac A and update dac B with buffer content mask
#define WRDACA          0x80

//#define DAC_CLK_SHADOW  _srPDDR     // DAC CLK line shadow register
#define DAC_CLK_SHADOW  _srPDDR ## !!    // DAC CLK line shadow register
#define DACCLK          3                   // DAC CLK bit

#define DAC_SD_SHADOW   _srPBDR ## !!       // DAC SDO line shadow register
#define DACDATA             7                   // DAC data bit

#define DAC_CS_SHADOW   _PE0Shadow    // Shadow reg for external latch
#define DACCS0          5                   // Chip select 0...channels 0 and 1
#define DACCS1          6                   // Chip select 1...channels 2 and 3

#define SLOW_MODE       0                   // DAC low power mode
#define FAST_MODE       1              // DAC fast access mode

#define POWER_NORMAL    0                   // DAC normal operation mode
#define POWER_DOWN      1                   // DAC power-down mode



#define AD_CLK_SHADOW   _srPDDR ## !!   //ADC CLK line, shadow register
#define AD_CLK              3               //ADC CLK line, Port D bit 3

#define AD_SD_SHADOW    _srPBDR ## !!  //Serial SDI line, shadow register
#define AD_SDI          7               //Serial SDI line, Port B bit 7
#define AD_SDO              0               //Serial SDO line, Port B bit 0

#define AD_CS_SHADOW     _PE0Shadow    //ADC chip-select, shadow register
#define AD_CS               7           //ADC chip-select, external reg bit 4
#define AD_CS_EXTADDR   0x0004      //Addr of the external latch for bit 4

#define AD12BITS        0   // number of output bits for the TLC2543
#define AD8BITS     1  //   "            "

#define MSB_FIRST   0
#define UNIPOLAR    0


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
digOutConfig                <BL21XX.LIB>

SYNTAX:         void digOutConfig(unsigned int outputMode);

DESCRIPTION:    The BL2100 high-current outputs have both sinking and
                sourcing capability for channels 0 - 15. This function
                must be executed at the start of your application to
                initially set channels 0-15 to be OFF for the type of
                circuit that you have, either sinking or sourcing.

                To properly set your high-current outputs you'll need to
                decide for each channel whether the output is sinking or
                sourcing. The function will then assure that each output
                remains OFF when the digital output control interface is
                initialized.

                Note: The brdInit function must be executed before calling
                      this function.

PARAMETER1:    outputMode is a 16 bit parameter for which each bit
               corresponds to one of the high-current outputs.

                    Bit 15 = high-current output channel OUT15
                    Bit 14 = high-current output channel OUT14
                    Bit 13 = high-current output channel OUT13
                    Bit 12 = high-current output channel OUT12
                    Bit 11 = high-current output channel OUT11
                    Bit 10 = high-current output channel OUT10
                    Bit 9  = high-current output channel OUT09
                    Bit 8  = high-current output channel OUT08
                    Bit 7  = high-current output channel OUT07
                    Bit 6  = high-current output channel OUT06
                    Bit 5  = high-current output channel OUT05
                    Bit 4  = high-current output channel OUT04
                    Bit 3  = high-current output channel OUT03
                    Bit 2  = high-current output channel OUT02
                    Bit 1  = high-current output channel OUT01
                    Bit 0  = high-current output channel OUT00

                    To set the outputs set the corresponding bit to one of
                    the following states:

                    0 = Sinking type circuit
                    1 = Sourcing type circuit

               EXAMPLE:
               --------
               outputMode = 0x0ff1; // Outputs OUT15 - 12  = Sinking
                                    // Outputs OUT11 - 08  = Sourcing
                                    // Outputs OUT07 - 04  = Sourcing
                                    // Outputs OUT03 - 01  = Sinking
                                    // Output  OUT00       = Sourcing

RETURN VALUE:   None.

SEE ALSO:       brdInit, digOut

END DESCRIPTION **********************************************************/

void digOutConfig(ulong outputMode)
{
  if (! __brdInitFlag)  return;

//--- initialize the shadow registers for the high-current outputs ---
  oute(0x0001, DOUTLShadow = (char) ~outputMode);
  oute(0x0003, DOUTHShadow = (char) ~(outputMode >> 8) );

//--- enable the external high-current data latch (0-7) ---
  oute(0x0004, PE0Shadow &= ~0x02);

//--- enable the external high-current data latch (8-15) ---
  oute(0x0004, PE0Shadow &= ~0x04);
}


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
Function: int _readExtioPort ( int port );

NOTE:
This function is setup to provide additional hold time when
accessing external I/O addresses 0000 and 0002. This routine
must stay and not be changed, for the additional address
hold-time to be valid.

Improved by Softools to not require a copy to RAM because
real tools have align and other stuff for real programmers.
************************************************************/
#pragma warn -764  // turn off warning about not returning a value

unsigned readExtIO (unsigned port)      // Port is 0 or 2
{
#asm
    push    ip

    ipset   3
    ld      a,l
    and     a,0x03
    jr      z,..read0
    jr      ..skip
    .align  4           ; Align to clear A0 & A1
..read0:
    nop                 ; A0 = 0, A1 = 0
    nop                 ; A0 = 1, A1 = 0
    ioe     ld l,(hl)   ; A0 = 0, A1 = 1
    pop     ip          ; A0 = 0, A1 = 0
    jr      ..exit
    .align  4           ; Align to clear A0 & A1
..skip:
    ioe     ld l,(hl)   ; A0 = 0, A1 = 0 (2 bytes)

    pop     ip          ; A0 = 0, A1 = 1
..exit:
#endasm
}


//-----------------------------------------------------------------------------
//  Determine the device ID based on external hardware settings
//
//  Assumes digital inputs 08-15 are connected to switches with the other
//  side of each switch connected to ground
//-----------------------------------------------------------------------------
ulong ReadDeviceID()
{
  return (ulong) (~(uchar)readExtIO(2));
}

//-----------------------------------------------------------------------------
//  return the current state of the digital inputs
//-----------------------------------------------------------------------------
ulong ReadDigInputs()
{
  ulong inBits;

  inBits  = (ulong) readExtIO(0);
  inBits |= (ulong) readExtIO(2) << 8;
  inBits |= (ulong) RdPortI(PADR) << 16;

  inBits = (inBits & ~forcedDigInMask) | forcedDigInBits;

  return inBits;
}

//-----------------------------------------------------------------------------
//  set the state of the digital outputs
//-----------------------------------------------------------------------------
void WriteDigOutputs(ulong digOut)
{
  digOut = (digOut & ~forcedDigOutMask) | forcedDigOutBits;

  oute(1, DOUTLShadow = (uchar)(digOut) );
  oute(3, DOUTHShadow = (uchar)(digOut >> 8) );
}


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
anaInDriver                 <BL21XX.LIB>

SYNTAX:         uint anaInDriver(uchar cmd, uchar len)

DESCRIPTION:    Reads the voltage of an analog input channel by serial
                    clocking out an 8 bit command to the A/D device of the
                    following formats: This function is non-reentrant


PARAMETER1:     The cmd parameter is as follows:

                    TLC2543 command
                  -------------------
                  D7-D4  D3-D2  D1 D0
                   |      |      |  |
                   |      |      |  +- Mode of operation
                   |      |      |      0 Unipolar (normally used as default)
                   |      |      |      1 Bipolar
                   |      |      |
                   |      |      +---- Output data format
                   |      |             0 MSB first
                   |      |             1 LSB first (not supported by driver)
                   |      |
                   |      +----------- Output data length
                   |                    01 8 bits
                   |                    00 12 bits (normally used as default)
                   |                    11 16 bits (not supported by driver)
                   |
                   +----------------- Channel 0 - 10
                                      Channel 11 = (Vref+  -  Vref-)/2
                                      Channel 12 = Vref-
                                      Channel 13 = Vref+
                                      Channel 14 = Software power-down

PARAMETER2:    Output data length
                  0  = 12 bit mode
                  1 =  8 bit mode

RETURN VALUE:   A value corresponding to the voltage on the analog input
                    channel, which will be:

                    0-4095   for 12 bit A/D conversions
                    0-255    for  8 bit A/D conversions
                    -1       indicates an EOC timeout error

SEE ALSO:       anaIn, anaInVolts, brdInit

END DESCRIPTION **********************************************************/

//------------------------------------------------------------------------
//  This driver has been designed for TI's TLC2543 ADC.
// - TLC2543 using Timing for 12-Clock transfer using CS with MSB first
// - TLC2543 using Timing for  8-Clock transfer using CS with MSB first
//
//  For complete detailed timing information, please refer to the data
//  sheets.
//------------------------------------------------------------------------
//
//    A/D DRIVER INFOR:
//
//      Primary Registers                           Alternate Registers
//    -----------------                         -------------------
//    af = General access                       af' = Port D shadow register
//    b  = Counter/timing                       bc' = Not used
//    c  = External latch shadow register
//    de = PDDR address                         de' = ADC rawcount
//    hl = Used to return ADC rawcount          hl' = ADC CLK shadow register
//
//    iy = External latch address and as a counter
//    ix = PBDR address
//
//    Shadow Register usage:
//    ----------------------
//    AD_CLK_SHADOW  = PDDRShadow  (standard register)
//    AD_SD_SHADOW   = PBDRShadow  (standard register)
//    AD_CS_SHADOW   = PE0Shadow   (created for external latch)
//
//------------------------------------------------------------------------
#pragma offset_labels on
//    uint anaInDriver(uchar cmd, uchar len)
    uint anaInDriver(unsigned char cmd, unsigned char len)
{
#asm
.checkeoc:
    ; Reg B = Desired Delay/(period * instruction cycles)
    ld      b,14               ;timeout period is 14usec

.eocdelay:
    ioi ld  a,(PBDR)           ; get EOC status
    and     0x04               ; mask out unwanted bits
    jr      nz,.adcdriver      ; check if EOC has completed
    djnz    .eocdelay          ; (int)13.4 = 1.425us/(45.211ns * 23 cycles)
                               ; requirement is >=14us so bump value to 14

    ld      hl,-1              ; ADC is not working right, exit with error
    jp      .exit_routine

    ;set to interrupt level 1 to protect the shadow registers
.adcdriver:
    push    ip                 ; save off IP state
    ipset   1                  ; set interrupt priority to level 1

    ;initialize registers and set CLK and CS low
    ld      hl,AD_CLK_SHADOW   ; load alt register w/addr of shadow reg
    ld      a,(hl)             ; get copy of port D shadow register
    ld      de,0               ; initialize AD rawcount to zero
    exx                        ; save-off alternate registers

    ex      af,af'             ; save-off port D shadow register
    ld      a,(AD_SD_SHADOW)   ; get copy of Port B shadow register
    ld      c,a                ; save-off port B shadow register
    ex      af,af'             ; retrieve port D shadow register
    ld      de,PDDR            ; load de with register to control CLK
    res     AD_CLK,a           ; clear CLK bit in register
    ioi ld  (de),a             ; set CLK line low

    ex      af,af'             ; save-off copy of port D shadow register
    ld      iy,AD_CS_EXTADDR   ; load address of external latch
    ld      a,(AD_CS_SHADOW)   ; get copy of external latch shadow register
    res     AD_CS,a            ; clear chip select bit
    ld      (AD_CS_SHADOW),a   ; update external latch shadow register
    ioe ld  (iy),a             ; set CS low

    ; load register with the command to send to the ADC chip
    ld      l,(ix+.cmd)        ; load AD command

    ;
    ; A 1.3us delay is required before reading the A/D data and 1.425us is
    ; required before clocking data into the A/D for both the TLC2543 and
    ; the TLC1543 AD converters. So worst case parameter tsu(CS)1.425us
    ; is being used to cover both requirements.
    ;
    ; The delay itself meets the timing requirement, no other instruction
    ; cycles were included in the calculation so the driver could be easily
    ; maintained.
    ;
    ; Reg B = Desired Delay/(period * instruction cycles)
    ld      b,7               ; delay based on 22.1184MHz CLK
.delay1:
    djnz    .delay1           ; (int)6.3 = 1.425us/(45.211ns * 5 cycles)
                              ; requirement is >=1.425us, bump value to 7
;;  push    ix
    ld      a,(ix+.len)       ; get A/D output length parameter
    cp      0x00              ; check for 12 bit data transfer
    jr      nz,.shift8Bit     ; jump to 8 bit data transfer

.shift12Bit:
    ld      b,4               ; setup for 4 MSB bits to be shifted
    ld      iy,0x08           ; counter, will rotate right 4 counts
    ld      ix,PBDR           ; load addr of Port B

.shiftMSBain:
    exx                       ; save-off primary registers
    ioi ld  a,(ix)            ; read A/D SDO data line, PD0
    and     0x01              ; check if data is one or zero
    or      d                 ; OR'in previous value
    ld      d,a               ; copy results to register D
    rr      iy                ; decrement counter
    jr      c,.bypass1        ; jump if counter has reached zero
    sla     d                 ; make room for the next bit
.bypass1:
    exx                       ; retrieve primary registers
    sla     l                 ; shift to next cmd bit to be written
    jr      nc,.clrbit1ain    ; check if cmd bit is high or low
    ld      a,c               ; get copy of Port B shadow register
    set     AD_SDI,a          ; set data bit high
    ioi ld  (ix),a            ; set DATA high to AD
    ld      c,a               ; update Port B shadow register
    jr      .clkdata1an       ; jump to CLK data code section
.clrbit1ain:
    ld      a,c               ; get copy of port B shadow register
    res     AD_SDI,a          ; set data bit low
    ioi ld  (ix),a            ; set data low to AD
    ld      c,a               ; update port B shadow register
.clkdata1an:
    ex      af,af'            ; retrieve copy of port D shadow register
    set     AD_CLK,a          ; set CLK bit high
    ioi ld  (de),a            ; CLK next SDO bit out of AD
    res     AD_CLK,a          ; set CLK bit low in register
    ioi ld  (de),a            ; CLK SDI data bit into AD
    ex      af,af'            ; save-off copy of port D shadow reg
    djnz    .shiftMSBain      ; loop until B register reaches zero

.shift8Bit:
    ld      b,8               ; setup for 8 LSB bits to be shifted
    ld      iy,0x80           ; counter, will rotate right 8 counts
    ld      ix,PBDR           ; load addr of Port B

.shiftLSBain:
    exx                       ; save-off primary registers
    ioi ld  a,(ix)            ; read A/D SDO data line, PD0
    and     0x01              ; check if data is one or zero
    or      e                 ; OR'in previous value
    ld      e,a               ; copy results to register D
    rr      iy                ; decrement counter
    jr      c,.bypass2        ; jump if counter has reached zero
    sla e                     ; make room for the next bit
.bypass2:
    exx                       ; retrieve primary registers
    sla     l                 ; shift to next cmd bit to be written
    jr      nc,.clrbit2ain    ; check if cmd bit is high or low
    ld      a,c               ; get copy of port B shadow register
    set     AD_SDI,a          ; set DATA bit high in register
    ioi ld  (ix),a            ; set DATA high to AD
    ld      c,a               ; update Port B shadow register
    jr      .clkdata2ain      ; jump to CLK data code section
.clrbit2ain:
    ld      a,c               ; get copy of port B shadow register
    res     AD_SDI,a          ; set data bit low in register
    ioi     ld (ix),a         ; set data low to AD
    ld      c,a               ; update port B shadow register
.clkdata2ain:
    ex      af,af'            ; retrieve copy of port D shadow reg
    set     AD_CLK,a          ; set CLK bit high
    ioi ld  (de),a            ; CLK next SDO bit out of AD
    res     AD_CLK,a          ; set CLK bit low
    ioi ld  (de),a            ; CLK SDI data bit into AD
    ex      af,af'            ; save-off copy of port D shadow reg
    djnz    .shiftLSBain      ; loop until B register reaches zero

    ;update Port B and Port D shadow registers
    ld      (AD_SD_SHADOW),a  ; update port B shadow register
    exx                       ; retrieve alternate registers
    ex      af,af'            ; retrieve copy of port D shadow reg
    ld      (hl),a            ; update port D shadow register

    ;un-assert the ADC chip select and update external latch register
    ld      iy,AD_CS_EXTADDR  ; load register with addr of external latch
    ld      a,(AD_CS_SHADOW)  ; get copy of external latch shadow register
    set     AD_CS,a           ; set AD CS bit high
    ld      (AD_CS_SHADOW),a  ; update external latch shadow register
    ioe ld  (iy),a            ; set AD chip select high

    ;save data for it to be returned to the calling function
    ex      de,hl

    ld      b,0xff
.waiteoc:
    ioi     ld a,(PBDR)       ; get EOC status
    and     0x04              ; mask out unwanted bits
    jr      z,.done           ; wait for EOC to be asserted
    djnz    .waiteoc          ; timeout if EOC never asserts
    ld      hl,-1
.done:
;;  pop ix
    pop ip                    ; restore IP to the previous state


; Due to the DC compiler...can't use embedded ret for the EOC routine
; at the beginning of this function, so jumping to here instead!
.exit_routine:
#endasm
}


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
anaIn                           <BL21XX.LIB>

SYNTAX:         int anaIn(unsigned int channel)

DESCRIPTION:    Reads the state of an analog input channel. This function
                    is non-reentrant.

PARAMETER1:     The analog input channel 0 - 10 to read.

RETURN VALUE:   A value corresponding to the voltage on the analog input
                    channel, which will be:

                    0-4095  for 12 bit A/D conversions

SEE ALSO:       anaInVolts, anaInCalib, anaInfast, brdInit

END DESCRIPTION **********************************************************/

/*
uint anaIn(uint chan)
{
  uchar  len, adcCmd, bitorder, mode;


  if (!__brdInitFlag OR (chan > 13))  return 2048;

//--- setup the A/D command ---
  len = AD12BITS;
  bitorder = MSB_FIRST;
  mode = UNIPOLAR;
  adcCmd = (chan<<4) + (len<<2) + (bitorder<<1) + mode;

//--- tell the ADC which signal to convert and read the result from the ---
//--- conversion (which is useless here) ---
  anaInDriver(adcCmd, len);

//--- now do it again, only this time the "last" conversion was the one ---
//--- done for the above call, so the data will be recent and for the ---
//--- desired signal ---
  return (anaInDriver(adcCmd, len));
}
*/

//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
anaOutDriver                <BL21XX.LIB>

SYNTAX:         void anaOutDriver(int power_control, int speed_control,
                                 int channel, unsigned int rawcount)

DESCRIPTION:   Sets the voltage of an analog output channel by serial
               clocking in 16-bits to a DAC device of the following
               format: (This function is non-reentrant)

                    D15 D14 D13 D12 D11-D0
                     |   |   |   |   |  |
                     |   |   |   |   +--+---- data bits, msb-lsb (0-4095)
                     |   |   |   |
                     |   |   +---|----------- power control
                     |   |       |              0 = normal (default)
                     |   |       |              1 = power down
                     |   |       |
                     |   +-------|----------- speed control,
                     |           |              0 = slow
                     |           |              1 = fast (default)
                     |           |
                     |           +----------- register R0
                     +----------------------- register R1
                                             R1 R0
                                              0  0  write data to dacB
                                              0  1  write data to buffer
                                              1  0  write data to dacA
                                              1  1  reserved

PARAMETER1:    DAC power mode option (0 or 1)
                    0 = Normal operation
                    1 = Power-down mode

                    When the power-down mode is selected, the only other
                    parameter that is used is the DAC channel (PARAMETER3).
                    All other parameters are don't cares.

                    Also when putting a DAC output in power-down or normal
                    mode, two DACchannels are affected, for example:

                    POWER-DOWN MODE:
                    ----------------
                    a) When power_control equals 1 and PARAMETER3 is channel
                       0 or 1, then both DAC channels 0 and 1 are put in power-
                       down mode. (channels 2 and 3 not affected)

                    b) When power_control equals 1 and PARAMETER3 is channel
                       2 or 3, then both DAC channels 2 and 3 are put in power-
                       down mode. (channels 0 and 1 not affected)

                    NORMAL OPERATION MODE:
                    ----------------------
                    a) When power_control equals 0 and PARAMETER3 is channel
                       0 or 1, then both DAC channels 0 and 1 are put in normal
                       operational mode. (channels 2 and 3 not affected)

                    b) When power_control equals 0 and PARAMETER3 is channel
                       2 or 3, then both DAC channels 2 and 3 are put in normal
                       operational mode. (channels 0 and 1 not affected)

PARAMETER2:    DAC speed control (0 or 1)

                    Mode of operation  Speed versus power dissipation
                    ----------------   ------------------------------
                    0 = Slow mode        12us access vs. 1 mA (see Note)
                    1 = Fast mode         3us access vs. 2.3 mA (see Note)

               Note: Test conditions from TI's data sheet for the speed
                     control option:
                     - No load.
                     - All inputs are at GND or VDD.
                     - DAC latch = 0x800.

PARAMETER3:     The DAC output channel to write (0-3).
PARAMETER4:     A data value corresponding to the desired voltage on the
               analog output channel (0-4095).


RETURN VALUE:   None

SEE ALSO:       anaOut, anaOutVolts, anaOutCalib

END DESCRIPTION **********************************************************/

#pragma offset_labels on
           void anaOutDriver(int power_control, int speed_control,
                              int channel, unsigned int rawcount)
{
//------------------------------------------------------------------------
//  This driver has been designed for TI's TLV5618A DAC chip.
//
//    Note: Upon power-up the DAC A and DAC B are in normal operational
//          mode and the outputs will be at approx zero volts.
//
//------------------------------------------------------------------------
//
//    DAC DRIVER INFOR:
//
//    Primary Registers                    Alternate Registers
//    -----------------                    -------------------
//    af = General access                  af' = Not used
//    b  = Used as a counter               b'  = Not used
//    c  = DAC CLK shadow register         c'  = External latch shadow reg
//    de = DAC CLK Port D address          de' = External latch address
//    hl = DAC command word                hl' = Not used
//
//    iy = DAC SD port B address
//    ix = Index for function parameters
//
//    Shadow Register usage:
//    ----------------------
//    DAC_CLK_SHADOW  = PDDRShadow  (standard register)
//    DAC_SD_SHADOW   = PBDRShadow  (standard register)
//    DAC_CS_SHADOW   = PE0Shadow   (created for external latch)
//
//------------------------------------------------------------------------
#asm
    ;set to interrupt level 1 to protect the shadow registers
    push  ip                    ; save off IP state
    ipset   1                   ; set interrupt priority to level 1
    ld    de,0x0004             ; load de with external latch address
    ld    a,(DAC_CS_SHADOW)     ; get copy of external latch shadow reg
    ld    c,a
    exx                         ; save-off alternate registers

    ld    a,(DAC_CLK_SHADOW)    ; get copy of port D shadow register
    ld    c,a
    ld    de,PDDR               ; load de w/address of Port D
    ld    iy,PBDR               ; load iy w/address of port B

    ;determine the DAC access mode fast vs. slow
    ld      hl,(ix+.rawcount)   ; setup data format for shift

.speedcntrl:
    ld    a,(ix+.speed_control) ;get speed_control value
    cp    0                     ; check for slow mode operation
    jr    z,.slow               ; Jump for slow mode of operation
    set DACSCB,h                ; set to be in fast speed mode
    jr    .powercntrl           ; mode of operation is fast mode
.slow:
    res   DACSCB,h              ; mode of operation is slow mode

    ;determine the DAC operating mode
.powercntrl:
    ld    a,(ix+.power_control) ; get power_control value
    cp    0                     ; check for normal mode of operation
    jr    nz,.powerdown         ; jump to set DAC in power-down mode
    res DACPCB,h                ; set DAC for normal operation
    jr    .dacselect            ; go determine which DAC to access
.powerdown:
    set   DACPCB,h              ; put DAC chip into power-down mode

    ; determine which channel to access on a given DAC chip, DAC A or
    ; DAC B (this doesn't determine which DAC chip to access, thats
    ; done in the chip select section)
.dacselect:
    ld      a,(ix+.channel)     ; get DAC channel to be accessed
    and 0x01                    ; mask out all bits except D0
    cp    0                     ; determine if DACA or DACB is to be accessed
    jr      z,.wrtoa            ; check on which DAC is going to be accessed
    ld    a,h                   ; get the DAC command
    and     0x5f                ; set the command to access DACB
    jp      .clockhigh          ; go set the CLK line high
.wrtoa:
    ld  a,h                     ; get a copy of MSB of the DAC command
    or      WRDACA              ; set the command to access DACA

    ;preset CLK high before asserting the chip select line
.clockhigh:
    ld  h,a                     ; put DAC command back into h register
    ld    a,c                   ; get copy of port D shadow register
    set DACCLK,a                ; clock low
    ioi ld (de),a               ; latch
    ld    c,a                   ; update port D shadow register

    ;determine which DAC chip to access and assert its chip select line
.chipselect:
;   ld      h,a                 ; update command with channel to be accessed
    exx                         ; save-off primary registers
    ld      a,(ix+.channel)     ; determine which chip select
    srl a                       ; shift channel to see the LSBit is 0 or 1
    cp      0                   ; select dac 0
    jr      nz,.cs1             ; go setup command for DAC 1 access
    res DACCS0,c                ; setup command for a DAC 0 access
    jr      .assertcs           ; go assert the chip select line
.cs1:
    res DACCS1,c                ; setup command for a DAC 1 access
.assertcs:
    ld    a,c
    ioe   ld (de),a             ; assert cs for the DAC selected

.shiftdataMSB:
    exx                         ; retrieve primary registers
    ld      b,8                 ; MSB of command word
    ld    a,(DAC_SD_SHADOW)     ; get copy of Port B shadow register

    ;shift the upper 8 bits (MSB) of the command out to the DAC
.shiftMSB:
    sla h
    jr      nc,.clrbit1         ; check if command bit is high or low
    set DACDATA,a               ; set command data bit high
    jp      .wrdata1            ; go CLK data into the DAC
.clrbit1:
    res DACDATA,a               ; set command data bit low
.wrdata1:
    ioi ld (iy),a               ; output command data bit

    ;data is preset now CLK the data into the DAC
.clkdata1:
    ex      af,af'
    ld    a,c
    res DACCLK,a                ; set the CLK bit low
    ioi ld (de),a               ; CLK data into the DAC
    set DACCLK,a                ; set the CLK bit high
    ioi ld (de),a               ; set CLK back to a high level
    ld    c,a
    ex      af,af'
    djnz    .shiftMSB           ; jump to send next command bit

    ;shift the lower 8 bits (LSB) of the command out to the DAC
.shiftdataLSB:
    ld      b,8                 ; LSB of command word
.shiftLSB:
    sla l                       ; shift MSB of command
    jr      nc,.clrbit2         ; check if command bit is high or low
    set DACDATA,a               ; set command data bit high
    jp      .wrdata2            ; jump to output data code section
.clrbit2:
    res DACDATA,a               ; set command data bit low

.wrdata2:
    ioi ld (iy),a               ; output command data bit

    ;data is preset now CLK the data into the DAC
.clkdata2:
    ex      af,af'
    ld     c,a
    res DACCLK,a                ; set the CLK bit low
    ioi ld (de),a               ; CLK data into the DAC
    set DACCLK,a                ; set the CLK bit high
    ioi ld (de),a               ; return CLK back to a high level
    ld    c,a
    ex      af,af'
    djnz    .shiftLSB           ; jump to send next command bit

    ;access is completed un-assert the chip select and update shadow reg
.cshigh:
    ex      af,af'
    ld    (DAC_SD_SHADOW),a     ; update port B shadow register
    ld    a,c                   ; get copy of port D shadow register
    ld    (DAC_CLK_SHADOW),a    ; update port D shadow register

    exx                         ; retrieve alternate registers
    ld      a,c                 ; get copy of external latch shadow reg
    set   DACCS0,a              ; set DAC chip selects high
    set     DACCS1,a
    ioe ld (de),a               ; unassert the chip-select lines
    ld    (DAC_CS_SHADOW),a     ; update external latch shadow register
    pop ip                      ; restore IP to the previous state
#endasm
}



//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
anaOut                      <BL21XX.LIB>

SYNTAX:         int anaOut(unsigned int channel, unsigned int rawcount)

DESCRIPTION:    Sets the voltage on a given analog output channel. This
               function is non-reentrant.

PARAMETER1:     The DAC output channel to write (0 - 3).
PARAMETER2:     A data value corresponding to the desired voltage on the
               analog output channel (0-4095).

RETURN VALUE:   0 if successful
                    -1 if rawcount parameter is greater than 4095

SEE ALSO:       anaOutDriver, anaOutVolts, anaOutCalib

END DESCRIPTION **********************************************************/

int anaOut_(unsigned int channel, unsigned int rawcount)
{
    auto unsigned int power_control, speed_control;

    if (!__brdInitFlag OR (channel > 3))  return -1;

    if (rawcount > 4095)  return -1;

    power_control = POWER_NORMAL;
    speed_control = FAST_MODE;
    anaOutDriver(power_control, speed_control, channel, rawcount);
}

#pragma warn 764  // turn back on warning about not returning a value


//-----------------------------------------------------------------------------
//  Read 1 or more analog inputs.  Each input is represented by a bit in the
//  mask parameter.  For each non-zero bit, the corresponding ADC signal is
//  read and the resulting raw value stored in the adcRaw[] array.
//-----------------------------------------------------------------------------
void readAnalogInputs(ulong mask)
{
  uchar  len, adcCmd, bitorder, mode;
  uint  i, prevIdx, rawVal, bitTest;


  if (!__brdInitFlag OR !mask)  return;

//--- setup the A/D command ---
  len = AD12BITS;
  bitorder = MSB_FIRST;
  mode = UNIPOLAR;
  adcCmd = (len<<2) + (bitorder<<1) + mode;

//--- Each I/O cycle on the ADC consists of telling it which signal to read ---
//--- and convert (starting at the end of the cycle) and reading the result ---
//--- of the previous conversion.  Therefore, we have to do n+1 I/O cycles ---
//---- to read n signals ---
  prevIdx = 255;
  for (i=0,bitTest=1; i<NUM_ADC; ++i,bitTest<<=1)  {
    if (!(bitTest & (uint)mask))
      adcRaw[i] = 0;
    else  {
      adcCmd = (i << 4) | (adcCmd & 0x0f);
      rawVal = anaInDriver(adcCmd, len);
      if ((prevIdx < NUM_ADC) AND !(rawVal & 0xF000)) adcRaw[prevIdx] = rawVal;
      prevIdx = i;
    }
  }
  if (prevIdx < NUM_ADC) {
    rawVal = anaInDriver(adcCmd, len);
    if (!(rawVal & 0xF000))  adcRaw[prevIdx] = rawVal;
  }

}

//-----------------------------------------------------------------------------
void writeAnalogOutputs(ulong mask)
{
  uint  i, bitTest;

  for (i=0,bitTest=1; i<NumHdwDAC; ++i,bitTest<<=1)
    if (bitTest & (uint)mask)  anaOut_(i, dacRaw[i]);
}


//-----------------------------------------------------------------------------
/*
float ADCtoVolts(uchar chan, uint rawADC)
{
  return (ADCOffset(chan) - (float)(rawADC)) * ADCGain(chan);

//  return  (float) (((long)(adcOffset_[chan]-(int)(rawADC)) * adcGain_[chan]));
//                   / 4194304.0;
}
*/

//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to transmit mode
//-----------------------------------------------------------------------------
void setExt485Tx(void)
{
  ulong  digOut;

  _EnterCritical_();

//--- set external 2/4-wire adapter to Transmit mode (uses DO-07) ---
  forcedDigOutMask |= 0x000000C0L;
  forcedDigOutBits = forcedDigOutBits & 0xFFFF3FL | 0x00000080L;

//  WriteDigOutputs(digitalOut);

  digOut = (digitalOut & ~forcedDigOutMask) | forcedDigOutBits;

  oute(1, DOUTLShadow = (uchar)(digOut) );
//  oute(3, DOUTHShadow = (uchar)((uint)digOut >> 8) );

  _ExitCritical_();
}


//-----------------------------------------------------------------------------
//  Functions to enable/disable the transmitter for an RS-485 connection.
//
//  WARNING:
//    These functions are called from an interrupt handler, so they must be
//    short and sweet!
//-----------------------------------------------------------------------------

#pragma offset_labels on

#define  EXT_485_RX      0x000040  // 06: enable Tx on extern 2/4-wire adapter
#define  EXT_485_TX      0x000080  // 07: enable Rx on extern 2/4-wire adapter

#define  EXT_485_RX_TX      0x0000C0  // both Tx and Rx bits
#define  NOT_EXT_485_RX_TX  0x00003F  // all BUT Tx & Rx bits (for 8-bit value)

extern ulong digitalOut;


//------------------------------------------
void _nearcall SerSetDirA(int enableXmit)  {  }

//------------------------------------------
void _nearcall SerSetDirB(int enableXmit)  {  }

//------------------------------------------
void _nearcall SerSetDirC(int enableXmit)  {  }

//-------------------------------------------
//  uses Digital Outputs 6 and 7 to ctl
//  xmit and rcv modes on extern 2/4 adapter
//  WARNING:
//    Usually called by a _fastint function so
//    save all registers used!!!
//-------------------------------------------
void _nearcall SerSetDirD(int enableXmit)
{
  ulong  digOut;

  INT_SAVE_ALL();

  _EnterCritical_();

  if (enableXmit)  {
  //--- enable the driver in the on-board SP483EN transceiver ---
    oute(0x0004, PE0Shadow |= 0x01);
//    ioSetBitE(0x0004, 1);

  //--- set external 2/4-wire adapter to Transmit mode (uses DO-07) ---
/* for some odd timing reason I don't yet understand, this actually takes
   effect LATER (?!?!?!?) so for now it is in setExt485Tx() which must be
   called before actually writing new data to the serial port
    forcedDigOutMask |= 0x000000C0L;
    forcedDigOutBits = forcedDigOutBits & 0xFFFF3FL | 0x00000080L;

    digOut = (digitalOut & ~forcedDigOutMask) | forcedDigOutBits;

    oute(1, DOUTLShadow = (uchar)(digOut) );
*/
  }
  else  {  // called by serial lib's isrSendCommon() ISR
  //--- disable the driver in the on-board SP483EN transceiver ---
    oute(0x0004, PE0Shadow &= 0xfe);
//    ioClearBitE(0x0004, 1);

  //--- set external 2/4-wire adapter to Receive mode (uses DO-06) ---
    forcedDigOutMask |= 0x000000C0L;
    forcedDigOutBits = forcedDigOutBits & 0xFFFF3FL | 0x00000040L;

    digOut = (digitalOut & ~forcedDigOutMask) | forcedDigOutBits;

    oute(1, DOUTLShadow = (uchar)(digOut) );
  }

  _ExitCritical_();

  INT_RESTORE_ALL();

/*
  if (enableXmit)  {  ;
#asm
    push  ip                ; Save off IP state
    ipset  2                ; Set interrupt priority to level 2

;;--- enable the driver in the on-board SP843EN transceiver ---
    ld    a,(_PE0Shadow)    ; Get copy of shadow register
    or    01h               ; Set shadow register bit 0
    ld    (_PE0Shadow),a    ; Update shadow register
    ld    bc, 0x0004        ; Load BC w/ext addr of the control latch
    ioe ld  (bc),a          ; Set bit 0 on ext latch via addr 0x0004

    pop   ip                ; Restore IP to the previous state
#endasm
  }
  else  {  ;
#asm
    push  ip                ; Save off IP state
    ipset  2                ; Set interrupt priority to level 2
#endasm

  //--- set external 2/4-wire adapter to Transmit mode ---
    forcedDigOutMask |= 0x000000C0L;
    forcedDigOutBits = forcedDigOutBits & 0xFFFF3FL | 0x00000040L;

    digOut = (digitalOut & ~forcedDigOutMask) | forcedDigOutBits;

    oute(1, DOUTLShadow = (uchar)(digOut) );
    oute(3, DOUTHShadow = (uchar)((uint)digOut >> 8) );
    ;
#asm
;;--- disable the driver in the on-board SP483EN transceiver ---
    ld    a,(_PE0Shadow)    ; Get copy of shadow reg
    and ~0x01               ; Clear shadow register bit 0
    ld    (_PE0Shadow),a    ; Update shadow reg
    ld    bc,0x0004         ; Load BC w/addr of external control latch
    ioe ld  (bc),a          ; Set bit 0 low on ext latch at addr 0x0004

    pop   ip                ; Restore IP to the previous state
#endasm
  }
*/
}

//-----------------------------------------------------------------------------
/*
SYNTAX:         void hdwInit(unsigned long digOutState);

DESCRIPTION:    Call this function at the beginning of application code.
                    1.Initializes the system I/O ports.
                    2.Loads all the A/D and DAC calibration constants from
                      flash and puts them into memory, ready for application
                      program use.
                    3.Will turn-off LED DS1 to indicate that the controller
                      was successfully initialized. This only applies if the
                      LCD/KEYPAD module is installed.

                   Note: This function is non-reentrant.

               *** Rabbit Ports A, B, C, D, and E Initialization ***

               Port  IO   Function               I/O State
               ----  ---  ------------------     ------------
               PA.0  In   Digital input,IN16     Pulled-up
               PA.1  In   Digital input,IN17     Pulled-up
               PA.2  In   Digital input,IN18     Pulled-up
               PA.3  In   Digital input,IN19     Pulled-up
               PA.4  In   Digital input,IN20     Pulled-up
               PA.5  In   Digital input,IN21     Pulled-up
               PA.6  In   Digital input,IN22     Pulled-up
               PA.7  In   Digital input,IN23     Pulled-up

               ----  ---  ------------------     ------------
               PB.0  In   DAC_ADC_SDO            Pulled-up
               PB.1  In   Not used               Pulled-up
               PB.2  In   ADC EOC line           Driven by ADC
               PB.3  In   Not used               Pulled-up
               PB.4  In   I/O config option      1 = standard  (JP2)
               "     In   "                      0 = custom
               PB.5  In   Analog circuit option  1 = installed (JP3)
               "     In   "                      0 = not installed
               PB.6  Out  Not used               Low
               "     Out  LED control (RCM2300)  Low...led OFF
               PB.7  Out  DAC_ADC_SDI            high

               ----  ---  ------------------     ------------
               PC.0  Out  TXD (RS485)            Inactive high
               PC.1  In   RXD (RS485)            Inactive high
               PC.2  Out  TXC (RS232)            Inactive high
               PC.3  In   RXC (RS232)            Inactive high
               PC.4  Out  Realtek reset          Initial'ed by sock_init()
               PC.5  In   Realtek INT0           Pulled-up
               PC.6  Out  TXA                    Inactive high
               PC.7  In   RXA                    Inactive high

               ----  ---  ------------------     ------------
               PD.0  In   Realtek CLK (RCM2200)  Initial'ed by sock_init()
               PD.1  In   Realtek SDO (RCM2200)  Initial'ed by sock_init()
               PD.2  Out  Not used               High
               PD.3  Out  DAC CLK line           High
               PD.4  Out  ATXB (RS232)           Inactive high
               PD.5  In   ARXB (RS232)           Inactive high
               PD.6  Out  Not used               High
               PD.7  Out  Not used               High

               ----  ---  ------------------     ------------
               PE.0  Out  Digital I/O strobe     High
               PE.1  Out  External I/O enable    High
               PE.2  N/A  Realtek IORB strobe    Initial'ed by sock_init()
               PE.3  N/A  Realtek SDI line       Initial'ed by sock_init()
               PE.4  In   INT0B                  Tied to PE5 via 1k resistor
               PE.5  In   INT1B                  User interrupt input **
               PE.6  N/A  Realtek IOWB strobe    Initial'ed by sock_init()
               PE.7  Out  LCD_KEYPAD strobe      High

               ** PE5 is driven by PE4, if the interrupt is not being used.

PARAMETER:      None

RETURN VALUE:   None

END DESCRIPTION **********************************************************/

//-----------------------------------------------------------------------------
void hdwInit(ulong digOutState)
{
  hdwType = HT_BL2105;

  __brdInitFlag = TRUE;

// Initialize variables using the global initialization chain
  PE0Shadow = 0xEE;
  DOUTLShadow = 0x00;
  DOUTHShadow = 0x00;
  LEDShadow = 0x00;

// (NI) = Normal Input
// (NO) = Normal Output
// (AI) = Alternate Input
// (AO) = Alternate Output
// (xx) = Used by ZWorld core - do not change

  //--- Port A ---
  //  PA7 - (NI)  DI-23                  (pulled up)
  //  PA6 - (NI)  DI-22                  (pulled up)
  //  PA5 - (NI)  DI-21                  (pulled up)
  //  PA4 - (NI)  DI-20                  (pulled up)
  //  PA3 - (NI)  DI-19                  (pulled up)
  //  PA2 - (NI)  DI-18                  (pulled up)
  //  PA1 - (NI)  DI-17                  (pulled up)
  //  PA0 - (NI)  DI-16                  (pulled up)
    outi(SPCR, srSPCR = 0x80);      // PortA: byte input

  //--- Port B ---
  //  PB7 - (-O)  DAC_ADC_SDI            (high)
  //  PB6 - (-O)  ---not used---         (low)
  //  PB5 - (-I)  Analog circuit option  (1 = installed) (JP3)
  //  PB4 - (-I)  I/O config option      (1 = standard)  (JP2)
  //  PB3 - (-I)  ---not used---         (pulled up)
  //  PB2 - (-I)  ADC EOC line           (driven by ADC)
  //  PB1 - (-I)  ---not used---         (pulled up)
  //  PB0 - (-I)  DAC_ADC_SDO            (pulled up)
    outi(PBDR, srPBDR |= 0x80);    // SDI line high

  //--- Port C ---
  //  PC7 - (-I)  Serial-A receive        (Diag/Prog/Console port)
  //  PC6 - (AO)  Serial-A transmit       (Diag/Prog/Console port)
  //  PC5 - (-I)  Realtek INT0            (pulled up)
  //  PC4 - (NO)  Realtek Ethernet RESET  (initialized by sock_init())
  //  PC3 - (-I)  Serial-C receive        (RS-232)
  //  PC2 - (AO)  Serial-C transmit       (RS-232)
  //  PC1 - (-I)  Serial-D receive        (RS-485)
  //  PC0 - (AO)  Serial-D transmit       (RS-485)
    outi(PCFR, srPCFR = 0x45);  // func: xAxN-xAxA

  //--- Port D ---
  //  PD7 - (NO)  --- not used ---        (high)
  //  PD6 - (NO)  --- not used ---        (high)
  //  PD5 - (NI)  Serial-B receive        (RS-232)
  //  PD4 - (AO)  Serial-B transmit       (RS-232)
  //  PD3 - (NO)  DAC CLK line            (high)
  //  PD2 - (NO)  --- not used ---        (high)
  //  PD1 - (NI)  Realtek SDO             (initialized by sock_init())
  //  PD0 - (NI)  Realtek CLK             (initialized by sock_init())
    outi(PDCR,  srPDCR = 0x00);      // outputs CLK'ed by PCLK/2
    outi(PDDDR, srPDDDR = 0xDC);     // dir:  OOIO-OOII
    outi(PDFR,  srPDFR = 0x10);      // func: NNNA-NNNN
    outi(PDDR,  srPDDR |= 0xDC);     // all outputs will start high

  //--- Port E ---
  //  PE7 - (AO)  LCD_KEYPAD strobe      (high)
  //  PE6 - (xx)  Realtek IOWB strobe    (initialized by sock_init())
  //  PE5 - (NI)  INT1B (Ext. interrupt) (input)
  //  PE4 - (NO)  --- not used ---       (tied to PE5 via 1K resistor)
  //  PE3 - (xx)  Realtek SDI line       (initialized by sock_init())
  //  PE2 - (xx)  Realtek IORB strobe    (initialized by sock_init())
  //  PE1 - (NO)  External I/O enable    (high)
  //  PE0 - (AO)  Digital I/O strobe     (high)
  // ** PE5 is driven by PE4, if the interrupt is not being used.

    outi(PEDDR, srPEDDR = (srPEDDR & 0x4C) | 0x93);   // dir:  OxIO-xxOO
    outi(PEFR,  srPEFR  = (srPEFR  & 0x4C) | 0x81);   // func: NxNN-xxNA
    outi(IB0CR, srIB0CR = 0xF8);    // PE0: 1WS, I/O Strobe, writes enabled
    outi(PECR,  srPECR  = 0x00);    // outputs CLK'ed by PCLK/2


//--- all parallel port outputs driven hi/lo (no open-drain outputs) ---
  outi(PDDCR, srPDDCR = 0x00);
  outi(PFDCR, srPFDCR = 0x00);
  outi(PGDCR, srPGDCR = 0x00);



  //---------------------------------------------------------------------
  // Initialize external latch
  //---------------------------------------------------------------------
  // set RS485 TX control low all others high
  oute(0x0004, PE0Shadow &= ~0x01);

  // enable external latch
  outi(PEDR, srPEDR &= ~0x02);

  // Preset all analog channel low before enabling the analog power supply
  if (ini(PBDR) & 0x20) {
//    anaOutDriver(POWER_NORMAL, FAST_MODE, 0, 4095);
//    anaOutDriver(POWER_NORMAL, FAST_MODE, 1, 4095);
//    anaOutDriver(POWER_NORMAL, FAST_MODE, 2, 4095);
//    anaOutDriver(POWER_NORMAL, FAST_MODE, 3, 4095);

    // Set RS485 TX off, and turn-on DAC power supply
    oute(0x0004, PE0Shadow &= ~0x11);
  }


  //---------------------------------------------------------------------
  // Initialize PE7 to turn OFF the bad CPU indicator
  //---------------------------------------------------------------------
  // Initialize Port E bit to be a chip select.
  // Set to be a I/O Read and Write strobe with 7 wait states.
  // and writes allowed.

  outi(LCDCSREGISTER, LCDCSSHADOW = LCDCSCONFIG);

  // Turn-off the CPU bad indicator
  oute(LCDLEDADDR, LEDShadow = 0x00);

  // Enable LED control register
  oute(LCDLEDREG, 0x00);

//--- enable the external high-current data latch (0-7) ---
  oute(0x0004, PE0Shadow &= ~0x02);

//--- enable the external high-current data latch (8-15) ---
  oute(0x0004, PE0Shadow &= ~0x04);

  WriteDigOutputs(digOutState);


  flashSize = 524288L;

  printf("\r\nInitialized %s hardware\r\n", ArchName);
}


//-----------------------------------------------------------------------------
//  Initialize and register hardware-dependent file systems
//-----------------------------------------------------------------------------
void registerFileSystems(void)
{
  socketInit_();
  srfsInit_();
  serialInit_();

  flashFileInit_();
}

//-----------------------------------------------------------------------------
//  Initialize the network hardware
//-----------------------------------------------------------------------------
void netHdwInit()
{
//  Select ethernet hardware:
//  usingAsix, usingDevKit, usingRealtek, usingSMSC
//  usingAll() can be used to support all cmd core modules, but req's more code

  usingRealtek();
}


//-----------------------------------------------------------------------------
//  allocate any resources that should be owned by the kernel process
//-----------------------------------------------------------------------------
void hdwInit2(void)
{

}


//-----------------------------------------------------------------------------

