//----- rfMonitor.c ----- 02/07/10 -----(03/26/06)-----

//-----------------------------------------------------------------------------
//  Code to handle hardware-specific implementation of generic hdw functions
//-----------------------------------------------------------------------------
//
//  2010-02-07 - Mark Davis -
//               - Fixes to void convertADS7825() to avoid race conditions
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <sys/backup.h>

#include <sys/spi.h>
#include <sys/hdwDefs.h>

#include <busybox.h>

#include <Rabbit.h>
#include <dcdefs.h>

#include <stcpip.h>

#include <sys/socket.h>
#include <sys/srfs.h>
#include <sys/flashFile.h>

#include <sys/farprintf.h>

//--- see hdwInit() for Digital I/O pin usage ---

const char ArchName[] = { "RFMonitor" };

#define  NUM_ADC    8
#define  NUM_DAC    2

#define  MAX_DIN   12
#define  MAX_DOUT  11


//--- these values can be set LOWER by the application ---
const uint  NumHdwADC = NUM_ADC;
const uint  NumHdwDAC = NUM_DAC;

const uint  MaxDIN =  MAX_DIN;
const uint  MaxDOUT = MAX_DOUT;


//--- rfMonitor 16-bit analog RCM3200-based ctlrs use port B for SPI to ADCs, DACs
#define  ADC_DAC_SPIPort  1  // for most RCM3200-based NSCL ctlr's w/Analog


unsigned char  __brdInitFlag;

#include <hdwDefs.c>


SerialInfo  *spiPort;


//=============================================================================
_nearcall float adcGetRaw(uint ui)  { return (float) adcRaw[ui]; }

_nearcall float adcMinRaw(uint ui)  { return 0.0; }
_nearcall float adcMaxRaw(uint ui)  { return (float)0xffff; }

_nearcall void adcDefaults(uint ui, float *gain, float *offset)
{ *gain = 0.0003046;  *offset = 32786.0; }

_nearcall void adcCalibPoints(uint ui, float *low, float *high)
{ *low = -9.5;  *high = 9.5; }

_nearcall char *adcUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================
_nearcall float dacGetRaw(uint ui)  { return (float) dacRaw[ui]; }

_nearcall float dacMinRaw(uint ui)  { return 0.0; }
_nearcall float dacMaxRaw(uint ui)  { return (float)0xffffL; }

_nearcall void dacDefaults(uint ui, float *gain, float *offset)
{  *gain = 0.000311;  *offset = 32768.0;  }

_nearcall void dacCalibPoints(uint ui, float *low, float *high)
  {  *low = -9.5;  *high = 9.5; }

_nearcall char *dacUnits(uint ui, char *buf)  { return "V"; }

//=============================================================================


//-----------------------------------------------------------------------------

/* START FUNCTION DESCRIPTION ********************************************
Function: int _readExtioPort ( int port );

Note:
This function is setup to provide additional hold time when
accessing external I/O addresses 0000 and 0002. This routine
must stay         and not be changed, for the additional
address hold-time to be valid.

Improved by Softools to not require a copy to RAM because
real tools have align and other stuff for real programmers.
************************************************************/
/*
unsigned readExtIO (unsigned port)      // Port is 0 or 2
{
#asm
    push    ip
    ipset   3
    ld      a,l
    and     a,0x03
    jr      z,..read0
    jr      ..skip
    .align  4           ; Align to clear A0 & A1
..read0:
    nop                 ; A0 = 0, A1 = 0
    nop                 ; A0 = 1, A1 = 0
    ioe     ld l,(hl)   ; A0 = 0, A1 = 1
    pop     ip          ; A0 = 0, A1 = 0
    jr      ..exit
    .align  4           ; Align to clear A0 & A1
..skip:
    ioe     ld l,(hl)   ; A0 = 0, A1 = 0 (2 bytes)
    pop     ip          ; A0 = 0, A1 = 1
..exit:
#endasm
}
*/

//-----------------------------------------------------------------------------
//  determine the device ID based on external hardware settings
//-----------------------------------------------------------------------------
ulong ReadDeviceID()
{
  uchar  inA, devID;


  inA = RdPortI(PADR);

//--- device ID comes from 8-bit switch input ----
  devID = ((inA & 0x02) << 6)
        | ((inA & 0x01) << 6)
        | ((inA & 0x08) << 2)
        | ((inA & 0x04) << 2)
        | ((inA & 0x20) >> 2)
        | ((inA & 0x10) >> 2)
        | ((inA & 0x80) >> 6)
        | ((inA & 0x40) >> 6);
  devID = ~devID;

  return (ulong) devID;
}

//-----------------------------------------------------------------------------
//  return the current state of the digital inputs
//-----------------------------------------------------------------------------
ulong ReadDigInputs()
{
  uchar  inB, inC, inD, inF;
  ulong  inBits;


//--- read state of each of the digital inputs we need bits from ---
  inB = RdPortI(PBDR);
  inC = RdPortI(PCDR);
  inD = RdPortI(PDDR);
  inF = RdPortI(PFDR);

//--- PortF: xxxx-xx##  ==>  inBits: xxxx-xxxx xxxx-xxxx xxxx-xx## ---
  inBits  = (uint)inF & 0x03;

//--- PortB: xxx#-##xx  ==>  inBits: xxxx-xxxx xxxx-xxxx xxx#-##xx
  inBits |= (uint)inB & 0x1c;

//--- PortC: xxxx-#xxx  ==>  inBits: xxxx-xxxx xxxx-xxxx xx#x-xxxx
  inBits |= ((uint)inC & 0x08) << 2;

//--- PortD: xx##-xxxx  ==>  inBits: xxxx-xxxx xxxx-xxxx ##xx-xxxx
  inBits |= ((uint)inD & 0x30) << 2;

//--- PortF: ####-xxxx  ==>  inBits: xxxx-xxxx xxxx-#### xxxx-xxxx
  inBits |= ((uint)inF & 0xf0) << 4;

  inBits = (inBits & ~forcedDigInMask) | forcedDigInBits;

  return inBits;
}


//-----------------------------------------------------------------------------
//  set the state of the digital outputs
//-----------------------------------------------------------------------------
void WriteDigOutputs(ulong digOut_)
{
  ulong  digOut;


  digOut = (digOut_ & ~forcedDigOutMask) | forcedDigOutBits;

  digOut = ~digOut;

//--- digOut: xxxx-xxxx xxxx-xxxx ####-####  ==>  PortG: ####-####
  WrPortI(PGDR, &PGDRShadow, (uchar)digOut );

//--- digOut: xxxx-xxxx xxxx-x### xxxx-xxxx  ==>  PortB: ###x-xxxx
  WrPortI(PBDR, &PBDRShadow,
                (PBDRShadow & 0x1F) | (uint)((digOut & 0x0700) >> 3) );
}


//-----------------------------------------------------------------------------
//  Function to communicate with an ADS7825 chip (a 4-channel, 16-bit ADC)
//
//  Call:
//    convertADS7825(uint nextChan)
//
//    Input:
//       nextChan: Channel # to sample between calls and to do conversion
//                 on during the next call.
//
//    Result:
//       The raw value for the channel specified during the previous call
//       is updated.
//
//
//  To maximize the conversion rate, the ADS7825 begins acquiring/sampling
//  the signal to be converted as soon as the previous conversion is
//  completed.  This requires that it already know what channel to begin
//  sampling, so you have to tell it what channel to sample for the NEXT
//  conversion, NOT the current conversion.
//
//  NOTE that this does NOT mean that the conversion initiated by this call
//  is for outdated data.  The ADS7825 is sampling the channel specified
//  during the last call right up to the time it is told to start the
//  conversion.  The only catch is you have to know what channel that was
//  so when you read the conversion data you know what channel it is for.
//
//  The ADC7825 is fairly unique in that it has a TAG input that allows
//  multiple ADC's to be connected in a daisy-chain configuration, allowing
//  you to initiate a conversion on multiple chips at the same time and then
//  read the results from all of them in a single operation.  This greatly
//  decreases the additonal time needed to support more than 4 channels when
//  the chips are connected in this manner.
//
//  NOTE:  The downside, of course, is the cost per input, both in $ and in
//         board space.  Only 4 inputs means never more than 4 passes to
//         read any # of channels.  It also means one more (expensive) chip
//         for each additional group of 4 channels.
//
//-----------------------------------------------------------------------------
void convertADS7825(uint nextChan)
{
  uchar  buf[10];
  static uint  prevChan = 0;


//--- Set the bits that will tell the ADC which channel to sample between ---
//--- calls to this func, and to hold and convert during the next call ---
//--- (PE6 and PE7 are connected to ADR0 and ADR1 on all ADC's) ---
//--- NOTE:  PE6/7 are also used for CS- on two of the DAC's.  Because ---
//--- the ADC input selection bits are latched when a conversion starts, ---
//--- we can get away with this, so long as we don't generate any SPI ---
//--- clock pulses for the ADC chip while a DAC chip is selected ---
  WrPortI(PEDR, &PEDRShadow, (PEDRShadow & 0x3F) | ((nextChan & 3) << 6));

//--- set R/C- low first so we can use CS- to start conversion ---
//--- NOTE:  PE4 also used for CS- on one of the DAC's ---
  WrPortI(PEDR, &PEDRShadow, PEDRShadow & 0xEF);  // R/C- (PE4) low (convert)

//--- Toggle /CS low then high very quickly.  This starts the conversion ---
//--- on the previously specified input and latches the input # to be used ---
//--- for the next conversion.  /CS must be low for at least 40ns, but must ---
//--- go high again before BUSY does (aprox 25us).  If /CS is still low when ---
//--- BUSY goes high, the next conversion will start and the results of the ---
//--- previous one will be lost ---
  _EnterCritical_();
  WrPortI(PFDR, &PFDRShadow, PFDRShadow & 0xFB);  // /CS (PF2) low
  WrPortI(PFDR, &PFDRShadow, PFDRShadow | 0x04);  // /CS (PF2) high
  _ExitCritical_();

//--- set R/C- high to enable read mode ---
  WrPortI(PEDR, &PEDRShadow, PEDRShadow | 0x10);  // R/C- (PE4) high

//--- The design did not include the ability to read the BUSY signal, so ---
//--- add a short delay to avoid trying to read the data before it is ready ---
  for (uint xx=0; xx<20; ++xx);

//--- set both addr lines (shared with DAC CS- lines) back high ---
//--- to avoid having any DAC's reacting to the serial clock pulses ---
//--- about to be generated ---
  WrPortI(PEDR, &PEDRShadow, PEDRShadow | 0xC0);

//--- set /CS low again to enable transmission of data ---
  WrPortI(PFDR, &PFDRShadow, PFDRShadow & 0xFB);  // /CS (PF2) low

//--- The NSCL16 controller supports an external ADC chained to the one on ---
//--- the SBC, so we will get values for 2 channels (the same one from each ---
//--- ADC) at once: ---
//--- NOTE: Each ADC sends 17 bits (for a total of 34 bits, or 4.25 bytes). ---
//---       The first bit of each 17 is NOT part of the data ---
  _spiRead_(spiPort, buf, 5, false);

//--- set /CS high so ADC does not react to clock pulses not meant for it ---
  WrPortI(PFDR, &PFDRShadow, PFDRShadow | 0x04);  // /CS (PF2) high

//--- extract the raw value from the bytes just read ---
  adcRaw[prevChan] = (uint) (0x8000 ^ (
      ((uint)buf[0] << 9) |
      ((uint)buf[1] << 1) |
      ((uint)buf[2] >> 7)
    ) );
  adcRaw[prevChan+4] = (uint) (0x8000 ^ (
    ((uint)buf[2] << 10) |
    ((uint)buf[3] << 2) |
    ((uint)buf[4] >> 6)
  ) );

//--- keep track of what channel we started the conversion on, cause that ---
//--- is the channel whose value will be read during the next call ---
  prevChan = nextChan;
}


//-----------------------------------------------------------------------------
//  Read analog inputs.  The PS8Rack controller has 4 ADS7825 ADC's.  Each
//  ADC has 4 16-bit inputs and the 4 chips are daisy-chained together so
//  data can be read from all 4 chips in a single operation.
//-----------------------------------------------------------------------------
void readAnalogInputs(ulong mask)
{
  if (! spiPort)  return;

//--- Remember we have to tell the ADC which channel's data we will read ---
//--- NEXT time we call it, so consistently using the sequence 1, 2, 3, 0 ---
//--- each time means we get the data for 0, 1, 2, 3.  Also remember that ---
//--- this means channel 0 will be continously sampled (READY for conversion) ---
//--- between calls to this function ---

  _spiConfig_(spiPort, CLK_1MHz, NORM_LOW_LATCH_FALL);

  convertADS7825(1);  // chan to convert NEXT call
  convertADS7825(2);  // chan to convert NEXT call
  convertADS7825(3);  // chan to convert NEXT call
  convertADS7825(0);  // chan to convert NEXT call
}


//-----------------------------------------------------------------------------
//  Update the state of the AD5542 DAC chips
//-----------------------------------------------------------------------------
void writeAnalogOutputs(ulong mask)
{
  uchar  buf[2];
  uint   rawVal, bitTest;
  uchar  i;
  static const uchar  DAC_CS[4] = { 0x08, 0x10, 0x40, 0x80 };


  if (! spiPort)  return;

//--- set SPI clock mode ---
  _spiConfig_(spiPort, CLK_1MHz, NORM_LOW_LATCH_RISE);

  for (i=0,bitTest=1; i<NUM_DAC; ++i,bitTest<<=1)  if (bitTest & (uint)mask)  {

  //--- set /CS low for appropriate chip ---
    WrPortI(PEDR, &PEDRShadow, PEDRShadow & ~DAC_CS[i]);

  //--- send cmd to DAC chip ---
    rawVal = dacRaw[i];  buf[0] = rawVal >> 8;  buf[1] = rawVal & 0xff;
    _spiWrite_(spiPort, buf, 2, false);

  //--- set /CS high for appropriate chip ---
    WrPortI(PEDR, &PEDRShadow, PEDRShadow | DAC_CS[i]);
  }

}


//-----------------------------------------------------------------------------
/*
float ADCtoVolts(uchar chan, uint rawADC)
{
  return (ADCOffset(chan) - (float)(rawADC)) * ADCGain(chan);
}
*/

//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to transmit mode
//-----------------------------------------------------------------------------
void setExt485Tx(void)
{
}
//-----------------------------------------------------------------------------
//  Set an External 2/4-wire RS-485 adapter to receive mode
//-----------------------------------------------------------------------------
void setExt485Rx(void)
{
}

//-----------------------------------------------------------------------------
//  functions to enable/disable the transmitter for an RS-485 connection
//-----------------------------------------------------------------------------
void _nearcall SerSetDirA(int enableXmit)  {  }
void _nearcall SerSetDirB(int enableXmit)  {  }
void _nearcall SerSetDirC(int enableXmit)  {  }
void _nearcall SerSetDirD(int enableXmit)
{  // PE1 controls RS-485 line driver for port D(3) RS-485 connections
  char  val;
  val = enableXmit ? (PEDRShadow | 0x02) : (PEDRShadow & ~0x02);
  WrPortI(PEDR, &PEDRShadow, val);
}


//-----------------------------------------------------------------------------
void hdwInit(ulong digOutState)
{
  hdwType = HT_RFMON;

  __brdInitFlag = TRUE;

// (NI) = Normal Input
// (NO) = Normal Output
// (AI) = Alternate Input
// (AO) = Alternate Output
// (xx) = Used by ZWorld core - do not change

  //--- Port A --- (*must be inputs due to bug in earlier Rabbit 3000's)
  //  PA0 - (NI)  DI-16*
  //  PA1 - (NI)  DI-17*
  //  PA2 - (NI)  DI-18*
  //  PA3 - (NI)  DI-19*
  //  PA4 - (NI)  DI-20*
  //  PA5 - (NI)  DI-21*
  //  PA6 - (NI)  DI-22*
  //  PA7 - (NI)  DI-23*
    WrPortI(SPCR, &SPCRShadow, SPCRShadow & ~0x0c);  // PortA: byte input

  //--- Port B ---
  //  PB0 - (--)  Serial-B (SPI) clock (output)
  //  PB1 - (--)  Serial-A (Diag/Prog) clock (output)
  //  PB2 - (NI)  DI-02
  //  PB3 - (NI)  DI-03
  //  PB4 - (NI)  DI-04
  //  PB5 - (NO)  DO-08
  //  PB6 - (NO)  DO-09
  //  PB7 - (NO)  DO-10
    WrPortI(PBDDR, &PBDDRShadow, (PBDDRShadow|0xE0) & ~0x1C); // dir: OOOI-IIxx

  //--- Port C ---
  //  PC0 - (AO)  Serial-D transmit
  //  PC1 - (--)  Serial-D receive
  //  PC2 - (NO)  enable/disable Digital I/O buffers (output)
  //  PC3 - (--)  DI-05
  //  PC4 - (AO)  Serial-B (SPI) transmit
  //  PC5 - (--)  Serial-B (SPI) receive
  //  PC6 - (AO)  Serial-A (Diag/Prog) transmit
  //  PC7 - (--)  Serial-A (Diag/Prog) receive
    WrPortI(PCFR, &PCFRShadow, 0x51);  // func: xAxA-xOxA

  //--- Port D ---
  //  PD0 - (xx)  Ethernet Controller Reset (output)
  //  PD1 - (xx)  Not Avail (pulled high - otherwise not connected???)
  //  PD2 - (xx)  Ethernet Controller (TPOUT-)
  //  PD3 - (xx)  Ethernet Controller (TPOUT+)
  //  PD4 - (NI)  DI-06
  //  PD5 - (NI)  DI-07
  //  PD6 - (xx)  Ethernet Controller (TPIN-)
  //  PD7 - (xx)  Ethernet Controller (TPIN+)
    WrPortI(PDDDR, &PDDDRShadow, PDDDRShadow & ~0x30);  // dir:  xxII-xxxx
    WrPortI(PDFR, &PDFRShadow, PDFRShadow & ~0x30);     // func: xxNN-xxxx

  //--- Port E ---
  //  PE0 - (xx)  Ethernet Controller Interrupt (input)
  //  PE1 - (NO)  RS485 xmit/driver enable (output)
  //  PE2 - (xx)  Ethernet Controller Chip Select (output)
  //  PE3 - (NO)  DAC-00 chip select (output)
  //  PE4 - (NO)  DAC-01 chip select (output)
  //  PE5 - (--)  External interrupt (input)
  //  PE6 - (NO)  DAC-02 chip select (output)
  //  PE7 - (NO)  DAC-03 chip select (output)
    WrPortI(PEDDR, &PEDDRShadow, PEDDRShadow | 0xDA);  // dir:  OOxO-OxOx
    WrPortI(PEFR, &PEFRShadow, PEFRShadow & ~0xDA);    // func: NNxN-NxNx

  //--- Port F ---
  //  PF0 - (NI)  DI-00
  //  PF1 - (NI)  DI-01
  //  PF2 - (NO)  ADC Chip Select (output)
  //  PF3 - (NI)  ADC Ready (input)
  //  PF4 - (NI)  DI-08
  //  PF5 - (NI)  DI-09
  //  PF6 - (NI)  DI-10
  //  PF7 - (NI)  DI-11
    WrPortI(PFDDR, &PFDDRShadow, 0x04);  // dir:  IIII-IOII
    WrPortI(PFFR, &PFFRShadow, 0x00);    // func: NNNN-NNNN

  //--- Port G ---
  //  PG0 - (NO)  DO-00
  //  PG1 - (NO)  DO-01
  //  PG2 - (NO)  D0-02
  //  PG3 - (NO)  DO-03
  //  PG4 - (NO)  DO-04
  //  PG5 - (NO)  DO-05
  //  PG6 - (NO)  DO-06
  //  PG7 - (NO)  DO-07
    WrPortI(PGDDR, &PGDDRShadow, 0xFF);  // dir:  OOOO-OOOO
    WrPortI(PGFR, &PGFRShadow, 0x00);    // func: NNNN-NNNN


  //--- all parallel port outputs driven hi/lo (no open-drain outputs) ---
    WrPortI(PDDCR, &PDDCRShadow, 0x00);
    WrPortI(PFDCR, &PFDCRShadow, 0x00);
    WrPortI(PGDCR, &PGDCRShadow, 0x00);

  //--- set state of digital outputs ---
    digitalOut = digOutState;  WriteDigOutputs(digOutState);

  //--- enable input and output buffers ---
    WrPortI(PCDR, &PCDRShadow, PCDRShadow & ~0x04);  // bring PC2 low

  //--- deselect all DAC's ---
    WrPortI(PEDR, &PEDRShadow, PEDRShadow | 0xD8);  // set PE3,4,6,7 high

//    ser485Rx();  // disable the RS485 transmitter
//    setExt485Rx();  // set external 2/4-wire RS-485 adapter to Rx mode

    spiPort = NULL;  // set later by a call from the kernel process



  flashSize = 524288L;

  printf("\r\nInitialized %s hardware\r\n", ArchName);
}


//-----------------------------------------------------------------------------
//  Initialize and register hardware-dependent file systems
//-----------------------------------------------------------------------------
void registerFileSystems(void)
{
  socketInit_();
  srfsInit_();
  serialInit_();

  flashFileInit_();
}

//-----------------------------------------------------------------------------
//  Initialize the network hardware
//-----------------------------------------------------------------------------
void netHdwInit()
{
//  Select ethernet hardware:
//  usingAsix, usingDevKit, usingRealtek, usingSMSC
//  usingAll() can be used to support all cmd core modules, but req's more code

  usingAsix();
}


//-----------------------------------------------------------------------------
//  allocate any resources that should be owned by the kernel process
//
//  To be called ONLY by the kernel process!!!!
//-----------------------------------------------------------------------------
void hdwInit2(void)
{
//  char  *argv_[3];
//  uint  argc_;
  fd_t  spifd;


  spifd = _serial_(ADC_DAC_SPIPort, 0, SER_SPI, SER_NO_FLOW);

  spiPort = _FileFromFD_(spifd)->ioDev;

printf("spiPort: %04lX (%d)\r\n", (ulong)spiPort, spiPort->port);  //tdebug

//  argv_[0] = "getty";  argv_[1] = "19200";  argv_[2] = "tty3";  argc_ = 3;
//  createProc(getty, 5, argc_, argv_);
}

//-----------------------------------------------------------------------------

