//----- asflash.c ----- 02/08/12 -----(05/22/06)-----

//-----------------------------------------------------------------------------
//  Standard DEBROS virtual file/device I/O API for reading and writing to
//  an AMTEL serial flash chip
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2012-02-08 - Mark Davis
//               - errno is now set to EBADF rather than ENXIO when a function
//                 is passed an invalid device pointer.
//
//  2008-04-07 - Mark Davis
//               - began work on new module using fifo.c as a starting point
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <errno.h>
#include <common.h>

#include <string.h>

#include <sys/asflash.h>

#include <sys/serial.h>
#include <sys/spi.h>

#include <sys/farprintf.h>

#include <linux/limits.h>


#define  MaxFDs   2   // max # of open file descriptors for this device

#define  DevDescr     "asflash"
#define  NodeDescr    "node"


typedef struct  {
  bool      inUse;
  char far *name;
  ulong     offset;  // offset for next r/w operation
} FileInfo;


static FileInfo  openFiles[MaxFDs];

char far  fileNames[MaxFDs][NAME_MAX];   // name associated with the FIFO pipe


FILESYS_FUNCS(asflash)

IODEV_FUNCS(asflash)

#define FileRef  ((FileInfo *)ioDev)


#define  EEPROM_SIZE  524288L  // size of serial EEPROM attached to spi2
//#define  EEPROM_SIZE  131072L  // size of serial EEPROM attached to spi2
//#define  EEPROM_SIZE   32768L  // size of serial EEPROM attached to spi2

//--- instruction codes for ATMEL SPI Serial EEPROMs ---
#define  EEPROM_WREN   0x06   // WRite ENable
#define  EEPROM_WRDI   0x04   // WRite DIsable latch
#define  EEPROM_RDSR   0x05   // ReaD Status Register
#define  EEPROM_WRSR   0x01   // WRite Status Register
#define  EEPROM_READ   0x03   // Read bytes from memory
#define  EEPROM_WRITE  0x02   // Write a page to memory

#define  EEPROM_Enable()   \
    WrPortI(PFDR, &PFDRShadow, PFDRShadow & 0xFE)  // Chip Select (PF0) low
#define  EEPROM_Disable()   \
    WrPortI(PFDR, &PFDRShadow, PFDRShadow | 0x01)  // Chip Select (PF0) high


#define  WriteSPI2(buf, len)    _spiWrite_(spi2Port, buf, len, false)


// The fd is set by the kernel task on startup (by the call to asflashInit_()
// so it is useless when other tasks call fcntl functions which in turn call
// functions in this module.

// So we save a pointer to the SPI I/O device itself so these functions can
// work for all tasks.

fd_t spi2fd;
SerialInfo  *spi2Port;

//-----------------------------------------------------------------------------
//  Initialize data structures for managing named pipes (asflash streams)
//-----------------------------------------------------------------------------
void asflashInit_(void)
{
  uint  u;
  int  clockRate;


//--- initialize the list of open "files" ---
  BZERO(openFiles);
  for (u=0; u<MaxFDs; ++u)  {
    openFiles[u].name = fileNames[u];
  }


  clockRate = CLK_1MHz;

  spi2fd = open("spi2", O_WRONLY, 0660);  spi2Port = NULL;
  printf("spi2fd: %d\r\n", spi2fd);
  if (spi2fd < 0)
    printf("*** Unable to open 'spi2' ***\r\n");
  else  {
    ioctl(spi2fd, IOC_SET_CLK_RATE, &clockRate);
    spi2Port = _FileFromFD_(spi2fd)->ioDev;
  }


  asflashFSysFuncs.tick = NULL;

  registerFileSys(&asflashFSysFuncs);
}

//-----------------------------------------------------------------------------
//  Returns true if pointer is a valid, in-use file ref for this device
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validFileRef(void *ioDev)
{
  if (ioDev)
    if (FileRef == openFiles + (int)(FileRef-openFiles))
      if (FileRef->inUse)  return true;

  return false;
}

//-----------------------------------------------------------------------------
//  return a file descriptor for a file stored on this device
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
fd_t _asflashOpen_(const char *path, int flags, int mode)
{
  uint  u;
  fd_t  fd;
  FileInfo *fileRef;


//--- ignore names that don't start with "/asf/" ---
  if (strncmp(path, "/asf/", 5))  { errno = EBADF;  return -1; }


  KLOCK();

//--- look for the specified name in the list of Flash fd's ---
  for (u=0; u<MaxFDs; ++u)  {
    fileRef = openFiles + u;
    if (fileRef->inUse)  continue;
    //... do whatever other checks are needed here ...
    break;
  }
  if (u == MaxFDs)  { KUNLOCK();  ERRNO = ENFILE;  return -1; }

//--- allocate a file descriptor to reference the device file ---
  if ((fd = _allocFile_(3, &asflashIOFuncs, fileRef)) < 0)  {
    KUNLOCK();  return -1;  }

  fileRef->offset = 0;
  fileRef->inUse = true;
  //... do whatever other initiailization is needed here ...

  KUNLOCK();

//--- everything OK - return the descriptor assigned for this process ---
  return fd;
}

//-----------------------------------------------------------------------------
//  func called each kernel tick to provide file system with CPU cycles
//  Should ONLY be called by the kernel task!
//-----------------------------------------------------------------------------
void _asflashTick_(void)
{
  //  if this func needs to actually be called in the future, the line in
  //  asflashInit_() that sets the pointer to this func to NULL must be
  //  commented out...
}


//-----------------------------------------------------------------------------
int _asflashGetQueues_(void *ioDev, WaitQueues far *queues)
{
//--- use NULL rather than sleepQueue if a request cannot fail one time ---
//--- and succeed later on ---
  queues->readQueue = queues->writeQueue = NULL;
  return 0;
}


//-----------------------------------------------------------------------------
//  Read bytes from an open file
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int  _asflashRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  int  i;


  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (numBytes <= 0)  { KUNLOCK();  return 0; }

//--- replace with code to do a real read ---
  for (i=0; i<numBytes; ++i)  {
    buf[i] = (char)(FileRef->offset & 0xff);
    ++FileRef->offset;
  }

  KUNLOCK();

  return numBytes;
}

//-----------------------------------------------------------------------------
//  Write bytes to an open file
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int  _asflashWrite_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (numBytes <= 0)  { KUNLOCK();  return 0; }

//--- add code to do an actual write and return a valid count ---

  KUNLOCK();

  return numBytes;
}


//-----------------------------------------------------------------------------
//  close a asflash
//-----------------------------------------------------------------------------
int  _asflashClose_(void *ioDev)
{
  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  FileRef->inUse = false;

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to reposition a file offset
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
off_t  _asflashLseek_(void *ioDev, off_t offset, int whence)
{
  KLOCK();

  if (! validFileRef(ioDev))  { ERRNO = EBADF;  return -1; }

  FileRef->offset = offset;

  KUNLOCK();

  return offset;
}


//-----------------------------------------------------------------------------
//  Get information about an open device file
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int  _asflashFstat_(void *ioDev, struct stat *fsbuf)
{
  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

//finish
  fsbuf->st_size  = 12345;         // total size in bytes
  fsbuf->st_atime = 1207594478L;   // time of last access
  fsbuf->st_mtime = 1207594478L;   // time of last modification
  fsbuf->st_ctime = 1207594478L;   // time of last status change

  fsbuf->st_mode = S_IFREG | 00777;

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for this device's files
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _asflashioctl_(void *ioDev, int request, void *arg)
{
  int  stat;


  KLOCK();

  if (! validFileRef(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  stat = -1;

  switch (request)  {

//    case TIOCINQ:
//      *((int *)arg) = FileRef->bytesUsed;  stat = 0;  break;

    default:
      errno = EINVAL;
  }

  KUNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to get a string description of a device file
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
char *_asflashDescr_(void *ioDev, char *buf, int bufSize)
{
  char  *ptr;

  sprintf(buf, "%s 0x%04lX ", DevDescr, (ulong)ioDev);

  ptr = buf + strlen(buf);  strcpy(ptr, "bad bad");

//finish: limit output to bufSize

  if (validFileRef(ioDev))
    sprintf(ptr, "%s[%d] %ls",
            NodeDescr, (int)(FileRef - openFiles), FileRef->name);

  return buf;
}

//-----------------------------------------------------------------------------
