//----- fifo.c ----- 02/08/12 -----(05/22/06)-----

//-----------------------------------------------------------------------------
//  First-In-First-Out circular buffer devices.  Also used to implement named
//  pipes.
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2012-02-08 - Mark Davis
//               - errno is now set to EBADF rather than ENXIO when a function
//                 is passed an invalid device pointer.
//
//  2009-04-02 - Mark Davis
//               - Finished support for using FIFOs as pipes.
//
//  2009-03-25 - Mark Davis
//               - Added unlink and rename functions
//
//  2009-03-23 - Mark Davis
//               - Changed the inUse flag to an openCount to properly handle
//                 and support having multiple global FD's reference the same
//                 FIFO
//
//  2006-05-xx - Mark Davis
//               - original version started
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <errno.h>
#include <common.h>

#include <string.h>

#include <sys/fifo.h>

#include <sys/farprintf.h>

#include <linux/limits.h>


typedef struct  {
  int       openCount;     // # FDEVs that point to this ioDev struct
  int       maxOpenCount;  // max # of fds that were open at once
  char far *name;          // 1st byte == 0 if entry not in use
  char far *buf;           // circular buffer
  uint      bytesUsed;     // # of bytes of buf[] in use
  uint      inIdx;         // index for next char to be written
  uint      outIdx;        // index for next char to be read
  bool      pipeMode;      // treat as a pipe
} FIFOInfo;


static char far  fifoFileName[MaxFIFOs][NAME_MAX];   // name associated with the FIFO pipe

static FIFOInfo  fifoTbl[MaxFIFOs];
static char far fifoBuf[MaxFIFOs][FIFOSize];


IODEV_FUNCS(fifo)

FILESYS_FUNCS(fifo)

#define FIFO  ((FIFOInfo *)ioDev)


//-----------------------------------------------------------------------------
//  Initialize data structures for managing named pipes (FIFO streams)
//-----------------------------------------------------------------------------
void fifoInit_(void)
{
  uint  u;
  FIFOInfo  *fifo;


  BZERO(fifoTbl);

  for (u=0; u<MaxFIFOs; ++u)  {
    fifo = fifoTbl + u;
    fifo->buf = fifoBuf[u];
    fifo->name = fifoFileName[u];
    fifo->name[0] = 0;  // compiler screws this up...
//    fifoFileName[u][0] = 0;
  }

  registerFileSys(&fifoFSysFuncs);
}

//-----------------------------------------------------------------------------
//  Returns true if provided pointer is a valid, in-use FIFO device.
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validFIFO(void *ioDev)
{
  char far *name;

  if (ioDev)
    if (FIFO == fifoTbl + (int)(FIFO - fifoTbl))  {
      name = FIFO->name;  // to get around compiler bug
      if (*name)  return true;
    }

  return false;
}

//-----------------------------------------------------------------------------
//  Return ptr into table of existing FIFO's, or NULL if name not found
//-----------------------------------------------------------------------------
FIFOInfo *fifoPtr(const char *path)
{
  uint  u;
  FIFOInfo  *fifo;

//--- look for the specified name in the list of FIFO's ---
  for (u=0; u<MaxFIFOs; ++u)  {
    fifo = fifoTbl + u;
    if (fstrcmp(fifo->name, path) == 0)  return fifo;
  }

  return NULL;
}


//-----------------------------------------------------------------------------
//  Open a FIFO created by mkfifo
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
fd_t _fifoOpen_(const char *path, int flags, int mode)
{
  fd_t  fd;
  FIFOInfo  *fifo;


  KLOCK();

  fifo = fifoPtr(path);  if (!fifo)  { KUNLOCK();  ERRNO = ENOENT;  return -1; }

//--- allocate a file descriptor to reference the FIFO ---
  if ((fd = _allocFile_(3, &fifoIOFuncs, fifo)) < 0)  { KUNLOCK(); return -1; }

//--- keep track of how many global FD's are using this same FIFO ---
  ++fifo->openCount;

  if (fifo->openCount > fifo->maxOpenCount)
    fifo->maxOpenCount = fifo->openCount;

  KUNLOCK();

  return fd;
}

//-----------------------------------------------------------------------------
//  func called each kernel tick to provide file system with CPU cycles
//  Should ONLY be called by the kernel task!
//-----------------------------------------------------------------------------
void _fifoTick_(void)
{

}

//-----------------------------------------------------------------------------
//  Return the state of the specified ioDev entry or NULL if not a valid
//  entry #.
//
//  For valid entrys, the entryState struct is filled in
//-----------------------------------------------------------------------------
void *_fifoEntry_(uint idx, EntryState *entryState)
{
  char far *name;
  FIFOInfo  *fifo;

  if (idx >= MaxFIFOs)  return NULL;

  fifo = fifoTbl + idx;  name = fifo->name;

  entryState->openCount = fifo->openCount;
  entryState->unused = (*name == 0);  // compiler can't handle "*fifo->name"

  return fifo;
}

//-----------------------------------------------------------------------------
//  Delete a named pipe (FIFO)
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _fifoUnlink_(char *path)
{
  FIFOInfo  *fifo;


  KLOCK();

  fifo = fifoPtr(path);  if (!fifo)  { KUNLOCK();  ERRNO = ENOENT;  return -1; }

//--- don't delete it if it is in use ---
  if (fifo->openCount > 0)  { KUNLOCK();  ERRNO = EBUSY;  return -1; }

  fifo->name[0] = 0;  // compiler does this OK because it is an assignment?

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Change the name of a named pipe (FIFO)
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _fifoRename_(const char *oldPath, const char *newPath)
{
  FIFOInfo  *fifo;


  KLOCK();

//--- can't do it if the new name is already in use ---
  if (fifoPtr(newPath))  { KUNLOCK();  ERRNO = EEXIST;  return -1; }

  fifo = fifoPtr(oldPath);
  if (!fifo)  { KUNLOCK();  ERRNO = ENOENT;  return -1; }

  fstrcpy(fifo->name, newPath);

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Create a named pipe (FIFO)
//-----------------------------------------------------------------------------
int mkfifo(const char *pathName, int mode)
{
  uint  u, unusedIdx;
  char far *name;
  FIFOInfo  *fifo;


  if (strlen(pathName) > NAME_MAX-1)  { errno = ENAMETOOLONG;  return -1; }

  KLOCK();

//--- check for the name already in use ---
  unusedIdx = MaxFIFOs;
  for (u=0; u<MaxFIFOs; ++u)  {
    fifo = fifoTbl + u;
    name = fifo->name;  // to get around compiler bug
    if (! *name)  { unusedIdx = u;  continue; }
    if (fstrcmp(name, pathName) == 0)  break;
  }
  if (u != MaxFIFOs)  { KUNLOCK();  ERRNO = EEXIST;  return -1; }

//--- did we run across an unused entry? ---
  if (unusedIdx == MaxFIFOs)  { KUNLOCK();  ERRNO = ENFILE;  return -1; }

//--- initialize the new pipe ---
  fifo = fifoTbl + unusedIdx;
  fstrcpy(fifo->name, pathName);
  fifo->inIdx = fifo->outIdx = fifo->bytesUsed = 0;
  fifo->openCount = fifo->maxOpenCount = 0;

  fifo->pipeMode = false;

  KUNLOCK();  return 0;
}


//-----------------------------------------------------------------------------
//  open a pipe and return a pair of file descriptors for it (one for reading
//  and one for writing)
//-----------------------------------------------------------------------------
int pipe(int fd[2])
{
  int  isPipe = 1;
  char  fifoName[NAME_MAX];


  sprintf(fifoName, "fifo_%u_%u", getpid(), (uint)GetMS());
  if (mkfifo(fifoName, 0666))  return -1;

  fd[0] = open(fifoName, O_RDONLY, 0);
  fd[1] = open(fifoName, O_WRONLY, 0);

  if ((fd[0] < 0) OR (fd[1] < 0))  {
    if (fd[0] >= 0)  { close(fd[0]);  fd[0] = -1; }
    if (fd[1] >= 1)  { close(fd[1]);  fd[1] = -1; }
    return -1;
  }

  ioctl(fd[0], PIPE_MODE, &isPipe);

  return 0;
}

//-----------------------------------------------------------------------------
int _fifoGetQueues_(void *ioDev, WaitQueues far *queues)
{
  queues->readQueue = queues->writeQueue = &sleepQueue;
  return 0;
}


//-----------------------------------------------------------------------------
//  Attempt to read bytes from a FIFO stream
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int  _fifoRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  uint  i, count;


  KLOCK();

  if (! validFIFO(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (numBytes <= 0)  { KUNLOCK();  return 0; }

//--- blocking/non-blocking handled by fcntl read() call ---
//--- If there are no bytes left to read, and the FIFO is being used as a ---
//--- pipe, and the writer has gone away, then return 0 (EOF) ---
//--- Otherwise, indicate that trying again may provide some data ---
  if (! FIFO->bytesUsed)
    if (FIFO->pipeMode AND (FIFO->openCount < FIFO->maxOpenCount))  {
      KUNLOCK();  return 0;  }
    else  {
      KUNLOCK();  ERRNO = EAGAIN;  return -1;  }

  if ((count = FIFO->bytesUsed) > numBytes)  count = numBytes;

//finish - do 1 or 2 block copies rather than one-char-at-a-time (ZZZzzz...)
  for (i=0; i<count; ++i)  {
    buf[i] = FIFO->buf[FIFO->outIdx];
    if (++FIFO->outIdx == FIFOSize)  FIFO->outIdx = 0;
  }
  FIFO->bytesUsed -= count;

  KUNLOCK();

  return count;
}

//-----------------------------------------------------------------------------
//  Attempt to write bytes to a FIFO stream
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int  _fifoWrite_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  uint  i, j;
  int  unused;


  KLOCK();

  if (! validFIFO(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (numBytes <= 0)  { KUNLOCK();  return 0; }

//--- NOTE: blocking/non-blocking handled by fcntl write() call ---

//--- all writes must be atomic, so its all or nothing! ---
  unused = FIFOSize - FIFO->bytesUsed;
  if (unused < numBytes)
    if (FIFO->pipeMode AND (FIFO->openCount < FIFO->maxOpenCount))  {
      KUNLOCK();  kill(getpid(), SIGPIPE);  ERRNO = EPIPE;  return -1;  }
    else  {
      KUNLOCK();  ERRNO = EAGAIN;  return -1;  }

  for (i=0,j=FIFO->inIdx; i<numBytes; ++i)  {
    FIFO->buf[j] = buf[i];
    if (++j == FIFOSize)  j = 0;
  }
  FIFO->inIdx = j;  FIFO->bytesUsed += numBytes;

  KUNLOCK();

  return numBytes;
}


//-----------------------------------------------------------------------------
//  close a FIFO
//-----------------------------------------------------------------------------
int  _fifoClose_(void *ioDev)
{
  KLOCK();

  if (! validFIFO(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (FIFO->openCount > 0)  --FIFO->openCount;

//--- if both "ends" of a pipe are closed, "delete" the FIFO ---
  if (FIFO->pipeMode AND (FIFO->openCount < 1))  FIFO->name[0] = 0;

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to reposition a file offset
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
off_t  _fifoLseek_(void *ioDev, off_t offset, int whence)
{
  KLOCK();

  if (! validFIFO(ioDev))  { ERRNO = EBADF;  return -1; }

  KUNLOCK();

  ERRNO = ESPIPE;  return (off_t)-1;
}


//-----------------------------------------------------------------------------
//  Get information about the FIFO
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int  _fifoFstat_(void *ioDev, struct stat *fsbuf)
{
  KLOCK();

  if (! validFIFO(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  fsbuf->st_mode = S_IFIFO | 00777;

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for fifo pipes
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _fifoioctl_(void *ioDev, int request, void *arg)
{
  int  stat;


  KLOCK();

  if (! validFIFO(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  stat = -1;

  switch (request)  {

    case TIOCINQ:
      *((int *)arg) = FIFO->bytesUsed;  stat = 0;  break;

    case PIPE_MODE:
      FIFO->pipeMode = true;  stat = 0;  break;

    default:
      errno = EINVAL;
  }

  KUNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to get a string description of an FIFO
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
char *_fifoDescr_(void *ioDev, char *buf, int bufSize)
{
  char  *ptr;

  sprintf(buf, "fifo 0x%04lX ", (ulong)ioDev);

  ptr = buf + strlen(buf);  strcpy(ptr, "bad bad");

//finish: limit output to bufSize

  if (validFIFO(ioDev))
    sprintf(ptr, "fifo[%d] %ls", (int)(FIFO - fifoTbl), FIFO->name);

  return buf;
}

//-----------------------------------------------------------------------------
