//----- serial.c ----- 03/21/12 -----(04/26/06)-----

//-----------------------------------------------------------------------------
//  Wrapper to provide standard fcntl API to Softools interrupt-driven serial
//  functions.
//
//  The individual port-specific functions in the Softools libs are mostly
//  thin wrappers for a common set of functions used for all ports.  The
//  port-specific functions set pointers to the port-specific funcs and data
//  structures and then call the generic routine to do the work.
//
//  One issue of particular note is that the Softools interface does not
//  provide a means of enabling or disabling in-band (xon/xoff) flow control
//  on an already intitialized port.  Whether or not to use flow control must
//  be known at the time the port is initialized.
//
//  So, for the time being, specifying use of flow control on a port is
//  handled in the fcntl layer by using the prefix ttyx rather than tty for
//  the device file name.
//
//finish:
//  Eventually, this should all be handled by supporting the standard
//  termios(3) API.
//
//  For details on Softools code, see notes below.
//
//-----------------------------------------------------------------------------
//  Update Log (most recent first):
//
//
//  2012-03-21 - Mark Davis
//               - flushConsole() now ensures that interrupts are enabled
//                 until it returns.  On return it restores the IP register to
//                 what it was before it was called.  Prior to this, the
//                 function did nothing (except generate an error message) if
//                 it was called during the portion of startup where interrupts
//                 are normally disabled.
//
//  2012-02-08 - Mark Davis
//               - errno is now set to EBADF rather than ENXIO when a function
//                 is passed an invalid device pointer.
//
//  2011-11-22 - Mark Davis
//               - Removed the static specifier from several items so they
//                 would show up in the map file.
//
//  2010-05-14 - Mark Davis
//               - ioctl function now support FIONSPACE request.
//
//  2010-04-23 - Mark Davis
//               - Added flushConsole() function to better encapsulate the
//                 fact that serial port A is the console.
//
//  2010-03-04 - Mark Davis
//               - set inUse flag for the port to true at start of _serial_()
//                 init function so the rest of the initialization process will
//                 not return errors.  The flag is still cleared if an error
//                 occurs during the rest of the initialization.
//
//  2010-03-03 - Mark Davis
//               - Changed SetSetDirX() funcs from near to _nearcall
//
//  2010-02-02 - Mark Davis
//               - Added check for port B in the init function to set the
//                 bits that use the alternate I/O pins PD4 and PD5 (since
//                 the normal I/O pins for B are used for other things - in
//                 particular PC4 is used for the Ethernet RSTDRV pin!)
//
//  2009-12-18 - Mark Davis
//               - Made several near functions _nearcall to free up root space
//
//  2009-05-11 - Mark Davis
//               - The write func once again ignores any pending schedule()
//                 calls for ports in packet mode (this minimizes the chance
//                 that task switches could result in a pause in the middle
//                 of a packet transmission).  With this change, a pause should
//                 never happen if packets are small enough to fit in the
//                 xmit buffer and a single write call is used for each packet.
//
//  2009-05-08 - Mark Davis
//               - improved efficiency of _serialRead_() and _serialWrite_()
//                 and reduced their effect on scheduler latency.
//
//  2009-05-06 - Mark Davis
//               - changes to address RS485 I/O issues
//
//  2009-03-26 - Mark Davis
//               - Added placeholder unlink and rename functions
//
//  2009-03-24 - Mark Davis
//               - added _SerialEntry_() function for getting info about serial
//                 driver fds
//
//  2006-04-26 - Mark Davis
//               - started first version
//
//-----------------------------------------------------------------------------
//  To-Do:
//
//   - This module currently relies on the sleep wait queue for blocking
//     operations.  It would be more efficient if it provided real wait queues
//     of its own.
//
//-----------------------------------------------------------------------------


/*-----------------------------------------------------------------------------
************  Notes on Softools Rabbit serial lib code  **********

Code located at:
    C:\Program Files\Softools WinIDE\Rabbit 1.70\Rabbit runtime\Rabbit serial


PORT/DEVICE INDEPENDENT FUNCTIONS:

============
 serialIO.c
============
  _SerialPutc() - puts char in to device's xmit queue

  _SerialGetc() - removes char from device's rcv queue

=============
 serialInt.c
=============
  _serialIsr() - called for a rcv interrupt (when not using xon/xoff)
               - put rcvd char in to device's rcv queue

  _isrSendCommon() - called for an xmit interrupt
               - remove char from device's queue and xmit it

================
 serialIntXon.c
================
  _serialIsrXon() - called for a rcv interrupt (when using xon/xoff)
                  - put rcvd char in to device's rcv queue
                  - enables/disables xmit for Xon/Xoff chars
                  - xmits Xon/Xoff based on rcv queue space


PORT-SPECIFIC FUNCTIONS

  ("A" specifies port 0, "B" port 1, ...)

==============
 serialAcom.c
==============
  SerialInitAcom() - initialization not related to Xon/Xoff mode
  + all the other funcs for using port A, all independent of Xoff/Xoff mode

===========
 serialA.c
===========
  SerialInitA() - initialization for port A without Xon/Xoff support


==============
 serialAxon.c
==============
  SerialInitAxon() - initialization for port A with Xon/Xoff xupport

=================
 serialArtsCts.c
=================
  SerialInitCtsRtsA() - set RTS/CTS flow control parameters

==============
 SerialA485.c
==============
  void  SerialInit485D(char myAddr, char use9thBit, void (*enableFunc)(int))
  int SerialStart485txD(char addr, int store)
  void SerialEnd485txD(void)
  void SerialEnd485rxD(void)




------------------------------------------------------------------------------*/

// SerialInitAxon

#include <serial.h>  // to get "Rabbit runtime\Rabbit serial\serial.h"

#include <common.h>

#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <errno.h>

#include <sys/serial.h>
#include <sys/spi.h>

#include <Rabbit.h>
#include <dcdefs.h>

#include <sys/farprintf.h>


extern _serialDevice  __serialA;
extern _serialDevice  __serialB;
extern _serialDevice  __serialC;
extern _serialDevice  __serialD;



SerialInfo  serPort[NumSerialPorts];

//--- modes allowed for each port ---
static uchar   serModes[NumSerialPorts] = {
  #if (_CPU_ID_ >= R3000)
    SER_RS232 | SER_SPI,
    SER_RS232 | SER_SPI,
    SER_RS232 | SER_SPI,
    SER_RS232 | SER_SPI | SER_RS485
  #else
    SER_RS232 | SER_SPI,
    SER_RS232 | SER_SPI,
    SER_RS232,
    SER_RS232 | SER_RS485
  #endif
};


IODEV_FUNCS(serial)

FILESYS_FUNCS(serial)

extern IODevFuncs  spiIOFuncs;


typedef  int (*GetCFunc)(void);
typedef  int (*PutCFunc)(int ch);



//--------- near functions to call lib's far funcs --------

#define  NearInitFunc(PORT)  \
  char _nearcall SerInit##PORT(serialRate rate, int mode, int ip,  \
           char *rcvBuf, int rcvSize, char *xmitBuf, int xmitSize, int flow)  \
{  \
  if (flow == SER_XON_XOFF)  {  \
    printf("*** xon init ***\r\n");  \
    return  \
       SerialInit##PORT##xon(rate, mode, ip, \
                        rcvBuf, rcvSize,  rcvSize*7/8, rcvSize/8,  \
                        xmitBuf, xmitSize);  \
  }  \
  \
  return  \
       SerialInit##PORT(rate, mode, ip, rcvBuf, rcvSize, xmitBuf, xmitSize); \
}
NearInitFunc(A)
NearInitFunc(B)
NearInitFunc(C)
NearInitFunc(D)


#define  NearSetPktModeFunc(PORT)  \
  void _nearcall SerSetPktMode##PORT(  \
       char startCh, char use9thBit, void (*enableFunc)(int) )  \
  {  SerialInit485##PORT(startCh, use9thBit, enableFunc); }
NearSetPktModeFunc(A)
NearSetPktModeFunc(B)
NearSetPktModeFunc(C)
NearSetPktModeFunc(D)

#define  NearEndPktRXFunc(PORT)  \
   void _nearcall SerEndPktRX##PORT(void) { SerialEnd485rx##PORT(); }
NearEndPktRXFunc(A)
NearEndPktRXFunc(B)
NearEndPktRXFunc(C)
NearEndPktRXFunc(D)

#define  NearStartPktTXFunc(PORT)  \
  int _nearcall SetStartPktTX##PORT(char startCh, int store)  \
  { return  SerialStart485tx##PORT(startCh, store);  }
NearStartPktTXFunc(A)
NearStartPktTXFunc(B)
NearStartPktTXFunc(C)
NearStartPktTXFunc(D)

#define  NearEndPktTXFunc(PORT)  \
   void _nearcall SerEndPktTX##PORT(void) { SerialEnd485tx##PORT(); }
NearEndPktTXFunc(A)
NearEndPktTXFunc(B)
NearEndPktTXFunc(C)
NearEndPktTXFunc(D)


#define  NearSet7O1Func(PORT)  \
   void _nearcall SerSet7O1##PORT(void) {  \
     printf("set format to 7O1\r\n"); \
   SerialSetOddParity##PORT(); }
NearSet7O1Func(A)
NearSet7O1Func(B)
NearSet7O1Func(C)
NearSet7O1Func(D)

#define  NearSet7E1Func(PORT)  \
   void _nearcall SerSet7E1##PORT(void) {  \
     printf("set format to 7E1\r\n");  \
   SerialSetEvenParity##PORT(); }
NearSet7E1Func(A)
NearSet7E1Func(B)
NearSet7E1Func(C)
NearSet7E1Func(D)

#define  NearSet8O1Func(PORT)  \
   void _nearcall SerSet8O1##PORT(void) {  \
     printf("set format to 8O1\r\n");  \
   SerialSet8OddParity##PORT(); }
NearSet8O1Func(A)
NearSet8O1Func(B)
NearSet8O1Func(C)
NearSet8O1Func(D)

#define  NearSet8E1Func(PORT)  \
   void _nearcall SerSet8E1##PORT(void) {  \
     printf("set format to 8E1\r\n");  \
   SerialSet8EvenParity##PORT(); }
NearSet8E1Func(A)
NearSet8E1Func(B)
NearSet8E1Func(C)
NearSet8E1Func(D)

//--- defined in the hdw-platform-specific arch modules ---
void _nearcall SerSetDirA(int enableXmit);
void _nearcall SerSetDirB(int enableXmit);
void _nearcall SerSetDirC(int enableXmit);
void _nearcall SerSetDirD(int enableXmit);


#define  NearXmitBufUsedFunc(PORT)  \
  uint _nearcall XmitBufUsed##PORT(void)  \
  { return  SerialSendCount##PORT();  }
NearXmitBufUsedFunc(A)
NearXmitBufUsedFunc(B)
NearXmitBufUsedFunc(C)
NearXmitBufUsedFunc(D)

#define  NearFlushXmitFunc(PORT)  \
  void _nearcall FlushXmit##PORT(void)  \
  { SerialSendFlush##PORT();  }
NearFlushXmitFunc(A)
NearFlushXmitFunc(B)
NearFlushXmitFunc(C)
NearFlushXmitFunc(D)


#define  NearRcvAvailFunc(PORT)  \
  uint _nearcall RcvAvail##PORT(void)  \
  { return  SerialRecvCount##PORT();  }
NearRcvAvailFunc(A)
NearRcvAvailFunc(B)
NearRcvAvailFunc(C)
NearRcvAvailFunc(D)

#define  NearFlushRcvFunc(PORT)  \
  void _nearcall FlushRcv##PORT(void)  \
  { SerialRecvFlush##PORT();  }
NearFlushRcvFunc(A)
NearFlushRcvFunc(B)
NearFlushRcvFunc(C)
NearFlushRcvFunc(D)



typedef struct  {
  char (*init)( serialRate rate, int mode, int ip,
                char *recvBuffer, int recvSize,
                char *tranBuffer, int tranSize, int flowCtl );
  int  (*getc)(void);
  int  (*putc)(int ch);
  void (*set7O1)(void);
  void (*set7E1)(void);
  void (*set8O1)(void);
  void (*set8E1)(void);
  void (*setPktMode)(char startCh, char use9thBit, void (*_enableFunc)(int) );
  void (*setDir)(int);  // non-zero ==> xmit mode, 0 ==> rcv mode
  void (*endPktRX)(void);
  int  (*startPktTX)(char startCh, int store);
  void (*endPktTX)(void);
  uint (*xmitBufUsed)(void);
  void (*flushXmit)(void);
  uint (*rcvAvail)(void);
  void (*flushRcv)(void);
} SerialFuncs;


#define  SER_FUNCS(port)  {  \
    SerInit##port, SerialGetc##port, SerialPutc##port,  \
    SerSet7O1##port, SerSet7E1##port, SerSet8O1##port, SerSet8E1##port, \
    SerSetPktMode##port, SerSetDir##port, SerEndPktRX##port,  \
    SetStartPktTX##port, SerEndPktTX##port,  \
    XmitBufUsed##port, FlushXmit##port, RcvAvail##port, FlushRcv##port  \
}

SerialFuncs  serFuncs[NumSerialPorts] =
{ SER_FUNCS(A), SER_FUNCS(B), SER_FUNCS(C), SER_FUNCS(D) };

serialPort serPortRef[] = { SER_A, SER_B, SER_C, SER_D };


#define SerOutBufSize  256
#define SerInBufSize   512

char serInBufs[NumSerialPorts][SerInBufSize],
     serOutBufs[NumSerialPorts][SerOutBufSize];


#define SER   ((SerialInfo *)ioDev)
#define SDEV  ((_serialDevice *)(SER->sDev))


//-----------------------------------------------------------------------------
//  func called once by init process to initialize data structures
//-----------------------------------------------------------------------------
void serialInit_()
{
//  uchar  i;


  BZERO(serPort);

  serPort[0].sDev = &__serialA;
  serPort[1].sDev = &__serialB;
  serPort[2].sDev = &__serialC;
  serPort[3].sDev = &__serialD;

/*
  for (i=0; i<NumSerialPorts; ++i)  {
    InitList(serPort[i].wrBlocked, blockedList);
    InitList(serPort[i].rdBlocked, blockedList);
  }
*/

  spiInUse = false;

//  serialFSysFuncs.tick = NULL;

  registerFileSys(&serialFSysFuncs);
}


//-----------------------------------------------------------------------------
//  Returns true if provided pointer is a valid, in-use Serial device.
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validSER(void *ioDev)
{
  if (ioDev)
    if (SER == serPort + (int)(SER - serPort))
      if (SER->inUse)  return true;

  return false;
}

//-----------------------------------------------------------------------------
//  Wait for the serial console buffer to empty.
//
//  NOTE this is called at startup before the timer interrupt is started so
//  we can't use things like GetMS() or time() yet.  Also, it temporarily
//  changes the IP register to ensure chars in the buffer CAN be sent, so
//  do NOT call it at times when interrupts must be disabled!
//-----------------------------------------------------------------------------
void flushConsole(void)
{
  int  charsLeft, lastCharsLeft, noProgressCount = 0;

  WDT_2S();  charsLeft = lastCharsLeft = SerialSendCountA();

  ipset1();

  while (charsLeft)  {
    if (charsLeft == lastCharsLeft)
      ++noProgressCount;
    else  {
      lastCharsLeft = charsLeft;  noProgressCount = 0;  }
    if (noProgressCount > 2000)  {
      printf("flushConsole() failed\r\n");  break;  }
    charsLeft = SerialSendCountA();
  }

  ipres();

}

//-----------------------------------------------------------------------------
void initConsole(void)
{
  SerialInitA( BR_115200, SER_8BITS, SER_IP2,
               serInBufs[0], SerInBufSize,
               serOutBufs[0], SerOutBufSize );
  _stdio = SerialA;                        //  redirect printf to SerialA

  printf("\r\n\n\n\n\n\n\n\n");  flushConsole();
}

//-----------------------------------------------------------------------------
//  check for file name indicating a serial port and attempt to open it
//-----------------------------------------------------------------------------
fd_t _serialOpen_(const char *path, int flags, int mode)
{
  char  *name;

  if (strncmp(path, "/dev/", 5))  { errno = ENXIO;  return -1; }

  name = (char *)path + 5;

  if (strncmp(name, "ttyx", 4) == 0)
    return  _serial_(atoi(name+4), 0, SER_RS232, SER_XON_XOFF);

  if (strncmp(name, "tty", 3) == 0)
    return  _serial_(atoi(name+3), 0, SER_RS232, SER_NO_FLOW);

  if (strncmp(name, "spi", 3) == 0)
    return  _serial_(atoi(name+3), 0, SER_SPI, SER_NO_FLOW);

  errno = ENXIO;  return -1;
}

//-----------------------------------------------------------------------------
//  func called each kernel tick to provide file system with CPU cycles
//  Should ONLY be called by the kernel task!
//-----------------------------------------------------------------------------
void _serialTick_(void)
{

}

//-----------------------------------------------------------------------------
//  Return the state of the specified ioDev entry or NULL if not a valid
//  entry #.
//
//  For valid entrys, the entryState struct is filled in
//-----------------------------------------------------------------------------
void *_serialEntry_(uint idx, EntryState *entryState)
{
  SerialInfo  *serDev;

  if (idx >= NumSerialPorts)  return NULL;

  serDev = serPort + idx;

  entryState->openCount = (serDev->inUse ? 1 : 0);
  entryState->unused = (entryState->openCount == 0);

  return serDev;
}

//-----------------------------------------------------------------------------
//  Placeholder, do-nothing unlink function
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _serialUnlink_(char *path)
{
  return -1;
}


//-----------------------------------------------------------------------------
//  Placeholder, do-nothing rename function
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _serialRename_(const char *oldPath, const char *newPath)
{
  return -1;
}



//-----------------------------------------------------------------------------
//  Kernel-space function to initialize a serial port and return a file
//  descriptor for it
//-----------------------------------------------------------------------------
fd_t _serial_(uchar port, uchar rateIdx, uchar mode, uchar flowCtl)
{
  int  stat, hdwMode;
  fd_t  fd;
  SerialInfo  *si;


  if (port >= NumSerialPorts)  { ERRNO = ENOENT;  return -1; }
  if (! (mode & serModes[port]))  { ERRNO = ENOENT;  return -1; }

  KLOCK();

  if (serPort[port].inUse)  { KUNLOCK();  ERRNO = EBUSY;  return -1; }

//--- initialize the specified serial port ---
  si = serPort + port;

  si->mode = mode;
  si->port = port;
  si->flowCtl = flowCtl;
  si->closed = si->pktMode = false;
  si->inUse = true;  // needs to be true during rest of init

  switch (mode)  {
    case SER_RS232:
    case SER_RS485:
      si->baudRate = 115200UL;  //finish
      if (port == 0)
        stat = 1;
      else  {
        hdwMode = SER_8BITS;
        if ((hdwType == HT_BL2105) AND (port == 1))  hdwMode |= SER_USE_D;
        stat =  (serFuncs[port].init)(si->baudRate, hdwMode, SER_IP2,
                                      serInBufs[port], SerInBufSize,
                                      serOutBufs[port], SerOutBufSize, flowCtl);
      }
      if (! stat)  {
        si->inUse = false;  KUNLOCK();  ERRNO = EINVAL;  return -1; }
      if ((fd = _allocFile_(3, &serialIOFuncs, si)) < 0)  {
        si->inUse = false;  KUNLOCK();  return -1;  }
      break;

    case SER_SPI:
      //finish - use rateIdx?
      if (! _spiInit_(si, CLK_1MHz, NORM_LOW_LATCH_RISE))  {
        si->inUse = false;  KUNLOCK();  ERRNO = EINVAL;  return -1; }
      if ((fd = _allocFile_(3, &spiIOFuncs, si)) < 0)  {
        si->inUse = false;  KUNLOCK();  return -1;  }
      break;

    default:
      si->inUse = false;  KUNLOCK();  return -1;
  }

  KUNLOCK();

//--- everything OK - return the descriptor assigned for this process ---
  return fd;
}


//=============================================================================
//  IODev interface functions (for use by fcntl module)
//=============================================================================

//-----------------------------------------------------------------------------
int _serialGetQueues_(void *ioDev, WaitQueues far *queues)
{
  queues->readQueue = queues->writeQueue = &sleepQueue;
  return 0;
}

//-----------------------------------------------------------------------------
//  Attempt to read bytes from a serial port
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _serialRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  int  i, ch;
  char  *ptr;
  GetCFunc  getcFunc;


  KLOCK();

  if (! validSER(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (SER->closed)  { KUNLOCK();  ERRNO = EPIPE;  return -1; }

  getcFunc = SDEV->serialGetc;  ptr = buf;

  for (i=0; i<numBytes; )  {
    ch = (*getcFunc)();  if (ch == -1)  break;
    *ptr++ = (uchar) ch;  ++i;
    if (_pending)  break;
  }

  KUNLOCK();

  if (!i)  { ERRNO = EAGAIN;  return -1; }

  return i;
}

//-----------------------------------------------------------------------------
//  Attempt to write bytes to serial port
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _serialWrite_(void *ioDev, const char *buf, int numBytes, bool dontBlock)
{
  int  i;
  bool  pktMode;
  const char  *ptr;
  PutCFunc  putcFunc;


  KLOCK();

  if (! validSER(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (SER->closed)  { KUNLOCK();  ERRNO = EPIPE;  return -1; }

  putcFunc = SDEV->serialPutc;  ptr = buf;  pktMode = SER->pktMode;

  for (i=0; i<numBytes; )  {
    if (((*putcFunc)(*ptr++)) < 0)  break;
    ++i;
    if (_pending AND !pktMode) break;
  }

  KUNLOCK();

  if (!i)  { ERRNO = EAGAIN;  return -1; }

  return i;
}


//-----------------------------------------------------------------------------
//  release a serial port for use by another process
//-----------------------------------------------------------------------------
int _serialClose_(void *ioDev)
{
  KLOCK();

  if (! validSER(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  if (SER->closed)  { KUNLOCK();  ERRNO = EPIPE;  return -1; }

  SER->inUse = false;

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to reposition a file offset
//
//  WARNING!  To be called ONLY by KERNELFUNC functions (or functions called BY
//            KERNELFUNC functions) !!!!
//-----------------------------------------------------------------------------
off_t  _serialLseek_(void *ioDev, off_t offset, int whence)
{
  KLOCK();

  if (! validSER(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  KUNLOCK();

  ERRNO = ESPIPE;  return (off_t)-1;
}


//-----------------------------------------------------------------------------
//  Get information about ...
//-----------------------------------------------------------------------------
int  _serialFstat_(void *ioDev, struct stat *fsbuf)
{
  KLOCK();

  if (! validSER(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  fsbuf->st_mode = S_IFCHR | 00777;

  KUNLOCK();

  return 0;
}

extern void setExt485Tx(void);
//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for serial ports.
//
//  Notes on Packet-mode interfaces:
//
//    The packet mode of operation provides a means to couple the control
//    of the hardware interface for a serial bus (e.g. RS-485) to the timing
//    of char transmission managed by the serial library.  This allows the
//    serial library to meet the timing requirements of most serial
//    interfaces.
//
//    This coupling is managed as follows:
//
//      - Set the packet mode for a serial connection (IOC_SET_PKTMODE).
//        The argument given for this request is the char that signals the
//        start of a received packet.  If non-zero, incoming data is ignored
//        until the specified char is received.  If zero, all incoming data
//        is saved.  For RS-485 connections, this character is often
//        used to signify a node's "address".
//
//      - To transmit a packet, do the following:
//          - Set IOC_START_PKT_TX mode
//          - Write a packet of bytes to the device
//          - Set IOC_END_PKT_TX mode
//
//      - To receive a packet, read characters from the device until an
//        entire packet is received, a timeout or other error occurs, etc,
//        and then set IOC_END_PKT_RX mode.
//
//    The purpose of the IOC_START_PKT_TX and IOC_END_PKT_TX modes are to
//    get the serial lib to invoke a platform-specific callback function
//    at the start and end of the transmission of the group of bytes that
//    make up a packet.  The actual callback functions (one per port) are
//    provided by the code for the specific hardware platform (the BSP, or
//    Board Support Package, found under the "arch" branch).
//
//    If told to do so (using the above sequence), the serial lib ISR that
//    handles transmitting characters will call the board-specific function
//    just before transmitting the first character and just after sending
//    the last character.  The function has one argument indicating if the
//    interface should be set to xmit or rcv mode.
//
//    The callback function is usually (bot not always) called directly from
//    the ISR, so it must be fast, and it must save and restore any registers
//    it uses.  Calling it from the ISR is what allows it to be called with
//    very little latency (i.e. not so long before or after the bytes are
//    transmitted that we risk losing data received).
//
//    Setting the IOC_END_PKT_RX mode for the device tells it that it should
//    begin looking for the start of a new packet again.  If the start char
//    given for packets was 0, this has no effect.  But for non-zero start
//    characters, this allows the library to ignore received characters until
//    the start of a new packet, greatly reducing the number of CPU cycles used
//    for incoming characters prior to the start char (i.e. the ISR doesn't
//    have to save them in the queue, and the application code can simply
//    block waiting for the next packet or an error, rather than having read
//    each character back out of the queue and check for the start character).
//-----------------------------------------------------------------------------
int _serialioctl_(void *ioDev, int request, void *arg)
{
  int  stat;
  uint  flowChars, used;
  SerialFuncs  *funcs;

#define CHAR_ARG   (*((char *)arg))
#define INT_ARG    (*((int *)arg))
#define UINT_ARG   (*((uint *)arg))
#define ULONG_ARG  (*((ulong *)arg))


  KLOCK();

  if (! validSER(ioDev))  { KUNLOCK();  ERRNO = EBADF;  return -1; }

  funcs = serFuncs + SER->port;  stat = -1;

  switch (request)  {

    case FIONREAD:
      _EnterCritical_();  INT_ARG = (funcs->rcvAvail)();  _ExitCritical_();
      stat = 0;  break;

    case FIONSPACE:  // return # of free bytes in the xmit buffer
      _EnterCritical_();  used = (funcs->xmitBufUsed)();  _ExitCritical_();
      INT_ARG = SerOutBufSize - used;  stat = 0;  break;

    case IOC_GET_BAUD:
      ULONG_ARG = SER->baudRate;   stat = 0;  break;

    case IOC_SET_BAUD:
      SER->baudRate = ULONG_ARG;
      SerialSetRate(serPortRef[SER->port], SER->baudRate);
      stat = 0;  break;

  //--- Enable RS845 packet-mode on the port.  The argument is the 8-bit ---
  //--- "address" that we want to receive packets for.  Incoming bytes are ---
  //--- ignored until the specified char is received.  Once the specified ---
  //--- char is received, the port enters receive mode and all subsequent ---
  //--- bytes are saved in the input buffer until the application calls ---
  //--- ioctl() with the IOC_END_PKT_RX request ---
    case IOC_SET_PKTMODE:
      (funcs->setPktMode)(CHAR_ARG, 0, funcs->setDir);
      SER->pktMode = true;  stat = 0;  break;

  //--- Tells the serial port to stop saving bytes for the current packet. ---
  //--- The application should read bytes as they come in looking for the ---
  //--- end of a packet.  When it finds the end, or determines there is a ---
  //--- problem with the packet, it should invoke this function to tell the ---
  //--- lib to stop saving bytes and go back to waiting for the char that ---
  //--- signals the start of a new packet (our "address") ---
    case IOC_END_PKT_RX:
      (funcs->endPktRX)();  stat = 0;  break;

  //--- Prepare for the start of the transmission of a "packet".  This ---
  //--- tells the serial lib that, just before it transmits the next char, ---
  //--- it must call the setDir() func given in the IOC_SET_PKTMODE request ---
  //--- to enable the hardware to xmit chars on the RS-485 bus.  This is ---
  //--- done by the ISR that handles transmitting chars in a timely fashion ---
  //--- so the func must be fast and protect any registers it uses. ---
  //--- The
  //--- must do whatever is required to configure the hdw for xmit mode ---
    case IOC_START_PKT_TX:
      setExt485Tx();  // required to be here due to BL2105 I/O timing issues
      (funcs->flushRcv)();
      (funcs->startPktTX)(0,0);  stat = 0;  break;

  //--- Tells the serial lib that, when the last byte of the current packet ---
  //--- has been transmitted, it should call the setDir() func given in ---
  //--- the IOC_SET_PKTMODE call (with a value of 0).  The called func is ---
  //--- responsible for configuring the hdw so new packets can be rcvd ---
    case IOC_END_PKT_TX:
      (funcs->endPktTX)();  stat = 0;  break;

    case IOC_SET_ASYNC_FMT:
      switch (UINT_ARG)  {
        case ASYNC_FMT_7O1:  (funcs->set7O1)();  stat = 0;  break;
        case ASYNC_FMT_7E1:  (funcs->set7E1)();  stat = 0;  break;
        case ASYNC_FMT_8O1:  (funcs->set8O1)();  stat = 0;  break;
        case ASYNC_FMT_8E1:  (funcs->set8E1)();  stat = 0;  break;
      }
      break;

    case IOC_SET_FLOW_CHARS:
      flowChars = (UINT_ARG);
printf("Before: XON: [%02X]  XOFF: [%02X]\r\n", SDEV->xOn, SDEV->xOff);  //tdebug
      _EnterCritical_();
      SDEV->xOn = (uchar)(flowChars >> 8);
      SDEV->xOff = (uchar)flowChars;
      _ExitCritical_();
printf("After:  XON: [%02X]  XOFF: [%02X]\r\n", SDEV->xOn, SDEV->xOff);  //tdebug
      stat = 0;  break;

    default:
      ERRNO = EINVAL;
  }

  KUNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to return a description of the serial port
//
//  WARNING!  To be called ONLY by KERNELFUNC functions (or functions called BY
//            KERNELFUNC functions) !!!!
//-----------------------------------------------------------------------------
char *_serialDescr_(void *ioDev, char *buf, int bufSize)
{
  char  *ptr, *suffix;


  sprintf(buf, "serial 0x%04lX ", (ulong)ioDev);

  ptr = buf + strlen(buf);  strcpy(ptr, "bad bad");

//finish: limit output to bufSize

  suffix = "";
  switch (SER->flowCtl)  {
    case SER_XON_XOFF:  suffix = "x";  break;
    case SER_RTS_CTS:   suffix = "r";  break;
  }

  if (validSER(ioDev))
    sprintf(ptr, "ser[%d] /dev/tty%s%d %ld baud",
           (uint)(SER - serPort), suffix, SER->port, SER->baudRate);

  return buf;
}


//-----------------------------------------------------------------------------
void _serialDumpObject_(uint port)
{
  _serialDevice  *sd;


  sd = (_serialDevice *) serPort[port].sDev;

  printf("Port: %u at [%04lx], dev at [%04lx]\r\n",
         port, (ulong)(serPort + port), (ulong)sd );

  printf("  dPort: %02x\r\n", (uint)sd->dPort);
  printf("  sPort: %02x\r\n", (uint)sd->dPort);
  printf("  InQueue:\r\n");
  printf("    in:    [%04x]\r\n", (uint)sd->serialInQueue.q_in    );
  printf("    out:   [%04x]\r\n", (uint)sd->serialInQueue.q_out   );
  printf("    start: [%04x]\r\n", (uint)sd->serialInQueue.q_start );
  printf("    end:   [%04x]\r\n", (uint)sd->serialInQueue.q_end   );
  printf("    start_int: [%04x]\r\n", (uint)sd->serialInQueue.q_start_int );
  printf("    end_int:   [%04x]\r\n", (uint)sd->serialInQueue.q_end_int );
  printf("    flags: [%04x]\r\n", (uint)sd->serialInQueue.q_flags   );
  printf("  OutQueue:\r\n");
  printf("    in:    [%04x]\r\n", (uint)sd->serialOutQueue.q_in    );
  printf("    out:   [%04x]\r\n", (uint)sd->serialOutQueue.q_out   );
  printf("    start: [%04x]\r\n", (uint)sd->serialOutQueue.q_start );
  printf("    end:   [%04x]\r\n", (uint)sd->serialOutQueue.q_end   );
  printf("    start_int: [%04x]\r\n", (uint)sd->serialOutQueue.q_start_int );
  printf("    end_int:   [%04x]\r\n", (uint)sd->serialOutQueue.q_end_int );
  printf("    flags: [%04x]\r\n", (uint)sd->serialOutQueue.q_flags   );
  printf("  getc(): [%04x]\r\n", (uint)sd->serialGetc);
  printf("  putc(): [%04x]\r\n", (uint)sd->serialPutc);
}

//-----------------------------------------------------------------------------

