//----- socket.c ----- 07/31/12 -----(03/12/06)-----

//-----------------------------------------------------------------------------
//  NOTE:  See the _README_.txt file for changes that need to be made to the
//         Softools/ZWorld Stcpip source files .
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//  Standard DEBROS virtual file/device I/O API for reading and writing to
//  sockets
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2012-07-31 - Mark Davis (*)
//               - An ENETDOWN error is now only asserted for functions that
//                 return it if the link was up then went down for at least
//                 30 seconds.
//
//  2012-06-20 - Mark Davis (*)
//               - _socketRead_(), _socketWrite(), send(), sendto(), recv(),
//                 and recvfrom() now return an error if the network lib
//                 functions don't even after the link does down for the
//                 interface the socket is attached to.  Prior to this,
//                 attempting to read (and write?) to a socket after the
//                 link goes down did not return an error even if its write
//                 buffer was full (or its read buffer empty).
//
//  2012-06-06 - Mark Davis
//               - When a new connection is established for a listening socket,
//                 the socket that replaces the listening one now uses the
//                 local addresses originally set for the listening socket,
//                 rather than the one for the local interface the new
//                 connection was established on.
//               - TCP Listening sockets are now "Virtual".  This means they
//                 do not use a tcp_Socket structure or the related buffers
//                 until they are actually connected. This avoids the need to
//                 tie up a tcp_Socket structure and its I/O buffers until the
//                 socket is actually connected by a successful call to
//                 accept() or connect().  This frees up several sockets that
//                 were tied up as listeners.
//               - Changed SockInfo.closeCalled to noFDEVrefs to clarify the
//                 meaning of this flag.
//
//  2012-04-20 - Mark Davis
//               - Added tcpCounts() and udpCounts() to get total and free
//                 counts for the global pool of sockets.
//
//  2012-02-24 - Mark Davis
//               - Changed SockInfo.state to conState and expanded its use to
//                 better track the state of a connection.
//               - sendto() now checks the connection state of TCP sockets and
//                 whether an address is given or not and returns appropriate
//                 error states as needed.
//               - _socketWrite_() now checks the connection state of TCP
//                  sockets and returns an error if the socket is not connected.
//               - accept() now sets the conState when the connection is
//                 readable
//               - Reduced the # of UDP socket buffers by 1.  This was made
//                 possible by eliminating some unneeded items from the
//                 Softools/ZWorld socket library that were not needed (but
//                 were still being linked in).
//               - The broadcast address returned by getIFConfig() is now
//                 based on the netmask and IP address (previously it simply
//                 returned 0xffffffff).
//
//  2012-02-08 - Mark Davis
//               - errno is now set to EBADF rather than ENXIO when a function
//                 is passed an invalid device pointer.
//
//  2012-02-07 - Mark Davis
//               - The broadcast addr for an interface is now computed based
//                 on the address and the netmask.
//
//  2012-01-25 - Mark Davis
//               - Minor mod to simplify if condition in sendto()
//
//  2011-05-02 - Mark Davis (*)
//               - Added a reset of the watchdog timer to startUpIF().  Prior
//                 to this the updated startup sequence and non-blocking
//                 handling of starting up and taking down the network could
//                 result in the watchdog timer expiring during startup.
//
//  2011-03-30 - Mark Davis (*)
//               - Fixed bug in socket() that caused it to use the wrong memory
//                 address when allocating any UDP connection but the 1st one.
//                 Prior to this it would use the correct addr for the 1st one,
//                 the addr of the 3rd one for the 2nd one, and addresses past
//                 the end of the array for the 3rd and 4th ones (so basically
//                 anything but the 1st one could cause problems).
//               - _fdSocket_() now returns EBADF rather than ENOTSOCK if the
//                 fd references a SockInfo structure but it is not currently
//                 in use.
//
//  2011-03-25 - Mark Davis (*)
//               - Fixed diagnostic code in _socketClose_() so it doesn't
//                 try to dump info on an empty wait queue (which ends up
//                 causing a watchdog timeout when it gets stuck in a loop!)
//               - validSOCK() now returns false for invalid pointers (it was
//                 always returning true!)
//               - Added closeCalled to avoid freeing up a SockInfo struct when
//                 it is still in use by an FDEV.  This can happen when socket
//                 connections are closed by the kernel (e.g. in preperation
//                 for shutting down the network interface to change its
//                 address or to reboot).
//
//  2011-01-31 - Mark Davis
//               - When enabled, the syslog diagnostic messages generated by
//                 tcpSockEvent_() now includes any errors for codes stored in
//                 the network library's socket structure.
//
//  2011-01-26 - Mark Davis
//               - getIFConfig() and setIFConfig() now use the new gateway
//                 value in the IFConfig structure.
//
//  2010-09-09 - Mark Davis (*)
//               - Changed validSOCK() to use the new INVALID() macro to better
//                 check for invalid pointers, and to produce log msgs when
//                 an attempt to use one occurs.
//               - _socketTick_() now does its processing each time it is
//                 called if a shutdown is in process.  This helps to fully
//                 close and cleanup the socket connections during a controlled
//                 shutdown.
//
//  2010-07-07 - Mark Davis (*)
//               - tcpSockEvent() now checks to make sure the passed socket
//                 is a reference to a valid, in-use socket.
//               - _socketRead_() and _socketWrite_() now make sure the
//                 passed device pointer is valid before setting the device's
//                 diagFlags value.
//
//  2010-05-21 - Mark Davis (*)
//               - When a socket connection is reset, the rdOK flag is now set
//                 based on whether there are any unread chars remaining, and
//                 any tasks in the read queue are woken up.
//               - The ioctl call now returns -1 with errno = ENOLINK for the
//                 FIONREAD/FIONSPACE requests if rdOK/wrOK is false.
//
//  2010-05-14 - Mark Davis
//               - Commented out change for 05-11 as it did not change anything
//                 and it just adds overhead.
//               - ioctl function now supports FIONSPACE request
//
//  2010-05-12 - Mark Davis (*)
//               - accept() now checks to make sure the socket isn't closed
//
//  2010-05-11 - Mark Davis (*)
//               - Added call to sock_tbleft() in _socketWrite_() function
//                 to see if I could eliminate the problem where the
//                 sock_fastwrite() func appears to accept many writes but
//                 not actually deliver the packets.  This was showing up in
//                 the scope command which would have its output "freeze"
//                 even though the lsof command (which shows the writeCount for
//                 the FDEV device) showed that data continued to be written
//                 to the socket.
//
//  2010-05-02 - Mark Davis (*)
//               - Added locks to _socketRead_() and _socketWrite_() functions
//                 in case the locking calls done within the socket library are
//                 not sufficient to safeguard the socket structures.
//
//  2010-04-26 - Mark Davis
//               - Changes to accomodate more capable wait queue handling
//
//  2010-04-09 - Mark Davis (*)
//               - Moved the call to sock_flush() in _socketClose_() to within
//                 KLOCK()/KUNLOCK() sequence.  Without this, it is possible
//                 for the close call to be preempted after the KUNLOCK() and
//                 the socket and socket table entry to be in various states
//                 between being closed and open and in use by another task
//                 when _socketClose_() resumes and calls sock_flush().
//               - Added check to _socketClose_() to make sure no tasks are
//                 still in the wait queue.
//               - Added lastClosed to SockInfo and use of it to make sure when
//                 a socket entry is allocated it is the least recently used
//                 of the ones available (reducing the risk that some bug in
//                 the socket library is causing a problem when reusing a
//                 socket that it claims is closed).
//
//  2010-03-04 - Mark Davis
//               - recv(), recvfrom(), send(), and sendto() now all take a
//                 far pointer for the buffer.
//
//  2010-03-02 - Mark Davis
//               - recv(), recvfrom(), send(), and sendto() now all update
//                 the readCount/writeCount value in the FDEV structure.
//               - sendto() now uses the port number (if provided) when sending
//                 UDP broadcast packets.
//
//  2010-01-12 - Mark Davis (*)
//               - Added breaks in switch statement and locking to
//                 checkSockets()
//
//  2009-12-10 - Mark Davis
//               - Removed call to DLM_Tick() at least for now. It is one
//                 more possible cause of some unexplained instabilities and
//                 mostly just another opportunity for human error to cause
//                 problems.  With this commented out, the device will not
//                 show up in the SHD Update Manager when not running the DLM
//                 so there is no chance for someone to accidentially click
//                 on the wrong device and mess up a live controller.
//
//  2009-11-05 - Mark Davis
//               - The configd task's announcementInterval is now reset
//                 whenever the Ethernet link comes up.
//
//  2009-10-01 - Mark Davis
//               - _socketTick_() now calls tcp_tick() each time but only
//                 does the other "housekeeping" stuff (like managing closed
//                 sockets) about every 50ms.  This adds a bit of overhead
//                 but reduces latency and improves response time for tasks
//                 that use socket connections (like modbusTCPd).
//
//  2009-09-30 - Mark Davis
//               - _socketTick_() now updates netReady.  This eliminates
//                 errors claiming the Network is down if it was down when
//                 it was first initialized but came up after initialization.
//               - _socketTick_() also checks the physical layer link status
//                 and logs a message when the status changes.
//
//  2009-09-29 - Mark Davis
//               - _socketTick_() now uses time rather than ticks to determine
//                 when to skip processing.  This improves response time for
//                 slower tick rates, although it means the overhead of the
//                 network processing does not change when the tick rate is
//                 lowered.
//
//  2009-08-26 - Mark Davis
//               - Removed definition of references to noRefs attribute which
//                 apparently was only set to false in the close call (i.e.
//                 it had no affect).
//
//  2009-08-25 - Mark Davis
//               -_socketClose_() now calls sock_flush() for TCP socket types
//                to try to insure the data from any recent writes actually
//                make it to the other end before the socket is reused.
//
//  2009-08-05 - Mark Davis
//               - Moved responsiblity for not messing up an fd's flags
//                 settings from _newListener_() to _File_Swap_() (in the
//                 fcntl.c module).  _newListener_() was actually
//                 just clearing the O_NONBLOCK bit in flags for the new
//                 socket fd, which did not leave the original fd with the
//                 correct settings.
//
//  2009-08-04 - Mark Davis
//               - removed handling of blocking mode from accept() for now
//                 as this can cause problems until I move task-specific
//                 descriptor settings out of the global descriptors and
//                 in to the task-specific ones.
//
//  2009-07-31 - Mark Davis
//               - setsockopt() and getsockopt() now check for a NULL pointer
//                 for the val parameter.
//               - getsockopt() now checks the value referenced by valLen
//                 to insure the buffer pointed to by val is large enough.
//               - listen() now returns an error if the socket type is not
//                 SOCK_STREAM.
//
//  2009-07-30 - Mark Davis
//               - setsockopt() now pays attention to the val passed so the
//                 broadcast option can be set or reset
//               - getsockopt() now allows the state of the broadcast option
//                 to be read
//               - Functions that can return a new socket now return an error
//                 if the network is not ready or is going down
//               - accept() now handles blocking mode
//
//  2009-06-23 - Mark Davis
//               - Changed WDT_2S() calls to resetWDT()
//
//  2009-06-19 - Mark Davis
//               - Modified upInterface() so it doesn't cause a reboot if
//                 it takes too long (a recently introduced problem related
//                 to the changes to dump diagnostic data when the watchdog
//                 timer is about to expire and cause a reboot).
//
//  2009-05-01 - Mark Davis
//               - set skipTicks to false at end of socketInit_() so the
//                 _socketTick_() function will actually do something
//
//  2009-03-26 - Mark Davis
//               - Added placeholder unlink and rename functions
//
//  2009-03-24 - Mark Davis
//               - added _socketEntry_() function for getting info about socket
//                 fds
//
//  2009-02-20 - Mark Davis
//               - Modified recvfrom() so it doesn't do the work of getting
//                 the address and port of the sender if the fromAddr is NULL.
//                 This should eliminate lots of ARP traffic when receiving
//                 broadcast messages.
//
//  2006-03-12 - Mark Davis
//               - began work on new module
//
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
//  Overview of basics of creating/using socket fd's:
//
//   - Call socket() to create a socket and get a fd for it
//
//   - To open a connection to a specific port+addr:
//        - Call connect()
//
//   - To listen for and accept incoming connections or packets:
//
//        - Call bind() to specify the local port and which interface(s)
//          (indirectly) by specifying a local addr (or INADDR_ANY for all
//          interfaces)
//
//        For TCP (stream-based) sockets:
//          - Call listen() to mark the socket as a passive listener, create
//            a queue for incoming connection requests, and begin accepting
//            connections on the port+addr/interface(s) specified by bind()
//
//          - Call accept() to create a new fd for a new open socket connection
//            for the first pending connection in the queue for the listening
//            socket.
//
//        For UDP (packet-based message) sockets:
//          - Call recv() or recvfrom() to get the next packet sent to the
//            port+addr specified in the call to bind()
//
//-----------------------------------------------------------------------------


/*-----------------------------------------------------------------------------
NOTES on DynamicC network stack code (came with Softools WinIDE):

Functions that call tcp_tick():
   tcp_open(), udp_open()
   sock_read(), sock_write()
   sock_close(), sock_abort()
   All blocking functions

   In general, non-blocking functions do NOT call tcp_tick()


To manually associate a buffer with a socket:
   tcp_extlisten(),  tcp_extopen(), udp_extopen()
   sock_xfastread(), sock_xfastwrite(), udp_xsendto()


Reserved listening ports:

   (include/Stcpip/tcp.h)
      extern word tcp_reserveports[];
      extern const char MAX_RESERVEPORTS;

   #define USE_RESERVEDPORTS        (inet.h)

   call to tcp_initreserve()        (tcp_02.c)

   tcp_reserveport(port)            (tcp_38.c)
      tcp_reserveports[]

   tcp_initreserve()                (tcp_39.c)

   tcp_clearreserve()               (tcp_40.c)

   void tcp_handler(in_Header *ip)  (tcp_41.c)

------------------------------------------------------------------------------

TO-DO:

  - Would it help at all to switch MAX_ETH_BUFS and ETH_MTU back to their
    defaults of 10 and 600 (rather than the 4 and 1500 Softools sets them to)?
    While the smaller MTU might be a problem, having MORE buffers might
    improve throughput.  (And what am I using for the MAX_ETH_BUFS?)


  - Is there a logical place/circumstance that I could call sock_flush() to
    avoid any possible delay in sending data when I know I won't be writing
    to a socket for a bit?  Perhaps where I currently check # of unused
    bytes in the xmit buffer I should call sock_flush() (and then tcp_tick(),
    if sock_flush() doesn't already?)


  - Section 3.8.2: "Incoming packets are process whenever tcp_tick() is called."
    What about outgoing packets?  Particularly when the nagle algorithm is
    enabled (which it is by default)?


  - The send(), sendto(), recv(), and recvfrom() funcs don't currently handle
    blocking mode (they behave as if the fd has non-blocking mode set).


  - I don't believe blocking mode currently works at ALL for UDP sockets!
    Waking up a task blocked on a read from or write to a TCP socket is handled
    by the event callback mechanism for TCP sockets, but I don't yet use the
    callback mechanism for UDP sockets.  At the moment, doing a blocking-mode
    read() or write() on an fd for a UDP socket would probably block the
    caller forever!

    NOTE:  For UDP sockets, you get a UDP_DH_INDATA event whenever a new,
    unread packet has arrived.  There is no equivalent to TCP_DH_OUTBUF to
    wake up a task blocked waiting to write to a UDP socket.

    Two possibilities come to mind to handle this:  Either use the default
    method for drivers that don't handle the wait queues on their own (which
    is to use the sleep wait queue), or have the _socketTick_() function
    check the state of any UDP sockets with blocked tasks and wake them up
    as appropriate.

    The later is certainly the preferred method, as it means that tasks are
    woken up only when they need to be.  When it comes right down to it, the
    callback mechanism for TCP sockets really isn't any different.  It can
    only do the callbacks during a call to tcp_tick(), which is done by
    _socketTick_(), so all we are doing is adding the same functionality for
    UDP packets.

    In fact, this way the event checking is limited to sockets that need it,
    whereas the current method for TCP sockets does the callback regardless of
    whether or not there are tasks blocked waiting for the event.  But I
    suppose those callbacks are done as just an extension of the processing
    that has to happen for the TCP sockets in any case, so it really isn't
    adding all that much to the processing to perform the callbacks.

    And it IS handy for debugging, since I can see all the events for a
    socket (at least TCP ones).


  - Add support for "overflow" buffers.  When the network lib xmit buffer for
    a TCP socket is full, then assign an overflow buffer to the socket and
    copy the data there.  Then the TCP_DH_OUTBUF event in tcpSockEvent_() can
    call sock_xfastwrite() to move data from the overflow buffer to the
    network lib's buffer.

    The idea is that this would increase the max throughput rate for a
    socket for the following reasons:

      - Without overflow buffers, a call to write to a socket when the
        network lib's internal xmit buffer for the socket is nearly full
        will result in the task being blocked (or returning with very
        little or none of the data being processed).

        Then, during a call to tcp_tick() (during the kernel's periodic
        background processing), when the network lib has emptied its internal
        xmit buffer for a socket, it will not have any more data to send for
        the socket until the following occurs:
          - tcp_tick() returns
          - All the rest of the kernel's background processing finishes
          - The task that owns the socket is resumed and has time to
            write more data to the socket before being preempted.
          - The kernel executes again and does it background processing
            (which includes a call to tcp_tick())

      - With overflow buffers, a call to write to a socket when the
        network lib's internal xmit buffer for the socket is nearly full
        will result in at least some of the extra data being copied to an
        overflow buffer before the task is blocked (or returning to the task).

        The rest scenario is the same with one exception:  When tcpSockEvent_()
        is called for the TCP_DH_OUTBUF event, it can refill the network lib's
        internal buffer by moving data from the overflow buffer.  Assuming the
        call to tcp_tick() will process this new data after tcpSockEvent()
        returns, this should greatly decrease the time (and most likely the #
        of CPU cycles) required to send the same amount of data.

        Of course, this assumes that the overhead of the additional logic
        for the overflow buffers is less than going through the whole
        cycle shown above.

        A related question:  Will tcp_tick() call tcpSockEvent_() multiple
        times for the same socket when applicable?

        The one (possibly significant) downside:  tcp_tick() calls will run
        longer, which will increase the kernel's CPU usage, which will
        decrease the number of cycles available each scheduler tick for other
        tasks.

        Of course, that is a generic issue that applies to any process you
        make more efficient by doing the processing in larger chunks:  While
        it may take fewer CPU cycles per unit of work (in this case, the # of
        bytes transmitted), if you make use of the increased efficiency to
        get more done per unit of time, then the # of CPU cycles used per
        unit of time is the same or even more.

        So the tradeoff is the same as for choosing the interval for the
        periodic scheduler tick:

          Shorter intervals:
             More task switches per second decreases latency, but lower
             overall efficiency due to more time spent switching tasks.

          Longer intervals:
             Fewer task switches per second increases latency, but higher
             overall efficient due to less time spent switching tasks.

        Adding overflow buffering is just this same concept applied to
        socket connections:  Each call to tcp_tick() can do more, but will
        require more CPU time to do it.  The result can be 1 of 3 things:

          - The rate of sending data is not allowed to increase, so the
            increased efficiency means the CPU usage of the tasks making
            use of the overflow buffering decreases (by more than the
            averagae increase for tcp_tick() calls), so there are more
            CPU cycles available for other tasks.

          - The rate of sending data is allowed to increase but only to
            the point where the combination of the CPU usage of tcp_tick()
            and the sending tasks are the same as they were before.  The
            result is higher throughput with the same CPU usage (although
            the increase in tcp_tick()'s usage will affect latency a bit
            for other tasks).

          - The rate of sending data is allowed to increase to the new
            maximum.  The CPU usage of tcp_tick() will be higher, and the
            CPU usage of the sending tasks will probably also increase a
            bit (since they will being doing a bit more work before they
            block, making use of the overflow buffers).  The result will
            be an increase in CPU usage, but also (hopefully) a significant
            increase in throughput for the socket connections.


         But all of this is speculation; An educated guess.  If the call to
         tcp_tick() does NOT process new data put in to the xmit buffers by
         a call to the event handler, then overflow buffering probably won't
         improve throughput much.  After the first call, the # of bytes
         processed for each call will be (on average) the same as it was
         without the overflow buffering.  A signficant gain requires that the
         average per call goes UP by an amount roughly equivalent to the
         amount of overflow buffer used.

         And keep in mind that we are also adding ANOTHER buffer-to-buffer
         copy for each use of the overflow buffer.  If data added to the
         network lib's xmit buffer for a socket by tcpSockEvent_() is NOT
         processed before tcp_tick returns, then this is mostly just wasted
         additional overhead.  There MIGHT be a small gain by virtue of the
         fact that the library's internal buffer will more often be full at
         the start of a tcp_tick() call, which would result in fewer/larger
         packets on average.  But I doubt any such gain would be as large
         as from processing it ASAP.


-----------------------------------------------------------------------------*/


#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include <errno.h>

#include <string.h>

#include <busybox.h>

//#include <sys/RegVars.h>
#include <SHD_DLM.h>

#include <sys/configd.h>

#include <sys/farprintf.h>




#define ConnectTimeout  10000  // max time (in ms) for connect() to finish



const char MAX_RESERVEPORTS = MaxListenTCP;
word tcp_reserveports[MaxListenTCP];


const unsigned  TCP_BUF_SIZE  = 4096;
const unsigned  UDP_BUF_SIZE  = 2048;

const char MAX_TCP_SOCKET_BUFFERS = MaxTCPCons;
const char MAX_UDP_SOCKET_BUFFERS = MaxUDPSockets + 1;

//--- Rabbit TCPIP lib structures for keeping track of each socket ---
tcp_Socket tcpSockDef[MaxTCPCons];
udp_Socket udpSockDef[MaxUDPSockets];


//--- DEBROS structures for keeping track of the state of each socket ---
SockInfo  tcpSockets[MaxTCPSockets];
SockInfo  udpSockets[MaxUDPSockets];

//--- index in to list for a given SockInfo structure ---
#define TCP_SI_Idx(ptr)  ( (uint)((ptr) - tcpSockets) )
#define UDP_SI_Idx(ptr)  ( (uint)((ptr) - udpSockets) )

//--- get ptr to TCP/UDP SockInfo object from a ptr to a tcp_/udp_Socket ---
#define TCP_SD_To_SI(sp) ( (SockInfo *) ((tcp_Socket *)(sp))->user_data )
#define UDP_SD_To_SI(sp) ( udpSockets + (uint)((udp_Socket *)(sp)-udpSockDef) )



IODEV_FUNCS(socket)

FILESYS_FUNCS(socket)

#define SI  ((SockInfo *)ioDev)


#define SockClosed(sock)    \
        ( (sock_alive(sock) == 0) OR (strcmp(sockstate(sock), "Closed") == 0) )

//-----------------------------------------------------------------------------
bool  skipTicks, netReady, linkWentDown, netGoingDown, closingSockets;
ulong  startedDownMS, linkDownMS;



//*********************************************************************
//***** Must match list found in stcpip/neterrno tcpip lib source *****
//*********************************************************************

#define NETERR_NONE                 0
#define NETERR_NOHOST_ARP           1     // Local host not reachable (ARP could not resolve)
#define NETERR_NOHOST_RTE           2     // Host not reachable (Router not resolved)
#define NETERR_HOST_REFUSED         3     // Host refused connection
#define NETERR_OPEN_TIMEOUT         4     // Timout on open or close
#define NETERR_CONN_TIMEOUT         5     // Connection timed out (keepalive etc.)
#define NETERR_ABORT                6     // Active abort sent from this peer
#define NETERR_INACTIVE_TIMEOUT     7     // Timed out due to inactivity
#define NETERR_LEASE_EXPIRED        8     // DHCP lease expired
#define NETERR_ICMP                 9     // ICMP reported trouble
#define NETERR_PROTOCOL            10     // Protocol error e.g. bad ack number for TCP
#define NETERR_REMOTE_RESET        11     // Remote peer reset connection
#define NETERR_NET_UNREACH         12     // ICMP errors...
#define NETERR_HOST_UNREACH        13     //
#define NETERR_PROTO_UNREACH       14     //
#define NETERR_PORT_UNREACH        15     //
#define NETERR_FRAG_NEEDED         16     //
#define NETERR_SOURCE_ROUTE_FAIL   17     //
#define NETERR_DEST_NET_UNKNOWN    18     //
#define NETERR_DEST_HOST_UNKNOWN   19     //
#define NETERR_ISOLATED            20     //
#define NETERR_DEST_NET_PRHB       21     //
#define NETERR_DEST_HOST_PRHB      22     //
#define NETERR_NET_TOS_UNREACH     23     //
#define NETERR_HOST_TOS_UNREACH    24     //
#define NETERR_TTL_EXC             25     //
#define NETERR_FRAG_REASM_EXC      26     //
#define NETERR_REDIRECT_NET        27     //
#define NETERR_REDIRECT_HOST       28     //
#define NETERR_REDIRECT_TOS_NET    29     //
#define NETERR_REDIRECT_TOS_HOST   30     //
#define NETERR_PARAM_PROB          31     //
#define NETERR_PARAM_PROB_REQOPT   32     //
#define NETERR_SOURCE_QUENCH       33     // ...last ICMP error
#define NETERR_IPADDR_CHANGE       34     // IP address changed
#define NETERR_OUT_OF_MEMORY       35     // No memory for buffer etc.
#define NETERR_IPADDR_CONFLICT     36     // IP address conflict detected
#define NETERR_IFDOWN              37     // Interface down or deactivated


#pragma seg (FARDATA,FARCONST)

#pragma warn -787

const char far * far NetErrStrings[] = {
  "OK",                                      //  0
  "Local host not reachable (ARP)",          //  1
  "Host not reachable (Router)",             //  2
  "Host refused connection",                 //  3
  "Timout on open or close",                 //  4
  "Connection timed out",                    //  5
  "Active abort sent from this peer",        //  6
  "Timed out due to inactivity",             //  7
  "DHCP lease expired",                      //  8
  "ICMP reported trouble",                   //  9
  "Protocol error",                          // 10
  "Remote peer reset connection",            // 11
  "Network unreachable",                     // 12
  "Host unreachable",                        // 13
  "Protocol unreachable",                    // 14
  "Port unreachable",                        // 15
  "Fragmentation needed but DF",             // 16
  "Source route failed",                     // 17
  "Destination network unknown",             // 18
  "Destination host unknown",                // 19
  "This host isolated",                      // 20
  "Destination network prohibited",          // 21
  "Destination host prohibited",             // 22
  "Network type-of-service unreachable",     // 23
  "Host type-of-service unreachable",        // 24
  "Time-to-live exceeded",                   // 25
  "Fragment reassembly time exceeded",       // 26
  "Redirect for network",                    // 27
  "Redirect for host",                       // 28
  "Redirect for TOS and network",            // 29
  "Redirect for TOS and host",               // 30
  "Parameter problem",                       // 31
  "Parameter problem: option missing",       // 32
  "Source quench",                           // 33
  "IP address changed",                      // 34
  "Out of memory",                           // 35
  "IP address conflict",                     // 26
  "Interface down or deactivated"            // 37
};


//--- descr of events for TCP event/datahandler (tcpSockEvent_()) ---
const char far * far EventDescTCP[] = {
  "passive open call",            // TCP_DH_LISTEN
  "active open call",             // TCP_DH_OPEN
  "established connection",       // TCP_DH_ESTAB
  NULL,                           // TCP_DH_INDATA  // new data has arrived
  NULL,                           // TCP_DH_OUTBUF  // more space in xmit buf
  "FIN rcvd (no more reads)",     // TCP_DH_INCLOSE
  "local close (no more writes)", // TCP_DH_OUTCLOSE
  "Closed",                       // TCP_DH_CLOSED
  "local call to sock_abort()",   // TCP_DH_ABORT
  "rcvd RST",                     // TCP_DH_RESET
  "ICMP rcvd"                     // TCP_DH_ICMPMSG
};


#pragma warn 787

#pragma seg (FARDATA)


#define NETERR_MAX_STRINGS  (sizeof(NetErrStrings) / sizeof(NetErrStrings[0]))

#define NUM_TCP_EVENTS  (sizeof(EventDescTCP) / sizeof(EventDescTCP[0]))



#define LogSockMsg(si, msg)  \
  fprintf(DiagLog, "socket (%04X)[%d]: %ls\n",  \
          (uint)(si), TCP_SI_Idx(si), (msg) )


#define  SetSockLink(si, tcpSock)  \
  { ((tcp_Socket *)(tcpSock))->user_data = (si);  (si)->sock = (tcpSock); }

#define ClearSockLink(si)  \
  {  if ((si)->sock)  \
       ((tcp_Socket *)((si)->sock))->user_data = NULL;  (si)->sock = NULL; }


//-----------------------------------------------------------------------------
//  return a far pointer to a string description of the network error #
//-----------------------------------------------------------------------------
const char far *sockErrStr(uint errNum)
{

  if (errNum >= NETERR_MAX_STRINGS)  return "Unknown error";

  return NetErrStrings[errNum];
}

//-----------------------------------------------------------------------------
//  Log the most recent errors (if any) for the specified socket connection.
//
//  The DC network lib currently in use saves the codes for the 2 most recent
//  errors.
//-----------------------------------------------------------------------------
void showErrs(SockInfo *si, sock_type *sock)
{
  int errCode0, errCode1;


  errCode0 = sock_error(sock, 1);
  errCode1 = sock_error(sock, 1);

  if (!errCode1)  return;

  LogSockMsg(si, sockErrStr(errCode1));

  if (!errCode0)  return;

  LogSockMsg(si, sockErrStr(errCode0));
}


//-----------------------------------------------------------------------------
//  Get the total and free counts for the global pool of TCP connections.
//  The *ttl is set to the total, and the free count is the return value.
//-----------------------------------------------------------------------------
uint tcpCounts(uint *ttl)
{
  uint  u, freeCount = 0;
  tcp_Socket  *sock;


  for (u=0,sock=tcpSockDef; u<MaxTCPCons; ++u,++sock)
    if (!sock->user_data)  ++freeCount;

  *ttl = MaxTCPCons;

  return freeCount;
}


//-----------------------------------------------------------------------------
//  Get the total and free counts for the global pool of UDP sockets.
//  The *ttl is set to the total, and the free count is the return value.
//-----------------------------------------------------------------------------
uint udpCounts(uint *ttl)
{
  uint  u, freeCount = 0;
  SockInfo  *si;


  for (u=0,si=udpSockets; u<MaxUDPSockets; ++u,++si)
    if (!si->inUse)  ++freeCount;

  *ttl = MaxUDPSockets;

  return freeCount;
}


//-----------------------------------------------------------------------------
/*
void dumpSockInfo(char *context, SockInfo *si)
{
  char  temp[80];

  _socketDescr_(si, temp, sizeof(temp));

  printf("%s\r\n"
         "%s\r\n"
         "rdOK:%d  wrOk:%d  conState:%d  closeState:%d  opened:%d\r\n",
         context, temp,
         si->rdOK, si->wrOK, si->conState, si->closeState, si->opened);
}
*/
//-----------------------------------------------------------------------------
void initSockInfo(SockInfo *si, sock_type *sock, char type)
{
  if (!KLOCKED())  klockErr(__func__);

  memset(si, 0, sizeof(SockInfo));

  si->type = type;
  si->rdOK = si->wrOK = true;

  si->lastClosed = GetMS();

  init_waitqueue_head(&si->readQueue);
  init_waitqueue_head(&si->writeQueue);

  si->myAddr = si;

  if (sock)  {  // can be NULL for TCP SockInfo structures...
    si->sock = sock;
    if (type == SOCK_STREAM)  ((tcp_Socket *)sock)->user_data = si;
  }


//  if (ShowSocketEvents)  dumpSockInfo("initSockInfo()", si);  //tdebug
}


//-----------------------------------------------------------------------------
//  func called once by init process to initialize data structures
//-----------------------------------------------------------------------------
void socketInit_()
{
  int  i;

  skipTicks = true;
  netReady = linkWentDown = netGoingDown = closingSockets = false;

  linkDownMS = 0;

  for (i=0; i<MaxTCPSockets; ++i)
    initSockInfo(tcpSockets+i, NULL, SOCK_STREAM );

  for (i=0; i<MaxUDPSockets; ++i)
    initSockInfo(udpSockets+i, (sock_type *)(udpSockDef + i), SOCK_DGRAM );

  for (i=0; i<MaxTCPCons; ++i)  tcpSockDef[i].user_data = NULL;

  registerFileSys(&socketFSysFuncs);

  disableIObus();

// no point until interface is up and _socketTick_() is being called
//  skipTicks = false;
}


//-----------------------------------------------------------------------------
//  Returns true if provided pointer is a valid, in-use socket device.
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validSOCK(void *ioDev)
{
  if (!INVALID(SockInfo, ioDev))  if (SI->inUse)  return true;

  return false;
}

//-----------------------------------------------------------------------------
//  Display a message and change the active IP address
//-----------------------------------------------------------------------------
void chgIPAddr(int iface, ulong newAddr)
{
  int  stat;
  char  temp[20];
  IFConfig  ifConfig;


  printf("Set IP address to: %s\r\n",
          inet_htop(AF_INET, &newAddr, temp, sizeof(temp)) );

  stat = ifpending(iface);

  if ((stat == IF_UP) OR (stat == IF_COMING_UP))  startNetDown(iface);

  for (;;)  {
    if (ifpending(iface) == IF_DOWN)  break;
    if (current != kernelTask)
      sleepMS(100);
    else  {
      WDT_2S();  tcp_tick(NULL);  }
  }

  getIFConfig(iface, &ifConfig);
  ifConfig.ipAddr = newAddr;
  setIFConfig(iface, &ifConfig);

  startNetUp(iface);
}

//-----------------------------------------------------------------------------
//  Get the current network interface configuration
//-----------------------------------------------------------------------------
int getIFConfig(int iface, IFConfig *ifConfig)
{
  memset(ifConfig, 0, sizeof(IFConfig));

  KLOCK();

  ifConfig->state = ifpending(iface);

  ifconfig( iface,
            IFG_IPADDR, &(ifConfig->ipAddr),
            IFG_NETMASK, &(ifConfig->netmask),
            IFG_ROUTER_DEFAULT, &(ifConfig->gateway),
            IFG_MTU, &(ifConfig->mtu),
            IFG_HWA, &(ifConfig->macAddr),
            IFS_END
          );

  KUNLOCK();

//  ifConfig->bcAddr = 0xffffffffL;
  ifConfig->bcAddr =
                  (ifConfig->netmask & ifConfig->ipAddr) | ~ifConfig->netmask;

  return 0;
}


//-----------------------------------------------------------------------------
//  set the current network interface configuration
//-----------------------------------------------------------------------------
int setIFConfig(int iface, IFConfig *ifConfig)
{
  IFConfig  curConfig;


  getIFConfig(iface, &curConfig);

  if (curConfig.state != IF_DOWN)  return -1;

  KLOCK();

  ifconfig( iface,
            IFS_IPADDR, ifConfig->ipAddr,
            IFS_NETMASK, ifConfig->netmask,
            IFS_ROUTER_SET, ifConfig->gateway,
            IFS_END
          );

  writeIPAddr(ifConfig->ipAddr);

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  Start bringing up the specified network interface.  Should ONLY be called
//  by startNetUp()!
//
//  Returns the current state of the interface.
//
//  Use ifpending() to check on the state.
//-----------------------------------------------------------------------------
int startUpIF(int iface)
{
  int  stat;


  KLOCK();

  WDT_2S();  stat = ifpending(iface);

  if ((stat != IF_UP) AND (stat != IF_COMING_UP))  {
    printf("Bringing up network interface...\r\n");
    ifconfig(iface, IFS_UP, IFS_END);  stat = ifpending(iface);
  }

  KUNLOCK();

  return stat;
}

//-----------------------------------------------------------------------------
//  Start shutting down the specified network interface.  Should ONLY be
//  called by _socketTick_()!
//
//  Returns the current state of the interface.
//
//  Use ifpending() to check on the state.
//-----------------------------------------------------------------------------
int startDownIF(int iface)
{
  int  stat;

  KLOCK();

  stat = ifpending(iface);

  if ((stat != IF_DOWN) AND (stat != IF_COMING_DOWN)) {
    printf("Taking down network interface...\r\n");
    ifdown(iface);  stat = ifpending(iface);
  }

  KUNLOCK();

  return stat;
}


//-----------------------------------------------------------------------------
//  Configure the network interface using the current settings and start
//  bringing it up.
//
//  Returns the current state of the interface.
//-----------------------------------------------------------------------------
int startNetUp(int iface)
{
  IFConfig  ifConfig;


  printf("\r\nConfiguring network interface...\r\n");

  skipTicks = false;

  SetHostName(hostName, strlen(hostName));

  memset(&ifConfig, 0, sizeof(IFConfig));

  ifConfig.ipAddr = readIPAddr();
  ifConfig.netmask = netmask;
  ifConfig.gateway = gateway;

  ifConfig.mtu = ETH_MTU;

  setIFConfig(IF_ETH0, &ifConfig);

  return startUpIF(iface);
}

//-----------------------------------------------------------------------------
//  Set close flag for all in-use sockets, give them a moment to close, and
//  then shut down the network interface.
//-----------------------------------------------------------------------------
int startNetDown(int iface)
{
  uint  i;
  SockInfo  *si;


  if (closingSockets)  return 0;

  printf("Closing all sockets...\r\n");

  KLOCK();

  for (i=0; i<MaxTCPSockets; ++i)
    if ((si = tcpSockets+i)->inUse)  si->closeState = CS_CALL_CLOSE;

  for (i=0; i<MaxUDPSockets; ++i)
    if ((si = udpSockets+i)->inUse)  si->closeState = CS_CALL_CLOSE;

  netReady = false;  netGoingDown = closingSockets = true;

  startedDownMS = GetMS();

  KUNLOCK();

  return 0;
}

//-----------------------------------------------------------------------------
//  Display the IP routing info
//-----------------------------------------------------------------------------
typedef struct {
  char  dest[16];
  char  gateway[16];
  char  mask[16];
  char  flags[6];
  char  iface[16];
} RouteStrings;

int getRouteStrs(word r, RouteStrings *strs)
{
  int  rsvd;
  word  iface;
  RTEntry  *rte;
  ATEntry  *ate;
  char  *ptr;
  char  ch;


  memset(strs, 0, sizeof(RouteStrings));

  if (r >= ARP_ROUTER_TABLE_SIZE OR !_arp_gate_data[r].ath)  return -1;

  rte = _arp_gate_data + r;

  if (ATH_IS_P2P(rte->ath)) {
    iface = (word)rte->ath - ATH_P2P;
    ate = NULL;  rsvd = 1;
    inet_ntoa(strs->gateway, _if_tab[iface].u.ppp.peeraddr);
    sprintf(strs->iface, "p2p%u", iface);
  }
  else {
    iface = (word) ate->iface;
    ate = _arp_data + ATH2INDEX(rte->ath);
    rsvd = ate->flags & ATE_RESOLVED;
    inet_ntoa(strs->gateway, ate->ip);
    sprintf(strs->iface, "eth%u", iface);
  }

  ptr = strs->flags;
  ch = ATH_IS_P2P(rte->ath) ? '@' :
       rte->flags & RTE_TRANSIENT ? 'T' :
       !(rte->flags & (RTE_TRANSIENT|RTE_DHCP|RTE_REDIRECT)) ? 'P' : 0;
  if (ch)  *ptr++ = ch;
  if (rte->flags & RTE_DHCP)  *ptr++ = 'D';
  if (rte->flags & RTE_REDIRECT)  *ptr++ = 'R';
  if (rte->flags & RTE_DEAD)  *ptr++ = '?';
  if (rsvd)  *ptr++ = 'H';

  if (rte->flags & (RTE_TRANSIENT | RTE_DHCP | RTE_REDIRECT))  {
    strcpy(strs->dest, "???.???.???.???");
    strcpy(strs->mask, "???.???.???.???");
//      rte->u.transient.preference,
//      (long)(rte->u.transient.timestamp - MS_TIMER) / 1000
  }
  else  {
    inet_ntoa(strs->dest, rte->u.preconfig.subnet & rte->u.preconfig.mask);
    inet_ntoa(strs->mask, rte->u.preconfig.mask);
  }

  return 0;
}

//-----------------------------------------------------------------------------
void printRouteStrs(RouteStrings *strs)
{
  printf("%-15s %-15s %-15s %-5s xxxxxx xxxxxx xxx %s\r\n",
    strs->dest, strs->gateway, strs->mask, strs->flags, strs->iface );
}

//-----------------------------------------------------------------------------
void showRoutes(void)
{
  word r;
  IFConfig  curConfig;
  RouteStrings  strs;


  printf(
"\r\n"
"Destination     Gateway         Genmask         Flags Metric Ref    Use Iface\r\n");

//--- first show the route for the local subnet ---
  getIFConfig(IF_ETH0, &curConfig);

  memset(&strs, 0, sizeof(RouteStrings));

  inet_ntoa(strs.dest, curConfig.ipAddr & curConfig.netmask),
  strcpy(strs.gateway, "0.0.0.0");
  inet_ntoa(strs.mask, curConfig.netmask),
  *strs.flags = (((int)curConfig.state == IF_UP) ? 'U' : 'D');
  strcpy(strs.iface, "eth0");
  printRouteStrs(&strs);

//--- and then any routes to addresses not on the local subnet ---
  for (r=0; r<ARP_ROUTER_TABLE_SIZE; r++)  if (_arp_gate_data[r].ath) break;

  if (r == ARP_ROUTER_TABLE_SIZE)  return;

  for (r=0; r<ARP_ROUTER_TABLE_SIZE; r++)
    if (!getRouteStrs(r, &strs))  printRouteStrs(&strs);


/* from a Linux system:
Kernel IP routing table
Destination     Gateway         Genmask         Flags Metric Ref    Use Iface
169.254.0.0     0.0.0.0         255.255.0.0     U     1000   0        0 eth0
192.168.0.0     0.0.0.0         255.255.0.0     U     1      0        0 eth0
0.0.0.0         192.168.0.1     0.0.0.0         UG    0      0        0 eth0
*/

/*
  RTEntry  *rte;
  ATHandle  ath;
  uint  i;

// extern RTEntry _arp_gate_data[ARP_ROUTER_TABLE_SIZE];

  for (i=0, rte=_arp_gate_data; i<ARP_ROUTER_TABLE_SIZE; i++,rte++)  {
    if (!rte->ath)  continue;
      ((!rte->flags AND !rte->u.preconfig.mask)
             OR (rte->flags & RTE_DHCP)) ) {
        ath = rte->ath;
        rx = i;
        break;
     }
*/
}

//-----------------------------------------------------------------------------
//  Check the given descriptor to see if it refers to a socket.
//
//  If it is valid and refers to a socket, then **file is set to the addr of
//  of the device object that joins the descriptor to a socket.
//-----------------------------------------------------------------------------
static SockInfo *_fdSocket_(fd_t sockfd, FDEV far **file)
{
  SockInfo  *si;

//--- does it point to a valid file object? ---
  if (! (*file = _FileFromFD_(sockfd)))  { errno = EBADF;  return NULL; }
  if (! File_OpenCount(*file))  { errno = EBADF;  return NULL; }

//--- does the file object point to a socket? ---
  if (File_Funcs(*file) != &socketIOFuncs)  { errno = ENOTSOCK;  return NULL; }
  if (! (si = File_Dev(*file)))  { errno = EBADF;  return NULL; }
  if (! si->inUse)  { errno = EBADF;  return NULL; }

  return si;
}


//-----------------------------------------------------------------------------
//  Returns a pointer to an unused tcp_Socket structure, or NULL if all of
//  them are in use.
//-----------------------------------------------------------------------------
tcp_Socket *unusedTCP(void)
{
  uint  ui;
  tcp_Socket  *sock;


  if (!KLOCKED())  klockErr(__func__);

  for (ui=0,sock=tcpSockDef; ui<MaxTCPCons; ++ui,++sock)  {
    if (sock->user_data)  continue;
    if (SockClosed(sock))  return sock;
  }

  return NULL;
}

//-----------------------------------------------------------------------------
//  Func to manage the process of closing sockets marked for closing.
//
//  Returns the # of sockets from the given pool that are in still use
//-----------------------------------------------------------------------------
int checkSockets(SockInfo *sockets, uint count)
{
  uint  i;
  int  inUseCount;
  SockInfo  *si;
  sock_type  *sock;
  char  buf[64];


  KLOCK();

  for (i=0,inUseCount=0; i<count; ++i)  {
    if (! (si = sockets + i)->inUse)  continue;

//    if (si->closeState)  {  //tdebug
//      idx = ((int)si->type == SOCK_STREAM) ? TCP_SI_Idx(si) : UDP_SI_Idx(si);
//      fprintf(DiagLog, "socket (%04X)[%d]: closeState: %d\n",
//              (uint)si, idx, si->closeState);  //tdebug
//    }

    sock = si->sock;

    switch (si->closeState)  {

      case CS_CALL_CLOSE:  // initiate a close
        si->rdOK = si->wrOK = false;  si->conState = SS_DISCONNECTED;
        if (si->listener)  {
          tcp_clearreserve(si->ourPort);  si->listener = false; }
        if (!sock)
          si->closeState = CS_IS_CLOSED;
        else {
          if (si->type == SOCK_DGRAM)  {
            sock_close(sock);  si->closeState = CS_IS_CLOSED; }
          else  {
            if (sock_rbused(sock) > 0)  {
              sock_fastread((tcp_Socket *)sock, buf, sizeof(buf));  break; }
            sock_close(sock);  si->closeState = CS_CLOSE_WAIT;
          }
        }
        break;

      case CS_CLOSE_WAIT:  // waiting for close of TCP socket to complete
        if (SockClosed(sock))  si->closeState = CS_IS_CLOSED;
        break;

      case CS_IS_CLOSED:  // close completed
        if (si->type == SOCK_STREAM)  ClearSockLink(si);
        si->lastClosed = GetMS();  si->closeState = CS_WAIT_FREE;

      case CS_WAIT_FREE:  // wait for fcntl layer to call _socketClose_()
        if (!si->noFDEVrefs)  break; // SockInfo struct now safe to reuse?
        si->inUse = false;
    }

    if (si->inUse)  ++inUseCount;
  }

  KUNLOCK();

  return inUseCount;
}

//-----------------------------------------------------------------------------
//  There is no standard way to create an fd for a socket via a call to open().
//
//  socket() is used to create an fd for a socket, which can then be used as
//  follows:
//
//   - To open a connection to a specific port+addr:
//        - Call connect()
//
//   - To listen for and accept incoming connections or packets:
//
//        - Call bind() to specify the local port and which interface(s)
//          (indirectly) by specifying a local addr (or INADDR_ANY for all
//          interfaces)
//
//        For TCP (stream-based) sockets:
//          - Call listen() to mark the socket as a passive listener, create
//            a queue for incoming connection requests, and begin accepting
//            connections on the port+addr/interface(s) specified by bind()
//          - Call accept() to create a new fd for a new open socket connection
//            for the first pending connection in the queue for the listening
//            socket.
//
//        For UDP (packet-based message) sockets:
//          - Call recv() or recvfrom() to get the next packet sent to the
//            port+addr specified in the call to bind()
//
//-----------------------------------------------------------------------------
fd_t _socketOpen_(const char *path, int flags, int mode)
{
  errno = ENXIO;  return -1;
}

//-----------------------------------------------------------------------------
//  Called each kernel tick to provide the driver with CPU cycles.
//
//  Must ONLY be called by the kernel task!!!
//-----------------------------------------------------------------------------
void _socketTick_(void)
{
  uint  curTime, openSockets;
  int  haveLink, ifState;
  static uint  lastCheckTime = 0;
  static int  lastHaveLink = 0, lastState = -1;


  if ((current != kernelTask) OR skipTicks) return;

  tcp_tick(NULL);

  ifState = ifpending(IF_ETH0);
  haveLink = pd_havelink(0);

  if (ifState != lastState)  {
    printf("=== network state chgd to %d from %d ===\r\n", lastState, ifState);
    lastState = ifState;
  }

  if (haveLink != lastHaveLink)  {
    printf("Ethernet Link %s\r\n", haveLink ? "UP" : "DOWN");
    lastHaveLink = haveLink;
    if (haveLink)  {
      resetAnnounceInterval();  linkDownMS = 0; }
    else
      linkDownMS = GetMS();
  }

  netReady = haveLink AND !netGoingDown AND (ifState == IF_UP);

  linkWentDown = !netReady AND linkDownMS AND (msSince(linkDownMS) > 30);

  if (netGoingDown AND (ifState == IF_DOWN))  netGoingDown = false;

  curTime = GetMS_();

  if (!doShutdown)  if (curTime - lastCheckTime < 50)  return;



  openSockets = checkSockets(tcpSockets, MaxTCPSockets)
              + checkSockets(udpSockets, MaxUDPSockets);

  if (netGoingDown)  {
    if (closingSockets AND (!openSockets OR (msSince(startedDownMS) > 5000))) {
      printf("# of sockets still open: %d\r\n", openSockets);
      closingSockets = false;  startDownIF(IF_ETH0);
    }
  }

  #ifdef SHD_DLM
//    if (! doShutdown)  doShutdown |= DLM_Tick();
  #endif

  lastCheckTime = curTime;
}

//-----------------------------------------------------------------------------
//  Return the state of the specified ioDev entry or NULL if not a valid
//  entry #.
//
//  For valid entrys, the entryState struct is filled in
//-----------------------------------------------------------------------------
void *_socketEntry_(uint idx, EntryState *entryState)
{
  SockInfo  *si;

  if (idx < MaxTCPSockets)
    si = tcpSockets + idx;
  else  {
    idx -= MaxTCPSockets;
    if (idx >= MaxUDPSockets)  return NULL;
    si = udpSockets + idx;
  }

  entryState->openCount = (si->inUse ? 1 : 0);
  entryState->unused = (entryState->openCount == 0);

  return si;
}


//-----------------------------------------------------------------------------
//  Placeholder, do-nothing unlink function
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _socketUnlink_(char *path)
{
  return -1;
}


//-----------------------------------------------------------------------------
//  Placeholder, do-nothing rename function
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int _socketRename_(const char *oldPath, const char *newPath)
{
  return -1;
}


//-----------------------------------------------------------------------------
//  return a count of the # of sockets in use
//-----------------------------------------------------------------------------
int numOpenSockets(void)
{
  return
     checkSockets(tcpSockets, MaxTCPSockets)
   + checkSockets(udpSockets, MaxUDPSockets);

}

//-----------------------------------------------------------------------------
//  Callback function for TCP events.  Called from within a call to tcp_tick()
//-----------------------------------------------------------------------------
int _nearcall tcpSockEvent_(int event, void *sock, ll_Gather *g, void *info)
{
  bool  logSockEvents;
  SockInfo  *si;
  const char far *eventDesc;


  if (!KLOCKED())  klockErr(__func__);


  si = TCP_SD_To_SI(sock);

  if (INVALID(SockInfo, si))  return 0;

  if ((si->sock != sock) OR !si->inUse)  {
    if (!doShutdown)  fprintf(DiagLog,
            "*** tcpSockEvent(): TCP[%u]: Event %d\r\n"
            "    Invalid socket pointer: %05lX\r\n",
            TCP_SI_Idx(si), event, (ulong)sock);
    return 0;
  }


  logSockEvents = BOOL(si->diagFlags & SocketEventBits);


  if (logSockEvents)  showErrs(si, sock);


  if ((uint)event >= NUM_TCP_EVENTS)  return 0;

  if (logSockEvents)  {
    eventDesc = EventDescTCP[(uint)event];
    if (eventDesc)  LogSockMsg(si, eventDesc);
  }


  switch (event)  {

    case TCP_DH_LISTEN:  // passive open call (listen)
      break;

    case TCP_DH_OPEN:    // active open call (sent FIN)
      break;

    case TCP_DH_ESTAB:  // Connection established
      break;

    case TCP_DH_INDATA:  // new data has arrived
      wakeTasks(&si->readQueue);  break;

    case TCP_DH_OUTBUF:  // more space in output buffer
      wakeTasks(&si->writeQueue);  break;

    case TCP_DH_INCLOSE:  // rcvd FIN - reads will fail
      si->rdOK = false;  wakeTasks(&si->readQueue);  break;

    case TCP_DH_OUTCLOSE:  // local appl closed, FIN sent: writes will fail
      si->wrOK = false;  si->rdOK = (sock_rbused(sock) > 0);
      wakeTasks(&si->readQueue);  wakeTasks(&si->writeQueue);  break;

    case TCP_DH_CLOSED:  // Socket completed closed
      si->wrOK = false;  si->rdOK = (sock_rbused(sock) > 0);
      wakeTasks(&si->readQueue);  wakeTasks(&si->writeQueue);  break;

    case TCP_DH_ABORT:   // local call to sock_abort()
      si->rdOK = si->wrOK = false;
      wakeTasks(&si->readQueue);  wakeTasks(&si->writeQueue);  break;

    case TCP_DH_RESET:   // peer sent RST flag
      si->rdOK = si->wrOK = false;
      wakeTasks(&si->readQueue);  wakeTasks(&si->writeQueue);  break;

    case TCP_DH_ICMPMSG:  // Incoming ICMP message
      break;
  }


  return 0;
}


//-----------------------------------------------------------------------------
//  Create a new listening socket to replace the one we just got a connection
//  on and swap what the descriptor's file objects point to so the new
//  descriptor refers to the socket just opened
//
//  WARNING!  KLOCK() must be in effect when this function is called!
//-----------------------------------------------------------------------------
static fd_t _newListener_(fd_t curSockfd)
{
  fd_t  newSockfd;
  FDEV far  *newFile, *curFile;
  SockInfo  *newSI, *curSI;


  if (!KLOCKED())  klockErr(__func__);

  curFile = _FileFromFD_(curSockfd);  curSI = File_Dev(curFile);


  if (curSI->type == SOCK_STREAM)  {
  //--- allocates new fd-->file-->socket triplet ---
    if ((newSockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)  return -1;

  //===== copy existing socket and fd settings to new socket and fd =====

    newFile = _FileFromFD_(newSockfd);  newSI = File_Dev(newFile);

    newSI->ourPort = curSI->ourPort;
    newSI->ourAddr = curSI->ourAddr;

    newSI->listener = true;

  //--- Swap what the global fds point to so the new fd refers to the open ---
  //--- socket and the existing one refers to the new listening socket. ---
  //--- Swapping what the global fds point to means that links from the ---
  //--- task's fds to the global fds will remain intact (and there can be
  //--- multiple task fds pointing to the same global fd) ---
  //--- NOTE:  _File_Swap_() does NOT swap the task-level fd values ---
  //--- (openCount and flags) ---
    _File_Swap_(curFile, newFile);

    return newSockfd;
  }

//--- error - unsupported type ---
  errno = ESOCKTNOSUPPORT;  return -1;
}


//=============================================================================
//  IODev interface functions (for use by fcntl module)
//=============================================================================

//-----------------------------------------------------------------------------
int _socketGetQueues_(void *ioDev, WaitQueues far *queues)
{
  queues->readQueue = &SI->readQueue;
  queues->writeQueue = &SI->writeQueue;

  return 0;
}

//-----------------------------------------------------------------------------
//  Attempt to read bytes from a socket
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and errno == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _socketRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  int  stat, idx;


  stat = -1;

  KLOCK();

  if (! validSOCK(ioDev))  { KUNLOCK();  errno = EBADF;  return -1; }

  SI->diagFlags = current->diagFlags;

  if (!SI->sock)  { KUNLOCK();  errno = EBADF;  return -1; }


//NOTE:  A read can suceed from a socket that is "closed" if there is still
//       data in its buffer.

  switch (SI->type)  {
    case SOCK_STREAM:
      stat = sock_fastread((tcp_Socket *)SI->sock, buf, numBytes);
      idx = TCP_SI_Idx(SI);
      break;

    case SOCK_DGRAM:
      stat = udp_recv( (udp_Socket *)SI->sock, buf, numBytes );
      if (stat == -1)  { KUNLOCK();  errno = EAGAIN;  return -1; }  // no datagram
      idx = UDP_SI_Idx(SI);
      break;

    default:
      KUNLOCK();  errno = ESOCKTNOSUPPORT;  return -1;  // unsupported type
  }

  KUNLOCK();

  if (stat > 0)  return stat;

  if (stat < 0)  {
    if (! SI->rdOK)  {
      if (ShowSocketEvents)
        fprintf(DiagLog, "socket (%04X)[%d]: read from closed socket ...\n",
                (uint)SI, idx );
    }
    errno = EPIPE;  return -1;
  }

//--- For network lib funcs that don't return an error even after the ---
//--- network goes down (?!?!?!?) ---
  if (linkWentDown)  { errno = ENETDOWN;  return -1; }

//--- no error - but no new data to read ---
  errno = EAGAIN;  return -1;
}


//-----------------------------------------------------------------------------
//  Attempt to write bytes to socket connection
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and errno == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _socketWrite_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  int  stat, idx; //, avail;


  stat = -1;

  KLOCK();

  if (! validSOCK(ioDev))  { KUNLOCK();  errno = EBADF;  return -1; }

  SI->diagFlags = current->diagFlags;

  if (!SI->sock)  { KUNLOCK();  errno = EBADF;  return -1; }


  switch (SI->type)  {
    case SOCK_STREAM:
      idx = TCP_SI_Idx(SI);
      if (!SI->wrOK)  break;
      if (SI->conState != SS_CONNECTED)  {
        KUNLOCK();  errno = ENOTCONN;  return -1; }
//    avail = sock_tbleft((tcp_Socket *)SI->sock);
//    if (avail <= 0)  { stat = 0;  break; }
//    if (avail < numBytes)  numBytes = avail;
//      if (sock_tbleft((tcp_Socket *)SI->sock) < 128)  tcp_tick(NULL);  //test
      stat = sock_fastwrite((tcp_Socket *)SI->sock, buf, numBytes);
//      if (sock_tbleft((tcp_Socket *)SI->sock) < 128)  tcp_tick(NULL);  //test
      break;

    case SOCK_DGRAM:
      if (! SI->opened)  { KUNLOCK();  errno = EBADF;  return -1; }
      stat = udp_send( (udp_Socket *)SI->sock, buf, numBytes);
      idx = UDP_SI_Idx(SI);
      break;

    default:
      KUNLOCK();  errno = ESOCKTNOSUPPORT;  return -1;  // unsupported type
  }

  KUNLOCK();

//--- if any successfully buffered, then return the # ---
  if (stat > 0)  return stat;

  if (stat < 0)  {
    if (! SI->wrOK)  {
      if (ShowSocketEvents)
        fprintf(DiagLog, "socket (%04X)[%d]: write to closed socket ...\n",
                (uint)SI, idx );
    }
    errno = EPIPE;  return -1;
  }


//--- For network lib funcs that don't return an error even after the ---
//--- network goes down (?!?!?!?) ---
  if (linkWentDown)  { errno = ENETDOWN;  return -1; }

//--- no error - but output buffer is full ---
  errno = EAGAIN;  return -1;
}


//-----------------------------------------------------------------------------
//  Function to reposition the current offset for a file descriptor
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
off_t  _socketLseek_(void *ioDev, off_t offset, int whence)
{
  errno = ESPIPE;  return (off_t) -1;
}


//-----------------------------------------------------------------------------
//  Get information about the socket
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
int  _socketFstat_(void *ioDev, struct stat *fsbuf)
{
  fsbuf->st_mode = S_IFSOCK | 00777;

  return 0;
}


//-----------------------------------------------------------------------------
//  Start the close process for a socket connection
//
//  Should NOT be called until ALL descriptors for the socket have been closed!
//-----------------------------------------------------------------------------
int _socketClose_(void *ioDev)
{
  KLOCK();

  if (! validSOCK(ioDev))  { KUNLOCK();  errno = EBADF;  return -1; }

  if (!list_empty(&SI->readQueue.listHead)
      OR !list_empty(&SI->writeQueue.listHead))  {
    if (!list_empty(&SI->readQueue.listHead))  {
      fprintf(DiagLog,
              "***** tasks in %sQueue for closed socket *****\n", "read");
      wqListTasks(&SI->readQueue);  wakeTasks(&SI->readQueue);
    }
    if (!list_empty(&SI->writeQueue.listHead))  {
      fprintf(DiagLog,
              "***** tasks in %sQueue for closed socket *****\n", "write");
      wqListTasks(&SI->writeQueue);  wakeTasks(&SI->writeQueue);
    }
  }

  SI->diagFlags = current->diagFlags;

  if (SI->type == SOCK_STREAM)
    if (SI->sock)  sock_flush((tcp_Socket *)SI->sock);

  SI->closeState = CS_CALL_CLOSE;  SI->noFDEVrefs = true;
//  SI->rdOK = SI->wrOK = false;  // done in checkSockets()

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  format an addr and port number in host-byte-order to a string
//-----------------------------------------------------------------------------
char *fmtAddrPort(char *buf, ulong addr, uint port)
{
  char  *ptr;

  if (addr)
    inet_htop(AF_INET, &addr, buf, 4);
  else
    strcpy(buf, "*");

  ptr = buf + strlen(buf);

  if (port)  sprintf(ptr, ":%u", port);  else  strcpy(ptr, ":*");

  return buf;
}


//-----------------------------------------------------------------------------
//  Return a string description of the state of a socket
//-----------------------------------------------------------------------------
const char *sockState(SockInfo *si)
{
  if (! VALID(SockInfo, si))  return "*INVALID*";

  if (si->sock)  return sockstate(si->sock);

  if (si->listener)  return "Listen";

  return "*INVALID*";
}


//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for sockets
//-----------------------------------------------------------------------------
int _socketioctl_(void *ioDev, int request, void *arg)
{
  int  stat;

#define CHAR_ARG   (*((char *)arg))
#define INT_ARG    (*((int *)arg))
#define UINT_ARG   (*((uint *)arg))
#define ULONG_ARG  (*((ulong *)arg))


  if (! validSOCK(ioDev))  { errno = EBADF;  return -1; }

  SI->diagFlags = current->diagFlags;

  stat = -1;

  switch (request)  {
    case FIONREAD:  // return # of unread bytes
      if (!SI->rdOK) { INT_ARG = -1;  errno = ENOLINK;  break; }
      KLOCK();  INT_ARG = (SI->sock) ? sock_rbused(SI->sock) : 0;  KUNLOCK();
      stat = 0;  break;

    case FIONSPACE:  // return # of free bytes in the xmit buffer
      if (!SI->wrOK) { INT_ARG = -1;  errno = ENOLINK;  break; }
      KLOCK();  INT_ARG = (SI->sock) ? sock_tbleft(SI->sock) : 0;  KUNLOCK();
      stat = 0;  break;

    default:
      errno = EINVAL;
  }

  return stat;
}


//-----------------------------------------------------------------------------
//  Return a description for a socket fd.
//
//  To be called ONLY by fcntl functions
//-----------------------------------------------------------------------------
char *_socketDescr_(void *ioDev, char *buf, int bufSize)
{
  char  *ptr;


  if (bufSize < 80)  { *buf = 0;  return buf; }

//IPv4 0x1234 TCP[1] 192.168.250.247:502 -> 192.168.125.102:57159 Established

  sprintf(buf, "IPv4 0x%04lX ", (ulong)ioDev);

  ptr = buf + strlen(buf);  strcpy(ptr, "bad bad");

//finish: limit output to bufSize

  if (! validSOCK(ioDev))  return buf;

  if (SI->type == SOCK_STREAM)
    sprintf(ptr, "TCP[%d] ", SI - tcpSockets);
  else
    sprintf(ptr, "UDP[%d] ", SI - udpSockets);
  while (*ptr) ++ptr;

  fmtAddrPort(ptr, SI->ourAddr, SI->ourPort);
  strcat(ptr, " -> ");  while (*ptr) ++ptr;

  fmtAddrPort(ptr, RemoteAddr(SI), RemotePort(SI));
  strcat(ptr, " ");  strcat(ptr, sockState(SI));

  return buf;
}



//=============================================================================
//  Unix compatible standard BSD socket functions
//=============================================================================

//-----------------------------------------------------------------------------
//  Allocate a socket and a file descriptor for the specified type of socket
//-----------------------------------------------------------------------------
fd_t socket(int family, int type, int protocol)
{
  fd_t  fd;
  uint  ui;
  SockInfo  *si, *test;
  void  *sockPtr;


  if (family != AF_INET)  { errno = EAFNOSUPPORT;  return -1; }

//--- get an unused socket of the correct type ---
  si = NULL;

  KLOCK();

//NOTE:  When there are multiple not-in-use sockets, we choose the one with
//       the earliest lastClosed time

  switch (type)  {

    case SOCK_STREAM:
      for (ui=0; ui<MaxTCPSockets; ++ui)  {
        test = tcpSockets + ui;
        if (!test->inUse)  if (!si OR (test->lastClosed < si->lastClosed))  {
          si = test;  sockPtr = NULL; }
      }
      break;

    case SOCK_DGRAM:
      for (ui=0; ui<MaxUDPSockets; ++ui)  {
        test = udpSockets + ui;
        if (!test->inUse)  if (!si OR (test->lastClosed < si->lastClosed))  {
            si = test;  sockPtr = udpSockDef + ui; }
      }
      break;

    default:
      KUNLOCK();  errno = ESOCKTNOSUPPORT;  return -1;
  }

  if (! si)  { KUNLOCK();  errno = ENOBUFS;  return -1; }

//--- initialize SockInfo object ---
  initSockInfo(si, sockPtr, type);
  si->inUse = true;  si->diagFlags = current->diagFlags;

//--- allocate a file descriptor to ref the socket data ---
  if ((fd = _allocFile_(3, &socketIOFuncs, si)) < 0)  {
    si->inUse = false;  KUNLOCK();  return -1; }


//--- everything OK - return the descriptor assigned for this process ---
  KUNLOCK();    return fd;
}



//-----------------------------------------------------------------------------
//  Bind a socket to a local address (or 0 for all) and port
//-----------------------------------------------------------------------------
int bind(fd_t sockfd, const struct sockaddr *localAddr, int nameLen)
{
  FDEV  far *file;
  SockInfo  *si;
  int  stat;


  KLOCK();

  if (! (si = _fdSocket_(sockfd, &file)))  { KUNLOCK();  return -1; }

  si->diagFlags = current->diagFlags;

  stat = -1;

  switch (si->type)  {

    case SOCK_STREAM:
      si->ourPort = ntohs(localAddr->sin_port);
      si->ourAddr = ntohl(localAddr->sin_addr);
      stat = 0;  break;

    case SOCK_DGRAM:
      si->ourPort = ntohs(localAddr->sin_port);
      si->ourAddr = ntohl(localAddr->sin_addr);
      stat = 0;  break;

    default:
      errno = ESOCKTNOSUPPORT;
  }

  KUNLOCK();  return stat;
}


//-----------------------------------------------------------------------------
//  Set a socket to listen mode so it can accept incoming connections with the
//  accept() call
//-----------------------------------------------------------------------------
int listen(fd_t sockfd, int backlog)
{
  FDEV  far *file;
  SockInfo  *si;
  word  netPort;
  uint  x;
  bool  allUsed;


  KLOCK();

  if (! (si = _fdSocket_(sockfd, &file)))  { KUNLOCK();  return -1; }

  if (si->type != SOCK_STREAM)  { KUNLOCK();  errno = EOPNOTSUPP;  return -1; }

//--- Check for unused and duplicate listener slots ---
  netPort = intel16(si->ourPort);  allUsed = true;
  for (x=0; x<MAX_RESERVEPORTS; ++x)  {
    if (tcp_reserveports[x] == netPort) {
      KUNLOCK();  errno = EADDRINUSE;  return -1; }
    if (!tcp_reserveports[x])  allUsed = false;
  }

  if (allUsed)  {  KUNLOCK();  errno = ENOSR;  return -1; }

  si->diagFlags = current->diagFlags;

  tcp_reserveport(si->ourPort);

  si->listener = true;

  KUNLOCK();

  if (ShowSocketEvents)
    fprintf(DiagLog, "socket (%04lX)[%d]: listening on port %d\n",
            (ulong)si, TCP_SI_Idx(si), si->ourPort);

  return 0;
}


//-----------------------------------------------------------------------------
//  Check for a new connection to a listening socket.
//
//  Returns:
//  - if new connection, a new fd for the new connection and clientAddr filled
//    in with the addr and port of the remote end of the connection
//
//  - if no new connection and blocking disabled, -1 and errno = EAGAIN
//
//  - if an error, -1 and errno indicates the error
//-----------------------------------------------------------------------------
fd_t accept(fd_t sockfd, struct sockaddr *clientAddr, int *addrLen)
{
  FDEV  far *file;
  fd_t  newSockfd;
  SockInfo  *si;
  sock_type  *sock;
  struct sockaddr addr;
  int  len, fflags;


  KLOCK();

  if (! (si = _fdSocket_(sockfd, &file)))  { KUNLOCK();  return -1; }

  if (!si->listener)  { KUNLOCK();  errno = EINVAL;  return -1; }

  if (! (sock = (sock_type *)unusedTCP()))  {
    KUNLOCK();  errno = ENOBUFS;  return -1; }

  SetSockLink(si, sock);

  if (! tcp_listen((tcp_Socket *)sock, si->ourPort, 0, 0, tcpSockEvent_, 0))  {
    showErrs(si, sock);  ClearSockLink(si);
    KUNLOCK();  errno = EIO;  return -1;
  }

  if (SockClosed(sock))  {
    ClearSockLink(si);  KUNLOCK();  errno = EPIPE;  return -1; }

  si->diagFlags = current->diagFlags;

  if (si->type == SOCK_STREAM)  for (;;)  {
//    if (tcp_tick(sock) == 0)  { KUNLOCK();  errno = ECONNRESET;  return -1; }

    if ( sock_established(sock) OR (sock_rbused(sock) > 0) )  {
    /*
       If _newListener_() returns >= 0, then newSockfd refers to the just
       accepted/established connection, and sockfd now refers to a new
       listener socket for the same port.

       If _newListener_() returns < 0, then there is no new listening socket
       and sockfd still refers to the just accepted/established connection.
       When that occurs, we need to close the new connection and leave the
       existing socket as a listener.
    */
      if ((newSockfd = _newListener_(sockfd)) < 0)  {
        ClearSockLink(si);  sock_abort(sock);
        KUNLOCK();  errno = ENOBUFS;  return -1;
      }

      len = sizeof(struct sockaddr);
      getsockname(sock, &addr, &len);

      si->ourPort = addr.s_port;
      si->ourAddr = addr.s_ip;

      si->rdOK = si->wrOK = true;

      si->listener = false;

      si->conState = SS_CONNECTED;

      if (clientAddr)  {
        if (*addrLen >= 2)  clientAddr->sin_family = AF_INET;
        if (*addrLen >= 4)  clientAddr->sin_port = htons(RemotePort(si));
        if (*addrLen >= 8)  clientAddr->sin_addr = htonl(RemoteAddr(si));
        *addrLen = sizeof(struct sockaddr);
      }

      KUNLOCK();

      if (ShowSocketEvents)   fprintf(DiagLog,
            "socket (%04lX)[%d]: new socket connection, port %d\n",
            (ulong)si, TCP_SI_Idx(si), si->ourPort);

      return newSockfd;
    }

  //--- No new connection, so reset the tcp_Socket we used to check for one ---
    sock_abort(sock);

//    KUNLOCK();  errno = EAGAIN;  return -1;  //tdebug - blocking causes problems right now

    fflags = fcntl(sockfd, F_GETFL, 0);

    ClearSockLink(si);

    KUNLOCK();

    if (fflags & O_NONBLOCK)  { errno = EAGAIN;  return -1; }

    if (linkWentDown)  { KUNLOCK();  errno = ENETDOWN;  return -1; }

    //finish - blocking mode should be handled using a wait queue
    sleepMS(200);  KLOCK();
  }


  KUNLOCK();  ERRNO = ESOCKTNOSUPPORT;  return -1;  // unsupported type
}



//-----------------------------------------------------------------------------
//  Initiate a connection on a socket
//-----------------------------------------------------------------------------
int connect(fd_t sockfd, const struct sockaddr *remoteAddr, int addrLen)
{
  FDEV  far *file;
  SockInfo  *si;
  sock_type  *sock;
  ulong  remAddr;
  short  remPort;
  struct sockaddr addr;
  int  len;
  char  descr[80];


  KLOCK();

  if (! (si = _fdSocket_(sockfd, &file)))  { KUNLOCK();  return -1; }

  si->diagFlags = current->diagFlags;

  if (! (sock = (sock_type *)unusedTCP()))  {
    KUNLOCK();  errno = ENOBUFS;  return -1; }

  SetSockLink(si, sock);

  if (si->type == SOCK_STREAM)  {  // TCP socket

    if (si->conState == SS_UNCONNECTED)  {
      remAddr = ntohl(remoteAddr->sin_addr);
      remPort = ntohs(remoteAddr->sin_port);

      if (ShowSocketEvents)
        fprintf(DiagLog, "socket (%04lX)[%d]: connect(): %s\n",
                (ulong)si, TCP_SI_Idx(si),
                fmtAddrPort(descr, remAddr, remPort) );

      if (! tcp_open((tcp_Socket *)sock, 0,
                      remAddr, remPort, tcpSockEvent_))  {
        ClearSockLink(si);  KUNLOCK();  errno = ENETUNREACH;  return -1;  }

      si->startTime = MS_TIMER;  si->conState = SS_CONNECTING;

//finish - support for non-blocking mode (errno == EINPROGRESS or EALREADY)
//         (i.e. allow return of -1 with errno set while not established)
      errno = EINPROGRESS;
    }

    while (si->conState == SS_CONNECTING)  {
      if ( sock_established(sock) OR (sock_rbused(sock) > 0) )  {
        len = sizeof(struct sockaddr);
        getsockname(sock, &addr, &len);
        si->ourPort = addr.s_port;
        si->ourAddr = addr.s_ip;
        si->conState = SS_CONNECTED;
        KUNLOCK();  return 0;
      }

      if (msSince(si->startTime) < ConnectTimeout)  {
        KUNLOCK();  sleepMS(10);  KLOCK();  continue;  }

      KUNLOCK();  errno = ETIMEDOUT;  return -1;
    }

    errno = (si->rdOK AND si->wrOK) ? EALREADY : ECONNREFUSED;
    KUNLOCK();  return -1;
  }

  KUNLOCK();  errno = ESOCKTNOSUPPORT;  return -1;  // unsupported type
}


//-----------------------------------------------------------------------------
//  Read a msg from a socket
//-----------------------------------------------------------------------------
int recvfrom(fd_t sockfd, void far *buf, int numBytes, int flags,
             struct sockaddr *fromAddr, int *addrLen)
{
  int  stat, idx;
  longword  fromIP;
  word  fromPort;
  SockInfo  *si;
  sock_type  *sock;
  FDEV  far *file;
  bool  returnFromAddr;


//finish: Implement values in "flags"

//finish: Need to handle blocking mode.  NOTE: Blocking and non-blocking
//        mode is normally handled in the fcntl layer/module, but this func
//        only applies to sockets, so probably should be done in the driver.

#define FromAddr  ((struct sockaddr_in *)fromAddr)


  KLOCK();

  if (! (si = _fdSocket_(sockfd, &file)))  { KUNLOCK();  return -1; }

  if (! (sock = si->sock))  { KUNLOCK();  errno = EBADF;  return -1; }


  si->diagFlags = current->diagFlags;

  returnFromAddr = (FromAddr AND (*addrLen >= sizeof(struct sockaddr)));

  switch (si->type)  {
    case SOCK_STREAM:
      stat = sock_xfastread((tcp_Socket *)sock, (ulong)buf, numBytes);
      idx = TCP_SI_Idx(si);
      break;

    case SOCK_DGRAM:
      if (! si->opened)  {
//        printf("recvfrom() udp open: lport: %d, remIP: -1, remPort: 0\r\n", si->ourPort);  //tdebug
        if (! udp_open((udp_Socket *)sock, si->ourPort, -1, 0, NULL))  {
          KUNLOCK();  errno = EINVAL;  return -1; }  // == 0 means failed
        si->opened = true;
      }
      if (returnFromAddr)
        stat = udp_recvfrom((udp_Socket *)sock, buf, numBytes,
                            &fromIP, &fromPort);
      else
        stat = udp_recv((udp_Socket *)sock, buf, numBytes);
      if (stat == -1) {  // no datagram, possibly an error
        if (! si->rdOK)  errno = EPIPE;  else  errno = EAGAIN;
        KUNLOCK();  return -1;
      }
      idx = UDP_SI_Idx(si);
      break;

    default:
      KUNLOCK();  errno = ESOCKTNOSUPPORT;  return -1;  // unsupported type
  }

  if (stat > 0)  {  // got some data
    if (returnFromAddr)  {
      FromAddr->sin_family = AF_INET;
      FromAddr->sin_port   = htons(fromPort);
      FromAddr->sin_addr   = htonl(fromIP);
      *addrLen = sizeof(struct sockaddr);
    }
    file->readCount += stat;
    KUNLOCK();  return stat;
  }

  if (stat < 0)  {  // error
    if (! si->rdOK)  {
      if (ShowSocketEvents)
        fprintf(DiagLog, "socket (%04X)[%d]: read from closed socket ...\n",
                 (uint)si, idx );
      errno = EBADF;
    }
    else
      errno = EPIPE;
    KUNLOCK();  return -1;
  }

  if (stat == 0)  if (! si->rdOK)  {
    if (ShowSocketEvents)
      fprintf(DiagLog, "socket (%04X)[%d]: read from closed socket ...\n",
            (uint)si, idx );
    KUNLOCK();  errno = EPIPE;  return -1;
  }


//--- For network lib funcs that don't return an error even after the ---
//--- network goes down (?!?!?!?) ---
  if (linkWentDown)  { KUNLOCK();  errno = ENETDOWN;  return -1; }

//--- no error - but no new data to read ---
  KUNLOCK();  errno = EAGAIN;  return -1;
}


//-----------------------------------------------------------------------------
//  Read a msg from a socket
//-----------------------------------------------------------------------------
int recv(fd_t sockfd, void far *buf, int numBytes, int flags)
{
  int  addrLen;

  return recvfrom(sockfd, buf, numBytes, flags, NULL, &addrLen);
}


//-----------------------------------------------------------------------------
//  Write a msg to a socket
//-----------------------------------------------------------------------------
int sendto(fd_t sockfd, const void far *buf, int numBytes, int flags,
           const struct sockaddr *toAddr, int addrLen)
{
  int  stat, idx;
  longword  remIP, saveIP;
  word  remPort;
  bool  addrGiven, nonBlocking;
  SockInfo  *si;
  sock_type  *sock;
  FDEV  far *file;


//finish: Implement values in "flags"

//finish: Need to handle blocking mode.  NOTE: Blocking and non-blocking
//        mode is normally handled in the fcntl layer/module, but this func
//        only applies to sockets, so probably should be done in the driver.


#define ToAddr  ((struct sockaddr_in *)toAddr)

  if (ShowRWCalls)
    fprintf(DiagLog, "sendto(%d, 0x%05lX, %d, %04X, 0x%04X, %d)\r\n",
            sockfd, (ulong)buf, numBytes, flags, (uint)toAddr, addrLen);

  if (numBytes <= 0)  return 0;

  if (!buf)  { errno = EINVAL;  return -1; }

  KLOCK();

  if (! (si = _fdSocket_(sockfd, &file)))  { KUNLOCK(); return -1; }

  if (! (sock = si->sock))  { KUNLOCK();  errno = EBADF;  return -1; }


  nonBlocking = file->flags & O_NONBLOCK;
  addrGiven = (ToAddr AND (addrLen > 0));

  si->diagFlags = current->diagFlags;


  switch (si->type)  {

    case SOCK_STREAM:
      if (si->conState != SS_CONNECTED)  {
        KUNLOCK();  errno = ENOTCONN;  return -1; }
      if (addrGiven)  { KUNLOCK();  errno = EISCONN;  return -1; }
//      if (sock_tbleft((tcp_Socket *)sock) < 128)  tcp_tick(NULL);  //test
      stat = sock_xfastwrite((tcp_Socket *)sock, (ulong)buf, numBytes);
//      if (sock_tbleft((tcp_Socket *)sock) < 128)  tcp_tick(NULL);  //test
      idx = TCP_SI_Idx(si);
      break;


    case SOCK_DGRAM:
      remIP = -1;  remPort = -1;  idx = UDP_SI_Idx(si);
      if (addrGiven)  {
        remIP = ntohl(ToAddr->sin_addr);  remPort = ntohs(ToAddr->sin_port);  }
      else  {
        remIP = 0;  remPort = 0;  }
      if (! si->opened)  {
        if (!si->bcastMode AND !addrGiven)  {
          KUNLOCK();  errno = EDESTADDRREQ;  return -1; }
//      printf("sendto() udp open: lport: %d, remIP: 0x%08lx, remPort: %d\r\n", //tdebug
//              si->ourPort, remIP, remPort);  //tdebug
        if (si->bcastMode)  { saveIP = remIP;  remIP = -1; }
        if (udp_open((udp_Socket *)sock, si->ourPort, remIP, remPort, NULL)
          == 0)  { KUNLOCK();  errno = EINVAL;  return -1; }  // == 0 means failed
        if (si->bcastMode)  remIP = saveIP;
        si->opened = true;
      }
      stat = udp_xsendto( (udp_Socket *)sock, (ulong)buf, numBytes,
                         remIP, remPort );
      break;


    default:
      KUNLOCK();  errno = ESOCKTNOSUPPORT;  return -1;  // unsupported type
  }


//--- if any successfully buffered, then return the # ---
  if (stat > 0)  {
    file->writeCount += stat;  KUNLOCK();
    if (ShowRWCalls)  fprintf(DiagLog, "  sendto returned: %d\n", stat);
    return stat;
  }

  if (stat < 0)  {
    if (! si->wrOK)  {
      if (ShowSocketEvents)
        fprintf(DiagLog, "socket (%04X)[%d]: write to a closed socket ...\n",
                (uint)si, idx );
      errno = EBADF;
    }
    else
      errno = EINVAL;
    KUNLOCK();  return -1;
  }


//--- For network lib funcs that don't return an error even after the ---
//--- network goes down (?!?!?!?) ---
  if (linkWentDown)  { KUNLOCK();  errno = ENETDOWN;  return -1; }

//--- No other error, so output buffer must be full ---
  KUNLOCK();  errno = EAGAIN;  return -1;
}


//-----------------------------------------------------------------------------
//  Write a msg to a socket
//-----------------------------------------------------------------------------
int send(fd_t sockfd, const void far *buf, int numBytes, int flags)
{
  return sendto(sockfd, buf, numBytes, flags, NULL, 0);
}



//-----------------------------------------------------------------------------
//  set options on a socket fd
//-----------------------------------------------------------------------------
int setsockopt(int sockfd, int level, int opt,
               const void *val, socklen_t valLen)
{
  SockInfo  *si;
  FDEV  far *file;


  if (level != SOL_SOCKET)  { errno = ENOPROTOOPT;  return -1; }

  if (! (si = _fdSocket_(sockfd, &file)))  return -1;

  if (!val)  return -1;

  si->diagFlags = current->diagFlags;

  if (opt == SO_BROADCAST)  {
    if (valLen != sizeof(int))  { errno = EINVAL;  return -1; }
    si->bcastMode = BOOL(*((int *)val)) ? true : false;
    return 0;
  }

//finish:  Add support for other opt values such as SO_RCVTIMEO, which sets a
//         timeout for calls to readv(), rcv(), recvfrom(), recvmsg()
//         (but what about read()?  And does it apply only to socket fd's
//         that are not set to non-blocking mode?)

  errno = ENOPROTOOPT;  return -1;
}

//-----------------------------------------------------------------------------
//  get value of an option on a socket fd
//-----------------------------------------------------------------------------
int getsockopt(int sockfd, int level, int opt,
               void *val, socklen_t *valLen)
{
  SockInfo  *si;
  FDEV  far *file;


  if (level != SOL_SOCKET)  { errno = ENOPROTOOPT;  return -1; }

  if (! (si = _fdSocket_(sockfd, &file)))  return -1;

  if (!val OR !valLen)  return -1;

  si->diagFlags = current->diagFlags;

  if (opt == SO_BROADCAST)  {
    if (*valLen >= sizeof(int))  {
      (*(int *)val) = si->bcastMode ? 1 : 0;  *valLen = sizeof(int);  }
    return 0;
  }

  errno = ENOPROTOOPT;  return -1;
}

//-----------------------------------------------------------------------------

