//----- srfs.c ----- 02/05/10 -----(07/31/06)-----

/*-----------------------------------------------------------------------------
   Client interface for Simple Remote File System (SRFS), an inefficient, but
   dirt simple protocol for accessing files remotely over a TCP connection.
  -----------------------------------------------------------------------------
  Calls made by the client:

  int  srfsOpen(const char *path, int oflag, int mode);
  int  srfsClose(int fd);
  int  srfsRead(int fd, void *buf, int nbyte);
  int  srfsWrite(int fd, void *buf, int nbyte);
  long srfsLseek(int fd, long offset, int whence);
  int  srfsUnlink(const char *pathname);
  int  srfsRename(const char *oldName, const char *newName);

  The protocol is kept very simple:

  - All parameters are sent in network (Most Significant 1st) byte order

  - Each request from a client consists of a 2-byte Func #, a 2-byte Size
    value, then Size # of bytes (if any).

  - Each response from the server consists of 6 result bytes:
    - 2 byte return value
    - 2 byte errno value
    - 2 byte data len
    The 6 result bytes are followed by "len" bytes of data (if any)
    For the lseek function, the 4-byte return value is in the data bytes
===============================================================================
 Update log (most recent first):

  2010-02-05 - Mark Davis (*)
               - Added fatalError flag and set it if the conversation with the
                 SRFS server fails at any point.  Since failing to read or
                 write the expected # of bytes means the conversation is
                 basically out of sync, this means the connection is no
                 longer usable, so an EIO error is returned for the call that
                 had the problem and all subsequent calls.  Prior to this fix,
                 the results of a failed read or write on the socket connection
                 could mean all kinds of bizarre behavior following the error.
               - filefd now initialized to -1 and checked in the close call
                 to avoid sending a close command for a remote file that was
                 never opened.  This gets rid of all the "Bad file number"
                 errors in the srfsd.log on the server.


  2010-01-22 - Mark Davis (*)
               - Changed all read()/write() calls to readBytes()/writeBytes()
                 with a timeout.

  2009-04-15 - Mark Davis
               - Fixed unlink() so it works again (the pathnames it is called
                 with now include the "/srfs/" prefix, which I believe where
                 removed by an earlier version of fcntl.c

  2009-04-13 - Mark Davis
               - Fixed rename() so it works again (the pathnames it is called
                 with now includes the "/srfs/" prefix, which I believe was
                 removed by an earlier version of fcntl.c

  2009-03-24 - Mark Davis
               - added _srfsEntry_() function for getting info about SRFS fds

  2007-10-05 - Mark Davis
               - added support for unlink() and lseek() calls

===============================================================================
Notes:

  Opening a connection to the SRFS server is currently driver-specific (i.e.
  it only supports using the socket driver right now).

  Opening an SRFS file results in 2 fd's:  One for the SRFS driver, and one
  for the driver that the SRFS layer used to connect to the SRFS server

  Once the connection is open, it uses the standard fcntl-layer funcs (read,
  write) to send requests to and wait for responses from the SRFS server.
  This means that adding support for establishing the connection via other
  ioDev drivers is all that is needed to support other types of connections.

  LOCKING:
    The big-kernel-lock (KLOCK) cannot be asserted around any call that might
    block, as this could lock up the system.

    Because SRFS fd's rely on another fd (currently a socket) to communicate
    with the SRFS server, we must be careful not to assert a lock around any
    call that might block, such as use of the fd used to communicate with
    the server.

    Currently this isn't too bad as there is a 1-to-1 relationship between
    the srfs fd and the socket fd.  But the disadvantages to this are that
    each socket requires minutes to cleanly shut down, so any activity that
    involves creating and destroying SRFS fds (such as backupd comparing the
    date/time on a local files with backup copies on the server) will cause
    ports to be tied up in WAIT states.

    It DOES appear that the DC IP stack is smart enough to allow me to reuse
    a socket buffer while the last connection it was used for is still in
    a WAIT state, so I am guessing the only real issue for the controller
    is how many ESTABLISHED connections I have at once, NOT how many
    previous connections are still in a WAIT state (although there must
    still be SOME sort of limit, as I doubt it can keep track of thousands
    or even hundreds of connections that have entered a WAIT state).

    The same issues apply to the server.  Although it has far more
    resources, it is multiplied by the the number of controllers connecting
    to it, so it too can end up with many ports in a WAIT state.  And
    while not a big issue, the server must currently fork off a new srfsd
    process every time a controller creates a new SRFS fd.

    A better system would be to have a separate srfsd task that keeps a
    single connection open to the SRFS server, and have this driver send
    and receive msgs to it as needed.  This would eliminate the problems that
    can be caused by frequent opening and closing of SRFS fds, since a single
    connection to the server would be opened once and left open (aside from
    connection errors, of course).

=============================================================================*/


#include <common.h>

#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>

#include <errno.h>

#include <sys/farprintf.h>

#include <services.h>

#include <linux/limits.h>


typedef struct {
  bool   inUse;
  bool   connected;       // socket connection established
  bool   busy;            // lock to avoid keepalive pkt interference
  bool   fatalError;      // sequence of conversation broken, unable to resume
  int    sockfd;          // descriptor for socket connection
  FDEV far *sockFile;     // global ftable entry for connection to srfs server
  int    filefd;          // descriptor for file on the server
  char far *name;         // name of remote file
  struct sockaddr_in  srvAddr;
  ulong  lastSendTime;    // time of last xmit on sockfd
} SRFS_File;


char far srfsFileName[MaxSRFSFiles][NAME_MAX];


IODEV_FUNCS(srfs)

FILESYS_FUNCS(srfs)


static SRFS_File srfsFile[MaxSRFSFiles];


#define SRFS  ((SRFS_File *)ioDev)

//--- request IDs for SRFS function "calls" ---
#define SRFS_KEEPALIVE  0
#define SRFS_OPEN       1
#define SRFS_CLOSE      2
#define SRFS_READ       3
#define SRFS_WRITE      4
#define SRFS_LSEEK      5
#define SRFS_FSTAT      6
#define SRFS_UNLINK     7
#define SRFS_RENAME     8
#define SRFS_BYE       99  // close connection to server


#define  ShowProcess  (DiagFlags & 0x00000001)

ulong  srfsIPAddr, srfsAltAddr;

#define  DEF_SRFS_ADDR   "192.168.4.34"  // icarus
#define  DEF_ALT_ADDR    "192.168.4.33"  // epics3rd


//-----------------------------------------------------------------------------
//  func called once by init process to initialize data structures
//-----------------------------------------------------------------------------
void srfsInit_()
{
  BZERO(srfsFile);

  inet_ptoh(AF_INET, DEF_SRFS_ADDR, &srfsIPAddr);
  inet_ptoh(AF_INET, DEF_ALT_ADDR, &srfsAltAddr);

  registerFileSys(&srfsFSysFuncs);
}


//-----------------------------------------------------------------------------
//  Returns true if provided pointer is a valid, in-use SRFS device.
//  Returns false if is not.
//-----------------------------------------------------------------------------
bool  validSRFS(void *ioDev)
{
  if (ioDev)
    if (SRFS == srfsFile + (int)(SRFS - srfsFile))
      if (SRFS->inUse)  return true;

  return false;
}


//-----------------------------------------------------------------------------
//  establish a connection to the SRFS server
//  Uses caller-provided sockaddr_in structure so it can be reentrant
//-----------------------------------------------------------------------------
int srfsConnect(struct sockaddr_in *srvAddr)
{
  int  sockfd;


//--- create a socket for it ---
  if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)  return -1;

//--- set addr info for remote end of connection ---
  srvAddr->sin_family = AF_INET;
  srvAddr->sin_port   = htons(SRFS_PORT);
  srvAddr->sin_addr   = htonl(srfsIPAddr);

//--- establish a connection with the SRFS server ---
  if (connect(sockfd, (struct sockaddr *) srvAddr,
              sizeof(struct sockaddr_in) ) < 0)  {
    _closeFD_(current, sockfd);  return -1; }

  return sockfd;
}


//-----------------------------------------------------------------------------
//  Read the response to last SRFS request
//
//  Returns -1 if an error
//
//  On success:
//  - returns 0
//  - *stat = status returned by the server
//  - *dataLen = length of any data returned by request
//  - dataBuf contains any data returned by the request
//-----------------------------------------------------------------------------
int readResp(int sockfd, int *stat, int *errnum, int *dataLen,
             char *dataBuf, int maxLen)
{
  char  buf[6];
  char  *ptr;


  if (read_(sockfd, buf, 6, 10000) < 6)  {
    perror("srfs: readResp(): read() error");
    *stat = -1;  *errnum = ERRNO;  *dataLen = 0;  return -1;
  }

  ptr = buf;
  *stat = getShort(&ptr);  *errnum = getShort(&ptr);  *dataLen = getShort(&ptr);

  if (ShowProcess)
    fprintf(DiagLog, "  stat: %d, errnum: %d, dataLen: %d\r\n",
            *stat, *errnum, *dataLen);

  if (*dataLen > maxLen)  {
    if (ShowErrors)
      fprintf(DiagLog, "srfs: readResp(): response data too large for buffer\r\n");
    *stat = -1;  *errnum = ENOMEM;  *dataLen = 0;  return -1;
  }

  if (*dataLen > 0)
    if (read_(sockfd, dataBuf, *dataLen, 10000) != *dataLen)  {
      perror("srfs: readResp(): read() error");
      *stat = -1;  *errnum = ERRNO;  *dataLen = 0;  return -1;
    }

  return 0;
}


//=============================================================================
//  FileSysFuncs that do not pass an ioDev object
//=============================================================================

//-----------------------------------------------------------------------------
//  open a connection to an SRFS server
//-----------------------------------------------------------------------------
fd_t _srfsOpen_(const char *path, int flags, int mode)
{
  int  i, fd, nameLen, stat, errnum, respLen, rdStat;
  SRFS_File  *srfs;
//  FDEV far  *file;
  char  *ptr;
  char  buf[16];


//--- ignore names that do not start with "/srfs/" ---
  if (strncmp(path, "/srfs/", 6))  { errno = ENXIO;  return -1; }
  path += 6;  // skip over "/srfs/" portion

  KLOCK();

//--- locate an unused entry in the list ---
  for (i=0; i<MaxSRFSFiles; ++i)  if (! srfsFile[i].inUse)  break;
  if (i == MaxSRFSFiles)  { KUNLOCK();  ERRNO = ENFILE;  return -1; }
  srfs = srfsFile + i;  srfs->name = srfsFileName[(uint)i];

//--- alloc a file descriptor to ref the SRFS data ---
  if ((fd = _allocFile_(3, &srfsIOFuncs, srfs)) < 0)  { KUNLOCK();  return -1; }

  srfs->inUse = true;  srfs->sockfd = -1;  srfs->filefd = -1;
  srfs->connected = srfs->fatalError = false;
  fstrcpy(srfs->name, path);

  KUNLOCK();

  if ((srfs->sockfd = srfsConnect(&srfs->srvAddr)) < 0)  {
    close(fd);  return -1; }

  KLOCK();
  srfs->sockFile = _fdToOpenFile_(srfs->sockfd);
  srfs->connected = true;  srfs->lastSendTime = GetMS();
  KUNLOCK();

//temp means of telling srfs server that we will send keepalive packets
#define SRFS_O_TIMEOUT   0x8000   // abort if idle for too long

//--- send OPEN request ---
  ptr = buf;  nameLen = strlen(path);
  addShort(&ptr, SRFS_OPEN);  addShort(&ptr, 4+nameLen);
  addShort(&ptr, flags | SRFS_O_TIMEOUT);  addShort(&ptr, mode);

  srfs->busy = true;
  if (write_(srfs->sockfd, buf, 8, 10000) < 8)  {
    perror("_srfsOpen_(): write_() error");  close(fd);  return -1; }
  if (write_(srfs->sockfd, (char *)path, nameLen, 10000) < nameLen)  {
    perror("_srfsOpen_(): write_() error");  close(fd);  return -1; }

//--- wait for response to OPEN request ---
  rdStat = readResp(srfs->sockfd, &stat, &errnum, &respLen, buf, sizeof(buf));
  if (rdStat OR (stat < 0))  {
    close(fd);  KLOCK();  ERRNO = errnum;  KUNLOCK();  return -1; }
  srfs->filefd = stat;

//--- everything OK - return the descriptor assigned for this process ---
  srfs->busy = false;  return fd;
}

//-----------------------------------------------------------------------------
//  func called each kernel tick to provide file system with CPU cycles
//  Should ONLY be called by the kernel task!
//-----------------------------------------------------------------------------
void _srfsTick_(void)
{
  static uint  tickCount;
  uint  ui;
  int  stat;
  SRFS_File  *srfs;
  char  *ptr;
  FDEV far  *file;
  char  buf[4];


  if (current != kernelTask)  return;

  if (++tickCount & 0x3F)  return;  // don't need to check very often

//===== send "keepalive" packet for any (otherwise) idle SRFS connection =====

//--- check idle time for each open connection ---
  for (ui=0; ui<MaxSRFSFiles; ++ui)  {
    srfs = srfsFile + ui;
    if (! srfs->inUse)  continue;
    if (msSince(srfs->lastSendTime) < 10000)  continue;
    if (! srfs->connected)  continue;
    if (srfs->busy OR srfs->fatalError)  continue;

  //--- send KEEPALIVE "cmd" (no response sent by server) ---
  //--- Remember: This is called by the kernel, so we can't call write() ---
  //--- with the owning process' fd for the socket ---
  //--- Also, the busy flag avoids concurrency issues (i.e. it makes sure ---
  //--- the kernel does not preempt some other task's use of the socket) ---
    ptr = buf;  addShort(&ptr, SRFS_KEEPALIVE);  addShort(&ptr, 0);
    file = srfs->sockFile;
    stat = (file->ioFuncs->write)(file->ioDev, buf, 4, true);
    if (stat == 4)
      srfs->lastSendTime = GetMS();
    else  {
      srfs->fatalError = true;
      fprintf(DiagLog, "_srfsTick_(): keepalive failed for [%d][%ls]\n",
              ui, srfsFileName[ui]);
    }
  }
}


//-----------------------------------------------------------------------------
//  Return the state of the specified ioDev entry or NULL if not a valid
//  entry #.
//
//  For valid entrys, the entryState struct is filled in
//-----------------------------------------------------------------------------
void *_srfsEntry_(uint idx, EntryState *entryState)
{
  SRFS_File  *srfs;

  if (idx >= MaxSRFSFiles)  return NULL;

  srfs = srfsFile + idx;

  entryState->openCount = (srfs->inUse ? 1 : 0);
  entryState->unused = (entryState->openCount == 0);

  return srfs;
}


//-----------------------------------------------------------------------------
//  Attempt to delete a remote file via the SRFS protocol
//
//  Should only be called from the fcntl.c/unlink() function!
//
//  RETURN VALUEs:
//  - if successful: 0
//  - if an error: -1 and ERRNO set accordingly
//-----------------------------------------------------------------------------
int _srfsUnlink_(const char *path)
{
  char  localBuf[4];
  int  sockfd, stat, respLen, errnum, pathLen;
  char  *ptr;
  struct sockaddr_in  srvAddr;


  if (ShowProcess)
    fprintf(DiagLog, "srfsUnlink(%.*s)\r\n", PATH_MAX, path);

//--- ignore names that do not start with "/srfs/" ---
  if (strncmp(path, "/srfs/", 6))  { errno = ENXIO;  return -1; }
  path += 6;  // skip over "/srfs/" portion

  if ((sockfd = srfsConnect(&srvAddr)) < 0)  return -1;

  pathLen = strlen(path) + 1;

//--- send UNLINK request and the data ---
  ptr = localBuf;
  addShort(&ptr, SRFS_UNLINK);  addShort(&ptr, pathLen);
  if (write_(sockfd, localBuf, 4, 10000) != 4)  {
    perror("_srfsUnlink_()");  return -1; }
  if (write_(sockfd, (char *)path, pathLen, 10000) != pathLen)  {
    perror("_srfsUnlink_()");  close(sockfd);  return -1;  }

//--- wait for response to UNLINK request ---
  readResp(sockfd, &stat, &errnum, &respLen, localBuf, sizeof(localBuf));

//--- do a clean, polite close of the connection ---
  ptr = localBuf;
  addShort(&ptr, SRFS_BYE);  addShort(&ptr, 0);
  if (write_(sockfd, localBuf, 4, 10000) != 4)
    perror("_srfsUnlink_(): closing srfs connection");
  close(sockfd);

  if (errnum)  ERRNO = errnum;

  return stat;
}


//-----------------------------------------------------------------------------
//  Attempt to rename a remote file via the SRFS protocol
//
//  Should only be called from the fcntl.c/rename() function!
//
//  RETURN VALUEs:
//  - if successful: 0
//  - if an error: -1 and ERRNO set accordingly
//-----------------------------------------------------------------------------
int _srfsRename_(const char *oldPath, const char *newPath)
{
  char  localBuf[4];
  int  sockfd, stat, respLen, errnum, oldPathLen, newPathLen;
  char  *ptr;
  struct sockaddr_in  srvAddr;


  if (ShowProcess)
    fprintf(DiagLog, "srfsRename(%.*s, %.*s)\r\n",
           PATH_MAX, oldPath, PATH_MAX, newPath);

  if ((sockfd = srfsConnect(&srvAddr)) < 0)  return -1;

//--- ignore names that do not start with "/srfs/" ---
  if (strncmp(oldPath, "/srfs/", 6))  { errno = ENXIO;  return -1; }
  if (strncmp(newPath, "/srfs/", 6))  { errno = ENXIO;  return -1; }
  oldPath += 6;  newPath += 6;  // skip over "/srfs/" portion


  oldPathLen = strlen(oldPath) + 1;  newPathLen = strlen(newPath) + 1;

//--- send WRITE request and the data ---
  ptr = localBuf;
  addShort(&ptr, SRFS_RENAME);  addShort(&ptr, oldPathLen + newPathLen);
  if (write_(sockfd, localBuf, 4, 10000) != 4)  {
    perror("_srfsRename_()");  return -1; }
  if (write_(sockfd, (char *)oldPath, oldPathLen, 10000) != oldPathLen)  {
    perror("_srfsRename_()");  close(sockfd);  return -1;  }
  if (write_(sockfd, (char *)newPath, newPathLen, 10000) != newPathLen)  {
    perror("_srfsRename_()");  close(sockfd);  return -1;  }

//--- wait for response to RENAME request ---
  readResp(sockfd, &stat, &errnum, &respLen, localBuf, sizeof(localBuf));
  if (errnum)  ERRNO = errnum;

//--- do a clean, polite close of the connection ---
  ptr = localBuf;
  addShort(&ptr, SRFS_BYE);  addShort(&ptr, 0);
  if (write_(sockfd, localBuf, 4, 10000) != 4)
    perror("closing srfs connection");

  close(sockfd);

  return stat;
}



//=============================================================================
//  IODev interface functions (for use by fcntl module)
//=============================================================================

//-----------------------------------------------------------------------------
int _srfsGetQueues_(void *ioDev, WaitQueues far *queues)
{
  queues->readQueue = queues->writeQueue = NULL;
  return 0;
}

//-----------------------------------------------------------------------------
//  close an SRFS connection to a remote file
//-----------------------------------------------------------------------------
int _srfsClose_(void *ioDev)
{
  char  localBuf[16];
  char  *ptr;
  bool  connected;
  int  sockfd, filefd;


  if (!validSRFS(ioDev))  { ERRNO = ENXIO;  return -1; }

  if (ShowProcess)  fprintf(DiagLog, "_srfsClose_(%04X)\r\n", (uint)ioDev);

//--- make a copy of the kernel-space values we need to do the close via ---
//--- user-space calls and then clear the kernel-space values ---
  KLOCK();
  connected = SRFS->connected;  SRFS->connected = false;
  sockfd = SRFS->sockfd;  SRFS->sockfd = -1;
  filefd = SRFS->filefd;  SRFS->filefd = -1;
  SRFS->inUse = false;
  KUNLOCK();

  if (connected)  {
    if (filefd >= 0)  {
      ptr = localBuf;
      addShort(&ptr, SRFS_CLOSE);  addShort(&ptr, 2);  addShort(&ptr, filefd);
      if (write_(sockfd, localBuf, 6, 10000) != 6)  perror("_srfsClose_()");
    }

    ptr = localBuf;
    addShort(&ptr, SRFS_BYE);  addShort(&ptr, 0);
    if (write_(sockfd, localBuf, 4, 10000) != 4)  perror("_srfsClose_()");

    close(sockfd);
  }

  return 0;
}

//-----------------------------------------------------------------------------
//  Attempt to read bytes from a remote file via the SRFS protocol
//
//  Should only be called from the fcntl.c/read() function!
//
//  RETURN VALUEs:
//  - if bytes read: # of bytes read
//  - if no bytes read:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _srfsRead_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  char  localBuf[16];
  int  sockfd, stat, respLen, errnum;
  char  *ptr;


  if (ShowProcess)  fprintf(DiagLog, "_srfsRead_(.., .., %d)\r\n", numBytes);

  if (!validSRFS(ioDev))  { ERRNO = ENXIO;  return -1; }

  if (SRFS->fatalError)  { errno = EIO;  return -1; }

  sockfd = SRFS->sockfd;

//--- send READ request and read parameters ---
  ptr = localBuf;
  addShort(&ptr, SRFS_READ);  addShort(&ptr, 4);
  addShort(&ptr, SRFS->filefd);  addShort(&ptr, numBytes);

  SRFS->busy = true;
  if (write_(sockfd, localBuf, 8, 10000) != 8)  {
    perror("_srfsRead_()");  SRFS->fatalError = true;
    if (errno == EAGAIN)  errno = EIO;
    SRFS->busy = false;  return -1;
  }

  SRFS->lastSendTime = GetMS();

//--- wait for response to READ request ---
  readResp(sockfd, &stat, &errnum, &respLen, buf, numBytes);
  if (errnum)  ERRNO = errnum;
//fprintf(DiagLog, "srfsd-read returned: %d/%d\r\n", stat, errno);  //tdebug

  SRFS->busy = false;  return stat;
}

//-----------------------------------------------------------------------------
//  Attempt to write bytes to remote file via the SRFS protocol
//
//  Should only be called from the fcntl.c/write() function!
//
//  RETURN VALUEs:
//  - if bytes written: # of bytes written
//  - if no bytes written:
//    - if more possible in the future: -1 and ERRNO == EAGAIN
//    - if no more possible (EOF):  0
//-----------------------------------------------------------------------------
int _srfsWrite_(void *ioDev, char *buf, int numBytes, bool dontBlock)
{
  char  localBuf[16];
  int  sockfd, stat, respLen, errnum;
  char  *ptr;


  if (ShowProcess)  fprintf(DiagLog, "_srfsWrite_(.., .., %d)\r\n", numBytes);

  if (!validSRFS(ioDev))  { ERRNO = ENXIO;  return -1; }

  if (SRFS->fatalError)  { errno = EIO;  return -1; }

  sockfd = SRFS->sockfd;

//--- send WRITE request and the data ---
  ptr = localBuf;
  addShort(&ptr, SRFS_WRITE);  addShort(&ptr, 4+numBytes);
  addShort(&ptr, SRFS->filefd);  addShort(&ptr, numBytes);

  SRFS->busy = true;
  if (write_(sockfd, localBuf, 8, 10000) != 8)  {
    perror("_srfsWrite_()");  SRFS->fatalError = true;
    if (errno == EAGAIN)  errno = EIO;
    SRFS->busy = false;  return -1;
  }
  if (write_(sockfd,  buf, numBytes, 10000) != numBytes)  {
    perror("_srfsWrite_()");  SRFS->fatalError = true;
    if (errno == EAGAIN)  errno = EIO;
    SRFS->busy = false;  return -1;
  }

  SRFS->lastSendTime = GetMS();

//--- wait for response to WRITE request ---
  readResp(sockfd, &stat, &errnum, &respLen, localBuf, sizeof(localBuf));
  if (errnum)  ERRNO = errnum;

  SRFS->busy = false;  return stat;
}


//-----------------------------------------------------------------------------
//  Set the current offset for a remote file using the SRFS protocol
//-----------------------------------------------------------------------------
off_t  _srfsLseek_(void *ioDev, off_t offset, int whence)
{
  char  localBuf[16];
  int  sockfd, stat, respLen, errnum;
  char  *ptr;


  if (ShowProcess)
    fprintf(DiagLog, "_srfsLseek_(.., %lu, %d)\r\n", (ulong)offset, whence);

  if (!validSRFS(ioDev))  { ERRNO = ENXIO;  return -1; }

  if (SRFS->fatalError)  { errno = EIO;  return -1; }

  sockfd = SRFS->sockfd;

//--- send LSEEK request and the data ---
  ptr = localBuf;
  addShort(&ptr, SRFS_LSEEK);  addShort(&ptr, 8);
  addShort(&ptr, SRFS->filefd);
  addLong(&ptr, offset);  addShort(&ptr, whence);

  SRFS->busy = true;
  if (write_(sockfd, localBuf, 12, 10000) != 12)  {
    perror("_srfsLseek_()");  SRFS->fatalError = true;
    if (errno == EAGAIN)  errno = EIO;
    SRFS->busy = false;  return -1;
  }

  SRFS->lastSendTime = GetMS();

//--- wait for response to LSEEK request ---
  readResp(sockfd, &stat, &errnum, &respLen, localBuf, sizeof(localBuf));
  if (errnum)  ERRNO = errnum;

  ptr = localBuf;  offset = getLong(&ptr);

  SRFS->busy = false;  return offset;
}


//-----------------------------------------------------------------------------
//  Get information about a remote file using the SRFS protocol
//-----------------------------------------------------------------------------
int  _srfsFstat_(void *ioDev, struct stat *fsbuf)
{
  char  buf[sizeof(struct stat)];
  int  sockfd, stat, respLen, errnum;
  char  *ptr;


  if (ShowProcess)  fprintf(DiagLog, "_srfsFstat_(.., ..)\r\n");

  if (!validSRFS(ioDev))  { ERRNO = ENXIO;  return -1; }

  if (SRFS->fatalError)  { errno = EIO;  return -1; }

  sockfd = SRFS->sockfd;

//--- send FSTAT request ---
  ptr = buf;
  addShort(&ptr, SRFS_FSTAT);  addShort(&ptr, 4);
  addShort(&ptr, SRFS->filefd);  addShort(&ptr, sizeof(buf));

  SRFS->busy = true;
  if (write_(sockfd, buf, 8, 10000) != 8)  {
    perror("_srfsFstat_()");  SRFS->fatalError = true;
    if (errno == EAGAIN)  errno = EIO;
    SRFS->busy = false;  return -1;
  }

  SRFS->lastSendTime = GetMS();

//--- wait for response to FSTAT request ---
  readResp(sockfd, &stat, &errnum, &respLen, buf, sizeof(buf));

  if (stat)
    ERRNO = errnum;
  else  {
    ptr = (char *)buf;
    fsbuf->st_mode  = getLong(&ptr);  // protection and file type bits
    fsbuf->st_rdev  = getLong(&ptr);  // device type
    fsbuf->st_size  = getLong(&ptr);  // total size in bytes
    fsbuf->st_atime = getLong(&ptr);  // time of last access
    fsbuf->st_mtime = getLong(&ptr);  // time of last modification
    fsbuf->st_ctime = getLong(&ptr);  // time of last status change
  }

  SRFS->busy = false;  return stat;
}

//-----------------------------------------------------------------------------
//  Kernel-space function to handle ioctl calls for srfs ports
//-----------------------------------------------------------------------------
int _srfsioctl_(void *ioDev, int request, void *arg)
{
  if (!validSRFS(ioDev))  { ERRNO = ENXIO;  return -1; }

  if (SRFS->fatalError)  { errno = EIO;  return -1; }

//finish:  need to implement this...

  ERRNO = EINVAL;  return -1;
}


//-----------------------------------------------------------------------------
//  Kernel-space function to return a description of the srfs port
//
//  WARNING!  To be called ONLY by KERNELFUNC functions (or functions called BY
//            KERNELFUNC functions) !!!!
//-----------------------------------------------------------------------------
char *_srfsDescr_(void *ioDev, char *buf, int bufSize)
{
  char  *ptr;

  sprintf(buf, "srfs 0x%04lX ", (ulong)ioDev);

  ptr = buf + strlen(buf);  strcpy(ptr, "bad bad");

//finish: limit output to bufSize

  if (validSRFS(ioDev))
    sprintf(ptr, "srfs[%d] %ls", (int)(SRFS - srfsFile), SRFS->name);

  return buf;
}


//-----------------------------------------------------------------------------

