//----- linux/wait.h ----- 04/20/12 -----(06/17/07)-----

//-----------------------------------------------------------------------------
//  Significantly modified version of standard linux header file
//
//   See wait.c for details.
//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2012-04-20 - Mark Davis
//               - Changed the name of listAll() to wqListTasks().
//               - Added wqCounts() to get total and free counts for the
//                 global pool of waitqueues.
//
//  2010-09-09 - Mark Davis (*)
//               - Added myAddr ptr to the WaitQueueNode structure to allow
//                 checking for invalid pointers.
//
//  2010-08-11 - Mark Davis
//               - changed the first argument to removeFromQueues() to a
//                 pointer to a task rather the task's list of waitqueue nodes.
//
//  2010-08-06 - Mark Davis
//               - changed 'void *private' to 'task_struct *task' in the
//                 __wait_queue_node structure.
//
//  2010-07-23 - Mark Davis
//               - Added __first_node() macro
//               - Added obj and removeFunc to __wait_queue_node structure.
//
//  2010-05-12 - Mark Davis
//               - replaced clearQueue() with wakeTasks() and removeFromQueues()
//
//  2010-04-29 - Mark Davis
//               - Significant overhaul to simplify and support the ability
//                 to have one task block on multiple wait queues (which is
//                 needed to support funcs like poll() and select(), which
//                 are needed for other projects).
//
//  2007-06-17 - Mark Davis
//               - Original version (?)
//
//-----------------------------------------------------------------------------

#ifndef _LINUX_WAIT_H
#define _LINUX_WAIT_H


#include <linux/dummy.h>
#include <sys/types.h>


#define WNOHANG     0x00000001
#define WUNTRACED   0x00000002
#define WSTOPPED    WUNTRACED
#define WEXITED     0x00000004
#define WCONTINUED  0x00000008
#define WNOWAIT     0x01000000  /* Don't reap, just poll status.  */

#define __WNOTHREAD 0x20000000  /* Don't wait on children of other threads in this group */
#define __WALL      0x40000000  /* Wait on all children, regardless of type */
#define __WCLONE    0x80000000  /* Wait only on non-SIGCHLD children */

/* First argument to waitid: */
#define P_ALL   0
#define P_PID   1
#define P_PGID  2

#ifdef __KERNEL__

#include <linux/list.h>
//#include <linux/stddef.h>
//#include <linux/spinlock.h>
//#include <asm/system.h>
//#include <asm/current.h>


typedef struct __wait_queue_node  WaitQueueNode;


typedef int (*WaitQueueFunc)(WaitQueueNode *node, unsigned mode, int sync, void *key);
int default_wake_function(WaitQueueNode *node, unsigned mode, int sync, void *key);


typedef void (*RemoveNode)(WaitQueueNode *node);


#define WQ_FLAG_EXCLUSIVE 0x01

struct task_struct;

//------------------------------------------------------------------------
//  In the DEBROS model each node belongs to 2 lists (in the Linux model,
//  they belong to just 1).
//
//   taskList:
//      The list of tasks waiting in a given wait queue.  The list head
//      is part of the structure used to keep track of the state of the
//      object that represents the resource the task is waiting for.  This
//      is the one used in the Linux implementation.
//
//   queueList:
//      The list of wait queues (resources) a task is blocked on.  The
//      list head is part of the task structure.  This is used to remove
//      all the nodes for a task from the lists they are in.  Used for
//      the poll() function and when killing a task.
//
//  This method provides a practical (and efficient) means of supporting
//  the ability for a single task to be in multiple wait queues (which is
//  necessary to support functions like poll(), select(), and semaphores).
//
//  Other changes/additions for the DEBROS model:
//    - obj:
//         The object/structure the list belongs to (or NULL)
//    - removeFunc:
//         A pointer to the function to call when removing the node.  This
//         allows the object to change its state when nodes are removed from
//         one of its lists.  If the object has multiple lists, you may need
//         to have one function per list to efficiently determine which list
//         is being updated (and therefore how the object is affected).
//------------------------------------------------------------------------
struct __wait_queue_node {
  uint                 flags;
  struct task_struct  *task;        // the task that added the node
  void                *obj;         // the device/resource the list belongs to
  RemoveNode           removeFunc;  // func to call when removing this node
//  WaitQueueFunc        func;       // not implemented at this time
  struct list_head     taskList;    // links all the tasks in the same queue
  struct list_head     queueList;   // links all the nodes for a given task
  void                *myAddr;      // to validate ptrs to this structure
};


struct __wait_queue_head {
//  spinlock_t  lock;
  struct list_head  listHead;  // list of WaitQueueNode's
};
typedef struct __wait_queue_head  WaitQueue;



/*----------------------------------------------------------------------------*
 * Macros for declaration and initializaton of the datatypes
 *----------------------------------------------------------------------------*/

#define __WAITQUEUE_INITIALIZER(name, tsk) {  \
  /* .flags = */  0,                       \
  /* .task  = */ tsk,                   \
  /* .func   = */ default_wake_function,   \
  /* .taskList  = */ { NULL, NULL },      \
  /* .devList = */   { NULL, NULL }       \
}

#define DECLARE_WAITQUEUE(name, tsk)          \
  WaitQueueNode  name = __WAITQUEUE_INITIALIZER(name, tsk)


#define __WAIT_QUEUE_HEAD_INITIALIZER(name) {       \
  /* .listHead  = */  { &(name).listHead, &(name).listHead } }

#define DECLARE_WAIT_QUEUE_HEAD(name) \
  WaitQueue  name = __WAIT_QUEUE_HEAD_INITIALIZER(name)


extern void init_waitqueue_head(WaitQueue *q);


#define init_waitqueue_entry(node, task_)  \
{  \
  (node)->flags = 0;  \
  (node)->task = (task_);  \
  (node)->obj = 0;  \
  (node)->removeFunc = NULL;  \
  (node)->func = default_wake_function; \
}


#define init_waitqueue_func_entry(node, func_)  \
{  \
  (node)->flags = 0;  \
  (node)->task = NULL;  \
  (node)->obj = 0;  \
  (node)->removeFunc = NULL;  \
  (node)->func = (func_);  \
}


#define waitqueue_active(q)  ( (int) (! list_empty(&(q)->listHead)) )


/*
 * Used to distinguish between sync and async io wait context:
 * sync i/o typically specifies a NULL wait queue entry or a wait
 * queue entry bound to a task (current task) to wake up.
 * aio specifies a wait queue entry with an async notification
 * callback routine, not associated with any task.
 */
#define is_sync_wait(node)  (!(node) || ((node)->task))


//--- funcs that expect caller to provide the lock that protects the lists ---

void  __add_wait_queue(WaitQueue *wq, WaitQueueNode *node);
void  __add_wait_queue_tail(WaitQueue *wq, WaitQueueNode *node);

extern int  __removeWaitQueue(WaitQueue *q, WaitQueueNode *node);


//---- funcs that provide the lock that protects the lists ---
extern void  add_wait_queue(WaitQueue *q, WaitQueueNode *node);
extern void  add_wait_queue_exclusive(WaitQueue *q, WaitQueueNode *node);

extern void  removeWaitQueue(WaitQueue *q, WaitQueueNode *node);



#define  NodeTask(node)  ((task_t *)((node)->task))


//void FASTCALL(__wake_up(WaitQueue *q, unsigned int mode, int nr, void *key));
//extern void FASTCALL(__wake_up_locked(WaitQueue *q, unsigned int mode));
//extern void FASTCALL(__wake_up_sync(WaitQueue *q, unsigned int mode, int nr));
//void FASTCALL(__wake_up_bit(WaitQueue *, void *, int));
//int FASTCALL(__wait_on_bit(WaitQueue *, struct wait_bit_queue *, int (*)(void *), unsigned));
//int FASTCALL(__wait_on_bit_lock(WaitQueue *, struct wait_bit_queue *, int (*)(void *), unsigned));
//void FASTCALL(wake_up_bit(void *, int));
//int FASTCALL(out_of_line_wait_on_bit(void *, int, int (*)(void *), unsigned));
//int FASTCALL(out_of_line_wait_on_bit_lock(void *, int, int (*)(void *), unsigned));
//WaitQueue *FASTCALL(bit_waitqueue(void *, int));

//#define wake_up(x)      __wake_up(x, TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE, 1, NULL)
//#define wake_up_nr(x, nr)   __wake_up(x, TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE, nr, NULL)
//#define wake_up_all(x)      __wake_up(x, TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE, 0, NULL)
//#define wake_up_interruptible(x)  __wake_up(x, TASK_INTERRUPTIBLE, 1, NULL)
//#define wake_up_interruptible_nr(x, nr) __wake_up(x, TASK_INTERRUPTIBLE, nr, NULL)
//#define wake_up_interruptible_all(x)  __wake_up(x, TASK_INTERRUPTIBLE, 0, NULL)
//#define wake_up_locked(x)   __wake_up_locked((x), TASK_UNINTERRUPTIBLE | TASK_INTERRUPTIBLE)
//#define wake_up_interruptible_sync(x)   __wake_up_sync((x),TASK_INTERRUPTIBLE, 1)

/*
DEFINE_WAIT(name_)  \
  WaitQueueNode  name_;  \
  name_.flags = 0;  name_.task  = current;  \
  name_.func = autoremove_wake_function;  \
  name_.taskList.next = name_.taskList.prev = &name_.taskList;   \
  name_.devList.next = name_.devList.prev = &name_.devList;
*/


//-----------------------------------------------------------------------------
//  Use the specified WaitQueue to efficiently wait for a specified condition
//-----------------------------------------------------------------------------
#define __wait_event(queue, condition)   \
do {    \
  WaitQueueNode *node = alloc_WaitQueueNode();  \
  for (;;)  {  \
    prepare_to_wait(&(queue), node, TASK_UNINTERRUPTIBLE);  \
    if (condition)  break;  \
    schedule();  \
  }  \
  finish_wait(&(queue), node);  \
} while (0)


/*----------------------------------------------------------------------------*
 * wait_event - sleep until a condition gets true
 * @wq: the waitqueue to wait on
 * @condition: a C expression for the event to wait for
 *
 * The process is put to sleep (TASK_UNINTERRUPTIBLE) until the
 * @condition evaluates to true. The @condition is checked each time
 * the waitqueue @wq is woken up.
 *
 * wake_up() has to be called after changing any variable that could
 * change the result of the wait condition.
 *----------------------------------------------------------------------------*/
#define wait_event(wq, condition)           \
do {                  \
  if (condition)  break;         \
  __wait_event(wq, condition);          \
} while (0)


/*----------------------------------------------------------------------------*
 *
 *----------------------------------------------------------------------------*
#define __wait_event_timeout(wq, condition, ret)      \
do {                  \
  DEFINE_WAIT(__wait);            \
                  \
  for (;;) {              \
    prepare_to_wait(&wq, &__wait, TASK_UNINTERRUPTIBLE);  \
    if (condition) break;            \
    ret = schedule_timeout(ret);        \
    if (!ret)  break;       \
  }       \
  finish_wait(&wq, &__wait);          \
} while (0)
*/


/*----------------------------------------------------------------------------*
 * wait_event_timeout - sleep until a condition gets true or a timeout elapses
 * @wq: the waitqueue to wait on
 * @condition: a C expression for the event to wait for
 * @timeout: timeout, in jiffies
 *
 * The process is put to sleep (TASK_UNINTERRUPTIBLE) until the
 * @condition evaluates to true. The @condition is checked each time
 * the waitqueue @wq is woken up.
 *
 * wake_up() has to be called after changing any variable that could
 * change the result of the wait condition.
 *
 * The function returns 0 if the @timeout elapsed, and the remaining
 * jiffies if the condition evaluated to true before the timeout elapsed.
 *----------------------------------------------------------------------------*/
#define wait_event_timeout(wq, condition, timeout)      \
({                  \
  long __ret = timeout;           \
  if (!(condition))             \
    __wait_event_timeout(wq, condition, __ret);   \
  __ret;                \
})


#define __wait_event_interruptible(wq, condition, ret)      \
do {                  \
  DEFINE_WAIT(__wait);            \
                  \
  for (;;) {              \
    prepare_to_wait(&wq, &__wait, TASK_INTERRUPTIBLE);  \
    if (condition)            \
      break;            \
    if (!signal_pending(current)) {       \
      schedule();         \
      continue;         \
    }             \
    ret = -ERESTARTSYS;         \
    break;              \
  }               \
  finish_wait(&wq, &__wait);          \
} while (0)


/*----------------------------------------------------------------------------*
 * wait_event_interruptible - sleep until a condition gets true
 * @wq: the waitqueue to wait on
 * @condition: a C expression for the event to wait for
 *
 * The process is put to sleep (TASK_INTERRUPTIBLE) until the
 * @condition evaluates to true or a signal is received.
 * The @condition is checked each time the waitqueue @wq is woken up.
 *
 * wake_up() has to be called after changing any variable that could
 * change the result of the wait condition.
 *
 * The function will return -ERESTARTSYS if it was interrupted by a
 * signal and 0 if @condition evaluated to true.
 *----------------------------------------------------------------------------*/
#define wait_event_interruptible(wq, condition)       \
({                  \
  int __ret = 0;              \
  if (!(condition))  __wait_event_interruptible(wq, condition, __ret); \
  __ret;                \
})


#define __wait_event_interruptible_timeout(wq, condition, ret)    \
do {                  \
  DEFINE_WAIT(__wait);            \
                  \
  for (;;) {              \
    prepare_to_wait(&wq, &__wait, TASK_INTERRUPTIBLE);  \
    if (condition)            \
      break;            \
    if (!signal_pending(current)) {       \
      ret = schedule_timeout(ret);      \
      if (!ret)   break;          \
      continue;         \
    }             \
    ret = -ERESTARTSYS;         \
    break;              \
  }               \
  finish_wait(&wq, &__wait);          \
} while (0)


/*----------------------------------------------------------------------------*
 * wait_event_interruptible_timeout - sleep until a condition is true or a
 *                                    timeout elapses
 * @wq: the waitqueue to wait on
 * @condition: a C expression for the event to wait for
 * @timeout: timeout, in jiffies
 *
 * The process is put to sleep (TASK_INTERRUPTIBLE) until the
 * @condition evaluates to true or a signal is received.
 * The @condition is checked each time the waitqueue @wq is woken up.
 *
 * wake_up() has to be called after changing any variable that could
 * change the result of the wait condition.
 *
 * The function returns 0 if the @timeout elapsed, -ERESTARTSYS if it
 * was interrupted by a signal, and the remaining jiffies otherwise
 * if the condition evaluated to true before the timeout elapsed.
 *----------------------------------------------------------------------------*/
#define wait_event_interruptible_timeout(wq, condition, timeout)  \
({                  \
  long __ret = timeout;           \
  if (!(condition)) __wait_event_interruptible_timeout(wq, condition, __ret); \
  __ret;                \
})


#define __wait_event_interruptible_exclusive(wq, condition, ret)  \
do {                  \
  DEFINE_WAIT(__wait);            \
                  \
  for (;;) {              \
    prepare_to_wait_exclusive(&wq, &__wait, TASK_INTERRUPTIBLE);    \
    if (condition)  break;            \
    if (!signal_pending(current)) { schedule();  continue; }             \
    ret = -ERESTARTSYS;         \
    break;              \
  }               \
  finish_wait(&wq, &__wait);          \
} while (0)

#define wait_event_interruptible_exclusive(wq, condition)   \
({                  \
  int __ret = 0;              \
  if (!(condition)) __wait_event_interruptible_exclusive(wq, condition, __ret);\
  __ret;                \
})


/*----------------------------------------------------------------------------*
 * Must be called with the spinlock in the WaitQueue held.
 *----------------------------------------------------------------------------*/
//static inline void add_wait_queue_exclusive_locked(WaitQueue *q,
//               WaitQueueNode * wait)
//{
//  wait->flags |= WQ_FLAG_EXCLUSIVE;
//  __add_wait_queue_tail(q,  wait);
//}
#define add_wait_queue_exclusive_locked(q, wait_)  \
{  \
  (wait_)->flags |= WQ_FLAG_EXCLUSIVE;  \
  __add_wait_queue_tail(q,  (wait_));  \
}

/*----------------------------------------------------------------------------*
 * Must be called with the spinlock in the WaitQueue held.
 *----------------------------------------------------------------------------*/
//static inline void remove_wait_queue_locked(WaitQueue *q,
//              WaitQueueNode * wait)
//{
//  __remove_wait_queue(q,  wait);
//}
#define remove_wait_queue_locked(q, wait_)  \
{  \
  __remove_wait_queue(q, (wait_));  \
}



//--- WaitQueueNodes which are removed from the lists at wakeup time ---

void prepare_to_wait(WaitQueue *q, WaitQueueNode *wait, int state);

void prepare_to_wait_exclusive(WaitQueue *q, WaitQueueNode *wait, int state);


void finish_wait(WaitQueue *q, WaitQueueNode *wait);


int autoremove_wake_function(WaitQueueNode *wait,
                             unsigned mode, int sync, void *key);

//int wake_bit_function(WaitQueueNode *wait, unsigned mode, int sync, void *key);


#define DEFINE_WAIT(name)           \
  WaitQueueNode name = {           \
    /* .flags = */  0,  \
    /* .task  = */  (void *) current,        \
//    /* .func   = */   autoremove_wake_function,   \
    /* .task_list = */  { &(name).task_list, &(name).task_list },  \
    /* .dev_list = */   { &(name).dev_list, &(name).dev_list }     \
  }

#define init_wait(wait)             \
  do {                \
    (wait)->task = current;        \
//    (wait)->func = autoremove_wake_function;    \
    INIT_LIST_HEAD(&(wait)->task_list);     \
  } while (0)


/*----------------------------------------------------------------------------*
 * wait_on_bit - wait for a bit to be cleared
 * @word: the word being waited on, a kernel virtual address
 * @bit: the bit of the word being waited on
 * @action: the function used to sleep, which may take special actions
 * @mode: the task state to sleep in
 *
 * There is a standard hashed waitqueue table for generic use. This
 * is the part of the hashtable's accessor API that waits on a bit.
 * For instance, if one were to have waiters on a bitflag, one would
 * call wait_on_bit() in threads waiting for the bit to clear.
 * One uses wait_on_bit() where one is waiting for the bit to clear,
 * but has no intention of setting it.
 *----------------------------------------------------------------------------*/
/*
static inline int wait_on_bit(void *word, int bit,
        int (*action)(void *), unsigned mode)
{
  if (!test_bit(bit, word))
    return 0;
  return out_of_line_wait_on_bit(word, bit, action, mode);
}
*/

/*----------------------------------------------------------------------------*
 * wait_on_bit_lock - wait for a bit to be cleared, when wanting to set it
 * @word: the word being waited on, a kernel virtual address
 * @bit: the bit of the word being waited on
 * @action: the function used to sleep, which may take special actions
 * @mode: the task state to sleep in
 *
 * There is a standard hashed waitqueue table for generic use. This
 * is the part of the hashtable's accessor API that waits on a bit
 * when one intends to set it, for instance, trying to lock bitflags.
 * For instance, if one were to have waiters trying to set bitflag
 * and waiting for it to clear before setting it, one would call
 * wait_on_bit() in threads waiting to be able to set the bit.
 * One uses wait_on_bit_lock() where one is waiting for the bit to
 * clear with the intention of setting it, and when done, clearing it.
 *----------------------------------------------------------------------------*/
/*
static inline int wait_on_bit_lock(void *word, int bit,
        int (*action)(void *), unsigned mode)
{
  if (!test_and_set_bit(bit, word))
    return 0;
  return out_of_line_wait_on_bit_lock(word, bit, action, mode);
}
*/


//=============================================================================

// NSCL additions

#include <common.h>


#define TaskListOffset   offsetof(WaitQueueNode, taskList)
#define QueueListOffset  offsetof(WaitQueueNode, queueList)

#define __first_task(wq)  \
   list_entry((wq)->listHead.next, WaitQueueNode, taskList)


void initWaitQueues(void);


WaitQueueNode *alloc_WaitQueueNode(void);

void free_WaitQueueNode(WaitQueueNode *node);


void wakeTasks(WaitQueue *wq);  // wakes all tasks in a driver's wait queue

void removeFromQueues(struct task_struct *task, bool freeNodes);  // remove a task from all wait queues


void wqListTasks(WaitQueue *wq);

uint _nearcall wqCounts(uint *ttl);


#endif /* __KERNEL__ */

#endif
