//----- semaphore.c ----- 04/23/12 -----(07/22/10)-----

//=============================================================================
//  POSIX style semaphores
//=============================================================================
// NOTES
//
//=============================================================================
//  Change log (most recent first):
//
//
//  2012-04-23 - Mark Davis
//               - Added semCounts() to get total and free counts for the
//                 global pool of named semaphores.
//
//  2012-04-19 - Mark Davis (*)
//               - Fixed some bugs in sem_wait_() (used by all cases of getting
//                 a lock on a semaphore).  Prior to this, if a task was
//                 resumed but did not get the semaphore, it would still be
//                 in the semaphores blocked queue, which could not only
//                 cause the task to resume when it shouldn't, but could
//                 use up the free waitqueue nodes.
//                 NOTE that this could affect all uses of semaphores (which
//                 includes the poll() function).
//
//  2010-08-12 - Mark Davis
//               - First version to pass basic tests (see sem_test command in
//                 the test application).
//
//  2010-07-22 - Mark Davis
//               - Began work on first version.
//
//=============================================================================

#include <common.h>

#include <linux/sched.h>

#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <signal.h>

#include <sys/time.h>
#include <semaphore.h>

#include <limits.h>


//#define  SEM_DEBUG


static sem_t  semaphores[MAX_NUM_SEMS];
static char far  names[MAX_NUM_SEMS][NAME_MAX];


//-----------------------------------------------------------------------------
//  Notes on software logic:
//-----------------------------
//
//  Managing the state of the semaphores is handled by maintaining 3 lists
//  for each semaphore:
//
//    - The tasks that have opened the semaphore (for named semaphores)
//    - The locks in use
//    - The tasks blocked waiting for a lock
//
//  Locking:
//
//    To avoid concurrency problems and corrupted lists, all scans of and
//    changes to a list must be done using a lock that insures no other task
//    modifies a list while other tasks are reading it or modifying it.  At
//    present this is done by disabling the scheduler, although in the future
//    support for semaphore-specific locking (and even RO vs RW locks) would
//    reduce the effect on latency.  Although this is sort of a chicken-&-egg
//    problem since you would need something like a semaphore to do it...
//
//
//  Adding/Removing nodes to/from the queues/lists:
//
//    It is important to add nodes to the END/tail of queues.  The reason for
//    this is that each node belongs to 2 lists:  One for the resource the
//    node represents, and one for the task that owns or is waiting for the
//    resource.  Because you have to open a named semaphore before you can get
//    a lock (or block) on it, adding nodes to the end of the lists ensures
//    the 'open" nodes will always come before any 'lock' or 'blocked' nodes
//    in each task's list.
//
//    This is important because when a task is killed, all its nodes are
//    removed (and processed) in reverse order (from the tail to the head),
//    causing any locks the task has for a semaphore to be released before its
//    connection to the semaphore is closed.  While doing it in the other
//    order is possible, it would require more time searching through lists,
//    which has to be done inside a lock, which would have a larger impact
//    on latency.
//
//
//  Granting locks in the order they were requested:
//
//    When there are tasks blocked waiting for a lock, we want to insure that
//    the next one in the blocked list gets the next lock that is released.
//
//    To insure this happens, the release (posting) of a lock is handled as
//    follows:
//
//      - The locked node for the task that called post is removed from the
//        lock node queue for the semaphore.
//
//      - If there are no blocked tasks waiting for a lock, the value of the
//        semaphore is increased so the next request for a lock can be
//        satisfied without waiting or blocking.
//
//      - If there are blocked tasks waiting for a lock, the value of the
//        semaphore is NOT increased.  Instead, the node for a blocked task is
//        simply moved from the blocked list to the locked list and the blocked
//        task is resumed.  This insures that the unblocked task owns the
//        just-released lock and that it cannot be given to another task before
//        the unblocked one actually resumes execution.
//
//   NOTE:
//     Currently, the first task (if any) in a semaphore's list of blocked
//     tasks is the one that receives a lock when it is released.  While
//     this a simple and efficient approach that ensures a degree of "fairness"
//     (first come, first serve), it can result in priority-inversion and it
//     violates the priority-based model of the scheduler.  A later version
//     should really call a scheduler function that determines which task
//     should be resumed.
//
//
//  Semaphore pointers:
//
//    The POSIX-style semaphore functions use a pointer to a sem_t value to
//    reference each semaphore.
//
//    For named semaphores, each task gets the pointer value needed to access
//    the semaphore by calling sem_open(), so the pointer value returned could
//    be different for different tasks.  So long as the correct semaphore is
//    accessed for a given pointer, this works just fine.
//
//    For unnamed semaphores, there is no standard way to get a pointer value
//    for the semaphore.  The pointer value must be known BEFORE calling ANY
//    of the semaphore functions.  This includes sem_init(), which must
//    be called just once to initialize an unnamed semaphore before any other
//    call that references the semaphore.
//
//    As with named semaphores, so long as the semaphore functions access the
//    correct semaphore, the value of the pointer COULD be different for each
//    task.  But for there to be any hope of portability, there must be some
//    standard way for multiple tasks to determine what pointer value to use
//    to access a given semaphore.
//
//    The most natural assumption is that the sem_t pointer actually points
//    to a sem_t value in shared memory and is accessable by all the tasks
//    that need to use the semaphore.
//
//    At the time of this writing, DEBROS does not (yet) support the shared
//    memory functions (shm_open, shm_unlink, shmat, shmctl, shmdt, shmget,
//    shmop) which is the only portable way I know of for multiple tasks to
//    get a pointer to an object in shared memory.
//
//    Of course, in DEBROS (and I suspect many Real-Time OS's), all of memory
//    is shared, so any variable declared in global memory is automatically
//    accessible by all tasks.
//
//    So, for DEBROS, an unnamed semaphore can be shared by multiple tasks
//    simply by declaring it in global memory (as opposed to a local variable
//    that resides in a task's stack) and then using its address when
//    calling any of the semaphore functions.
//
//    Note that an unnamed semaphore CAN reside on the stack if it is NOT
//    shared with other tasks.  Although, since DEBROS does not support
//    multiple threads within a task, such a semaphore would probably not be
//    particularly useful.
//
//
//  Assumptions regarding unspecified behavior:
//
//    The manual pages I have found that describe these functions leave out a
//    number of details which I have decided to implement in a manner that
//    makes the most sense to me and which I believe will minimize the chance
//    for problems:
//
//    - Can a task aquire more than one lock on a semaphore?  Nothing I
//      found says one way or the other, and I can envision scenarios where
//      a single task might need to tie up multiple instances of the resource
//      the semaphore represents, so the DEBROS implementation supports this.
//
//    - If a task closes a semaphore when it has 1 or more active locks on
//      the semaphore, what should happen to the locks?  The manual page
//      says that it deallocates any system resources allocated for the
//      process' use of the semaphore.  That would seem to support my belief
//      that the only thing that makes sense is to release the locks before
//      returning from the sem_close call, so that is what the DEBROS
//      implementation does.
//
//    - In many places the manual pages say that the effect of some set of
//      circumstances is undefined (e.g. calling sem_destroy() for a named
//      semaphore).  Wherever possible, I try to detect such circumstances
//      and return an error without altering the state of any semaphore.
//
//-----------------------------------------------------------------------------


#ifdef SEM_DEBUG
//-----------------------------------------------------------------------------
//  For debugging purposes
//-----------------------------------------------------------------------------
void SemMsg(sem_t *sem, const char far *msg)
{
  int  semNum;
  const char far *semName;


  if (current == kernelTask)  return;  //tdebug

  if (sem->name)  {
    semNum = sem - semaphores;  semName = sem->name; }
  else  {
    semNum = -1, semName = "<unnamed>"; }

  fprintf(DiagLog, "Semaphore [%04lX] '%ls'[%d] %ls\r\n",
          (ulong)sem, semName, semNum, msg);
}

#endif


//-----------------------------------------------------------------------------
//  Check a pointer to see if it points to a valid, named semaphore.
//
//  Returns non-0 if the pointer refers to a named semaphore structure that has
//  been initialized via a sem_open() call and is still in use.
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
bool  namedSemaphore(sem_t *sem)
{
  return (!INVALID(sem_t, sem) AND sem->name AND sem->inUse);
}

//-----------------------------------------------------------------------------
//  Check a pointer to see if it points to a valid, unnamed semaphore.
//
//  Returns non-0 if the pointer refers to an unnamed semaphore structure that
//  has been initialized via a sem_init() call and is still in use.
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
bool  unnamedSemaphore(sem_t *sem)
{
  return (!INVALID(sem_t, sem) AND !sem->name AND sem->inUse);
}

//-----------------------------------------------------------------------------
int qListSize(WaitQueue *wq)
{
  int  count;
  list_head_t  *cur, *list;


  list = &wq->listHead;  count = 0;

  list_for_each(cur, list)  ++count;

  return count;
}

//-----------------------------------------------------------------------------
//  Scan the nodes in a queue looking for one that points to a specific task.
//
//  Returns a pointer to the first node found (if any), or NULL if none of the
//  nodes in the list point to the task.
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
WaitQueueNode *taskInQList(WaitQueue *wq, task_t *task)
{
  list_head_t  *cur, *list;
  WaitQueueNode  *node;


  list = &wq->listHead;

  list_for_each(cur, list)  {
    node = list_entry(cur, WaitQueueNode, taskList);
    if (node->task == task)  return node;
  }

  return NULL;
}


//-----------------------------------------------------------------------------
//  Determine if a node is one of the ones in a specified queue.
//
//  This is done by scanning the queue list the node is in to see if one of
//  the nodes is the listhead node for the specified queue.
//
//  Returns true if the node is in the specified queue,  false if it is not.
//
//  NOTE:  This works best if the node is the LAST node in the queue (i.e.
//  the node's next pointer points to the node for the queue's list head).
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
bool nodeInQList(WaitQueue *wq, WaitQueueNode *node)
{
  list_head_t  *cur, *start, *list;


  list = &wq->listHead;  start = &node->taskList;

  for (cur=start->next; cur!=start; cur=cur->next)
    if (cur == list)  return true;

  return false;
}


//-----------------------------------------------------------------------------
//  Free up any locks a task has for a semaphore.  If other tasks are blocked
//  waiting for a lock, one of them is given the lock and unblocked.
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
void semFreeLocks(sem_t *sem, task_t *task)
{
  list_head_t  *cur, *next, *list;
  WaitQueueNode  *node;
  WaitQueue  *locked;


//--- free up any locks the task may have had ---
  locked = &sem->locked;  list = &locked->listHead;

//--- NOTE that removal of a node from the locked list causes ---
//--- semRemovedFromLockedList() to be called, which gives the just-released ---
//---- lock to another task if one is blocked, waiting for a lock ---

  list_for_each_safe(cur, next, list)  {
    node = list_entry(cur, WaitQueueNode, taskList);
    if (node->task == task)  {
      __removeWaitQueue(locked, node);  free_WaitQueueNode(node);  }
  }

}


//-----------------------------------------------------------------------------
//  Called to see if a named, unlinked semaphore can be freed
//
//  Must only be called for named, unlinked semaphores.
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
void semFreeIfNotOpen(sem_t *sem)
{
  WaitQueue  *opened, *locked, *blocked;


  opened = &sem->opened;  if (waitqueue_active(opened))  return;

  blocked = &sem->blocked;  locked = &sem->locked;

  if (waitqueue_active(blocked))  {
    fprintf(DiagLog, "*** %s on fully closed, unlinked semaphore ***\n",
            "blocked tasks");
    return;
  }
  if (waitqueue_active(locked))  {
    fprintf(DiagLog, "*** %s on fully closed, unlinked semaphore ***\n",
            "locks");
    return;
  }

  #ifdef SEM_DEBUG
    SemMsg(sem, "freed");
  #endif

  sem->inUse = sem->unlinked = false;
}



//-----------------------------------------------------------------------------
//  Called just after a node is removed from the "open" list for a semaphore.
//
//  If the node was the last one in the semaphore's list of open connections,
//  and sem_unlink() was called for the semaphore, then the semaphore is
//  destroyed (freed up for use in a sem_open() or sem_init() call).
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
void _nearcall semRemovedFromOpenList(WaitQueueNode *node)
{
  sem_t  *sem = (sem_t *)node->obj;

  if (!namedSemaphore(sem))  return;

  semFreeLocks(sem, node->task);

  if (sem->unlinked)  semFreeIfNotOpen(sem);
}

//-----------------------------------------------------------------------------
//  Release the lock represented by the specified node that was just removed
//  from the locked list.  If others tasks are blocked waiting for a lock, one
//  of them is given the just-released lock and then unblocked.
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
void _nearcall semRemovedFromLockedList(WaitQueueNode *node)
{
  WaitQueue  *blocked;
  sem_t  *sem = (sem_t *)node->obj;
  #ifdef SEM_DEBUG
    char msg[20];
  #endif



  if (INVALID(sem_t, sem) OR !sem->inUse)  { errno = EINVAL;  return; }

  blocked = &sem->blocked;

  #ifdef SEM_DEBUG
    SemMsg(sem, "lock released");
  #endif


//--- If no blocked tasks, increment the semaphore value ---
  if (!waitqueue_active(blocked))  {
    #ifdef SEM_DEBUG
      SemMsg(sem, "lock available");
    #endif
    ++sem->value;  return;
  }

//--- Otherwise move the node for the first blocked task from the blocked ---
//--- list to the locked list.  Note that we add the node to the end of ---
//--- locked list because the unblocked task will call nodeInQList() to ---
//--- see if it has the lock, and nodeInQList() is very fast if the node ---
//--- is at the end of the queue. ---

//--- Also note that we are no longer working with the node that was passed ---
//--- to use (which will most likely be free up in a moment) ---

  node = __first_task(blocked);

  #ifdef SEM_DEBUG
    sprintf(msg, "unblocking [%u]", node->task->pid);
    SemMsg(sem, msg);
  #endif

  __removeWaitQueue(blocked, node);

  node->removeFunc = semRemovedFromLockedList;
  __add_wait_queue_tail(&sem->locked, node);

  node->task->state = TASK_RUNNING;

  #ifdef SEM_DEBUG
//    sprintf(msg, "unblocked [%u]", node->task->pid);
//    SemMsg(sem, msg);
  #endif
}



//-----------------------------------------------------------------------------
//  Initialize the static array of semaphores
//-----------------------------------------------------------------------------
void  initSemaphores(void)
{
  uint  u;
  sem_t  *sem;


  BZERO(semaphores);

  for (u=0,sem=semaphores; u<MAX_NUM_SEMS; ++u,++sem)  {
    init_waitqueue_head(&sem->opened);
    init_waitqueue_head(&sem->locked);
    init_waitqueue_head(&sem->blocked);
    sem->name = names[u];
    sem->myAddr = sem;
  }

}

//-----------------------------------------------------------------------------
//  Get the total and free counts for the global pool of named semaphores.
//  The *ttl is set to the total, and the free count is the return value.
//-----------------------------------------------------------------------------
uint semCounts(uint *ttl)
{
  uint  u, freeCount = 0;
  sem_t  *sem;


  KLOCK();
  for (u=0,sem=semaphores; u<MAX_NUM_SEMS; ++u,++sem)
    if (!sem->inUse)  ++freeCount;
  KUNLOCK();

  *ttl = MAX_NUM_SEMS;

  return freeCount;
}



//-----------------------------------------------------------------------------
//  Return a pointer to the requested named semaphore.
//
//  If the requested semaphore exists and sem_unlink() has not been called
//  for it, a pointer to the semaphore is returned.  Otherwise NULL is
//  returned.
//
//  If an active semaphore with the specified name does not exist, and the
//  unused argument is not NULL, then *unused is set to the address of an
//  unused semaphore (if one exists).
//
//  Locking must be handled by caller.
//-----------------------------------------------------------------------------
sem_t *getNamedSemaphore(const char *name, sem_t **unused)
{
  uint  u;
  sem_t  *sem;


  if (unused)  *unused = NULL;

//===== scan the list of semaphores looking for a match and an unused one =====
  for (u=0,sem=semaphores; u<MAX_NUM_SEMS; ++u,++sem)  {
    if (INVALID(sem_t, sem))  { errno = EBADSTRUCT;  return NULL; }

    if (!sem->inUse)  { if (unused)  *unused = sem;  continue; }

    if (sem->unlinked)  continue;  // unlinked ones are inaccessible

    if (fstrcmp(name, sem->name))  continue;  // different name?

    return sem;
  }

  return NULL;
}


//==================== The standard POSIX style functions =====================

//-----------------------------------------------------------------------------
//  Create a connection to a named semaphore
//
//  Arguments:
//    name:   Name to use for a new semaphore or the name of an existing
//            semaphore
//    oflag:  Specifies open mode:  O_CREAT alone means create the semaphore
//            if it doesn't already exist.  O_CREAT | O_EXCL means return
//            and error if the semaphore already exists.
//    mode:   Traditional mask of access rights (currently ignored)
//    value:  Initial value for new semaphores.
//
//  Returns a pointer to the requested semaphore, or NULL if an error (with
//  errno set to indicate the error)
//-----------------------------------------------------------------------------
sem_t *sem_open(const char *name, int oflag, mode_t mode, uint value)
{
  sem_t  *sem, *unused = NULL;
  uint  len;
  WaitQueueNode  *node;


  oflag &= (O_CREAT | O_EXCL);

  KLOCK();

  sem = getNamedSemaphore(name, &unused);

  if (sem)  {
  //--- error if caller expected to create a new one and it already exists ---
    if (oflag == (O_CREAT | O_EXCL))  {
      KUNLOCK();  errno = EEXIST;  return SEM_FAILED; }

  //--- If caller already in the semaphore's 'opened' list, we are done ---
    if (taskInQList(&sem->opened, current))  { KUNLOCK();  return sem; }

  //--- try to add the caller to the semaphore's 'opened' list ---
    if (!(node = alloc_WaitQueueNode()))  { KUNLOCK();  return NULL; }

    node->obj = sem;  node->removeFunc = semRemovedFromOpenList;
    __add_wait_queue_tail(&sem->opened, node);

    KUNLOCK();

    #ifdef SEM_DEBUG
      SemMsg(sem, "opened");
    #endif

    return sem;
  }

//===== No match, so create a new one if requested =====
  do  {
    sem = NULL;

  //--- doesn't exist and caller did not request to create it ---
    if (!(oflag & O_CREAT))  { errno = ENOENT;  break; }

  //--- name too long or empty? ---
    len = strlen(name);
    if (len > NAME_MAX)  { errno = ENAMETOOLONG;  break; }
    if (len < 1)  { errno = EINVAL;  break; }

  //--- do we have the resources to create a new semaphore? ---
    if (!unused)  { errno = ENFILE;  break; }
    if (!(node = alloc_WaitQueueNode()))  break;

    sem = unused;

    node->obj = sem;  node->removeFunc = semRemovedFromOpenList;
    __add_wait_queue_tail(&sem->opened, node);

    fstrcpy(sem->name, name);
    sem->value = value;  sem->inUse = true;
  }  while(0);


  KUNLOCK();

  #ifdef SEM_DEBUG
    if (sem)  SemMsg(sem, "created");
  #endif

  return sem;
}


//-----------------------------------------------------------------------------
//  Initialize an unnamed semaphore
//
//  Arguments:
//    sem:      ptr to sem_t variable accessible to all tasks that will use it
//    pshared:  true if other tasks should be able to use the semaphore
//    value:    Initial value for the semaphore
//
//  Returns 0 on success, -1 if an error (with errno set to indicate the error)
//-----------------------------------------------------------------------------
int sem_init(sem_t *sem, int pshared, uint value)
{
#define  ADDR  ((ulong)sem)


//--- shared ones must be in global memory where other tasks can access them ---
  if (pshared)
    if (! ( ((ADDR >= dataStart) AND (ADDR <= dataEnd))
         OR ((ADDR >= bssStart)  AND (ADDR <= bssEnd)) ) )  {
      errno = EFAULT;  return -1;
    }

  if (sem->inUse)
    fprintf(DiagLog, "***** sem_init() on inUse semaphore! *****\n");

#ifdef SEM_DEBUG
  SemMsg(sem, "sem_init");
#endif

  KLOCK();

  BZERO(sem);

  init_waitqueue_head(&sem->opened);
  init_waitqueue_head(&sem->locked);
  init_waitqueue_head(&sem->blocked);

  sem->myAddr = sem;

  sem->value = value;  sem->inUse = true;

  KUNLOCK();

  return 0;
}


//-----------------------------------------------------------------------------
//  Close the connection between a task and a named semaphore.
//
//  If the caller has an open connection to the semaphore, then release any
//  of the locks it may have and close the connection.  If unlink was called
//  for the semaphore, then it cannot be opened again, and when the last
//  open connection to it is closed it will be destroyed.
//
//  Returns 0 on success, -1 if an error (with errno set to indicate the error)
//-----------------------------------------------------------------------------
int sem_close(sem_t *sem)
{
  int  stat = -1;
  WaitQueueNode  *node;
  WaitQueue  *opened;


  KLOCK();

  do  {
    if (!namedSemaphore(sem))  { errno = EINVAL;  break; }

    opened = &sem->opened;

  //--- does the caller have an open connection to the semaphore? ---
    if (! (node = taskInQList(opened, current)))  { errno = ENOLINK;  break; }

  //--- Remove the task from the list of those that have the semaphore open. ---
  //--- Note that the removal of the node from the opened list will cause ---
  //--- semRemovedFromOpenList() to be called, which will free up any locks ---
  //--- the task may have on the semaphore and free up the semaphore if it ---
  //--- was unlinked and has no more open connections ---
    __removeWaitQueue(opened, node);  free_WaitQueueNode(node);

    stat = 0;
  } while(0);

  KUNLOCK();

  #ifdef SEM_DEBUG
    SemMsg(sem, "closed");
  #endif

  return stat;
}


//-----------------------------------------------------------------------------
//  Attempt to get a lock on a semaphore, blocking up to the specified number
//  of "jiffies" (scheduler ticks) before giving up.  A timeout of 0 means
//  return right away if no lock is available, while MAX_SCHEDULE_TIMEOUT means
//  block indefinitely.  In any case, the function may return without a lock
//  if it was interrupted by a signal.
//
//  Returns 0 on success, -1 if an error (with errno set to indicate the error)
//-----------------------------------------------------------------------------
//  NOTE:  sem_wait() is a macro that calls this with MAX_SCHEDULE_TIMEOUT
//-----------------------------------------------------------------------------
int sem_wait_(sem_t *sem, ulong timeoutTicks)
{
  int  stat = -1;
  WaitQueueNode *node;


  do  {
    KLOCK();

  //--- valid semaphore? ---
    if (INVALID(sem_t, sem) OR !sem->inUse)  {
      KUNLOCK();  errno = EINVAL;  break; }

  //--- if a named semaphore, does the task have an open connection? ---
    if (sem->name)  if (! (node = taskInQList(&sem->opened, current)))  {
      KUNLOCK();  errno = ENOLINK;  break; }

  //--- no locks available and blocking not requested? ---
    if (!sem->value AND !timeoutTicks)  {
      KUNLOCK();  errno = EAGAIN;  break; }

  //--- allocate a node to use in the locked or blocked list ---
    if (!(node = alloc_WaitQueueNode()))  { KUNLOCK();  break; }
    node->obj = sem;

  //--- lock available, so take it ---
    if (sem->value)  {
      node->removeFunc = semRemovedFromLockedList;
      __add_wait_queue_tail(&sem->locked, node);  --sem->value;
      KUNLOCK();
      #ifdef SEM_DEBUG
        SemMsg(sem, "got lock");
      #endif
      stat = 0;  break;
    }


  //--- no lock available, so block waiting for it ---
    __add_wait_queue_tail(&sem->blocked, node);
    __set_current_state(TASK_INTERRUPTIBLE);

    KUNLOCK();

    #ifdef SEM_DEBUG
      SemMsg(sem, "blocked");
    #endif

    timeoutTicks = schedule_timeout(timeoutTicks);

    KLOCK();

  //--- We were woken up either because of a signal or because a lock was ---
  //--- given to us when another task's lock was released ---

  //--- If we got the lock, then the node we put in to the blocked list was ---
  //--- moved to the locked list by semRemovedFromLockedList().  Otherwise, ---
  //--- we were woken by a signal and we need to remove the node from the ---
  //--- blocked list, free it, and (possibly) try again. ---

    if (! nodeInQList(&sem->locked, node))  {
      __removeWaitQueue(&sem->blocked, node);  free_WaitQueueNode(node);
      // if timeout elapsed, return an error
      if (!timeoutTicks)  { KUNLOCK();  errno = ETIMEDOUT;  break; }
      // if still some ticks left, try again...
      if (timeoutTicks != MAX_SCHEDULE_TIMEOUT)  { KUNLOCK();  continue; }
      // No timeout was given, and we didn't get the lock, so we must have
      // been woken for some other reason.
      errno = EINTR;  break;
    }

    KUNLOCK();

  //--- we have a lock! ---
    stat = 0;

  #ifdef SEM_DEBUG
    SemMsg(sem, "resumed with lock");
  #endif

  } while(0);

  return stat;
}


//-----------------------------------------------------------------------------
//  Attempt to get a lock on a semaphore, blocking if necessary until a
//  specified time.
//
//  NOTE:  It is generally much more efficient to simply do this:
//
//            sem_wait_(sem, msToTicks(ms));
//
//         which avoids all the math for computing a timeoutTicks value from
//         a timespec value (msToTicks() does some, but MUCH less then here).
//
//         While sem_wait_(sem, timeoutTicks) is not a standard, it would be
//         easy enough to implement on a Unix/Linux system.
//
//
//  Returns 0 on success, -1 if an error (with errno set)
//-----------------------------------------------------------------------------
int sem_timedwait(sem_t *sem, const struct timespec *abs_timeout)
{
  ulong  secs = 0;
  uint  curMS, timeoutMS, ms;
  ulong  timeoutTicks = 0, maxSecs, ticksPerSec;
  struct timeval  tv;


  gettimeofday(&tv, NULL);

//--- compute interval from now to abs_timeout in "jiffies" (timer ticks) ---
  do  {
    if (abs_timeout->tv_sec < tv.tv_sec)  break;

    curMS = (uint) (tv.tv_usec / 1000);
    timeoutMS = (uint) (abs_timeout->tv_nsec / 1000000L);

    if (abs_timeout->tv_sec == tv.tv_sec)  if (timeoutMS <= curMS)  break;

    ticksPerSec = msToTicks(1000);  maxSecs = ULONG_MAX / ticksPerSec - 1;

    secs = abs_timeout->tv_sec - tv.tv_sec;
    if (secs > maxSecs)  secs = maxSecs;

    ms = timeoutMS - curMS;

    timeoutTicks = secs * ticksPerSec + msToTicks(ms);
  }  while (0);


  return sem_wait_(sem, timeoutTicks);
}


//-----------------------------------------------------------------------------
//  Get a lock on a semaphore if it can be obtained witout blocking.
//
//  Returns 0 on success, -1 if not (errno == EAGAIN if no errors, but all
//  locks in use).
//-----------------------------------------------------------------------------
int sem_trywait(sem_t *sem)
{
  return sem_wait_(sem, 0);
}


//-----------------------------------------------------------------------------
//  Release a lock on a semaphore
//-----------------------------------------------------------------------------
int sem_post(sem_t *sem)
{
  int  stat = -1;
  WaitQueueNode  *node;
  WaitQueue  *locked;


  KLOCK();

  do  {
//    if (!namedSemaphore(sem))  { errno = EINVAL;  break; }

    locked = &sem->locked;

  //--- does the task have a lock on the semaphore? ---
    if (! (node = taskInQList(locked, current)))  { errno = ENOLINK;  break; }

  //--- Remove the first node in the locked list for the calling task ---
  //--- Note that the removal of the node in the locked list will cause ---
  //--- semRemovedFromLockedList() to be called, which will transfer the ---
  //--- lock to the first blocked task (if any) and resume it ---
    __removeWaitQueue(locked, node);  free_WaitQueueNode(node);

    stat = 0;

  } while(0);

  KUNLOCK();

  return stat;
}


//-----------------------------------------------------------------------------
//  Destroy (free for future use) an unnamed semaphore that was initialized
//  by an earlier call to sem_init()
//
//  Returns 0 on success, -1 if an error (with errno set to indicate the error)
//-----------------------------------------------------------------------------
int sem_destroy(sem_t *sem)
{
  WaitQueue  *blocked, *locked;


  KLOCK();

  if (!unnamedSemaphore(sem))  { KUNLOCK();  errno = EINVAL;  return -1; }

  blocked = &sem->blocked;  locked = &sem->locked;

  if (waitqueue_active(blocked) OR waitqueue_active(locked))  {
    KUNLOCK();  errno = EBUSY;  return -1; }

  sem->inUse = false;

  KUNLOCK();

  #ifdef SEM_DEBUG
    SemMsg(sem, "destroyed");
  #endif

  return 0;
}


//-----------------------------------------------------------------------------
//  Get the recent state of a semaphore.
//
//  *sval is set to a value indicating the state of the semaphore at some time
//  during the call.
//
//  A value of 0 or < 0 indicates there were no locks available, and the
//  absolute value represents the number of tasks that were waiting for a lock.
//
//  A value > 0 indicates that a lock was available.
//
//  Remember that the result only represents that state the semaphore WAS in
//  at some time during the call.  It is very possible the state will have
//  changed before the caller even sees this result.
//
//  A return value of 0 indicates success and *sval was set.  A return value
//  of -1 indicates there was an error occured and the value of *sval is
//  undefined.
//-----------------------------------------------------------------------------
int sem_getvalue(sem_t *sem, int *sval)
{
  WaitQueue  *blocked;


  KLOCK();

//--- invalid pointer or corrupted semaphore? ---
  if (INVALID(sem_t, sem))  { KUNLOCK();  return -1; }

  if (sem->value)  // locks available?
    *sval = sem->value;
  else  {
    blocked = &sem->blocked;  *sval = qListSize(blocked);  }

  KUNLOCK();  return 0;
}


//-----------------------------------------------------------------------------
//  Unlink (remove) a named semaphore.
//
//  If an active, named semaphore with the specified name exists, and there
//  are no open connections to it, it is released/destroyed (i.e. made
//  available for use by another sem_open() call).
//
//  If it exists, but has open connections, then it is marked as unlinked and
//  it is not accessible via any sem_xxx() functions except sem_post() and
//  sem_close().
//
//  Once all open connections to an unlinked semaphore are closed, the
//  semaphore is released/destroyed.
//-----------------------------------------------------------------------------
int sem_unlink(const char *name)
{
  sem_t *sem;


  KLOCK();

  sem = getNamedSemaphore(name, NULL);

  if (!sem)  { KUNLOCK();  errno = ENOENT;  return -1; }

  sem->unlinked = true;  semFreeIfNotOpen(sem);

  KUNLOCK();

  return -1;
}


//-----------------------------------------------------------------------------

