//----- wait.c ----- 04/20/12 -----(06/17/07)-----

//-----------------------------------------------------------------------------
// DEBROS of standard Linux wait.c source file
//-----------------------------------------------------------------------------
//  Change log (most recent first):
//
//
//  2012-04-20 - Mark Davis
//               - Changed the name of listAll() to wqListTasks().
//               - Added wqCounts() to get total and free counts for the
//                 global pool of waitqueues.
//
//  2010-03-25 - Mark Davis (*)
//               - __removeWaitQueue() now returns a value indicating success
//                 or an error.
//
//  2010-09-09 - Mark Davis (*)
//               - __removeWaitQueue() now checks to make sure the node pointer
//                 is valid.
//
//  2010-08-10 - Mark Davis
//               - Made __removeWaitQueue() a real function to reduce the
//                 amount of inline code generated each place it is used.
//
//  2010-07-28 - Mark Davis (*)
//               - initWaitQueues() and free_WaitQueueNodes now sets the task
//                 pointer for a node to the kernel.  Without this, the
//                 previous change to how the task pointer is handled would
//                 result in functions/macros using a NULL task pointer for
//                 some operations, corrupting memory in the process.
//
//  2010-07-23 - Mark Davis
//               - Added obj and removeFunc to __wait_queue_node structure.
//                 These are used to support the ability to do special-case
//                 processing when a node is removed from one of the lists
//                 an object keeps.
//               - With the exception of the alloc_WaitQueueNode() and
//                 free_WaitQueueNode() the functions, the functions and macros
//                 that manipulate WaitQueueNodes now use the task pointer
//                 stored in the node rather than just assuming it is or
//                 setting it to 'current'.  This allows other tasks to call
//                 functions that need to manipulate nodes that belong to
//                 other tasks (such as the semaphore module which needs to
//                 remove a node from one queue and add it to another without
//                 changing what task the node points to.
//
//  2010-06-30 - Mark Davis
//               - alloc_WaitQueueNode() now sets errno if it cannot allocate
//                 a node.
//
//  2010-05-12 - Mark Davis
//               - replaced clearQueue() with wakeTasks() and removeFromQueues()
//
//  2010-04-29 - Mark Davis
//               - Significant overhaul to simplify and support the ability
//                 to have one task block on multiple wait queues (which is
//                 needed to support funcs like poll() and select(), which
//                 are needed for other projects).
//
//  2007-06-17 - Mark Davis
//               - original version (?)
//
//-----------------------------------------------------------------------------

#include <common.h>

#include <linux/dummy.h>

#include <sys/types.h>
#include <string.h>

#include <errno.h>


//#include <linux/init.h>
//#include <linux/module.h>
#include <linux/sched.h>
//#include <linux/mm.h>
#include <linux/wait.h>
//#include <linux/hash.h>

#include <unistd.h>

#include <limits.h>

#include <linux/sched.h>

/*-----------------------------------------------------------------------------

  Wait queues are the means by which a task can wait for an event without
  having to waste CPU cycles while doing so.

  When waiting for an event, a task must use a loop in which it tests for the
  occurance of the event and, if it has not happened, yet, sleep for a bit.
  Aside for other considerations like timeouts, the task will generally
  do this indefinitely until it finds that the event it has been testing
  for occurs.

  But there are two basic problems with this method:

    - The task uses CPU time each time it tests for the event.  Each time
      the test fails means we wasted some CPU time.

    - The time it takes for the task to respond to the event after it happens
      can be as long as the length of time each sleeps each pass through the
      loop.  Even the average time will be half of that.

  Wait queues address both of these issues.

  A wait queue provides a way for a task to tell a device driver that it is
  waiting for an event that the driver understands.  To block on an event
  without using of CPU while waiting for the event, the task adds itself to
  one of the driver's wait queues and then suspends itself.  When the driver
  sees an event that corresponds to one of its wait queues, it removes one or
  more tasks from the queue and wakes them up.

  The resumed task then tests for the occurance of the event again.  If the
  event doesn't satisfy its need, then it puts itself back in to the queue
  and suspends itself again.

  Assuming that the driver does a reaonable job and only wakes up a task
  when there is a good chance of it satisfying its needs, then the result
  is that the task uses no CPU time while waiting for the event, and the
  average time between when the event occurs and when the task wakes up
  and processes it is much shorter.

  ---------------------------------------------------------------------------

  The following provides a "pool" of nodes to use in the wait queues.

  NOTE that the DEBROS implementation is a bit different than the standard
  Linux one.  In Linux, each node is part of just one list whose head is
  owned by the device driver.  Each node contains a pointer that is used
  to identify which task it represents, so the driver can quickly determine
  which task to wake up when it removes a node from one of its wait queues.

  The problem with this is that it does not provide a way to determine
  which wait queues a task has added itself to when you need to remove it
  from all of them at once.

  For this reason, the earlier DEBROS implementation limited each task to
  a single wait queue node that was part of the task table structure.  This
  meant each task could only block on one event at a time, but it also meant
  that the possibility of orphaned wait queue nodes was not an issue.

  To get around this limitation without adding too much overhead, the newer
  approach puts each wait queue node in two lists:

    taskList:  The list of tasks that are in the same wait queue
    queueList:  The list of event queues a task is waiting in

  The 1st set of links (taskList) are used like the standard task_list
  links in Linux, which allows a driver to quickly resume a task and
  remove it from the queue.  The taskList links are the ONLY ones that
  a driver uses.

  The 2nd set of links (queueList) was added to link together all the nodes
  for a given task.  The task itself (or, if we are killing a task, the
  kernel task) is responsible for using the queueList links to remove each
  node belonging to a task from any wait queues they may still be in and
  from the queueList set.

  Without the 2nd set of links, there is no efficient way to locate all
  the nodes that belong to a given task.  This is important becasue any
  time we wake up or kill a task we need to insure that it is removed from
  all wait queues.  If we didn't, it could be woken up for some event it is
  no longer interested in, or a driver could try to wake up a task that
  no longer exists.

  The primary motivation for the change in approach was to make it possible
  to support functions like poll() and select() without having to do a
  sequential search of all the wait queues of all the drivers each time
  a task is resumed after adding itself to a one or more wait queues.

  Haven't discovered yet how Linux deals with this issue, much less if it
  is as efficient as this approach.


NOTES:

  - Wait queue nodes will often have to be accessed by tasks other than the
    one that they point to (e.g. when the kernel does some background
    processing for the driver, or when a task is terminated).

  - The only time a task's stack is mapped in to the logical 16-bit address
    space is when the task is running.  Access by any other task would
    require computing a far pointer using the task's STACKSEG value and
    the logical 16-bit address and then using it to read or write the value,
    all of which would be quite time consuming.

  - Therefore, the memory used for all wait queue nodes and heads must have
    a fixed memory location easily usable by other tasks.  In addition, the
    functions that use wait queues are called fairly frequently so they
    must be efficient.  For this reason, the memory for the nodes should
    reside in root memory.


-----------------------------------------------------------------------------*/

// A pre-allocated set of wait queue nodes with fixed addresses in root memory.
// The initWaitQueues() function adds them all to a linked list so locating
// and removing an unused one and putting one back in to the pool can all
// be done efficiently (i.e. without having to search a list).

WaitQueueNode  wqNodes[WQ_POOL_SIZE];
WaitQueue  wqNodePool;


//-----------------------------------------------------------------------------
//  Initialize the list of unused wait queue nodes.  This creates a list of
//  of unused WaitQueueNodes belonging to the kernel.  Keeping them in a
//  linked-list means we don't have to scan a linear list looking for one
//  that is not in use - we just take the next on in the list (or if the list
//  is empty, we know they are all used up).
//-----------------------------------------------------------------------------
void  initWaitQueues(void)
{
  WaitQueueNode *node;

//--- Create a linked list of all the unused wait queue nodes ---
  init_waitqueue_head(&wqNodePool);  BZERO(wqNodes);

  for (node=wqNodes; node<wqNodes+WQ_POOL_SIZE; ++node)  {
    INIT_LIST_HEAD(&node->queueList);
    node->task = kernelTask;  node->myAddr = node;
    __add_wait_queue(&wqNodePool, node);
  }

}

//-----------------------------------------------------------------------------
//  Get an initialized WaitQueueNode structure
//
//  Returns NULL if all of them are in use.
//-----------------------------------------------------------------------------
WaitQueueNode *alloc_WaitQueueNode(void)
{
  WaitQueueNode *node = NULL;
  list_head_t  *taskListEntry;


  KLOCK();

  if (list_empty(&wqNodePool.listHead))  {
    KUNLOCK();
    fprintf(DiagLog, "***** out of wait queue nodes *****\r\n");
    errno = ENONODES;  return NULL;
  }

//--- remove the first entry in the list of unused nodes ---
  taskListEntry = wqNodePool.listHead.next;  list_del(taskListEntry);

  KUNLOCK();

//--- When we removed the node from the list, we got the address of one ---
//--- of the list_head structures that is embedded in the WaitQueueNode ---
//--- structure.  Now we get the addr of the WaitQueueNode it is a part of ---
  node = list_entry(taskListEntry, WaitQueueNode, taskList);

//--- initialize the new node ---
  node->flags = 0;  node->task = current;
  node->obj = NULL;  node->removeFunc = NULL;
//  node->func = autoremove_wake_function;
  INIT_LIST_HEAD(taskListEntry);
  INIT_LIST_HEAD(&node->queueList);

  if (ShowWQNodes)
    fprintf(DiagLog, "--- alloc_WaitQueueNode: %04X\r\n", (uint)node);

  return node;
}

//-----------------------------------------------------------------------------
//  Free a node by adding it back in to the list of unused nodes.  The node
//  must have ben removed from any lists it was in by __removeWaitQueue()
//  or removeWaitQueue().
//-----------------------------------------------------------------------------
void free_WaitQueueNode(WaitQueueNode *node)
{
  KLOCK();

  INIT_LIST_HEAD(&node->queueList);

  node->task = kernelTask;

  list_add(&node->taskList, &wqNodePool.listHead);

  KUNLOCK();

  if (ShowWQNodes)
    fprintf(DiagLog, "--- free_WaitQueueNode: %04X\r\n", (uint)node);
}


//-----------------------------------------------------------------------------
//  for diagnostic use only - list all tasks in the specified wait queue
//-----------------------------------------------------------------------------
void wqListTasks(WaitQueue *wq)
{
  task_t  *task;
  list_head_t  *cur, *next, *list;
  int  pid;


  list = &wq->listHead;    // The addr doesn't chg, so lock not yet needed

  fprintf(DiagLog, "\r\nwqListTasks[%04x]\n", (uint)wq);  //tdebug

  KLOCK();

//--- list all the tasks in a wait_queue's task_list ---
  list_for_each_safe(cur, next, list)  {
    task = list_entry(cur, WaitQueueNode, taskList)->task;
    pid = (task ? task->pid : -1);
    fprintf(DiagLog, "  [%04lx][%04lx][%d]\n", (ulong)cur, (ulong)task, pid);
    if (INVALID(task_t, task))  break;
  }

  KUNLOCK();
}

//-----------------------------------------------------------------------------
//  Get the total and free counts for the global pool of waitQueues.
//  The *ttl is set to the total, and the free count is the return value.
//-----------------------------------------------------------------------------
uint wqCounts(uint *ttl)
{
  uint  freeCount = 0;
  list_head_t  *cur;


//--- count the # of nodes left in the pool ---
  KLOCK();
  list_for_each(cur, &wqNodePool.listHead)  ++freeCount;
  KUNLOCK();

  *ttl = WQ_POOL_SIZE;

  return freeCount;
}



//-----------------------------------------------------------------------------
//  Wake all the tasks in a wait queue.
//
//  This function should only be called by a device driver to wake all the
//  tasks in one of its wait queues.  The wait queue node for each task is
//  removed from the queue but is NOT freed.  This allows the task to reuse
//  the node without the overhead of having to allocate a new node.
//
//  WARNING:  wq MUST point to a wait queue for a device/driver as the list
//            will be traversed via the taskList links in the queue's nodes.
//-----------------------------------------------------------------------------
void wakeTasks(WaitQueue *wq)
{
  task_t  *task;
  list_head_t  *cur, *next, *list;
  WaitQueueNode  *node;


  list = &wq->listHead;  // The addr doesn't chg, so lock not yet needed

//  fprintf(DiagLog, "wakeTasks[%04lx]\n", (ulong)wq);  //tdebug

  KLOCK();  // but the contents can change, so here we need a lock

//--- remove each node from the lists it is in but don't free it ---
  list_for_each_safe(cur, next, list)  {
    node = list_entry(cur, WaitQueueNode, taskList);

    __removeWaitQueue(wq, node);  // does not "free" the node

    task = NodeTask(node);  if (INVALID(task_t, task))  break;

  //--- resume the task ---
    if (_activeTaskIdx_(task->pid) >= 0)  {
      if (task->diagFlags & DebugBits)
        fprintf(DiagLog, "unblock task: [%.26s][%d] state: %d\n",
                task->argv[0], task->pid, task->state);
      task->state = TASK_RUNNING;
    }
  }

  KUNLOCK();
}


//-----------------------------------------------------------------------------
//  Remove the task from any and all wait queues it may be in.  If freeNodes
//  is non-0, the nodes will be freed as they are removed.
//
//  For each node that is removed, its removeFunc (if not NULL) is called after
//  the node has been removed from the list.
//
//  NOTE:  The queue is traversed in reverse order so processes that add nodes
//         to the tail of a list (such as the semaphore module) can be sure
//         that the nodes are removed (and their removeFunc's called) in
//         the opposite order that they were created/added.
//-----------------------------------------------------------------------------
void removeFromQueues(task_t *task, bool freeNodes)
{
  list_head_t  *cur, *prev, *list;
  WaitQueueNode  *node;
  WaitQueue  *wq;


  wq = &task->waitQs;  list = &wq->listHead;

  KLOCK();

//--- remove each node in the list (which is either a list of tasks blocked ---
//--- on a queue or a list of queues a task is blocked on ---
  list_for_each_prev_safe(cur, prev, list)  {
    node = list_entry(cur, WaitQueueNode, queueList);
    __removeWaitQueue(wq, node);  // does not "free" the node
    if (freeNodes)  free_WaitQueueNode(node);
  }

  KUNLOCK();
}


//-----------------------------------------------------------------------------
//  Initializes a WaitQueue head to an empty list
//-----------------------------------------------------------------------------
void  init_waitqueue_head(WaitQueue *wq)
{
  spin_lock_init(&wq->lock);
  INIT_LIST_HEAD(&wq->listHead);
}


//-----------------------------------------------------------------------------
//  Add a node to the front of a WaitQueue and clear the EXCLUSIVE bit in the
//  node's flags.
//
//  This function should be used when the appropriate locking is not already
//  asserted.
//-----------------------------------------------------------------------------
void  add_wait_queue(WaitQueue *wq, WaitQueueNode *node)
{
//  unsigned long flags;

  if (!wq OR !node)  return;

  node->flags &= ~WQ_FLAG_EXCLUSIVE;

  KLOCK();  // spin_lock_irqsave(&wq->lock, flags);

  __add_wait_queue(wq, node);

  KUNLOCK();  // spin_unlock_irqrestore(&wq->lock, flags);
}


//-----------------------------------------------------------------------------
//  Add a node to the front of a WaitQueue.
//
//  This function should ONLY be used when the appropriate locking is already
//  asserted!
//-----------------------------------------------------------------------------
void  __add_wait_queue(WaitQueue *wq, WaitQueueNode *node)
{
  list_add(&node->taskList, &wq->listHead);

  if (NodeTask(node))
    list_add(&node->queueList, &NodeTask(node)->waitQs.listHead);
}


//-----------------------------------------------------------------------------
//  Add a node to the end of a WaitQueue.
//
//  This function should ONLY be used when the appropriate locking is already
//  asserted!
//-----------------------------------------------------------------------------
void  __add_wait_queue_tail(WaitQueue *wq, WaitQueueNode *node)
{
  list_add_tail(&node->taskList, &wq->listHead);

  if (NodeTask(node))
    list_add_tail(&node->queueList, &NodeTask(node)->waitQs.listHead);
}


//-----------------------------------------------------------------------------
//  Add a node to the end of a WaitQueue and set the EXCLUSIVE bit in the
//  node's flags.
//
//  This function should be used when the appropriate locking is not already
//  asserted.
//-----------------------------------------------------------------------------
void  add_wait_queue_exclusive(WaitQueue *wq, WaitQueueNode *node)
{
//  unsigned long flags;

  if (!wq OR !node)  return;

  node->flags |= WQ_FLAG_EXCLUSIVE;

  KLOCK();  // spin_lock_irqsave(&wq->lock, flags);

  __add_wait_queue_tail(wq, node);

  KUNLOCK();  // spin_unlock_irqrestore(&wq->lock, flags);
}


//-----------------------------------------------------------------------------
//  Remove a WaitQueueNode from any lists it is in but do not free it.
//
//  This function should be used in cases where the appropriate locking is
//  not already asserted.
//-----------------------------------------------------------------------------
void removeWaitQueue(WaitQueue *wq, WaitQueueNode *node)
{
//  unsigned long flags;

  if (!wq OR !node)  return;

  KLOCK();  //  spin_lock_irqsave(&wq->lock, flags);

  __removeWaitQueue(wq, node);

  KUNLOCK();  // spin_unlock_irqrestore(&wq->lock, flags);
}


//-----------------------------------------------------------------------------
//  Remove a WaitQueueNode from any lists it is in but do not free it.
//
//  This function must ONLY be called when the appropriate locking is already
//  asserted.
//-----------------------------------------------------------------------------
int __removeWaitQueue(WaitQueue *wq, WaitQueueNode *node)
{
  if (INVALID(WaitQueueNode, node))  return -1;

  list_del_init(&node->taskList);
  list_del_init(&node->queueList);

  if (node->removeFunc)  (*node->removeFunc)(node);

  return 0;
}


//-----------------------------------------------------------------------------
//  Add a WaitQueueNode to two lists:
//    - The specified WaitQueue (a list of tasks waiting for the event
//      represented by the queue)
//    - The list of wait queues the current task is in
//-----------------------------------------------------------------------------
void prepare_to_wait(WaitQueue *wq, WaitQueueNode *node, int state)
{
  //  unsigned long flags;

  if (!wq OR !node)  return;

  node->flags &= ~WQ_FLAG_EXCLUSIVE;

  KLOCK();  //  spin_lock_irqsave(&q->lock, flags);

  if (!list_empty(&node->taskList) OR !list_empty(&node->queueList))
    __removeWaitQueue(wq, node);

  __add_wait_queue(wq, node);
  __set_current_state(state);

  KUNLOCK();  // spin_unlock_irqrestore(&wq->lock, flags);
}



//-----------------------------------------------------------------------------
//  Same as prepare_to_wait() except that the EXCLUSIVE bit is set in the
//  flags value of the WaitQueueNode, indicating the driver should wake up
//  only ONE task in the queue when the event occurs.
//-----------------------------------------------------------------------------
void prepare_to_wait_exclusive(WaitQueue *wq, WaitQueueNode *node, int state)
{
//  unsigned long flags;

  if (!wq OR !node)  return;

  node->flags |= WQ_FLAG_EXCLUSIVE;

  KLOCK();  //  spin_lock_irqsave(&wq->lock, flags);

  if (list_empty(&node->taskList) AND list_empty(&node->queueList))  {
    __add_wait_queue_tail(wq, node);
    __set_current_state(state);
  }

  KUNLOCK();  // spin_unlock_irqrestore(&wq->lock, flags);
}


//-----------------------------------------------------------------------------
//  Set the current task's state to RUNNING, remove the specified WaitQueueNode
//  from any lists it may be in, and free the node
//-----------------------------------------------------------------------------
void finish_wait(WaitQueue *wq, WaitQueueNode *node)
{
  //  unsigned long flags;

  __set_current_state(TASK_RUNNING);

  KLOCK();  //  spin_lock_irqsave(&wq->lock, flags);

  __removeWaitQueue(wq, node);  free_WaitQueueNode(node);

  KUNLOCK();  // spin_unlock_irqrestore(&wq->lock, flags);
}



//-----------------------------------------------------------------------------
//  Support for use of a callback function when removing a WaitQueueNode from
//  the lists it is in (not supported at this time)
//-----------------------------------------------------------------------------
/*
int _nearcall autoremove_wake_function(
                      WaitQueueNode *node, unsigned mode, int sync, void *key)
{
  int ret = default_wake_function(node, mode, sync, key);

  if (ret)  { KLOCK();  free_WaitQueueNode(node);  KUNLOCK();  }

  return ret;
}
*/


//-----------------------------------------------------------------------------

