//--- common.c --- 11/08/13 ---(05/06/05)---

//-----------------------------------------------------------------------------
//
//  Copyright (c) 2003-2013 National Superconducting Cyclotron Laboratory,
//            Michigan State University, and Mark Davis
//
//=============================================================================
//  Misc functions common to almost all NSCL ZWorld code
//=============================================================================
//  Change log (most recent first):
//
//
//  2013-11-08 - Mark Davis (*)
//               - For non-Rabbit platforms only:
//                 - read_() now returns with errno == EAGAIN if the timeout
//                   value given was 0.  Prior to this it would return with
//                   errno == ETIME if it failed to read the requested # of
//                   bytes, even if the timeout value was 0 (indicating it
//                   should be a non-blocking read).
//                 - read_() now returns a value of -1 if it has read no bytes
//                   at all and EOF condition was not reached
//
//  2012-12-12 - Mark Davis
//               - Fixed compileTime() so it works with date strings that do
//                 not include the day of the week.
//
//  2012-12-05 - Mark Davis
//               - Removed Force/Release/Dig/In/Out() functions
//               - Moved the following from fcntl to common:
//                    vt100 escape sequence definitions
//                    DispStr() and DispStrAttr()
//                    read_() and write_()
//                    redirectStdio() and logStdout()
//
//  2012-02-14 - Mark Davis
//               - Modified fwrite_() to call writeFDEV() rather than write_()
//                 to reduce its overhead.
//
//  2011-03-29 - Mark Davis
//               - showNetState() now takes the interface ID as an argument
//
//  2011-03-25 - Mark Davis (*)
//               - ValidatePtr() now checks for pointer values of 0 and -1
//                 (0xFFFF)
//
//  2010-12-14 - Mark Davis
//               - Changed _TimerSeconds + secTimerOffset to TIME()
//
//  2010-07-13 - Mark Davis
//               - Commented out the fmult() function and replaced it with a
//                 macro that just does a regular multiply.
//                 fmult() is no longer necessary because I created a
//                 modified version of fmult() that can be substituted for the
//                 Softools internal __fmlt() function so ALL floating point
//                 multiples will use the faster (but less accurate) version.
//
//  2010-06-15 - Mark Davis (*)
//               - renamed fWriteBytes() to fwrite_()
//               - read_() and write_() now check for a NULL buffer pointer
//               - added skipBytes() to use instead of calling new read_()
//                 with a NULL buffer pointer.
//
//  2010-06-15 - Mark Davis
//               - renamed fWriteBytes() to fwrite_()
//
//  2010-05-14 - Mark Davis (*)
//               - Fixed fWriteBytes() so it no longer gets stuck in a loop
//                 if write_() returns < 0 each time.
//
//  2010-05-13 - Mark Davis
//               - Removed readBytes() and writeBytes() (they are replaced by
//                 read_() and write_() in the fcntl module).
//
//  2010-03-29 - Mark Davis
//               - fWriteBytes() now uses the fd passed to it (it was always
//                 writing to stdout!)
//               - fWriteBytes() now uses a buffer passed by the caller (to
//                 avoid possibly overflowing the stack of the caller)
//
//  2010-03-11 - Mark Davis
//               - CheckETX() now tosses away any additional input characters
//                 available at the time it is called after the first one,
//                 unless one of them is a Ctrl-C.
//
//  2010-03-08 - Mark Davis
//               - Added fGetShort(), fGetLong(), fAddShort(), fAddLong()
//
//  2010-03-03 - Mark Davis
//               - Added sendBCastPkt() (moved it here from configd.c)
//
//  2010-02-15 - Mark Davis
//               - Moved readBufCh() from the "busybox" shared code in to this
//                 module
//
//  2010-01-29 - Mark Davis (*)
//               - moved resetWDT() in readBytes() and writeBytes() to avoid
//                 problems if they are called by the applic task
//               - setBlockTimeout() is now called to reset the timeout to 0
//                 if an unexpected error occurs during a read() call in
//                 readBytes() or a write() call in writeBytes()
//               - Changed writeBytes() so, like readBytes(), it uses the
//                 timeout mechanism supported by write() for each call.  Prior
//                 to this change it was basically polling (calling write()
//                 with blocking disabled, sleeping if the timeout had not
//                 been reached without progress, repeat...)
//
//  2009-06-23- Mark Davis -
//               - Changed WDT_2S() calls to resetWDT()
//
//  2009-05-07- Mark Davis
//               - fixes to readBytes()
//
//  2007-10-30- Mark Davis
//               - readBytes() now uses blocking reads with a timeout handled
//                 by the kernel (if a timeout specified)
//
//  2007-10-02- Mark Davis
//               - moved rounding from fmtFloat() (where it could alter the
//                 result so it doesn't fit in specified space) to the
//                 Engineering() func (which calls fmtFloat())
//
//  2007-03-23- Mark Davis
//               - added strstri() - a case insensitive version of strstr()
//
//  2007-03-08 - Mark Davis
//               - added writeBytes()
//
//  2006-03-07 - Mark Davis
//               - started port of selected portions to Softools compiler
//
//  2005-07-15 - Mark Davis
//               - Added ForceDigOut() and ReleaseDigOut()
//
//  2005-06-01 - Mark Davis
//               - Changed Engineering() so it works with larger exponents
//
//  2005-05-06 - Mark Davis
//               - Extracted various pieces from older code
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <ctype.h>

#include <common.h>
#include <unistd.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <math.h>

#include <sys/stat.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <sys/backup.h>

#include <sys/hdwDefs.h>

#ifdef RABBIT
  #include <dcdefs.h>
#endif
#include <limits.h>



typedef  struct in_addr      INET_ADDR;
typedef  struct sockaddr    *SAP;
typedef  struct sockaddr_in  SAI;

typedef  struct ether_addr  EtherAddr;

#ifdef RABBIT
  #define  SIN_ADDR  sin_addr
#else
  #define  SIN_ADDR  sin_addr.s_addr
#endif

#define  MaxPow10  24
#define  Pow0Idx   MaxPow10

extern int  timeZone;


//--- common vt100 commands ---
const char clsCmd[] = {"\033[2J\033[00;00H"};  // CLear Screen
const char cllCmd[] = {"\033[K"};              // CLear to end-of-Line
const char clpCmd[] = {"\033[J"};              // CLear to end-of-Page



float far  powOf10[MaxPow10*2+1];
#define  PowOf10(pow)  (powOf10[Pow0Idx+(pow)])


const unsigned short far crc16_rev_table[256] =
{   0x0000, 0xC0C1, 0xC181, 0x0140, 0xC301, 0x03C0, 0x0280, 0xC241,
    0xC601, 0x06C0, 0x0780, 0xC741, 0x0500, 0xC5C1, 0xC481, 0x0440,
    0xCC01, 0x0CC0, 0x0D80, 0xCD41, 0x0F00, 0xCFC1, 0xCE81, 0x0E40,
    0x0A00, 0xCAC1, 0xCB81, 0x0B40, 0xC901, 0x09C0, 0x0880, 0xC841,
    0xD801, 0x18C0, 0x1980, 0xD941, 0x1B00, 0xDBC1, 0xDA81, 0x1A40,
    0x1E00, 0xDEC1, 0xDF81, 0x1F40, 0xDD01, 0x1DC0, 0x1C80, 0xDC41,
    0x1400, 0xD4C1, 0xD581, 0x1540, 0xD701, 0x17C0, 0x1680, 0xD641,
    0xD201, 0x12C0, 0x1380, 0xD341, 0x1100, 0xD1C1, 0xD081, 0x1040,
    0xF001, 0x30C0, 0x3180, 0xF141, 0x3300, 0xF3C1, 0xF281, 0x3240,
    0x3600, 0xF6C1, 0xF781, 0x3740, 0xF501, 0x35C0, 0x3480, 0xF441,
    0x3C00, 0xFCC1, 0xFD81, 0x3D40, 0xFF01, 0x3FC0, 0x3E80, 0xFE41,
    0xFA01, 0x3AC0, 0x3B80, 0xFB41, 0x3900, 0xF9C1, 0xF881, 0x3840,
    0x2800, 0xE8C1, 0xE981, 0x2940, 0xEB01, 0x2BC0, 0x2A80, 0xEA41,
    0xEE01, 0x2EC0, 0x2F80, 0xEF41, 0x2D00, 0xEDC1, 0xEC81, 0x2C40,
    0xE401, 0x24C0, 0x2580, 0xE541, 0x2700, 0xE7C1, 0xE681, 0x2640,
    0x2200, 0xE2C1, 0xE381, 0x2340, 0xE101, 0x21C0, 0x2080, 0xE041,
    0xA001, 0x60C0, 0x6180, 0xA141, 0x6300, 0xA3C1, 0xA281, 0x6240,
    0x6600, 0xA6C1, 0xA781, 0x6740, 0xA501, 0x65C0, 0x6480, 0xA441,
    0x6C00, 0xACC1, 0xAD81, 0x6D40, 0xAF01, 0x6FC0, 0x6E80, 0xAE41,
    0xAA01, 0x6AC0, 0x6B80, 0xAB41, 0x6900, 0xA9C1, 0xA881, 0x6840,
    0x7800, 0xB8C1, 0xB981, 0x7940, 0xBB01, 0x7BC0, 0x7A80, 0xBA41,
    0xBE01, 0x7EC0, 0x7F80, 0xBF41, 0x7D00, 0xBDC1, 0xBC81, 0x7C40,
    0xB401, 0x74C0, 0x7580, 0xB541, 0x7700, 0xB7C1, 0xB681, 0x7640,
    0x7200, 0xB2C1, 0xB381, 0x7340, 0xB101, 0x71C0, 0x7080, 0xB041,
    0x5000, 0x90C1, 0x9181, 0x5140, 0x9301, 0x53C0, 0x5280, 0x9241,
    0x9601, 0x56C0, 0x5780, 0x9741, 0x5500, 0x95C1, 0x9481, 0x5440,
    0x9C01, 0x5CC0, 0x5D80, 0x9D41, 0x5F00, 0x9FC1, 0x9E81, 0x5E40,
    0x5A00, 0x9AC1, 0x9B81, 0x5B40, 0x9901, 0x59C0, 0x5880, 0x9841,
    0x8801, 0x48C0, 0x4980, 0x8941, 0x4B00, 0x8BC1, 0x8A81, 0x4A40,
    0x4E00, 0x8EC1, 0x8F81, 0x4F40, 0x8D01, 0x4DC0, 0x4C80, 0x8C41,
    0x4400, 0x84C1, 0x8581, 0x4540, 0x8701, 0x47C0, 0x4680, 0x8641,
    0x8201, 0x42C0, 0x4380, 0x8341, 0x4100, 0x81C1, 0x8081, 0x4040
};


const char *months[12] = { "Jan", "Feb", "Mar",
                           "Apr", "May", "Jun",
                           "Jul", "Aug", "Sep",
                           "Oct", "Nov", "Dec" };

#ifdef RABBIT
  #include <sys/farprintf.h>
#endif


//-----------------------------------------------------------------------------
//  Redirect buffered IO (printf, perror(), etc) so it does not use fds 0,1,2
//  If successful, all buffered output goes to a log file instead.
//-----------------------------------------------------------------------------
int redirectStdio(void)
{
  #ifdef RABBIT
    current->logStdout = current->redirStdio = true;  return 0;
  #else
    return -1;
//finish:  If/When "real" FILE support is implemented, this should call
//         fclose(stdout), and do freopen() for stdout, stderr
//     OR: Better yet?  If syslog had its own fd, then I could use the
//         existing dup2()...
  #endif
}

//-----------------------------------------------------------------------------
//  Cause everything written to stdout to be written to the log file as well
//-----------------------------------------------------------------------------
int logStdout(void)
{
  #ifdef RABBIT
    current->logStdout = true;  return 0;
  #else
    //finish
    return -1;
  #endif
}


//-----------------------------------------------------------------------------
//  given an array of elements of specified size and # elements and the offset
//  within each element to a byte that acts as an "in Use" flag, return a ptr
//  to the 1st element that was not in use (or NULL, if all the elements are
//  in use) (returned element is filled with 0's and marked as in use)
//-----------------------------------------------------------------------------
/*
char *unused(void *arrayOf, int elemSize, int numElements, int inUseOffset)
{
  int  i;
  char  *ptr;


  for (i=0; i<numElements; ++i)  {
    ptr = (char *)arrayOf + (uint)(elemSize * i);
    if (! ptr[inUseOffset])  break;
  }
  if (i == numElements)  return NULL;

  memset(ptr, 0, elemSize);  ptr[inUseOffset] = true;

  return ptr;
}
*/




/*-----------------------------------------------------------------------------
 *  Functions for adding to and extracting data from a byte array in network
 *  (most significant byte 1st) order.
 *  All functions update the value of the pointer into the array.
 *---------------------------------------------------------------------------*/
void addShort(uchar **ptr, short val)
{
/*
  uchar bufPtr = *ptr + 1;  *ptr += 2;

  *bufPtr = (uchar)(val & 0xff);  --bufPtr;  val >>= 8;
  *bufPtr = (uchar)(val & 0xff);
*/
//  *((short *)(*ptr)) = htons(val);  *ptr += 2;

  ushort  netShort = htons(val);
  uchar  *nsp = (uchar *)&netShort;

  **ptr = *nsp++;  ++(*ptr);
  **ptr = *nsp++;  ++(*ptr);
}

/*---------------------------------------------------------------------------*/
void addLong(uchar **ptr, long val)
{
/*
  uchar bufPtr = *ptr + 3;  *ptr += 4;

  *bufPtr = (uchar)(val & 0xff);  --bufPtr;  val >>= 8;
  *bufPtr = (uchar)(val & 0xff);  --bufPtr;  val >>= 8;
  *bufPtr = (uchar)(val & 0xff);  --bufPtr;  val >>= 8;
  *bufPtr = (uchar)(val & 0xff);
*/
//  *((long *)(*ptr)) = htonl(val);  *ptr += 4;

  ulong  netLong = htonl(val);
  uchar  *nlp = (uchar *)&netLong;

  **ptr = *nlp++;  ++(*ptr);
  **ptr = *nlp++;  ++(*ptr);
  **ptr = *nlp++;  ++(*ptr);
  **ptr = *nlp++;  ++(*ptr);
}

//-----------------------------------------------------------------------------
void fAddShort(uchar far **ptr, short val)
{
  ushort  netShort = htons(val);
  uchar  *nsp = (uchar *)&netShort;

  **ptr = *nsp++;  ++(*ptr);
  **ptr = *nsp++;  ++(*ptr);
}

/*---------------------------------------------------------------------------*/
void fAddLong(uchar far **ptr, long val)
{
  ulong  netLong = htonl(val);
  uchar  *nlp = (uchar *)&netLong;

  **ptr = *nlp++;  ++(*ptr);
  **ptr = *nlp++;  ++(*ptr);
  **ptr = *nlp++;  ++(*ptr);
  **ptr = *nlp++;  ++(*ptr);
}

/*---------------------------------------------------------------------------*
 *  Return the 2-byte integer stored in network byte order (MSB first) at    *
 *  the specified address, and increment the a pointer by 2 bytes.           *
 *---------------------------------------------------------------------------*/
short getShort(uchar **ptr)
{
  short netShort = *((short *)(*ptr));  *ptr += 2;
/*
  ushort  netShort;
  uchar  *nsp = (uchar *)&netShort;

  *nsp++ = **ptr;  ++(*ptr);
  *nsp++ = **ptr;  ++(*ptr);
*/
  return ntohs(netShort);
}

/*---------------------------------------------------------------------------*
 *  Return the 4-byte integer stored in network byte order (MSB first) at    *
 *  the specified address, and increment the a pointer by 4 bytes.           *
 *---------------------------------------------------------------------------*/
long getLong(uchar **ptr)
{
  long netLong = *((long *)(*ptr));  *ptr += 4;
/*
  ulong  netLong;
  uchar  *nlp = (uchar *)&netLong;

  *nlp++ = **ptr;  ++(*ptr);
  *nlp++ = **ptr;  ++(*ptr);
  *nlp++ = **ptr;  ++(*ptr);
  *nlp++ = **ptr;  ++(*ptr);
*/
  return ntohl(netLong);
}

/*---------------------------------------------------------------------------*
 *  Return the 2-byte integer stored in network byte order (MSB first) at    *
 *  the specified address, and increment the a pointer by 2 bytes.           *
 *---------------------------------------------------------------------------*/
short fGetShort(uchar far **ptr)
{
  short netShort = *((short far *)(*ptr));  *ptr += 2;

  return ntohs(netShort);
}

/*---------------------------------------------------------------------------*
 *  Return the 4-byte integer stored in network byte order (MSB first) at    *
 *  the specified address, and increment the a pointer by 4 bytes.           *
 *---------------------------------------------------------------------------*/
long fGetLong(uchar far **ptr)
{
  long netLong = *((long far *)(*ptr));  *ptr += 4;

  return ntohl(netLong);
}


//-----------------------------------------------------------------------------
short GetShort_(uchar *buf, int idx)
{
  return ((ushort)(buf[idx]) * 256) + (ushort)(buf[idx+1]);
}

/*----------------------------------------------------------------------------*
 *  Functions to add "tagged" parameter values to a buffer, where each value
 *  is prefaced with a 2-byte ID and a 2-byte length.
 *----------------------------------------------------------------------------*/
void addShortParam(uchar **ptr, short id, short val)
{
  addShort(ptr, id);  addShort(ptr, 2);  addShort(ptr, val);
}

/*---------------------------------------------------------------------------*/
void addLongParam(uchar **ptr, short id, long val)
{
  addShort(ptr, id);  addShort(ptr, 4);  addLong(ptr, val);
}

/*---------------------------------------------------------------------------*/
void addStrParam(uchar **ptr, short id, char *buf, short len)
{
  short  i;

  addShort(ptr, id);  addShort(ptr, len);

  for (i=0; i<len; ++i)  { **ptr = (uchar)*buf;  ++(*ptr);  ++buf; }
}


//-----------------------------------------------------------------------------
//  return the time as a millisecond counter that wraps around
//-----------------------------------------------------------------------------
#ifndef RABBIT
ulong GetMS(void)
{
  struct timespec  timeData;


  clock_gettime(CLOCK_REALTIME, &timeData);

  return (timeData.tv_sec * 1000L) + (timeData.tv_nsec / 1000000L);
}
#endif


//-----------------------------------------------------------------------------
//  return the time in seconds
//-----------------------------------------------------------------------------
#ifndef RABBIT
ulong GetSeconds(void)
{
  struct timespec  timeData;


  clock_gettime(CLOCK_REALTIME, &timeData);

  return timeData.tv_sec;
}
#endif


/*---------------------------------------------------------------------------*
 *  format a MAC addr as an ASCII string
 *---------------------------------------------------------------------------*/
char *mactoa(char *str, uchar far *macAddr)
{
  sprintf(str, "%02x:%02x:%02x:%02x:%02x:%02x",
      macAddr[0], macAddr[1], macAddr[2], macAddr[3], macAddr[4], macAddr[5] );

  return str;
}


//-----------------------------------------------------------------------------
//  convert host-byte-ordered inet value to octet-dot-delimited string
//-----------------------------------------------------------------------------
const char *inet_htop(int af, const void *src, char *dst, socklen_t cnt)
{
  ulong  netLong;
  uchar  *ptr;

  if (af != AF_INET)  { ERRNO = EAFNOSUPPORT;  return NULL; }

  netLong = htonl(*(ulong *)src);
  ptr = (uchar *)&netLong;

  sprintf(dst, "%d.%d.%d.%d", ptr[0], ptr[1], ptr[2], ptr[3]);

  return dst;
}


//-----------------------------------------------------------------------------
//  convert octet-dot-delimited string to host-byte-ordered inet value
//-----------------------------------------------------------------------------
int inet_ptoh(int af, const char *src, void *dst)
{
  if (af != AF_INET)  { ERRNO = EAFNOSUPPORT;  return -1; }

  inet_pton(af, src, dst);
  *((ulong *)dst) = ntohl(*(ulong *)dst);

  return 1;
}



//-----------------------------------------------------------------------------
//  return unsigned-long value for network-ordered (MSB-1st) IPV4 IP address
//-----------------------------------------------------------------------------
ulong ipv4_ntohl(uchar *ipAddr)
{
  return
    (ulong)ipAddr[0] << 24 |
    (ulong)ipAddr[1] << 16 |
    (ulong)ipAddr[2] << 8  |
    (ulong)ipAddr[3];
}

//-----------------------------------------------------------------------------
//  copy host-style IPV4 unsigned-long IP addr to network-ordered (MSB-1st)
//  string of bytes
//-----------------------------------------------------------------------------
void ipv4_htonl(ulong ipAddrLong, uchar *ipAddrStr)
{
  ipAddrStr[0] = (uchar) (ipAddrLong >> 24);
  ipAddrStr[1] = (uchar) (ipAddrLong >> 16);
  ipAddrStr[2] = (uchar) (ipAddrLong >>  8);
  ipAddrStr[3] = (uchar) ipAddrLong;
}


//-----------------------------------------------------------------------------
uint fastCRC16(uchar *ucpBuf, int nSize, ushort accum)
{
  ushort  x, crc;
  int  i;


  for (crc=accum,i=0; i<nSize; ++i)  {
    x = crc ^ (ushort)ucpBuf[i];
    crc = (crc >> 8) ^ crc16_rev_table[x & 0x00FF];
  }

  return crc;
}

//-----------------------------------------------------------------------------
//  compute a 16-bit CRC on the specified block of memory
//-----------------------------------------------------------------------------
uint memCRC16(ulong startAddr, ulong size)
{
  ushort  x, crc;
  uchar far *ptr, *endPtr;


  ptr = (uchar far *)startAddr;  endPtr = ptr + size;

  for (crc=0; ptr<endPtr; ++ptr)  {
    x = crc ^ (ushort)(*ptr);
    crc = (crc >> 8) ^ crc16_rev_table[x & 0x00FF];
    WDT_2S();
  }

  return crc;
}


//-----------------------------------------------------------------------------
//  Format a string for display purposes - changes CR's to '[' and LF's to ']'
//  All other non-displayable chars are changed to '*'
//-----------------------------------------------------------------------------
char *FmtStrCmd(char *cmdBuf, char *dispBuf, short dispBufSize)
{
  short  i;
  char  ch;


  for (i=0; cmdBuf[i] AND (i<dispBufSize-1); ++i)  {
    dispBuf[i] = ch = cmdBuf[i];
    if ((ch < ' ') OR (ch > '~'))  {
      if (ch == 0x0a)  dispBuf[i] = '[';  else
      if (ch == 0x0d)  dispBuf[i] = ']';  else
      dispBuf[i] = '*';
    }
  }

  dispBuf[i] = 0;

  return dispBuf;
}

//-----------------------------------------------------------------------------
//  Given a pointer to a char buffer and the number of chars within it, the
//  order of the characters in that buffer are reversed. This function destroys
//  the original string and store the resulting string in the same buffer.
//-----------------------------------------------------------------------------
char* ReverseStr(char *buffer, char dataCount)
{
  char  temp, *left, *right;

  left = buffer;  right = buffer + dataCount - 1;
  while ( left < right ) {
    temp = *left;  *left = *right;  *right = temp;
    ++left;  --right;
  }
  return buffer;
}

//-----------------------------------------------------------------------------
//  pad a string with blanks until it is the specified length (does nothing if
//  the string is already >= the specified length).  Returns the address of
//  the string.
//-----------------------------------------------------------------------------
char *padStr(char *str, int toLen)
{
  int  i;

  for (i=0; i<toLen; ++i)  if (! str[i])  break;
  if (i == toLen)  return str;
  while (i < toLen)  str[i++] = ' ';
  str[toLen] = 0;

  return str;
}


//-----------------------------------------------------------------------------
uchar far *fstrncat(uchar far *dst, uchar far *src, size_t size)
{
  uint  i;
  uchar far  *ptr;

  ptr = dst;
  while (*ptr) ++ptr;
  for (i=0; (i<size) AND *src; ++i)  *ptr++ = *src++;
  *ptr = 0;

  return dst;
}


//-----------------------------------------------------------------------------
//  case insensitive version of strstr()
//-----------------------------------------------------------------------------
char *strstri(char *inStr, char *lookfor)
{
  char *start, *p1, *p2;

  for (start=inStr; *start; ++start)  {
    p1 = start;  p2 = lookfor;
    while (*p1 AND *p2 AND (tolower(*p1) == tolower(*p2)))  { ++p1;  ++p2; }
    if (!*p2)  return start;
  }

  return NULL;
}

//-----------------------------------------------------------------------------
//  Parse a given string in to a list of NULL-terminated strings
//
//  Returns the # of blank-separated strings found
//-----------------------------------------------------------------------------
int parseArgs(char *buf, int maxArgs, char **argv, int *argc)
{
  char  *ptr;
  char  endCh;


  argv[0] = NULL;  ptr = buf;

  for (*argc=0; *argc<maxArgs; )  {
  //--- find start of next argument ---
    endCh = ' ';
    while (*ptr AND (*ptr <= ' '))  ++ptr;
    if (*ptr < ' ')  break;

  //--- check for start of quoted string ---
    if ((*ptr == '"') OR (*ptr == '\''))  {
      endCh = *ptr;  ++ptr;  if (*ptr < ' ')  break;
    }

  //--- save start of new argument ---
    argv[*argc] = ptr;  ++*argc;

  //--- locate end of argument ---
    while ((*ptr >= ' ') AND (*ptr != endCh))  ++ptr;

  //--- mark end of argument ---
    if (*ptr)  { *ptr = 0;  ++ptr; }
  }

  return *argc;
}


//-----------------------------------------------------------------------------
//  Converts the given long integer to a decimal format we can send back
//  in a debug-poll reply.
//-----------------------------------------------------------------------------
long DecimalRep(long num)
{
  long z, k, r;
  int ii;

  z = num;
  r = 0;
  for (ii = 0; ii < 8; ii++) {
    k = z % 10;
    z = z / 10;
    k <<= 28;
    r >>= 4;
    r &= 0x0FFFFFFFL;
    r |= k;
  }
  ReverseStr((char *)&r,4);
  return(r);
}

//---------------------------------------------------------------------------
//  initialize power of 10 table
//---------------------------------------------------------------------------
void initPow10(void)
{
  uint  ui;
  float  f;

  for (ui=0,f=1.0; ui<=MaxPow10; ++ui,f*=10.0)  powOf10[Pow0Idx+ui] = f;
  for (ui=1,f=0.1; ui<=MaxPow10; ++ui,f*=0.1)   powOf10[Pow0Idx-ui] = f;
}

//---------------------------------------------------------------------------
//  Passed a real number, returns the number expressed in engineering
//  notation format  [+/-]XXX.X(..)E[+/-]XX
//---------------------------------------------------------------------------
char *Engineering(float num, int sigDigs, char *str)
{
  uint  ui, idigs;
  int  pow;
  char  numSign, expSign;
  char  temp[32];
  float  absNum;


//--- make sure our power-of-10 table is initialized ---
  if (powOf10[Pow0Idx+1] != 10.0)  initPow10();

//--- check for 0 ---
  if (num == 0.0)  {
    strcpy(str, " 0.");
    for (ui=3; ui<sigDigs+2; ++ui)  str[ui] = '0';
    strcpy(str+sigDigs+2, "E+00");
    return str;
  }

  if (sigDigs > 8)  sigDigs = 8;

  memcpy(str, "*ERR*", 5);  memset(str+5, ' ', sigDigs+1);  str[sigDigs+6] = 0;

//--- determine exponent to get normalized value between 0 and 10 ---
  if (num < 0.0)  {
    absNum = -num;  numSign = '-'; }
  else {
    absNum = num;  numSign = ' '; }

  if (absNum < 1.0)  {
    for (ui=Pow0Idx; powOf10[ui]>absNum; --ui)
      if (ui < sigDigs)  return str;
  }
  else  {
    for (ui=Pow0Idx; powOf10[ui]<=absNum; ++ui)
      if (ui > (MaxPow10<<1))  return str;
    --ui;
  }
  absNum /= powOf10[ui];
//printf("exponent (before rounding): %d (%ldE-6)\r\n",
//       Pow0Idx - (int)ui, (long)(absNum*1000000.0) );  //tdebug

//--- round the result and adjust exponent if needed ---
  absNum += 5.0 / PowOf10((uint)sigDigs);
  if (absNum >= 10.0)  { absNum /= 10.0;  ++ui; }
//printf("exponent (after rounding):  %d (%ldE-6)\r\n",
//       Pow0Idx - (int)ui, (long)(absNum*1000000.0) );  //tdebug

  pow = (int)ui - Pow0Idx;  idigs = 1;

//--- if at least 4 significant digits, adjust exponent to a mult of 3 ---
  if (sigDigs >= 4)
    while (pow % 3)  { absNum = fmult(absNum, 10.0); --pow;  ++idigs; }

//--- format the # without the exponent ---
  fmtFloat(absNum, sigDigs+1, sigDigs-idigs, temp);

//--- add the exponent portion ---
  if (pow >= 0)  expSign = '+';  else { expSign = '-';  pow = -pow; }
  sprintf(str, "%c%sE%c%02d", numSign, temp, expSign, pow);


  return(str);
}

//---------------------------------------------------------------------------
//  Format a floating point for display (much faster than printf %#.#f)
//  Limitations:
//  - ttlLen > 1 and < 9
//  - decDigs > 0 and < ttlLen
//  - any rounding must be done before calling this func
//---------------------------------------------------------------------------
char *fmtFloat(float num, int ttlLen, int decDigs, char *str)
{
  int  i, j, decIdx, intDigs;
  char  save1st;

//tdebug
//  fprintf(DiagLog, "len: %d, dec:%d, num: %5d.%04d\r\n",
//                    ttlLen, decDigs, (int)num,
//                    abs((int)((num-(float)((int)num))*10000.0))
//         );  //tdebug

//--- make sure our power-of-10 table is initialized ---
  if (powOf10[Pow0Idx+1] != 10.0)  initPow10();


//--- if # won't fit in specified format, return a str of '*' ---
  intDigs = ttlLen - decDigs - 1;
  for (;;)  {
    if (num < 0.0)  {
      if (fabs(num) < PowOf10((uint)intDigs-1))  break;  }
    else
      if (num < PowOf10((uint)intDigs))  break;
    for (i=0; i<ttlLen; ++i)  str[i] = '*';
    str[ttlLen] = 0;  return str;
  }

//--- get all the digits we want in to a long ---
;;  sprintf(str, "%0*ld", ttlLen-1, (long)fmult(num, PowOf10((uint)decDigs)) );
  sprintf(str, "%0*ld", ttlLen-1, (long)(num * PowOf10((uint)decDigs)) );

//--- insert a decimal point at the appropriate spot ---
  decIdx = ttlLen - 1 - decDigs;
  for (i=ttlLen; i>decIdx; --i)  str[i] = str[i-1];
  str[decIdx] = '.';

//--- eliminate unwanted leading zeros and move sign char---
  save1st = *str;  if (save1st == '-')  *str = '0';
  for (i=0,j=-1; (i<decIdx-1) AND (str[i]=='0'); ++i,++j)  str[i] = ' ';
  if (j >= 0)  if (save1st == '-')  str[j] = '-';

  return str;
}


//-----------------------------------------------------------------------------
//  Format the specified Least significant # of bits of an integer value as
//  a binary number.  The left-most bit is the Most significant bit of the
//  selected # least significant bits (e.g. numBits = 4 means format the 4
//  least significant bits, left-most = num & 16, rightmost = num & 1)
//-----------------------------------------------------------------------------
char *LongToBinMS1st(long num, char *buf, int numBits)
{
  uint  i;
  char  *ptr;

  ptr = buf + numBits;  *ptr-- = 0;
  for (i=numBits; i; --i,num>>=1)  *ptr-- = '0' + ((uchar)num & 1);
  return buf;
}


//-----------------------------------------------------------------------------
//  Format the specified Least significant # of bits of an integer value as
//  a binary number.  The left-most bit is the Least significant bit of the
//  selected # least significant bits (e.g. numBits = 4 means format the 4
//  least significant bits, left-most = num & 1, rightmost = num & 16)
//-----------------------------------------------------------------------------
char *LongToBinLS1st(long num, char *buf, int numBits)
{
  uint  i;
  char  *ptr;

  ptr = buf;
  for (i=numBits; i; --i,num>>=1)  *ptr++ = '0' + ((uchar)num & 1);
  *ptr = 0;
  return buf;
}


//-----------------------------------------------------------------------------
//  format # of seconds as days, hours, mins, secs
//-----------------------------------------------------------------------------
char *fmtDHMS(char *str, ulong count)
{
  int  days, hours, mins, secs;

  secs = (int)(count % 60);  count /= 60;
  mins = (int)(count % 60);  count /= 60;
  hours = (int)(count % 24);
  days = (int)(count / 24);

  sprintf(str, "%dd %02d:%02d:%02d", days, hours, mins, secs);

  return str;
}


//-----------------------------------------------------------------------------
//  convert a date string (of the form "Day Mon dd yyyy" or "Mon dd yyyy") and
//  a time string (in the form "hh:mm:ss") in to a time_t value (seconds since
//  the epoch: Jan 01 1970 00:00:00)
//-----------------------------------------------------------------------------
time_t  compileTime(char *dateStr, char *timeStr)
{
  uchar  i;
  uint  offset = 0;
  char *ptr;
  struct tm  tim;

  for (i=0; i<12; ++i)
    if ((ptr = strstr(dateStr, months[i])) != NULL)  {
      offset = ptr - dateStr;  break; }

  tim.tm_sec = atoi(timeStr+6);
  tim.tm_min = atoi(timeStr+3);
  tim.tm_hour = atoi(timeStr);

  tim.tm_mday = atoi(dateStr+4+offset);
  tim.tm_mon = i;
  tim.tm_year = atoi(dateStr+7+offset) - 1900;

  tim.tm_wday = tim.tm_yday = tim.tm_isdst = 0;

  return  mktime(&tim);
}


//-----------------------------------------------------------------------------
//  Get the # of bytes available for reading on the specified descriptor
//
//  On success, returns the # of unread bytes.  Returns < 0 if an error
//-----------------------------------------------------------------------------
int bytesAvail(fd_t fd)
{
  int  count;

  if (ioctl(fd, FIONREAD, &count) != 0)  return -1;

  return count;
}


//-----------------------------------------------------------------------------
//  check the std input fd[0] for the char ETX (Ctl-C).  If found, exit the
//  calling process with a status of -3
//
//Finish:  This is a temporary workaround for tty shells.  Once I finish
//  implementing signals, the shell process can check for this sort of thing
//  and send a terminate signal to its child process(es).
//-----------------------------------------------------------------------------
int CheckETX(void)
{
  int  count, fflags;
  uchar  ch, toss;


  ch = 0;

  fflags = fcntl(0, F_GETFL, 0);
  if ((fflags == -1) AND (errno == EBADF))  return -1;

  fcntl(0, F_SETFL, fflags | O_NONBLOCK);

  count = read(0, &ch, 1);  // read the first character (if any)

  if (count > 0)  // toss away any additional chars except a Ctrl-C
    while (read(0, &toss, 1) > 0)  if ((toss & 0x7f) == 0x03)  Exit(-3);

  fcntl(0, F_SETFL, fflags);

  if (count == 0)  { fprintf(DiagLog, "connection closed\r\n");  Exit(-1); }

  if (count < 0)  {
    if (errno == EAGAIN)  return -1;
    perror("CheckETX()/stdin");  Exit(-2);
  }


  if ((ch & 0x7f) == 0x03)  Exit(-3);

  return ch;
}


//-----------------------------------------------------------------------------
//  Read and ignore bytes from an fd.
//
//  timeOut values > 0 tell it the max # of ms that can pass without making
//  any progress toward reading the specified # of bytes.  Values < 0 means
//  block until all the bytes are read or an error occurs.
//
//  Returns the # of bytes read or -1 if an error
//-----------------------------------------------------------------------------
int  skipBytes(fd_t fd, int numBytes, ulong timeOut)
{
  int  remain, count, stat;
  char  buf[64];


  for (remain=numBytes; remain>0; remain-=stat)  {
    count = (remain > sizeof(buf) ? sizeof(buf) : remain);
    if ((stat = read_(fd, buf, count, timeOut)) <= 0)  break;
  }

  return (numBytes - remain);
}

//-----------------------------------------------------------------------------
//  Wrapper around write_() to write from a buffer located in extended
//  memory
//-----------------------------------------------------------------------------
int fwrite_(fd_t fd, char far *from, int numBytes,
            char *buf, uint bufSize, ulong inactiveTimeout)
{
  int  count, remain, stat;
  char far  *ptr;
  FDEV far  *file;


  ptr = from;  count = bufSize;  file = _fdToOpenFile_(fd);

  for (remain=numBytes; remain; remain-=count)  {
    if (count > remain)  count = remain;
    fmemcpy(buf, ptr, count);  ptr += (uint)count;
    stat = writeFDEV(file, buf, count, inactiveTimeout);
    if (stat == count)  continue;  // no timeout, no error
    if (stat < 0)  return stat;  // an error occured
    remain -= stat;  break;  // timeout expired
  }

  return  (numBytes - remain);
}

//-----------------------------------------------------------------------------
//  Function to do buffered reading from a file descriptor and return the next
//  unused char from the buffer.  Returns -1 if no chars left in the buffer
//  and no more chars could be read from the descriptor.
//-----------------------------------------------------------------------------
int  readBufCh(BufInfo *bufInfo)
{
  int  ch, stat;

  for (;;)  {
  //--- return next unread char in the buffer if there is one ---
    if (bufInfo->nextOut < bufInfo->count)  {
      ch = (uint) bufInfo->buf[bufInfo->nextOut];
      ++bufInfo->nextOut;  return ch;
    }

  //--- read as many more chars as we can in to the buffer ---
  //--- (assumes the fd is in non-blocking mode) ---
    if ((stat = read(bufInfo->fd, bufInfo->buf, bufInfo->bufSize)) < 1)
      return -1;

    bufInfo->nextOut = 0;  bufInfo->count = stat;
  }

}


//-----------------------------------------------------------------------------
//  sleepMS() implementation for Linux/Unix systems
//-----------------------------------------------------------------------------
#ifndef RABBIT
ulong sleepMS(ulong ms)
{
  ulong  remainMS;
  struct timespec  sleepFor, sleepRemain;


  sleepFor.tv_sec = ms / 1000;  sleepFor.tv_nsec = (ms % 1000) * 1000000L;

  if (!nanosleep(&sleepFor, &sleepRemain))  {
    remainMS = sleepRemain.tv_sec * 1000 + sleepRemain.tv_nsec / 1000000L;
    return (remainMS > ms) ? 0 : (ms - remainMS);
  }

  return ms;
}
#endif


//-----------------------------------------------------------------------------
//  Attempt to read specified # of bytes from a device
//
//  maxIdleMS determines blocking mode as follows:
//    < 0: indefinite timeout (i.e. block until done or error)
//      0: non-blocking:  read as much as possible and return
//    > 0: max # of ms to wait with no progress
//
//  Note that a positive value does NOT place an upper limit on how long this
//  function will wait to read the requested # of bytes, but rather the max
//  time it will wait while NO PROGRESS is being made at reading the bytes.
//
//  Note that if non blocking mode is set for the file descriptor then the
//  maxIdleMS value is ignored (treated as if it were 0)
//
//  If a timeout > 0 expires, errno is set to ETIME (regardless of what the
//  return value is, which could be > 0 if some bytes were read).
//-----------------------------------------------------------------------------
int read_(fd_t fd, void *buf, int numBytes, long maxIdleMS)
{
  #ifdef RABBIT
    return readFDEV(_fdToOpenFile_(fd), buf, numBytes, maxIdleMS);
  #else
    int  fflags, count, remain, numRead;
    long  idleTime;

    fflags = fcntl(fd, F_GETFL, 0);
    fcntl(fd, F_SETFL, fflags | O_NONBLOCK);  // insure non-blocking reads

    numRead = 0;  remain = numBytes;  idleTime = 0;
    while (remain > 0)  {
      errno = 0;
      count = read(fd, buf, remain);
      if (count > 0)  {
        remain -= count;  numRead += count;  buf += count;
        idleTime = 0;  continue;
      }
      if (count == 0)  break;  // EOF
      if (errno != EAGAIN)  break;
      if (!maxIdleMS)  break;  // errno left as EAGAIN if non-blocking read
      if (idleTime > maxIdleMS)  { errno = ETIME;  break; }
      idleTime += sleepMS(10);
    }

    fcntl(fd, F_SETFL, fflags);  // restore fd flags

    if ((count < 0) AND (numRead == 0))  return -1;  // NOT EOF & nothing read

    return numRead;
  #endif
}

//-----------------------------------------------------------------------------
//  Attempt to write specified # of bytes to a device.
//
//  maxIdleMS determines blocking mode as follows:
//    < 0: indefinite timeout (i.e. block until done or error)
//      0: non-blocking:  write as much as possible and return
//    > 0: max # of ms to wait with no progress
//
//  Note that a positive value does NOT place an upper limit on how long this
//  function will wait to write the specified # of bytes, but rather the max
//  time it will wait while NO PROGRESS is being made at writing the bytes.
//
//  Note that if non blocking mode is set for the file descriptor then the
//  maxIdleMS value is ignored (treated as if it were 0)
//
//  If a timeout > 0 expires, errno is set to ETIME (regardless of what the
//  return value is, which could be > 0 if some bytes were written).
//-----------------------------------------------------------------------------
int write_(fd_t fd, void *buf, int numBytes, long maxIdleMS)
{
  #ifdef RABBIT
    return writeFDEV(_fdToOpenFile_(fd), buf, numBytes, maxIdleMS);
  #else
    int  fflags, count, remain, numSent;
    long  idleTime;

    fflags = fcntl(fd, F_GETFL, 0);
    fcntl(fd, F_SETFL, fflags | O_NONBLOCK);  // insure non-blocking reads

    numSent = 0;  remain = numBytes;  idleTime = 0;
    while (remain > 0)  {
      errno = 0;
      count = write(fd, buf, remain);
      if (count > 0)  {
        remain -= count;  numSent += count;  buf += count;
        idleTime = 0;  continue;
      }
      if (count == 0)  break;  // EOF
      if (errno != EAGAIN)  break;
      if (idleTime > maxIdleMS)  { errno = ETIME;  break; }
      idleTime += sleepMS(10);
    }

    fcntl(fd, F_SETFL, fflags);  // restore fd flags

    return numSent;
  #endif
}


//-----------------------------------------------------------------------------
//  Copy a file by name.  Deletes an existing file with the toName before
//  doing the copy.
//
//  On success, returns 0
//  On error, returns -1 and *errContext points to name of file that produced
//  the error.
//-----------------------------------------------------------------------------
int copyFileByName(char *fromName, char *toName, bool showProgress,
                   char *buf, int bufSize, char **errContext)
{
  ulong  remain, fromSize;
  struct stat  fsInfo;
  int  fromfd, tofd, count, reqCount;


  *errContext = "Error";

  if ((fromfd = open(fromName, O_RDONLY, 0)) < 0)  {
    *errContext = fromName;  return -1; }
  if (fstat(fromfd, &fsInfo) < 0)  { *errContext = fromName;  Exit(-2); }
  fromSize = fsInfo.st_size;

  unlink(toName);

  if ((tofd = open(toName, O_RDWR | O_CREAT, 0666)) < 0)  {
      close(fromfd);  *errContext = toName;  return -1;  }

  remain = fromSize;  reqCount = bufSize;
  while (remain)  {
    if (remain < (ulong)bufSize)  reqCount = (uint)remain;
    if ((count = read(fromfd, buf, reqCount)) < reqCount)  {
      *errContext = fromName;  break;  }
    if (write(tofd, buf, count) != count)  {
      *errContext = toName;  break;  }
    remain -= count;
    if (showProgress)  { printf("%10lu\r", fromSize-remain);  fflush(stdout); }
  }
  if (showProgress)  printf("\r\n");

  close(fromfd);  close(tofd);

  if (remain)  return -1;

  return 0;
}


//-----------------------------------------------------------------------------
//  Manage the process of copying bytes from an arbitray source to an
//  arbitrary destination.  It calls the provided functions to read and
//  write bytes in groups until an error occurs or the specified # of bytes
//  has been processed.
//
//  Parameters:
//    readFunc:      Pointer to function to read bytes
//    readContext:   Pointer to information needed by the readFunc to manage
//                   the reading of a group of bytes
//
//    writeFunc:     Pointer to a function to write bytes
//    writeContext:  Pointer to information needed by the writeFunc to manage
//                   the writing of a group of bytes
//
//    *buf:          Pointer to a block of memory to be used for transferring
//                   bytes from the readFunc to the writeFunc
//    bufSize:       The size of the memory block pointed to *buf
//
//    showProgres:   Specifies whether or not a count of the # of bytes
//                   transferred so far should be displayed.
//
//  The read and write functions must be written to read or write the # of
//  bytes they are told to and return the # of bytes they successfully copied
//  to/from the buffer.
//
//  The read and write Context pointers are used to provide the read and write
//  functions with whatever context or state information they need to carry
//  out their tasks.  The values pointed to by each should be initialized as
//  needed before calling copyBytes, and the read and write functions can do
//  whatever they need to the values pointed to.
//
//  Returns the # of bytes successfully copied.
//-----------------------------------------------------------------------------
ulong copyBytes(ReadFunc readFunc, void *readContext,
                WriteFunc writeFunc, void *writeContext,
                char *buf, int bufSize, ulong copyCount, bool showProgress)
{
  ulong  remain;
  int  count;


  remain = copyCount;  count = bufSize;
  while (remain)  {
    if (remain < (ulong)bufSize)  count = (uint)remain;
    if ((readFunc)(readContext, buf, count) != count)  break;
    if ((writeFunc)(writeContext, buf, count) != count)  break;
    remain -= count;
    if (showProgress)  { printf("%10lu\r", copyCount-remain);  fflush(stdout); }
  }
  if (showProgress)  printf("\r\n");

  return copyCount - remain;
}


//----------------------------------------------------------------------------
//  Send a pakcet using a UDP socket set for broadcast mode.
//
//  Returns 0 if successful.  If an error occurs, it closes the socket and
//  returns -1.
//-----------------------------------------------------------------------------
int sendBCastPkt(int *sockfd, int port, ulong bcIPv4Addr,
                 char far *pktBuf, int pktLen)
{
  int  stat, addrLen, fflags;
  SAI  bcAddr;


  addrLen = sizeof(SAI);

/*--- Form the broadcast address ---*/
  memset(&bcAddr, 0, addrLen);
  #ifdef  vxWorks
    bcAddr.sin_len = addrLen;
  #endif
  bcAddr.sin_family = AF_INET;
  bcAddr.sin_port = htons(port);
  bcAddr.SIN_ADDR = htonl(bcIPv4Addr);

//--- set the socket descriptor to blocking mode ---
  fflags = fcntl(*sockfd, F_GETFL, 0);
  fcntl(*sockfd, F_SETFL, fflags & ~O_NONBLOCK);

//--- send the broadcast packet ---
  stat = sendto(*sockfd, (void far *)pktBuf, pktLen, 0, (SAP)&bcAddr, addrLen);
  if (stat < 0)  {
    perror("sendBCastPkt()");  close(*sockfd);  *sockfd = -1;  return -1; }

//--- reset the flags for the socket descriptor ---
  fcntl(*sockfd, F_SETFL, fflags);

  return 0;
}



//-----------------------------------------------------------------------------
//  Check a list of strings to see if a given string is in the list
//  The end of strList is denoted by a NULL ptr
//
//  Returns the index in to the list if found, or -1 if not
//-----------------------------------------------------------------------------
int strListIdx(char **listPtr, char *testStr)
{
  int  i;


  if (!listPtr OR !testStr)  return -1;

  for (i=0; listPtr[i]; ++i)
    if (stricmp(testStr, listPtr[i]) == 0)  return i;

  return -1;
}


//-----------------------------------------------------------------------------
//  format the current time in to the supplied buffer in the format:
//      yyyy-dd-mm hh:mm:ss.mmm
//-----------------------------------------------------------------------------
char *dateTimeMS(char *buf, const time_t dateTime, uint ms)
{
  struct tm  tms;

  localtime_r(&dateTime, &tms);

  sprintf(buf, "%04d-%02d-%02d %02d:%02d:%02d.%03u",
         tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday,
         tms.tm_hour, tms.tm_min, tms.tm_sec, ms );

  return buf;
}

//-----------------------------------------------------------------------------
//  print a "header" line common to several user commands
//-----------------------------------------------------------------------------
void stdHeader(void)
{
  uint  ctlRate;
  ulong  upSecs;
  time_t  dateTime;
  struct tm  tms;
  char  msg[16];


//2011-12-29 13:46:49   Idle: 42%  Control: 100/s  [51:0026]   Up: 0d 02:57:02

  CursorTo(0, 0);

  _EnterCritical_();
  ctlRate = ctlLoopTicksPerSec;  dateTime = TIME();
    _ExitCritical_();

  upSecs = GetSeconds();;

  localtime_r(&dateTime, &tms);
  printf("%04d-%02d-%02d %02d:%02d:%02d",
         tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday,
         tms.tm_hour, tms.tm_min, tms.tm_sec );

  printf("   Idle: %2d%%  Control: %3u/s  [%02X:%04lX]   Up: %s %s\r\n",
    percentIdle, ctlRate, devType, deviceID, fmtDHMS(msg, upSecs), cllCmd);
}

//-----------------------------------------------------------------------------
void showNetState(int iface)
{
  IFConfig  ifConfig;
  char  *stateDesc;
  char  macStr[18], ipStr[16], bcStr[16], maskStr[16];

/* Sample output for ifconfig command on a Linux box:

eth0      Link encap:Ethernet  HWaddr 00:11:09:65:E8:15
          inet addr:35.9.57.207  Bcast:35.15.255.255  Mask:255.248.0.0
          inet6 addr: fe80::211:9ff:fe65:e815/64 Scope:Link
          UP BROADCAST RUNNING MULTICAST  MTU:1500  Metric:1
          RX packets:273501680 errors:0 dropped:0 overruns:0 frame:0
          TX packets:45036584 errors:0 dropped:0 overruns:0 carrier:0
          collisions:0 txqueuelen:1000
          RX bytes:2732679160 (2.5 GiB)  TX bytes:3498902465 (3.2 GiB)
          Interrupt:169 Base address:0x6f00
*/

  getIFConfig(iface, &ifConfig);

  mactoa(macStr, ifConfig.macAddr);
  inet_htop(AF_INET, &ifConfig.ipAddr,  ipStr,   sizeof(ipStr));
  inet_htop(AF_INET, &ifConfig.bcAddr,  bcStr,   sizeof(bcStr));
  inet_htop(AF_INET, &ifConfig.netmask, maskStr, sizeof(maskStr));

  switch (ifConfig.state)  {
    case IF_DOWN:         stateDesc = "DOWN";         break;
    case IF_COMING_UP:    stateDesc = "COMING UP";    break;
    case IF_UP:           stateDesc = "UP";           break;
    case IF_COMING_DOWN:  stateDesc = "COMING DOWN";  break;
    default:              stateDesc = "UNKNOWN";
  }


  printf("eth%d      Link encap:Ethernet  HWaddr %.18s\r\n", iface, macStr);
  printf("          inet addr:%.16s  Bcast:%.16s  Mask:%.16s\r\n",
         ipStr, bcStr, maskStr);
  printf("          %s BROADCAST MTU:%d\r\n", stateDesc, ifConfig.mtu);
}

//-----------------------------------------------------------------------------
//  Add an EventMsg to the end of an event queue FIFO
//-----------------------------------------------------------------------------
// PARAMTERS:
//
//   eventsfd:
//      The file descriptor to which the EventMsg is to be written.  If < 0,
//      this function will attempt to write the msg in to the standard
//      application process' event queue.
//
//   eventFunc:
//      A pointer to the function to be called by the task that processes
//      the events in the queue (normally the main application process).
//      A pointer to a copy of the message will be passed to the function.
//
//   chan:
//      For controllers that manage multiple devices, the channel or device
//      that the message applies to.  Use 0 if the command does not apply to
//      a specific channel (or there IS only 1 channel).
//
//   eventMsg:
//      A pointer to a buffer to use for constructing an EventMsg structure.
//
//   data:
//      A pointer to any data for the message.  Specify NULL if there is no
//      data for the message, or the data is already in the EventMsg structure
//      pointed to by the eventMsg pointer.
//
//   dataLen:
//      The # of bytes of data to be sent in the message.  If dataLen > 0 and
//      the data pointer is NULL, the assumption is the data already resides
//      in the EventMsg structure pointed to by the eventMsg parameter.
//
//
// RETURNS:
//     0 on success, < 0 if an error
//
//
//  NOTES:
//    - The data for a message must NOT contain pointers to local (auto) data
//      values since these pointers will not be valid in the context of the
//      process that actually processes the message.  Pointers to global
//      shared memory values are allowed (alothough keep in mind that the
//      value pointed to could change by the time the msg is processed).
//
//    - Use of separate pointers for the EventMsg structure and the msg data
//      provides 2 important features for msgs with larger data sizes:
//      - No need to provide stack large enough for a 2nd copy of the msg data
//      - No need to spend time copying the msg data from one buffer to another
//
//      The intent is that, for larger messages, the calling function will
//      simply read the msg data into a buffer holding an EventMsg structure
//      and specify NULL for the data pointer (even though the dataLen is > 0).
//
//    - Specifying -1 for the eventsfd and/or specifying a non-NULL pointer
//      for the data allows the calling function to minimize its size in
//      exchange for a bit more CPU time (to open the FIFO and/or copy the
//      data in to the EventMsg structure).
//
//-----------------------------------------------------------------------------
int putMsg(fd_t eventsfd, EventHandler eventFunc, int chan,
           EventMsg *eventMsg, void *data, int dataLen)
{
  bool  closefd;


//--- if the caller did not provide an fd to write to, then assume the ---
//--- msg should go in the standard application events FIFO queue ---
  closefd = false;
  if (eventsfd < 0)  {
    if ((eventsfd = open(EventsFIFO, O_WRONLY, 0)) < 0)  {
      perror(EventsFIFO);  return -1;  }
    closefd = true;
  }

  eventMsg->dataLen = dataLen;
  eventMsg->eventFunc = eventFunc;
  eventMsg->chanNum = chan;
  if (dataLen AND (data != NULL))  memcpy(eventMsg->data, data, dataLen);

  write(eventsfd, (char *)eventMsg, dataLen + MSG_HDR_SIZE);

  if (closefd)  close(eventsfd);

  return 0;
}


//-----------------------------------------------------------------------------
//  process the next event msg in the queue (if any)
//-----------------------------------------------------------------------------
void processMsg(fd_t eventsfd, EventMsg *eventMsg, int bufSize)
{
  int  ttlLen, readLen;


  if (read(eventsfd, (char *)eventMsg, 2) != 2)  return;

  ttlLen = eventMsg->dataLen + MSG_HDR_SIZE;
  readLen = ttlLen  - 2;

  if (ttlLen > bufSize)  {
    fprintf(DiagLog, "event msg too long: %d\n", ttlLen);
    skipBytes(eventsfd, readLen, 500);  return;
  }

  if (ttlLen < MSG_HDR_SIZE)  {
    fprintf(DiagLog, "event msg too short: %d\n", ttlLen);  return;  }

  if (read(eventsfd, (char *)eventMsg+2, readLen) != readLen)  {
    perror("processMsg");  return;  }

  if (ShowEventMsgs)
    fprintf(DiagLog, "EventMsg: Addr: 0x%08lx  dataLen: %d\n",
           (ulong)eventMsg->eventFunc, eventMsg->dataLen);

  (*(EventHandler)(eventMsg->eventFunc))(eventMsg);
}


/*
//-----------------------------------------------------------------------------
//  (slow) test of algorithm for floating point multiply
//-----------------------------------------------------------------------------
float near fmult(float f1, float f2)
{
  int  exp;
  ulong  result;
  uchar  exp1, exp2;
  uint  man1, man2;


#define  f1_0  (*(((uchar *)&f1)+3))
#define  f1_1  (*(((uchar *)&f1)+2))
#define  f1_2  (*(((uchar *)&f1)+1))
#define  f1_3  (*(((uchar *)&f1)+0))

#define  f2_0  (*(((uchar *)&f2)+3))
#define  f2_1  (*(((uchar *)&f2)+2))
#define  f2_2  (*(((uchar *)&f2)+1))
#define  f2_3  (*(((uchar *)&f2)+0))

#define  result_0  (*(((uchar *)&result)+3))
#define  result_1  (*(((uchar *)&result)+2))
#define  result_2  (*(((uchar *)&result)+1))
#define  result_3  (*(((uchar *)&result)+0))

#define  man1_0 (*(((uchar *)&man1)+1))
#define  man1_1 (*(((uchar *)&man1)+0))

#define  man2_0 (*(((uchar *)&man2)+1))
#define  man2_1 (*(((uchar *)&man2)+0))


  man1_0 = (f1_1 | 0x80);  man1_1 = f1_2;
  man2_0 = (f2_1 | 0x80);  man2_1 = f2_2;

  result =  ((ulong)man1 * man2);

//  result += (((ulong)m1l * (uint)f2_3) + ((ulong)m2l * (uint)f1_3)) >> 8;

//  result = result >> 8;
  result_3 = result_2;  result_2 = result_1;  result_1 = result_0;

  exp1 = ((f1_0 << 1) | (f1_1 >> 7));
  exp2 = ((f2_0 << 1) | (f2_1 >> 7));
  exp = (int)((uint)exp1 + (uint)exp2) - 254;
  while (!(result_1 & 0x80))  { result <<= 1; --exp; }

//  result_1 &= 0x7f;  result |= ((ulong)(exp + 128)) << 23;
  exp += 128;
  result_0 = ((uchar)exp >> 1);
  result_1 &= 0x7f;  result_1 |= ((uchar)exp << 7);

  if ((f1_0 ^ f2_0) & 0x80)  result_0 |= 0x80;


  return *((float *)&result);
}
*/



//-------------------------------------------------------------------------
//  faster but less accurate floating point multiply than the Softools C
//  version (this uses "mul" instructions like the DynamicC libs, while the
//  Softools C version uses shifts and meets IEEE standards)
//-------------------------------------------------------------------------
/*
#pragma warn -764  // turn off warning about not returning a value

float near fmult(float f1, float f2)
{
//  if ((f1 == 0.0) OR (f2 == 0.0))  return 0.0;  //tdebug
  ;
#asm
;;        .cseg

.result     equ   -20     ;; temp 32-bit mantissa for result
.man2A      equ   -16     ;; mantissa 2  (16-bits)
.man1A      equ   -14     ;; mantissa 1  (16-bits)
.resultExp  EQU   -8      ;; 1 byte
.f1         EQU    4      ;; IEEE float (4 bytes)
.f2         EQU    8      ;; IEEE float (4 bytes)


;;----- provided by compiler -----
;;        push    ix
;;
;;        ld      ix,0
;;        add     ix,sp
;;--------------------------------

        add     sp,-12             ; additional local variables

;  if ((f1 == 0.0) OR (f2 == 0.0))  return 0.0;
        ld      hl,(ix+.f1)
        ex      de,hl
        ld      hl,(ix+.f1+2)
        ld      bc,hl
        ld      a,b
        and     a,7Fh
        or      a,c
        or      a,d
        or      a,e
        jr      z,@501x
        ld      hl,(ix+.f2)
        ex      de,hl
        ld      hl,(ix+.f2+2)
        ld      bc,hl
        ld      a,b
        and     a,7Fh
        or      a,c
        or      a,d
        or      a,e
        jr      nz,@502x
@501x:
        ld      de,0
        ld      c,e
        ld      b,e
        jp      done
@502x:


;#define  f1_0  (*(((uchar *)&f1)+3))
;#define  f1_1  (*(((uchar *)&f1)+2))
;#define  f1_2  (*(((uchar *)&f1)+1))
;#define  f1_3  (*(((uchar *)&f1)+0))

;#define  f2_0  (*(((uchar *)&f2)+3))
;#define  f2_1  (*(((uchar *)&f2)+2))
;#define  f2_2  (*(((uchar *)&f2)+1))
;#define  f2_3  (*(((uchar *)&f2)+0))
;
;#define  FirstByte  (*(((uchar *)&u3)+2))


;=== D = ((f1_0 << 1) | (f1_1 >> 7));  // f1 exponent
    ld      hl,(ix+.f1+2)      ;[9]  HL = MSW(f1)
    ld      a,h                ;[2]  save byte with f1's sign bit
    ex      de,hl              ;[2]  DE = HL
    rl      de                 ;[2]  D = MSB(DE << 1) == f1 exponent
    ld      c,d                ;[2]  C = f1 exponent

;=== H = ((f2_0 << 1) | (f2_1 >> 7));  // f2 exponent
    ld      hl,(ix+.f2+2)      ;[9]  HL = MSW(f2)
    xor     h                  ;[2]  msbit of A has sign bit for result
    sla     a                  ;[4]  now carry bit has sign bit
    ex      af,af'             ;[2]  save carry bit with sign bit for result
    ex      de,hl              ;[2]  DE = HL
    rl      de                 ;[2]  D = MSB(DE << 1) == f2 exponent

;=== exp = (int)((uint)exp1 + (uint)exp2) - 254;
    bool    hl                 ;[2]  H = 0
    ld      b,h                ;[2]  BC = f1 exponent
    ld      l,d                ;[2]  HL = f2 exponent
    add     hl,bc              ;[2]  HL = f1 exp + f2 exp
    ld      de,0xff82          ;[6]  DE = 128-254
    add     hl,de              ;[2]  HL = result exponent
    ld      (ix+.resultExp),l  ;[10] save 8 bit result exponent

;; close approximation of result Mantissa computed as:
;;   man1A*man2A + (man1A*man2B + man2A*man1B) >> 8
;;
;; where man1A, man2A are 15 msb of each mantissa (incl implied 1 bit)
;;       man2B, man2B are  9 lsb of each mantissa
;;
;;   s-eeeeeee e-
;;             1-AAAAAAA AAAAAAA-B BBBBBBB

;=== BC = ((uint)(f1_1 | 0x80)<<8 | (uint)f1_2) >> 1;  // 15 msb of f1 mantisa
    bool    hl                 ;[2]  to clear carry flag
    ld      hl,(ix+.f1+1)      ;[9]  2nd and 3rd bytes of f1
    set     7,h                ;[4]  force implied 1 bit
    rr      hl                 ;[2]  carry -> [HL] -> carry
    ld      bc,hl              ;[4]  BC = 01xxxxxx xxxxxxxx
    push    bc                 ;[10] save result at (ix-14)

;=== DE = ((uint)(f2_1 | 0x80)<<8 | (uint)f2_2) >> 1;  // 15 msb of f2 mantisa
    bool    hl                 ;[2]  to clear carry flag
    ld      hl,(ix+.f2+1)      ;[9]  2nd and 3rd bytes of f2
    set     7,h                ;[4]  force implied 1 bit
    rr      hl                 ;[2]  carry -> [HL] -> carry
    ld      de,hl              ;[4]  DE = 01xxxxxx xxxxxxxx
    push    de                 ;[10] save result at (ix-16)

;=== multiply 2 15-bit mantissa values (msb is 0 for both values)
    mul                        ;[16] HL:BC = BC * DE

    push    hl                 ;[10] msw at (ix-18)
    push    bc                 ;[10] lsw at (ix-20)

;=== add in (man1A * man2B) >> 8
    ld      hl,(ix+.f2)        ;[9]  HL = 3rd and 4th bytes of f2
    ld      a,h                ;[2]
    and     a,0x01             ;[4]  keep on 9th bit
    ld      b,a                ;[2]  A
    ld      c,l                ;[2]  BC = 9 lsb of f1 mantissa
    ld      hl,(ix+.man1A)     ;[9]
    ex      de,hl              ;[2]  DE = 15 msb of f1 mantissa (with implied 1)
    mul                        ;[16] HL:BC = man1A * man2B
    ld      e,b                ;[2]
    ld      d,l                ;[2]  DE has 12-msb of 20-bit result
    rr      de                 ;[2]
    res     7,d                ;[4]
    ; add to 32-bit result mantissa
    ld      hl,(ix+.result)    ;[9]  HL = LSW(result)
    add     hl,de              ;[2]
    ld      (ix+.result),hl    ;[11]
    jr      nc,nocarry1        ;[5]
    ld      hl,(ix+.result+2)  ;[9]  HL = MSW(result)
    inc     hl                 ;[2]
    ld      (ix+.result+2),hl  ;[9]
nocarry1:
;=== add in (man2A * man1B) >> 8
    ld      hl,(ix+.f1)        ;[9]  HL = LSW(f1)
    ld      a,h                ;[2]
    and     a,0x01             ;[4]  keep on 9th bit
    ld      b,a                ;[2]  A
    ld      c,l                ;[2]  BC = 9 lsb of f1 mantissa
    ld      hl,(ix+.man2A)     ;[9]
    ex      de,hl              ;[2]  DE = 15 msb of f1 mantissa (with implied 1)
    mul                        ;[16] HL:BC = man2A * man1B
    ld      e,b                ;[2]
    ld      d,l                ;[2]  DE has 12-msb of 20-bit result
    rr      de                 ;[2]
    res     7,d                ;[4]
  ; add to 32-bit result mantissa
    ld      hl,(ix+.result)    ;[9]  HL = lsw of 32-bit result
    add     hl,de              ;[2]
    ld      (ix+.result),hl    ;[11]
    jr      nc,nocarry2        ;[5]
    ld      hl,(ix+.result+2)  ;[9]  HL = msw of 32-bit mantissa
    inc     hl                 ;[2]
    ld      (ix+.result+2),hl  ;[9]
nocarry2:

;=== set HL:DE = 32-bit result
    ld      hl,(ix+.result)    ;[9]
    ex      de,hl              ;[2]
    ld      hl,(ix+.result+2)  ;[9]

;=== shift HL:DE left 2 or 3 bits (until msb is a 1)

    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]
    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]

    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]
    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]

    ld      b,(ix+.resultExp)  ;[9]  B = 8-bit result exponent

    ;--- if msb is not 1, then shift one more time ---
    bit     7,h                ;[4]  F.z
    jr      nz,clear1stbit     ;[5]

    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]
    rl      de                 ;[2]  carry <- [DE] <- carry
    ex      de,hl              ;[2]

    dec     b                  ;[2]  and decr exponent

clear1stbit:
    res     7,h                ;[4]  reset msb (the implied 1 bit for mantissa)

;=== put result in to BC:DE
    ex      af,af'             ;[2]  carry bit contains sign bit for result
    rr      b                  ;[4]  carry -> [B] -> carry
    ld      c,0                ;[4]
    rr      c                  ;[4]  carry -> [C] -> carry
    ld      a,c                ;[2]
    or      h                  ;[2]
    ld      c,a                ;[2]
    ld      e,d                ;[2]
    ld      d,l                ;[2]

.COMMENT ~
;// ... a LOT more work for a TINY bit more accuracy....
;  u3 += (((ulong)m1l * (uint)f2_3) + ((ulong)m2l * (uint)f1_3)) >> 8;
    ld      a,(ix+8)
    bool    hl
    ld      l,a
    ex      de,hl
    ld      hl,(ix-10)
    ex      de,hl
    ld      b,h
    ld      c,h
    push    bc
    push    de
    ex      de,hl
    call    __ulmlt##
    ld      a,(ix+4)
    bool    hl
    ld      l,a
    push    bc
    push    de
    ex      de,hl
    ld      hl,(ix-12)
    ex      de,hl
    ld      b,h
    ld      c,h
    push    bc
    push    de
    ex      de,hl
    call    __ulmlt##
    call    __ladd##
    ld      e,d
    ld      d,c
    ld      c,b
    ld      b,0
    push    bc
    push    de
    ld      hl,(ix-4)
    ex      de,hl
    ld      hl,(ix-2)
    ld      bc,hl
    call    __ladd##
    ld      hl,bc
    ld      (ix-2),hl
    ex      de,hl
    ld      (ix-4),hl
    ex      de,hl

;  u3 = u3 >> 8;
    ld      e,d
    ld      d,c
    ld      c,b
    ld      b,0
    ld      hl,bc
    ld      (ix-2),hl
    ex      de,hl
    ld      (ix-4),hl
    ex      de,hl

;  if ((f1_0 ^ f2_0) & 0x80)  u3 |= 0x80000000L;
    ld      a,(ix+7)
    ld      hl,(ix+11)
    xor     a,l
    and     a,128
    jr      z,..endif
    ld      hl,(ix-2)
    ld      bc,hl
    set     7,b
    ld      hl,bc
    ld      (ix-2),hl
    ex      de,hl
    ld      (ix-4),hl
    ex      de,hl
..endif:
~


done:
    ld      sp,ix

;;    pop     ix
;;        ret

#endasm
}

#pragma warn 764  // turn back on warning about not returning a value
*/

//-----------------------------------------------------------------------------
//  Function to compare the copy of a structure's address stored in the
//  structure to the address of the structure as a means of verifying that
//  the address does indeed point to the desired type of structure.
//
//  Returns 0 if the values match.
//
//  Logs a message and returns -1 if the values do NOT match.
//-----------------------------------------------------------------------------
int ValidatePtr(void *structPtr, void *copyOfPtr,
                const char *typeName, const char *context)
{
  if (structPtr AND ((int)structPtr != -1) AND (structPtr == copyOfPtr) )
    return 0;

  fprintf(DiagLog, "*** %s: Invalid '%s' ptr: [0x%04X] ***\r\n",
          context, typeName, (uint)structPtr );

  return -1;
}



//-----------------------------------------------------------------------------
//  Functions to display strings and/or set display attributes
//  (assumes ANSI/vt100 compatible terminal)
//-----------------------------------------------------------------------------
void DispStr(uchar x, uchar y, char *msg, bool blankToEnd)
{
  printf("\033[%02d;%02dH%s", y+1, x+1, msg);
  if (blankToEnd)  printf("\033[K");
}

//-----------------------------------------------------------------------------
void DispStrAttr(uchar x, uchar y, char *attrs, char *msg, bool blankToEnd)
{
  char  *ptr;

  printf("\033[%02d;%02dH", y+1, x+1);

  ptr = attrs;
  if (*ptr)  {
    printf("\033[%d", *ptr);
    while (*(++ptr))  printf(";%d", *ptr);
    printf("m");
  }

  printf("%s\033[0m", msg);

  if (blankToEnd)  printf("\033[K");
}

//-----------------------------------------------------------------------------


//=============================================================================
//  some standard Linux functions
//=============================================================================

#ifdef RABBIT

//-----------------------------------------------------------------------------
//  return the specified time broken down in the provided tm structure
//-----------------------------------------------------------------------------
struct tm *gmtime_r(const time_t *secs, struct tm *dts)
{
  KLOCK();  *dts = *gmtime(secs);  KUNLOCK();

  return dts;
}

//-----------------------------------------------------------------------------
//  return the specified time broken down in the provided tm structure as a
//  local time
//-----------------------------------------------------------------------------
struct tm *localtime_r(const time_t *utcSecs, struct tm *dts)
{
  time_t  localSecs;

  KLOCK();

  localSecs = *utcSecs + (long)timeZone * 3600;
  *dts = *gmtime(&localSecs);

  KUNLOCK();

  return dts;
}

//-----------------------------------------------------------------------------
//  format a broken-down time value into a string of the following format:
//    "Wed Jun 30 21:49:08 1993\n"
//-----------------------------------------------------------------------------
char *asctime_r(const struct tm *dts, char *str)
{
  KLOCK();
  strcpy(str, asctime(dts));
  KUNLOCK();

  return str;
}

//-----------------------------------------------------------------------------
//  format a calendar time value into a string of the following format:
//    "Wed Jun 30 21:49:08 1993\n"
//-----------------------------------------------------------------------------
char *ctime_r(const time_t *utcSecs, char *str)
{
  struct tm  dts;

  localtime_r(utcSecs, &dts);  asctime_r((const struct tm *)&dts, str);

  return str;
}


#endif

