//----- fmemcpy.c ----- 03/29/12 -----(06/17/10)-----

//-----------------------------------------------------------------------------
//  A version of fmemcpy() that avoids single write operations that span a
//  4K boundary
//-----------------------------------------------------------------------------
//  Version log (most recent first):
//
//
//  2012-03-29 - Mark Davis (*)
//               - The most significant byte of the source and destination
//                 pointer values are now forced to be 0.  Prior to this, I
//                 ran in to at least one case where one of them was not 0,
//                 which caused the code to malfunction because some of the
//                 code forced that byte to 0 after fetching it and some did
//                 not.
//
//  2010-07-09 - Mark Davis (*)
//               - Several bugs found and fixed.  Some more refinements that,
//                 along with the bug fixes, resulted in some significant
//                 speed improvements.
//
//  2010-06-25 - Mark Davis
//               - mostly optimized all-assembly code version
//
//  2010-06-24 - Mark Davis
//               - First working hybrid C/asm version of fmemcpy() that is
//                 compatible with hardware write protection.
//
//  2010-06-17 - Mark Davis
//               - Temporary (and slow) version is compatible with hardware
//                 write protection (because the compiled code calls _farst1()
//                 which has been replaced with a hardware write protection
//                 comptabile version).
//
//-----------------------------------------------------------------------------

#include <sys/types.h>
#include <unistd.h>


#pragma offset_labels on

//-----------------------------------------------------------------------------
//  A version of fmemcpy() that uses the safe version of __farst1 to avoid
//  writing to 4K pages outside the specified range.
//-----------------------------------------------------------------------------
/*
void far * far fmemcpy (void far *to, void const far *from, ulong size)
{
  uchar far  *to_ = (uchar far *)to, *from_ = (uchar far *)from;

  while (size--)  *to_++ = *from_++;

  return to;
}
*/

#pragma warn -758
#pragma warn -759

//-----------------------------------------------------------------------------
//  DEBROS version of fmemcpy() function that is both faster than the Softools
//  version and compatible with Rabbit 3000A memory write protection
//-----------------------------------------------------------------------------
void far * far fmemcpy(void far *to, void const far *from, ulong size)
{
  uchar far *saveTo;
  uint   byteCount, safeToCount;
  ulong  fromEnd, toEnd, safeFromEnd, safeToEnd;

/*=======================================================
   Aprox CPU cycles for each stage:

      1-time regardless of # of bytes:        300
      once for each 4K segment touched:       484
      per byte within each <= 4K block:        22.5

   For the best case (all bytes in 1 4K block):

                             784 cycles + 22.5/byte  (~0.0178 ms @ 44MHz)

   For blocks that don't cross a 4K boundary:

      512 bytes:   ~12,304 cycles  (~0.28  ms @ 44MHz)
     1024 bytes:   ~23,824 cycles  (~0.54  ms @ 44MHz)
     2048 bytes:   ~46,864 cycles  (~1.065 ms @ 44MHz)


   For a typical "full" Ethernet packet (MSS data bytes):

     1460 bytes:   ~33,634 cycles  (~0.764 ms @ 44MHz)

=========================================================*/

;
#asm
;   make sure 24-bit addrs have 0 in the MS byte
        ld      (ix+.to+3),0         ;[10]
        ld      (ix+.from+3),0       ;[10]

;   saveTo = to;
        ld      hl,(ix+.to)          ;[9]
        ld      (ix+.saveTo),hl      ;[11]
        ld      hl,(ix+.to+2)        ;[9]
        ld      (ix+.saveTo+2),hl    ;[11]

;   fromEnd = (ulong)from + size - 1;
        ld      hl,(ix+.from+2)      ;[9]
        push    hl                   ;[10]
        ld      hl,(ix+.from)        ;[9]
        push    hl                   ;[10]
        ld      hl,(ix+.size+2)      ;[9]
        ld      bc,hl                ;[4]
        ld      hl,(ix+.size)        ;[9]
        ex      de,hl                ;[2]
        call    __ladd##             ;[12]+??  BC:DE = BC:DE + pushed value
        call    __ldec##             ;[12]+??  -- BC:DE
        ld      hl,bc                ;[4]
        ld      (ix+.fromEnd+2),hl   ;[11]
        ex      de,hl                ;[2]
        ld      (ix+.fromEnd),hl     ;[11]

;   toEnd = (ulong)to + size - 1;
        ld      hl,(ix+.to+2)        ;[9]
        push    hl                   ;[10]
        ld      hl,(ix+.to)          ;[9]
        push    hl                   ;[10]
        ld      hl,(ix+.size+2)      ;[9]
        ld      bc,hl                ;[4]
        ld      hl,(ix+.size)        ;[9]
        ex      de,hl                ;[2]
        call    __ladd##             ;[12]+??  BC:DE = BC:DE + pushed value
        call    __ldec##             ;[12]+??  -- BC:DE
        ld      hl,bc                ;[4]
        ld      (ix+.toEnd+2),hl     ;[11]
        ex      de,hl                ;[2]
        ld      (ix+.toEnd),hl       ;[11]

;;===== 288+ cycles up to this point =====

;   while (size)  {
  while:
        ld      hl,(ix+.size)        ;[9]
        ex      de,hl                ;[2]
        ld      hl,(ix+.size+2)      ;[9]
        or      hl,de                ;[2]
        jp      z,done               ;[7]

;   //--- # of source bytes that won't cross a 4K boundary ---
;     safeFromEnd = (ulong)from | 0x0FFF;
        ld      hl,(ix+.from)        ;[9]
        ld      de,0x0FFF            ;[6]
        or      hl,de                ;[2]
        ld      (ix+.safeFromEnd),hl ;[11]
        ld      hl,(ix+.from+2)      ;[9]
        ld      (ix+.safeFromEnd+2),hl ;[11]

;     if (safeFromEnd > fromEnd)  safeFromEnd = fromEnd;
        ex      de,hl                ;[2]  DE = MSW(safeFromEnd) from above
        ld      hl,(ix+.fromEnd+2)   ;[9]  HL = MSW(fromEnd)
        ex      de,hl                ;[2]  HL = MSW(safeFromEnd), DE = MSW(fromEnd)
        or      a,a                  ;[2]  clear Carry (for DE == HL case)
        sbc     hl,de                ;[4]  Carry set if HL < DE
        jr      c,fromOK             ;[5]  MSW(safeFromEnd) < MSW(fromEnd) ?
        jr      nz,copyFromEnd       ;[5]  MSW(safeFromEnd) > MSW(fromEnd) ?

        ld      hl,(ix+.safeFromEnd) ;[9]  MSW's equal, so compare LSW's
        ex      de,hl                ;[2]  DE = LSW(safeFromEnd)
        ld      hl,(ix+.fromEnd)     ;[9]  HL = LSW(fromEnd)
        scf                          ;[2]  set Carry (for DE == HL case)
        sbc     hl,de                ;[4]  Carry set if DE > HL
        jr      nc,fromOK            ;[5]  LSW(safeFromEnd) <= LSW(fromEnd)

    copyFromEnd:                     ;     set safeFromEnd = fromEnd
        ld      hl,(ix+.fromEnd+2)   ;[9]
        ld      (ix+.safeFromEnd+2),hl ;[11]
        ld      hl,(ix+.fromEnd)     ;[9]
        ld      (ix+.safeFromEnd),hl ;[11]
    fromOK:

;     byteCount = (uint)safeFromEnd - (uint)from + 1;
        ld      hl,(ix+.safeFromEnd) ;[9]
        ex      de,hl                ;[2]
        ld      hl,(ix+.from)        ;[9]
        ex      de,hl                ;[2]
        or      a,a                  ;[2]
        sbc     hl,de                ;[4]
        inc     hl                   ;[2]
        ld      (ix+.byteCount),hl   ;[11]

;   //--- # of target bytes that won't cross a 4K boundary ---
;     safeToEnd = (ulong)to | 0x0FFF;
        ld      hl,(ix+.to)          ;[9]
        ld      de,0x0FFF            ;[6]
        or      hl,de                ;[2]
        ld      (ix+.safeToEnd),hl   ;[11]
        ld      hl,(ix+.to+2)        ;[9]
        ld      (ix+.safeToEnd+2),hl ;[11]

;     if (safeToEnd > toEnd)  safeToEnd = toEnd;
        ex      de,hl                ;[2]   DE = MSW(safeToEnd) from above
        ld      hl,(ix+.toEnd+2)     ;[9]   HL = MSW(toEnd)
        ex      de,hl                ;[2]   HL = MSW(safeToEnd), DE = MSW(toEnd)
        or      a,a                  ;[2]   clear Carry (for DE == HL case)
        sbc     hl,de                ;[4]   Carry set if HL < DE
        jr      c,toOK               ;[5]   MSW(safeToEnd) < MSW(toEnd) ?
        jr      nz,copyToEnd         ;[5]   MSW(safeToEnd) > MSW(toEnd) ?

        ld      hl,(ix+.safeToEnd)   ;[9]   MSW's equal, so compare LSW's
        ex      de,hl                ;[2]   DE = LSW(safeToEnd)
        ld      hl,(ix+.toEnd)       ;[9]   HL = LSW(toEnd)
        scf                          ;[2]   set Carry (for DE == HL case)
        sbc     hl,de                ;[4]   Carry set if DE > HL
        jr      nc,toOK              ;[5]   LSW(safeToEnd) <= LSW(toEnd)

    copyToEnd:                       ;      set safeToEnd = toEnd
        ld      hl,(ix+.toEnd+2)     ;[9]
        ld      (ix+.safeToEnd+2),hl ;[11]
        ld      hl,(ix+.toEnd)       ;[9]
        ld      (ix+.safeToEnd),hl   ;[11]
    toOK:

;     safeToCount = (uint)safeToEnd - (uint)to + 1;
        ld      hl,(ix+.to)          ;[9]
        ex      de,hl                ;[2]
        ld      hl,(ix+.safeToEnd)   ;[9]
        or      a,a                  ;[2]  clear Carry
        sbc     hl,de                ;[4]
        inc     hl                   ;[2]
        ld      (ix+.safeToCount),hl ;[11]

;   //--- use smaller of the 2 counts ---
;     if (safeToCount < byteCount)  byteCount = safeToCount;
        ex      de,hl                ;[2]   DE = HL (safeToCount from above)
        ld      hl,(ix+.byteCount)   ;[9]   HL = byteCount
        scf                          ;[2]   set Carry
        sbc     hl,de                ;[4]   Carry cleared if DE < HL
        jr      c,countOK            ;[5]   ! (safeToCount < byteCount) ?
        ex      de,hl                ;[2]   HL = saved copy of safeToCount
        ld      (ix+.byteCount),hl   ;[11]  byteCount = safeToCount
    countOK:

;   //--- if at least 2 bytes, use loop to copy 2 bytes at a time ---
;     if (byteCount >> 1)  {
        ld      hl,(ix+.byteCount)   ;[9]
        or      a,a                  ;[2]
        rr      hl                   ;[2]
        jp      z,copy1              ;[7]  HL == 0 is assumed by copy1!!!!


;===== copy up to 4K bytes, 2 bytes at a time =====
;  //--- set up register for loop ---
;  // reg use:  A[3-0]:IX = from,  A'[3-0]:IY = to,  DE = # of words to copy

       ex    de,hl                   ;[2]   DE = # words

       push  ix                      ;[12]

     ;; to address
       ld    a,(ix+.to+2)            ;[9]
       ex    af,af'                  ;[2]   A' = MSB(to)
       ld    hl,(ix+.to)             ;[9]
       ld    iy,hl                   ;[4]   IY = LSW(to)

     ;; from address
       ld    a,(ix+.from+2)          ;[9]   A = MSB(from)
       ld    hl,(ix+.from)           ;[9]
       ld    ix,hl                   ;[4]   IX = LSW(from)

       ld    bc,2                    ;[6]   BC = 2 (# of bytes copied each pass)

;;===== around 484 cycles from start of per-block loop to here =====
;;===== (which means that many cycles per block UP to 4K + 22.5/byte) =====
;;===== plus the

;;===== 45 cycles per word (22.5 cycles per byte) =====
;;------ tight loop that copies up to 4K bytes, 2 bytes each pass ------
  loop:
       ldp   hl,(ix)           ;[10]   HL = *(uint *)from;
       add   ix,bc             ;[4]    from += 2;

       ex    af,af'            ;[2]    switch A to target addr's 4 MSB

       ldp   (iy),hl           ;[12]   *(uint *)to = HL;
       add   iy,bc             ;[4]    to += 2;

       ex    af,af'            ;[2]    switch A back to source addr's 4 MSB

       dec   de                ;[2]    --count
       rl    de                ;[2]    set Z flag if DE is 0
       rr    de                ;[2]
       jr    nz,loop           ;[5]    if (count) continue
;;----------------------------------------------------------------------

       ld    hl,ix                 ;[4]   save source addr

       pop   ix                    ;[9]   restore Stack Frame

       ld    (ix+.from),hl         ;[11]  update LSW(from)
       or    hl,de                 ;[2]   DE 0 from above loop
       jr    nz,saveTarget         ;[5]   did LSW(from) wrap to 0?
       inc   (ix+.from+2)          ;[12]  incr MSW(from)

   saveTarget:
       ld    hl,iy                 ;[4]   update LSW(to)
       ld    (ix+.to),hl           ;[11]
       or    hl,de                 ;[2]   DE still 0 from above
       jr    nz,updSize            ;[5]   did LSW(to) wrap to 0?
       inc   (ix+.to+2)            ;[12]  inc MSW(to)

   updSize:   ;; size -= (byteCount & 0x0FFE);
       ld    hl,(ix+.byteCount)    ;[9]   HL = byteCount
       ex    de,hl                 ;[2]   DE = byteCount
       res   0,e                   ;[4]   DE = byteCount & 0xFFFE

       ld    hl,(ix+.size)         ;[9]   upate LSW(size)
       or    a,a                   ;[2]   clear Carry
       sbc   hl,de                 ;[4]   Carry set if LSW(size) < DE
       ld    (ix+.size),hl         ;[11]
       jp    nc,while              ;[7]
       dec   (ix+.size+2)          ;[12]  update MSW(size)
       jp    while                 ;[7]


;;===== can only copy one byte without crossing a 4K boundary =====
;;   *TO++ = *FROM++;
  copy1:
       ex    de,hl                 ;[2]   DE = 0 (because HL is 0 from the jp)

   ; B = *from;
       ld    a,(ix+.from+2)        ;[9]   A = MSB(from)
       ld    hl,(ix+.from)         ;[9]   HL = LSW(from)
       ld    iy,hl                 ;[4]   IY = LSW(from)
       ldp   hl,(hl)               ;[10]  L = byte to be copied
       ld    b,l                   ;[2]   B = byte to be copied

   ; ++from;
       ld    hl,iy                 ;[4]
       inc   hl                    ;[2]   ++LSW(from)
       ld    (ix+.from),hl         ;[11]
       or    hl,de                 ;[2]   DE 0 from above
       jr    nz,getTo              ;[5]   did LSW(from) wrap to 0?
       inc   (ix+.from+2)          ;[12]  incr MSW(from)

   ; IY = to;
    getTo:
       ld    a,(ix+.to+2)          ;[9]   A = MSB(to)
       ld    hl,(ix+.to)           ;[9]   HL = LSW(to)
       ld    iy,hl                 ;[4]   IY = LSW(to)

   ; ++to;
       inc   hl                    ;[2]   ++LSW(to)
       ld    (ix+.to),hl           ;[11]
       or    hl,de                 ;[2]   DE still 0 from above
       jr    nz,testTo             ;[5]   did LSW(to) wrap to 0?
       inc   (ix+.to+2)            ;[12]  incr MSB(to)

   ; is target addr even or odd? (remember addr already incremented!)
    testTo:
       bit   0,l                   ;[4]   test bit that is already reversed
       jr    nz,evenAddr           ;[5]   nz ==> bit is 1, ==> it WAS 0 b4 ++!

;; The target addr is an odd #, so we decrement it to insure both bytes
;; we are about to write will be in the same 4K block.  Using the decremented
;; address, L = the preeceding byte, H = the target byte
       dec    iy                   ;[4]   target addr now even
       ldp    hl,(iy)              ;[10]  read word from target addr - 1
       ld     h,b                  ;[2]   replace byte to be updated
       jr     writeWord            ;[5]

;; The passed addr is an even #, so the read/write will not span a 4K boundary.
;; L represents the target addr, H represents the one after it
    evenAddr:
       ldp    hl,(iy)              ;[10]  read word from target addr
       ld     l,b                  ;[2]   replace byte to be updated

    writeWord:
       ldp    (iy),hl              ;[12]  write copied and updated bytes


;;   --size;
       ld     hl,(ix+.size)        ;[9]  -- LSW(size)
       dec    hl                   ;[2]
       ld     (ix+.size),hl        ;[11]

       ld     a,h                  ;[4]  is HL now 0xFFFF?
       and    l                    ;[2]
       inc    a                    ;[2]  Z set if HL is 0xFFFF
       jp     nz,while             ;[7]

       ld     hl,(ix+.size+2)      ;[9]  -- MSW(size)
       dec    hl                   ;[2]
       ld     (ix+.size+2),hl      ;[11]
       jp     while

  done:

#endasm

  return saveTo;
}

#pragma warn +758
#pragma warn +759

//-----------------------------------------------------------------------------

#pragma offset_labels off

