//----- xalloc.c ----- 11/26/12 -----(06/29/10)-----

//-----------------------------------------------------------------------------
//  Replacement for Softools xalloc module.  This version logs the result so
//  we know what each portion of memory is being used for.  It also adds
//  functions to allocate memory from the portion of memory that can be
//  battery-powered so it retains its value even across short-term power
//  outage or shutdowns, and provides a means to provide a description of
//  what memory was allocated for.
//-----------------------------------------------------------------------------
//  Update log (most recent first)
//
//
//  2012-11-26 - Mark Davis (*)
//               - Uncommented the check for overlap of the RUN_INFO block.
//                 If this block is used for something else, it could interfere
//                 with the ability to do a clean shutdown (or recover after
//                 an unexpected reset).  (NOTE: I believe that this is NOT the
//                 case, as it is used at startup before any other use, and set
//                 at shutdown after any other use.)
//
//  2010-07-07 - Mark Davis
//               - xalloc() now updates the mask for write protection of
//                 extended memory blocks
//
//  2010-06-29 - Mark Davis
//               - First version of DEBROS-comptaible replacement for
//                 Softools xalloc.c module, with extensions.
//
//-----------------------------------------------------------------------------

#include <string.h>
#include <Rabbit.h>

#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/serial.h>  //tdebug

//--- Defined and initialized in the CStart.asm ---
extern ulong  _xmemStart;   //  Start of available space
extern ulong  _xmemSize;    //  Size of available space
extern ulong  _xmemHeap;    //  Address of available space
extern ulong  _xmemEnd;     //  End of buffer

//-----------------------------------------------------------------------------
void logAllocation(ulong start, ulong size)
{
  ulong  end;

  end = start + size - 1;

  fprintf(DiagLog, "  Allocated %6ld bytes (%05lX-%05lX)  Free: %05lX-%05lX\r\n",
          size, start, end, _xmemHeap, _xmemEnd);

  flushConsole();  //tdebug

  if (! ((start >= 0x80000UL) OR (end < 0x7FF80UL)))  //tdebug
    fprintf(DiagLog, "  *** Intersects RUN_INFO area ***\r\n");  //tdebug
}

//-----------------------------------------------------------------------------
//  Allocate a portion of unused extended memory.
//
//  Returns a pointer to the allocated block of memory or NULL if there is
//  insufficient memory.
//-----------------------------------------------------------------------------
char far *xalloc(ulong size)
{
//  uchar  seg, firstSeg, lastSeg;
//  uint  mask;
  ulong  start;  //, end;


  KLOCK();

  start = _xmemHeap;  _xmemHeap += size;

  if (_xmemEnd < _xmemHeap)  {  _xmemHeap = start;  KUNLOCK();  return 0; }

  _xmemSize -= size;

  KUNLOCK();

  logAllocation(start, size);

/*
//--- update write protection mask for 64K extended memory blocks ---
  end = start + size - 1;
  firstSeg = (uchar) MSWL(start);  lastSeg = (uchar) MSWL(end);

  mask = 1 << firstSeg;
  for (seg=firstSeg; seg<=lastSeg; ++seg)  { wpSegMask &= ~mask;  mask <<= 1; }

  fprintf(DiagLog, "  --- new mask: 0x%04X ---\r\n", wpSegMask);  //tdebug

  _EnterCritical_();
  outi(MMIDR, 0x80);  // enable access to registers > 0x00FF
  outi(WPHR, MSBI(wpSegMask));
  outi(WPLR, LSBI(wpSegMask));
  outi(MMIDR, 0x00);  // disable access to registers > 0x00FF
  _ExitCritical_();
*/

  return (char far *)start;
}


//-----------------------------------------------------------------------------
//  Return the current amount of unallocated memory
//-----------------------------------------------------------------------------
ulong xavail(void)
{
  ulong  freeCount;

  _EnterCritical_();  freeCount = _xmemEnd - _xmemHeap;  _ExitCritical_();

  return freeCount;
}

//-----------------------------------------------------------------------------
//  Allocate the specified # of bytes from the end of xmem and update the
//  values used by xalloc() to reflect the allocated bytes.
//
//  On RCM3200's, only the upper (slower) 256K is battery backed.  On BL2105's
//  all RAM is battery backed.
//
//  Memory allocated via bb_xalloc() cannot be freed up.  The intent is to
//  allocate blocks during initialization that would normally be handled with
//  hardcoded declarations.  bb_xalloc() simply provides a way to allocate
//  memory from the end of available extended memory.  As long as the total
//  amount allocated via bb_xalloc() is less than the amount of memory that is
//  battery-backed (256 on the RCM3200), then the allocated memory will retain
//  its values across a reboot or a reset.
//-----------------------------------------------------------------------------
char far *bb_xalloc(ulong size)
{
  ulong start;


  KLOCK();

//--- not enough free xmem ---
  if ((_xmemEnd - _xmemHeap) < size)  {  KUNLOCK();  return 0; }

//--- move end of free space down and adjust size of free space ---
  _xmemSize -= size;  _xmemEnd -= size;

  bbAllocatedCount += size;  // keep track of ttl # BB bytes allocated

  start = _xmemEnd;  // since we allocated from the end rather than the beginning

  KUNLOCK();

  logAllocation(start, size);

  return (char far *) start;
}

//-----------------------------------------------------------------------------

