//----- _shared_.c ----- 09/07/12 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//  Module to house versions of the standard command-line commands and ones
//  that are indepent of the application type.
//-----------------------------------------------------------------------------
//  UPDATE LOG (most recent first)
//
//
//  2012-09-07 - Mark Davis
//               - Removed leading 0's in the output of fmtHMSUsed()
//
//  2010-10-21 - Mark Davis
//               - Fixed problem that caused calibrations for the wrong signal
//                 to be shown if its name rather than its number was used in
//                 the calibration command.
//
//  2010-05-11 - Mark Davis
//               - Added echo argument to copyData() for use by tip command and
//                 half-duplex devices
//
//  2010-02-15 - Mark Davis
//               - Moved readBufCh() from this module to the "common" module.
//
//  2010-02-09 - Mark Davis
//               - calibrate() function now always show the resulting gain,
//                 offset, and range of any changes made before asking if
//                 you want to save the updated values.
//
//  2009-10-27 - Mark Davis
//               - The range show for calib commands is now sorted (with the
//                 numerically smaller value shown for the minimum) so the
//                 min and max values are intuatively correct independent of
//                 the sign on the gain.  NOTE that external functions are
//                 are called to set global values that hold the min/max
//                 values so application code does not have to concern itself
//                 with this issue.
//
//  2009-10-06 - Mark Davis
//               - Updated call to saveParams() to include new parameter that
//                 tells it to save to flash (which was the only option before)
//
//  2009-05-13 - Mark Davis
//               - Added option to calibrate() to dump values out in format of
//                 commands that can be run as a script.
//
//  2009-05-11 - Mark Davis
//               - calibrate() now allows you to specify signals via their
//                 index rather than just their name.
//
//  2009-03-13 - Mark Davis
//               - Changed fmtHexLine so it prefaces '%' chars with a '\'
//                 (used by od, rdflash, rdmem, rduserblk).  Before this fix
//                 the output of these commands would be corrupted when they
//                 encountered a byte value of '%'.
//
//  2009-02-04 - Mark Davis
//               - The current value of digital input bits with a forced value
//                 are shown for ios cmd when using digital or input mode even
//                 if the bit doesn't represent a real hardware bit.
//
//  2008-12-03 - Mark Davis
//               - Added ability to set the testpoints used to calibrate
//                 analog outputs
//
//  2008-11-13 - Mark Davis
//               - Added ability to set calibration data by specifying the
//                 range of measurable values
//
//  2008-05-21 - Mark Davis
//               - split in to 2 modules to address linkage problems
//
//  2008-03-29 - Mark Davis
//               - added -t option to update command to download new boot image
//                 from /srfs/test directory
//
//  2008-02-11 - Mark Davis
//               - added "zero" option to xxxcalib commands to set offset to
//                 the latest reading
//
//  2008-01-31 - Mark Davis
//               - added "adc" cmd to allow forcing a value for an ADC reading
//
//  2007-07-02 - Mark Davis
//               - Added calibrate() func, used by cmds for viewing/setting
//                 device-specific analog calibrations
//
//  2006-03-14 - Mark Davis
//               - original version started
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/backup.h>

#include <sys/hdwDefs.h>

#include <sys/srfs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

extern ulong _xmemSize;


#include <tsh.h>

#include <sys/modbus.h>

#include <sys/backup.h>

#include <applic.h>

#include <sys/farprintf.h>


extern void _task configd(int argc, char **argv);
extern void _task inetd(int argc, char **argv);
//extern void _task timeClient(int argc, char **argv);


//--- a "forgotten" prototype (from sys/stat.h) ---
int stat(const char *file_name, struct stat *buf);

//#include <user_block.h>
//int   readUserBlock(void *dest, unsigned addr, unsigned numbytes);


//=============================================================================
const CalibInfo far adcCalibInfo = {
  adcGetRaw, adcMinRaw, adcMaxRaw,
  adcDefaults, adcCalibPoints,
  adcUnits, ainName, logicalOrderAIN,
  newADCCalib,
  NULL,
  &hdwCfgFile,
  adcGain, adcOffset,
  &NumHdwADC,
  "ADC"
};

const CalibInfo far dacCalibInfo = {
  dacGetRaw, dacMinRaw, dacMaxRaw,
  dacDefaults, dacCalibPoints,
  dacUnits, aoutName, logicalOrderAOUT,
  newDACCalib,
  setStateDAC,
  &hdwCfgFile,
  dacGain, dacOffset,
  &NumHdwDAC,
  "DAC"
};

const CalibInfo far readCalibInfo = {
  readGetRaw, readMinRaw, readMaxRaw,
  readDefaults, readCalibPoints,
  readUnits, ainName, logicalOrderAIN,
  newADCCalib,  // newCalib(ui)
  NULL,
  &devDataFile,
  readGain, readOffset,
  &NumApplAIN,
  "Device Readings"
};

const CalibInfo far setCalibInfo = {
  setGetRaw, setMinRaw, setMaxRaw,
  setDefaults, setCalibPoints,
  setUnits, aoutName, logicalOrderAOUT,
  newDACCalib,
  NULL,
  &devDataFile,
  setGain, setOffset,
  &NumApplAOUT,
  "Device Settings"
};



//-----------------------------------------------------------------------------
//  show or set the level of diagnostic output for processes
//-----------------------------------------------------------------------------
//  show description of the specified command
//-----------------------------------------------------------------------------
void showDesc(char *cmdStr)
{
  uint  u;
  const CmdInfo far *cmdInfo;


  for (u=0; u<NumCommonCmds; ++u)  {
    cmdInfo = commonCmds + u;
    if (fstrcmp((char far *)cmdStr, cmdInfo->cmdName) == 0)  {
      printf("\r\n%ls\r\n", cmdInfo->descr);  return;  }
  }

  for (u=0; u<NumApplCmds; ++u)  {
    cmdInfo = applCmds + u;
    if (fstrcmp((char far *)cmdStr, cmdInfo->cmdName) == 0)  {
      printf("\r\n%ls\r\n", cmdInfo->descr);  return;  }
  }

}

//-----------------------------------------------------------------------------
//  returns true if any of the arguments are one of the typical formats for
//  asking for help on command syntax, or the number of arguments is less than
//  the minimum needed by the command.
//-----------------------------------------------------------------------------
bool showHelp(int minArgs, int argc, char **argv)
{
  int  i, j;
  static char  *helpStrs[] = { "?", "-?", "/?",
                                "help", "-help", "/help", NULL };

  if (argc < minArgs)  { showDesc(argv[0]);  return true; }

  for (i=0; i<argc; ++i)
    for (j=0; helpStrs[j]; ++j)
      if (StriEq(argv[i], helpStrs[j]))  { showDesc(argv[0]);  return true; }

  return false;
}


//-----------------------------------------------------------------------------
//  return info about specified file in a string of the form:
//    <size> YYYY-MM-DD HH:MM
//-----------------------------------------------------------------------------
int fileInfoStr(char *fileName, char *descr)
{
  time_t  fileTime;
  struct tm  tms;
  struct stat  fsInfo;


  strcpy(descr, "       0 unknown");

  if (stat(fileName, &fsInfo))  return -1;

  fileTime = fsInfo.st_mtime;  fileTime += (long)(timeZone * 3600);
  gmtime_r(&fileTime, &tms);

  sprintf(descr, "%8lu %04d-%02d-%02d %02d:%02d",
          (ulong)fsInfo.st_size,
          tms.tm_year+1900, tms.tm_mon+1, tms.tm_mday,
          tms.tm_hour, tms.tm_min);

  return 0;
}


//-----------------------------------------------------------------------------
//  convert # of ms to a string of the format: ####ms, ####s, ####m, or ####h
//-----------------------------------------------------------------------------
char *fmtIdleTime(char *str, ulong idle)
{
  if (idle < 1000)
    sprintf(str, "%4ums  ", (uint)idle);
  else if (idle < 60000L)
    sprintf(str, "%4us   ", (uint)(idle / 1000));
  else if (idle < 3600000L)
    sprintf(str, "%4um   ", (uint)(idle / 60000L));
  else
    sprintf(str, "%4uh   ", (uint)(idle / 3600000L));

  return str;
}

//-----------------------------------------------------------------------------
//  convert # of seconds to a string of the format HHHH:MM:SS
//-----------------------------------------------------------------------------
char *fmtHMSUsed(char *str, ulong secs)
{
  uint  hours, mins;

  hours = (uint)(secs / 3600);
  mins = (uint)((secs % 3600) / 60);
  secs %= 60;

  sprintf(str, "%4u:%02u:%02u ", hours, mins, (uint)secs);

  return str;
}

//-----------------------------------------------------------------------------
//  Write a hex-dump style representation of an array of bytes in to the
//  provided string using the following format:
//
//  hhhhhh: xx xx xx xx xx xx xx xx xx xx xx xx xx xx xx xx   AAAAAAAAAAAAAAAA
//
//  where hhhhhhhh   is the value of offset in hexidecimal format
//        xx xx ..   are the bytes in buf, each shown as a 2-digit hex value
//        AAAAA...   are the chars in buf shown in ASCII (with a '.' shown
//                   in places of any non-printable chars)
//
//  The result is in "printf-compatible" format, meaning that '%' characters
//  are changed in to "%%" for the ASCII part.
//-----------------------------------------------------------------------------
char *fmtHexLine(char *str, ulong offset, char *buf, uchar count)
{
  char  *ptr;
  uchar  i;
  char  ch;
  uint  remain;


  ptr = str;  sprintf(ptr, "%06lX: ", offset);  ptr += 8;

  for (i=0; i<count; ++i)  { sprintf(ptr, "%02X ", buf[i]);  ptr += 3; }
  remain = (16 - (uint)i) * 3;
  while (remain)  { *ptr++ = ' ';  --remain; }

  strcpy(ptr, "   ");  ptr += 3;

  for (i=0; i<count; ++i)  {
    ch = buf[i];
    if ((ch < ' ') OR (ch > '~'))  ch = '.';
    if (ch == '\%')  *ptr++ = '\%';
    *ptr++ = ch;
  }
  *ptr = 0;

  return str;
}

//-----------------------------------------------------------------------------
//  Copy any unread data from one fd to another.  Does not block if there is
//  nothing to read for fromfd, but does block until any data read has been
//  written to tofd.
//
//  returns # of bytes copied on success ( >= 0), < 0 if an error
//-----------------------------------------------------------------------------
int  copyData(fd_t fromfd, int fromFlags, fd_t tofd, int toFlags,
              bool checkForETX, bool echo)
{
  int  i, readCount, writeCount;
  char  *ptr;
  char  buf[128];


  fcntl(fromfd, F_SETFL, fromFlags | O_NONBLOCK);  // don't block on read
  readCount = read(fromfd, buf, sizeof(buf));
  fcntl(fromfd, F_SETFL, fromFlags);  // restore flags

  if (readCount < 0)  {
    if (errno == EAGAIN)  return 0;
    printf("read error: %d\r\n", readCount);  return -1;
  }

  if (readCount == 0)  { printf("EOF\r\n");  return -2; }

  if (checkForETX)  if (*buf == 0x03)  { printf("KILL\r\n");  return -3; }

  for (i=0; i<readCount; ++i)  buf[i] = (int)buf[i] & 0x7f;

  if (echo)  for (ptr=buf,i=0; i<readCount; ++i,++ptr)  {
    write(1, ptr, 1);
    if (*ptr == 0x0D)  write(1, "\012", 1);
  }

  fcntl(tofd, F_SETFL, toFlags & (~O_NONBLOCK));  // block on write
  writeCount = write(tofd, buf, readCount);
  fcntl(tofd, F_SETFL, toFlags);  // restore flags

  if (writeCount < readCount)  { printf("write error\r\n");  return -4; }

  return writeCount;
}

//----------------------------------------------------------------------------
#define  EGUMin(ui)   ( (ci->minRaw(ui) - ci->offsets[ui]) * ci->gains[ui] )
#define  EGUMax(ui)   ( (ci->maxRaw(ui) - ci->offsets[ui]) * ci->gains[ui] )

const char far CalibPrompt[] = {
    "\r\nApply setting for '%.24s' test point #%u"
    "\r\nEnter known/measured value in %.16s: (default %s): " };


//-----------------------------------------------------------------------------
//  function to show the Gain, Offset, and Range for an analog signal
//-----------------------------------------------------------------------------
void showGainOffsetRange(uint ui, CalibInfo far *ci,
                         char *buf, bool showHeader)
{
  float  eguMin, eguMax, hold;


  if (showHeader)  printf("\r\n"
"       ----Gain----  ---Offset---  ------- Range -------  ------Name------\r\n");
//####:  -##.####E-##  -##.####E-##  -##.##E-## -##.##E-##  ################

  eguMin = EGUMin(ui);  eguMax = EGUMax(ui);
  if (eguMin > eguMax)  { hold = eguMin;  eguMin = eguMax;  eguMax = hold; }

  printf("%4u:  %s  %s  %s %s  %.24s\r\n", ui,
          Engineering(ci->gains[ui], 6, buf),
          Engineering(ci->offsets[ui], 6, buf+16),
          Engineering(eguMin, 4, buf+32),
          Engineering(eguMax, 4, buf+48),
          ci->getName(ui, buf+64)
        );
}


//-----------------------------------------------------------------------------
//  Prompt user for actual (externally measured) engineering units for 2 test
//  points for a signal and update the gain and offset based on the 2 points.
//-----------------------------------------------------------------------------
bool calibValue(CalibInfo far *ci, uint ui, char *temp, float egu1, float egu2)
{
  float  raw1, raw2, calibLow, calibHigh;
  static float  lastCalibLow, lastCalibHigh, lastegu1, lastegu2;


  if (egu1 == egu2)  {
    ci->calibPoints(ui, &calibLow, &calibHigh);
    egu1 = (calibLow == lastCalibLow) ? lastegu1 : calibLow;
    egu2 = (calibHigh == lastCalibHigh) ? lastegu2 : calibHigh;
  }


#define  ForceValue    false
#define  ClearForced   true

//--- test point 1 ---
  if (ci->setOutput)  ci->setOutput(ui, egu1, ForceValue);

  printf(CalibPrompt, ci->getName(ui, temp), 1,
         ci->units(ui, temp+32), Engineering(egu1, 4, temp+64) );
  GETS(temp);  raw1 = ci->getRaw(ui);
  if (*temp)  egu1 = atof(temp);


//--- test point 2 ---
  if (ci->setOutput)  ci->setOutput(ui, egu2, ForceValue);

  printf(CalibPrompt,  ci->getName(ui, temp), 2,
         ci->units(ui, temp+32), Engineering(egu2, 4, temp+64) );
  GETS(temp);  raw2 = ci->getRaw(ui);
  if (*temp)  egu2 = atof(temp);


  if (ci->setOutput)  ci->setOutput(ui, 0.0, ClearForced);

//printf("raw: %s:%s   EGU: %s:%s\r\n",  //tdebug
//       Engineering(raw1, 6, temp), Engineering(raw2, 6, temp+16),
//       Engineering(egu1, 6, temp+32), Engineering(egu2, 6, temp+48) );

  if ((raw1 == raw2) OR (egu2 == egu1))  {
    printf("\r\nUnusable calibration data\r\n");  return false; }


  ci->gains[ui] = (egu2 - egu1) / (raw2 - raw1);
  ci->offsets[ui] = raw1 - egu1/ci->gains[ui];

//  showGainOffsetRange(ui, ci, temp, true);  // now done by caller

  lastegu1 = egu1;  lastCalibLow = calibLow;
  lastegu2 = egu2;  lastCalibHigh = calibHigh;

  return true;
}


//-----------------------------------------------------------------------------
//  function to manage calibration information
//-----------------------------------------------------------------------------
void calibrate(uint argc, char **argv, CalibInfo far *ci)
{
  char  ch;
  uint  ui, ut, hdwIdx, numValues, interval, numSamples;
  bool  allValues, foundMatch, chgsMade, useDefaults, setZeroOffset, setRange,
        updatedCalib;
  float  gain, offset, minValue, maxValue, minRaw, maxRaw, quietMode,
         egu1, egu2;
  char  temp[128];



  quietMode = StriEq(argv[argc-1], "-quiet");
  if (quietMode)  --argc;

  if (showHelp(2, argc, argv))  {
    printf("\nusage: %s ...\r\n", argv[0]);
    printf("     all|<name>|#\r\n"
           "     all|<name>|# defaults\r\n"
           "     all|<name>|# range <min> <max>\r\n"
           "     all|<name>|# <gain> <offset>\r\n");
    if (ci->setOutput)  printf(
           "     all|<name>|# setpoints <point1> <point2>\r\n");
    printf("     all|<name>|# zero [interval [#samples]]\r\n"
           "     dump\r\n"
           "     show\r\n"
           "     reload\r\n"
           "     save\r\n"
           "  (-quiet at end will skip prompts)\r\n"
          );
    Exit(-1);
  }


  numValues = *(ci->numValues);   egu1 = egu2 = 0.0;

//--- reload the data from a file ---
  if (argc == 2)  if (StriEq(argv[1], "reload"))  {
    loadParams(temp, sizeof(temp), ci->paramFile, false);   Exit(0);  }


//--- save the data to a file ---
  if (argc == 2)  if (StriEq(argv[1], "save"))  {
    saveParams(temp, sizeof(temp), ci->paramFile, ForceFlush);  Exit(0);  }


//--- dump in format that can be run as commands ---
  if (argc == 2)  if (StriEq(argv[1], "dump"))  {
    for (ui=0; ui<numValues; ++ui)  {
      printf("%s %2d %s %s -quiet\r\n", argv[0], ui,
              Engineering(ci->gains[ui], 7, temp),
              Engineering(ci->offsets[ui], 7, temp+16)
            );
    }
    printf("\r\n");  Exit(0);
  }

//--- show the data ---
  if (argc == 2)  if (StriEq(argv[1], "show"))  {
    printf("\r\n%s calibration data:\r\n", ci->setDescr);
    for (ui=0; ui<128; ++ui)  {
      if ((hdwIdx = (ci->order)(ui)) >= numValues)  continue;
      showGainOffsetRange(hdwIdx, ci, temp, !ui);
    }
    printf("\r\n");  Exit(0);
  }


  if (argc < 2)  Exit(-1);


//--- set calibration values for one or more of the I/O signals ---
  useDefaults = setZeroOffset = setRange = false;
  gain = 0.0;
  allValues = (strcmp(argv[1], "all") == 0);

  if (argc > 2)  if (StriEq(argv[2], "range"))  {
    if (argc < 5)  { printf("\r\nInvalid range command\r\n");  Exit(-1); }
    minValue = atof(argv[3]);  maxValue = atof(argv[4]);
    if (minValue == maxValue)  { printf("\r\nInvalid range\r\n");  Exit(-1); }
    setRange = true;
  }

  if (argc > 2)  if (StriEq(argv[2], "setpoints"))  {
    if (argc < 5)  { printf("\r\nInvalid setpoints command\r\n");  Exit(-1); }
    egu1 = atof(argv[3]);  egu2 = atof(argv[4]);
    if (egu1 == egu2)  { printf("\r\nInvalid setpoints\r\n");  Exit(-1); }
  }

  if (argc == 4)  {
    gain = atof(argv[2]);  offset = atof(argv[3]);
    if (gain == 0.0)  {
      printf("\r\nUnusable gain: %s\r\n", argv[2]);  Exit(-1); }
  }

  if (argc > 2)  {
    if (StriEq(argv[2], "defaults"))  useDefaults = true;

    if (StriEq(argv[2], "zero"))  {
      setZeroOffset = true;  interval = 250;  numSamples = 10;
      if (argc > 3)  interval = atoi(argv[3]);
      if (interval < 10)  interval = 10;
      if (argc > 4)  numSamples = atoi(argv[4]);
      if (numSamples < 1)  numSamples = 1;
    }
  }


  foundMatch = chgsMade = false;

//--- perform specified action for matching signal names (or all) ---
  for (ui=0; ui<128; ++ui)  {

    updatedCalib = false;

    if (isdigit(argv[1][0]))  {
      hdwIdx = atoi(argv[1]);  if (hdwIdx != ui)  continue;
      ci->getName(hdwIdx, temp);
    }
    else  {
      hdwIdx = (ci->order)(ui);  ci->getName(hdwIdx, temp);
      if (!allValues AND !strstri(temp, argv[1]))  continue;
    }

    if (hdwIdx >= numValues)  continue;

    foundMatch = true;

    if (!quietMode)  {
      printf("  %s (Y/N/Q): ", temp);  GETS(temp);  ch = toupper(*temp);
      if (ch == 'Q')  Exit(-1);
      if (ch != 'Y')  continue;
    }

    if (useDefaults)
      ci->defaults(hdwIdx, &gain, &offset);

    if (setZeroOffset)  {
      gain = ci->gains[hdwIdx];  offset = 0.0;
      for (ut=0; ut<numSamples; ++ut)  {
        offset += ci->getRaw(hdwIdx);
        printf("    %s \r", Engineering(offset/(float)(ut+1), 6, temp));
        fflush(stdout);  sleepMS(interval);
      }
      offset /= (float) numSamples;  printf("\n");
    }

    if (setRange)  {
      minRaw = ci->minRaw(hdwIdx);  maxRaw = ci->maxRaw(hdwIdx);
      offset = (maxRaw * minValue - maxValue * minRaw) / (minValue - maxValue);
      gain = maxValue / (maxRaw - offset);
    }


    if (gain == 0.0)
      updatedCalib = calibValue(ci, hdwIdx, temp, egu1, egu2);
    else  {
      ci->gains[hdwIdx] = gain;  ci->offsets[hdwIdx] = offset;
      updatedCalib = true;
    }

  //--- If calibrations where changed, then we need to update any values ---
  //--- that are dependent on them, and to force the recompute of the ---
  //--- reading or the hdw output value ---
    if (updatedCalib)  {
      if (ci->newCalib)  ci->newCalib(hdwIdx);
      chgsMade = true;
    }

    if (updatedCalib AND !quietMode)
      showGainOffsetRange(hdwIdx, ci, temp, true);
  }


  if (! foundMatch)  { printf("No such value: %s\r\n", argv[1]);  Exit(-1); }

  if (chgsMade AND !quietMode)  {
    printf("\r\nSave updated calibrations (Y/N): ");  GETS(temp);
    if (toupper(*temp) == 'Y')
      saveParams(temp, sizeof(temp), ci->paramFile, ForceFlush);
  }

  Exit(0);
}


//-----------------------------------------------------------------------------
