//----- update.c ----- 11/26/13 -----(03/14/06)-----

//-----------------------------------------------------------------------------
//
//  2013-11-26 - Mark Davis
//               - Modified 'usage' options to be a bit clearer
//
//  2011-04-29 - Mark Davis
//               - Any bootimg.new file left over from an aborted/failed
//                 previous update attempt is now deleted before checking the
//                 amount of free file space.
//
//  2010-12-21 - Mark Davis
//               - Addes -rs option for doing a Soft Reset rather than a CPU
//                 Reset to reboot the OS.  This attempts to maintain the
//                 majority of the application state across the reboot.
//
//  2010-05-16 - Mark Davis (*)
//               - Added scheduler lock around call to Writeflash()
//
//  2010-04-09 - Mark Davis
//               - If a file name is given and it starts with a /, then it is
//                 considered to be an absolute path and the default use of
//                 /srfs/release or /srfs/test is ignored.
//
//  2010-01-01 - Mark Davis (*)
//               - Added scheduler lock around calls to write to the flash
//                 when updating the DLM (DownLoad Manager).
//
//  2009-12-25 - Mark Davis
//               - Added ability to update the DLM (DownLoad Manager)
//
//  2009-04-13 - Mark Davis
//               - "update" command now looks in /srfs/release (rather than
//                  /srfs/) for bootimg file if the -t command line option is
//                  not specified.  It continues to look in /srfs/test if the
//                  -t option is specified.
//
//  2009-xx-xx - Mark Davis
//               - original version of ...
//
//-----------------------------------------------------------------------------

#include <time.h>
#include <common.h>

#include <sys/types.h>
#include <unistd.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#include <utime.h>


#include <busybox.h>

#include <sys/hdwDefs.h>

#include <sys/srfs.h>

#include <dcdefs.h>

#include <ctype.h>
#include <limits.h>

#include <sys/SHD_headers.h>

#include <sys/modbus.h>

#include <tsh.h>

#include <applic.h>

#include <sys/farprintf.h>


//-----------------------------------------------------------------------------
//  insure that given file name ends with the specified suffix
//-----------------------------------------------------------------------------
char *setSuffix(char *name, char *suffix)
{
  char  *ptr;


  if (! name)  return name;

  if (! (ptr = strrchr(name, '.')))  ptr = name + strlen(name);

  strcpy(ptr, suffix);

  return name;
}


//-----------------------------------------------------------------------------
//  update boot image file
//-----------------------------------------------------------------------------
void _task _update(int argc, char **argv)
{
  int  i, count, stat;
  bool  autoReboot, testVers, updateDLM;
  fd_t  fromfd, tofd;
  ulong  ttl, newSize, fromTime;
  struct stat  fsInfo;
  struct utimbuf  fileTimes;
  char  *ptr, *primaryName, *backupName, *temp, *baseName;
  char  fromName[NAME_MAX+12], fileName[NAME_MAX+12];
  char  buf[400];  // must be at least 256 bytes!!!


//finish:  need to use file locks to insure only one process trying to change
//         the boot image files (which means I first have to add support
//         for file locking to fcntl.c)


  if (showHelp(1, argc, argv))  {
    printf("\nusage:  update [-t] [-r[s]] [bootImgFile.bin]\r\n"
             "               [-dlm]\r\n");
    Exit(-1);
  }


//--- parse arguments ---
  autoReboot = testVers = updateDLM = false;  *fileName = 0;
  for (i=1; i<argc; ++i)  {
    if (StrEq(argv[i], "-t"))  testVers = true;

    if (StrEq(argv[i], "-r"))  autoReboot = true;

    if (StrEq(argv[i], "-rs"))  autoReboot = useSoftReset = true;

    if (StrEq(argv[i], "-dlm"))  {
      strcpy(fileName, "DLM-III-D.bin");  updateDLM = true; }

    if (argv[i][0] != '-')  {
      strncpy(fileName, argv[i], sizeof(fileName));
      fileName[sizeof(fileName)-1] = 0;
    }
  }


//--- use the buffer temporarily for file names ---
  ptr = buf;
  primaryName = ptr;  ptr += NAME_MAX+12;
  backupName = ptr;  ptr += NAME_MAX+12;
  temp = ptr;  ptr += 64;
  if (ptr-buf > sizeof(buf))  {
    printf("Internal buffer too small\r\n");  Exit(-1); }

  memset(buf, 0, sizeof(buf));


  if (!updateDLM)  {
  //finish: need generic means of dealing with primary/backup file names
    bootFileNames(primaryName, backupName, true);  printf("\r\n");

  //===== make whichever image we are running the primary =====
  //--- if we are running the primary image, then delete any backup one ---
    if (runningPrimary AND *primaryName)  {
      if (*backupName)
        if (unlink(backupName))  { perror(backupName);  Exit(-1); }
    }
  //--- otherwise, make the backup image the primary one ---
    else  if (*backupName)  {
      if (*primaryName)
        if (unlink(primaryName))  { perror(primaryName);  Exit(-1); }
      ffile_makePrimary(backupName);
      strcpy(primaryName, backupName);
      setSuffix(primaryName, ".bin");
      rename(backupName, primaryName);  *backupName = 0;
      runningPrimary = true;
      bootFileNames(primaryName, backupName, false);
    }

  //--- use name of current primary bootimg if no name given on cmd line ---
    if (!*fileName)
      if (*primaryName)
        strcpy(fileName, primaryName);
      else  {
        printf("No name given and no existing file to replace!\r\n");  Exit(-1); }
  }

  *fromName = 0;
  if (*fileName != '/')  {
    strcpy(fromName, "/srfs/");
    strcat(fromName, testVers ? "test/" : "release/");
  }
  strcat(fromName, fileName);
  baseName = strrchr(fileName, '/');
  if (baseName)  ++baseName;  else  baseName = fileName;


//--- open remote file to be downloaded and make sure we have space for it ---
  fileInfoStr(fromName, temp);
//--- get size of new file ---
  if ((fromfd = open(fromName, O_RDONLY, 0)) < 0)  {
    perror(fromName);  Exit(-4);  }
  if (fstat(fromfd, &fsInfo) < 0)  { perror(fromName);  Exit(-2); }
  newSize = (ulong)fsInfo.st_size;
  fromTime = fsInfo.st_mtime;

  printf("Download: %s %s\r\n", temp, fromName);

//--- make sure we have enough space to download it ---
//finish --- need generic API method to determine free space ---
  if (updateDLM)  {
    if ((newSize < 0xF000) OR (newSize > 65370))  {
      printf("Invalid DLM file size\r\n");  Exit(-3); }
  }
  else  {
    unlink("bootimg.new");  // delete any file left from prev aborted attempt
    if (newSize > ffile_free())  {
      printf("Insufficient space to copy %s\r\n", fromName);  Exit(-3); }
  }


//--- if updating the DLM, copy tofile directly to flash ---
  if (updateDLM)  {
    ttl = 0;
    while ((count = read(fromfd, buf, 256)) > 0)  {
      _FlashUnlock = ~ttl;
      KLOCK();
      stat = WriteFlash((FlByte far *)ttl, buf, (ulong)count);
      KUNLOCK();
      if (stat)  {
        printf("Error [%d] writing to flash at addr: 0x%04lX\r\n", stat, ttl);
        Exit(-4);
      }
      ttl += count;  printf("%10lu  \r", ttl);  fflush(stdout);  CheckETX();
      if (ttl >= 0xF000L)  break; // don't overwrite FAT and other info
    }
    printf("\r\n");  Exit(0);
  }


//--- copy new file to local flash --
  ttl = 0;
  if ((tofd = open("bootimg.new", O_RDWR | O_CREAT, 0774)) < 0)  {
    perror("bootimg.new");  Exit(-2); }
  while ((count = read(fromfd, buf, sizeof(buf))) > 0)  {
    if (write(tofd, buf, count) != count)  { perror("write");  break; }
    ttl += count;  printf("%10lu  \r", ttl);  fflush(stdout);  CheckETX();
  }
  printf("\r\n");

  close(fromfd);  close(tofd);

  if (ttl != newSize)  { printf("Download failed\r\n");  Exit(-3); }

//--- download succeeded - chg cur primary to backup ---
  ffile_getPrimaryName(primaryName, PATH_MAX);
  strcpy(temp, primaryName);  setSuffix(temp, ".bak");
  rename(primaryName, temp);
  if (ffile_makeBackup(temp))  perror(temp);
//--- and make the new file the new primary ---
  rename("bootimg.new", baseName);  ffile_makePrimary(baseName);
  runningPrimary = false;

//--- set date/time on new file to match remote one we just downloaded ---
  fileTimes.actime = fileTimes.modtime = fromTime;
  utime(baseName, &fileTimes);


//--- show end result ---
  bootFileNames(primaryName, backupName, true);


//--- if reboot option was specified, do a shutdown and reboot now ---
  if (autoReboot)  {
    printf("shutting down and restarting OS...\r\n\n");  fflush(stdout);
    fprintf(DiagLog, "shutting down and restarting OS...\r\n\n");
    sleepMS(1000);  doShutdown = doReboot = true;
  }

  Exit(0);
}


//-----------------------------------------------------------------------------
