//----- ModularPSAppl.c ----- 12/04/13 -----(03/05/06)----

// EE Documents bundle # 211 (Older FET design) and # 160 (newer SCR design)


// command for temp "soft scope" traces needed for debugging control loop:

//    scope 5 ampApply ampRead ttlbulkcur bjtbias vctrl vout polcmd outmode

//-----------------------------------------------------------------------------
//  Application specific code for the main application task that is run at
//  a fixed interval and application specific commands run on demand.
//
//  Application code for the Modular Power Supply
//-----------------------------------------------------------------------------
//  TO-DO:
//
//  - With the VLIM settings having much more affect on the gain of the
//    regulation circuit, can they be used to manage the oscillation better?
//    Could they, in fact, be used to ALMOST disable the feedback circuit
//    while changing the polarity of the connection to the load?
//
//
//  - More intelligent detection of calibration problem and load state:
//
//    While it is true that all controllers require properly calibrated
//    readings in order to provide accurate information to the control system
//    and to function as designed, the ModularPS controller is more dependent
//    on the relationships between such readings than most.  An offset or gain
//    error on any one of the readings can easily cause to control logic to
//    conclude the system is in a different state than the one it is really
//    in which can result in a system that does not work smoothly, does not
//    work at ALL, or damages the equipment.
//
//    For that reason, the ability to monitor the behavior of the system and
//    log or otherwise react to any anomolies it discovers can be a big help
//    in avoiding or at least diagnosing problems.  The more checks and
//    balances we can incorporate, the less likely we are to have unexpected
//    problems, and easier it will be to diagnose them when they do occur.
//
//    One possibility that is not (yet) implemented would be to determine the
//    approximate slope of certain readings and use that to double-check the
//    conclusions that are currently based almost entirely on the most recent
//    values (i.e. historical data plays little or no part in most of the
//    logic).  Such an approach could help detect and warn of offset errors
//    in the calibrations or possibly a hardware failure.
//
//    Of course, this requires an efficient means of tracking the slope.
//    It would require a method that could update the value with each new
//    reading while using very little CPU time.
//
//   -----------------------------------------------------------------------
//
//    Assuming the bulk/bridge current, load current, and load voltage readings
//    are correct, then the resistance of the dump resistor should equal:
//
//         load volts / (bridge cur - load cur)
//
//    And the resistance of the load at any given moment is:
//
//         load volts / load cur
//
//
//    In theory, it should be possible to use the 1st to compensate for
//    calibration errors as follows:
//
//       - Whenever the load and bridge currents are stable, compute the
//         apparent dump resistor value and save it (log a message if I have a
//         previous value that is significantly different).
//
//       - If the computed value is fairly consistent for multiple current
//         values, then the calibration must be fairly good.
//
//         If the values are different, then compute what gain and offset
//         values (probably just for the bridge current) would result in a
//         consistent value.  At the very least, these values would be logged.
//         If the method proves to be reliable, then we could use them instead
//         of the existing calibrations to determine what mode the load and
//         supply are in (regulating, dumping, sinking, ramping, etc).
//
//    Questions:
//
//         Is there anything about the behavior of the load itself that would
//         affect this logic?  Even if the load resistance was different for
//         different (steady state) current values, that would not change the
//         above relationship for a given current. Of course, this assumes the
//         resistance of the dump resistor doesn't change much as the current
//         varies over the normally used range of values.
//
//         While the ttlBulkCur/bridge current is the reading that seems to be
//         most frequently in error, it is possible that the load current and
//         voltage readings are wrong as well.  Can changing the (assumed) gain
//         and offset for just the bridge current make up for one or both being
//         wrong?
//
//   ---------------------------------------------------------------------
//
//  When logging a message about an observed error in some reading, it
//  would help if the message listed the possible causes (i.e. which
//  calibration values for which signals could result in the observed
//  error).
//
//  - Log the following as calibration errors:
//
//     - The load current is changing and the polarity (sign) of the slope
//       is different than that of the load voltage.
//
//     - The bridge/bulk current is not close to the load current and the
//       polarity (sign) of the difference is not the same as the
//       polarity of the load voltage.
//
//
//     - The following is true:
//         - The supply is on
//         - The mode has been stable in REG+ or REG-
//         - The load current is stable
//         - The readback error or total bulk current error is > a
//           specified percentage of the maximum reading
//
// --------------------------------------------------------------------------
//  NOTES:
//
//     (Most of this refers to behavior for inductive loads)
//
//     Regardless of what the readings tell us and how we may have been
//     able to correct for calibration errors, regulation won't work unless
//     the feedback regulation circuit is properly adjusted.  So we
//     might as well assume it IS (since nothing will work if it isn't),
//     and also make the additional assumption that follows from that:
//
//        The controller is changing the polarity of the bridge to
//        match the polarity of the output of regulation circuit, and
//        the polarity is oscillating, then the bridge current is
//        close to 0.
//
//     Of course, this assumes that the readback of the regulation
//     circuit's output is correct (or at least its offset is).
//
//     But this is also another opportunity for double-checking the accuracy
//     of the calibrations:
//
//       - The absolute value of the regulation circuit's output voltage
//         should be roughly the same just before and just after it switches
//         polarity.  If it isn't, then the offset is probably wrong.
//
//       - When the polarity of the setpoint for the load current is changed,
//         and the previous and new values are not close to 0, then we can
//         record the absolute value of the load current at the time the
//         polarity of the regulation circuit changes.  Then, at a later time
//         when the polarity of the setpoint changes back (again, to a value
//         not close to 0), then the absolute value of the load current at
//         the time the polarity of the regulation circuit changes back
//         should be about the same (i.e. the absolute value of the load
//         current should always be about the same when the polarity of the
//         regulation circuit first changes).
//           NOTE:  Do not confuse changing the polarity of the connection
//         to the load with the polarity of the regulation circuit.  The
//         changing of the polarity of the connection to the load is done by
//         the software and always occurs after the change in the regulation
//         polarity (LONG after, in terms of how fast the regulation circuit
//         reacts to changes - this lag is what makes avoiding oscillation
//         during a change in polarity nearly impossible).
//
// -------------------------------------------------------------------------
//
//    - Add code to track the observed vout (load voltage) and determine when
//      it appears to be "railed".  This differs from the current method for
//      setting "voltageLimited", which assumes that the hardware and software
//      gains are all correct.
//
//    - The per-supply error checking includes comparing the load current to
//      the supply currents, but doesn't yet take in to account the difference
//      that occurs when ramping or sinking because some of the current is
//      going through the dump resistor.
//
//    - The magnitude of the load voltage when "railed" should be the same
//      for both postive and negative values.  If they are significantly
//      different, then log a message that indicates a problem (i.e. one or
//      more of the following are incorrect):
//         - The hardware gains for vout
//         - The adcCalib or readCalib values for vout
//         - The dacCalib or setCalib for vlimPlus and/or vlimMinus
//
//    - Add code to track and log the observed ratio of vctrl to ifdbk when
//      iset, ifdbk, ttlBulkCur, bjtBias, vctrl, vout, and polCmd are all
//      stable.
//
//    - Add code to track and log the observed gain for the iset/ifdbk portion
//      of the feedback circuit (i.e. the ratio of vctrl to iset-ifdbk during
//      the VERY short period between a steady state and when vout is railed
//      after a change in the setpoint). (Q: Is the period actually TOO short
//      to get useful data to do this?  Especially with the change from a 10ms
//      to a 20ms scheduler tick interval?)
//
//    - Add code to track and log the actual ramp rate of the load current.
//      The ideal ramp rate would be the minimum rate required to cause
//      vout to "rail" (i.e. for the vlim settings to take affect) as this
//      minimizes the oscillation that occurs around a polarity change without
//      slowing down the rate of change.
//
// Question:
//    Are the gain settings for vlimMinus, vlimPlus, and vout dependant on the
//    load?
//
//
//-----------------------------------------------------------------------------
//  Update Log (most recent first)
//
//
//  2013-12-04 - Mark Davis
//               - Added logic to update new onStatesBitMap in the DiagLogInfo
//                 structure.
//               - Added checkPrevOnStates() for setting offDueToReset, which
//                 is used to assert an interlock if the supply shut off due to
//                 an unexpected reset.
//
//  2013-07-01 - Mark Davis
//               - Added PS_CTLR_RESET bit to statusBitsAppl.  This is set
//                 during startup whenever the previous state could not be
//                 restored, and cleared the first time one of the supplies is
//                 successfully turned on.
//
//  2012-11-16 - Mark Davis (*)
//               - Added debounceMS setting and the debounceDigIn() function
//                 developed for the TCP controller.  Prior to this, there
//                 was at least one observed instance of anomalous readings on
//                 the digital inputs (where nearly all the inputs changed from
//                 a 0 to a 1 state for one pass).
//
//  2012-10-19 - Mark Davis (*)
//               - Readings that indicate a bulk supply is shorted must be
//                 consistent for > 700 ms (which covers just the last 3
//                 readings for bulk supplies) before it triggers a shorted
//                 error (and an interlock that shuts off the supplies).
//                 Prior to this, many of the controllers experienced what
//                 appeared to be spurious one-time readings that triggered
//                 the error when there wasn't actually a problem.
//
//  2012-03-22 - Mark Davis
//               - Added logic to limit polarity and mode change messages
//                 while waiting to turn the supplies on.  Prior to this,
//                 there could be up to 500 messages logged while waiting
//                 (even more if new setpoints were received while waiting).
//
//  2012-03-19 - Mark Davis
//               - Added loop at beginning of applic task to insure that we
//                 have read the state of ALL the bulk supply signals prior
//                 to normal initialization.
//
//  2012-03-12 - Mark Davis
//               - The flag indicating the code detected a load type config
//                 error (i.e. inductive load behavior when the type was not
//                 set to inductive) is now reset when current conditions do
//                 not indicate this.  Prior to this change, the flag would
//                 stay set (and generate periodic error messages in the log)
//                 until the conflict was resolved (as opposed to waiting
//                 until the error was seen again).
//
//  2012-02-17 - Mark Davis
//               - Moved some of the code from updatecurState() to a seperate
//                 function because updateCurState() was large enough to cause
//                 linker problems (it was larger than 4K which, depending on
//                 its location, can result in pieces residing in 3 different
//                 4K blocks, while the Rabbit extended program counter used
//                 for executing code in extended memory spans only 2 4K blocks
//                 at a time).
//
//  2011-12-28 - Mark Davis (*)
//               - Added bulkSupErr flag that considers only relevant bulk
//                 supplies when checking for interlock errors.  Recent changes
//                 prior to this would result in unclearable interlocks.
//               - Added bridgeStateStable flag and used it to avoid updating
//                 readings for bulk supplies until after the bride has been
//                 on for a short period.  This reduces the number of errors
//                 (and resulting log messages) about the current operating
//                 state.
//               - Added ampReadMax (set to the largest load current reading
//                 possible using the current analog calibrations) and used
//                 it to set the bulkCurClose flag.  The later is used to help
//                 detect when the system is in ramping or sinking modes.
//               - Updated the logic that keeps track of the number of msgs
//                 logged between setpoint and on/off changes and disables
//                 many of the messages if too many have occured since the
//                 last change (to avoid filling the log with useless info).
//               - Added bits to vctrlStatus for regPolStable and inductiveLoad
//               - The inductiveLoad flag is once again set from the loadType
//                 configuration value.  However, new logic was added that
//                 periodically logs a warning message if the load behaves
//                 like an inductive one with a load resistor but the loadType
//                 setting does not specify this type of load.
//               - Messages are now periodically written to the syslog
//                 indicating the state of key values.  The interval varies
//                 based on the current operating mode, and the same messages
//                 are triggered by changes in various values.
//
//  2011-12-13 - Mark Davis (*)
//               - The bias and current errors for bulk supplies are still
//                 logged but no longer cause an interlock.
//               - All changes to the individual bulk supply error states are
//                 now logged.  Prior to this, the clearing of error conditions
//                 were not logged.
//               - The error checking for the bulk supplies now treats a
//                 change in polarity the same as the supplies being turned
//                 on (in terms of giving the bulk supply currents time to
//                 stablize under the new conditions).  Prior to this change,
//                 the checking resumed too quickly after a polarity change
//                 and would see bulk supply currents of 0 amps when the load
//                 current was still quite high, which would cause an interlock.
//               - Changed the ampApply code so it uses the actual sched tick
//                 rate.  Previous to this, it assumed a fixed 10ms interval
//                 which resulted in a slower then specified rate of change to
//                 the setpoint.
//
//  2011-12-08 - Mark Davis (*)
//               - Updated code to comply with the newer version that has a
//                 regulation circuit that reduces the magnitude of the
//                 control signal oscillations around 0.  These changes
//                 include the following:
//                   - The input that used to read vlim now reads bjtBias
//                   - The new design appears to use the pos and neg vlim
//                     settings to limit the effect of the set/read error on
//                     the control voltage.  It is not clear how much influence
//                     it still has on actually limiting the load voltage.
//               - The UNCAL output mode is asserted only if it continues for
//                 at least maxErrTime milliseconds.  Prior to this, it
//                 would be asserted frequently between mode changes, filling
//                 up the log with misleading messages (it is STILL doing this,
//                 although hopefully to a lesser degree).
//               - Like a polarity change for any other mode, changing the
//                 bridge polarity for SINK mode now requires the polarity of
//                 the control voltage to be stable for a period before the
//                 change.  Prior to this change, the software would allow
//                 rapid changes in the bridge polarity while in SINK mode.
//               - The error checking for the state of each bulk supply now
//                 takes the OFF state of the bridges in to account.  Prior to
//                 this an interlock would occur when it was off and the load
//                 current was non-0 (i.e. it was dumping energy via the dump
//                 resistor while the bridge was off).
//
//  2011-06-24 - Mark Davis
//               - Changed the order of the data elements in the devData file
//                 to deal with order-of-restoration dependencies.
//
//  2010-02-08 - Mark Davis
//               - Added msgs for the individual bulk supplies turning on and
//                 off
//
//  2010-02-04 - Mark Davis
//               - maxErrtime is actually used now!
//
//  2009-12-14 - Mark Davis
//               - Changed order of param files to deal with problems caused
//                 by the order in which they were restored at startup.
//
//  2009-11-02 - Mark Davis
//               - Moved contrast to settings persistent file.
//
//  2009-09-23 - Mark Davis
//               - If the setpoint has not changed in several minutes and the
//                 rate at which certain state changes happen exceeds some
//                 threshold, then the logging of the state changes is disabled.
//                 Prior to this leaving the supplies on with a near 0 setting
//                 would cause an endless stream of log messages.
//
//  2009-07-01 - Mark Davis -
//                - The enabled/disabled setting for each supply is now part
//                  of the persistent configuration data.  This was necessary
//                  because many of the controller boards are not completely
//                  stuffed resulting in input values that are erroneous.
//
//  2009-04-23 - Mark Davis -
//                - Fixed macros used to toggle the ActivityLED output bit
//
//  2009-03-20 - Mark Davis -
//                - Added digInDescr[] and digOutDescr[] to support the newer
//                  generic methof for logging changes of interest in the state
//                  of the digital I/O signals.
//
//  2009-02-03 - Mark Davis
//               - added code to skip logging msgs for most input bit changes
//                 if the applic task is starving other processes for CPU cycles
//
//  2009-01-20 - Mark Davis
//               - modified formatting of ISET and IFDBK for "state" command
//                 and LCD interface to deal with changes in how we calibrate
//                 those values.
//
//  2008-11-13 - Mark Davis
//               - made the numeric portion of the labels for bulk supply
//                 signals 2 digits for all cases (to avoid ambiguity or
//                 mismatches when specifying signal names)
//               - reversed polarity of VLIM_POS and VLIM_NEG to deal with
//                 new calibration regime
//
//  2008-08-26 - Mark Davis
//               - added check for a psCount value of 0
//
//  2008-07-02 - Mark Davis
//               - ttlBulkCur register now reads +/- again depending upon
//                 bridge polarity
//
//  2008-06-17 - Mark Davis
//               - new loadType setting is now actually used (in place of
//                 logic that tried to determine load type automatically)
//               - scaled voltage reading (readVal[]) rather than "raw"
//                 reading (anaIn[]) is used to compare load voltage to
//                 voltage limits (which is important to detect RAMP and
//                 SINK modes)
//
//  2008-06-11 - Mark Davis
//               - vlimMinus setting now checks the limit of its setting
//                 properly so it doesn't force the setting to the wrong
//                 value
//               - changed default labels for temp OK bits for bulk supplies
//
//  2008-06-05 - Mark Davis
//               - loadType is now a configurable value, and both loadType
//                 and supplyType are settable via the LCD screen (althogh
//                 the supplyType setting doesn't yet do anything...)
//
//  2008-05-21 - Mark Davis
//               - new values for vlimMinus and vlimPlus now compared to
//                 min/max set (scaled) values rather than min/max DAC values
//
//  2008-04-28 - Mark Davis
//               - changes for current-to-voltage-loop version of regulation
//                 circuit
//
//  2008-02-11 - Mark Davis
//               - changes for supporting bulk supplies that use SCR's rather
//                 than BJT/FET bridges on their outputs.
//
//---------------------------------------------
//  2008-01-30 - Mark Davis
//               - brought up to date in terms of latest code structure
//               - Added minBSVolts setting to detect when a bulk supply
//                 is not working properly.  Any bulk supply whose voltage
//                 drops below this value for any significant period of time
//                 while "ON" causes an interlock and a message is added to
//                 the log.
//
//  2007-09-19 - Mark Davis
//               - first properly functioning prototype!
//
//  2007-05-08 - Mark Davis -
//               - Started new project.
//
//-----------------------------------------------------------------------------

/*
===============================================================================
                   Description of persistent register values
===============================================================================

  The main concepts to keep in mind:

    - Regulation is done by an analog feeback loop

    - The Rabbit's job is to control the bridges that connect the load to the
      power supplies and change the polarity of the bridges as necessary to
      match the polarity of the control voltage from the feedback loop.

    - The Rabbit has no way to stop the feedback loop from TRYING to regulate
      the current through the load, even when the load is not connected.


Basic design flaw:

  The above indicates a basic flaw in the design:  The bridge connecting the
  supplies to the load has to be turned off for some period of time to switch
  the polarity.  During this time the regulation circuit tries to keep
  regulating the current through the load.  But of course it can't when the
  load isn't connected.  The result is that, whenever the load is disconnected,
  the control voltage (VCTRL) spikes.

  This really doesn't matter while the load is disconnected.  But the moment
  you reconnect it, the control voltage again, in the opposite direction.
  The Rabbit responds by disconnecting and then reconnecting the load in the
  new polarity, which just causes the cycle to repeat.

  In earlier versions, this resulted in wild oscillations as the Rabbit would
  keep disconnecting and reconnecting the load, trying to keep the polarity
  the same as the control voltage from the feedback loop's output.  Because
  the Rabbit was so much slower, it couldn't keep up, and the oscillation could
  go on for quite some time.

  Even worse, in the original version, the feeback loop's control voltage
  set the polarity for the lower half of the bridges, resulting in a short
  across the bulk supply outputs until the Rabbit swapped the top half to
  match.  And because the shorted state meant the load current was not being
  affected by the feedback loop, these spikes could produce considerable
  current, potentially damaging the supplies and/or the bridges.

  Since the Rabbit has no way to disable the regulation circuit, it cannot
  eliminate the spike that happens each time it switches the polarity of the
  connection to the load.  The best that can be done is to try to minimize
  them.  Instead of trying to do what is CAN'T do (i.e. change the bridge
  polarity faster than the feedback loop can respond), it does the opposite:
  It figures out when it should IGNORE the polarity of VCTRL.


  In the latest version, the flow of current through the bridge is disabled
  when the polarity of VCTRL is different than what the bridge is configured
  for.  So when the polarity of VCTRL suddenly changes, at least it does not
  result in a power spike through the bridge, which greatly reduces the risk
  of damaging the load or the supplies.

  However, on the older FET/BJT bridge design, this is a major issue.  If
  the Rabbit simply waits for the polarity of VCTRL to change, by the time
  it reacts, the bottom portion of the bridge (controlled by the polarity
  of VCTRL) has already changed, causing a short through the bridge that
  bypasses the load.  Trying to keep up with the changes in VCTRL just
  prolongs the oscillation.  Turning off both sides of the upper portion
  of the bridge stops the oscillation, but eventually it has to be turned
  back on, at which point the oscillation will start again.

  Unfortunately, "If it hurts, don't do it!" isn't an option.  We can't
  just leave the load disconnected all the time.  The only viable option is
  for the Rabbit to switch the polarity of the bridge when needed, but then
  ignore changes in the VCTRL polarity for a short period.

  For the older version, this means that the bottom half of the bridge will
  keep oscillating back and forth between causing the bridge to be a short
  (with the load disconnected), and connecting the load using the polarity
  needed to nudge it further along in the direction that will eventually lead
  to VCTRL stablizing again.

  While this does (eventually) lead to stable regulation again, there is
  a modification on this theme that drastically reduces the magnitude and
  period of the oscillation:  If the Rabbit changes the polarity of the
  top half of the bridge JUST BEFORE the instant when VCTRL WOULD have
  changed polarity on its own, this avoids the huge initial spike that
  occurs if the Rabbit waits and simply reacts to the change AFTER it has
  happened.  It also helps a great deal if the Rabbit temporarily reduces
  the VLIM settings at the same time, which helps dampen the initial
  oscillation.

  The Rabbit does this by monitoring the state of the load and the feedback
  loop and trying to determine when a SINK mode is about to end (a SINK mode
  being the state where the magnet is dumping energy through the bridge and
  the dump resistor).  But of course, for this to work, the values the
  Rabbit is using must be accurate, which means the calibrations for the
  signals that provide those values must be accurate.

  NOTE:  One very logical alternative that was tried was to change the active
         setpoint to the match the readback when oscillation started (or was
         about to start) and then to slowly restore it until the system had
         passed through the region where oscillation occurs.  But the gain on
         the feedback loop made this impractical, as even a TINY difference
         between the setpoint and the readback would result in a large
         control voltage.  And given how much slower the Rabbit is, even if
         it HAD the needed accuracy in its settings and readings, it could
         not adjust the setpoint QUICKLY enough to avoid the oscillation.


Load Voltage limits:

  The controller also supports voltage limit settings for the voltage across
  the load.  The Rabbit supplies these to the feedback loop, which uses them
  to limit the control voltage (VCTRL) so as not to exceed these limits for
  any significant period of time.

  In the latest version, the voltage limit settings are applied to the
  IFDBK-ISET error as well as the load voltage.  This helps by putting a cap
  on the magnitude of the oscillations that occur around a change in the
  polarity of the bridge.

  Perhaps even more significant is the addition of switches between the
  control voltage from the analog feedback circuit (VCTRL) and the voltage that
  gets applied to the BJTs in series with the bridges.  The new arrangement
  is designed to ensure that the BJTs (now in series with the bridges) and
  the bridge itself (now made up entirely of SCRs) are conducting ONLY when
  the polarity command asserted by the Rabbit agrees with the polarity of the
  VCTRL signal.

  Since the Rabbit is WAY too slow to keep up with polarity changes in the
  VCTRL signal, this all but ensures that the bridge will not bypass the
  load (i.e. cause a short across the output of bulk supplies) while any
  significant amount of current is flowing through the BJTs.  And while it
  doesn't eliminate the problem of the VCTRL signal spiking in the opposite
  direction when the bridge is turned back on, it stops such spikes from
  causing a jump in current through the bridge.


The REAL solution (never implemented):

  While all of this means the latest version is a big improvement over the
  earlier versions, what would be even better is if the Rabbit could control
  the gain of the feedback loop, which would allow it to effectively eliminate
  the root of the problem:  A feedback loop that keeps trying to regulate the
  current through the load even when the load isn't connected.  If the
  gain on the feedback loop could be reduced far enough, the Rabbit could
  simply AVOID the conditions that result in the initial spike and the
  oscillation.



The importance of accurate calibrations:

  But of course NONE of this will work unless the Rabbit has accurate data,
  which means everything must be calibrated BEFORE this will work - which leads
  to a bit of a Catch-22, since such calibrations generally require running
  the controller connected to the device it has not yet been calibrated for.

  The best compromise has been to start with some default calibrations that
  allow the controller to operate well enough to complete the more precise
  field-level calibrations.


The settings:

  Roughly in order of importance (i.e. the ones closer to the start of the
  list are more likely to result in damage to the load, controller, and/or
  supplies if they are not reasonably close to the optimal values, followed
  by ones that determine whether the ctlr will or will not operate properly):

  bulkShortCur:      Value that indicates a short across a bulk supply's output
                     (i.e. both SCR's on one side of its output bridge are ON
                     so most or all the current is going through the 2 SCR's
                     on one side of the bridge and NOT contributing to the
                     current through the load).  This can save you from the
                     time consuming and expensive process of replacing the
                     bridge components if you get something else wrong.

  loadType:          Specifies if the load is a resistor (DC) or an inductor
                     (i.e. a magnet) with a dump resistor (RL).

  supplyType:        Specifies what type of bridge the supplies have: SCR's
                     or FET's with BJT's (not yet used, as this version of the
                     code is designed for the latest version with the SCR
                     bridges and polarity cutoff switches).

  sinkEndAmps:       The optimal total bulk-supply currrent value at which to
                     switch the polarity of the bridge at the end of SINK mode.
                     Affects the size of the unavoidable voltage spike that
                     occurs while the bridge is off (because the regulation
                     circuit, which is still active, keeps trying to regulate
                     but can't while the bridge is off).

  vlimReduction:     The amount the voltage limits are temporary reduced by
                     during sink mode to minimize the voltage spike when
                     sink mode ends.

  minBridgeOffTime:  Min time (in ms) the bridge must be off before being
                     turned back on (by reasserting a polarity command).  Also
                     affects the size of the unavoidable voltage spike that
                     occurs while the bridge is off.

  minPolStableTime:  Min time (in ms) the regulation polarity (inverse of VCTRL
                     polarity) must be stable before the bridge polarity is
                     changed to match.

  vlimPlus:          The maximum positive voltage across the load

  vlimMinus:         The maximum negative voltage across the load

  vctrlHyst:         The magnitude of the "deadman" range around 0 that VCTRL
                     must cross for changes in its polarity to be recognized.
                     This avoids constantly turning the bridge on and off when
                     VCTRL is near zero (and avoids initiating the very kind
                     of oscillation we are trying so hard to avoid).

  maxErrTime:        The maximum contiguous time normal operating parameters
                     are allowed to be exceeded before an Interlock condition
                     is asserted (in ms)

  rampRate:          The rate at which the active setpoint changes until it
                     reaches the requested setpoint (in Amps/second)

  maxAmpSet:         The maximum allowable setpoint (in amps)

  minPSonTime:       The minimum amount of time the bulk supplies and bridges
                     must be on before normal operation is allowed (in ms)

  minBSVolts:        The minimum Bias voltage that each Bulk Supply should have
                     while turned on and working properly.  If a Bulk Supply's
                     Bias voltage drops below this value for a significant
                     period of time, then an interlock state is asserted
                     (another guard against shorts).

  maxBSFracCurDiff:  Max allowed differrence in the current reading for each
                     active Bulk Supply while in regulation mode.  Expressed
                     as a fraction of the max current reading for each supply.

  knobSelectRate:    Controls how much the knob on the front panel has be be
                     turned to change what value is selected.


===============================================================================
              Notes on how the Modular Power Supply operates
===============================================================================

    The ModularPS controller supports up to 10 "bulk" power supplies.  The
    output is normally connected to an inductive load (a magnet) through
    bridges that acts as switches to control the direction current from the
    supplies can flow through the bridges and affect the load.

    The magnitude of the current flow from the supplies through the bridges is
    set by a control voltage produced by an analog feedback control loop which
    reacts much faster than the Rabbit can.

    The inputs to the feedback circuit consist of the following:
      - The current through the load.  This does NOT include the current
        through the dump resistor which is in parallel with the load.
      - The voltage across the load
      - The setpoint
      - Positive and Negative limits for the voltage across the load

    The analog feedback loop sets the control voltage based on the difference
    between the setpoint and the load current and how close the load voltage
    is to one of the limits.  As long as the load voltage is well below the
    set limits, the result is due primarily to the difference between the
    setpoint and load current.  But when the load voltage gets close to one
    of the set limits, the result is a "clipped" value which effectively
    limits the bridge current so as not to exceed the set voltage limits.

    Inputs to the Rabbit allow it to monitor the following:
      - The current through the load
      - The voltage across the load
      - The control voltage produced by the analog feedback circuit
      - The output current and voltage from each of the bulk supplies


    In the original version of the controller, the bridge consisted of FET's
    on the "top" and BJT's on the "bottom".  Which "side" of the bottom of
    the bridge was conducting was determined by the polarity of the control
    voltage, while which side of the top was controlled by the Rabbit.

    The Rabbit's primary task was to monitor the polarity of the control voltage
    (which told it which side of the bottom of the bridge was conducting) and
    to try to insure that the correct side on the top was conducting to insure
    that the output from the supplies was flowing through the load (and the
    dump resistor) rather than just through the bridge.  Because the Rabbit
    is so much slower than the feedback loop, the result would be a temporary
    mismatch, causing the bridges to be configured as a short across the
    outputs of the supplies.

    In the latest design includes several changes/enhancements:

      - The bridges are made up entirely of SCRs with one set of BJTs in
        series with the bridges (i.e. not one set for each polarity, as was
        the case when the BJT's were part of the bridges).

      - Switches that enable current to flow through the bridges only when
        the polarity settings for the bridge match the polarity of the control
        voltage from the feedback loop.  This guards against shorts, and
        provides a way for the Rabbit to turn off the bridge.  Once enabled,
        SCRs will continue to conduct after the enable signal is removed,
        unless the current flow drops below some fairly low level.  So even
        if the Rabbit stops asserting both polarity commands, the bridge
        would stay on if the current flow was not also stopped.

      - The voltage limit settings now place a cap on the gain applied to
        the setpoint/readback difference (ISET-IFDBK).  This helps to
        minimize the magnitude of spikes in the control voltage (VCTRL).


The control loop in more detail:

    The controller has an analog feedback loop that uses the following inputs:

      ISET:  Active setpoint for the analog feedback loop

      IFDBK: The current through the load
      VOUT:  The voltage across the load

      VLIM_NEG:  The largest negative voltage allowed across the load
      VLIM_POS:  The largest positive voltage allowed across the load

    and produces the following outputs:

      VCTRL:  The control signal whose polarity indicates the polarity the
              bridges should be set to.  This signal passes through switches
              and buffers that effectively do an absolute value operation on
              it and feed it to BJTBIAS, but ONLY when the polarity commands
              to the bridge match the polarity of VCTRL.

      BJTBIAS: The control signal (not including a hdw adjustable offset)
               applied to the BJTs that are in series with all the bridges
               for the individual supplies (i.e. this signal controls how
               much current flows through the bridges).  See note below.
               In effect:
                   BJTBIAS = (POL(VCTRL) == POL(BRIDGE)) ? ABS(VCTRL) : 0


    NOTES:

      - The total current flowing through the bridges is generally different
        than the load current when ramping or sinking an inductive load
        (because some of the current will be going through the dump resistor
        that is in parallel with the load)

      - In the latest design (bundle #160 rather than the older design in
        bundle # 211), the feedback loop limit settings VLIM_POS and VLIM_NEG
        seem to be used more to limit the effect of the iset/ifdbk difference,
        rather than limiting the actual load voltage.



    The software controls the following outputs:

       Analog:
          ISET, VLIM_NEG, VLIM_POS

       Digital:
          Positive and Negative polarity enable for the SCR bridges
          Positive and Negative polarity indicator LEDs
          On enable for each bulk supply
          Activity LED

    and reads the following inputs:

       Analog:
          IFDBK, VOUT, VCTRL, BJTBIAS
          Current and Voltage for each of the bulk supplies

       Digital:
          On/Off, Available, and Temp OK states from each of the bulk supplies
          Holec OK
          External Interlock


    The software controls the setpoint for the analog feedback circuit, the
    on/off state of the bulk supplies, the pos/neg gain limits, and
    (indirectly) the state of the SCR bridges connecting the load to the
    bulk supplies.

    The primary duties of the software are as follows:

      - Keep track of what state the load and the bulk supplies are in

      - Minimize the chaos that results when the polarity of the current
        through the bridge needs to be changed

      - Protect the load and the power supplies from harm


    To accomplish this, the software must be able to determine what state
    the supplies, load, and bridge are in and change its outputs at the
    correct time.

    The following describe the various states the code must be able to
    recognize:

    NOTES:
       The following discussion applies mostly to an inductive load with a
       "dump resistor" connected in parallel to the load.  The purpose of the
       dump resistor is to provide a controlled means of converting the energy
       stored as a magnetic field in to heat energy.

       A sudden collapse of the magnetic field must be avoided, as this would
       exert considerable force on the conductors within the coil, which would
       likely damage them.  By having a dump resistor connected in parallel
       to the coil, there is always a path for the electrons to flow, so a
       sudden collapse of the field is avoided.  The conductivity of the dump
       resistor determines how quickly the energy of the (controlled) collapse
       of the magnetic field is converted in to heat eminating from the dump
       resistor.

       In addition to protecting the load from excessive physical stress, the
       presense of the dump resistor means that, under certain conditions,
       there will be a significant difference between the current flowing
       through the bridge and the current flowing through an inductive load.

       This difference shows up when the current through the load is changing,
       which causes the magnetic field around the load to increase or decrease,
       which in turn exerts an opposing force on the electron flow (which
       means even a superconducting electro-magnet will have an effective
       resistance).

       At these times, a significant amount of current will flow through
       the dump resistor, causing a significant difference between the
       current flowing through the bridge and the current flowing through
       the load.


    REGULATION:
       This state is characterized by a steady state for all the major
       inputs and outputs:  VCTRL, BJTBIAS, IFDBK, VOUT, ttlBulkCur, ISET,
       polCmd.  In this state, the magnetic field around the inductor is
       also steady, so the inductor has little resistance and most of the
       current flowing through the bridges flows through the inductive load.

       The result is very little difference between the bridge/bulk supplies
       current and the load current.


    RAMPING:
       Ramping occurs when the energy stored in the load is being increased.

       During RAMPING, the difference between the setpoint and the load current
       causes the current output from the supplies to be increased, raising
       the current flowing through the bridges.  This causes an increase in the
       supply current flowing through the inductor, which enlarges the magnetic
       field, which increases the effective resistance of the load.  All of
       which means more of the increasing supply/bridge current will flow
       through the dump resistor, causing a significant increase in the
       difference between the load current and the bridge/supplies current, as
       well as a rise in the voltage across the load and dump resistor.

       In short, RAMPING mode is indicated by the following conditions:
        - The bridge polarity is the same as the load voltage polarity
        - The value (ttlBulkCur-ifdbk) is significantly shifted in the direction
          of the polarity of the load voltage polarity compared to the same
          value during REGULATION:
             i.e. POLARITY( CUR(ttlBulkCur-ifdbk) - REG(ttlBulkCur-ifdbk) )
                    == POLARITY(vout)
        - The load voltage is signifcantly larger than it is during REGULATION


    SINKING:
       Sinking occurs when the energy stored in the load is being decreased.

       During SINKING, the difference between the setpoint and the load current
       causes the current output from the supplies to be decreased, reducing
       the current flow through the bridges.  This causes a decrease in the
       supply current flowing through the inductor and dump resistor, which
       shrinks the magnetic field, which resists the change by increasing the
       current flow through the inductor.  Some of this current flows through
       the dump resistor in the opposite direction the bridge/supply current
       normally flows through it, so the voltage across it (and the load)
       reverses.

       In short, SINKING mode is indicated by the following conditions:
        - The bridge polarity is the opposite of the load voltage polarity
        - The value (ttlBulkCur-ifdbk) is significantly shifted in the direction
          of the polarity of the load voltage polarity compared to the same
          value during REGULATION:
             i.e. POLARITY( CUR(ttlBulkCur-ifdbk) - REG(ttlBulkCur-ifdbk) )
                    == POLARITY(vout)
        - The load voltage is signifcantly larger than it is during REGULATION




===============================================================================
*** Need to review/revise pretty much everything below this.  Some of it is ***
*** outdated, some of it is misleading, and some of it only applies to the  ***
*** older BJT+FET bridge design that predates both the SCR bridge designs)  ***
===============================================================================

Question:

  Shouldn't we be able to determine the effective resistance of the magnet
  based on the available readings and determine if a magnet has "quenched"
  (i.e. a portion of the windings got too warm to be superconduncting, so
  the resistance jumped?)


NOTES:

  The ModularPS controller contains hardware configurable gains for 2 of the 3
  possible sources for the IFDBK signal.  For the internal HOLEC, the gain is
  set by 6 jumpers that determine which of 6 resistors are connected to the
  feedback loop for an op-amp.  For the external shunt, a variable resistor
  sets the gain.

  For the control logic to work correctly, the read/set gains (and to a lesser
  degree, the offsets) must be set to match the hardware configuration.
  Unfortunately, the design does not provide any means for the sofware to
  know where the IFDBK signal is coming from or what the gain on it is.

  The only information related to the gain that the software has available is
  the number of bulk supplies that are connected.  The software COULD assume
  that the hardware gain is configured properly for the number of supplies
  connected, but all you would need to do is connect or disconnect a single
  supply without changing the hardware settings and that assumption would be
  incorrect and the controller would malfunction.

  The best that can be done (short of correcting this design flaw) is to add
  a software configurable parameter that tells the software how the hardware
  is configured, and compare that to the number of supplies connected.  If
  the two do not match, then log a warning and set a bit read by the control
  system to signal a possible error.

-------------------------------------------------------------------------------

  VOUT jumping to zero at the end of a sink period/start of the unstable
  period can only happen if both sides of the bridge are on, placing a
  short across the dump resistor and the magnet.

  VLIM is largely influenced by ISET-IFDBK.  The value of VLIM cannot be
  used (at least by itself) to determine things like how much current is
  flowing through the dump resistor, or when the regulation is primarily
  due to the voltage limit or to the current error.

  Comparing the polarity of VCTRL to VOUT when VOUT is close to or greater
  than the applied voltage limit is how you can tell if regulation is due
  to the voltage limit or the current error.  When VOUT is close to (or
  greater than) the applied voltage limit, and VCTRL specifies a polarity
  opposite the polarity of VOUT, then regulation is due to the applied
  voltage limit.

-------------------------------------------------------------------------------
  Remember:  A zero VCTRL has nothing to do with a zero load current.  The
             magnitude and polarity of VCTRL indicates how much CHANGE we
             want in the load current, not the VALUE of the load current!

             VCTRL is driven one of 2 values:
              - The difference between ISET and IFDBK
              - The voltage across the load exceeding the set limits

             The 1st case is in effect when the load voltage is below the
             specified limits, and the current is being ramped toward or
             maintained at a specific value.

             The 2nd case is in effect when the load voltage exceeds one of
             the specified limits, and some of the current is being diverted
             from the load resistor (through the bridge and - for the SCR
             type bridges - through the supplies themselves) to control
             the rate at which power is dumped through the resistor and
             turned in to heat.

             NOTE:  The dump resistor can handle an off condition in which
                    all the current is going through the resistor, but this
                    results in a faster rate of change in the current in
                    the magnet, which puts more stress on the magnet and
                    can damage it, so this is avoided whenever possible.


-------------------------------------------------------------------------------

  For the BJT/FET type bridges, control of the bridge is split between the
  regulation circuit and the software.  The output of the regulation circuit
  is VCTRL.  The polarity of VCTRL sets which of side of the lower (BJT)
  portion of the bridge is conducting, and its magnitude determines the
  amount of current.  The software determines which side of the upper (FET)
  portion of the bridge is enabled, and it monitors the VCTRL signal to
  determine which one should be enabled to keep current flowing through the
  load.

  One problem with this approach is that the software is FAR slower than the
  regulation circuit.  As long as the polarity of the VCTRL signal does not
  change, this is not a problem.  But when it does change (which it does
  whenever SINK mode ends), the output of the power supplies is shorted
  through one side of the bridge until the software sees the change and
  switches which set of FETs are enabled.

  Since this condition normally happens only at lower currents, it generally
  doesn't present a danger to the equipment, but the result is oscillation
  as the software keeps lagging behind the changes in the VCTRL polarity.

  What happens is this:

      VCTRL changes polarity, changing which side of the BJT portion of the
      bridge is conducting.  The software hasn't yet switched which FETs are
      conducting, so there is a temporary short through the bridge.

      During this period, the regulation circuit has no effect on the current
      it is trying to regulate.


    VOUT (the voltage across the load?) is (mostly?) determined by the
    current through the dump resistor that is in parallel with the magnet.

    When the FETs are disabled during SINK mode, the power is dumped
    primarily through the dump resistor.


    As the current through the load drops, the portion of the current ...

  ---------------------------------------------------------------------------

    Eash supply has its own bridge with the load connected to the center
    of the bridge.

  ==================================
    Supplies with FET/BJT bridges
  ==================================

           +---------------------------------------+
           |                                       |
           |                              +--------+-------+
           |                              |                |
           |                            |-+                +-|
           |              Enable(+) ---|                      |--- Enable(-)
           |                            |-+                +-|
           |                    [+] FET   |                |   FET [-]
           |                              +----- load -----+
           | (+)                          |                |
           |                              +--- resistor ---+
          bulk                            |                |
         supply                           +---+        +---+
           |                              |  _|_      _|_  |
           | (-)          abs(VCTRL) ___|/    +        +    \|___ abs(VCTRL)
           |                            |\   / \      / \   /|
           |                    [-] BJT   v  ===      ===  v   BJT [+]
           |                              |   |        |   |
           |                              +---+        +---+
           |                              |                |
           |                              +--------+-------+
           |                                       |
           +---------------------------------------+


     - FET forms "upper" half of bridges, BJTs the "lower" half

     - The software controls the FETs, while the regulation circuit controls
       the BJTs

     - The software monitors the state of the BJTs and the load and determines
       which (if either) side of the FET portion of the bridges should be
       enabled.

     - The software has no control over the BJTs.  The polarity of the VCTRL
       signal (the output of the regulation circuit) determines which side
       of the BJT portion of the bridge should be enabled.  The software must
       monitor the VCTRL value and keep the correct FET side enabled to
       avoid a short through the bridge.

     - The amplitude of the VCTRL signal determines how much current flows
       through the BJT portion of the bridge.  If one of the FET sides
       is enabled, it also determines how much current flows through each
       bulk supply.

     - The software is much slower (100Hz) than the regulation circuit.  Any
       time the polarity of VCTRL changes while the FETs are enabled, there
       will be a short period (before the software can detect and react to
       the change) when there is a short across the bridge.  This must NOT
       happen when there is any significant current through the supplies.


     - SINK mode operation:

        - The polarity commands are both disabled while in SINK mode, which
          turns off all the FETs

        - BJTs are still controlled by VCTRL to limit the dump rate.  During
          this time, the voltage across the load/dump resistor is effectively
          in control of the VCTRL signal, which adjusts the current flowing
          through the BJTs to maintain the voltage set by the VLIM values (by
          controlling how much current does NOT flow through the dump resistor).

     - Coming out of SINK mode:

        - When enough power has been dumped through the resitor that the
          voltage across it drops below the VLIM setting, the difference
          between ISET and IFDBK regains control of the VCTRL value, which
          then changes polarity, trying to drive the load current in the
          desired direction (by changing the BJT polarity).

        - At this point VCTRL begins to spike, trying to regulate the
          feedback value (the load current) which it does not yet have
          control over.

        - To minimize the magnitude of the spike, and magnitude and period
          of oscillation will occur when the FETs are turned back on, ISET
          is set equal to IFBDK.

        - The FETs are turned on based on the polarity indicated by VCTRL
          (before ISET was changed to match IFDBK).  At this point there will
          some osciallation of VCTRL (including its polarity).  Because the
          software cannot keep pace with the speed of these changes, it
          must "ride out" this period of oscillation, allowing the bridge
          to temporarily short at low current until VCTRL settles down.

        - For a fixed period of time after coming out of SINK mode:
           - Do NOT turn off or change the polarity of the FETs after they
             are turned on
           - Do NOT reassert SINK mode


  ==============================
    Supplies with SCR bridges
  ==============================

           +---------------------------------------+
           |                                       |
           |                              +--------+-------+
           |                    [+] SCR  _|_              _|_  SCR [-]
           |                        ---+ \ /              \ / +---
           |                            \_v_              _v_/
           | (+)                          | \            / |
           |                              +----- load -----+
          bulk                            |                |
         supply                           +--- resistor ---+
           |                              |                |
           | (-)                [-] SCR  _|_              _|_  SCR [+]
           |                        ---+ \ /              \ / +---
           |  abs(VCTRL)                \_v_              _v_/
           |       ---+                   | \            / |
           |         _|_                  +--------+-------+
           |         / \                           |
           +--------v   ---------------------------+


     - The bridge is made up entirely of SCRs.

     - The polarity commands turn on SCRs, but they cannot (directly) turn
       off SCRs that are conducting more than about 200ma of current (but
       they can indirectly as described below).

     - The BJTs are now in series with the bridge, controlling the amount
       of current flowing through the supply AND the bridge (in the FET/BJT
       bridge supplies, the BJTs controlled the current flowing through the
       bottom half of the bridge, which could be different than what was
       flowing through the supplies).

     - The only way the SCRs can be turned off is to insure no current
       through the bridge.  This is done by turing off the BJTs.

     - The VCTRL signal passes through an additional circuit that insures
       that the BJTs are enabled only when one of the polarity commands is
       asserted, and it is in agreement with the polarity specified by the
       VCTRL signal.  This insures that the bridge is disabled until the
       software detects and reacts to changes in the VCTRL polarity.


     - SINK mode operation:

       - The polarity command continues to be asserted in SINK mode to insure
         the bulk supplies are still connected.  This allows the regulation
         circuit to maintain the specified voltage across the load/dump
         resistor by controlling how much flows through the supplies rather
         than the dump resistor.

     - Coming out of SINK mode:

        - Treated the same as for the FET/BJT type bridges.


  ------------------------------------
    The end result?

    The only difference is that, for the SCR type bridges, we MUST leave
    the polarity command enabled to avoid disconnecting supplies from the
    bridge and turning off the bridge.  Doing so causes the dump resistor
    to absorb all the power being dumped by the magnet, with no control
    over the current going through it.  While the dump resistors are
    intended to handle this when needed, it should NOT be the normal mode
    of operation as it increases the likelihood of the magnet "quenching".

      NOTE:  The polarity commands themselves cannot turn off the SCRs.
             While disabling a polarity command disables the "on" signal
             to the SCRs, the SCR's will continue to conduct if there is
             any appreciable current going through them.  The only reason
             the SCR's DO go off (soon) after the polarity commands are
             disabled is that the circuit is set up to force the BJT's
             off unless a regulation polarity command is asserted that
             agrees with the regulation polarity.

===============================================================================
    At the end of a SINK period, there is a period where the polarity of
    VCTRL is not stable.  This period is is caused by the following:

      There are 2 factors that determine the magnitude and polarity
      of VCTRL:  The voltage across load and dump resistor, and the
      differece between ISET and IFDBK.

      As long

      Whenever the load voltage reaches one of the set limits, the
      voltage limit portion of the feedback control circuit applies
      an offset to VCTRL in the direction needed to keep the load
      voltage from exceeding the set limits.

      If the portion of VCTRL

      The current in the magnet has dropped just low enough that the
        voltage across the dump resistor dips below the set limits

      - The difference

    between ISET and IFDBK

--- When


--- Are we starting or ending that period at the end of SINK mode ---
--- where we have to keep the bridge off to avoid oscillation? ---
--- This period is a special "crossover" case because the following ---
--- conditions exist at this time: ---

      - The voltage limit we are/were at during SINK mode requires
        the bridge be set for one polarity to allow the bridge to
        divert some of the current flow from the magnet away from
        the dump resistor

      - The different between ISET and IFDBK requires the bridge to
        to be set to the opposite polarity to ramp the magnet current
        in the needed direction

      - When the load voltage dips below the set limit, the polarity
        of VCTRL (which specifies what polarity we need for the bridge)
        switches to that needed to ramp.  But attempting to ramp using
        the indicated polarity will bump the load voltage back to the
        limit it was just at.  Because the bulk supply's contribution
        to the current through the resistor is so small, the voltage
        limit portion of the control circuit once again overwhelms it
        and causes the polarity of VCTRL switch back.

    During this period, the only way to avoid this oscillation in the
    polarity of VCTRL is to wait for the power level in the magent to
    drop to the point where the VLIM portion of the feedback control
    simply reduces the magnitude of VCTRL, rather than causes it to
    change polarity.

    SO:  We detect that start of this period by watching the polarity of
    VCTRL.  If it switches while we are at one of the voltage limits,
    then we are at the start of one of these unstable periods.  When the
    voltage drops significantly below the limit, we are no longer in one
    of these unstable periods.


===============================================================================

  The feedback control circuit assumes that the polarity of the bridges
  follow what it is trying to do.  If they are backwards from what the
  polarity of the VCTRL signal requires, the result is runaway.  If they are
  OFF while it is trying to regulate, the result is a large VCTRL that
  will take effect as soon as the bridges are turned on.

  Properly controlling the bridges requires you know what state they are in.

  For the BJT/FET bridges, the state of the BJTs is determined by the VCTRL
  signal, which the controller reads, and the state of the FETs is set by
  the controller.

-------------------------------------------------------------------------------
  Load current vs bulk currents:

    The load current measures just what is going through the load (NOT the
    dump resistor in parallel with the load).

    The bulk currents measure how much is flowing through each bulk supply.

    A voltage across the load means there is a current through the dump
    resistor.

    If there is a voltage across the load, then the sum of the currents
    through the bulk supplies will be LARGER than the current through the
    load, because the load current does not include the current through
    the dump resistor.

    In fact, the difference between the sum of the bulk supply currents
    and the load current multiplied by the voltage across the load provides
    a ballpark value for how much power the dump resistor is disapating
    (ballpark because the bulk supply currents aren't always well calibrated,
    the values are lower resolution and lower sample rate, so the sum of all
    these errors is included in the difference).


  Characteristics of various modes:

    REG+/-

       The current through the load (noramlly a superconducting magnet) and
       the bulk supplies is fairly stable.  Because the magnet has almost
       no resistance to a steady current, the voltage across the load is
       nearly zero and VCTRL is at some small, steady value.

    RAMP+/-

       The ISET output is gradually ramped toward the new setpoint, and
       IFDBK is changing in the same direction.  For an inductive load,
       IFDBK lags behind ISET quite a bit, which is important, as this
       fact is used to detect RAMP and SINK modes.  For RAMP mode, the
       polarity for the bridge (the inverse of VCTRL's polarity) and the
       slope of IFBDK are the same so the bridge is left on so that it can
       actively drive the load in the desired direction.

    SINK+/-

       Just like RAMP mode, except that the polarity for the bridge is
       opposite the slope of IFDBK.  For the BJT/FET-type bridges, the
       bridge is turned off so the power supplies do not have to
       dissapate some of the power being dumped by the magnet.  For the
       SCR-type bridges, the SCR's will be locked on during SINK mode
       so turning off the bridges is not an option.



-------------------------------------------------------------------------------
  Misc thoughts:

    - What if instead of trying to "ride out" the glitch that happens when the
      FET's are turned back on at the end of SINK mode, we could eliminate
      the cause?  What if, just before turning on the FET's, we set ISET
      equal to IFDBK?  Would that eliminate the glitch?

    - Coming out of SINK mode seems to happen at a certain IFDBK voltage.
      Why is that?  Should it be based on the relative difference between
      IFDBK and ISET?  If the setpoint goes from a large value down to one
      that results in an IFDBK value well above 1.25 volts, it still has
      to go into and back out of SINK mode.  Does the glitch still happen
      in such cases?

    - Difference between load in SINK mode and supply in SINK mode (?)
      True or false:  The BJT switches polarity when VOUT drops below the
      VLIM_POS/_NEG limits, so the supply then tries to drive the load in
      the correct direction, even though it has some power to dissapate in
      the wrong polarity (?)  So the IFDBK voltage at which this occurs is
      dependant upon the VLIM_XXX settings?


===============================================================================

    The combination of ISET and IFDBK provides regulation of the load current,
    while the VLIM value imposes a limit based on the voltage across the load.
    The combined result (VCTRL) is used to bias the BJT's which control the
    combined power output of the bulk supplies.


    The software's role in the control loop is to monitor the state of the
    load and the BJT's and determine when one side of the bridge should be
    turned on via the positive or negative FET driver.  The VCTRL, IFDBK,
    ISET, and VOUT values provide the needed information.


    The polarity of the VCTRL signal determines which polarity BJT driver is
    enabled.  This value, and the difference between ISET and IFDBK allow
    the software to determine when the power in the load is decreasing (SINK
    mode) and when power should be increased or maintained (RAMP and REG
    modes).  This works as follows:

    - If the difference between ISET and IFDBK exceeds some preset level,
      then the power in the load is changing.

      If the BJT polarity is consistent with the direction of change (+ for a
      value that is getting more positive, - for a value that is getting more
      negative), then the power supply is in RAMP mode.  In this mode the
      load voltage (VOUT) is normally at the VLIM_NEG or VLIM_POS limit.

      If the BJT polarity is NOT consistent with the direction of change, then
      the supply is in SINK mode.

      If there is little difference between ISET and IFDBK, then the supply is
      in REGulation mode.  At this point, the voltage across the load (VOUT)
      will also be significantly smaller than the VLIM_NEG/VLIM_POS values.

    - When the supply is in SINK mode, the dump resistors on the bulk supplies
      are sinking (absorbing) power from the load.


    When in RAMP or REG mode, the bridge should be turned on and the polarity
    of the enabled FET driver normally must match the active BJT driver.

    In SINK mode, the FET's are disabled to avoid dumping lots of power
    through the bridge.

    When the mode changes from SINK back to RAMP or REG, the appropriate FET
    must be re-enabled.  However, this is where a bit of a problem occurs.

    SINK mode ends when the polarity of the active BJT changes.  But when the
    load is reconnected at the end of the SINK mode, VCTRL spikes, causing the
    active BJT to switch back.  If the software attempts to keep the active
    FET in sync with the active BJT at this time, the result is oscillation.

    To avoid this problem, the changes in the BJT state are ignored for a
    (configurable) period time after coming out of SINK mode.  This avoids
    having the control loop oscillate between SINK and RAMP or REG mode.
    Only a change in polarity of the ISET-IFDBK difference will shorten this
    period and cause the control loop to resume using the BJT state to trigger
    another SINK mode.

--------------------------------------------------------------------------
NOTES:
--------------------------------------------------------------------------
Signals for each of the 10 supplies:

12-Analog Inputs:
- HOLEC Current
- Bulk Supply Volt

Digital Inputs:
- On/Off status
- Temperature
- Roll Call Jumper

Digital Outputs:
- On/Off Command

-------------------------------------------------
Signals for all supplies:

16-bit Analog Inputs:
- IFDBK     current through load
- VOUT      voltage across load
- VCTRL     output of current regulation feedback-ctl circuit
- VLIM      scaled/limited load voltage used for feedback-ctl circuit

16-bit Analog Outputs:
- ISet      setpoint for current feeback-ctl circuit
- VLimPos   positive limit for VOUT feedback ctl
- VLimNeg   negative limit for VOUT feedback ctl

Digital Inputs:
- HOLEC OK
- External Interlock

Digital Outputs:
- Activity LED

- Negative Polarity
- Positive Polarity
- Negative Polarity LED
- Positive Polarity LED

------------------------------------------------------------------------------*/
#include <common.h>

#include <unistd.h>
#include <sys/types.h>



#ifndef  ApplicSource
  #define  ApplicSource  "ModularPSAppl.c"
#endif

#ifndef DidRegDefs
  #define  DidRegDefs
  #include <modbusRegs.c>
#endif
#ifndef RegDefs


//APPLCODE - change to the application name
const char far applName[] = { "ModularPS" };
const char ApplAbbrev[] = { "ModPS" };

//APPLCODE - add ID to unistd.h and place it here
#define DeviceType  DT_ModularPS

//APPLCODE - set the # of input and outputs used
#define  NUM_APPL_AIN     26   // analog inputs
#define  NUM_APPL_AOUT     3   // analog outputs
#define  NUM_APPL_DIN     32   // digital inputs
#define  NUM_APPL_DOUT    16   // digital outputs

#define  ADC_MASK_APPL  0x03FFFFFFL
#define  DAC_MASK_APPL  0x00000007L

//APPLCODE - set to state we want outputs in during initialization
#define InitialDigOutState  ( 0x0000 )


#define SharedINC
#include <sharedAppl.c>


#define  NumPS  10  // # of Bulk Power Supplies supported



//=============================================================================
//  definitions related to use of the Digital and Analog I/O signals
//=============================================================================

//=============== digitalIn bit usage ===============
#define  PS5_IS_ON       0x00000001    // DI-00
#define  PS5_AVAIL       0x00000002    // DI-01
#define  PS5_TEMP_OK     0x00000004    // DI-02

#define  PS10_IS_ON      0x00000008    // DI-03
#define  PS10_AVAIL      0x00000010    // DI-04
#define  PS10_TEMP_OK    0x00000020    // DI-05

#define  PS4_IS_ON       0x00000040    // DI-06
#define  PS4_AVAIL       0x00000080    // DI-07
#define  PS4_TEMP_OK     0x00000100    // DI-08

#define  PS9_IS_ON       0x00000200    // DI-09
#define  PS9_AVAIL       0x00000400    // DI-10
#define  PS9_TEMP_OK     0x00000800    // DI-11

#define  PS3_IS_ON       0x00001000    // DI-12
#define  PS3_AVAIL       0x00002000    // DI-13
#define  PS3_TEMP_OK     0x00004000    // DI-14

#define  PS8_IS_ON       0x00008000    // DI-15
#define  PS8_AVAIL       0x00010000L   // DI-16
#define  PS8_TEMP_OK     0x00020000L   // DI-17

#define  PS2_IS_ON       0x00040000L   // DI-18
#define  PS2_AVAIL       0x00080000L   // DI-19
#define  PS2_TEMP_OK     0x00100000L   // DI-20

#define  PS7_IS_ON       0x00200000L   // DI-21
#define  PS7_AVAIL       0x00400000L   // DI-22
#define  PS7_TEMP_OK     0x00800000L   // DI-23

#define  PS1_IS_ON       0x01000000L   // DI-24
#define  PS1_AVAIL       0x02000000L   // DI-25
#define  PS1_TEMP_OK     0x04000000L   // DI-26

#define  PS6_IS_ON       0x08000000L   // DI-27
#define  PS6_AVAIL       0x10000000L   // DI-28
#define  PS6_TEMP_OK     0x20000000L   // DI-29

#define  HOLEC_OK        0x40000000L   // DI-30
#define  NO_INTLK        0x80000000L   // DI-31

#define  PS_AVAIL_MASK   0x12492492L  // mask for all AVAIL bits

//--- map from Bulk PS index to first bit in a group of digitalIn bits ---
const uint  psDigInIdx[NumPS] = { 24, 18, 12, 6, 0, 27, 21, 15, 9, 3 };

//--- sensible, human-readable order for displaying digital inputs ---
const uchar  LogicOrderDIn[NUM_APPL_DIN] = {
  24,25,26, 18,19,20, 12,13,14,  6,7,8,   0,1,2,
  27,28,29, 21,22,23, 15,16,17,  9,10,11, 3,4,5,
  30, 31
};

//--- msgs to log automatically when a digital input changes state ---
const char far *digInDescr[NUM_APPL_DIN*2] = {
//-0 state descr-    -1 state descr-
  "PS1 is OFF",      "PS1 is ON",
  "PS5 NOT Avail",   "PS5 Avail",
  "PS5 Temp BAD",    "PS5 Temp OK",

  "PS10 is OFF",     "PS10 is ON",
  "PS10 NOT Avail",  "PS10 Avail",
  "PS10 Temp BAD",   "PS10 Temp OK",

  "PS4 is OFF",      "PS4 is ON",
  "PS4 NOT Avail",   "PS4 Avail",
  "PS4 Temp BAD",    "PS4 Temp OK",

  "PS9 is OFF",      "PS9 is ON",
  "PS9 NOT Avail",   "PS9 Avail",
  "PS9 Temp BAD",    "PS9 Temp OK",

  "PS3 is OFF",      "PS3 is ON",
  "PS3 NOT Avail",   "PS3 Avail",
  "PS3 Temp BAD",    "PS3 Temp OK",

  "PS8 is OFF",      "PS8 is ON",
  "PS8 NOT Avail",   "PS8 Avail",
  "PS8 Temp BAD",    "PS8 Temp OK",

  "PS2 is OFF",      "PS2 is ON",
  "PS2 NOT Avail",   "PS2 Avail",
  "PS2 Temp BAD",    "PS2 Temp OK",

  "PS7 is OFF",      "PS7 is ON",
  "PS7 NOT Avail",   "PS7 Avail",
  "PS7 Temp BAD",    "PS7 Temp OK",

  "PS1 is OFF",      "PS1 is ON",
  "PS1 NOT Avail",   "PS1 Avail",
  "PS1 Temp BAD",    "PS1 Temp OK",

  "PS6 is OFF",      "PS6 is ON",
  "PS6 NOT Avail",   "PS6 Avail",
  "PS6 Temp BAD",    "PS6 Temp OK",

  "*** HOLEC BAD ***",  "HOLEC OK",
  "*** External Interlock ***",  "External Interlock cleared"
};


//=============== digitalOut bit usage ===============
#define  PS10_SET_ON      0x00000001  // PS 10 ON ctl      DO-00
#define  PS5_SET_ON       0x00000002  // PS  5 ON ctl      DO-01
#define  PS9_SET_ON       0x00000004  // PS  9 ON ctl      DO-02
#define  PS4_SET_ON       0x00000008  // PS  4 ON ctl      DO-03
#define  PS8_SET_ON       0x00000010  // PS  8 ON ctl      DO-04
#define  PS3_SET_ON       0x00000020  // PS  3 ON ctl      DO-05
#define  PS7_SET_ON       0x00000040  // PS  7 ON ctl      DO-06
#define  UNUSED_OUT       0x00000080  // unused output     DO-07

#define  PS2_SET_ON       0x00000100  // PS  2 ON ctl      DO-08
#define  PS6_SET_ON       0x00000200  // PS  6 ON ctl      DO-09
#define  PS1_SET_ON       0x00000400  // PS  1 ON ctl      DO-10
#define  ENABLE_POL_NEG   0x00000800  // FET- ctl          DO-11
#define  ENABLE_POL_POS   0x00001000  // FET+ ctl          DO-12
#define  LED_POS          0x00002000  // Positive LED ctl  DO-13
#define  LED_NEG          0x00004000  // Negative LED ctl  DO-14
#define  ActivityLED      0x00008000  // Activity LED      DO-15

#define  PS_SET_ON_MASK      0x077F  // mask for all SET_ON bits

//--- map from Bulk PS index to digitalOut bits index ---
const uint psDigOutIdx[NumPS] = { 10, 8, 5, 3, 1, 9, 6, 4, 2, 0 };

//--- sensible, human-readable order for displaying digital outputs ---
const uchar  LogicOrderDOut[NUM_APPL_DOUT] = {
  10, 8, 5, 3, 1, 9, 6, 4, 2, 0, 11, 12, 13, 14, 15, 7 };

#define  TurnOnActivityLED()    ClearBits(Next.digitalOut, ActivityLED)
#define  TurnOffActivityLED()   SetBits(Next.digitalOut, ActivityLED)

//--- msgs to log automatically when a digital output changes state ---
//--- use NULL for bits that we don't want to log changes for ---
const char far *digOutDescr[NUM_APPL_DOUT*2] = {
//--0 state descr--    --1 state descr--
  "Set PS10 OFF",       "Set PS10 ON",
  "Set PS05 OFF",       "Set PS05 ON",
  "Set PS09 OFF",       "Set PS09 ON",
  "Set PS04 OFF",       "Set PS04 ON",
  "Set PS08 OFF",       "Set PS08 ON",
  "Set PS03 OFF",       "Set PS03 ON",
  "Set PS07 OFF",       "Set PS07 ON",
  NULL,                 NULL,
  "Set PS02 OFF",       "Set PS02 ON",
  "Set PS06 OFF",       "Set PS06 ON",
  "Set PS01 OFF",       "Set PS01 ON",
  NULL,                 NULL,   // "POL [-] Disabled",   "POL [-] Enabled",
  NULL,                 NULL,   // "POL [+] Disabled",   "POL [+] Enabled",
  NULL,                 NULL,
  NULL,                 NULL,
  NULL,                 NULL
};


//--- extraIn bit usage ---
#define  SELECT_BTN      0x0001  // select LCD page to show on LCD
#define  SET_BTN         0x0002  // change value on LCD display
#define  START_BTN       0x0004  // turn ON
#define  STOP_BTN        0x0008  // turn OFF
#define  RESET_BTN       0x0010  // reset interlock (?)


//=============== Analog input indexes ================
//--- 16-bit ADC channels ---
#define  AI_IFDBK    0
#define  AI_VOUT     1
#define  AI_VCTRL    2
#define  AI_BJT_BIAS 3
//--- 1st 12-bit ADC channels ---
#define  AI_BSV6     4
#define  AI_HECT6    5
#define  AI_BSV1     6
#define  AI_HECT1    7
#define  AI_BSV7     8
#define  AI_HECT7    9
#define  AI_BSV2    10
#define  AI_HECT2   11
#define  AI_BSV8    12
#define  AI_SPARE0  13
#define  AI_SPARE1  14
//--- 2nd 12-bit ADC channels ---
#define  AI_HECT8   15
#define  AI_BSV3    16
#define  AI_HECT3   17
#define  AI_BSV9    18
#define  AI_HECT9   19
#define  AI_BSV4    20
#define  AI_HECT4   21
#define  AI_BSV10   22
#define  AI_HECT10  23
#define  AI_HECT5   24
#define  AI_BSV5    25

//--- sensible, human-readable order for displaying analog inputs ---
const uchar  LogicOrderAIn[NUM_APPL_AIN] = {
   7, 11, 17, 21, 24, 5, 9, 15, 19, 23,  // cur1 - cur10
   6, 10, 16, 20, 25, 4, 8, 12, 18, 22,  // bias1 - bias10
   0,  1,  2,  3,   // ifdbk, vout, vctrl, bjtBias
  13, 14  // SPARE0, SPARE1
};


//=============== Analog output indexes ===============
#define  AO_ISET       0
#define  AO_VLIM_POS   1
#define  AO_VLIM_NEG   2

//--- sensible, human-readable order for displaying analog outputs ---
const uchar  LogicOrderAOut[NUM_APPL_AOUT] = { 0, 1, 2 };


//=============================================================================
//  List of daemon processes.  The inetd process listens for connection
//  requests on each port in the list and creates a new instance of the
//  specified process to service each new connection.
//=============================================================================
#include <services.h>

//--- service funcs declared the same as standard command funcs:
//     e.g.:   void _task fastdatad(int argc, char **argv);

void _task tsh(int argc, char **argv);


ServiceInfo inetdConf[] = {
//--- add any application-specific inetd-type network services here ---
//--- (add port # names to services.h) ---
//  NAME          PORT           PROTOCOL  FUNC       PRI  -1  Config Params
//  ------------  -------------  --------  ---------- ---  --  ---------------

//--- example of standard ones defined in sharedAppl.c ---
//{ "console",    TTY0,          ASYNC,    tsh,        10, -1, &consCfg },
//{ "fastdatad",  FASTDATA_PORT, ProtoTCP, fastdatad,   5, -1, NULL },
//{ "modbusTCPd", MODBUS_PORT,   ProtoTCP, modbusTCPd,  6, -1, NULL },
//{ "tsh",        TELNET_PORT,   ProtoTCP, tsh,        10, -1, NULL }

  #define SharedServices
  #include <sharedAppl.c>
};

const int  NumServices = sizeof(inetdConf) / sizeof(ServiceInfo);



//=============================================================================
//  Definition of data elements needed to keep track of the current state of
//  the device and the controller logic
//=============================================================================
typedef struct  {
//--- values updated in "cur" instance at start of each pass ---
  bool   isOn;       // bulk supply is ON
  bool   isAvail;    // bulk supply is available
  bool   tempOK;     // no temp problem with this supply

  float  biasVolts;  // bias voltage readback
  float  current;    // current output readback
  float  curDiff;    // diff from nominal value

  ulong  lastTimeBiasOK; // last time biasVolts was nominal
  ulong  lastTimeCurOK;  // last time current was nominal

  bool   biasErr;     // bias voltage exceeded max
  bool   curErr;      // current out of nominal range for too long
  bool   shortErr;    // current exceeded upper limit

//--- values updated in "next" instance during each pass ---
  bool   enabled;    // supply not manually or automatically disabled
  bool   setOn;      // assert ON for this supply
}  BulkSupplyState;


//-------------------------------------------------------------------------


// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****
//  Changing the applState structure means that a soft reset will fail!  A
//  change to the size will cause a soft restart to abort and do a CPU
//  restart instead.  And while a change that does NOT result in a new
//  size will not cause the soft restart to abort, it does mean that any
//  shift in the position of values can cause all kinds of mayhem.
// ***** WARNING ***** WARNING ***** WARNING ***** WARNING *****


typedef struct  {
  BulkSupplyState  bs[NumPS];  // state of each bulk supply

//--- values updated in "next" instance by tasks other than "applic" ---
//--- these are copied to "cur" at start of the pass and then "reset" ---
//--- (as needed) so we can tell if they are set again ---
  bool   onRequest;
  bool   offRequest;
  bool   resetRequest;

  float  ampSet;      // most recent setpoint

//--- values updated in "cur" instance at start of each pass ---
  ulong  digitalIn;

  bool   externIntlk;

  bool   regPol;      // current polarity indicated by VCTRL
  bool   hystRegPol;  // current polarity indicated by VCTRL with hysteresis

  bool   regPolStable;  // is the regulation polarity stable?
  bool   sinkOrDump;  // bridge polarity is opposite the load voltage polarity

  uint   btnBits;
  uint   psCount;     // # of avail and enabled bulk supplies

  float  ifdbk;       // unscaled IFDBK voltage reading
  float  vout;        // unscaled VOUT voltage reading
  float  vctrl;       // unscaled VCTRL voltage reading
  float  bjtBias;     // unscaled BJT BIAS voltage reading

  float  iset;        // unscaled ctl volt setting (scaled from ampApply)

  float  ctlErr;      //

  uint  outMode;      // OFF, WAIT, REG+, REG-, SINK+, SINK-, RAMP+, RAMP-, UNCAL

//--- values updated in "next" instance during each pass ---
  uint   state;       // STATE_INIT, STATE_OFF, STATE_ON, STATE_INTLK;

  float  ampApply;    // in-use value, ramping toward ampSet

  uint   polCmd;      // polarity cmd asserted for the bridges (OFF, -, +)

  ulong  digitalOut;  // state for digital outputs
}  ApplState;


ApplState  prevState, curState, nextState;

#define  Prev  prevState
#define  Cur   curState
#define  Next  nextState


const uint ApplStateDataSize =  ( sizeof(ApplState)*3 + sizeof(setVal) );



//===== "cur" values that we don't need to compare to "prev" instance =====
//----- (so we don't need them to be in the ApplState structure) -----

//ulong  msSincePrev;

uint   numLoggedChgMsgs; // # pol/mode chgs since last ampSet or on/off/int chg
uint  lastLoggableStateChg, lastEnabledChgMsg;
bool  logStateChgMsgs;

uint  psOnTime;  // the time the bulk supplies where turned on
bool  psOnAndReady;    // bulk supplies on long enough to be stable

uint  bridgeChgdTime;  // time of last bridge state change
bool  bridgeStateStable;  // bridge state unchanged for at least minPSonTime

float  absErr;  // error in load current

bool  tempRegPol, bridgeJustOff, voltageLimited, inductiveLoad, bulkSupErr,
      loadTypeWrong;

ulong  bridgeOffTime, unstableStartTime, lastModeSetTime,
       lastRegPolChgTime, lastNotWaitingTime;

uint  lastLoadTypeWarnTime, lastValLogTime, logInterval;

bool  ramping, regulating, sinking, dumping, lowLoadCur, holecOK, bulkCurClose,
      usePolarity, offDueToReset = false, loggedResetIntlk = false;

float  perSupCur, maxCurDiff, vlimAdjust, bulkCurDiff, absBulkCurDiff,
       absAmpRead;

uint  enabledMask;  // bit-mapped persistent vers of BulkSupplyState[].enabled

uint  bulkShortStarted[NumPS];


#define  Pos_Pol  true
#define  Neg_Pol  false


//--- values for polCmd ---
#define  POL_NEG   0   // negative FET driver ON
#define  POL_OFF   1   // both FETs OFF
#define  POL_POS   2   // positive FET driver ON

const char *PolCmdDesc[] = { "[-]", "OFF", "[+]" };


//=============================================================================
//  Defintions for variables that are accessable as ModBus registers
//=============================================================================
#else  // end of before-RegDefs portion

/*
  The purpose of this section is to define a set of C variables that are
  accessable as ModBus registers.  This approach avoids the need to use
  special functions or macros throughout the code when accessing register
  values, which keeps the application code fast and simple.

  The model used provides support for multiple logical devices.  The 2nd
  parameter in each register definition macro indicates if one instanace
  applies to all logical devices (specified by an "x" for "don't care"), or
  if each logical device requires its own instance (specified by an under-
  score "_" to signify "fill in the blank").  In the 1st case, the variable
  is defined as a scalar.  In the 2nd case, it is defined as an array.

  The definition macros support data types (floats and longs in particular)
  that require more than one register to store them.  To avoid the possibility
  of corrupted values, all reads of and writes to such values must be done
  using the multi-register commands and including all the registers that
  make up the value in the command.  Differences in register order between
  different client types is handled by a table that automatically swaps
  register values as needed based on the client type.

  The definition for each writable register also includes the function to
  call when the value is set via the ModBus interface.  Many registers can
  use the default setXXX() functions, but you can specify your own to provide
  any special logic needed (including ignoring the new setting if invalid).

  For registers, the format for the function definitions is:

    bool mySetFunc(SetRegFuncParams);

       where SetRegFuncParams is currently defined as:
              void *info, uchar chanNum, void *newVal, fd_t eventsfd

  For coils, the format for the function definitions is:

     bool (*SetCoilFunc)(SetCoilFuncParams);

         where SetCoilFuncParams is currently define as:
               void *info, uchar chanNum, uchar newVal, fd_t eventsfd

  The parameters to the call are as follows:
    *info:     A pointer to the Coil's/Register's CoilInfo/RegInfo structure
    chanNum:   The logical device # (0 is the 1st)
    *newVal:   A pointer to the new value requested in the Modbus packet
               (changed to native-byte order where needed)
    eventsfd:  The fd for the FIFO used to send msgs to the main application
               process

  The function returns true if the new value was accepted, false if not.

  The register numbers for multiple "channels" or logical devices are
  handled automatically based on the xxChanMult values.  These values specify
  the offset between a register value for one logical device and the "same"
  register for the Next.

  Example:

    Assume the following definitions:
      #define  NumLogicalDevices  2
      #define  R4ChanMult  100
      REG4( motorPos,       _, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos[2];
      register pair:  40050,40051  refers to:  motorPos[0]
      register pair:  40150.40151  refers to:  motorPos[1]

    If instead we had specified:
      REG4( motorPos,       x, 40050, F32, RW, setMotorPos )

    Then we would get:
      global variable definition:  float motorPos;
      register pairs:  40050,40051 & 40150,40151  BOTH refer to:  motorPos


  To keep the definitions as simple as possible, and avoid requiring multiple
  macros for each register definition, each macro is used more than once,
  with a different definition (#define) each time (a practice generally
  referred to as X-Macros).  While trying to follow the order of inclusion
  of files (including a recursive reference back to this file) can be rather
  confusing, the X-Macro method keeps the source code simple to read and
  maintain by doing everything with a single macro for each register.

  ==========================================================================
                             !!! WARNING !!!
  ==========================================================================
    Each sequence of register definitions MUST be contiguous!  There must
    NOT be any holes (i.e. no unused register #'s between used ones) and no
    overlap!
-----------------------------------------------------------------------------*/


#ifndef NumLogicalDevices
  #define  NumLogicalDevices  1
#endif

//--------------------------------------------------------------
#ifndef NumC0Chans
  #define NumC0Chans_  NumLogicalDevices
  #define C0ChanMult_  100
#endif

#ifdef COIL0

COIL0( DoReboot,       x, 1, setCoil )
//COIL0( Coil00002,      _, 2, setCoil )
//COIL0( Coil00003,      _, 3, setCoil )

#undef COIL0
#endif


//--------------------------------------------------------------
#ifndef NumR3Chans
  #define NumR3Chans_  NumLogicalDevices
  #define R3ChanMult_  100
#endif

#ifdef REG3

REG3( dlmVersTime,    x, 30001, U32 )
REG3( rtosVersTime,   x, 30003, U32 )
REG3( applVersTime,   x, 30005, U32 )
REG3( devType_,       x, 30007, U16 )
REG3( deviceID_,      x, 30008, U32 )
REG3( upSecs,         x, 30010, U32 )
REG3( percentIdle_,   x, 30012, U16 )

REG3( numChans,       x, 30013, U16 )
REG3( coilChanMult,   x, 30014, U16 )
REG3( reg3ChanMult,   x, 30015, U16 )
REG3( reg4ChanMult,   x, 30016, U16 )

REG3( digInBits,      x, 30017, U32 )

REG3( statusBitsStdA, x, 30019, U16 )
REG3( statusBitsStdB, x, 30020, U16 )
REG3( statusBitsAppl, x, 30021, U32 )

/*  Bias Supply Voltage for each supply */
REG3( bsv_1,          x, 30023, F32 )
REG3( bsv_2,          x, 30025, F32 )
REG3( bsv_3,          x, 30027, F32 )
REG3( bsv_4,          x, 30029, F32 )
REG3( bsv_5,          x, 30031, F32 )
REG3( bsv_6,          x, 30033, F32 )
REG3( bsv_7,          x, 30035, F32 )
REG3( bsv_8,          x, 30037, F32 )
REG3( bsv_9,          x, 30039, F32 )
REG3( bsv_10,         x, 30041, F32 )

/* Current total for each supply */
REG3( amps_1,         x, 30043, F32 )
REG3( amps_2,         x, 30045, F32 )
REG3( amps_3,         x, 30047, F32 )
REG3( amps_4,         x, 30049, F32 )
REG3( amps_5,         x, 30051, F32 )
REG3( amps_6,         x, 30053, F32 )
REG3( amps_7,         x, 30055, F32 )
REG3( amps_8,         x, 30057, F32 )
REG3( amps_9,         x, 30059, F32 )
REG3( amps_10,        x, 30061, F32 )

REG3( ampRead,        x, 30063, F32 )
REG3( voltRead,       x, 30065, F32 )   // same as vout (below)
REG3( vctrl,          x, 30067, F32 )
REG3( bjtBias,        x, 30069, F32 )   // was vlim

REG3( state_1,        x, 30071, U16 )
REG3( state_2,        x, 30072, U16 )
REG3( state_3,        x, 30073, U16 )
REG3( state_4,        x, 30074, U16 )
REG3( state_5,        x, 30075, U16 )
REG3( state_6,        x, 30076, U16 )
REG3( state_7,        x, 30077, U16 )
REG3( state_8,        x, 30078, U16 )
REG3( state_9,        x, 30079, U16 )
REG3( state_10,       x, 30080, U16 )

REG3( outMode,        x, 30081, U16 )

REG3( ampApply,       x, 30082, F32 )

REG3( psCount,        x, 30084, U16 )

REG3( state,          x, 30085, U16 )

REG3( iset,           x, 30086, F32 )
REG3( ifdbk,          x, 30088, F32 )
REG3( vout,           x, 30090, F32 )  // same as voltRead (above)

REG3( polCmd,         x, 30092, U16 )

REG3( vctrlStatus,    x, 30093, U16 )  /* combines regPol and hystRegPol */

REG3( ampErr,         x, 30094, F32 )  /* ISET-IFDBK scaled to amps */

REG3( ttlBulkCur,     x, 30096, F32 )  /* total bulk supply current */

REG3( unused2,        x, 30098, F32 )  /* was VLIM  w/o ISET-IFDBK contribution */

REG3( latency,        x, 30100, U16 )  // actual vs scheduled appl start in ms
REG3( tcpTickTime,    x, 30101, U16 )  // # ms inside tcp_tick() call
REG3( applLoopTime,   x, 30102, U16 )  // # ms used by applic task loop
REG3( kernelLoopTime, x, 30103, U16 )  // # ms used by last kernel pass

REG3( loadRes,        x, 30104, F32 )  // apparent load resistance
REG3( dumpRes,        x, 30106, F32 )  // calculated dump resistor resistance

#undef REG3
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//--- values for statusBitsAppl (implemented 07/01/13) ---
#define  PS_CTLR_RESET  0x000000001   // PS off since last CPU reset


//--- values for state_# registers ---
#define  PS_IS_ON     0x0001
#define  PS_AVAIL     0x0002
#define  PS_TEMP_OK   0x0004
#define  PS_SET_ON    0x0008
#define  PS_ENABLED   0x0010
#define  PS_BIAS_OK   0x0020

#define  PS_STATE(idx)  ( *((uint *)((&state_1_RegInfo + (idx))->regVal)) )


//--- values for outMode ---
#define  OM_OFF     0   // supplies are off
#define  OM_WAIT    1   // wait for supplies to stabilize or bridge to turn off
#define  OM_REG_P   2   // steady with + current flow
#define  OM_REG_N   3   // steady with - current flow
#define  OM_SINK_P  4   // dissipating power, positive current flow
#define  OM_SINK_N  5   // dissipating power, negative current flow
#define  OM_RAMP_P  6   // incr current flow in positive dir
#define  OM_RAMP_N  7   // incr current flow in negative dir
#define  OM_DUMP    8   // dumping energy via dump resistor
#define  OM_UNCAL   9   // unknown due to soft bug or calibration error


//--- string descr corresponding to outMode values ---
const char *OutModeDesc[10] = {
  "OFF  ", "WAIT ", "REG+ ", "REG- ", "SINK+", "SINK-", "RAMP+", "RAMP-",
  "DUMPG", "UNCAL" };


//  #define  READING(idx)  REG_ARRAY(float, rdg_1, idx)

#endif


//--------------------------------------------------------------
#ifndef NumR4Chans
  #define NumR4Chans_  NumLogicalDevices
  #define R4ChanMult_  100
#endif

#ifdef REG4

REG4( timeZone_,        x, 40001, S16, RW, MB_timeZone )
REG4( curTime,          x, 40002, U32, RW, setU32 )

REG4( ipAddr,           x, 40004, U32, RW, setU32 )
REG4( netmask,          x, 40006, U32, RW, setU32 )
REG4( gateway,          x, 40008, U32, RW, setU32 )

REG4( setACoil,         _, 40010, U16, RW, setU16 )
REG4( resetACoil,       _, 40011, U16, RW, setU16 )

REG4( digOutBits,       x, 40012, U32, RW, setU32 )

REG4( cmdBitsStd,       x, 40014, U32, RW, MB_cmdBitsStd )
REG4( cmdBitsAppl,      x, 40016, U32, RW, MB_cmdBitsAppl )

REG4( ampSet,           x, 40018, F32, RW, MB_ampSet )
REG4( vlimPlus,         x, 40020, F32, RW, MB_vlimPlus )
REG4( vlimMinus,        x, 40022, F32, RW, MB_vlimMinus )
REG4( rampRate,         x, 40024, F32, RW, MB_rampRate )

REG4( vctrlHyst,        x, 40026, F32, RW, MB_vctrlHyst )
REG4( minPolStableTime, x, 40028, U16, RW, setU16 )
REG4( unused1,          x, 40029, U16, RW, setU16 )
REG4( minBSVolts,       x, 40030, F32, RW, setF32 )
REG4( maxBSFracCurDiff, x, 40032, F32, RW, setF32 )

REG4( vlimReduction,    x, 40034, F32, RW, setF32 )
REG4( sinkEndAmps,      x, 40036, F32, RW, setF32 )
REG4( unused3,          x, 40038, F32, RW, setF32 )

REG4( knobSelectRate,   x, 40040, U16, RW, setU16 )

REG4( maxAmpSet,        x, 40041, F32, RW, setF32 )

REG4( minPSonTime,      x, 40043, U16, RW, setU16 )
REG4( minBridgeOffTime, x, 40044, U16, RW, setU16 )
REG4( unused4,          x, 40045, U16, RW, setU16 )

REG4( unused5,          x, 40046, F32, RW, setF32 )

REG4( maxErrTime,       x, 40048, U16, RW, setU16 )

REG4( contrast,         x, 40049, U16, RW, setU16 )

REG4( bulkShortCur,     x, 40050, F32, RW, setF32 )

REG4( supplyType,       x, 40052, U16, RW, MB_supplyType )
REG4( loadType,         x, 40053, U16, RW, MB_loadType )

REG4( debounceMS,       x, 40054, U16, RW, setU16 )  /* ms period for digIn debouncing */

#undef REG4
#endif


#ifdef RegMacros
//===== application-specific defintions related to register values =====

//  #define  ISET(idx)   REG_ARRAY(float, iSet_1, idx)

#endif

//--------------------------------------------------------------
#undef RegMacros

#endif
#ifndef RegDefs // start of after-RegDefs portion


//----- values for state -----
#define  STATE_INIT   0
#define  STATE_OFF    1
#define  STATE_ON     2
#define  STATE_INTLK  3

const char *CtlStateDesc[4] = { "INIT ", "OFF  ", "ON   ", "INTLK" };


//----- values for cmdBitsAppl -----
#define  CBA_TURN_ON     0x00000001
#define  CBA_TURN_OFF    0x00000002
#define  CBA_CLR_INTLK   0x00000004

#define  CBA_ENABLE_1    0x00000008
#define  CBA_ENABLE_2    0x00000010
#define  CBA_ENABLE_3    0x00000020
#define  CBA_ENABLE_4    0x00000040
#define  CBA_ENABLE_5    0x00000080
#define  CBA_ENABLE_6    0x00000100
#define  CBA_ENABLE_7    0x00000200
#define  CBA_ENABLE_8    0x00000400
#define  CBA_ENABLE_9    0x00000800
#define  CBA_ENABLE_10   0x00001000

#define  CBA_DISABLE_1   0x00002000
#define  CBA_DISABLE_2   0x00004000
#define  CBA_DISABLE_3   0x00008000
#define  CBA_DISABLE_4   0x00010000
#define  CBA_DISABLE_5   0x00020000
#define  CBA_DISABLE_6   0x00040000
#define  CBA_DISABLE_7   0x00080000
#define  CBA_DISABLE_8   0x00100000
#define  CBA_DISABLE_9   0x00200000
#define  CBA_DISABLE_10  0x00400000


//--- values for supplyType --- (not used at this time)
#define  TYPE_FET    1   // FET upper/BJT lower bridges, each
#define  TYPE_SCR    2   // SCR upper/lower bridges, BJT ctl for all bridges

const char *SupplyTypeDesc[3] = { "???", "FET", "SCR" };

//--- values for loadType ---
#define  LOAD_RL   1   // inductive (electro-magnet) load
#define  LOAD_DC   2   // resistive (or open) load

const char *LoadTypeDesc[3] = { "??LOAD??", "INDUCTOR", "RESISTOR" };

//=============================================================================
//  defintions for "parameterized" data values - lists of values that we want
//  to be "persistent" (i.e. saved to a file and restored on startup)
//=============================================================================

float dummyFloat;

//--- parameter ID's for data in Device Data/calibration file ---
//--------------------------------------------------------------------------
// WARNING:  ID's can be added, but existing ID's (and the format of the
//           data the represent) should NEVER change!  NOTE:  Existing ones
//           can be "eliminated" by changing their name and no longer
//           using them, but never REUSE an old one for something new!!!
//--------------------------------------------------------------------------
//  NOTE:  ID's >= 1000 are rsvd for values shared by most applications
//--------------------------------------------------------------------------
RESTORE_FUNC(enabledMask);

const ParamData far  devDataParams[] = {
  #define SharedDevParams
  #include <sharedAppl.c>

// NOTE:  The application-specific ones need to come AFTER the generic ones
//        if (as in this case) there are values that depend on the generic
//        ones (in this case, the vlimPlus and vlimMinus depend on the
//        setGain and setOffset values)
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
  PERSISTENT_VAR(  1,  vlimPlus,         NULL ),
  PERSISTENT_VAR(  2,  vlimMinus,        NULL ),
  PERSISTENT_VAR(  3,  rampRate,         NULL ),
  PERSISTENT_VAR(  4,  vctrlHyst,        NULL ),
  PERSISTENT_VAR(  5,  maxErrTime,       NULL ),
//  PERSISTENT_VAR(  6,  maxVctrl,         NULL ),
//  PERSISTENT_VAR(  7,  forcePolTime,     NULL ),
//  PERSISTENT_VAR(  8,  maxCurDiff,       NULL ),
//  PERSISTENT_VAR(  9,  maxVoltDiff,      NULL ),
//  PERSISTENT_VAR( 10,  maxDiffSec,       NULL ),
  PERSISTENT_VAR( 11,  knobSelectRate,   NULL ),
  PERSISTENT_VAR( 12,  maxAmpSet,        NULL ),
  PERSISTENT_VAR( 13,  minPSonTime,      NULL ),
  PERSISTENT_VAR( 14,  minBridgeOffTime, NULL ),
  PERSISTENT_VAR( 15,  minPolStableTime, NULL ),
//  PERSISTENT_VAR( 16,  rampThreshold,    NULL ),
  PERSISTENT_VAR( 17,  minBSVolts,       NULL ),
  PERSISTENT_VAR( 18,  maxBSFracCurDiff, NULL ),
  PERSISTENT_VAR( 19,  bulkShortCur,     NULL ),
  PERSISTENT_VAR( 20,  supplyType,       NULL ),
  PERSISTENT_VAR( 21,  vlimReduction,    NULL ),
//  PERSISTENT_VAR( 22,  contrast,         NULL ),
  PERSISTENT_VAR( 23,  sinkEndAmps,      NULL ),
  PERSISTENT_VAR( 24,  loadType,         NULL ),
  PERSISTENT_VAR( 25,  enabledMask,      PVRF(enabledMask) ),
  PERSISTENT_VAR( 26,  debounceMS,       NULL ),

  END_PARAM_DATA
};

//----------------------------------------------------
RESTORE_FUNC(ampSet);

const ParamData far  settingsParams[] = {
//APPLDATA
//                ID   ---value---  --RESTORE_FUNC--
  PERSISTENT_VAR(  1,  ampSet,       PVRF(ampSet) ),
  PERSISTENT_VAR(  2,  contrast,     NULL ),

  END_PARAM_DATA
};


//-----------------------------------------------------------------
//  Application-specific values needed to restore the state of the
//  controller following a "soft" reset (i.e. a restart without a
//  CPU reset)
//-----------------------------------------------------------------
const ParamData far  ctlrStateParams[] = {
// sample values from sharedAppl.c, common to all applications
//                   --ID--  ---value,granularity---  --restoreFunc--
//    PERSISTENT_VARx(  1000,  prevState,               NULL ),
//    PERSISTENT_VARx(  1001,  curState,                NULL ),
//    PERSISTENT_VARx(  1002,  nextState,               NULL ),
//    PERSISTENT_TBL (  1003,  setVal,4,                NULL ),
// // PERSISTENT_TBL (  1004,  shadowRegValues,4,       NULL ),

//APPLDATA
//    PERSISTENT_VAR(  1,  myApplStateVar,        NULL ),


  #define SharedCtlrStateParams
  #include <sharedAppl.c>

  END_PARAM_DATA
};


//-----------------------------------------------------------------------------
//  A list of parameterized "persistent" data.  Backup copies of such values
//  are stored in files and restored on startup.
//-----------------------------------------------------------------------------

//    Base File Name    Category  ParamData
//    ----------------  --------  ---------
//--- add any application-specific parameterized file defs here ---
//ParamFile  myAppCfgFile =
//  { "myFileName",     PDC_CTLR, myCfgParams,   0 };

ParamFile  settingsFile =
    { "settings",       PDC_SET,  settingsParams  };


#define SharedParams
#include <sharedAppl.c>


ParamFile *paramFiles[] = {
  #define SharedParamsList
  #include <sharedAppl.c>
//--- add any application-specific ParamData references here ---
//  &myAppCfgFile,
  &settingsFile
};

const int NumParamFiles = sizeof(paramFiles) / sizeof(paramFiles[0]);


//=============================================================================
//  Declarations for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//=============================================================================

void onRequest(void);
void offRequest(void);
void resetRequest(void);

void enableBulkSupply(uint idx);
void disableBulkSupply(uint idx);

void newSetpoint(float newAmps);


//=============================================================================
//  This is where misc global variables should be defined
//=============================================================================
#define SharedGlobals
#include <sharedAppl.c>

//--- add definitions for application-specific global variables here ---

float * const bsvX[] = {
  &bsv_1, &bsv_2, &bsv_3, &bsv_4, &bsv_5,
  &bsv_6, &bsv_7, &bsv_8, &bsv_9, &bsv_10
};
float * const ampsX[] = {
  &amps_1, &amps_2, &amps_3, &amps_4, &amps_5,
  &amps_6, &amps_7, &amps_8, &amps_9, &amps_10
};

//--- mapping from supply index to Analog Input #'s ---
const uint BSVidx[NumPS] = {
  AI_BSV1, AI_BSV2, AI_BSV3, AI_BSV4, AI_BSV5,
  AI_BSV6, AI_BSV7, AI_BSV8, AI_BSV9, AI_BSV10
};
const uint HECTidx[NumPS] = {
  AI_HECT1, AI_HECT2, AI_HECT3, AI_HECT4, AI_HECT5,
  AI_HECT6, AI_HECT7, AI_HECT8, AI_HECT9, AI_HECT10
};

//--- list of bit-masks for testing state of each bulk PS ---
const uint SetOnBit[NumPS] = {
  PS1_SET_ON, PS2_SET_ON, PS3_SET_ON, PS4_SET_ON, PS5_SET_ON,
  PS6_SET_ON, PS7_SET_ON, PS8_SET_ON, PS9_SET_ON, PS10_SET_ON
};

const ulong IsOnBit[NumPS] = {
  PS1_IS_ON, PS2_IS_ON, PS3_IS_ON, PS4_IS_ON, PS5_IS_ON,
  PS6_IS_ON, PS7_IS_ON, PS8_IS_ON, PS9_IS_ON, PS10_IS_ON
};

const ulong AvailBit[NumPS] = {
  PS1_AVAIL, PS2_AVAIL, PS3_AVAIL, PS4_AVAIL, PS5_AVAIL,
  PS6_AVAIL, PS7_AVAIL, PS8_AVAIL, PS9_AVAIL, PS10_AVAIL
};

const ulong TempOKBit[NumPS] = {
  PS1_TEMP_OK, PS2_TEMP_OK, PS3_TEMP_OK, PS4_TEMP_OK, PS5_TEMP_OK,
  PS6_TEMP_OK, PS7_TEMP_OK, PS8_TEMP_OK, PS9_TEMP_OK, PS10_TEMP_OK
};


float absBulkCur;

//=============================================================================
//  Functions referenced in the Modbus register definitions to handle writes to
//  specific global values defined as ModBus registers
//=============================================================================

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for "cmdBitsAppl"
//
//  Check the bits that where set and trigger the corresponding events as
//  appropriate (but do NOT save the new value - just use it to trigger
//  events)
//-----------------------------------------------------------------------------
bool _nearcall MB_cmdBitsAppl(SetRegFuncParams)
{
  uint  ui;
  ulong  mask, cmdBits = MB_VAL(ulong);


  if (cmdBits & CBA_TURN_ON)    onRequest();
  if (cmdBits & CBA_TURN_OFF)   offRequest();
  if (cmdBits & CBA_CLR_INTLK)  resetRequest();

  mask = CBA_ENABLE_1;
  for (ui=0; ui<10; ++ui,mask<<=1)  if (cmdBits & mask)  enableBulkSupply(ui);

  mask = CBA_DISABLE_1;
  for (ui=0; ui<10; ++ui,mask<<=1)  if (cmdBits & mask)  disableBulkSupply(ui);

  return true;
}


//-----------------------------------------------------------------------------
//  External ModBus client sent a new setpoint
//
//  Check the new setting and alter output values as appropriate
//-----------------------------------------------------------------------------
bool MB_ampSet(SetRegFuncParams)
{
  newSetpoint(MB_VAL(float));

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_vlimPlus(SetRegFuncParams)
{
  float  limit, fval = MB_VAL(float);

//--- make sure value is achievable ---
  limit = SetMin(AO_VLIM_POS);  limit = FABS(limit);
  if (fval < 0.0)  fval = 0.0;
  if (fval > limit)  fval = limit;

  KLOCK();  vlimPlus = fval;  KUNLOCK();

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_vlimMinus(SetRegFuncParams)
{
  float  limit, fval = MB_VAL(float);

//--- make sure value is achievable ---
  limit = SetMax(AO_VLIM_NEG);  limit = FABS(limit);
  if (fval < 0.0)  fval = 0.0;
  if (fval > limit)  fval = limit;

  KLOCK();  vlimMinus = fval;  KUNLOCK();

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for xxx
//-----------------------------------------------------------------------------
bool MB_rampRate(SetRegFuncParams)
{
  rampRate = MB_VAL(float);

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for the VCTRL Hysterisis
//-----------------------------------------------------------------------------
bool MB_vctrlHyst(SetRegFuncParams)
{
  float fval = MB_VAL(float);

  fval = FABS(fval);

  if (fval > 0.50)  fval = 0.50;

  vctrlHyst = fval;

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for the supplyType
//-----------------------------------------------------------------------------
bool MB_supplyType(SetRegFuncParams)
{
  uint newType = MB_VAL(uint);

  if ((newType != TYPE_FET) AND (newType != TYPE_SCR))  newType = 0;

  supplyType = newType;

  return true;
}

//-----------------------------------------------------------------------------
//  External ModBus client sent a new value for the loadType
//-----------------------------------------------------------------------------
bool MB_loadType(SetRegFuncParams)
{
  uint newType = MB_VAL(uint);

  if ((newType != LOAD_RL) AND (newType != LOAD_DC))  newType = 0;

  loadType = newType;

  return true;
}

//-----------------------------------------------------------------------------




//=============================================================================
//  functions to handle restoration of saved "persistent" values
//=============================================================================

//-----------------------------------------------------------------------------
//  restore an ampSet value from the settings file
//-----------------------------------------------------------------------------
RESTORE_FUNC(ampSet)
{
  newSetpoint(NEW_VAL(float));
}


//-----------------------------------------------------------------------------
//  restore the enabled/disabled settings for each bulk supply
//-----------------------------------------------------------------------------
RESTORE_FUNC(enabledMask)
{
  uint  u, bit, mask;

  mask = NEW_VAL(uint);

  for (u=0,bit=1; u<NumPS; ++u,bit<<=1)
    if (mask & bit)  enableBulkSupply(u);  else  disableBulkSupply(u);
}


//-----------------------------------------------------------------------------
#define SharedFuncs
#include <sharedAppl.c>



//=============================================================================
//  Misc Application-specific functions
//=============================================================================

//-----------------------------------------------------------------------------
//  Return a # indicating the state of the FETs based on the passed digitalOut
//  value.
//
//  The return value is 0, 1, 2, or 3 with the following meaning:
//   0 - both FET drivers off
//   1 - negative FET driver on
//   2 - positive FET driver on
//   3 - both FET drivers on (an ERROR!)
//-----------------------------------------------------------------------------
uint bitStateFETS(ulong outBits)
{
  uint  state = 0;

  if (outBits & ENABLE_POL_NEG)  state = 1;
  if (outBits & ENABLE_POL_POS)  state |= 2;

  return state;
}




//=============================================================================
//  Task functions to handle application-specific commands
//=============================================================================

typedef struct LCDElem_  LCDElem;

typedef char * (*LCDFmtFunc)(LCDElem far *elem, void far *val, char *buf);
typedef bool   (*LCDChgFunc)(LCDElem far *elem, void far *val, float chgBy);
typedef void   (*LCDSetFunc)(LCDElem far *elem, void far *val);

typedef struct LCDElem_ {
  uchar        type;
  uchar        col, row;
  float        chgRate;
  void far    *val;
  int          valSize;
  long         p1, p2;
  LCDFmtFunc   fmtFunc;
  LCDChgFunc   chgFunc;
  LCDSetFunc   setFunc;
};

//--- values for LCDElem.type ---
#define  RO_DATA   0
#define  RW_DATA   1
#define  PAGE_REF  2


typedef struct {
  const char far   *name;      // name of the page
  const char far  **text;      // static text for page
  LCDElem far      *elems;     // elems that make up the page
  LCDElem far      *cur;       // ptr to current element
} PageDef;



#define LCD_ELEM_VAL(type)  ( *((type far *)elem->val) )

#define LCD_VAL(type)  ( *((type far *)val) )

#define VAR(name)   VFPV(name), sizeof(name)

#define ENABLED(idx)  VAR(Next.bs[idx-1].enabled)


#define LCD_FMT_FUNC(name)  \
  _nearcall char *name(LCDElem far *elem, void far *val, char *buf)

#define LCD_CHG_FUNC(name)  \
  _nearcall bool  name(LCDElem far *elem, void far *val, float chgBy)

#define LCD_SET_FUNC(name)  \
  _nearcall void  name(LCDElem far *elem, void far *val)

#define LCD_SET_REG_FUNC(funcname, reg, type)  \
  _nearcall void  funcname(LCDElem far *elem, void far *val)  { \
    type newVal = LCD_VAL(type);  \
    char  temp[64];  \
    printf("new value for %ls: %s\r\n", reg##_RegInfo.name,  \
           fmtRegVal(&reg##_RegInfo, (S16 *)&newVal, 0, temp) );  \
    (reg##_RegInfo.setRegFunc)(&reg##_RegInfo, 0, &newVal, -1);  \
  }

#define  PAGE_TEXT(name)   const char far * name##Text[]

#define  PAGE_ELEMS(name)  const LCDElem far name##Elems[]

#define  PAGE_DEF(name, title)  const PageDef far name##Page = {  \
                title, name##Text, VFP(name##Elems),  VFP(name##Elems) };


bool  lcdBlink;

//-----------------------------------------------------------------------------
LCD_FMT_FUNC(strF32)  {
  return fmtFloat(LCD_VAL(float), (int)elem->p1, (int)elem->p2, buf); }

LCD_CHG_FUNC(chgF32)  {  LCD_VAL(float) += chgBy;  return true; }


LCD_SET_FUNC(savF32)  {
  KLOCK();  LCD_ELEM_VAL(float) = LCD_VAL(float);  KUNLOCK(); }

LCD_FMT_FUNC(strU16)  {
  sprintf(buf, "%*u", (int)elem->p1, LCD_VAL(uint) );  return buf; }


LCD_SET_REG_FUNC(set_ampSet, ampSet, F32)
LCD_SET_REG_FUNC(set_plim, vlimPlus, F32)
LCD_SET_REG_FUNC(set_mlim, vlimMinus, F32)


LCD_FMT_FUNC(strCtlState) {
  uint ui = LCD_VAL(uint);
  if ((ui != STATE_INTLK) OR lcdBlink)  return (char *)CtlStateDesc[ui];
  return "     ";
}

LCD_FMT_FUNC(strOutMode) {
  return (char *) OutModeDesc[LCD_VAL(uint)]; }

LCD_FMT_FUNC(fetStr) {
  return (char *) PolCmdDesc[LCD_VAL(uint)]; }

LCD_FMT_FUNC(loadTypeStr) {
  return (char *) LoadTypeDesc[LCD_VAL(uint)]; }

LCD_FMT_FUNC(supTypeStr) {
  return (char *) SupplyTypeDesc[LCD_VAL(uint)]; }

LCD_FMT_FUNC(polStr) {
  return (LCD_VAL(bool)) ? "[+]" : "[-]"; }

LCD_FMT_FUNC(strHolec) {
  return (LCD_VAL(ulong) & HOLEC_OK) ? "OK " : (lcdBlink?"BAD":"   "); }

LCD_FMT_FUNC(strIlk)  {
  return (LCD_VAL(ulong)  & NO_INTLK) ? "OK   " : (lcdBlink?"FAULT":"     "); }

LCD_FMT_FUNC(pageName)  {
  fstrcpy(buf, LCD_VAL(PageDef).name);  return buf;  }

LCD_FMT_FUNC(bitYN)  {
  return (LCD_VAL(ulong)  & elem->p1) ? "Y" : "N"; }

LCD_FMT_FUNC(bitOB)  {
  return (LCD_VAL(ulong) & elem->p1) ? "OK " : (lcdBlink?"BAD":"   "); }

LCD_FMT_FUNC(bitYN_)  {
  return (LCD_VAL(ulong)  & elem->p1) ? "YES" : "NO "; }

LCD_FMT_FUNC(enaYN)  {
  return LCD_VAL(bool) ? "YES" : "NO "; }

//========================================================================
PAGE_TEXT(main) = {
//----------------------------------------
 "SET:-####.##A  CUR:-####.##A   V:-##.###",
 " ",
 "ISET:-###.##A   +LIM:-##.###V   SFT: ###",
 "BJTB:-##.###V   -LIM:-##.###V   HDW: ###",
 "IFDK:-###.##A  VCTRL:-##.###V      #####",
 "                LOAD:########  STYP: ###",
 "HOLEC: ###  EXTILK: #####  #####  #PS:##",
//----------------------------------------
 NULL
};

PAGE_ELEMS(main) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RW_DATA,  5,1,  0.001,  VAR(ampSet),      8,2, strF32, chgF32, set_ampSet },
  { RO_DATA, 20,1,  0.0,    VAR(ampRead),     8,2, strF32,       NULL, NULL },
  { RO_DATA, 34,1,  0.0,    VAR(voltRead),    7,3, strF32,       NULL, NULL },

  { RO_DATA,  6,3,  0.0,    VAR(iset),        7,2, strF32,       NULL, NULL },
  { RO_DATA,  6,4,  0.0,    VAR(bjtBias),     7,3, strF32,       NULL, NULL },
  { RO_DATA,  6,5,  0.0,    VAR(ifdbk),       7,2, strF32,       NULL, NULL },

  { RW_DATA, 22,3,  0.0001, VAR(vlimPlus),    7,3, strF32, chgF32, set_plim },
  { RW_DATA, 22,4,  0.0001, VAR(vlimMinus),   7,3, strF32, chgF32, set_mlim },
  { RO_DATA, 22,5,  0.0,    VAR(vctrl),       7,3, strF32,       NULL, NULL },

  { RO_DATA, 38,3,  0.0,    VAR(polCmd),     -1,1, fetStr,       NULL, NULL },
  { RO_DATA, 38,4,  0.0,    VAR(Cur.regPol), -1,1, polStr,       NULL, NULL },
  { RO_DATA, 36,5,  0.0,    VAR(outMode),    -5,5, strOutMode,   NULL, NULL },

  { RO_DATA, 22,6,  0.0,    VAR(loadType),    8,8, loadTypeStr,  NULL, NULL },
  { RO_DATA, 38,6,  0.0,    VAR(supplyType),  3,3, supTypeStr,   NULL, NULL },

  { RO_DATA,  8,7,  0.0,    VAR(digitalIn),  -3,3, strHolec,     NULL, NULL },
  { RO_DATA, 21,7,  0.0,    VAR(digitalIn),  -5,5, strIlk,       NULL, NULL },
  { RO_DATA, 28,7,  0.0,    VAR(state),      -4,4, strCtlState,  NULL, NULL },
  { RO_DATA, 39,7,  0.0,    VAR(psCount),     2,0, strU16,       NULL, NULL },


  { 0,        0,0,  0.0,    NULL,              0,0, NULL, NULL, NULL }
};
PAGE_DEF(main, "MAIN");

//========================================================================
PAGE_TEXT(ps1) = {
//----------------------------------------
 "    --Bias- --Cur--  On  Temp Avail Ena ",
 "  1.-##.### -###.##  #/#  ###  ###  ### ",
 "  2.-##.### -###.##  #/#  ###  ###  ### ",
 "  3.-##.### -###.##  #/#  ###  ###  ### ",
 "  4.-##.### -###.##  #/#  ###  ###  ### ",
 "  5.-##.### -###.##  #/#  ###  ###  ### ",
 " ",
//----------------------------------------
 NULL
};
PAGE_ELEMS(ps1) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RO_DATA,  5,2,  0.0,  VAR(bsv_1),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,3,  0.0,  VAR(bsv_2),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,4,  0.0,  VAR(bsv_3),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,5,  0.0,  VAR(bsv_4),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,6,  0.0,  VAR(bsv_5),       7,3, strF32, NULL, NULL },

  { RO_DATA, 13,2,  0.0,  VAR(amps_1),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,3,  0.0,  VAR(amps_2),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,4,  0.0,  VAR(amps_3),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,5,  0.0,  VAR(amps_4),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,6,  0.0,  VAR(amps_5),      7,3, strF32, NULL, NULL },

  { RO_DATA, 22,2,  0.0,  VAR(digitalOut), PS1_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 22,3,  0.0,  VAR(digitalOut), PS2_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 22,4,  0.0,  VAR(digitalOut), PS3_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 22,5,  0.0,  VAR(digitalOut), PS4_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 22,6,  0.0,  VAR(digitalOut), PS5_SET_ON,0, bitYN, NULL, NULL },

  { RO_DATA, 24,2,  0.0,  VAR(digitalIn), PS1_IS_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,3,  0.0,  VAR(digitalIn), PS2_IS_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,4,  0.0,  VAR(digitalIn), PS3_IS_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,5,  0.0,  VAR(digitalIn), PS4_IS_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,6,  0.0,  VAR(digitalIn), PS5_IS_ON,0, bitYN, NULL, NULL },

  { RO_DATA, 27,2,  0.0,  VAR(digitalIn), PS1_TEMP_OK,0, bitOB, NULL, NULL },
  { RO_DATA, 27,3,  0.0,  VAR(digitalIn), PS2_TEMP_OK,0, bitOB, NULL, NULL },
  { RO_DATA, 27,4,  0.0,  VAR(digitalIn), PS3_TEMP_OK,0, bitOB, NULL, NULL },
  { RO_DATA, 27,5,  0.0,  VAR(digitalIn), PS4_TEMP_OK,0, bitOB, NULL, NULL },
  { RO_DATA, 27,6,  0.0,  VAR(digitalIn), PS5_TEMP_OK,0, bitOB, NULL, NULL },

  { RO_DATA, 32,2,  0.0,  VAR(digitalIn), PS1_AVAIL,0, bitYN_, NULL, NULL },
  { RO_DATA, 32,3,  0.0,  VAR(digitalIn), PS2_AVAIL,0, bitYN_, NULL, NULL },
  { RO_DATA, 32,4,  0.0,  VAR(digitalIn), PS3_AVAIL,0, bitYN_, NULL, NULL },
  { RO_DATA, 32,5,  0.0,  VAR(digitalIn), PS4_AVAIL,0, bitYN_, NULL, NULL },
  { RO_DATA, 32,6,  0.0,  VAR(digitalIn), PS5_AVAIL,0, bitYN_, NULL, NULL },

  { RO_DATA, 37,2,  0.0,  ENABLED(1),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,3,  0.0,  ENABLED(2),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,4,  0.0,  ENABLED(3),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,5,  0.0,  ENABLED(4),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,6,  0.0,  ENABLED(5),       3,0, enaYN, NULL, NULL },

  { 0,        0,0,  0.0,  NULL,             0,0, NULL, NULL, NULL }
};
PAGE_DEF(ps1, "S1-5");


//========================================================================
PAGE_TEXT(ps2) = {
//----------------------------------------
 "    --Bias- --Cur--  On  Temp Avail Ena ",
 "  6.-##.### -###.##  #/#  ###  ###  ### ",
 "  7.-##.### -###.##  #/#  ###  ###  ### ",
 "  8.-##.### -###.##  #/#  ###  ###  ### ",
 "  9.-##.### -###.##  #/#  ###  ###  ### ",
 " 10.-##.### -###.##  #/#  ###  ###  ### ",
 " ",
//----------------------------------------
 NULL
};
PAGE_ELEMS(ps2) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RO_DATA,  5,2,  0.0,  VAR(bsv_6),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,3,  0.0,  VAR(bsv_7),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,4,  0.0,  VAR(bsv_8),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,5,  0.0,  VAR(bsv_9),       7,3, strF32, NULL, NULL },
  { RO_DATA,  5,6,  0.0,  VAR(bsv_10),      7,3, strF32, NULL, NULL },

  { RO_DATA, 13,2,  0.0,  VAR(amps_6),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,3,  0.0,  VAR(amps_7),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,4,  0.0,  VAR(amps_8),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,5,  0.0,  VAR(amps_9),      7,3, strF32, NULL, NULL },
  { RO_DATA, 13,6,  0.0,  VAR(amps_10),     7,3, strF32, NULL, NULL },

  { RO_DATA, 22,2,  0.0,  VAR(digitalOut), PS6_SET_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 22,3,  0.0,  VAR(digitalOut), PS7_SET_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 22,4,  0.0,  VAR(digitalOut), PS8_SET_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 22,5,  0.0,  VAR(digitalOut), PS9_SET_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 22,6,  0.0,  VAR(digitalOut), PS10_SET_ON,0, bitYN, NULL, NULL },
  { RO_DATA, 24,2,  0.0,  VAR(digitalIn), PS6_IS_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 24,3,  0.0,  VAR(digitalIn), PS7_IS_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 24,4,  0.0,  VAR(digitalIn), PS8_IS_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 24,5,  0.0,  VAR(digitalIn), PS9_IS_ON,0,  bitYN, NULL, NULL },
  { RO_DATA, 24,6,  0.0,  VAR(digitalIn), PS10_IS_ON,0, bitYN, NULL, NULL },

  { RO_DATA, 27,2,  0.0,  VAR(digitalIn), PS6_TEMP_OK,0,  bitOB, NULL, NULL },
  { RO_DATA, 27,3,  0.0,  VAR(digitalIn), PS7_TEMP_OK,0,  bitOB, NULL, NULL },
  { RO_DATA, 27,4,  0.0,  VAR(digitalIn), PS8_TEMP_OK,0,  bitOB, NULL, NULL },
  { RO_DATA, 27,5,  0.0,  VAR(digitalIn), PS9_TEMP_OK,0,  bitOB, NULL, NULL },
  { RO_DATA, 27,6,  0.0,  VAR(digitalIn), PS10_TEMP_OK,0, bitOB, NULL, NULL },

  { RO_DATA, 32,2,  0.0,  VAR(digitalIn), PS6_AVAIL,0,  bitYN_, NULL, NULL },
  { RO_DATA, 32,3,  0.0,  VAR(digitalIn), PS7_AVAIL,0,  bitYN_, NULL, NULL },
  { RO_DATA, 32,4,  0.0,  VAR(digitalIn), PS8_AVAIL,0,  bitYN_, NULL, NULL },
  { RO_DATA, 32,5,  0.0,  VAR(digitalIn), PS9_AVAIL,0,  bitYN_, NULL, NULL },
  { RO_DATA, 32,6,  0.0,  VAR(digitalIn), PS10_AVAIL,0, bitYN_, NULL, NULL },

  { RO_DATA, 37,2,  0.0,  ENABLED(6),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,3,  0.0,  ENABLED(7),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,4,  0.0,  ENABLED(8),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,5,  0.0,  ENABLED(9),       3,0, enaYN, NULL, NULL },
  { RO_DATA, 37,6,  0.0,  ENABLED(10),      3,0, enaYN, NULL, NULL },

  { 0,        0,0,  0.0,  NULL,             0,0, NULL, NULL, NULL }
};
PAGE_DEF(ps2, "S6-10");


//========================================================================
PAGE_TEXT(config) = {
//----------------------------------------
 "Contrast: ### ",
 "Load Type: ######## ",
 "Bridge Type: ### ",
 " ",
 " ",
 " ",
 " ",
//----------------------------------------
 NULL
};

//------------------------------------------------------------------------
//--- contrast ---
LCD_CHG_FUNC(chgContr)  {
  int ii = LCD_VAL(U16);
  ii += (int)chgBy;
  if (ii < 5)  ii = 5;
  if (ii > 200)  ii = 200;
  if ((uint)ii != LCD_VAL(U16))  { LCD_VAL(U16) = ii;  return true; }
  return false;
}
LCD_SET_REG_FUNC(setContr, contrast, U16)

//--- loadType ---
LCD_CHG_FUNC(chgLTyp)  {
  int ii = LCD_VAL(U16);
  ii += (int)chgBy;
  if (ii < 0)  ii = 0;
  if (ii > 2)  ii = 2;
  if ((uint)ii != LCD_VAL(U16))  { LCD_VAL(U16) = ii;  return true; }
  return false;
}
LCD_SET_REG_FUNC(setLTyp, loadType, U16)

//--- supplyType ---
LCD_CHG_FUNC(chgSTyp)  {
  int ii = LCD_VAL(U16);
  ii += (int)chgBy;
  if (ii < 0)  ii = 0;
  if (ii > 2)  ii = 2;
  if ((uint)ii != LCD_VAL(U16))  { LCD_VAL(U16) = ii;  return true; }
  return false;
}
LCD_SET_REG_FUNC(setSTyp, supplyType, U16)

//------------------------------------------------------------------------
PAGE_ELEMS(config) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RW_DATA, 11,1,  0.005,  VAR(contrast),    3,0, strU16,  chgContr,setContr},
  { RW_DATA, 12,2,  0.001,  VAR(loadType),    8,8, loadTypeStr,chgLTyp,setLTyp},
  { RW_DATA, 14,3,  0.001,  VAR(supplyType),  3,3, supTypeStr, chgSTyp,setSTyp},

  { 0,        0,0,  0.0,    NULL,              0,0, NULL, NULL, NULL }
};
PAGE_DEF(config, "CONFIG");


//========================================================================
/*
PAGE_TEXT(pageName) = {
//----------------------------------------
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
//----------------------------------------
 NULL
};

PAGE_ELEMS(pageName) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { RW_DATA, ##,#,  #.###,  VAR(varName),    #,#,  xxxxxx,  xxxxxx, xxxxxx },

  { 0,        0,0,  0.0,    NULL,              0,0, NULL, NULL, NULL }
};
PAGE_DEF(pageName, "PAGE NAME");

*/

//========================================================================
PAGE_TEXT(list) = { NULL };

PAGE_ELEMS(list) = {
//{ type,     C,R, chgRate, *val,valSize,   p1,p2, fmtFunc, chgFunc, setFunc }
  { PAGE_REF, 1,1,  0.0,   VAR(mainPage),    0,0,  pageName,  NULL, NULL },
  { PAGE_REF, 1,2,  0.0,   VAR(ps1Page),     0,0,  pageName,  NULL, NULL },
  { PAGE_REF, 1,3,  0.0,   VAR(ps2Page),     0,0,  pageName,  NULL, NULL },
  { PAGE_REF, 1,4,  0.0,   VAR(configPage),  0,0,  pageName,  NULL, NULL },

  { 0,        0,0,  0.0,  NULL,            0,0, NULL, NULL, NULL }
};
PAGE_DEF(list, "LIST");

//========================================================================

/*
//----------------------------------------
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
 "........................................",
//----------------------------------------
*/


//-----------------------------------------------------------------------------
  #define SetNormal(fd)  { \
        write(fd, "\376\061\001\376\062\000\000\000\001\100", 10); }
  #define SetReverse(fd)  { \
        write(fd, "\376\061\002\376\062\000\000\000\001\100", 10); }
  #define ResetLCD(fd)  {  \
      write(fd, "\376\223\001\376\071\147\376\120\200\376\223\000", 12);  \
      write(fd, "\376\072\024\024\376\130\376\122", 8);  SetNormal(fd);  \
}

//-----------------------------------------------------------------------------
//  Change the current element ptr for a page to the next/prev page element
//  that can be altered
//
//  If no such elements, then the ptr is left unchanged
//
//  delta determines direction (>=0 means next, < 0 means prev)
//-----------------------------------------------------------------------------
void chgFocus(PageDef far *page, int delta)
{
  LCDElem far *elem = page->cur;

  for (;;)  {
    if (delta < 0)  --elem;  else  ++elem;
    if ((elem < page->elems) OR !elem->col)  return;
    if (elem->type != RO_DATA)  { page->cur = elem;  return; }
  }
}


//-----------------------------------------------------------------------------
//  display the static text for an LCD display page
//-----------------------------------------------------------------------------
void updateLCD(fd_t fd, PageDef far *page, void far *modVal)
{
  uchar  i;
  bool  hasFocus;
  char far **ptr;
  char  *str;
  void far  *val;
  LCDElem far *elem;
  char  temp[50];
  static PageDef far *lastPage = NULL;


  SetNormal(fd);

//--- clear screen and draw static text if page changed ---
  if (page != lastPage)  {
  //--- Home + CLS, disable scrolling, set Normal font ---
    write(fd, "\376\130\376\122", 4);  SetNormal(fd);

    temp[41] = 0;   ptr = (char far **) page->text;
    for (i=1; *ptr; ++i,++ptr)  {
      sprintf(temp, "\376\107\001%c%.41ls", i, *ptr);
      write(fd, temp, strlen(temp));
    }

    lastPage = page;
  }

//--- display the data and link elements ---
  for (elem=page->elems; elem->col; ++elem)  {
    hasFocus = (elem == page->cur);
    if (elem->type == RO_DATA)  hasFocus = false;

    val = elem->val;
    if (hasFocus)  { SetReverse(fd);  if (modVal)  val = modVal;  }

    sprintf(temp, "\376\107%c%c", elem->col, elem->row);  // set cursor pos
    str = elem->fmtFunc(elem, val, temp+4);
    if (str != temp+4)  strncpy(temp+4, str, (sizeof temp)-5);
    temp[(sizeof temp)-1] = 0;  write(fd, temp, strlen(temp));

    if (hasFocus)  SetNormal(fd);
  }

}


//-----------------------------------------------------------------------------
//  Process to manage the front panel LCD display and monitor front panel
//  controls
//-----------------------------------------------------------------------------
void _task lcdUI(int argc, char **argv)
{
#pragma warn -775

  fd_t  lcdfd;
  int  delta, knobCount, newCount;
  uint  flowChars, btnBits, prevBtnBits, curContrast, prevContrast;
  bool  updateDisplay;
  ulong  baudRate, lastKnobTime, lastUpdateTime;
  float  chg;
  char  contrByte;
  void far  *modVal = NULL;
  LCDElem far  *elem;
  PageDef far  *page;
  char  *lcdDev;
  char  valBuf[64];


  redirectStdio();


//--- open port for LCD connection in Xon/Xoff mode ---
  lcdDev = (hdwType == HT_MODULAR_PS) ? "/dev/ttyx2" : "/dev/ttyx3";
  if ((lcdfd = open(lcdDev, O_RDWR, 0)) < 0)  {
    printf("Unable to open port for LCD\r\n");  Exit(-3); }

//--- set baud rate and Xon/Xoff chars ---
  //finish - use std termio interface instead of non-std ioctl values
  baudRate = 19200;  ioctl(lcdfd, IOC_SET_BAUD, &baudRate);
  flowChars = 0xFFFE;  ioctl(lcdfd, IOC_SET_FLOW_CHARS, &flowChars);

//--- Initialize LCD: ---
//---   Enable Flow Ctl, Xoff when 20 unused bytes, Xon when 20 unread bytes ---
//---   Home + CLS, disable scrolling
  write(lcdfd, "\376\072\024\024\376\130\376\122", 8);  SetNormal(lcdfd);

  page = (PageDef far *) &mainPage;  updateDisplay = true;

//--- get initial values ---
  prevContrast = 0;
  KLOCK();
  btnBits = Cur.btnBits;  curContrast = contrast;
  KUNLOCK();
  _EnterCritical_();  knobCount = qdCount = 0;  _ExitCritical_();

  lastUpdateTime = GetMS();  updateDisplay = true;

  for (;;)  {
    if (! updateDisplay)  {
      sleepMS(100);
      if (msSince(lastUpdateTime) >= 500)  updateDisplay = true;
    }

  //--- update display periodically or because of user action ---
    if (updateDisplay)  {
      lcdBlink = !lcdBlink;
      updateLCD(lcdfd, page, modVal);  updateDisplay = false;
      lastUpdateTime = GetMS();
    }

  //--- get last recorded state of buttons ---
    prevBtnBits = btnBits;
    KLOCK();
    btnBits = Cur.btnBits;  curContrast = contrast;
    KUNLOCK();

  //--- get # of knob ticks since last pass and reset count ---
    _EnterCritical_();  newCount = qdCount;  qdCount = 0;  _ExitCritical_();


    if (newCount)  {
      lastKnobTime = GetMS();  knobCount += newCount;  }
    else
      if (msSince(lastKnobTime) > 500)  knobCount = 0;

  //--- if knob turned while no buttons are being pressed, move the ---
  //--- highlight to the next or prev selectable element ---
    if (!btnBits AND (abs(knobCount) > knobSelectRate))  {
      delta = (knobCount < 0) ? -1 : 1;
      knobCount -= knobSelectRate * delta;
      chgFocus(page, delta);  updateDisplay = true;
    }

  //--- SELECT button:
  //---   if focus is on page ref, then switch to selected page
  //---   otherwise, switch to page with list of pages
    if ((btnBits == SELECT_BTN) AND (prevBtnBits != SELECT_BTN))  {
      elem = page->cur;
      if (elem->type != PAGE_REF)
        page = (PageDef far *) &listPage;
      else
        page = (PageDef far *) elem->val;
      updateDisplay = true;
    } while (0);


  //--- SET button (with focus on RW element):
  //---   if just pressed, make a copy of the cur elem's value
  //---   while pressed, apply knob movements to copy of cur's value
    if (btnBits == SET_BTN)  do  {
      if ((elem = page->cur)->type != RW_DATA)  break;
      if (!modVal)  {  // copy existing value if just pressed
        fmemcpy(valBuf, elem->val, elem->valSize);  modVal = valBuf;  }
      if (!knobCount)  break;
      chg = (float) knobCount;
      chg = SIGN(knobCount) * chg * chg * elem->chgRate;
      if (elem->chgFunc)  if (elem->chgFunc(elem, modVal, chg))  {
        updateDisplay = true;  knobCount = 0;  }
    } while (0);
    else  if (modVal)  {
      if (elem->setFunc)  elem->setFunc(elem, modVal);
//      printf("done modifying\r\n");
      modVal = NULL;  knobCount = 0;
    }


  //--- if contrast changed, update the display's contrast ---
    if (curContrast != prevContrast)  {
      write(lcdfd, "\376\120", 2);  contrByte = contrast;
      write(lcdfd, &contrByte, 1);
      prevContrast = curContrast;
    }


  //--- if SELECT and SET buttons both pressed at once, then reinitialize ---
  //--- the display (if manual override jumper was in when it was powered ---
  //--- up, this should force it to remember the 19200 baud rate when it ---
  //--- is next powered up without the jumper ---
    if (btnBits == (SELECT_BTN | SET_BTN))   {
      ResetLCD(lcdfd);  updateDisplay = true;  }

#pragma warn 775
  }

}


//-----------------------------------------------------------------------------
//  example func to process a command entered via a user shell
//-----------------------------------------------------------------------------
/*
void _task miscTest(int argc, char **argv)
{
  int  idx;
  bool  validCmd;

//used by EVENT macro
  fd_t  eventsfd = -1;
  EventMsg  eventMsg;


  validCmd = false;

  for (;;)  {
    if (argc < 3)  break;
    if (! isdigit(argv[1][0]))  break;

    idx = atoi(argv[1]);
    if ((idx < 1) OR (idx > NumChan))  {
      printf("Invalid chan #\n");  Exit(-2); }

    --idx;  validCmd = true;

    if (StriEq(argv[2], "x"))     { EVENT(MyEventX, idx, NULL, 0);  break; }
    if (StriEq(argv[2], "y"))     { EVENT(MyEventY, idx, NULL, 0);  break; }

    validCmd = false;  break;
  }

  if (! validCmd)  { printf("Usage:  misc # x|y\r\n");  Exit(-1); }

  Exit(0);
}
*/

//=============================================================================
//  Generic structure to handle application-specific presentation of the
//  digital and analog input/output values.  Code to handle additional
//  formatting that cannot be handled in this generic manner should be
//  placed in the showState() task routine.
//=============================================================================


//-----------------------------------------------------------------------------
_nearcall char *fmtState(DispElem far *elem, char *buf)
{
  return (char *) CtlStateDesc[ ELEM_VAL(uint) ];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtMode(DispElem far *elem, char *buf)
{
  return (char *) OutModeDesc[ ELEM_VAL(uint) ];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtFET(DispElem far *elem, char *buf)
{
  return (char *) PolCmdDesc[ELEM_VAL(uint)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtPol(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? "[+]" : "[-]");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtBSFault(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? "*" : "_");
}

//-----------------------------------------------------------------------------
_nearcall char *fmtSupType(DispElem far *elem, char *buf)
{
  return (char *) SupplyTypeDesc[ELEM_VAL(uint)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtLoadType(DispElem far *elem, char *buf)
{
  return (char *) LoadTypeDesc[ELEM_VAL(uint)];
}

//-----------------------------------------------------------------------------
_nearcall char *fmtEnabled(DispElem far *elem, char *buf)
{
  return (ELEM_VAL(bool) ? BlueText"YES" : RedText"NO " );
}

//-----------------------------------------------------------------------------
/*
_nearcall char *fmtBridgeState(DispElem far *elem, char *buf)
{
  return  (char *) BridgeStateDesc[ELEM_VAL(uint)];
}
*/

//-----------------------------------------------------------------------------
  const char far DevStateStr[] = {
" Set:-####.##A   Read:-####.##A       --Bias-  --Cur-- On   Temp Avail Enabled\r\n"
"                                   1. -##.### -###.##  #/#   ###  ###    ###\r\n"
"Volts:-##.###V   Err: -####.##A    2. -##.### -###.##  #/#   ###  ###    ###\r\n"
"                                   3. -##.### -###.##  #/#   ###  ###    ###\r\n"
"ISET:-####.##A   Sft: ###          4. -##.### -###.##  #/#   ###  ###    ###\r\n"
"BJTB: -##.###V   Hdw: ###          5. -##.### -###.##  #/#   ###  ###    ###\r\n"
"IFDK:-####.##A   #####/#####       6. -##.### -###.##  #/#   ###  ###    ###\r\n"
"                                   7. -##.### -###.##  #/#   ###  ###    ###\r\n"
"VCTRL:-##.###V  +LIM:-##.##V       8. -##.### -###.##  #/#   ###  ###    ###\r\n"
"#PS: ##         -LIM:-##.##V       9. -##.### -###.##  #/#   ###  ###    ###\r\n"
"                                  10. -##.### -###.##  #/#   ###  ###    ###\r\n"
"HOLEC: ###   ExtILK: ##### \r\n"
"                                     Nominal: -###.##  (TTL: -####.##) \r\n"
"SupplyType: ###  Load: ########      Max Diff:-###.## \r\n"
"\r\n"
"Apparent Resistance:\r\n"
"  Load:    -##.### ohms\r\n"
"  Dump Res:-##.### ohms\r\n"
};

const DispElem  far dispElems[] = {
  {  5, 3, VFPV(ampSet),      DARK_BLUE, fmtF32,    8,2 },
  { 22, 3, VFPV(ampRead),     DARK_BLUE, fmtF32,    8,2 },

  {  6, 5, VFPV(voltRead),    DARK_BLUE, fmtF32,    7,3 },
  { 22, 5, VFPV(ampErr),      DARK_BLUE, fmtF32,    8,2 },

  {  5, 7, VFPV(iset),        DARK_BLUE, fmtF32,    8,2 },
  {  6, 8, VFPV(bjtBias),     DARK_BLUE, fmtF32,    7,3 },
  {  5, 9, VFPV(ifdbk),       DARK_BLUE, fmtF32,    8,2 },

  { 22, 7, VFPV(polCmd),      DARK_BLUE, fmtFET,    3,3 },
  { 22, 8, VFPV(Cur.regPol),  DARK_BLUE, fmtPol,    3,3 },
  { 17, 9, VFPV(state),       DARK_BLUE, fmtState, -5,5 },
  { 23, 9, VFPV(outMode),     DARK_BLUE, fmtMode,  -5,5 },

  {  6,11, VFPV(vctrl),       DARK_BLUE, fmtF32,    7,3 },
  {  5,12, VFPV(psCount),     DARK_BLUE, fmtU16,    2,2 },

  { 21,11, VFPV(vlimPlus),    DARK_BLUE, fmtF32,    6,2 },
  { 21,12, VFPV(vlimMinus),   DARK_BLUE, fmtF32,    6,2 },

  { 12,16, VFPV(supplyType),  DARK_BLUE, fmtSupType,  3,3 },
  { 23,16, VFPV(loadType),    DARK_BLUE, fmtLoadType, 8,8 },

  { 38, 4, VFPV(bsv_1),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 5, VFPV(bsv_2),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 6, VFPV(bsv_3),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 7, VFPV(bsv_4),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 8, VFPV(bsv_5),       DARK_BLUE, fmtF32,    7,3 },
  { 38, 9, VFPV(bsv_6),       DARK_BLUE, fmtF32,    7,3 },
  { 38,10, VFPV(bsv_7),       DARK_BLUE, fmtF32,    7,3 },
  { 38,11, VFPV(bsv_8),       DARK_BLUE, fmtF32,    7,3 },
  { 38,12, VFPV(bsv_9),       DARK_BLUE, fmtF32,    7,3 },
  { 38,13, VFPV(bsv_10),      DARK_BLUE, fmtF32,    7,3 },

  { 45, 4, VFPV(amps_1),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 5, VFPV(amps_2),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 6, VFPV(amps_3),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 7, VFPV(amps_4),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 8, VFPV(amps_5),      DARK_BLUE, fmtF32,    8,2 },
  { 45, 9, VFPV(amps_6),      DARK_BLUE, fmtF32,    8,2 },
  { 45,10, VFPV(amps_7),      DARK_BLUE, fmtF32,    8,2 },
  { 45,11, VFPV(amps_8),      DARK_BLUE, fmtF32,    8,2 },
  { 45,12, VFPV(amps_9),      DARK_BLUE, fmtF32,    8,2 },
  { 45,13, VFPV(amps_10),     DARK_BLUE, fmtF32,    8,2 },

  { 73, 4, VFPV(Cur.bs[0].enabled), "", fmtEnabled, 3,3 },
  { 73, 5, VFPV(Cur.bs[1].enabled), "", fmtEnabled, 3,3 },
  { 73, 6, VFPV(Cur.bs[2].enabled), "", fmtEnabled, 3,3 },
  { 73, 7, VFPV(Cur.bs[3].enabled), "", fmtEnabled, 3,3 },
  { 73, 8, VFPV(Cur.bs[4].enabled), "", fmtEnabled, 3,3 },
  { 73, 9, VFPV(Cur.bs[5].enabled), "", fmtEnabled, 3,3 },
  { 73,10, VFPV(Cur.bs[6].enabled), "", fmtEnabled, 3,3 },
  { 73,11, VFPV(Cur.bs[7].enabled), "", fmtEnabled, 3,3 },
  { 73,12, VFPV(Cur.bs[8].enabled), "", fmtEnabled, 3,3 },
  { 73,13, VFPV(Cur.bs[9].enabled), "", fmtEnabled, 3,3 },

  { 45,15, VFPV(perSupCur),   DARK_BLUE, fmtF32, 8,2 },
  { 60,15, VFPV(ttlBulkCur),  DARK_BLUE, fmtF32, 9,2 },
  { 45,16, VFPV(maxCurDiff),  DARK_BLUE, fmtF32, 8,2 },

  { 11,19, VFPV(loadRes),     DARK_BLUE, fmtF32, 7,3 },
  { 11,20, VFPV(dumpRes),     DARK_BLUE, fmtF32, 7,3 },

  { 0,0, NULL, NULL, NULL, 0,0 }
};


const DigDispElem far digInElems[] = {
  {  7,14,  HOLEC_OK,  "BAD",DARK_RED,    "OK ",DARK_BLUE },
  { 21,14,  NO_INTLK,  "FAULT",DARK_RED,  "OK   ",DARK_BLUE },

  { 57, 4,  PS1_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 5,  PS2_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 6,  PS3_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 7,  PS4_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 8,  PS5_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57, 9,  PS6_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57,10,  PS7_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57,11,  PS8_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57,12,  PS9_IS_ON,  "N",DARK_RED,  "Y",DARK_BLUE },
  { 57,13,  PS10_IS_ON, "N",DARK_RED,  "Y",DARK_BLUE },

  { 61, 4,  PS1_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 5,  PS2_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 6,  PS3_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 7,  PS4_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 8,  PS5_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61, 9,  PS6_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61,10,  PS7_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61,11,  PS8_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61,12,  PS9_TEMP_OK,   "BAD",DARK_RED,  "OK ",DARK_BLUE },
  { 61,13,  PS10_TEMP_OK,  "BAD",DARK_RED,  "OK ",DARK_BLUE },

  { 66, 4,  PS1_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 5,  PS2_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 6,  PS3_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 7,  PS4_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 8,  PS5_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66, 9,  PS6_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66,10,  PS7_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66,11,  PS8_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66,12,  PS9_AVAIL,  "NO ",DARK_RED,  "YES",DARK_BLUE },
  { 66,13,  PS10_AVAIL, "NO ",DARK_RED,  "YES",DARK_BLUE },

  { 0,0, 0, NULL,NULL, NULL,NULL }
};

const DigDispElem far digOutElems[] = {
  { 55, 4,  PS1_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 5,  PS2_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 6,  PS3_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 7,  PS4_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 8,  PS5_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55, 9,  PS6_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55,10,  PS7_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55,11,  PS8_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55,12,  PS9_SET_ON,   "N",DARK_RED,  "Y",DARK_BLUE },
  { 55,13,  PS10_SET_ON,  "N",DARK_RED,  "Y",DARK_BLUE },

  { 0,0,  0,  NULL,NULL,  NULL,NULL }
};


const DigFmtInfo digInFmt[] = {
  { -1,-1, "",0, "",0 }
};

const DigFmtInfo digOutFmt[] = {
  { -1,-1, "",0, "",0 }
};

const AnalogFmtInfo  adcFmt[] = {
  { -1,-1,  0,0 }
};

const AnalogFmtInfo  dacFmt[] = {
  { -1,-1, 0,0 }
};

//-----------------------------------------------------------------------------
//  "state":  Show continuously updated state of the system
//-----------------------------------------------------------------------------
void _task showState(int argc, char **argv)
{
  bool  blinkOff;
  char  temp[128];


  printf(clsCmd);  blinkOff = false;

  CursorTo(0, 3);  printf(DevStateStr);

  for (;;)  {
    CursorTo(0, 0);  blinkOff = !blinkOff;

    stdHeader();  displayElems((DispElem far *)dispElems, blinkOff, temp);

    displayDigElems(digitalIn, (DigDispElem far *)digInElems, blinkOff, temp);
    displayDigElems(digitalOut, (DigDispElem far *)digOutElems, blinkOff, temp);

    CursorTo(0,22);  fflush(stdout);  CheckETX();  sleepMS(500);
  }

}


//-----------------------------------------------------------------------------
//  Turn supplies On/Off, enable/disable individual supplies, reset interlock
//  condition
//-----------------------------------------------------------------------------
void _task cmd(int argc, char **argv)
{
  bool  validCmd, enableCmd, disableCmd;
  uint  idx;
  char  *cmdStr;


  validCmd = false;  cmdStr = argv[1];

  do  {
    if (argc < 2)  break;

    if (StriEq(cmdStr, "on"))  { onRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "off"))  { offRequest();  validCmd = true;  break; }

    if (StriEq(cmdStr, "reset"))  { resetRequest();  validCmd = true;  break; }

    enableCmd = StriEq(cmdStr, "enable");
    disableCmd = StriEq(cmdStr, "disable");

    if (argc < 3)  break;

    if (enableCmd OR disableCmd)  {
      idx = ATOI(argv[2]);  if ((idx < 1) OR (idx > NumPS))  break;
      --idx;
      if (enableCmd)  enableBulkSupply(idx);  else  disableBulkSupply(idx);
      validCmd = true;  break;
    }

    if (StriEq(cmdStr, "set"))  {
      newSetpoint(atof(argv[2]));  validCmd = true;  break;  }

  }  while(0);


  if (! validCmd)  {
    printf("Usage:  cmd on|off|reset\r\n"
           "            enable|disable #\r\n"
           "            set #.#\r\n" );
    Exit(-1);
  }

  Exit(0);
}


//=============================================================================
//  Event-Handler functions called by the appl task to process event messages
//=============================================================================

//-----------------------------------------------------------------------------
//  Event-Handler for a request to ...
//-----------------------------------------------------------------------------
/*
void EVENT_MyEvent(EventMsg *event)
{
  uint  idx;
  float  setpoint;
  char  temp[16];


  idx = event->chanNum;

  if (error)  {
    fprintf(DiagLog, "error msg, chan:\n", idx+1);
    return;
  }

  setpoint = *((float *)event->data);

  fprintf(DiagLog, "New SetPoint for PS%d: %s\n",
     idx+1, fmtFloat(setpoint, 8,3, temp) );

  *((float *)((&iSet_1_RegInfo + idx)->regVal)) = setpoint;

  voltsOut[idx] = ISETtoVolts(idx, setpoint);
}
*/


//=============================================================================
//  Update the the state of the application logic in well-defined phases
//=============================================================================
// IMPORTANT NOTES and RULES:
//
//  - The code to change ANY value must be in ONE place.  NEVER spread the
//    logic for updating a single value around.  There should be ONE place
//    to look to see how a value gets set.
//
//  - "cur" values are basically "inputs" - literally (they are set based
//    on the just-read state of a digital or analog input) and/or figuratively
//    (they hold a value set by some external source, such as a cmd from a
//    remote client or another user-interface task).
//
//  - "next" values are basically "outputs" - literally (they determine the
//    next state of a digital or analog output) and/or figuratively (they
//    can affect the state of the logic for the next pass).
//
//  - The "cur" OR the "next" instance of a value may be changed - NEVER
//    both of them (with the exception of the first step below that copies
//    all of cur to prev and next to cur).
//
//  - A value MUST be in the ApplState struct if one of the following applies:
//     - The application logic is dependent on if or how much the value has
//       changed between passes.
//     - The value is used as both input (used in in a conditional and/or in
//       an expression that can affect the result of an assignment) and output
//       (the target of an assignment).
//
//  - Variables that are NOT in the ApplState struct but are USED (in
//    conditionals or other expressions) by the application logic must be
//    treated just like "cur" values.  Variables SET (the target in an
//    assignment statement) by the application logic must be treated just
//    like "next" values.
//
//  - When the logic requires comparing an "input" to a related "output", be
//    sure to use the "cur" value for the "input", and the "prev" value for
//    the "output".
//
//    This is because during the processing of the loop, the "cur" values
//    represent the input values just read (what they are "now") and the
//    "prev" values represent the values the outputs were set to at the
//    beginning of the prev pass (i.e. they are the values that "caused"
//    the inputs to have their "cur" values).
//
//    During a pass the output values in "cur" represent what the outputs
//    where set to AFTER the "cur" inputs were updated, so comparing them
//    makes no sense.  Doing so is the equivalent of comparing a past response
//    to a condition that didn't exist at the time.
//=============================================================================


//=============================================================================
//  Definitions for functions that handle setting "next" values that serve
//  as inputs to the application control logic (run by the "applic" task in
//  a periodically executed loop).
//
//  While care must be taken in these functions to use locks to avoid the
//  possibility of being preempted in the middle of an operation, such locks
//  must be kept to as short a time as possible to avoid excessively latency
//  in the main control loop.
//
//  For REALLY short locks, using a critical section is best, because this
//  has the least overhead (but it has to be REALLY short to avoid causing
//  problems for interrupt handlers).
//
//  For longer locks (ones that involve anything but the most trivial
//  conditional logic), use either KLOCK() and KUNLOCK() (which disable the
//  scheduler) or a more specific global semaphore/mutex lock.
//=============================================================================

//-----------------------------------------------------------------------------
void onRequest(void)
{
  _EnterCritical_();
  Next.onRequest = true;  Next.offRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "ON request\n");
}

//-----------------------------------------------------------------------------
void offRequest(void)
{
  _EnterCritical_();
  Next.offRequest = true;  Next.onRequest = false;
  _ExitCritical_();

  fprintf(DiagLog, "OFF request\n");
}

//-----------------------------------------------------------------------------
void resetRequest(void)
{
  _EnterCritical_();  Next.resetRequest = true;  _ExitCritical_();

  fprintf(DiagLog, "RESET request\n");
}

//-----------------------------------------------------------------------------
void enableBulkSupply(uint idx)
{
  _EnterCritical_();  Next.bs[idx].enabled = true;  _ExitCritical_();

  enabledMask |= (uint)1 << idx;

  fprintf(DiagLog, "Enable Bulk Supply #%d\n", idx+1);
}

//-----------------------------------------------------------------------------
void disableBulkSupply(uint idx)
{
  _EnterCritical_();  Next.bs[idx].enabled = false;  _ExitCritical_();

  enabledMask &= ~((uint)1 << idx);

  fprintf(DiagLog, "Disable Bulk Supply #%d\n", idx+1);
}

//-----------------------------------------------------------------------------
void newSetpoint(float newAmps)
{
  char  temp[16];


  KLOCK();

  if (FABS(newAmps) > maxAmpSet)  newAmps = (float)SIGN(newAmps) * maxAmpSet;

  Next.ampSet = newAmps;

  KUNLOCK();

  fprintf(DiagLog, "New Setpoint: %s\n", fmtFloat(newAmps, 10,3, temp) );
}


int  bitCounts[32];

//-----------------------------------------------------------------------------
//  Minimize the effect of noise on the digitial inputs in trade for some
//  added latency in response time to actual "legitimate" changes in state.
//
//  What this func does is add or subtract the periodic timer interval value
//  to/from a running total for each input bit, with a cap on the total being
//  the configured # of ms for debouncing.
//
//  The current state of each input is determined by whether the running total
//  for each input is positive or negative.  A positive value indicates a
//  predominately ON state, while a negative value indicates a predominiately
//  OFF state.
//
//  Inputs that are mostly stable (anomolies not very frequent) will quickly
//  reach the positive or negative limit and stay close to it until the "real"
//  state changes to the opposite state.  When that happens, the total will
//  shrink, cross 0, and grow toward the opposite limit.
//
//  NOTE the latency effect of this approach on the two extremes:
//    - No noise:  No errors, but Max latency (== the debounce period)
//    - Max noise:  Min latency, but constant errors.
//
//  NOTE that debounce period also effectively sets a limit on how short a
//  period that a change in state must last for it to be recognized:  Any
//  change that lasts for less than twice the debounceMS setting will
//  generally not be detected.
//
//  The setting for the debounce period effectively allows you to ajust the
//  inherent compromise between delays introduced for the correct readings vs
//  the probability of getting a false reading.  A setting of 0 effectively
//  disables the filter, while larger values reduce the effect of any noise.
//
//  This approach behaves aproximately like connecting a capacitor across a
//  noisy input.  As long as the noise level is low enough, the capacitor
//  will tend to charge up to the level of the ON voltage, or dischage to
//  0.  Noise will increase or decrease the voltage, but so long as it is
//  less frequent than the "real" value, the voltage across the capacitor
//  will tend to stay near the voltage level for the "real" state.
//-----------------------------------------------------------------------------
void debounceDigIn(void)
{
  int  i;
  int  *bitCount;
  ulong  mask, digIn;


  bitCount = bitCounts;  digIn = 0;  mask = 1;

  for (i=0; i<32;  ++i)  {
    if (digitalIn & mask)
      *bitCount += (int)SchedMS;
    else
      *bitCount -= (int)SchedMS;

    if (*bitCount >  (int)debounceMS)  *bitCount =  (int)debounceMS;
    if (*bitCount < -(int)debounceMS)  *bitCount = -(int)debounceMS;

    if (*bitCount > 0)  digIn |= mask;

    ++bitCount;  mask <<= 1;
  }

  digitalIn = digIn;
}


//-----------------------------------------------------------------------------
//  Check the state of the bulk supplies.
//-----------------------------------------------------------------------------
void checkBulkSupplies(void)
{
  uint  ui;
  float  absCur, absCurDiff, maxReadableCur;
  uint  *shortStarted;
  BulkSupplyState  *bsPrev, *bsCur;
  char  temp[32];


// NOTE:  Due to the choice of ADCs used for monitoring the bulk supplies and
//        the time required to get just a single value from them, the voltage
//        and current readings for the bulk supplies are updated only once
//        every 11 passes through the main application loop, which happens
//        every SchedMS milliseconds.  At the (typical) setting of 20ms, this
//        means it takes 220ms to update the current and voltage readings for
//        all the supplies (s sample rate of just over 4Hz)!

//        In contrast, the IFDBK, VOUT, VTCRL, and BJT_BIAS values are read
//        every pass (a sample rate of 50Hz when SchedMS = 20ms).


//finish:  We could reduce the overhead of these checks by taking in to
//         account the fact that we have new values only every 220ms.
//         To spread the (reduced) load out and avoid large jumps in the
//         time the appl tasks takes, we need to know WHICH inputs we have
//         new readings for, so I will need to have the hardware layer
//         provide this information (e.g. a bitmask that the hardware layer
//         sets bits in for each input it reads, and then applications like
//         this one could clear those bits when they have processed the
//         latest value).
//
// NOTE:  The inital reads to get the state of the inputs doesn't actually
//        work in this case, as that first call will only read 2 of the 22
//        bulk supply inputs (not even that, really, since the first call
//        will convert whatever channel the chip was sampling - if any -
//        BEFORE the first call, which could be the wrong channel (or nothing).
//
//        Note that, while we pass the readAnalogInputs a mask "telling" it
//        what inputs we want read, it generally isn't practical for the
//        hardware to read ONLY the specified inputs because you have to
//        tell the ADC's LAST time what channel you want to read the value
//        for THIS time.  So trying to honor the bitmask during the step
//        where we are sending commands to and reading data from the ADCs
//        could easily take LONGER than just converting and reading all the
//        inputs in sequence.
//


//----- check state of each bulk supply for problems -----
  maxCurDiff = 0.0;  bulkSupErr = false;
  for (ui=0; ui<NumPS; ++ui)  {
    bsPrev = Prev.bs + ui;
    bsCur = Cur.bs + ui;
    shortStarted = bulkShortStarted + ui;

    bsCur->curErr = bsCur->biasErr = bsCur->shortErr = false;

  //--- skip if no supply attached or the supply is disabled ---
    if (!bsCur->isAvail OR !bsCur->enabled)  continue;

  //--- wait until readings have had time to stablize ---
    if (!bridgeStateStable)  {
      bsCur->lastTimeCurOK = bsCur->lastTimeBiasOK = GetMS();  continue; }

  //--- compare current and voltage to nominal operating values ---
    bsCur->curDiff = bsCur->current - perSupCur;  // difference from nominal value
    absCurDiff = FABS(bsCur->curDiff);
    absCur = FABS(bsCur->current);

  //--- keep track of max diff for all active supplies ---
    if (absCurDiff > maxCurDiff)  maxCurDiff = absCurDiff;

    maxReadableCur = ReadMax(HECTidx[ui]);  // max possible reading

  //--- is supply operating in expected cur and volt range? ---
    if (absCurDiff/maxReadableCur < maxBSFracCurDiff)
      bsCur->lastTimeCurOK = GetMS();
    if (bsCur->biasVolts >= minBSVolts)
      bsCur->lastTimeBiasOK = GetMS();

  //--- error if nominal values exceeded for too long ---
    bsCur->curErr = (msSince(bsCur->lastTimeCurOK) > maxErrTime);
    bsCur->biasErr = (msSince(bsCur->lastTimeBiasOK) > maxErrTime);

  //--- error if current so high there is probably a short ---
    if ((absCur < bulkShortCur) AND (absCur/maxReadableCur < 0.99))  {
      bsCur->shortErr = false;  *shortStarted = 0;  }
    else
      if (!*shortStarted)
        *shortStarted = GetMS_();
      else
        // NOTE: The bulk current value is only updated once every 11 * SchedMS
        // ms (at SchedMS = 20 ms, that is only about 4Hz!)
        if (msSince_(*shortStarted) > 700)  // covers just 3 samples...
          bsCur->shortErr = true;


    if (bsCur->shortErr OR !bsCur->tempOK)  bulkSupErr = true;

  //===== log messages about supply-specific changes of state =====
    if (bsCur->curErr != bsPrev->curErr)
      printf("Supply #%d current %s nominal range\n"
             "  Actual: %sA  Nominal: %sA\n", ui+1,
             (bsCur->curErr ? "OUTSIDE OF" : "RETURNED TO"),
             fmtFloat(bsCur->current, 6,1, temp),
             fmtFloat(perSupCur, 6,1, temp+10) );

    if (bsCur->biasErr != bsPrev->biasErr)
      printf("Supply #%d bias voltage %s\n"
             "  Actual: %sV  Min: %sV\n", ui+1,
             (bsCur->biasErr ? "TOO LOW" : "OK"),
             fmtFloat(bsCur->biasVolts, 7,3, temp),
             fmtFloat(minBSVolts, 7,3, temp+10) );

    if (bsCur->shortErr != bsPrev->shortErr)
      printf("Supply #%d appears to %s SHORTED\n"
             "  Actual: %sA  Max: %sA or %sA\n", ui+1,
             (bsCur->shortErr ? "BE" : "NO LONGER BE"),
             fmtFloat(bsCur->current, 6,1, temp),
             fmtFloat(maxReadableCur, 6,1, temp+10),
             fmtFloat(bulkShortCur, 6,1, temp+20)  );

    if (bsCur->tempOK != bsPrev->tempOK)
      printf("Supply #%d temperature fault: %s\n", ui+1,
             (bsCur->tempOK ? "CLEARED" : "ASSERTED")  );
  }
}


//=============================================================================
//  Copy the "cur" state from last time to "prev", and update "cur"
//=============================================================================
//  Just before this function is called, readings were taken of the current
//  state of the hardware, and any changes to the output states specified by
//  the last pass were applied (just after taking the readings).
//
//  This func first copies "cur" to "prev" and then "next" to "cur" which does
//  2 things for us:
//
//    - We have a copy of the controller's logical state from the last pass
//      (so we can determine what has changed)
//
//    - We have a copy of values set by other tasks indicating events/requests
//      that happend/were made since the beginning of the last pass
//
//  We then reset the values (as needed) that are set by other tasks to
//  indicate events and requests, and update the current state based on the
//  readings just taken and the requests and events communicated by other
//  tasks.
//
//  All of this is in preperation for the updateNextState() function, which
//  implements the main "decision making" portion of the application logic,
//  determining what the next state should be based on the latest conditions
//  (as determined by the updateCurState() function).
//
//=============================================================================
void updateCurState(void)
{
  uint  ui;
  float  vlimOffset, ampReadMax;
  bool  loggableStateChg, loggedStateChg;
  BulkSupplyState  *bsCur;
  char  temp[64];

#define POLARITY(val)  ( (val) > 0.0 ? Pos_Pol : Neg_Pol )


  if (current != applTask)  return;  // must only be called by "applic" task

//-----------------------------------------------------------------
// The following conditions exist at this point:
// - The inputs have just been read
// - The outputs were just set to values computed during the last pass
// - "cur" contains the state the inputs were in at the START of
//   of the previous pass, and the state the outputs were in DURING
//   the previous pass
// - "next" contains the values that the outputs were just set
//   to and any "input" values that were changed by other tasks
//------------------------------------------------------------------

//  Next.processTime = GetMS();  curTime = time(NULL);

//  msSincePrev = msSince(Cur.processTime);

  debounceDigIn();  Next.digitalIn = digitalIn;


//--- clear the offDueToReset flag for ON or RESET requests ---
  if (Next.onRequest OR Next.resetRequest)  offDueToReset = false;


// Note that a copy is done (rather than just switching pointer values) to
// insure all values are set in both the source and destination.  If we
// simply kept pointers and swapped the pointer values, then EVERY value
// would have to be updated independently by the code, which is a pain to
// program, adds overhead and code, and is a VERY easy way to introduce odd
// bugs.  Overall, copying everything here is actually faster overall,
// simpler, and much safer.

  Prev = Cur;  Cur = Next;

//--- At this point the Cur copy contains what the inputs were at the start ---
//--- of this pass, and the Prev copy contains the output values from the ---
//--- the previous pass (i.e. the output values that resulted in the input ---
//--- values in Cur) ---


//--------------------------------------------------------------------
// Next we update the rest of the "cur" values needed by the control
// logic (the ones that do NOT come from "next" values set by other
// tasks)
//
// NOTES:
//  - This portion of the logic has order-of-execution dependencies,
//    so be sure that all values referenced in an expression are ones
//    set before they are used.
//
//  - References to "prev" values are always safe since they are just
//    a copy of values from the last pass
//--------------------------------------------------------------------


//--- log msgs for bit chgs only if not starving other processes for cycles ---
  if (SchedMS - applLoopTime > 2)  logDigIOChgs(false);


//APPLCODE

  if ((supplyType != TYPE_FET) AND (supplyType != TYPE_SCR))  supplyType = 0;
  if ((loadType != LOAD_DC) AND (loadType != LOAD_RL))  loadType = 0;

  Cur.btnBits = ~(uint)extraIn & 0x001F;


//--- check for front-panel button activity ---
//--- These bits are not in the normal digitalIn so they are not covered ---
//---  by the call to logInputChgs() ---
  if (Cur.btnBits != Prev.btnBits)  {
    if (Cur.btnBits == START_BTN)  {
      Cur.onRequest = true;  printf("\nSTART btn pressed\n");  }
    if (Cur.btnBits == STOP_BTN)   {
      Cur.offRequest = true;  printf("\nSTOP btn pressed\n");  }
    if (Cur.btnBits == RESET_BTN)  {
      Cur.resetRequest = true;  printf("\nRESET intlk btn pressed\n");  }
  }

  Cur.externIntlk = !(digitalIn & NO_INTLK);
  holecOK = BOOL(digitalIn & HOLEC_OK);

//--- unscaled input voltage readings ---
  Cur.ifdbk = anaIn[AI_IFDBK];
  Cur.vctrl = -anaIn[AI_VCTRL];
  Cur.bjtBias = anaIn[AI_BJT_BIAS];

//--- unscaled output voltages applied just after the readings were taken ---
  Cur.iset = -anaOut[AO_ISET];

//--- input voltages scaled to engineering units ---
  voltRead = Cur.vout = -readVal[AI_VOUT];
  ampRead = readVal[AI_IFDBK];
  absAmpRead = FABS(ampRead);

  ampReadMax = ReadMax(AI_IFDBK);

//--- have bulk supplies provided power long enough to have stable outputs? ---
  if (Cur.state != STATE_ON)  {
    psOnAndReady = false;  psOnTime = 0; }
  else  {
    if (!psOnTime)  psOnTime = GetMS_();
    if (!psOnAndReady)
      if (msSince_(psOnTime) > minPSonTime)  psOnAndReady = true;
  }

//--- have the bridges been in the same state long enough for stable rdgs? ---
  if (Cur.polCmd != Prev.polCmd)  {
    bridgeStateStable = false;  bridgeChgdTime = GetMS_(); }
  else
    if (!bridgeStateStable)
      if (msSince_(bridgeChgdTime) > minPSonTime)  bridgeStateStable = true;

//--- keep track of if and how long the bridges have been off ---
  if (Cur.polCmd != POL_OFF)
    bridgeOffTime = 0;   // bridges are on
  else  {
    if (!bridgeOffTime)  {  // bridges just changed from ON to OFF?
      bridgeOffTime = GetMS();  bridgeJustOff = true; }
    else
      if (msSince(bridgeOffTime) >= minBridgeOffTime)  bridgeJustOff = false;
  }



//=================================================
//===== update the state of the bulk supplies =====
//=================================================

//--- update readings for each bulk supply and determine # of supplies ---
//--- that should be operating and the total current they are handling ---
  Cur.psCount = 0;  ttlBulkCur = 0.0;
  for (ui=0; ui<NumPS; ++ui)  {
    bsCur = Cur.bs + ui;

    bsCur->biasVolts = readVal[BSVidx[ui]];
    bsCur->current = readVal[HECTidx[ui]];

    bsCur->isOn = BOOL(digitalIn & IsOnBit[ui]);
    bsCur->tempOK = BOOL(digitalIn & TempOKBit[ui]);
    bsCur->isAvail = BOOL(digitalIn & AvailBit[ui]);

    if (!bsCur->isAvail)  continue;

    ++Cur.psCount;  ttlBulkCur += bsCur->current;
  }
  absBulkCur = FABS(ttlBulkCur);
  if (polCmd == POL_NEG)  ttlBulkCur = -absBulkCur;


//--- how much current should we be reading for each active bulk supply ---
  if (polCmd == POL_OFF)
    perSupCur = 0.0;
  else  {
    perSupCur = (Cur.psCount ? ampRead / Cur.psCount : 0.0);
    perSupCur = FABS(perSupCur);
    //finish: need to incr/decr by amount that flows through dump resistor
    //  When ramping, some of the supply current is going through the dump
    //  resistor, so the perSupCur is larger than the load current/#supplies.
    //  When sinking, some of the load current is going through the dump
    //  resistor, so the perSupCur is smaller then the load current/#supplies.
  }


  checkBulkSupplies();


//===========================================================
//===== update values that are not bulk-supply-specific =====
//===========================================================

//--- polarity and magnitude of error in the load current ---
  ampErr = Prev.ampApply - ampRead;
  Cur.ctlErr = Prev.iset - Cur.ifdbk;
  absErr = FABS(Cur.ctlErr);

//--- what polarity is the regulation circuit asserting? ---
//--- NOTE:  VCTRL has been inverted from schematic to make logical sense ---
  Cur.regPol = POLARITY(Cur.vctrl);

//--- determine what the polarity is (ignoring changes within a small ---
//--- window around zero) ---
//--- NOTE:  VCTRL has been inverted from schematic to make logical sense ---
  Cur.hystRegPol = Prev.hystRegPol;
  if (FABS(Cur.vctrl) > vctrlHyst)  Cur.hystRegPol = Cur.regPol;

//--- keep track of whether regulation polarity is stable or not ---
  Cur.regPolStable = Prev.regPolStable;
  if (Cur.hystRegPol != Prev.hystRegPol)   {
    Cur.regPolStable = false;  lastRegPolChgTime = GetMS();  }
  else
    if (msSince(lastRegPolChgTime) > minPolStableTime)
      Cur.regPolStable = true;


//--- are we close to one of the voltage limits? ---
  vlimOffset = (Cur.vout < 0.0) ?
      (vlimMinus - vlimAdjust + Cur.vout) : (vlimPlus - vlimAdjust - Cur.vout);
  voltageLimited = (vlimOffset < 0.1);  // includes neg values for exceeding limit


//--- Is bulk current close to the magnet current? ---
  bulkCurDiff = ttlBulkCur - ampRead;  absBulkCurDiff = FABS(bulkCurDiff);
  bulkCurClose = absBulkCurDiff < ampReadMax * 0.10;

//--- Is the load current large enough to allow us to determine with any ---
//--- accuracy what mode the load and supplies are in? ---
  lowLoadCur = (absAmpRead < sinkEndAmps);


  if (!lowLoadCur)  {
    loadRes = Cur.vout / ampRead;
    dumpRes = Cur.vout / (ttlBulkCur - ampRead);
  }

//--- If the polarity of the voltage across the load and dump resistor is ---
//--- opposite that of the current through the bridge, then the dump ---
//--- resistor (and possibly the supplies) are absorbing power from an ---
//... inductive load ---
  sinking = !bulkCurClose AND
             ( ((Cur.vout > 0.0) AND (Cur.polCmd == POL_NEG))
            OR ((Cur.vout < 0.0) AND (Cur.polCmd == POL_POS)) );

//--- If the polarity of the voltage across the load and dump resistor is ---
//--- the same as the bridge polarity (i.e. the bridge is on and we aren't ---
//--- sinking power), and the supply/bridge current is significantly ---
//--- different than the load current, then we are ramping (i.e. the load ---
//--- is absorbing power from the bulk supplies) ---
  ramping = (!sinking AND (polCmd != POL_OFF) AND !bulkCurClose);

//--- If the bulk current is close to the load current, and the load voltage ---
//--- is not at the limits, then we are in a steady regulation state ---
  regulating = bulkCurClose AND !voltageLimited;

//--- If the bridge is OFF, and the load current is not close to 0, then ---
//--- the load is dumping energy via the dump resistor ---
  dumping = ((polCmd == POL_OFF) AND !lowLoadCur);


//--- reduce applied voltage limits during sink mode ---
  vlimAdjust = sinking ? vlimReduction : 0.0;

//--- combine output state logic values in to a single index ---
  Cur.outMode = OM_UNCAL;
  do  {
    if (Cur.state != STATE_ON)  {
      Cur.outMode = OM_OFF;  break; }
    if (!psOnAndReady OR !bridgeStateStable)  {
      Cur.outMode = OM_WAIT;  break; }
    if (sinking)  {
      Cur.outMode = ((Cur.vout > 0.0) ? OM_SINK_P : OM_SINK_N);  break; }
    if (ramping)  {
      Cur.outMode = (Cur.hystRegPol ? OM_RAMP_P : OM_RAMP_N);  break; }
    if (regulating)  {
      Cur.outMode = (Cur.hystRegPol ? OM_REG_P : OM_REG_N);  break; }
    if (dumping)  {
      Cur.outMode = OM_DUMP;  break; }
  } while(0);
  if (Cur.outMode != OM_UNCAL)
    lastModeSetTime = GetMS();
  else
    if (msSince(lastModeSetTime) < maxErrTime)  Cur.outMode = Prev.outMode;


//--- check to see if the loadType appears to be wrong ---
  inductiveLoad = (loadType == LOAD_RL);
  if (sinking OR ramping OR dumping)
    loadTypeWrong = !inductiveLoad;
  else
    loadTypeWrong = false;
  if (!loadTypeWrong)
    lastLoadTypeWarnTime = GetSeconds_() - 300;
  else
    if (SecSince_(lastLoadTypeWarnTime) > 300)  {
      printf("***** loadType appears to be set wrong ***\n\n");
      lastLoadTypeWarnTime = GetSeconds_();
    }



//===== log messages about non-supply-specific changes of state =====

//--- limit unwanted "noise" messages in the log ---
  if ((Cur.ampSet != Prev.ampSet) OR (Cur.ampApply != Prev.ampApply)
      OR (Cur.state != Prev.state) OR (SecSince_(lastLoggableStateChg) > 120)) {
    numLoggedChgMsgs = 0;  logStateChgMsgs = true;
    lastEnabledChgMsg = GetMS_();
  }
  else  if (logStateChgMsgs AND (numLoggedChgMsgs > 12)
            AND (msSince_(lastEnabledChgMsg) > 1000))  {
    logStateChgMsgs = false;
    printf("--- Polarity and Mode change msgs temporarily disabled ---\n");
  }


  loggableStateChg = loggedStateChg = false;

  if (Cur.state == STATE_ON)  {
    if (Cur.hystRegPol != Prev.hystRegPol)  {
      loggableStateChg = true;
      if (logStateChgMsgs)
        printf("Regulation polarity changed to %s\n",
               Cur.hystRegPol ? "POS" : "NEG" );
    }
    else  if (Cur.regPolStable != Prev.regPolStable)  {
      loggableStateChg = true;
      if (logStateChgMsgs)
        printf("Regulation polarity %sSTABLE\n", Cur.regPolStable ? "" : "UN");
    }
  }

  if (Cur.polCmd != Prev.polCmd)  {
    loggableStateChg = true;
    if (logStateChgMsgs)  printf("Polarity cmds: %s\n", PolCmdDesc[Cur.polCmd]);
  }

  if (Cur.outMode != Prev.outMode)  {
    loggableStateChg = true;
    if (logStateChgMsgs)  printf("Output Mode: %s\n", OutModeDesc[Cur.outMode]);
  }

  if (loggableStateChg)  {
    if (logStateChgMsgs)  { ++numLoggedChgMsgs;  loggedStateChg = true; }
    lastLoggableStateChg = GetSeconds_();
  }


  if (Cur.state != Prev.state)
    printf("Run state: %s\n", CtlStateDesc[Cur.state]);


  if (offDueToReset AND !loggedResetIntlk)  {
    printf("*** Supplies are OFF due to ctlr RESET ***\n");
    loggedResetIntlk = true;
  }


  logInterval = (Cur.state != STATE_ON) ? 3600
      : ((sinking OR ramping OR dumping) ? 20 : 300);

  if ( loggedStateChg OR (SecSince_(lastValLogTime) > logInterval) )  {
    if (!loggedStateChg)  printf("Current state:\n");
    printf("  ampApply:%s   ampRead:%s   ttlBulkCur:%s\n"
           "  bjtBias:%s   vctrl:%s   vout:%s   \n"
           "  polCmd: %s   outMode: %s\n\n",
           fmtFloat(ampApply, 7,2, temp),
           fmtFloat(ampRead, 7,2, temp+10),
           fmtFloat(ttlBulkCur, 7,2,  temp+20),
           fmtFloat(Cur.bjtBias, 6,2, temp+30),
           fmtFloat(Cur.vctrl, 6,2, temp+40),
           fmtFloat(Cur.vout, 6,2, temp+50),
           PolCmdDesc[Cur.polCmd], OutModeDesc[Cur.outMode]
          );
    lastValLogTime = GetSeconds_();
  }
}


//=============================================================================
//  Now we update the values in "next" based on the "cur" and "prev"
//  values.
//=============================================================================
//  All values changed by updateNextState() are "outputs" from this func.
//  All other values referenced in this function are "inputs" to this func.
//
//  All the values in "next" provide information about the next logical state
//  for the controller:  Values to be applied during the next pass of the
//  control logic.  All "next" values fall in to one of 2 categories:
//
//    - They are set/updated ONLY by this function (which is executed only
//      by the "applic" task)
//
//    - They are set ONLY by other tasks (within a lock), and possibly
//      reset/cleared by the "applic" task in updateCurState() (after copying
//      "next" to "cur")
//
//
//  All values that act as input to this func must fall in to one of these
//  categories:
//
//    - "prev" values
//
//    - "cur" values updated directly by statements in updateCurState()
//
//    - "cur" values updated indirectly by updateCurState() due to the copy
//      of "next" to "cur", which includes the values in "next" that are
//      set by other tasks
//
//    - a shared/global value that is the equivalent of a "cur" value but
//      for which we don't need to track changes
//
//
//  All "outputs" from this function must be altered ONLY within this function.
//
//  All "inputs" to this function must NEVER be altered within this function.
//
//  Each output value (or VERY small "set" of ouput values) must be updated
//  in a single, self-contained block of code.  NO output may be referenced
//  OR altered in ANOTHER block of code:  This is a declarative style of
//  programming and has multiple advantages:
//
//     - There are no order-of-execution dependencies across nodes/blocks
//       within a pass.  They can be executed in any order and the result
//       will be the same (get used to it - this method will be the norm as
//       computer languages and hardware add support for atomic transactions
//       and computers evolve in to neural-net-like processor arrays - FPGAs
//       are just the beginning of this trend).
//
//     - Encapsulating all logic for an output within a single node/block
//       leaves just one place to start when tracking a problem backward from
//       the result (far simpler than guessing at the possible causes and
//       trying to work forward, looking for paths to the observed result).
//
//     - Drastically reduces the complexity of the code by greatly reducing
//       the number of possible paths through the code.  This simplifies
//       testing and verification processes.
//
//  Each node/block-of-code should be prefaced with the following information:
//   - The names of the values it modifies
//   - A descr of the meaning of the values and the logic used to set them
//   - A list of the values that serve as "inputs" to the node
//
//=============================================================================
void updateNextState(void)
{
  bool  chgPolarity;
  uint  ui;
  BulkSupplyState  *bsCur, *bsNext;
  float  ampsPerPass, diff;


  if (current != applTask)  return;  // must only be called by "applic" task

//--------------------------------------------------------------------
// Next.state
// Next.onRequest, Next.offRequest, Next.resetRequest (resets)
//    Description:
//       Determines the next state of for the application logic
//       (STATE_ON, STATE_OFF, STATE_INTLK)
//    Inputs:
//       Cur.onRequest
//       Cur.offRequest
//       Cur.resetRequest
//       Cur.externIntlk
//       Cur.bulkSup[]
//       holecOK
//       offDueToReset
//----------------------------------------------------------------------

//--- chg state from OFF to ON? ---
  if (Cur.onRequest)  do  {
    Next.onRequest = false;
    if (!supplyType)  {
      printf("Must set supplyType 1st!\n");  break;  }
    if (!loadType)  {
      printf("Must set loadType 1st!\n");  break;  }
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_ON)  { printf("Already ON\n\n");  break; }
    Next.state = STATE_ON;  printf("Turning ON\n\n");
  } while(0);

//--- chg state from ON to OFF? ---
  if (Cur.offRequest)  do {
    Next.offRequest = false;
    if ((Cur.state == STATE_INTLK) OR Cur.externIntlk)  {
      printf("Must clear Interlock first\n\n");  break; }
    if (Cur.state == STATE_OFF)  { printf("Already OFF\n\n");  break; }
    Next.state = STATE_OFF;  printf("Turning OFF\n\n");
  } while(0);

//--- chg state from INTLK to OFF? ---
  if (Cur.resetRequest)  do {
    Next.resetRequest = false;
    if (Cur.state != STATE_INTLK)  {
      printf("No Interlock to clear\n\n");  break;  }
    if (Cur.externIntlk)  {
      printf("External Interlock still exists\n\n");  break; }
    if (!holecOK)  {
      printf("HOLEC still reads BAD\n\n");  break; }
    if (bulkSupErr)  {
      printf("Bulk supply error still exists\n\n");  break; }
    Next.state = STATE_OFF;  printf("Interlock cleared\n\n");
  } while(0);

//--- chg state to INTLK due to an error or external interlock? ---
  if (Cur.externIntlk OR !holecOK OR bulkSupErr OR offDueToReset)  do {
    if (Cur.state == STATE_INTLK)  break;
    Next.state = STATE_INTLK;
  } while(0);



//-------------------------------------------------------------------
// Next.polCmd
//    Description:
//       Determines which (if either) of the polarity commands should
//       asserted.
//
//       Sink mode ends when the regulation polarity changes so it is the
//       same as the polarity of the load voltage.
//
//       To avoid an oscillation caused by the time required to change the
//       bridge polarity, we have to make the change "just BEFORE" the
//       regulation polarity changes, so it can regulate the moment it does
//       change.  "Just before" is currently determined by comparing the load
//       current to a set threshold, a value that (unfortunately) is not
//       terribly consistent.
//
//       Note that changing the bridge polarity before the regulation polarity
//       changes causes the amplitutde of the control voltage to peak.  This
//       is not a problem ONLY if the power supply output is disconnected
//       from the bridge whenever the polarity commands are in conflict with
//       the regulation polarity.  Since this also means the bridge is no
//       longer sinking any of the power from the magnet, the load voltage
//       jumps a bit - a side effect we live with to avoid the oscillation
//       that occurs if we don't change the bridge polarity early
//
//    Inputs:
//       inductiveLoad
//       Cur.ampApply
//       Prev.regPolStable
//       sinking
//       absBulkCur
//       sinkEndAmps
//       Cur.hystRegPol
//       Cur.polCmd
//       psOnAndReady
//       bridgeJustOff
//-------------------------------------------------------------------
/*
  Two different types of loads with two different requirements:

  Inductive loads:
    The current through an inductive load responds slowly to changes in the
    bulk supply currents.  If the regulation circuit response time is too
    fast in comparison the result is oscillation in the bulk supply outputs.

  Resistive/DC loads:
    The current through a resistive load responds quickly to changes in the
    bulk supply currents.  If the regulation circuit response time is too
    slow in comparison, the result is oscillation as the regulation lags
    behind the load current.

  In case you think a compromise between these two is a reasonable
  approach, there is the problem of regulation near zero.  The load is
  connected to the bulk supply outputs by a bridge, and changing the
  polarity of current through the load requires the connections in the bridge
  to be changed.  But the current bridge uses SCR's, and so the bridge must
  be turned OFF for several ms (at least) to ensure one set is turned off
  before the other set is turned on.  Since the regulation circuit is NOT
  designed to handle a step-wise transfer function like this, regulation
  too close to zero is NOT possible.

  Besides making regulation close to zero impossible, this also has the
  effect of making oscillation near zero worse for one case when the
  regulation circuit is tuned for the other case.

  A fast response time (for resistive loads) means the loop tends to go in
  to regulation WITHOUT changing polarity, so oscillation caused by crossing
  zero is not common.  But when an inductive load is connected, the load
  current doesn't respond fast enough, and the regulation circuit naturally
  oscillates.  If this happens near zero, it is a disaster.

  A slow response time (for inductive loads) means the regulation loop does
  not try to change the load current faster than it can respond, so it tends
  to be stable when connected to an inductive load.  But when connected to
  a resistive load, then the regulation circuit lags behind the changes in
  the load and (again) naturally causes oscillation.  And again, if it occurs
  near zero, it is disaster.


  The "solution" is to tune the regulation circuit response for inductive
  loads, and simply follow the active setpoint polarity for resistive loads.

  Not only does the slower response time tend to dampen the osciallation near
  zero for inductive loads, but significant changes in the load current for
  inductive loads causes the voltage across the dump resistor in parallel to
  the inductor to reach the voltage limits, which puts an upper limit on how
  much the error in the load current can affect the regulation output.

  For resistive loads, which are always in phase with the regulation circuit,
  we avoid any significant oscillation by simply forcing the bridge polarity
  to the polarity of the active setpoint.  While a regulation polarity
  opposite the active setpoint polarity means the bridge still won't conduct,
  (and therefore not "actively" regulate), we avoid the usual delay incurred
  for changing the polarity of the bridge, so regulation resumes the moment
  the regulation polarity changes back in to agreement with the active
  setpoint polarity.
*/

  if (!inductiveLoad)
    usePolarity = POLARITY(Cur.ampApply);
  else
  //--- To minimize the oscillation that can occur when the polarity of the ---
  //--- bridge is changed, we change the bridge polarity (to match that of ---
  //--- the control voltage) only when a change in the control voltage ---
  //--- polarity has been preceeded by short period of no change.  By ---
  //--- ignoring changes that are NOT preceeded by a stable period, we avoid ---
  //--- at least the worst of the oscillation that normally occurs when we ---
  //--- reach the end of a sink period (i.e. the bulk supply currents get ---
  //--- near 0 and the polarity of the control voltage changes to match the ---
  //--- polarity of the load voltage) ---
    if (Prev.regPolStable)
      usePolarity = (sinking AND (absBulkCur < sinkEndAmps)) ?
                      !Cur.hystRegPol : Cur.hystRegPol;

//--- A change in polarity has to be preceeded by a short period where the ---
//--- bridge is off (to insure the SCR's have had TIME to turn off) ---
  chgPolarity = ( ((usePolarity == Pos_Pol) AND (Cur.polCmd == POL_NEG))
               OR ((usePolarity == Neg_Pol) AND (Cur.polCmd == POL_POS)) );

#define PolCmd(pol)  ( (pol) ? POL_POS : POL_NEG )

  Next.polCmd = (!psOnAndReady OR bridgeJustOff OR chgPolarity) ?
                        POL_OFF : PolCmd(usePolarity);


//-------------------------------------------------------------------
// Next.ampApply
//    Description:
//       Determines the active setpoint to apply at the start of
//       the next pass.  If we are in the period where the polarity
//       is not stable, we set the active setpoint to the current
//       reading to minimize the spike in VCTRL when we turn the
//       bridges back on.
//    Inputs:
//       Cur.state
//       Cur.ampset
//       Cur.ampApply
//       rampRate
//-------------------------------------------------------------------
  ampsPerPass = rampRate * SchedMS * 0.001;  // amps/sec --> amps/SchedMS

  diff = Cur.ampSet - Cur.ampApply;

  do  {
    if (Cur.state != STATE_ON)  { Next.ampApply = 0;  break; }

    if (FABS(diff) < ampsPerPass) { Next.ampApply = Cur.ampSet;  break; }

    if (diff < 0.0)
      Next.ampApply -= ampsPerPass;
    else
      Next.ampApply += ampsPerPass;
  }  while (0);



//-------------------------------------------------------------------
// bsNext->setOn,
//    Description:
//       Determines which bulk supplies should have the setOn output
//       enabled at the start of the next pass
//    Inputs:
//       Cur.state
//       bsCur->isAvail
//       bsCur->enabled
//       bsCur->shortErr
//       bsCur->tempOK
//-------------------------------------------------------------------
//--- set next state for each bulk supply ---
  for (ui=0; ui<NumPS; ++ui)  {
    bsCur = Cur.bs + ui;
    bsNext = Next.bs + ui;
    bsNext->setOn =
      ((Cur.state == STATE_ON) AND bsCur->isAvail AND bsCur->enabled
       AND !bsCur->shortErr AND bsCur->tempOK);

  }

}


//=============================================================================
//  Next we update output signal values to be written to the outputs at
//  the beginning of the next pass (just after the inputs are read).
//=============================================================================
void updateOutputState(void)
{
  uint  ui;


  if (current != applTask)  return;  // must only be called by "applic" task

//--- update output values that are based on current conditions (values ---
//--- in "cur" or any values used in a conditional in the applic logic) ---

/*EXAMPLE:
  // (insured earlier that only ONE of these can be true when setting them)
  if (initialBtnPressed)  setVal[FP_METER] = rfOffSetpoint;  else
  if (finalBtnPressed)    setVal[FP_METER] = rfOnSetpoint;  else
                          setVal[FP_METER] = cavityVolts;


  SetOffApplLEDs(Next.digitalOut);

  if (limited)  SetOnLimitLED(Next.digitalOut);
  if (regErr)   SetOnErrorLED(Next.digitalOut);
*/

//--- update output values based on values that were set by the application ---
//--- logic (e.g. values in "next" or variables that are the target of an ---
//--- assignment statement in the application logic) ---

//  setVal[CTL_VOLTS] = Next.ctlVolts;  // ctl voltage to DC Servo amplifier

//  setVal[EXTERN_METER] = Next.meterValue;


  Next.digitalOut = 0;

//APPLCODE

  for (ui=0; ui<NumPS; ++ui)
    if (Next.bs[ui].setOn)  Next.digitalOut |= SetOnBit[ui];

  if (Next.polCmd == POL_POS)  Next.digitalOut |= (ENABLE_POL_POS | LED_POS);
  if (Next.polCmd == POL_NEG)  Next.digitalOut |= (ENABLE_POL_NEG | LED_NEG);

  setVal[AO_ISET] = Next.ampApply;

//  setVal[AO_VLIM_POS] = -vlimPlus + vlimAdjust;
//  setVal[AO_VLIM_NEG] = vlimMinus - vlimAdjust;
  setVal[AO_VLIM_POS] = vlimPlus - vlimAdjust;
  setVal[AO_VLIM_NEG] = -vlimMinus + vlimAdjust;
}




//=============================================================================
// The last thing we do is update values read by other tasks
//=============================================================================
// IMPORTANT NOTES and WARNINGS:
//
//   To avoid concurrency issues, other tasks need to use locks when reading
//   values set here.  But be careful to keep such locks short to avoid
//   delaying the timing of the main application task.  If necessary, split
//   such code in to sections, getting and releasing the lock(s) in each
//   section to avoid increasing latency for the main application task.
//
//   Values in the "next" instance of the application state fall in to 2
//   categories:  Inputs and Outputs.
//
//   The "next" Outputs are those that may be changed in updateOutputState().
//   The reset are Inputs, and may be changed by other tasks within a lock.
//
//   Since each pass through the application logic starts by setting
//   cur = next, any "next" values set by other tasks become part of the
//   "cur" instance used by the rest of the applic logic.  Because the
//   "applic" task at a higher priority than any other real-time task, the
//   changes to "next" values by other tasks can ONLY happen between passes
//   through the applic logic loop.
//
//   NOTES:
//     - Some of these "next" input values values must be initialized in
//       updateCurState() in order to tell when a value is set by another
//       task.  For example, if we have a boolean flag that is set when we
//       get an ON command, then (after its "next" value is copied to "cur"),
//       we have to clear the flag in order to tell if another ON command is
//       received between passes.
//
//     - As with reads to these values, other tasks must use locks when
//       writing to the "next" input values to insure they are not preempted
//       by the "applic" task in the middle of a write.
//
//=============================================================================
void updateExternals(void)
{
  uint  ui, bsState, onStatesBitMap = 0;
  BulkSupplyState  *bsCur;


  if (current != applTask)  return;  // must only be called by "applic" task

  digInBits = Cur.digitalIn;
  digOutBits = Cur.digitalOut;

//APPLCODE

  iset = Cur.iset;
  polCmd = Cur.polCmd;
  ampApply = Cur.ampApply;

  state = Cur.state;

  ampSet = Cur.ampSet;


  vctrlStatus  = (Cur.regPol ? 1 : 0);
  vctrlStatus |= (Cur.hystRegPol ? 2 : 0);
  vctrlStatus |= (Cur.regPolStable ? 4 : 0);
  vctrlStatus |= (inductiveLoad ? 8 : 0);

  psCount = Cur.psCount;

  ifdbk = Cur.ifdbk;
  vout = Cur.vout;
  vctrl = Cur.vctrl;
  bjtBias = Cur.bjtBias;

  outMode = Cur.outMode;


  for (ui=0; ui<NumPS;  ++ui)  {
    bsCur = Cur.bs + ui;

    *bsvX[ui] = bsCur->biasVolts;
    *ampsX[ui] = bsCur->current;

    bsState = 0;

    if (bsCur->isOn)      bsState |= PS_IS_ON;
    if (bsCur->isAvail)   bsState |= PS_AVAIL;
    if (bsCur->tempOK)    bsState |= PS_TEMP_OK;
    if (!bsCur->biasErr)  bsState |= PS_BIAS_OK;

    if (bsCur->setOn)     bsState |= PS_SET_ON;
    if (bsCur->enabled)   bsState |= PS_ENABLED;

    PS_STATE(ui) = bsState;

    if (bsCur->setOn AND bsCur->isOn)
      ClearBits(statusBitsAppl, PS_CTLR_RESET);
  }

  if (Cur.state == STATE_ON)  onStatesBitMap = 1;

  diagLogInfo->onStatesBitMap = onStatesBitMap;
}



//=============================================================================
#define  BiasInputs    0x02551550L
#define  CurInputs     0x01aa8aa0L
#define  SpareInputs   0x00006000L

//=============================================================================
_nearcall float readGetRaw(uint ui)  {
  return (ui < NumHdwADC ? anaIn[ui] : 0.0);  }

_nearcall float readMinRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMinRaw(ui) - adcOffset[ui]) * adcGain[ui] : 0.0; }

_nearcall float readMaxRaw(uint ui)  {
  return (ui < NumHdwADC) ? (adcMaxRaw(ui) - adcOffset[ui]) * adcGain[ui] : 9.8; }

//------------------------------------------------------------------
_nearcall void readDefaults(uint ui, float *gain, float *offset)
{
  *gain = 1.0;  *offset = 0.0;
  if (ui == AI_IFDBK)        *gain =  100.0;
  if (ui == AI_VOUT)         *gain =    1.0;
  if ((1L<<ui) & CurInputs)  *gain = 1000.0;
}
//------------------------------------------------------------------
_nearcall void readCalibPoints(uint ui, float *low, float *high)
{
  *low = -9.5;  *high = 9.5;
  if (ui == AI_IFDBK)        { *low = -500.0;  *high = 500.0;  }
  if ((1L<<ui) & CurInputs)  { *low =  -70.0;  *high =  70.0;  }
}
//--------------------------------------------------------
_nearcall char *readUnits(uint ui, char *buf)  {
   return (((1L<<ui) & CurInputs) OR (ui == AI_IFDBK)) ? "A" : "V"; }


//=============================================================================
_nearcall float setGetRaw(uint ui)  {
  return (ui < NumHdwDAC ? anaOut[ui] : 0.0);  }

_nearcall float setMinRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMinRaw(ui) - dacOffset[ui]) * dacGain[ui] : 0.0; }

_nearcall float setMaxRaw(uint ui)  {
  return (ui < NumHdwDAC) ? (dacMaxRaw(ui) - dacOffset[ui]) * dacGain[ui] : 9.8; }

//-----------------------------------------------------------------------------
_nearcall void setDefaults(uint ui, float *gain, float *offset)
{  *gain = (ui == AO_ISET) ? 100.0 : 1.5;  *offset = 0.0;  }

_nearcall void setCalibPoints(uint ui, float *low, float *high)
{
  if (ui == AO_ISET)  { *low = -500.0;  *high = 500.0;  return; }
  *low = -9.5;  *high = 9.5;
}
_nearcall char *setUnits(uint ui, char *buf)  {
  return (ui == AO_ISET) ? "A" : "V"; }


//=============================================================================
//  Set default labels for all the I/O signals
//=============================================================================
void setDefaultLabels(void)
{
  uint  ui, uidx;


  printf("Setting default labels for I/O signals\r\n");

//--- set application-appropriate default labels ---
//APPLCODE

//--- set default labels for the I/O signals ---
  fstrcpy(ainLabel[AI_SPARE0], "SPARE0");
  fstrcpy(ainLabel[AI_SPARE1], "SPARE1");

  for (ui=0; ui<NumPS; ++ui)  {
    sprintf(ainLabel[BSVidx[ui]], "bias%02u", ui+1);
    sprintf(ainLabel[HECTidx[ui]], "cur%02u", ui+1);

    uidx = psDigInIdx[ui];
    sprintf(dinLabel[uidx],   "IS_ON_%02u", ui+1);
    sprintf(dinLabel[uidx+1], "AVAIL_%02u", ui+1);
    sprintf(dinLabel[uidx+2], "TEMPOK_%02u", ui+1);

    sprintf(doutLabel[psDigOutIdx[ui]], "SET_ON_%02u", ui+1);
  }
  fstrcpy(dinLabel[30], "HOLEC_OK");
  fstrcpy(dinLabel[31], "NO_INTLK");

  fstrcpy(ainLabel[AI_IFDBK], "IFDBK");
  fstrcpy(ainLabel[AI_VOUT], "VOUT");
  fstrcpy(ainLabel[AI_VCTRL], "VCTRL");
  fstrcpy(ainLabel[AI_BJT_BIAS], "BJT_BIAS");

  fstrcpy(aoutLabel[AO_ISET], "ISET");
  fstrcpy(aoutLabel[AO_VLIM_POS], "VLIM_POS");
  fstrcpy(aoutLabel[AO_VLIM_NEG], "VLIM_NEG");

  fstrcpy(doutLabel[7],  "unused");
  fstrcpy(doutLabel[11], "Enable_Pol_Minus");
  fstrcpy(doutLabel[12], "Enable_Pol_Plus");
  fstrcpy(doutLabel[13], "Pol_Plus_LED");
  fstrcpy(doutLabel[14], "Pol_Minus_LED");
  fstrcpy(doutLabel[15], "ActivityLED");
}

//=============================================================================
//  This function is called to force software configurable values to default
//  values.
//=============================================================================
void setDefaultConfig(void)
{
  uint  u;
  BulkSupplyState  *bsNext;


  printf("Setting in-RAM device config to default values\r\n");

//===== set default values for persistent data values =====
//APPLCODE

  vlimPlus = vlimMinus = 7.0;

  rampRate = 10.0;  // in amps/sec

  knobSelectRate = 20;  maxAmpSet = 2000.0;

  minPSonTime = 1000;  minBridgeOffTime = 20;  // in ms

  vctrlHyst = 0.100;  // in volts

  maxErrTime = 1000;  // in ms

  minBSVolts = 5.0;  // min Bulk Supply bias voltage (in volts)

  maxBSFracCurDiff = 0.10;  // max Bulk Supply current error (frac of full scale)

  bulkShortCur = 100.0;  // bulk supply current that indicates a short

  supplyType = 0;  // someone must set the type correctly to enable operation!
  loadType = 0;    // someone must set the type correctly to enable operation!

//  sinkEndAmps = 0.5;  // ttlBulkCur threshold that JUST preceeds end of sink

  minPolStableTime = 500;  // min time regulation polarity must be stable before
                           // bridge polarity is changed to match (in ms)

  vlimReduction = 0.5;  // amount to reduce voltage limits by during SINK mode
                        // to minimize voltage spike at end of sink mode

  contrast = 32;  // LCD contrast, range 0-255

  enabledMask = 0xFFFF;  // all PS enabled unless overridden by stored config
  for (u=0; u<NumPS; ++u)  {
    bsNext = Next.bs + u;
    bsNext->enabled = true;
  }

  debounceMS = 250;  // interval (in ms) for debouncing digitalIn bits
}


//=============================================================================
//  Set the initial state we want the controller to start with.  This includes
//  "next" values that will be used to set the output signal states as soon
//  as the main application loop begins.
//
//  WARNING:  This is run AFTER the persistent values are restored so do
//            NOT set default values for persistent values here.  That should
//            be done in setDefaultConfig() instead!
//=============================================================================
void setInitialState(void)
{
  uint  ui;
  BulkSupplyState  *bsNext;


  if (current != applTask)  return;  // must only be called by "applic" task

  inductiveLoad = (loadType == LOAD_RL);

  lastRegPolChgTime = lastModeSetTime = GetMS();

//---------------------------------------------------------------
//  At this point the analog and digital input values have been
//  read so they can be used in determining what state to start
//  with for the outputs and any other values.
//--------------------------------------------------------------

//--- update current state, both for use here, and so when updateCurState() ---
//--- is called near the start of the application loop, there will be valid ---
//--- data to copy to the previous state ---
  updateCurState();

  Next.ampSet = Next.ampApply = 0.0;

  Next.state = STATE_INIT;

  for (ui=0; ui<NumPS; ++ui)  {
    bsNext = Next.bs + ui;
    bsNext->setOn = false;
  }

  psOnTime = 0;  psOnAndReady = loadTypeWrong = false;

  Cur.regPol = Cur.hystRegPol = true;

  bridgeOffTime = 0;
  lastLoadTypeWarnTime = lastValLogTime = GetSeconds_();

  logStateChgMsgs = true;

  dummyFloat = 0;

  bridgeStateStable = false;  bridgeChgdTime = GetMS_();

  logInterval = 3600;
}


//=============================================================================
//  This function is called by the application loop just before a shutdown.
//  The appliation must do whatever is necessary for a clean shutdown and
//  then exit
//=============================================================================
void applicShutdown(void)
{
  printf("Application prepping for shutdown\r\n");  fflush(stdout);


  Exit(0);
}


//=============================================================================
//  Use the values in prevLogInfo to determine if we should assert interlocks
//  for any devices any devices that were shut off due to a non-soft reset.
//=============================================================================
void checkPrevOnStates(void)
{
  if (prevLogInfo.onStatesBitMap)  offDueToReset = true;
}


//=============================================================================
//  This is the highest priority task, usually resumed at a regular interval
//  so that it can do what it needs and sleep again until the start of the
//  next interval.
//=============================================================================
void _task applic(int argc, char **argv)
{
  #define SharedApplLocals
  #include <sharedAppl.c>

//--- add local variable definitions here ---


  #define SharedApplLocals2
  #include <sharedAppl.c>


//--- Required to insure that we have read ALL the bulk supply signals once ---
  for (u=0; u<11; ++u)  readAnalogInputs(ADC_Mask);

  #define SharedApplInit
  #include <sharedAppl.c>

  BZERO(bitCounts);  // counts for debouncing the digital inputs

  _EnterCritical_();  SchedMS = 20;  _ExitCritical_();

//--- workaround for older devData files with order-of-reload dependencies ---
  if (vlimPlus == 0.0)  loadAllParams(applBuf, sizeof(applBuf));


  setInitialState();

//--- add application initialization code here ---
//APPLCODE

  numLoggedChgMsgs = 0;

  BZERO(bulkShortStarted);


  #define SharedApplInit2
  #include <sharedAppl.c>

//--- start process to manage the LCD display ---
  argv_[0] = "lcdUI";  argv_[1] = NULL;
  createProc(lcdUI, 4, 1, argv_, 0, 1, 2);


  if (!restoredApplState)  SetBits(statusBitsAppl, PS_CTLR_RESET);


//--- The following loop is run at a periodic interval (every SchedMS ms). ---
//--- The application process consists of reading the current state of the ---
//--- digital and analog inputs, changing the state of the analog and ---
//--- digital outputs to the values computed during the previous pass, then ---
//--- computing new values for the outputs and updating values that outside ---
//--- clients can read. ---

  Next.state = STATE_OFF;

  for (;;)  {
    #define SharedApplLoopStart
    #include <sharedAppl.c>


  //--- read the current state of the analog and digital Input signals, and ---
  //--- write the values computed at the end of the last pass ---
    digitalOut = Next.digitalOut;  rwIO();

  //--- process next event message (if any) ---
    processMsg(eventsfd, (EventMsg *)buf, sizeof(buf));

  //--- update the current state and determine the next state ---
    updateCurState();  updateNextState();
    updateOutputState();  updateExternals();


    #define SharedApplLoopEnd
    #include <sharedAppl.c>
  }

}


//=============================================================================
//  list of application-specific commands that are normally entered via a
//  user shell
//=============================================================================
//void _task nsclSeriald(int argc, char **argv);

//--- add any application-specific commands here ---
const CmdInfo far applCmds[] = {
  { "lcdui",     lcdUI,       "LCD user interface daemon" },
  { "cmd",       cmd,         "Turn On/Off, Reset, enable/disable, etc" },
  { "state",     showState,   "Show state of controller" }
};

int const NumApplCmds = ( sizeof(applCmds) / sizeof(CmdInfo) );

//-----------------------------------------------------------------------------

#endif
